
//#include <iostream.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#include "global.h"

#include "action_forms.h"
#include "action_forms_aux.h"
#include "browser.h"
#include "cb.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "forms_aux.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "status_forms_aux.h"
#include "tree_forms_aux.h"

//static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

static void action_cancelled(void){
    fl_set_object_label(ActionWorkFormText, "O P E R A T I O N   C A N C E L L E D !");
    fl_set_object_lcol (ActionWorkFormText, FL_RED);
    sginap(200);
}

static char
    source_current_fname_buf[MAXPATHLEN+1] = "",
    target_current_fname_buf[MAXPATHLEN+1] = ""
;

/* if true is returned, forms stays visible and can be used for progress
 * indication. It needs then later to be switched off, e.g. by calling the
 * CancelButtonCB. This also reactivates the rest.
 */
  
static int
    action_continue_flag     = TRUE,   // for a delayed cancel.
    local_continue_flag      = TRUE,   // for the OK or cancel decision.
    multi_work_flag          = FALSE,  // wether one or many files
    target_dir_override_flag = FALSE   // for override functions.
;

int ask_action_work_question(
    const char *question, 
    const char *source_dir, 
    int n_files,
    const char **help_text,
    const char *help_title,
    int default_go_flag,
    int no_multi_work_flag,
    int danger_flag			     
){
    int dummy;
    return ask_action_work_question(
        question,
	source_dir,
	"",
	n_files,
	help_text,
	help_title,
	default_go_flag,
	dummy,
        no_multi_work_flag,
        danger_flag
    );
}

static const char
    **local_help_text = NULL,
    *local_help_title = NULL
;

static void local_help_func(void){
    if(local_help_text){
	load_help_browser(local_help_text, local_help_title);
    }
}

static void local_close_func(void){
    ActionWorkFormCancelButtonCB(NULL, 0);
}

void action_work_noform(void){
    action_continue_flag = TRUE;
}

static int action_work_raw_cb(FL_FORM *form, void *xev){
    if(form != ActionWorkForm){
	cerr << "invalid form in raw cb.\n";
	return FALSE;
    }
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;

    int press_flag = FALSE;
    KeySym keysym;
    char keybuf[2];
    XEvent *xevent = (XEvent *) xev;
    XKeyEvent *xkev= (XKeyEvent *) xev;

    switch(xevent->type){
    case KeyPress:
	press_flag = TRUE;
    case KeyRelease:
	XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
	if(keysym == XK_F4){
	    // always gobble up F4 events here. dangerous: This is not reentrant. 
	    if(press_flag){
		ActionWorkFormF4TreeButtonCB(NULL, 0);
	    }
	    return TRUE;
	}
	break;
    default:
	break;
    }
    return standard_form_event_cb(form, (XEvent *) xev, local_close_func, local_help_func);
}

void ActionWorkFormNewTargetButtonCB(FL_OBJECT *, long){
    if (!source_browser || !target_browser) return;
    const char *dir_text = target_browser->get_cwd();
    if(strlen(tree_new_target_dir) > 0){
	// a better default, if we're already changing things.
	dir_text = tree_new_target_dir;
    }
    if(!dir_text) return;
    const char *new_dir = OKCancel("type new target directory, if desired:", dir_text);
    if(!new_dir) return;

    const char *cwd_buf = source_browser->get_cwd();
    if (!cwd_buf) return;

    char result[MAXPATHLEN];
    if (new_dir[0] == '/') {
	strcpy(result, new_dir);
    } else {
	sprintf(result, "%s/%s", cwd_buf, new_dir);
    }
    
    switch(validate_source_target_dirs(cwd_buf, result)){
    case error_e:
	status_msg(1, "Error in directory validation.");
	return;
    case identical_e:
	timed_error_notifier("source and target dir's are identical. Try again.");
	return;
    case ok_e:
    default:
	break;
    }
    if(target_browser != source_browser){
	if(!tool_change_target_dir(result, TRUE, TRUE)){
	    timed_strerror_notifier("Couldn't change directory:");
	    return;
	}   // else verification success.
	strcpy(tree_new_target_dir, target_browser->get_cwd());  // remember canonical form.
    } else {
	if(!validate_dir(result)){
	    timed_strerror_notifier("Target directory invalid:");
	    return;
	}
	if(access(result, W_OK | X_OK) != 0){
	    timed_strerror_notifier("write/search permission denied for target directory.");
	    return;
	}
	strcpy(tree_new_target_dir, result);  // remember typed result.
    }
    action_work_form_target_sync(tree_new_target_dir);  // was result.
}

void ActionWorkFormF4TreeButtonCB(FL_OBJECT *, long){
    if(!local_continue_flag || !ActionWorkFormF4TreeButton->visible){
	// don\'t want a return while proceeding !
	return;
    }
    //tree_unrestrict_cbs();
    //do_tree(TRUE);
    if(do_tree_new_target()){
	// strcpy(target_dir, tree_new_target_dir);
    }
}

static int action_sync_default_go_flag = TRUE;

static void action_work_form_resize(void){
    int min_w, max_w, string_w;
    if(fl_scrw >= 1023){
	min_w = 560;
    } else if(fl_scrw >= 799){
	min_w = 480;
    } else {
	min_w = 400;
    }
    max_w = min_w;

    // calculate new width
    string_w = 10 + fl_get_string_width(
	ActionWorkFormText->lstyle,
	ActionWorkFormText->lsize, 
	ActionWorkFormText->label,
	strlen(ActionWorkFormText->label)
    );
    if(string_w > max_w) max_w = string_w;
    string_w = 10 + fl_get_string_width(
	ActionWorkFormSourceDirText->lstyle,
	ActionWorkFormSourceDirText->lsize, 
	ActionWorkFormSourceDirText->label,
	strlen(ActionWorkFormSourceDirText->label)
    );
    if(string_w > max_w) max_w = string_w;
    string_w = 10 + fl_get_string_width(
	ActionWorkFormTargetDirText->lstyle,
	ActionWorkFormTargetDirText->lsize, 
	ActionWorkFormTargetDirText->label,
	strlen(ActionWorkFormTargetDirText->label)
    );
    if(string_w > max_w) max_w = string_w;

    // determine wether resize is necessary
    /*
    int do_resize_flag = FALSE;
    if(max_w != ActionWorkForm->w-10){
	do_resize_flag = TRUE;
    }
    */
    /*
    if(ActionWorkForm->visible){
	// make only bigger (select target)
	if(max_w > ActionWorkForm->w-10){
	    do_resize_flag = TRUE;
	}
    } else {
	// always (minsize built-in)
	do_resize_flag = TRUE;
    }
    */


    /* buttons == 100, slider == 300 (+50) OR VARIABLE, half stuff 50, filetext = 300) */
	
    fl_freeze_form(ActionWorkForm);

    int form_h= 0,
    	row_h = ActionWorkFormGoReturnButton->h,
	row_y
    ;

    ActionWorkFormText->y          = 5;
    ActionWorkFormSourceDirText->y = ActionWorkFormText->y + ActionWorkFormText->h;
    ActionWorkFormTargetDirText->y = ActionWorkFormSourceDirText->y + ActionWorkFormSourceDirText->h;
    
    if(ActionWorkFormTargetDirText->visible){
	row_y = ActionWorkFormTargetDirText->y + ActionWorkFormTargetDirText->h;
    } else if(ActionWorkFormSourceDirText->visible){
	row_y = ActionWorkFormSourceDirText->y + ActionWorkFormSourceDirText->h;
    } else {
	row_y = ActionWorkFormText->y + ActionWorkFormText->h;
    }
    form_h = row_y + row_h + 5;
    
    fl_set_form_size(ActionWorkForm, max_w+10, form_h);

    // renorm y & h.
    ActionWorkFormGoReturnButton->h  = row_h;
    ActionWorkFormCancelButton->h    = row_h;
    ActionWorkFormNewTargetButton->h = row_h;
    ActionWorkFormF4TreeButton->h    = row_h;
    ActionWorkFormRotatorText->h     = row_h;
    ActionWorkFormValSlider->h       = row_h;
    ActionWorkFormFileNameText->h    = row_h;
    ActionWorkFormProgressText->h    = row_h;
    ActionWorkFormText->h            = row_h;
    ActionWorkFormSourceDirText->h   = row_h;
    ActionWorkFormTargetDirText->h   = row_h;
    ActionWorkFormFillerText->h      = row_h;

    ActionWorkFormGoReturnButton->y  = row_y;
    ActionWorkFormCancelButton->y    = row_y;
    ActionWorkFormNewTargetButton->y = row_y;
    ActionWorkFormF4TreeButton->y    = row_y;
    ActionWorkFormRotatorText->y     = row_y;
    ActionWorkFormValSlider->y       = row_y;
    ActionWorkFormFileNameText->y    = row_y;
    ActionWorkFormProgressText->y    = row_y;
    ActionWorkFormFillerText->y      = row_y;

    // repeat
    ActionWorkFormText->y          = 5;
    ActionWorkFormSourceDirText->y = ActionWorkFormText->y + ActionWorkFormText->h;
    ActionWorkFormTargetDirText->y = ActionWorkFormSourceDirText->y + ActionWorkFormSourceDirText->h;

    ActionWorkFormText->x = 5;
    ActionWorkFormText->w = max_w;
    ActionWorkFormSourceDirText->x = 5;
    ActionWorkFormSourceDirText->w = max_w;
    ActionWorkFormTargetDirText->x = 5;
    ActionWorkFormTargetDirText->w = max_w;
    
    const int button_w = 100;
    ActionWorkFormGoReturnButton->w = button_w;
    ActionWorkFormGoReturnButton->x = 5;
    ActionWorkFormFillerText->x     = ActionWorkFormGoReturnButton->x + ActionWorkFormGoReturnButton->w;
    ActionWorkFormFillerText->w     = button_w/2 + 10;

    ActionWorkFormCancelButton->w   = button_w;
    ActionWorkFormCancelButton->x   = 5 + max_w - button_w;
    
    // use gap
    ActionWorkFormF4TreeButton->w   = button_w;
    ActionWorkFormF4TreeButton->x   = (int) floor(5.0 + button_w + (max_w - 4*button_w)/3.0);

    ActionWorkFormNewTargetButton->w= button_w;
    ActionWorkFormNewTargetButton->x= ActionWorkFormF4TreeButton->x + button_w + (FL_Coord) ((max_w - 4*button_w)/3.0);

    ActionWorkFormRotatorText->w = ActionWorkFormFillerText->w;
    ActionWorkFormRotatorText->x = ActionWorkFormCancelButton->x - ActionWorkFormRotatorText->w;
    
    ActionWorkFormValSlider->w = max_w - 2 * button_w - ActionWorkFormRotatorText->w;
    ActionWorkFormValSlider->x = 5 + button_w;
    
    ActionWorkFormFileNameText->w = ActionWorkFormValSlider->w - ActionWorkFormRotatorText->w;
    ActionWorkFormFileNameText->x = 5 + button_w + ActionWorkFormRotatorText->w;
    
    ActionWorkFormProgressText->w = ActionWorkFormValSlider->w  - ActionWorkFormRotatorText->w;
    ActionWorkFormProgressText->x = ActionWorkFormRotatorText->x - ActionWorkFormProgressText->w;
    
    fl_unfreeze_form(ActionWorkForm);
}

void action_work_form_target_sync(const char *new_target_dir){
    if(!new_target_dir) return;
    if(ActionWorkForm->visible){
	if(ActionWorkFormTargetDirText->visible){
	    char buffer[MAXPATHLEN+64];
	    sprintf(buffer, "target dir `%s' ?", new_target_dir);
	    fl_freeze_form(ActionWorkForm);
	    fl_set_object_label(ActionWorkFormTargetDirText, buffer);
	    if(action_sync_default_go_flag){
		set_to_ret_button(ActionWorkFormGoReturnButton);
	    } else {
		set_to_ret_button(ActionWorkFormCancelButton);
	    }
	    action_work_form_resize();
	    fl_unfreeze_form(ActionWorkForm);
	}
    }
}

static void aaw_init(int n_files, const char **help_text, const char *help_title){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_button_shortcut(ActionWorkFormNewTargetButton, "Tt", TRUE);
	fl_set_object_dblbuffer(ActionWorkFormValSlider, FALSE);
	make_standard_ok_shortcut(ActionWorkFormGoReturnButton);	
        make_standard_cancel_shortcut(ActionWorkFormCancelButton);
	fl_register_call_back(ActionWorkForm, FL_ALL_EVENT, action_work_raw_cb);
    }
    fl_deactivate_all_forms();
    fl_hide_object(ActionWorkFormRotatorText);
    fl_hide_object(ActionWorkFormFillerText);
    fl_hide_object(ActionWorkFormProgressText);
    fl_hide_object(ActionWorkFormValSlider);
    fl_set_slider_bounds   (ActionWorkFormValSlider, 0.00, (double) n_files);
    fl_set_slider_value    (ActionWorkFormValSlider, 0.00);
    fl_set_object_lcol     (ActionWorkFormText, FL_BLACK);
    fl_set_slider_precision(ActionWorkFormValSlider, 0);

    if(help_text){
	local_help_text = help_text;
	local_help_title = help_title;
    } else {
	local_help_text = NULL;
	local_help_title = NULL;
    }
}

static void aaw_default_go_sync(int default_go_flag){
    if(default_go_flag){
	set_to_ret_button(ActionWorkFormGoReturnButton);
    } else {
	set_to_ret_button(ActionWorkFormCancelButton);
    }
}

int ask_action_work_question(
    const char *question, 
    const char *source_dir, 
    const char *target_dir,
    int n_files,
    const char **help_text,
    const char *help_title,
    int default_go_flag,
    int &new_target_dir_flag,
    int no_multi_work_flag,
    int danger_flag
){
    if(!question || !source_dir || !target_dir) return FALSE;
    if(no_multi_work_flag){
	multi_work_flag = FALSE;
    } else {
	multi_work_flag = (n_files > 1);
    }

    new_target_dir_flag = FALSE;
    action_sync_default_go_flag = default_go_flag;
    //strcpy(tree_new_target_dir, "");
    strcpy(tree_new_target_dir, target_dir);
    
    aaw_init(n_files, help_text, help_title);
    
    char buffer[MAXPATHLEN+64];
    if(strlen(target_dir) <= 0){
	sprintf(buffer, "%s in", question);
	fl_set_object_label(ActionWorkFormText, buffer);
	sprintf(buffer, "source dir `%s' ?", source_dir);
	fl_set_object_label(ActionWorkFormSourceDirText, buffer);
	fl_set_object_label(ActionWorkFormTargetDirText, "");
	fl_hide_object(ActionWorkFormF4TreeButton);
	fl_hide_object(ActionWorkFormNewTargetButton);
	fl_hide_object(ActionWorkFormTargetDirText);
    } else {
	sprintf(buffer, "%s from", question);
	fl_set_object_label(ActionWorkFormText, buffer);
	sprintf(buffer, "source dir `%s' to", source_dir);
	fl_set_object_label(ActionWorkFormSourceDirText, buffer);
	sprintf(buffer, "target dir `%s' ?", target_dir);
	fl_set_object_label(ActionWorkFormTargetDirText, buffer);
	//fl_show_object(ActionWorkFormNewTargetButton);
	fl_show_object(ActionWorkFormF4TreeButton);
	fl_show_object(ActionWorkFormNewTargetButton);
	fl_show_object(ActionWorkFormTargetDirText);
    }
    fl_show_object(ActionWorkFormSourceDirText); // it could be hidden.
    fl_set_object_label(ActionWorkFormFileNameText, "");
    aaw_default_go_sync(default_go_flag);
    action_work_form_resize();
    if(!help_title) help_title = "UC Operation";
    focus_save();
    fl_show_form(ActionWorkForm, FL_PLACE_CENTER, TRUE, help_title);
    if(danger_flag){
	fl_set_cursor_color(XC_pirate, FL_BLACK, FL_RED);
	fl_set_global_cursor(XC_pirate);
    }    
    tool_focus_move(ActionWorkForm, find_return_button(ActionWorkForm));

    action_continue_flag = TRUE;
    local_continue_flag = TRUE;
    callback_table.restrict_all();
    while(local_continue_flag){
	tool_mung_callbacks();
    }
    callback_table.unrestrict_all();
    if(strlen(tree_new_target_dir) > 0){
	new_target_dir_flag = TRUE;
    }
    // if we're still visible, that means not cancel.
    // cancel has reactivated our forms if necessary, otherwise
    // they stay off as desired.
    if(ActionWorkForm->visible){
	set_to_ret_button(ActionWorkFormCancelButton);
        return TRUE;
    } else {
	return FALSE;
    }
	    
}

void ActionWorkFormCancelButtonCB(FL_OBJECT *, long){
    if(!local_continue_flag){
	// it's a real cancel during an operation. (doublechecked here).
	action_cancelled();
    }   // else it was immediate, so don't complain.
    action_end();
}

void ActionWorkFormGoReturnButtonCB(FL_OBJECT *, long){
    if(!local_continue_flag){
	// don\'t want a return while proceeding !
	return;
    }
    local_continue_flag = FALSE;
    fl_show_object(ActionWorkFormProgressText);
    fl_show_object(ActionWorkFormRotatorText);
    fl_hide_object(ActionWorkFormF4TreeButton);
    fl_hide_object(ActionWorkFormNewTargetButton);
    if(multi_work_flag){
	fl_show_object(ActionWorkFormValSlider);	
    } else {
	fl_show_object(ActionWorkFormFillerText);
    }
    fl_set_cursor_color(XC_watch, FL_RED, FL_WHITE);
    fl_set_global_cursor(XC_watch);
}

void action_end(void){
    local_continue_flag = FALSE;
    action_continue_flag = FALSE;
    if(ActionWorkForm->visible){
	focus_restore_start();
	//cerr << "hide action\n";
	fl_hide_form(ActionWorkForm);
	focus_restore_finish();
	fl_reset_global_cursor();
	fl_activate_all_forms();
    }
}

void action_msg(char *buffer){
    fl_set_object_label(ActionWorkFormFileNameText, buffer);
}

static void next_state(char state_label[2], int &state){
    state_label[1] = '\0';
    switch(state){
    case 0: state_label[0] = '|'; break;
    case 1: state_label[0] = '/'; break;
    case 2: state_label[0] = '-'; break;
    case 3: state_label[0] = '\\'; break;
    }
    state++;
    if(state>3) state = 0;
}

int action_mini_step(int check_forms_flag){
    if(!action_continue_flag) return FALSE;   // must be first !!
    if(!ActionWorkForm->visible) return TRUE;
    // don't care about multi/mono
    static int state = 0;
    static char state_label[2];

    // X server desync buffering can make it useless/invisible.
    // this is a speed question...
    next_state(state_label, state);
    fl_set_object_label(ActionWorkFormRotatorText, state_label);
    
    if(check_forms_flag){
	if(!tool_mung_forms()){
	    action_cancelled();
	    action_end();
	    return FALSE;
	}
    }
    return TRUE;
}

int action_step(int check_forms_flag){
    if(!action_continue_flag) return FALSE;   // must be first !!
    if(!ActionWorkForm->visible) return TRUE;
    if(multi_work_flag){
	double val = fl_get_slider_value(ActionWorkFormValSlider);
	fl_set_slider_value(ActionWorkFormValSlider, val + 1.0);
	if(check_forms_flag){
	    if(!tool_mung_forms()){
		action_end();
		return FALSE;
	    } else {
		return TRUE;
	    }
	} else {
	    return TRUE;
	}
    } else {
	return action_mini_step(check_forms_flag);
    }
}

static const char *action_make_byte_string(void){    
    static char byte_string[64];
    long bytes = source_browser->get_selection_byte_size(TRUE);
    sprintf(byte_string, " (%ld byte", bytes);
    if(bytes != 1){
        strcat(byte_string, "s");
    }
    strcat(byte_string, ")");
    return byte_string;
}

int source_dir_action(
    source_dir_action_single_fpt source_dir_action_single,
    source_dir_action_many_fpt   source_dir_action_many,
    const char                  *action_string,
    int                          action_do_ask_one_file_flag,
    int                          action_default_go_flag,
    const char                   **help_text,
    int                          action_no_multi_work_flag,
    int                          action_no_file_string_flag,
    int                          danger_flag
){
    if(!source_is_stable()) return FALSE;
    if(!action_string){
	program_unstable("UC internal: action: no action string.");
	return FALSE;	
    }

    int do_action_flag;
    char fname_buf[MAXPATHLEN+1];
    const char
	*fname = NULL,
	*source_dir = source_browser->get_cwd()
    ;
    char  buffer[1024];

    if(!validate_source_dir(source_dir)){
	return FALSE;
    }

    int n_selected = source_browser->get_n_selected();
    int n_changes = 0; // record how much happened.
    const char *byte_string = action_make_byte_string();
    reset_source_current();

    /* both the case of no selection (= current) and only one selection
     * warrant a bit of special attention - we don't want option overkill.
     * always freeze/unfreeze completely for an operation like this. 
     */
    
    if(n_selected <= 1){
        /* current file only. always ask for overwrite, directory, special. */
        if(n_selected <= 0){
	    fname = source_browser->get_current_fname();
	    if(!fname){
                //program_unstable("Couldn't extract current filename from source browser.");
		return FALSE;
            } else {
                strcpy(fname_buf, fname);   
            }
	    sprintf(buffer, "%s current file `%s'", action_string, fname);
	    //cerr << "Debug: " << buffer << endl;
	} else {
	    fname = source_browser->get_first_selection();
	    if(!fname){
		program_unstable("Couldn't extract first selection filename from source browser.");
		return FALSE;
            } else {
                strcpy(fname_buf, fname);   
            }
	    sprintf(buffer, "%s selected file `%s'", action_string, fname);
	}
	if(source_browser->show_size_flag){
	    strcat(buffer, byte_string);
	}
	if(action_do_ask_one_file_flag){
            if(action_no_file_string_flag){
		strcpy(buffer, action_string);
	    }
	    do_action_flag = ask_action_work_question(buffer, source_dir,1,help_text,action_string,action_default_go_flag, FALSE, danger_flag);
	} else {
	    do_action_flag = TRUE;
	}
        source_browser->freeze();
	if(do_action_flag){
	    if((*source_dir_action_single)(fname_buf, source_dir)){
		//cerr << action_string << " of current/single selected file succeeded." << endl;
		n_changes = 1;
	    } else {
		//cerr << action_string << " failed." << endl;
	    }
	}
    } else {
	sprintf(buffer, "%s %d selected files", action_string, n_selected);
	if(source_browser->show_size_flag){
	    strcat(buffer, byte_string);
	}    
        source_browser->freeze();
        if(action_no_file_string_flag){
	    strcpy(buffer, action_string);
	}
	if(ask_action_work_question(
	    buffer,
	    source_dir,
	    n_selected,
	    help_text,
	    action_string,
	    action_default_go_flag,
	    action_no_multi_work_flag,
	    danger_flag
	)){
	    sprintf(buffer, "%s: working on %d files ...", action_string, n_selected);
	    n_changes = (*source_dir_action_many)(source_dir);
	    //cerr << n_changes << "files were removed." << endl;
	}
    }
    if(n_changes > 0 || toolop_cancelled_flag){
	//cerr << n_changes << " > 0, updating ..." << endl;
	// always ok to update the source browser, if n_changes > 0. Either
	// something failed (no stat, etc), or we did something. Both are ok.
	tool_update_matching_dirs(source_browser->get_cwd(), TRUE);

        // updating the target browser is not necessary, we do mono ops here.
	// old note: if the first
	// lstat on a selection list failed, but hey, that's very rare and
	// possibly a good precaution anyway.
	restore_source_current();
    }
    source_browser->unfreeze();
    action_end();
    return TRUE;
}

int source_and_target_dir_action(
    source_and_target_dir_action_single_fpt source_and_target_dir_action_single,
    source_and_target_dir_action_many_fpt   source_and_target_dir_action_many,
    const char                   *action_string,
    int                          action_do_ask_one_file_flag,
    const char                   **help_text		      
){
    // meta button func, equal dirs func ?
    if(!source_target_are_stable()) return FALSE;
    if(!action_string){
	program_unstable("UC internal: action: no action string.");
	return FALSE;	
    }

    int do_action_flag, target_override_flag;
    char fname_buf[MAXPATHLEN+1];
    const char
        *fname,
	*source_dir = source_browser->get_cwd()
    ;
    char 
        buffer[1024],
	target_dir[MAXPATHLEN+1]
    ;
    strcpy(target_dir,target_browser->get_cwd());
    target_override_flag = target_dir_get_override(target_dir);
    
    if(!fully_validate_source_target_dirs(source_dir, target_dir, action_string)) return FALSE;

    int n_selected = source_browser->get_n_selected();
    int n_changes = 0; // record how much happened.
    const char *byte_string = action_make_byte_string();
    
    reset_source_current();
    reset_target_current();

    if(n_selected <= 1){
	// current file only. always ask for overwrite, directory, special.
	if(n_selected <= 0){
	    fname = source_browser->get_current_fname();
	    if(!fname){
		//program_unstable("Couldn't extract current filename from source browser.");
		return FALSE;
            } else {
                strcpy(fname_buf, fname);   
            }
	    sprintf(buffer, "%s current file `%s'", action_string, fname);
	    //cerr << "Debug: " << buffer << endl;
	} else {
	    fname = source_browser->get_first_selection();
	    if(!fname){
		program_unstable("Couldn't extract first selection filename from source browser.");
		return FALSE;
            } else {
                strcpy(fname_buf, fname);   
            }
	    sprintf(buffer, "%s selected file `%s'", action_string, fname);
	}
	if(source_browser->show_size_flag){
	    strcat(buffer, byte_string);
	}
	if(action_do_ask_one_file_flag){
	    int new_target_dir_flag = FALSE;
	    do_action_flag = ask_action_work_question(
		buffer, source_dir, target_dir, 1, help_text, action_string, TRUE, new_target_dir_flag, FALSE, FALSE
	    );
	    if(new_target_dir_flag) strcpy(target_dir, tree_new_target_dir);
	} else {
	    do_action_flag = TRUE;
	}
        source_browser->freeze();
        target_browser->freeze();
	if(do_action_flag){
	    if((*source_and_target_dir_action_single)(fname_buf, source_dir, target_dir)){
		//cerr << action_string << " of current/single selected file succeeded." << endl;
		n_changes = 1;
	    } else {
		//cerr << action_string << " failed." << endl;
	    }
	}
    } else {
	sprintf(buffer, "%s %d selected files", action_string, n_selected);
	if(source_browser->show_size_flag){
	    strcat(buffer, byte_string);
	}
        source_browser->freeze();
        target_browser->freeze();
	int new_target_dir_flag = FALSE;
	if(ask_action_work_question(
	    buffer, source_dir, target_dir, n_selected, help_text, action_string, TRUE, new_target_dir_flag, FALSE, FALSE
	)){
	    if(new_target_dir_flag) strcpy(target_dir, tree_new_target_dir);
	    n_changes = (*source_and_target_dir_action_many)(source_dir, target_dir);
	}
    }
    if(n_changes > 0 || toolop_cancelled_flag){
	//cerr << n_changes << " > 0, updating ..." << endl;
	// always ok to update the source browser, if n_changes > 0. Either
	// something failed (no stat, etc), or we did something. Both are ok.
	tool_update_matching_dirs(source_browser->get_cwd(), TRUE);
	restore_source_current();

	if (!target_override_flag) {
	    tool_update_matching_dirs(target_browser->get_cwd(), TRUE);
	    restore_target_current();
	}
    }
    source_browser->unfreeze();
    target_browser->unfreeze();
    action_end();
    return TRUE;
}

void remember_source_current(const char *fname){
    if(fname){
    	strcpy(source_current_fname_buf, fname);
	//cerr << "source remembering: " << fname << endl;
    }
}

void remember_target_current(const char *fname){
    if(fname){
    	strcpy(target_current_fname_buf, fname);
	//cerr << "target remembering: " << fname << endl;
    }
}

void restore_source_current(void){
    if(strlen(source_current_fname_buf) > 0){
	source_browser->set_current_fname(source_current_fname_buf, TRUE);
	//cerr << "source restoring: " << source_current_fname_buf << endl;
    }
}

void restore_target_current(void){
    if(strlen(target_current_fname_buf) > 0){
	target_browser->set_current_fname(target_current_fname_buf, TRUE);
	//cerr << "target restoring: " << target_current_fname_buf << endl;
    }
}

void reset_source_current(void){
    remember_source_current("");
}

void reset_target_current(void){
    remember_target_current("");
}

int fully_validate_source_target_dirs(
    const char *source_dir,
    char *target_dir,
    const char *action_string
){
    char buffer[1024];
    const char *dir_text = NULL, *result = NULL;
    if(!action_string) action_string = "?";
    do {
	switch(validate_source_target_dirs(source_dir, target_dir)){
	case error_e:
	    status_msg(1, "Error in directory validation.");
	    return FALSE;
	case identical_e:
	    status_msg(1, "Source & target dirs identical,");
	    status_msg(1, "Please select a new target.");
	    sprintf(buffer, "%s: Source and target dirs identical. Select a new target ?", action_string);
	    switch(ask_multi_question(FALSE, 1, buffer, "F4 Tree", "", XK_F4, "New Target", "Tt", NoSymbol, "Cancel", "Cc")){
	    case -1:
		// esc/cancel
		return FALSE;
	    case 0:
		// tree
		if(do_tree_new_target()){
		    strcpy(target_dir, tree_new_target_dir);
		} else {
		    status_msg(1, "still identical, try again.");
		}
		break;
	    case 1:
		dir_text = target_browser->get_cwd();
		if(strlen(tree_new_target_dir) > 0){
		    // a better default, if we're already changing things.
		    dir_text = tree_new_target_dir;
		}
		if(!dir_text) return FALSE;
		result = OKCancel("Type new target directory, if desired:", dir_text);
		if(result){
		    if (result[0] == '/') {
			strcpy(target_dir, result);  
		    } else {
			sprintf(target_dir, "%s/%s", source_browser->get_cwd(), result); 
		    }
		    // remember!
		    strcpy(tree_new_target_dir, target_dir);
		}
		// always continue checking loop here. 
		break;
	    case 2:
		// real cancel, fall through.
	    default:
		return FALSE;
	    }
	    break;
	case ok_e:
	    return TRUE;
	default:
	    return TRUE;
	}
    } while (1);
}

void action_sync_defaults(void){
    fl_freeze_form(ActionWorkForm);
    fl_set_object_lsize(ActionWorkFormText, tool_fontsize);
    fl_set_object_lsize(ActionWorkFormSourceDirText, tool_fontsize);
    fl_set_object_lsize(ActionWorkFormTargetDirText, tool_fontsize);
    
    fl_set_object_lsize(ActionWorkFormGoReturnButton,  tool_fontsize);
    fl_set_object_lsize(ActionWorkFormCancelButton,    tool_fontsize);
    fl_set_object_lsize(ActionWorkFormNewTargetButton, tool_fontsize);
    fl_set_object_lsize(ActionWorkFormF4TreeButton,    tool_fontsize);
    fl_set_object_lsize(ActionWorkFormRotatorText,     tool_fontsize);
    fl_set_object_lsize(ActionWorkFormValSlider,       tool_fontsize);
    fl_set_object_lsize(ActionWorkFormFileNameText,    tool_fontsize);
    fl_set_object_lsize(ActionWorkFormProgressText,    tool_fontsize);
    fl_set_object_lsize(ActionWorkFormText,            tool_fontsize);

    fl_unfreeze_form(ActionWorkForm);
}

int fork_action(
    action_fpt action_fp, 
    const char *question,
    const char **help_text,
    const char *help_title,
    int   danger_flag
){
    if(!question) return FALSE;
    aaw_init(1, help_text, help_title);
    aaw_default_go_sync(!danger_flag);

    fl_set_object_label(ActionWorkFormText, question);    
    fl_set_object_label(ActionWorkFormFileNameText,  "");
    fl_set_object_label(ActionWorkFormSourceDirText, "");
    fl_set_object_label(ActionWorkFormTargetDirText, "");

    fl_hide_object(ActionWorkFormF4TreeButton);
    fl_hide_object(ActionWorkFormNewTargetButton);
    fl_hide_object(ActionWorkFormTargetDirText);
    fl_hide_object(ActionWorkFormSourceDirText);
    
    action_work_form_resize();

    if(!help_title) help_title = "UC Operation";
    focus_save();
    fl_show_form(ActionWorkForm, FL_PLACE_CENTER, TRUE, help_title);
    if(danger_flag){
	fl_set_cursor_color(XC_pirate, FL_BLACK, FL_RED);
	fl_set_global_cursor(XC_pirate);
    }    
    tool_focus_move(ActionWorkForm, find_return_button(ActionWorkForm));
   
    action_continue_flag = TRUE;
    local_continue_flag = TRUE;
    callback_table.restrict_all();
    while(local_continue_flag){
	tool_mung_callbacks();
    }
    callback_table.unrestrict_all();
    // if we're still visible, that means not cancel.
    // cancel has reactivated our forms if necessary, otherwise
    // they stay off as desired.
    if(ActionWorkForm->visible){
	set_to_ret_button(ActionWorkFormCancelButton);
	action_fp();
        // action_step(FALSE);
	action_end();
	return TRUE;
    } else {
	return FALSE;
    }
}

int target_dir_get_override(char target_dir[]){
    if (!target_browser) return FALSE;
    if (target_dir_override_flag) {
	target_dir_override_flag = FALSE;
	if (strcmp(tree_new_target_dir, target_browser->get_cwd()) == 0) {
	    // special: not a true override. Do target_browser updates.
	    return FALSE;
	}
	strcpy(target_dir, tree_new_target_dir);
	return TRUE;
    } else {
	strcpy(tree_new_target_dir, "");  // important reset!
	return FALSE;
    }
}

void target_dir_set_override(const char *override_dir){
    if (!override_dir) {
	cerr << "target_dir_set_override: bad pointer.\n";
	return;
    }
    // use existing tree buf.
    target_dir_override_flag = TRUE;
    strcpy(tree_new_target_dir, override_dir);
}
