
#include <ctype.h>
#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>      // getenv
#include <string.h>
#include <sys/param.h>   // for MAXPATHLEN
#include <sys/stat.h>
#include <sys/time.h>   // gettimeofday
#include <sys/types.h>
#include <X11/cursorfont.h>

#ifdef LINUX_GLIBC2
#       include <errno.h>
#endif

#include "action_forms_aux.h"
#include "bitmaps/browser.bm"
#include "browser.h"
#include "browser_menu.h"
#include "cb_key.h"
#include "cb_fkey.h"
#include "dir_menu.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "ghelp_forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "linklist.h"
#include "client_msg.h"
#include "fork_pipe.h"
#include "options_forms_aux.h"
#include "panel.h"
#include "sort_forms_aux.h"
#include "system.h"
#include "status_forms_aux.h"
#include "superls.h"
#include "toolbox_forms_aux.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "xshell.h"
#include "wildcard.h"
#include "wm_forms_aux.h"

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

static const char *line_icon_indent = "  ";

/* important shape constraining function. Essential for free scaling.
 * two important constants, the height of the strip, and the box width.
 * we fudge it a little bit: the target_box width is box_width + 1;
 * the menu is made 2 pixels longer than the boxes, so that it is nicely
 * centered. 
 */

static const int
    box_width    = 12,
    text_h_fudge = 2,                       // translate the darn thing up!
    text_x_base_fudge = 1
;

static int
//    text_x_fudge = FL_PANEL_BW + FL_PANEL_TEXTMARGIN,   // bound, frame, var 
    text_x_fudge = text_x_base_fudge,
    box_height = 25
;

static void
    column_hidden_button_cb(FL_OBJECT *column_hidden_button, long), 
    menu_hidden_button_cb  (FL_OBJECT *menu_hidden_button, long)
;
/*
 * a little auxiliary function for construction
 * of a pseudo command line.
 */

static void nextarg(int &argc, char *argv[], const char *arg){
    argv[argc] = strdup(arg);
    argc++;
}

// dnd variables
static browser_c *dnd_default_target_browser = NULL;
static int dnd_meta_flag = FALSE, dnd_copy_flag = FALSE;

/* some static non-member functions */


// if null, reset wintitles of all except source browser
// else, all except both source and this browser.

static void dnd_reset_wintitle(browser_c *browser){
    browser_c *tmp_browser = source_browser->next_browser;
    while (tmp_browser != source_browser) {
	if (tmp_browser != browser) {
	    tmp_browser->set_complete_wintitle();
	}
	tmp_browser = tmp_browser->next_browser;
    }
}
    
static void dnd_hide_other_current(browser_c* browser){
    if (!browser) {
	// NULL means hide all.  
	browser = source_browser;
	fl_hide_panel_current(browser->panel);
    } else {
	fl_unhide_panel_current(browser->panel);
    }
    browser_c *tmp_browser = browser->next_browser;
    while (tmp_browser != browser) {
	fl_hide_panel_current(tmp_browser->panel);
	tmp_browser = tmp_browser->next_browser;
    }
}

static void dnd_cleanup_func(FL_OBJECT *panel){
    if (!panel) {
	cerr << "dnd_cleanup: bad pointer.\n";
	return;
    }
    dnd_hide_other_current(source_browser);
    // wintitles are set below.

    browser_c *tmp_browser = source_browser;
    do {
	tmp_browser->set_complete_wintitle();
	fl_set_panel_cursor(tmp_browser->panel, fl_default_curname);
	tmp_browser = tmp_browser->next_browser;
    } while (tmp_browser != source_browser);

    tree_sync_to_target_browser();
    tree_set_complete_wintitle();
    //tree_form_sync();
}

static void dnd_exec_func(FL_OBJECT *panel){
    if (!panel) {
	cerr << "dnd_exec: bad pointer.\n";
	return;
    }
    browser_c *browser = tool_find_browser(panel->form);
    if (!browser) {
	cerr << "dnd_exec: no browser.\n";
	return;
    }
    //cerr << "exec." << outside_flag << "\n";
    if (!fl_get_panel_dnd_flag(panel)) {
	cerr << "dnd_exec called with dnd off.\n";
	return;
    }
    char dir_buf[1024];
    const char *tmp_buf = NULL;
    strcpy(dir_buf, "");

    FL_Coord mx, my;
    unsigned keymask;
    FL_FORM *mouse_form = fl_find_pointer_form(&mx, &my, &keymask);
    mx -= wm_forms_fudge_dx;
    my -= wm_forms_fudge_dy;

    if (!mouse_form) return;
    browser_c *slave_browser = tool_find_browser(mouse_form);
    
    if (mouse_form == TreeForm) {
	tmp_buf = tree_xdnd_track(mx, my);
	if (!tmp_buf) return;
	strcpy(dir_buf, tmp_buf);
    } else {
	if (!slave_browser) return;
	if (slave_browser == browser) {
	    if (!fl_get_panel_dnd_track_valid(panel)) {
		// cerr << "exec: not valid.\n";
		return; // was cancelled.
	    }
	    int current_line;
	    tmp_buf = browser->get_current_fname(current_line);
	    if (!tmp_buf) {
		cerr << "dnd_exec: no target .\n";
		return;
	    } else if (strcmp(tmp_buf, ".") == 0) {
		status_msg(2, "Can't drag & drop into the same directory (`.')."); 
		return;
	    }
	    if (fl_isselected_panel_line(panel, current_line)) {
		status_msg(2, "Cannot drag & drop `%s' onto itself.", tmp_buf); 
		return;
	    }
	    const char *cwd_buf = browser->get_cwd();
	    if (!cwd_buf) return;
	    sprintf(dir_buf, "%s/%s", cwd_buf, tmp_buf);
	} else {
	    //cerr << "other browser.\n";
	    tmp_buf = slave_browser->xdnd_track(mx, my);
	    if (!tmp_buf) return;
	    strcpy(dir_buf, tmp_buf);
	}
    }

    // dnd target verification
    switch(validate_source_target_dirs(source_browser->get_cwd(), dir_buf)){
    case error_e:
	status_msg(1, "Error in directory validation.");
	return;
    case identical_e:
	timed_error_notifier("source and target dir's are identical. Try again.");
	return;
    case ok_e:
    default:
	break;
    }
    // dnd target sync
    
    // useful for n_brwsers > 2: (this is mostly done by show already)
    if (slave_browser && slave_browser != target_browser &&
	slave_browser != source_browser
    ) {
	tool_set_target(slave_browser);
    }
    /*
    // search for a non-source target browser.
    if (target_browser == source_browser) {
	if (dnd_default_target_browser != target_browser) {
	    tool_set_target(dnd_default_target_browser);
	} else {
	    // surrender, choose next, whichever that may be:
	    tool_set_target(source_browser->next_browser);
	}
    }
    */
    // sync it's directory.
    if(target_browser != source_browser){
	if(!tool_change_target_dir(dir_buf, TRUE, TRUE)){
	    timed_strerror_notifier("Couldn't change to target directory:");
	    return;
	}
    } else {
	target_dir_set_override(dir_buf);
    }
    if (dnd_meta_flag) {
	tool_set_ctrl();
    }
    if (dnd_copy_flag) {
	CopyButtonCB(NULL, 0);
    } else {
	MoveButtonCB(NULL, 0);
    }
}

static int dnd_source_verify_func(FL_OBJECT *panel, int line, XEvent *xev) {
    if (!panel) {
	cerr << "dnd_source_verify: bad pointer.\n";
	return FALSE;
    }
    browser_c *browser = tool_find_browser(panel->form);
    if (!browser || !browser->linecheck(line)) {
	cerr << "dnd_source_verify: no browser or bad line.\n";
	return FALSE;
    }
    const char *fname = browser->extract_filename(line);
    if (!fname) {
	cerr << "dnd_source_verify: no fname.\n";
	return FALSE;
    }
    if (dot_dotdot_check(fname)) {
	dnd_default_target_browser = target_browser;
	dnd_meta_flag = tool_get_meta(xev);
	int button = (*xev).xbutton.button;
	dnd_copy_flag =
	    (button == 2 || button == 3) ||
	    (fl_keymask & Button2Mask) ||
	    (fl_keymask & Button3Mask) ||
	    (fl_keymask & ShiftMask)
	;
	if (dnd_copy_flag) {
	    if (dnd_meta_flag) {
		tool_dnd_op_str = "copy-rename";
	    } else {
		tool_dnd_op_str = "copy";
	    }
	} else {
	    if (dnd_meta_flag) {
		tool_dnd_op_str = "move-rename";
	    } else {
		tool_dnd_op_str = "move";
	    }
	}
	/*
	FL_Coord x, y;
	fl_get_mouse(&x, &y, &fl_keymask);
	dnd_meta_flag = tool_get_meta();
	*/
	//cerr << "meta, copy: " << dnd_meta_flag << ", " << dnd_copy_flag << endl;
	return TRUE;
    } else {
	//cerr << "FALSE for " << fname << endl;
	return FALSE;
    }
}

static void dnd_track_make_selbuf(char *sel_buf, char *track_buf){
    int n_selected = source_browser->get_n_selected();
    if (n_selected <= 0) {
	M_err("track show", "no selection.");
	strcpy(sel_buf, "<no selection>");
	return;
    } else if (n_selected > 1) {
	sprintf(sel_buf, "%d files %s", n_selected, track_buf);
    } else {
	const char *tmp_buf = source_browser->check_first_selection();
	if (!tmp_buf) tmp_buf = "<unknown>";
	sprintf(sel_buf, "`%s'%s", tmp_buf, track_buf);
    }
}

static int dnd_track_external_func(FL_OBJECT *panel) {
    char full_buf[MAXNAMELEN], sel_buf[MAXPATHLEN];
    //cerr << "external.\n";
    if (!panel) {
	M_err("dnd_track_external", "bad pointer.");
	return FALSE;
    }
    browser_c *browser = tool_find_browser(panel->form);
    if (!browser) { 
	M_err("dnd_track_external", "no browser or bad panel.");
	return FALSE;
    }
    if (browser != source_browser) {
	M_err("dnd_track_external", "not source browser");
	return FALSE;
    }

    dnd_track_make_selbuf(sel_buf, "");
    strcpy(full_buf, "");

    FL_Coord mx, my;
    unsigned keymask;
    FL_FORM *mouse_form = fl_find_pointer_form(&mx, &my, &keymask);
    mx -= wm_forms_fudge_dx;
    my -= wm_forms_fudge_dy;

    if (mouse_form && mouse_form != panel->form) {
	browser_c *slave_browser = tool_find_browser(mouse_form);
	if (slave_browser) {
	    const char *slave_target = slave_browser->xdnd_track(mx, my);
	    if (!slave_target) return FALSE;
	    tool_set_target(slave_browser);
	    
	    sprintf(full_buf, "#%d D&D %s from #%d to `%s'",
		slave_browser->get_ID(), tool_dnd_op_str,
		browser->get_ID(),
		slave_target
	    );
	    
	    slave_browser->set_complete_wintitle(full_buf);

	    sprintf(full_buf, "#%d D&D %s %s to #%d `%s'",
		browser->get_ID(), tool_dnd_op_str, sel_buf, 
		slave_browser->get_ID(),
		slave_target
	    );
	} else if (mouse_form == TreeForm) {
	    const char *tree_dir = tree_xdnd_track(mx, my);
	    if (tree_dir) {
		sprintf(full_buf, "#%d D&D %s %s to UCTree `%s'",
		    browser->get_ID(), tool_dnd_op_str, sel_buf, 
		    tree_dir
		);
	    } else {
		sprintf(full_buf, "#%d drag & drop %s %s to UCTree...",
		    browser->get_ID(), tool_dnd_op_str, sel_buf
		);
	    }
	}
	if (strlen(full_buf)) {
	    dnd_hide_other_current(slave_browser);  // if NULL, all hidden, ok.
	    dnd_reset_wintitle(slave_browser);  // if NULL, it's tree, ok.
	    if (mouse_form != TreeForm) {
		tree_set_complete_wintitle();
	    }
	    
	    // we only work on the source_browser's cursor - it's
	    // a server grab!
	    fl_set_panel_cursor(source_browser->panel, XC_fleur);
	    browser->set_complete_wintitle(full_buf);
	    return TRUE;
	}
	    
    }
    return FALSE;
}

static void dnd_track_show_func(FL_OBJECT *panel, int) {
    char full_buf[MAXPATHLEN + 64];
    if (!panel || !panel->form || !fl_get_panel_dnd_flag(panel)){
	cerr << "dnd_track_show: bad state or pointer.\n";
	return;
    }
    browser_c *browser = tool_find_browser(panel->form);
    if (!browser) {
	cerr << "dnd_track_show: no browser or bad panel.\n";
	return;
    }
    if (browser != source_browser) return;
    // undo possible external effects.
    dnd_hide_other_current(source_browser);
    dnd_reset_wintitle(source_browser);
    tree_set_complete_wintitle();
    
    // DND title
    sprintf(full_buf, "#%d drag & drop %s ", browser->get_ID(), tool_dnd_op_str);
    char
	sel_buf      [MAXPATHLEN],
	track_tmp_buf[MAXNAMELEN],
	track_buf    [MAXPATHLEN]
    ;
    strcpy(track_tmp_buf, "");
    if (fl_get_panel_dnd_track_valid(panel)) {
	if (!fl_isselected_panel_line(panel, fl_get_panel_current(panel))){
	    const char *tmp_buf = browser->get_current_fname();
	    if (tmp_buf){
		if (strcmp(tmp_buf, ".") != 0) {
		    strcpy(track_tmp_buf, tmp_buf);
		}   // else dot-exception. 
	    } else {
		strcpy(track_tmp_buf, "<unknown>");
	    }
	}
    }
    if (strlen(track_tmp_buf)) {
	tool_set_target(source_browser);
	sprintf(track_buf, " to `%s'", track_tmp_buf);
    } else {
	tool_set_target(dnd_default_target_browser);
	strcpy(track_buf, " ...");
    }
    dnd_track_make_selbuf(sel_buf, track_buf);
    strcat(full_buf, sel_buf);

    browser->set_complete_wintitle(full_buf);
}

/* browser_c data members */
int browser_c::auto_raise_flag = TRUE;
int browser_c::cwd_bufsize = MAXPATHLEN+2;
int browser_c::next_ID = 1;
int browser_c::directory_update_count = 40;

/* browser_c function members */
void browser_c::add_date_string(char *buf){
    strcat(buf, "\"%h %d ");
    if(show_time_years_flag) strcat(buf, "19%y "); 

    if(show_time_seconds_flag){
	strcat(buf, "%T\"");
    } else {
	strcat(buf, "%H:%M\"");
    }
}

int browser_c::selection_cmd(const char *path, const char *arg0, const char *arg1, const char *arg2){
    if(!path || !arg0){
	cerr << "invalid NULL path or arg for selection cmd!\n";
	return FALSE;
    }
    
    int n_selected = get_n_selected();  
    int n_args = 2; // name, NULL, and files.
    const char *fname = NULL;
    if(n_selected > 0){
	n_args += n_selected;
    } else {
	n_args++;
	fname = get_current_fname();
	if(!fname){
	    cerr << "couldn't get fname in selection cmd.\n";
	    return FALSE;
	}
    }
    if(arg1) n_args++;
    if(arg2) n_args++;

    
    char **argv = (char **) fl_calloc(n_args, sizeof(char *)); 
    int argc = 0;

    argv[argc] = strdup(arg0);
    argc++;

    if(arg1){
	argv[argc] = strdup(arg1);
	argc++;
    }
    if(arg2){
	argv[argc] = strdup(arg2);
	argc++;
    }


    if(n_selected > 0){
	const char *name_buf = get_first_selection();
	while(name_buf){
	    argv[argc] = strdup(name_buf);
	    argc++;
	    name_buf = get_next_selection();
	}
    } else {
	argv[argc] = strdup(fname);
	argc++;
    }
    argv[argc] = NULL;

    int retval = do_fork_cmd(path, (const char **)argv);
    for(int i=0; i<n_args; i++){
	if(argv[i]){
	    fl_free(argv[i]);
	    argv[i] = NULL;
	}
    }
#ifdef SUN    
    fl_free((char *)argv);
#else    
    fl_free(argv);
#endif    
    argv = NULL;
    
    return retval;
}

const char*
browser_c::get_cwd(void){
	if (cwd_buf) return cwd_buf;
	return "<empty>";
}

void browser_c::set_cwd_buf(const char *buf){
    if(amd_dir_kludge){
	if(fclose(amd_dir_kludge) != 0){
	    cerr << "UC: couldn't close atmp dir.\n";
	}
	amd_dir_kludge = NULL;
    }
    amd_dir_kludge = fopen(buf, "r");
    if(!amd_dir_kludge){
	cerr << "UC: couldn't open atmp dir.\n";
    } else {
	strcpy(cwd_buf, buf);
    }
}

void browser_c::set_fontsize(int newsize){
    freeze();
    fl_set_panel_fontsize(panel, newsize);
    fl_set_object_lsize(text, newsize);
    fl_set_object_lsize(underscores, newsize);
    fl_set_object_lsize(search_text, newsize);
    rescale(TRUE);
    fl_redraw_form(form);
    unfreeze();
	//cerr << "setting fontsize for " << get_ID() << " to " << newsize << endl;
}

void browser_c::set_fontstyle(int newstyle){
    if(newstyle > 0){
	freeze();
	fl_set_panel_fontstyle(panel, newstyle);
	fl_set_object_lstyle(text, newstyle);
	fl_set_object_lstyle(underscores, newstyle);
	fl_set_object_lstyle(search_text, newstyle);
	rescale(TRUE);
	fl_redraw_form(form);
	unfreeze();
    }
}

void browser_c::set_sort_defaults(void){
    /* sort options */
    fst_sort = filetype_e;
    snd_sort = name_e;
    sort_fst_reverse_flag = FALSE;
    sort_snd_reverse_flag = FALSE;
}

void browser_c::set_show_defaults(void){
    /* 
     * defaults to medium long listing, with directories
     * listed first, hidden files visible.
     */
    list_hidden_flag = TRUE;    
    show_filetype_flag = TRUE;
    show_inode_flag = FALSE;
    show_nlinks_flag = FALSE;
    show_owner_flag = TRUE;
    show_perms_flag = TRUE;
    show_symbolic_flag = TRUE;
    show_group_flag = FALSE;
    show_size_flag = TRUE;
    show_modific_flag = TRUE;
    show_access_flag = FALSE;
    show_time_seconds_flag = FALSE;
    show_time_years_flag = FALSE;
    show_type_char_flag = TRUE;

    follow_links_flag = FALSE;
    long_wintitle_flag = TRUE;
}

void browser_c::set_font_defaults(void){
    // FL_FIXEDBOLD_STYLE);	/* courier-bold */
    set_fontstyle(FL_FIXED_STYLE);	                /* courier */
    set_fontsize(FL_NORMAL_SIZE);	                /* courier-bold 12 ? */
}

browser_c::~browser_c(){
    // error_msg("browser_c::destructor called.");
    //cerr << "destroying browser #" << my_ID << endl;
    if(form){
	//cerr << "form is valid." << endl;
	if(form->visible){
	    fl_hide_form(form);
	}
	for(int line=1; line<= FL_PANEL_MAXLINE; line++){
	    SFILESTAT *client_data = (SFILESTAT *) fl_get_panel_line_client_data(panel, line);
	    if(client_data){
		if(client_data->f_name){
		    fl_free(client_data->f_name);
		    client_data->f_name = NULL;
		}
#ifdef SUN
		fl_free((char *)client_data);
#else
		fl_free(client_data);
#endif
		fl_set_panel_line_client_data(panel, line, NULL);

	    }
	}
	// This crashes again and again and again. (Feb 97). Leave it.
	//fl_free_form(form);  // HACK (when it was commented out)
	form = NULL;
	if(cwd_buf){
	    fl_free(cwd_buf);
	    cwd_buf = NULL;
	}
	if(filter_pattern){
	    fl_free(filter_pattern);
	    filter_pattern = NULL;
	}
	/*
	 * fix the ring data structure.
	 */
	prev_browser->next_browser = next_browser;
	next_browser->prev_browser = prev_browser;
    } else {
	cerr << "browser destructor: form is invalid." << endl;
    }
}

browser_c::browser_c(void){
    amd_dir_kludge = NULL;
    last_wintitle = NULL;

    fl_set_panel_dnd_cleanup_cb        (dnd_cleanup_func);
    fl_set_panel_dnd_execute_cb        (dnd_exec_func);
    fl_set_panel_dnd_source_verify_cb  (dnd_source_verify_func);
    fl_set_panel_dnd_track_show_cb     (dnd_track_show_func);
    fl_set_panel_dnd_track_external_cb (dnd_track_external_func);
    /* here we try to be very politically correct:
     * proper size, and initialized to '.'
     */

    cwd_buf = (char *)fl_calloc(cwd_bufsize, sizeof(char));
	assert(cwd_buf);
    strcpy(cwd_buf, ".");

    if(!getcwd(cwd_buf,cwd_bufsize)){
	program_unstable("UC internal: problems double-checking current directory.");
    }
    next_browser = this;     // be correct.
    prev_browser = this;     // be correct.
    
    selection_workline = -1; // safeguard
    reading_flag = FALSE;
    
    cur_up_flag = FALSE;
    tool_escape_flag	= FALSE;

    // window settings, just in case.
    
    my_ID = next_ID;
    if(dbg_flag){
	//cerr << "constructor: making browser #" << my_ID << endl;
    }
    next_ID++;

    // just in case.
    win_w = 600;
    win_h = 120;
    win_x = 10;
    win_y = 10;
    
    filter_pattern = strdup("*");    
    filter_full_line_flag = FALSE;
    set_show_defaults();
    set_sort_defaults();
    

    /* forms */
    create();
    set_dark_yellow_bright();    // new default colours.    
    fl_hide_object(search_text);
    fl_set_object_label(search_text, "");  
    /* initialization kludge. Allows set_source to be clever */
    source_flag = TRUE; set_source(FALSE);
    /* ok, we also allow set_target to be clever :-) */
    target_flag = TRUE; set_target(FALSE);

    // catches initialization bug.
    strcpy(search_last_fname_buf, "");

    set_font_defaults();
}

/*
 * try the supplied filename, or use default name in
 * the home directory.
 */

int browser_c::defaults_save(ofstream &ofile){
    if(!ofile) return FALSE;

    char bbuf[1024];
    sprintf(bbuf, "browser%d.", my_ID);
    ofile << endl;
    ofile << "# resources for " << bbuf << "\n";
    ofile << "# \n";
    ofile << bbuf << "list_hidden_flag:        " << list_hidden_flag << endl;
    ofile << bbuf << "show_access_flag:        " << show_access_flag << endl;
    ofile << bbuf << "show_filetype_flag:      " << show_filetype_flag << endl;
    ofile << bbuf << "show_group_flag:         " << show_group_flag << endl;    
    ofile << bbuf << "show_inode_flag:         " << show_inode_flag << endl;
    ofile << bbuf << "show_modific_flag:       " << show_modific_flag << endl;
    ofile << bbuf << "show_nlinks_flag:        " << show_nlinks_flag << endl;        
    ofile << bbuf << "show_owner_flag:         " << show_owner_flag << endl;
    ofile << bbuf << "show_perms_flag:         " << show_perms_flag << endl;
    ofile << bbuf << "show_size_flag:          " << show_size_flag << endl;        
    ofile << bbuf << "show_symbolic_flag:      " << show_symbolic_flag << endl;
    ofile << bbuf << "show_time_seconds_flag:  " << show_time_seconds_flag << endl;        
    ofile << bbuf << "show_time_years_flag:    " << show_time_years_flag << endl;
    ofile << bbuf << "show_type_char_flag:     " << show_type_char_flag << endl;
    ofile << bbuf << "follow_links_flag:       " << follow_links_flag << endl;
    ofile << bbuf << "long_wintitle_flag:      " << long_wintitle_flag << endl;
    ofile << endl;
    ofile << bbuf << "sort_fst_reverse_flag:   " << sort_fst_reverse_flag << endl;
    ofile << bbuf << "sort_snd_reverse_flag:   " << sort_snd_reverse_flag << endl;        
    ofile << bbuf << "fst_sort:                " << fst_sort << endl;
    ofile << bbuf << "snd_sort:                " << snd_sort << endl;        
    ofile << endl;

    ofile << bbuf << "fontsize:                " << ((int) fl_get_panel_fontsize(panel)) << endl;
    ofile << bbuf << "fontstyle:               " << fl_get_panel_fontstyle(panel) << endl;
    ofile << endl;     

    // browser colour parameters 
    int SelCurLabelCol, SelCurBackCol, SelLabelCol, SelBackCol, 
        CurLabelCol, CurBackCol, LabelCol, BackCol
    ;
    fl_get_panel_colors(panel, 
	&SelCurLabelCol, &SelCurBackCol, &SelLabelCol, &SelBackCol,
	&CurLabelCol, &CurBackCol, &LabelCol, &BackCol
    );

    ofile << bbuf << "SelCurLabelCol: " << SelCurLabelCol << endl;
    ofile << bbuf << "SelCurBackCol:  " << SelCurBackCol << endl;
    ofile << bbuf << "SelLabelCol:    " << SelLabelCol << endl;
    ofile << bbuf << "SelBackCol:     " << SelBackCol << endl;
    ofile << bbuf << "CurLabelCol:    " << CurLabelCol << endl;
    ofile << bbuf << "CurBackCol:     " << CurBackCol << endl;
    ofile << bbuf << "LabelCol:       " << LabelCol << endl;
    ofile << bbuf << "BackCol:        " << BackCol << endl;

    //window specs:	ofile << bbuf << "window_x:  " << x << endl;     
    ofile << bbuf << "window_x:       " << form->x << endl;
    ofile << bbuf << "window_y:       " << form->y << endl;
    ofile << bbuf << "window_w:       " << form->w << endl;
    ofile << bbuf << "window_h:       " << form->h << endl;

    ofile << endl;
    return TRUE;
}

int browser_c::defaults_load(const char *fname){
    // opens and closes.
    if(!fname) fname = tool_default_fname;
    if(!fname) return FALSE;
    ifstream ifile;
    ifile.open(fname);
    if(!ifile) return FALSE;
    int retval = defaults_load(ifile);
    ifile.close();
    return retval;
}

/*
 * note: this is inefficient, but works. As it is a non-crucial
 * function, I don't care. Classik Kludge.
 */

int browser_c::defaults_load(ifstream &ifile){
    // does not open and close.
    // input file stream operations 
    if(!ifile) return FALSE;
    
    if(ifile.rdbuf()->seekpos(0,0) < 0){
	cerr << "browser_c::couldn't rewind defaults file.\n";
    }
    // browser colour parameters 
    int SelCurLabelCol, SelCurBackCol, SelLabelCol, SelBackCol, 
        CurLabelCol, CurBackCol, LabelCol, BackCol
    ;

    SelCurLabelCol = SelCurBackCol = SelLabelCol = SelBackCol = 0;
    CurLabelCol = CurBackCol = LabelCol = BackCol = 0;

    int found_flag = FALSE,
	not_yet_found_flag = FALSE
    ;
    

    // temporary buffer 
    char bbuf[1024];

    // ID token - e.g. `browser1'
    char btokbuf[64];
    sprintf(btokbuf, "browser%d", my_ID);

    while(ifile.getline(bbuf, 1024) ){
	if(bbuf[0] == '#' || bbuf[0] == '\0'){
	    continue;
	    // it is a comment, or empty line.
	}
	// look for object ID token. 
	char *tok = strtok(bbuf, ".");
	if(!tok){
	    cerr << "UC: strange line in default database file - no object.\n";
	    continue;
	}
	if(strcmp(tok, btokbuf) != 0){
	    continue;
	    // it is for a different object
	}
	// look for item 
	tok = strtok(NULL, ":");
	if(!tok){
	    cerr << "UC: strange line in default database file - no item for `" << btokbuf << "'.\n";
	    continue;
	}

	// remember item, look for value 
	char *valtok = strtok(NULL, " ");
	if(!valtok){
	    cerr << "UC: strange line in default database file - no value for `"
		<< btokbuf << "." << tok << "'.\n";
	    continue;
	}
	not_yet_found_flag = found_flag;
	found_flag = TRUE;
	
	// decode value (only ints allowed) 
	int value = atoi(valtok);

	// decode item - I know there are fancier ways to do this. 
	if(strcmp(tok, "show_access_flag") == 0){
	    show_access_flag = value;
	} else if(strcmp(tok, "list_hidden_flag") == 0){
	    list_hidden_flag = value;
	} else if(strcmp(tok, "show_filetype_flag") == 0){
	    show_filetype_flag = value;
	} else if(strcmp(tok, "show_group_flag") == 0){
	    show_group_flag = value;    
	} else if(strcmp(tok, "show_inode_flag") == 0){
	    show_inode_flag = value;    
	} else if(strcmp(tok, "show_modific_flag") == 0){
	    show_modific_flag = value;    
	} else if(strcmp(tok, "show_nlinks_flag") == 0){
	    show_nlinks_flag = value;    
	} else if(strcmp(tok, "show_owner_flag") == 0){
	    show_owner_flag = value;    
	} else if(strcmp(tok, "show_perms_flag") == 0){
	    show_perms_flag = value;    
	} else if(strcmp(tok, "show_size_flag") == 0){
	    show_size_flag = value;    
	} else if(strcmp(tok, "show_symbolic_flag") == 0){
	    show_symbolic_flag = value;    
	} else if(strcmp(tok, "show_time_seconds_flag") == 0){
	    show_time_seconds_flag = value;    
	} else if(strcmp(tok, "show_time_years_flag") == 0){
	    show_time_years_flag = value;    
	} else if(strcmp(tok, "show_type_char_flag") == 0){
	    show_type_char_flag = value;    
	} else if(strcmp(tok, "follow_links_flag") == 0){
	    follow_links_flag = value;    
	} else if(strcmp(tok, "long_wintitle_flag") == 0){
	    long_wintitle_flag = value;    
	} else if(strcmp(tok, "sort_fst_reverse_flag") == 0){
	    sort_fst_reverse_flag = value;    
	} else if(strcmp(tok, "sort_snd_reverse_flag") == 0){
	    sort_snd_reverse_flag = value;    
	} else if(strcmp(tok, "fst_sort") == 0){
	    fst_sort = (sorttype_et) value;    
	} else if(strcmp(tok, "snd_sort") == 0){
	    snd_sort = (sorttype_et) value;    
	} else if(strcmp(tok, "window_x") == 0){
	    win_x = value;    
	} else if(strcmp(tok, "window_y") == 0){
	    win_y = value;    
	} else if(strcmp(tok, "window_w") == 0){
	    win_w = value;    
	} else if(strcmp(tok, "window_h") == 0){
	    win_h = value;    
	} else if(strcmp(tok, "SelCurLabelCol") == 0){
	    SelCurLabelCol = value;    
	} else if(strcmp(tok, "SelCurBackCol") == 0){
	    SelCurBackCol = value;    
	} else if(strcmp(tok, "SelLabelCol") == 0){
	    SelLabelCol = value;    
	} else if(strcmp(tok, "SelBackCol") == 0){
	    SelBackCol = value;    
	} else if(strcmp(tok, "CurLabelCol") == 0){
	    CurLabelCol = value;    
	} else if(strcmp(tok, "CurBackCol") == 0){
	    CurBackCol = value;    
	} else if(strcmp(tok, "LabelCol") == 0){
	    LabelCol = value;    
	} else if(strcmp(tok, "BackCol") == 0){
	    BackCol = value;    
	} else if(strcmp(tok, "fontsize") == 0){
	    set_fontsize(value);
	} else if(strcmp(tok, "fontstyle") == 0){
	    set_fontstyle(value);
	} else {
	    cerr << "unrecognized object.item in default database file: " << tok << endl;
	    cerr << "It may be a leftover from an older UC version. To get rid of this\n",
	    cerr << "message, simply save the configuration again.\n";
	    if(!not_yet_found_flag){
		found_flag = FALSE;
	    }
	}
    }

    if(!found_flag) return FALSE;

    fl_set_panel_colors(panel, 
	SelCurLabelCol, SelCurBackCol, SelLabelCol, SelBackCol,
	CurLabelCol, CurBackCol, LabelCol, BackCol
    );
    //fl_set_object_color(search_text, CurBackCol, CurBackCol);
    //fl_set_object_lcol (search_text, CurLabelCol);
    
    // this is slightly more intelligent: It checks the form->window,
    // so it doesn't interfere if this is the very first time
    // the defaults get loaded, i.e. before the first `show'.
    // additionally, `show' is clever enough to take the win
    // options into account anyway :-)

    window_props_update();

    return TRUE;
}

void browser_c::window_no_lunatic(void){
    // avoid totally lunatic settings.
    if(win_w < 50)win_w = 100;
    if(win_h < 50)win_h = 100;
}

void browser_c::window_props_update(void){
    if(form->window){
	/* update window specs */
	window_no_lunatic();
	//freeze();  // freezing may be harmfull.
	fl_set_form_position(form, win_x, win_y);
	fl_set_form_size(form, win_w, win_h);
	//XFlush(fl_display);
	//tool_mung_forms(FALSE);
	rescale(TRUE);
	fl_redraw_form(form);
	//unfreeze();
    }
}

const char * browser_c::xdnd_track(FL_Coord mx, FL_Coord my){
    static char track_buf[MAXPATHLEN];
    const char *cwd = get_cwd();
    if (!cwd) return NULL;
    strcpy(track_buf, cwd);
    if (fl_get_panel_xdnd_track(panel, mx, my)){
	//cerr << "valid subtrack." << endl;
	const char *fname_buf = get_current_fname();
	if (!fname_buf) return NULL;
	strcat(track_buf, "/");
	strcat(track_buf, fname_buf);
    }
    return track_buf;
}

static void browser_help_func(void){
    if(dbg_flag) cerr << "brwser context sensitive help request." << endl;
    if(source_browser->search_text->visible){
	load_help_browser(browser_search_help_text, "Browser Search");
    } else {
	ghelp_browser_cb(NULL, 0);
    }
}

int browser_handle_free(FL_OBJECT *obj, int event, FL_Coord, FL_Coord, int, void *){
    // int key and void *xev are not needed.
    tool_got_event_flag = TRUE;
    if(!obj){
	cerr << "browser: obj == NULL.\n";
	return FALSE;
    }
    if(event == FL_DRAW){
	if(dbg_flag) cerr << "browser FL_DRAW event.\n";
	tool_find_browser(obj->form)->rescale();
	return TRUE;
    }
    return FALSE;
}

int browser_c::handle_keypress(XKeyEvent *xkev){	
    if(!xkev){
	// cerr << "browser: xevent == NULL.\n";
	return FALSE;
    }

    if(dbg_flag)cerr << "browser handle keypress called.\n";
    KeySym keysym;
    char keybuf[2];
    keybuf[1] = '\0';  // saveguard.
    
    //cerr << "numlock: " << fl_keypressed(XK_Num_Lock) << endl;
    XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
    if(keysym >= XK_F2 && keysym <= XK_F35){
	default_actions();
    }
    if(is_end(keysym, xkev)){
	EndKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_down(keysym, xkev)){
	DownArrowKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_page_down(keysym, xkev)){
	PageDownKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_left(keysym, xkev)){
	LeftArrowKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_begin(keysym, xkev)){
	Pad5KeyCB(NULL, 0);
	return TRUE;
    }
    if(is_right(keysym, xkev)){
	RightArrowKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_home(keysym, xkev)){
	HomeKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_up(keysym, xkev)){
	UpArrowKeyCB(NULL, 0);
	return TRUE;
    }
    if(is_page_up(keysym, xkev)){
	PageUpKeyCB(NULL, 0);
	return TRUE;
    }
    
    // first switch is short and preemptive-only. (no default key handling)
    switch(keysym){
    case XK_g:
    case XK_G:
	if (xkev->state & ControlMask) {
	    source_browser->default_actions();	// get rid of search text.
	    return TRUE;
	}
	// else continue, so that we can work on the G/g normally.
	break;
    }

    switch(keysym){
    case XK_F1:
	help_event_cb(browser_help_func);
	return TRUE;
    case XK_F2:
	if(tool_get_meta()){
	    OptionsButtonCB(NULL,0);
	} else {
	    if(dbg_flag) cerr << "brwser context sensitive options request." << endl;
	    //tool_set_source(obj->form);
	    browser_menu_event_cb(NULL, 0);
	}
	return TRUE;
    case XK_F3:
	MkDirButtonCB(NULL,0);
	return TRUE;
    case XK_F4:
	TreeButtonCB(NULL,0);
	return TRUE;
    case XK_F5:
	QuickEditButtonCB(NULL,0);
	return TRUE;
    case XK_F6:
	EditButtonCB(NULL,0);
	return TRUE;
    case XK_F7:
	RenameButtonCB(NULL,0);
	return TRUE;
    case XK_F8:
	MoveButtonCB(NULL,0);
	return TRUE;
    case XK_F9:
	DuplicateButtonCB(NULL,0);
	return TRUE;
    case XK_F10:
	CopyButtonCB(NULL,0);
	return TRUE;
    case SunXK_F36: if(!sun_keyboard_flag) return FALSE;
#ifdef DEC_ALPHA
    case XK_F13:
#endif
    case XK_F11:
	DeleteButtonCB(NULL,0);
	return TRUE;
    case SunXK_F37: if(!sun_keyboard_flag) return FALSE;
    case XK_F12:
	ExitButtonCB(NULL,0);
	return TRUE;
#ifdef XK_KP_Insert
    case XK_KP_Insert:
#endif
    case XK_KP_0:
	if(fl_keypressed(XK_Num_Lock))break;
    case XK_Insert:
	InsertKeyCB      (NULL, 0);
	return TRUE;
    case XK_KP_Tab:
    case XK_Tab:
	if(dbg_flag)cerr << "tab.\n";
	if(tool_get_ctrl()){
	    xkev->state = xkev->state & ~ControlMask;
	    xshell_sendevent_callback(xkev, XK_Tab);
	} else {
	    TabKeyCB     (NULL, 0);
	}
	return TRUE;
    case XK_F26: if(!sun_keyboard_flag) return FALSE;
    case XK_KP_Multiply:
	PadAsteriskKeyCB (NULL, 0);
	return TRUE;
    case XK_F24: if(!sun_keyboard_flag) return FALSE;
    case XK_KP_Subtract:
	DeselectButtonCB (NULL, 0);
	return TRUE;
    case XK_KP_Add:
	SelectButtonCB   (NULL, 0);
	return TRUE;
    case XK_F25: if(!sun_keyboard_flag) return FALSE;
    case XK_KP_Divide:
	PadVirguleKeyCB  (NULL, 0);
	return TRUE;
    case XK_Escape:
	source_browser->default_actions();	// get rid of search text.
	tool_escape_flag = TRUE;
	//cerr << "escape true\n";
	return TRUE;
	/*
	case XK_KP_:
	case XK_:
	    KeyCB      (NULL, 0); return TRUE;
	    */
    case NoSymbol:
	// cerr << "no symbol in browser event processing.\n";
	break;
    default:
	if(direct_window_access_cb(xkev, keysym, TRUE)) return TRUE;
	if(source_browser->key_event_callback(keybuf[0], keysym)) return TRUE;
	return xshell_sendevent_callback(xkev, keysym);
    }
    if(dbg_flag)cerr << "browser event unprocessed.\n";
    return FALSE;
}

int browser_c::handle_keyrelease(XKeyEvent *xkev){	
    if(!xkev){
	// cerr << "browser: xevent == NULL.\n";
	return FALSE;
    }

    if(dbg_flag)cerr << "browser handle keyrelease called.\n";
    KeySym keysym;
    char keybuf[2];
    keybuf[1] = '\0';  // saveguard.
    
    XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
    if(keysym == XK_Escape){
	tool_escape_flag = FALSE;
	//cerr << "escape false\n";
	return TRUE;
    }
    if(dbg_flag)cerr << "browser keyrelease event unprocessed.\n";
    return FALSE;
}

int browser_raw_cb(FL_FORM *form, void *xev){
    //cerr << "raw cb called. \n";
    if(!form){
	cerr << "NULL form in raw cb.\n";
	return FALSE;
    }
    // cerr << form->deactivated << endl;
    browser_c *tmp_browser = tool_find_browser(form);
    if(!tmp_browser){
	cerr << "NULL browser in raw cb.\n";
	return FALSE;
    }
    static browser_c *last_browser = NULL;
    static int had_click0_flag = FALSE, had_click1_flag = FALSE, x = 0, y = 0;
    static struct timeval prev_click_time;
    
    XButtonEvent *x_butt = (XButtonEvent*) xev;
    XEvent *xevent = (XEvent*) xev;
    switch(xevent->type){
    case FocusIn:	
	//cerr << "B FocusIn.\n";
	toolbox_attach(FALSE);
	return FALSE;
    case FocusOut:	
	//cerr << "B FocusOut.\n";
	//toolbox_attach(FALSE);
	return FALSE;
    case EnterNotify:
	//cerr << "B EnterNotify.\n";
	if(xevent->xcrossing.focus){
	    toolbox_attach(FALSE);
	}
	return FALSE;
    case LeaveNotify:
	//cerr << "B LeaveNotify.\n";
	if(xevent->xcrossing.focus){
	    toolbox_attach(FALSE);
	}
	return FALSE;
    case ButtonPress:
	//cerr << "Press\n";
	// No set source here, else jumping of the current line.
	// do it in selection_event_cb.
	// but: click in the browser ALWAYS means set_source.
	
	tool_got_event_flag = TRUE;
	if(x_butt->button == Button3){
	    // gobble for menu.
	    return TRUE;
	}
	
	if(x_butt->state & (Button2Mask | Button3Mask)){
	    //cerr << "2/3 pressed.\n";
	}
	if(x_butt->button == Button1){
	    int retval;
	    //cerr << "triggered by 1.\n";
	    struct timeval click_time;
	    gettimeofday(&click_time, NULL);
	    long diff = msecs_diff(click_time, prev_click_time);
	    /* need a click before. Same browser. No pointer motion. quick.
	     * and not in the text area. 
	     */
	    //cerr << x_butt->x << ", " << x_butt->y << endl;
	    if(had_click0_flag && last_browser == tmp_browser &&
	       (abs(x - x_butt->x) < 5) && (abs(y - x_butt->y) < 5) &&
	       (diff < double_click_msecs_max) &&
	       (x > 2 * box_width + 5) &&
	       (x < last_browser->form->w) &&
	       (y > box_height + 2) &&
	       (y < last_browser->form->h)
	       ){
		//cerr << "double click - gobble.\n";
		had_click1_flag = TRUE;
		retval = TRUE;
	    } else {
		had_click0_flag = TRUE;
		x = x_butt->x;
		y = x_butt->y;
		last_browser = tmp_browser;
		retval = FALSE;
	    }
	    prev_click_time.tv_sec  = click_time.tv_sec;
	    prev_click_time.tv_usec = click_time.tv_usec;
	    return retval;
	} else {
	    had_click0_flag = FALSE;
	}
	break;
    case ButtonRelease:
	tool_got_event_flag = TRUE;
	if(x_butt->button == Button3){
	    // gobble for menu.
	    source_browser->default_actions();
	    tool_set_source(form);
		//toolbox_attach(FALSE);
	    browser_menu_event_cb(NULL, 0);
	    return TRUE;
	}
	//cerr << "Release\n";
	//cerr << x_butt->x << ", " << x_butt->y << endl;
	if(had_click1_flag){// && x_butt->button == Button1){
	    had_click1_flag = FALSE;
	    had_click0_flag = FALSE;
	    // gobble the last release. Exec if no move only.
	    //cerr << "gobble release and trigger exec.\n";
	    FL_Coord x2,y2;
	    unsigned int key;
	    fl_get_form_mouse(tmp_browser->form, &x2 , &y2, &key);
	    //cerr << x2 << ", " << y2 << endl;

	    // special case: gobble event, but do set source.
	    tool_set_source(tmp_browser);
		//toolbox_attach(FALSE);

	    if((abs(x - x2) < 5) && (abs(y - y2) < 5)){
		NewEnterKeyCB(NULL, 0);
	    } else {
		// cerr << "cancelled dblclick request.\n";
	    }
	    return TRUE;
	}
	break;
    case KeyPress:
	tmp_browser->handle_keypress((XKeyEvent *) xev);
	return TRUE;
    case KeyRelease:
	// allways gobble releases here.
	tmp_browser->handle_keyrelease((XKeyEvent *) xev);
	return TRUE;
    case MotionNotify:
	// fast panel processing.
	tool_got_event_flag = TRUE; 
	return FALSE;
    default:
	if(had_click1_flag){
	    //cerr << "gobble in-between events.\n";
	    return TRUE;
	}
	break;
    }

/*
    if (browser_handle_free(tmp_browser->free_obj, FL_OTHER, 0, 0, 0, xev)){
	//toolbox_attach(FALSE);
	return TRUE;
    } else {
	return FALSE;
    }
    */
    return FALSE;
}

void browser_c::create(void){
    if(!fl_display){
	M_err("browser_c::create","fl not initialized");
	exit(-1);
    }
    if(tool_bw_flag){
	text_x_fudge = text_x_base_fudge+FL_PANEL_TEXTMARGIN;
    }

    // auto generated parts of routine. 

    FL_OBJECT *obj;
    form = fl_bgn_form(FL_NO_BOX,600,336);
    //obj = fl_add_box(FL_UP_BOX,0,0,600,336,"");  // sppppeeeeeeed !
    
    free_obj = obj = fl_add_free(FL_NORMAL_FREE,0,0,1,1,"",browser_handle_free);
    // free_obj->input = TRUE;
    // free_obj->wantall = TRUE;
    
    panel = obj = fl_add_panel(FL_PANEL,0,0,600,310,"");
    fl_set_object_boxtype(obj,FL_UP_BOX);
    fl_set_object_align(obj,FL_ALIGN_TOP);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE);
    fl_set_object_resize(obj, FL_RESIZE_ALL);
    fl_set_object_gravity(obj, SouthWestGravity, NorthEastGravity);
  /* note this is special ! */
    fl_set_call_back(obj,browser_selection_event_cb,0);

  text_box = obj = fl_add_box(FL_UP_BOX,2 * box_width,310,600,box_height,"");
  
  text = obj = fl_add_text(FL_NORMAL_TEXT,2 * box_width + text_x_fudge,310,600,box_height,"");
    fl_set_object_boxtype(obj,FL_NO_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE);
    fl_set_object_align(obj,FL_ALIGN_LEFT);
    fl_set_object_resize(obj, FL_RESIZE_X);
    fl_set_object_gravity(obj, NorthWestGravity, NorthEastGravity);

  underscores = obj = fl_add_text(FL_NORMAL_TEXT,2 * box_width + text_x_fudge,310,600,box_height,"");
    fl_set_object_boxtype(obj,FL_NO_BOX);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE);
    fl_set_object_align(obj,FL_ALIGN_LEFT);
    fl_set_object_resize(obj, FL_RESIZE_X);
    fl_set_object_gravity(obj, NorthWestGravity, NorthEastGravity);

  column_hidden_button = obj = fl_add_button(FL_HIDDEN_BUTTON,2 * box_width,310,600,box_height,"");
    fl_set_call_back(obj,::column_hidden_button_cb,0);
    fl_set_object_resize(obj, FL_RESIZE_X);
    fl_set_object_gravity(obj, NorthWestGravity, NorthEastGravity);

  source_box = obj = fl_add_box(FL_UP_BOX,0,310,box_width,box_height,"");
    fl_set_object_color(obj,tool_source_colour,47);
    fl_set_object_resize(obj, FL_RESIZE_NONE);
    fl_set_object_gravity(obj, NorthWestGravity, NorthWestGravity);
  target_box = obj = fl_add_box(FL_UP_BOX,box_width,310,box_width+1,box_height,"");
    fl_set_object_color(obj,tool_target_colour,47);
    fl_set_object_resize(obj, FL_RESIZE_NONE);
    fl_set_object_gravity(obj, NorthWestGravity, NorthWestGravity);
  search_text = obj = fl_add_text(FL_NORMAL_TEXT,515,310,85,25,"...");
    fl_set_object_boxtype(obj,FL_FRAME_BOX);
    fl_set_object_color(obj,0,47);
    fl_set_object_lcol(obj,7);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE);
    fl_set_object_resize(obj, FL_RESIZE_NONE);
    fl_set_object_gravity(obj, NorthEastGravity, NorthEastGravity);

/*
  menu_text = obj = fl_add_text(FL_NORMAL_TEXT,1,310,2*box_width-1,box_height,"@menu");
    fl_set_object_boxtype(obj,FL_NO_BOX);
    fl_set_object_lcol(obj,FL_COL1);
    fl_set_object_align(obj,FL_ALIGN_CENTER);
    fl_set_object_resize(obj, FL_RESIZE_NONE);
    fl_set_object_gravity(obj, NorthWestGravity, NorthWestGravity);
*/
  
  menu_hidden_button = obj = fl_add_button(FL_HIDDEN_BUTTON,0,310,2*box_width,box_height,"");
    fl_set_call_back(obj,menu_hidden_button_cb,0);
    fl_set_object_resize(obj, FL_RESIZE_NONE);
    fl_set_object_gravity(obj, NorthWestGravity, NorthWestGravity);

  fl_end_form();
}

int browser_c::set_current_fname(const char *fname, int auto_center_flag){
    if(!set_current_exact_fname(fname, auto_center_flag)){
	char fname_buf[MAXPATHLEN+1];
	strcpy(fname_buf, fname);
	int len = strlen(fname_buf);
	if(len <= 0) return FALSE;
	if(len == 1 && fname_buf[0] == '*') return FALSE;
	if(fname_buf[len-1] != '*'){
	    // attempt extension with '*' first:
	    strcat(fname_buf, "*");
	    len++;
	    if(set_current_exact_fname(fname_buf, auto_center_flag)) return TRUE;
	}
	while(len > 1){
	    fname_buf[len-2] = '*';
	    fname_buf[len-1] = '\0';
	    if(set_current_exact_fname(fname_buf, auto_center_flag)){
		return TRUE;
	    } else {
		len--;
	    }
	}
    } else {
	return TRUE;
    }
    return FALSE;
}

int browser_c::set_current_exact_fname(const char *fname, int auto_center_flag){
    if(!fname) return FALSE;
    int maxline = fl_get_panel_maxline(panel);
    for(int i=1; i <= maxline; i++){
	if(wildcard_match(fname, extract_filename(i)) ){
	    set_current(i, auto_center_flag);
	    return TRUE;
	}
    }
    return FALSE;
}

int browser_c::set_current_below_match(const char *pattern, int below_line){
    if(!pattern) return FALSE;
    int maxline = fl_get_panel_maxline(panel);
    for(int i=below_line+1; i <= maxline; i++){
	if(wildcard_match(pattern, extract_filename(i)) ){
	    set_current(i);
	    return TRUE;
	}
    }
    return FALSE;
}

int browser_c::set_current_first_match(const char *pattern){
    if(!pattern) return FALSE;
    int maxline = fl_get_panel_maxline(panel);
    for(int i=1; i <= maxline; i++){
	if(wildcard_match(pattern, extract_filename(i)) ){
	    set_current(i);
	    return TRUE;
	}
    }
    return FALSE;
}



int browser_c::deselect_first_name_match(const char *pattern){
    int maxline = fl_get_panel_maxline(panel);
    for(int i=1; i <= maxline; i++){
	if(wildcard_match(pattern, extract_filename(i)) ){
	    if(dbg_flag) cerr << "deselection match at line" << i << endl;
	    if(fl_isselected_panel_line(panel,i)){
		fl_deselect_panel_line(panel,i);
		if(dbg_flag) cerr << "trying to deselect line" << i << endl;
		return TRUE;
	    }   /* else already deselected. */
	    return FALSE;
	}
    }
    return FALSE;
}

/*
 * checking wether a line is already in a state in order to
 * avoid redrawing _may_ work ... hard to estimate the true
 * benefits. What the hell.
 */

void browser_c::deselect_all_name_matches(const char *pattern){
    selection_remember();
    int maxline = fl_get_panel_maxline(panel);
    // printf("trying to deselect pattern %s.\n", pattern);
    freeze();
    for(int i=1; i <= maxline; i++){
	if(wildcard_match(pattern, extract_filename(i)) ){
	    if(dbg_flag){
		 //if(dbg_flag) cerr << "deselection match at line" << i << endl;
	    }
	    if(fl_isselected_panel_line(panel,i)){
		fl_deselect_panel_line(panel,i);
		if(dbg_flag){
		     //if(dbg_flag) cerr << "trying to deselect line" << i << endl;
                }
	    } /* else already deselected. */
	}
    }
    unfreeze();
    selection_change_msg();
}

void browser_c::deselect_all_line_matches(const char *pattern){
    selection_remember();
    int maxline = fl_get_panel_maxline(panel);
    // printf("trying to deselect pattern %s.\n", pattern);
    freeze();
    for(int i=1; i <= maxline; i++){
	if(wildcard_match(pattern, fl_get_panel_line(panel, i) ) ){
	    if(dbg_flag){
                //if(dbg_flag) cerr << "deselection match at line" << i << endl;
            }
	    if(fl_isselected_panel_line(panel,i)){
		fl_deselect_panel_line(panel,i);
		if(dbg_flag){
		    //if(dbg_flag) cerr << "trying to deselect line" << i << endl;
                }
	    } /* else already deselected. */
	}
    }
    unfreeze();		
    selection_change_msg();
}

int browser_c::calc_fieldnum(sorttype_et which){
    int field_num = 0;
    if(which == inode_e)    return show_inode_flag  ? field_num : -1;
    field_num += show_inode_flag;
    if(which == filetype_e) return show_filetype_flag ? field_num : -1;
    field_num += show_filetype_flag;
    if(which == perms_e)    return show_perms_flag  ? field_num : -1;
    field_num += show_perms_flag;
    if(which == nlinks_e)   return show_nlinks_flag ? field_num : -1;
    field_num += show_nlinks_flag;
    if(which == owner_e)    return show_owner_flag  ? field_num : -1;
    field_num += show_owner_flag;
    if(which == group_e)    return show_group_flag  ? field_num : -1;
    field_num += show_group_flag;
    if(which == size_e)     return show_size_flag   ? field_num : -1;
    cerr << "not implemented: extract modific, access or higher.\n";
    return -1;
}
    
char *browser_c::extract_item(int line, sorttype_et which){
    if(!linecheck(line)) return NULL;
    int field_num = calc_fieldnum(which);
	if (show_filetype_flag && show_perms_flag) {
		// concat, special case.
		field_num--;
	}
    if(field_num < 0) return NULL;

    const char *line_buf =  fl_get_panel_line(panel, line);
    if(!line_buf){
		error_msg("line %d empty.", line);
		return NULL;
    }

    static char buf[1024];
    char *tokbuf = "<nothing>";

    // jump over the leading magic icon info:
    if(line_buf[0] == '@' && line_buf[1] != '\0'){
		line_buf += 2;
    }
    strncpy(buf, line_buf, 1024);
    
    int first = TRUE;

    while(field_num>=0){
		if(first){
		    first = FALSE;
	    	tokbuf = strtok(buf, " ");
		} else {
	    	tokbuf = strtok(NULL, " ");
		}
		field_num--;
		if(!tokbuf){
	    	static int lfirst = TRUE;
	    	if(lfirst){
				error_msg("UC: consistency error, string exhausted to early in extract_item.");
				first = FALSE;
	    	}
	    	return NULL;
		}
    }
	// hackhack!
	if (show_filetype_flag && show_perms_flag && strlen(tokbuf)) {
		if (which == filetype_e) {
			tokbuf[1] = '\0';	// delimit the type char.
		} else if (which == perms_e) {
			tokbuf++;			// jump over leading type char.
		}
	}
    if(dbg_flag) cerr << "extracting item: " << tokbuf << endl;
    return tokbuf;
}

void browser_c::selection_remember(void){
    old_n_selected = get_n_selected();
}

void browser_c::selection_change_msg(void){
    int n_selected = get_n_selected(),
	diff
    ;
    diff = n_selected - old_n_selected;
    if(diff == 0){
	status_msg(1, "no change (%d total selected).", n_selected);
    } else if(diff < 0){
	status_msg(1, "%d new files deselected (makes %d total selected).", -diff, n_selected);	
    } else {
	status_msg(1, "%d new files selected (makes %d total selected).", diff, n_selected);	
    }
}

void browser_c::select_all_item_matches(const char *pattern, sorttype_et which){
    selection_remember();
    freeze();
    int maxline = fl_get_panel_maxline(panel);
    for(int line=1; line <= maxline; line++){
	if(wildcard_match(pattern, extract_item(line, which))){
	    fl_select_panel_line(panel, line);
	}
    }
    deselect_first_name_match(".");
    deselect_first_name_match("..");
    unfreeze();
    selection_change_msg();
}

void browser_c::deselect_all_item_matches(const char *pattern, sorttype_et which){
    selection_remember();
    freeze();
    int maxline = fl_get_panel_maxline(panel);
    for(int line=1; line <= maxline; line++){
	if(wildcard_match(pattern, extract_item(line, which))){
	    fl_deselect_panel_line(panel, line);
	}
    }
    unfreeze();
    selection_change_msg();
}

void browser_c::select_all_line_matches(const char *pattern){
    const char *fname = "";
    int do_file_flag = TRUE,
	have_dot_flag = FALSE,    // little optimization.
	have_dotdot_flag = FALSE
    ;	    
    
    selection_remember();
    int maxline = fl_get_panel_maxline(panel);
    // printf("trying to select pattern %s.\n", pattern);
    freeze();
    for(int i=1; i <= maxline; i++){
	if(!(have_dot_flag && have_dotdot_flag)){
	    fname = extract_filename(i);
	}   // else we don't need to check for the fname anymore. 
	do_file_flag = TRUE;
	
	if(!have_dot_flag){
	    if(strcmp(fname, ".")==0){
		do_file_flag = FALSE;
		have_dot_flag = TRUE;
	    }
	}
	if(!have_dotdot_flag){
	    if(strcmp(fname, "..")==0){
		do_file_flag = FALSE;
		have_dotdot_flag = TRUE;
	    }
	}
	if(do_file_flag){
	    if(wildcard_match(pattern, fl_get_panel_line(panel, i) ) ){
		if(dbg_flag){
		    //if(dbg_flag) cerr << "selection match at line" << i << endl;
		}
		if(!fl_isselected_panel_line(panel,i)){
		    fl_select_panel_line(panel,i);
		    if(dbg_flag){
			//if(dbg_flag) cerr << "trying to select line" << i << endl;
		    }
		} /* else already selected. */
	    }
	}
    }
    unfreeze();
    selection_change_msg();
}

void browser_c::select_all_name_matches(const char *pattern){
    const char *fname;
    int do_file_flag = TRUE,
	have_dot_flag = FALSE,    // little optimization.
	have_dotdot_flag = FALSE
    ;	    
    
    selection_remember();
    int maxline = fl_get_panel_maxline(panel);
    // printf("trying to select pattern %s.\n", pattern);
    freeze();
    for(int i=1; i <= maxline; i++){
	fname = extract_filename(i);
	do_file_flag = TRUE;
	
	if(!have_dot_flag){
	    if(strcmp(fname, ".")==0){
		do_file_flag = FALSE;
		have_dot_flag = TRUE;
	    }
	}
	if(!have_dotdot_flag){
	    if(strcmp(fname, "..")==0){
		do_file_flag = FALSE;
		have_dotdot_flag = TRUE;
	    }
	}
	if(do_file_flag){
	    if(wildcard_match(pattern, fname) ){
		if(dbg_flag){
		    //if(dbg_flag) cerr << "selection match at line" << i << endl;
		}
		if(!fl_isselected_panel_line(panel,i)){
		    fl_select_panel_line(panel,i);
		    if(dbg_flag){
			//if(dbg_flag) cerr << "trying to select line" << i << endl;
		    }
		} /* else already selected. */
	    }
	}
    }
    unfreeze();		
    selection_change_msg();
}

/* 
 * do an action function on all selected items. If none is selected,
 * grab the current line. An action function is defined to take 
 * a file name as its argument, returning wether the user wants
 * to continue = 1, or cancel all = 0. Thus, we can fine tune the
 * update requirements, because the function also takes a counter
 * as argument, which it increments if the particular file was
 * changed.
 * This function goes through major pains to avoid aliasing problems
 * of the internal buffers used in get_first_selection etc. 
 */

int browser_c::selection_action_loop(action_func_pt action){
    const char *current_selection = get_first_selection();
    char sel_buf[MAXPATHLEN+1];
    int n_changes = 0;
    if(!current_selection){
	/* look at current line */
	current_selection = get_current_fname();
	if(current_selection){
	    //if(dbg_flag) cerr << "loop: curr op on " << current_selection << endl;
	    strcpy(sel_buf, current_selection);
	    (*action)(sel_buf, n_changes);
	} else {
	    //if(dbg_flag) cerr << "loop: No current line ?" << endl;
	}
    } else {
	while(current_selection){
	    //if(dbg_flag) cerr << "loop: op on " << current_selection << endl;
	    strcpy(sel_buf, current_selection);
	    if((*action)(sel_buf, n_changes)){
		current_selection = get_next_selection();
	    } else {
		// cancel all called, leave loop.
		//if(dbg_flag) cerr << "operation cancelled." << endl;
		current_selection = NULL;
	    }
	}
    }
    /* the following is a little safeguard. 
     */
    selection_workline = -1;
    return n_changes;
}

int browser_c::get_current_line(void){
    return fl_get_panel_current(panel);
}

llong_t browser_c::get_selection_byte_size(int use_current_flag){
    if(!show_size_flag) return -1;
    int line_no;
    llong_t bytes = 0;
    const char *buf = get_first_selection(line_no);
    if(!buf && use_current_flag){
	bytes = extract_size(fl_get_panel_current(panel));
    } else {
	while(buf){
	    bytes += extract_size(line_no);
	    buf = get_next_selection(line_no);
	}
    }
    return bytes;
}
	
llong_t browser_c::get_total_byte_size(void){
    if(!show_size_flag) return -1;
    llong_t bytes = 0;
    int maxline = fl_get_panel_maxline(panel);

    for(int i=1; i<=maxline; i++){
	bytes += extract_size(i);
    }
    return bytes;
}

int browser_c::get_selection_workline(void){
    if(selection_workline == -1){
	return fl_get_panel_current(panel);
    } else {
	return selection_workline;
    }
}

int browser_c::get_n_selected(void){
    if (!panel) return -1;
    return fl_get_panel_n_selected(panel);
}

/*
 * in the following implementation, note that we 
 * return a pointer to a static character array.
 * this is potentially a source of problems.
 */

const char *browser_c::extract_filename(int line){
    if(!linecheck(line)) return NULL;

    const char *tmpbuf =  fl_get_panel_line(panel, line);
    if(!tmpbuf){
	error_msg("no file name in line %d, line empty.", line);
	return NULL;
    } else {
	SFILESTAT *client_data = (SFILESTAT *) fl_get_panel_line_client_data(panel, line);
	if(client_data){
	    if(client_data->f_name){
		if(dbg_flag) cerr << "new extracting: " << client_data->f_name << "\n";
		static char buf[MAXPATHLEN+1];
		strcpy(buf, client_data->f_name);
		return buf;
	    } else {
		cerr << "extract: emtpy fname in client data !\n";
		return NULL;
	    }
	} else {
	    cerr << "extract: no client data !\n";
	    return NULL;
	}
    }
}

llong_t browser_c::extract_size(int line){
    if(!linecheck(line)) return 0;
    SFILESTAT *client_data = (SFILESTAT *) fl_get_panel_line_client_data(panel, line);
    if (!client_data) {
	cerr << "extract: no client data !\n";
	return 0;
    }
    return client_data->f_stat.st_size;
}

char browser_c::extract_type(int line){
    if(!linecheck(line)) return '\0';
    SFILESTAT *client_data = (SFILESTAT *) fl_get_panel_line_client_data(panel, line);
    if (!client_data) {
	cerr << "extract: no client data !\n";
	return '?';
    }
    switch(client_data->f_stat.st_mode & S_IFMT){
    case S_IFBLK:	return 'b';
    case S_IFCHR:	return 'c';
    case S_IFDIR:	return 'd';
    case S_IFIFO:	return 'p';
    case S_IFLNK:	return 'l';
    case S_IFREG:	return '-';
    case S_IFSOCK:	return 's';
    default:
	cerr << "unknown type field for " << client_data->f_name << endl;
	return '?';
    }
}

const char *browser_c::get_current_fname(int &current_line){
    current_line = fl_get_panel_current(panel);
    if(!linecheck(current_line)) return NULL;
    return extract_filename(current_line);
}

void browser_c::deselect_file(const char *fname){
    int found_flag = FALSE;
    if(!fname) program_unstable("NULL fname in deselect file");
    if(linecheck(selection_workline, TRUE)){
	if(strcmp(fname , extract_filename(selection_workline)) == 0){
	    // found cache line.
	    //if(dbg_flag) cerr << "cache hit: " << fname << endl;
	    fl_deselect_panel_line(panel, selection_workline);
	    found_flag = TRUE;
	}
    }
    if(!found_flag){
	if(dbg_flag) cerr << "cache miss: " << fname << endl;
	deselect_first_name_match(fname);
    }
}

const char *browser_c::check_first_selection(int &first_selected_line){
    int maxline = fl_get_panel_maxline(panel);
    
    for(int i=1; i <= maxline; i++){
	if(fl_isselected_panel_line(panel,i)){
	    first_selected_line = i;
	    return extract_filename(i);
	}
    }
    first_selected_line = 0;
    return NULL;	// no selection.
}

const char *browser_c::get_first_selection(int &first_selected_line){
    int maxline = fl_get_panel_maxline(panel);
    
    for(int i=1; i <= maxline; i++){
	if(fl_isselected_panel_line(panel,i)){
	    selection_workline = i;
	    first_selected_line = i;
	    //if(dbg_flag) cerr << "first selected line" << i << endl;
	    return extract_filename(i);
	}
    }
    first_selected_line = 0;
    return NULL;	// no selection.
}

const char *browser_c::get_next_selection(int &next_selected_line){
    if(selection_workline == -1){
	error_msg("UC internal: get_next_selection without get_first_selection.");
	return NULL;
    }

    int maxline = fl_get_panel_maxline(panel);
    for(int i = selection_workline+1; i <= maxline; i++){
	if(fl_isselected_panel_line(panel,i)){
	    if(dbg_flag){
		//if(dbg_flag) cerr << "next selected line" << i << endl;
 	    }
	    selection_workline = i;
	    next_selected_line = i;
	    return extract_filename(i);
	}
    }
    /* 
     * the following is a little safeguard.
     */
    next_selected_line = 0;
    selection_workline = -1;    
    return NULL;	// no more selections
}

void browser_c::toggle_select_current(void){
    int line = fl_get_panel_current(panel);
    if(!linecheck(line)) return;
    if(fl_isselected_panel_line(panel,line)){
	fl_deselect_panel_line(panel,line);
	status_msg(1, "1 name deselected (makes %d total selected).", get_n_selected());
    } else {
	fl_select_panel_line(panel,line);
	status_msg(1, "1 new name selected (makes %d total selected).", get_n_selected());
    }
    update_selection_state();
}

void browser_c::headline(void){
    static char buf[1024], ubuf[1024];
    /*
     * extra indent kludge :-) hardwired to the font. Not necessary
     * anymore.
     */

    static const char
        *filetype_str        = "T ",
        *filetype_perms_str  = "T",
        *group_sym_str       = "   Group ",
        *group_num_str       = " GID ",       
        *inode_str           = " Inode ",
        *nlinks_str          = "#Lks ",
        *owner_sym_str       = "   Owner ",
        *owner_num_str       = " OID ",
        *perms_str           = " Permiss. ",
	*show_access_str     = "      Access ",
        *show_modific_str    = "Modification ",
        *show_time_years_str = "     ",
        *show_time_secs_str  = "   ",
        *size_str            = "     Size ",
	
        *filetype_ustr        = "_ ",
        *filetype_perms_ustr  = "_",
        *group_sym_ustr       = "   _     ",
        *group_num_ustr       = " _   ",       
        *inode_ustr           = " _     ",
        *nlinks_ustr          = " _   ",
        *owner_sym_ustr       = "   _     ",
        *owner_num_ustr       = " _   ",
        *perms_ustr           = " _        ",
	*show_access_ustr     = "      _      ",
        *show_modific_ustr    = "_            ",
        *show_time_years_ustr = "     ",
        *show_time_secs_ustr  = "   ",
        *size_ustr            = "     _    "    
    ;

    strcpy(buf, line_icon_indent); 
    strcpy(ubuf, line_icon_indent); 
    
    if(show_inode_flag){
		strcat(buf, inode_str);
		strcat(ubuf, inode_ustr);
		show_inode_len = strlen(inode_str);
    }
    if(show_filetype_flag){
		if(show_perms_flag){
		    strcat(buf, filetype_perms_str);
	    	strcat(ubuf, filetype_perms_ustr);
			show_filetype_len = strlen(filetype_perms_str);
		} else {
		    strcat(buf, filetype_str);
	    	strcat(ubuf, filetype_ustr);
	   		show_filetype_len = strlen(filetype_str);
		}
    }
    if(show_perms_flag){
		strcat(buf, perms_str);
		strcat(ubuf, perms_ustr);
		show_perms_len = strlen(perms_str);
    }
    if(show_nlinks_flag){
		strcat(buf, nlinks_str);
		strcat(ubuf, nlinks_ustr);
		show_nlinks_len = strlen(nlinks_str);
    }
    if(show_owner_flag){
	if(show_symbolic_flag){
	    strcat(buf, owner_sym_str);
	    strcat(ubuf, owner_sym_ustr);
	    show_owner_len = strlen(owner_sym_str);
	} else {
	    strcat(buf, owner_num_str);
	    strcat(ubuf, owner_num_ustr);
	    show_owner_len = strlen(owner_num_str);
	}
    }
    if(show_group_flag){
	if(show_symbolic_flag){
	    strcat(buf, group_sym_str);
	    strcat(ubuf, group_sym_ustr);
	    show_group_len = strlen(group_sym_str);
	} else {
	    strcat(buf, group_num_str);
	    strcat(ubuf, group_num_ustr);
	    show_group_len = strlen(group_num_str);
	}
    }
    if(show_size_flag){
	strcat(buf, size_str);
	strcat(ubuf, size_ustr);
	show_size_len = strlen(size_str);
    }
    if(show_modific_flag){
	show_modific_len = 0;
	if(show_time_years_flag){
	    strcat(buf, show_time_years_str);
	    strcat(ubuf, show_time_years_ustr);
	    show_modific_len = strlen(show_time_years_str);
	}
	if(show_time_seconds_flag){
	    strcat(buf, show_time_secs_str);
	    strcat(ubuf, show_time_secs_ustr);
	    show_modific_len = strlen(show_time_secs_str);
	}
        strcat(buf, show_modific_str);
        strcat(ubuf, show_modific_ustr);
	show_modific_len += strlen(show_modific_str);
    }
    if(show_access_flag){
	show_access_len = 0;
	if(show_time_years_flag){
	    strcat(buf, show_time_years_str);
	    strcat(ubuf, show_time_years_ustr);
	    show_access_len = strlen(show_time_years_str);
	    // raw = 5 (no add.)
	}
	if(show_time_seconds_flag){
	    strcat(buf, show_time_secs_str);
	    strcat(ubuf, show_time_secs_ustr);
	    show_access_len = strlen(show_time_secs_str);
	    // raw = 3 (no add.)
	}
        strcat(buf, show_access_str);
        strcat(ubuf, show_access_ustr);
	show_access_len += strlen(show_access_str);
	// raw = 12
    }
    strcat(buf, "Name ");
    strcat(ubuf, "_    ");
    fl_redraw_object(text_box);
    fl_set_object_label(text, buf);
    fl_set_object_label(underscores, ubuf);
    // fl_set_object_label(underscores, "_A_jyg");
    // attempt an explicit redraw here. Good hack.
    fl_redraw_object(text);
    fl_redraw_object(underscores);
}   // headline

int browser_c::line_is_dir(int line, const char *fname){
    if(linecheck(line, TRUE) && show_filetype_flag){
	char dir_char = extract_type(line);
	if(dir_char == 'd'){
	    return TRUE;
	}
    }
    if(!fname) return FALSE;
    struct stat statbuf;
    if(stat(fname, &statbuf) < 0){
	char buffer[1024];
	sprintf(buffer, "line_is_dir: cannot stat file %s !", fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    } 
    return S_ISDIR(statbuf.st_mode);
}

int browser_c::linecheck(int line, int quietflag){
   if(line >= 1 && line <= fl_get_panel_maxline(panel) ){
       return TRUE;
   } else {
        if(!quietflag){
#ifdef BETA	    
	    error_msg("browser_c::linecheck: line #%d out of range.\n", line);
#endif	    
	}
	return FALSE;
    }
}

void browser_c::mirror_selection(void){
    // toggle all.
    int maxline = fl_get_panel_maxline(panel),
	ds_count = 0,
	s_count  = 0
    ;
    freeze();
    for(int i=1; i <= maxline; i++){
	if(!dot_dotdot_check(extract_filename(i))){
	    fl_deselect_panel_line(panel,i);
	} else {
	    if(fl_isselected_panel_line(panel,i)){
		fl_deselect_panel_line(panel,i);
		ds_count++;
	    } else {
		fl_select_panel_line(panel,i);
		s_count++;
	    }
	}
    }
    unfreeze();
    status_msg(1, "%d new files selected, %d deselected (makes %d total selected).", s_count, ds_count, get_n_selected());    
    update_selection_state();
}

void browser_c::new_fst_sort(sorttype_et sorttype, int reverse_flag){
    if(sorttype < access_e || sorttype > typechar_e){
	cerr << "invalid sorttype in browser.\n";
	return;
    }
    fst_sort = sorttype;
    sort_fst_reverse_flag = reverse_flag;
    sort_form_sync();
    update(TRUE);
}

void browser_c::new_snd_sort(sorttype_et sorttype, int reverse_flag){
    if(sorttype < access_e || sorttype > typechar_e){
	cerr << "invalid sorttype in browser.\n";
	return;
    }
    snd_sort = sorttype;
    sort_snd_reverse_flag = reverse_flag;
    sort_form_sync();
    update(TRUE);
}

void browser_c::print_cmd(int &argc, char *argv[]){
    nextarg(argc, argv, "-p");

    char buf[128];
    strcpy(buf, line_icon_indent); // necessary !

    if(show_inode_flag)    strcat(buf, "%i ");
    if(show_filetype_flag){
	if(show_perms_flag){
	    strcat(buf, "%t");
	} else {
	    strcat(buf, "%t ");
	}
    }
    if(show_perms_flag)    strcat(buf, "%p ");
    if(show_nlinks_flag)   strcat(buf, "%l ");
    if(show_owner_flag){
	if(show_symbolic_flag){
	    strcat(buf, "%u ");
	} else {
	    strcat(buf, "%U ");
	}
    }
    if(show_group_flag){
	if(show_symbolic_flag){
	    strcat(buf, "%g ");
	} else {
	    strcat(buf, "%G ");
	}
    }
    if(show_size_flag)    strcat(buf, "%s ");
    if(show_modific_flag){
	strcat(buf, "%m");
	add_date_string(buf);
	strcat(buf, " ");
    }
    if(show_access_flag){
	strcat(buf, "%a");
	add_date_string(buf);
	strcat(buf, " ");
    }
    strcat(buf, "%Nb");
    if(dbg_flag){
	if(dbg_flag) cerr << "buf is: " << buf << "#" << endl;
    }
    if(show_type_char_flag){
	strcat(buf, "s");
    }
    nextarg(argc, argv, buf);
}

void browser_c::redraw(void){
    if(form){
	fl_redraw_form(form);
    }
}

/*
 * ringbuffer functions.
 */

browser_c* browser_c::rb_get_last_browser(void){
    browser_c *tmp_browser = this;
    while(tmp_browser->my_ID != next_ID-1){
	//if(dbg_flag) cerr << tmp_browser->my_ID << " " << next_ID-1 << endl;
	tmp_browser = tmp_browser->next_browser;
    }
    return tmp_browser;
}

void browser_c::rb_insert_after(browser_c* new_browser){
    next_browser->prev_browser = new_browser;    // next know new prev
    new_browser->next_browser = next_browser;    // new knows old next
    new_browser->prev_browser = this;            // new knows we're prev
    next_browser = new_browser;                  // we know new next
}

void browser_c::set_source_target(int source_val, int target_val){
    freeze();
    set_source(source_val, FALSE);
    set_target(target_val, FALSE);
    unfreeze();
    set_complete_wintitle();
}

void browser_c::sync_source_target_colour(void){
    if(target_flag){
	fl_set_object_color(target_box,tool_target_colour,47);
    }
    if(source_flag){
	fl_set_object_color(source_box,tool_source_colour,47);
    }
    //fl_redraw_object(menu_text);
}

void browser_c::set_target(int boolval, int update_title_flag){
    if(boolval == target_flag) return;
    target_flag = boolval;
    if(target_flag){
	fl_set_object_color(target_box,tool_target_colour,47);
    } else {
	fl_set_object_color(target_box,0,47);
    }
    //fl_redraw_object(menu_text);
    if(update_title_flag){
	set_complete_wintitle();
    }
}

void browser_c::set_source(int boolval, int update_title_flag){
    if(boolval == source_flag){
	return;
    } else {
	source_flag = boolval;
    }
    if(source_flag){
	fl_set_object_color(source_box,tool_source_colour,47);
	//fl_redraw_object(menu_text);
	fl_unhide_panel_current(panel);
    
	if(!tool_change_dir(cwd_buf, FALSE, FALSE, TRUE, TRUE, TRUE)){
	    char buffer[1024];
	    sprintf(buffer, "Unable to change directory: %s", strerror(errno));
	    program_unstable(buffer);
	}
    } else {
	fl_set_object_color(source_box,0,47);
	//fl_redraw_object(menu_text);
	fl_hide_panel_current(panel);
    }
    if(update_title_flag){
	set_complete_wintitle();
    }
}

const char * browser_c::adjust_length(const char *status_buf){
    int continue_flag = TRUE;
    long string_width; // , x_correction_factor = 25;
    static char work_buf[MAXPATHLEN + 32];
    char *next_slash;
    
    if(!form->window){
	strcpy(work_buf, status_buf);
	return work_buf;
    }
    FL_Coord w = form->w;

    sprintf(work_buf, "%s%s", status_buf, cwd_buf);
    string_width = (long) fl_get_string_width(1, FL_NORMAL_SIZE, work_buf, strlen(work_buf));
    //if(dbg_flag) cerr << "x_width: " << w << ", string_width: " << string_width << endl;

    // idea: jump from slash to slash.
    // on initialization, it will actually prepend ... before the
    // root '/', which is a bit of a waste :-)
    
    next_slash = cwd_buf;
    while (string_width > w - 65 && continue_flag){  // 65 is a mungval. 
	if(strlen(next_slash) > 1){
	    next_slash = strchr(next_slash+1, '/');  // step through.
	    if(next_slash){
		sprintf(work_buf, "%s...%s", status_buf, next_slash);
		string_width = (long) fl_get_string_width(1, FL_NORMAL_SIZE, work_buf, strlen(work_buf));
		//if(dbg_flag) cerr << "x_width: " << w << ", string_width: " << string_width << endl;
	    } else {
		continue_flag = FALSE; // what the hell.
	    }
	} else {
	    continue_flag = FALSE; // what the hell.
	}
    }

    return work_buf;
}


void browser_c::set_filter_pattern_and_update(const char *new_pattern){
    // be intelligent enough to only update if necessary.
    if(!new_pattern || !filter_pattern) return;
    if(strcmp(new_pattern, filter_pattern) == 0) return;
    fl_free(filter_pattern);
    filter_pattern = strdup(new_pattern);
    update();
}

void browser_c::set_complete_wintitle(const char *override_str){
	// sync the attached windows:
	toolbox_attach_title();

    /* make a little title for the icons. - the cwd is too clumsy
     * this version adjusts the string to fit the window title.
     * we have a buffer for the leading status stuff, and the directory.
     */
    
    char full_buf[MAXPATHLEN + 64], status_buf[32];
    char iconbuf[16];
    
    if(!cwd_buf)return;

    char root_buf[16];
    char root_icon_buf[16];
    if(geteuid() == 0){
	strcpy(root_buf, "Root: ");
	strcpy(root_icon_buf, "R: ");
    } else {
	strcpy(root_buf, "");
	strcpy(root_icon_buf, "");
    }
    if(reading_flag){
	sprintf(full_buf, "%s#%d Reading directory ...", root_buf, my_ID);
	sprintf(iconbuf, "%s#%d: R", root_icon_buf, my_ID);
    } else {
	if(source_flag){
	    if(target_flag){
		sprintf(status_buf, long_wintitle_flag ? "%s#%d Source&Target" : "%s#%d S&T", root_buf, my_ID);
		sprintf(iconbuf, "UC %s#%d: S&T", root_icon_buf, my_ID);
	    } else {
		/* source, not target */
		sprintf(status_buf, long_wintitle_flag ? "%s#%d Source" : "%s#%d S", root_buf, my_ID);
		sprintf(iconbuf, "UC %s#%d: S", root_icon_buf, my_ID);
	    }
	} else {
	    if(target_flag){
		sprintf(status_buf, long_wintitle_flag ? "%s#%d Target" : "%s#%d T", root_buf, my_ID);
		sprintf(iconbuf, "UC %s#%d: T", root_icon_buf, my_ID);
	    } else {
		/* passive, not target */
		sprintf(status_buf, long_wintitle_flag ? "%s#%d Passive" :  "%s#%d P", root_buf, my_ID);
		sprintf(iconbuf, "UC %s#%d: P", root_icon_buf, my_ID);
	    }
	}
	// append host string (can be just a ":")
	strcat(status_buf, tool_host_string);

        /* check for possible overflow, and shorten buffer to fit
	 * the window title bar. Plug in cwd.
	 */
	strcpy(full_buf, adjust_length(status_buf));
    }    
    if(strcmp(filter_pattern, "*") != 0){
	strcat(full_buf, ": ");
	if(filter_full_line_flag)strcat(full_buf, "[");
	strcat(full_buf, filter_pattern);
	if(filter_full_line_flag)strcat(full_buf, "]");
    }

    int n_lines = fl_get_panel_maxline(panel);
    int n_selected = get_n_selected();

    char aux_buf[64];    
    llong_t total_bytes = get_total_byte_size();
    if(n_selected>0){
	if(show_size_flag){
	    llong_t bytes = get_selection_byte_size();
	    if(long_wintitle_flag){
		sprintf(aux_buf, " (%d/%d %s, ", n_selected, n_lines, (n_lines == 1 ? "file" : "files"));
	    } else {
		sprintf(aux_buf, " (%d/%d, ", n_selected, n_lines);
	    }
	    strcat(aux_buf, bytes_to_string(bytes));
	    strcat(aux_buf, "/");
	    strcat(aux_buf, bytes_to_string(total_bytes));
	    strcat(aux_buf, ")");
	} else {
	    sprintf(aux_buf, " (%d/%d)", n_selected, n_lines);
	}
    } else {
	if(show_size_flag){
	    if(long_wintitle_flag){
		sprintf(aux_buf, " (%d %s, %s)", n_lines, (n_lines == 1 ? "file" : "files"), bytes_to_string(total_bytes));
	    } else {
		sprintf(aux_buf, " (%d, %s)", n_lines, bytes_to_string(total_bytes));
	    }
	} else {
	    sprintf(aux_buf, " (%d)", n_lines);
	}
    }
    strcat(full_buf, aux_buf);

    /* we try to be efficient here. The check is very necessary.
     * e.g., a browser can be set source/passive without being shown.
     */
    if(form->window){
	if (override_str) strcpy(full_buf, override_str);
	if (last_wintitle && (strcmp(last_wintitle, full_buf) == 0)) return;
	    
	XStoreName(fl_display, form->window, full_buf);
	XSetIconName(fl_display, form->window, iconbuf);		   

	if (last_wintitle) fl_free(last_wintitle);
	last_wintitle = strdup(full_buf);
    }
}

void browser_c::win_raise(void){
    if(form->window && form->visible){
	XMapRaised(fl_display, form->window);
	// no set input focus. that's extra !
    }
}

void browser_c::set_current(int line, int auto_center_flag){
    if(dbg_flag){
	//if(dbg_flag) cerr << "line " << line << endl;
    }
    if(!linecheck(line)) return;
    freeze();
    int oldtopline = fl_get_panel_topline(panel);
    fl_set_panel_current(panel, line);
    int topline = fl_get_panel_topline(panel);
    int screen_lines = fl_get_panel_screenlines(panel);
    if((auto_center_flag && oldtopline != topline) || 
       line < topline ||
       line >=topline + screen_lines
       ){
	center_current();
    }
    unfreeze();
}

void browser_c::center_current(void){
    fl_center_panel_current(panel);
}

/*
 * first try wether we can restore the defaults, else just go
 * for the vanilla size. Important: unconstrain anyway.
 * Note that we can avoid a load attempt by ensuring checked
 * defaults, signalled via the function argument.
 * this is useful for cloning, when there are no entries
 * in the database (or actually undesired entries (!)).
 */

static int browser_atclose_cb(FL_FORM* form, void*){
    tool_set_source(form);
    if(dbg_flag) cerr << "brwser winquit request." << endl;
    BrowserKillButtonCB(NULL, 0);    
    return FL_IGNORE;
}

void browser_c::show(int defaults_checked_flag){
    M_warn("browser::show", "my_ID = %d.", my_ID);
    if (!form) {
	error_exit("Cannot show a browser with a NULL form.\n");
    }
    int have_defaults_flag = FALSE;
    /* non-fdesign extras: (precaution) */
    if(defaults_checked_flag){
	have_defaults_flag = TRUE;
    } else {
	have_defaults_flag = defaults_load(NULL);
    }

    if(!have_defaults_flag && my_ID < 2){
	have_defaults_flag = TRUE;
    }
    fl_register_call_back(form, FL_ALL_EVENT, browser_raw_cb);
    M_info("browser_c::show", "x %d,y %d,w %d, h %d", win_x, win_y, win_w, win_h);
    fl_set_initial_placement(form, win_x, win_y, win_w, win_h);
    rescale(TRUE);
    fl_set_form_minsize(form, 40, 40);
    fl_set_form_maxsize(form, tool_max_ww, tool_max_wh);
    
    fl_set_form_atclose(form, browser_atclose_cb, NULL);
    fl_show_form(form, FL_PLACE_FREE, TRUE, "UC Browser:");

    static Pixmap pixmap =0;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, form->window, browser_bits, browser_width,
	    browser_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    if (pixmap > 0) {
	set_window_icon(form->window, pixmap);
    } else {
	cerr << "Failed to create pixmap for browser.\n";
    }

    // ICCCM
    char *nargv[2];
    nargv[0] = "";    // HP sucks dead bunnies through a bent straw.
    nargv[1] = NULL;
    XSetCommand(fl_display, form->window, nargv, 1);

    // read data.
    update();
}

void browser_c::sort_cmd(int &argc, char *argv[]){
    char sortbuf[64];
    strcpy(sortbuf, ""); // necessary !
    nextarg(argc, argv, "-s");
    sort_process_option(sortbuf, fst_sort, sort_fst_reverse_flag);
    if(fst_sort != snd_sort){
        sort_process_option(sortbuf, snd_sort, sort_snd_reverse_flag);
    }
    if(dbg_flag){
        //cerr << "cmp: " << strcmp(sortbuf, "%-t%nb") << " string: " << sortbuf << "#" << endl;
    }
    nextarg(argc, argv, sortbuf);    
}

void browser_c::sort_process_option(
    char *buf,
    sorttype_et sorttype,
    int reverse_flag)
{
    strcat(buf, "%");
    if(reverse_flag)strcat(buf, "-");
    switch(sorttype){
    case access_e	: strcat(buf, "a"); break;
    case filetype_e     : strcat(buf, "t"); break;
    case group_e 	:if(show_symbolic_flag){strcat(buf, "g"); } else {strcat(buf, "G");} break;
    case inode_e 	: strcat(buf, "i"); break;
    case modific_e	: strcat(buf, "m"); break;
    case name_e	        : strcat(buf, "nb"); break;
    case nlinks_e	: strcat(buf, "l"); break;
    case owner_e	: if(show_symbolic_flag){strcat(buf, "u"); } else {strcat(buf, "U");} break;
	// case perms_e        : strcat(buf, "p"); break; // does not work :-( 
    case size_e	        : strcat(buf, "s"); break;
    case extension_e    : strcat(buf, "x"); break;
    default             : 
	error_msg("UC internal: browser_c:: unknown sort option."); 
	cerr << "option value: " << sorttype << ", defaulting to name." << endl;
	strcat(buf, "nb"); 
	break;
    }
}

static FL_OBJECT *puts_panel = NULL;
static char *puts_filter_pattern = NULL;
static int puts_filter_full_line_flag = FALSE;

void panel_puts(char *string, SFILESTAT *sfstat_p){
    if(!string || !sfstat_p){
	program_unstable("Invalid string or stat in puts.\n");
	return;
    }
    int do_match_flag = TRUE;
    if(puts_panel && string){
	// we try to avoid many pattern tests;
	if(puts_filter_pattern){
	    if(puts_filter_pattern[0] == '*'){
		if(puts_filter_pattern[1] == '\0'){
		    do_match_flag = FALSE;
		}   // else longer expr.
	    }   // else different expr. 
	} else {
	    do_match_flag = FALSE;
	}
	const char *fbase_name = get_basename(sfstat_p->f_name);
	if(do_match_flag){
	    if(puts_filter_full_line_flag){
		if(!wildcard_match(puts_filter_pattern, string)) return;
	    } else {
		if(!wildcard_match(puts_filter_pattern, fbase_name)) return;
		// if(!wildcard_match(puts_filter_pattern, extract_filename(string))) return;
	    }
	}
	int line = fl_add_panel_line(puts_panel, string);
	SFILESTAT *client_data = (SFILESTAT *) fl_get_panel_line_client_data(puts_panel, line);
	if(!client_data){
	    client_data = (SFILESTAT *) fl_malloc(sizeof(SFILESTAT));
	    client_data->f_name = NULL;
	    fl_set_panel_line_client_data(puts_panel, line, (void *) client_data);
	}
	client_data->f_stat = sfstat_p->f_stat;  // copies entire struct.

	if(client_data->f_name){
	    fl_free(client_data->f_name);
	    client_data->f_name = NULL;
	}
	client_data->f_name = strdup(fbase_name);
	
	if(dbg_flag){
	    cerr << string << endl;
	}
    } else {
	error_msg("no current browser or empty string.");
    }
}

void browser_c::update(int keep_selection_current_flag){
    default_actions();
    char *tmp_current_fname = NULL;
    int i, topline = 0;
    linklist_c selection_list;
    reading_flag = TRUE;
    fl_set_cursor_color(XC_watch, FL_RED, FL_WHITE);
    fl_set_global_cursor(XC_watch);
    set_complete_wintitle();

    // XFlush(fl_display);
    // would force setting, which I usually don\'t need. 
    /*
     * remember selection 
     */
    if(keep_selection_current_flag){
	
	const char *tmp_fname = get_current_fname();
	topline = fl_get_panel_topline(panel);
	if(tmp_fname){
	    tmp_current_fname = strdup(tmp_fname);
	}
	int maxline = fl_get_panel_maxline(panel);
	for(i=1; i <= maxline; i++){
	    if(fl_isselected_panel_line(panel,i)){
		char *buf = strdup(extract_filename(i));
		selection_list.add((void *) buf);
		//if(dbg_flag) cerr << "adding " << buf << " to selection list." << endl;
	    }
	}
    } else {
        // reset this on change dir etc.
        cur_up_flag = FALSE;
    }
    
    // important to freeze the form.
    freeze();
    fl_clear_panel(panel);

    // kludges for the callback function passed sls_main.
    puts_panel = panel;
    puts_filter_pattern = filter_pattern;
    puts_filter_full_line_flag = filter_full_line_flag;
    
    /* Step 0: assemble a command line
     * malloc each string. We always malloc/free a command line from scratch.
     * Pre-allocing that array was to buggy.
     */
    char *argv[50];


    int argc = 0;
    nextarg(argc, argv, "dirread");
    if(list_hidden_flag){
	nextarg(argc, argv, "-a");
    }
    if(follow_links_flag){
	nextarg(argc, argv, "-L");  
    }
    sort_cmd(argc, argv);
    print_cmd(argc, argv);

    /* on SGI, stat doesn't like empty strings, so always
     * give it the complete directory.
     */

    nextarg(argc, argv, cwd_buf);


    if(dbg_flag){
	if(dbg_flag) cerr << "update_main string: " << endl;
	for(i=0; i<argc; i++){
	    if(dbg_flag) cerr << "argv[" << i << "] = " << argv[i] << "#" << endl;
	}
    }

    /*
     * Step 1: execute.
     */

    sls_main(argc, argv, panel_puts);
    set_current(1);


    // make shure nobody else messes up our data. 
    puts_panel = NULL;
    
    // go to top line (default) - this should change, if 
    // we have a preferred line.
    headline();

    /*
     * restore selection (remember to free data structures)
     * idea: go linearly through the browser. For each extracted
     * file name, search. If found, select it and remove list node.
     */

    if(keep_selection_current_flag){
	int maxline = fl_get_panel_maxline(panel);
	int found_flag, list_index, list_size;
	const char *fname_buf;
	for(i=1; i <= maxline; i++){
	    found_flag = FALSE;
	    fname_buf = extract_filename(i);
	    selection_list.headend();
	    list_index = 0;
	    list_size = selection_list.getsize();
	    while(!found_flag && (list_index < list_size)){
		char *selection_fname_buf = (char *)selection_list.getval();
		if(selection_fname_buf){
		    if(strcmp(selection_fname_buf, fname_buf) == 0){
			found_flag = TRUE;
			//if(dbg_flag) cerr << "found " << fname_buf << " in selection list." << endl;
			fl_select_panel_line(panel,i);
			fl_free(selection_fname_buf);
			selection_fname_buf = NULL;
			selection_list.remove();
		    }
		}
		list_index++;
		if(!found_flag && (list_index < list_size)){
		    selection_list.step();
		    /*
		     * note the 'not found'. If we remove, were are we ?
		     * well, we would be one step forward, except if we kill
		     * the last node.
		     */
		}
	    }
	}
	/*
	 * free list data. (rest handled by list destructor)
	 */
	list_size = selection_list.getsize();
	for(i=0; i<list_size; i++){
	    char *buf = (char *)selection_list.getval();
	    if(buf){
		fl_free(buf);
		buf = NULL;
	    } else {
		//cerr << "freeing null ptr." << endl;
	    }
	    selection_list.step();	    
	}

	fl_set_panel_topline(panel, topline);
	if(tmp_current_fname){
	    set_current_fname(tmp_current_fname, TRUE);
	    fl_free(tmp_current_fname);
	}
    }
    unfreeze();

    reading_flag = FALSE;
    fl_reset_global_cursor();
    set_complete_wintitle();
    for(i=0; i<argc; i++){
	fl_free(argv[i]);
    }
    time_stamp();
}

/*
 * callback look-alikes.
 */

void browser_c::home(void){
    default_actions();
    set_current(1);
    cur_up_flag = FALSE;
}

void browser_c::end(void){
    default_actions();
    set_current(fl_get_panel_maxline(panel));
    cur_up_flag = TRUE;
}

void browser_c::freeze(void){
    fl_freeze_form(form);
}

void browser_c::unfreeze(int full_redraw){
    fl_unfreeze_form(form);
    if(full_redraw){
	fl_redraw_form(form);
    }
}

void browser_c::down_arrow(int n_lines){
    default_actions();
    cur_up_flag = FALSE;
    if(n_lines <= 0) return;
    int newline = fl_get_panel_current(panel) + n_lines;
    set_current_with_margin(newline);
}

void browser_c::up_arrow(int n_lines){
    default_actions();
    if(n_lines <= 0) return;
    cur_up_flag = TRUE;
    int newline = fl_get_panel_current(panel) - n_lines;
    set_current_with_margin(newline);
}

void browser_c::set_current_with_margin(int newline){
    // this new version maintains the geometry in the lower & upper 2-line borders.

    int maxline, screenlines, topline;
    
    maxline      = fl_get_panel_maxline(panel);
    screenlines  = fl_get_panel_screenlines(panel);
    topline      = fl_get_panel_topline(panel);

    if(newline < 1){
	home();
	return;
    }

    if(newline > maxline){
	end();
	return;
    } 

    if(newline >= topline + 2 && newline < topline + screenlines - 2){
	// normal case, no scroll, no border, just move it.
	fl_set_panel_current(panel, newline);
	return;
    }
    
    if(newline <= 2 || newline > maxline - 2){
	// special case, we're at the start or end of the list and have no margin.
	set_current(newline);
	//fl_set_panel_current(panel, newline);
	return;
    }

    if(newline < topline+2){
	// going up.
	freeze();
	fl_set_panel_topline(panel, newline-2);
	fl_set_panel_current(panel, newline);
	unfreeze();
	return;
    }
    if(newline >= topline+screenlines-2){
	// going up.
	freeze();
	fl_set_panel_topline(panel, newline-(screenlines-3));
	fl_set_panel_current(panel, newline);
	unfreeze();
	return;
    }
}

void browser_c::page_down(void){
    default_actions();
    // attempt to maintain the _full_ geometry, including the geometric
    // position of the current bar. It scrolls by n-gap lines, so that
    // the last visible line becomes the first visible one after pagedown.

    int current_line, maxline, newline, new_topline, screenlines, topline;
    
    current_line = fl_get_panel_current(panel);
    maxline      = fl_get_panel_maxline(panel);
    screenlines  = fl_get_panel_screenlines(panel);
    topline      = fl_get_panel_topline(panel);
    newline      = current_line + screenlines - 2;

    int fudge_diff = current_line-topline;
    if(fudge_diff >= 0 && fudge_diff < 2){
	// special case when the current line was _in_ the top margin.
	newline += 2-fudge_diff;
    } else {
	fudge_diff = topline + screenlines -1 - current_line;
	if(fudge_diff >= 0 && fudge_diff < 2){
	    // special case when the current line was _in_ the bottom margin.
	    newline -= 2-fudge_diff;
	}
    }

    if(newline > maxline){
	end();
	return;
    } 

    if(newline <= current_line){
	// nasty artefact possible with small screens.
	down_arrow(1);
	return; 
    }
    cur_up_flag = FALSE;

    new_topline  = topline + screenlines - 2;
    if(new_topline < topline){
	// artefact
	new_topline = topline;
    } else if(new_topline > maxline-screenlines+1){
	new_topline = maxline-screenlines+1;
    }

    freeze();
    fl_set_panel_topline    (panel, new_topline);
    fl_set_panel_current(panel, newline);
    unfreeze();
}

void browser_c::page_up(void){
    default_actions();
    // attempt to maintain the _full_ geometry, including the geometric
    // position of the current bar. It scrolls by n-gap lines, so that
    // the last visible line becomes the first visible one after pagedown.

    int current_line, newline, new_topline, screenlines, topline;
    
    current_line = fl_get_panel_current(panel);
    screenlines  = fl_get_panel_screenlines(panel);
    topline      = fl_get_panel_topline(panel);
    newline      = current_line - screenlines + 2;

    int fudge_diff = current_line-topline;
    if(fudge_diff >= 0 && fudge_diff < 2){
	// special case when the current line was _in_ the top margin.
	newline += 2-fudge_diff;
    } else {
	fudge_diff = topline + screenlines -1 - current_line;
	if(fudge_diff >= 0 && fudge_diff < 2){
	    // special case when the current line was _in_ the bottom margin.
	    newline -= 2-fudge_diff;
	}
    }

    if(newline < 1){
	home();
	return;
    } 

    if(newline >= current_line){
	// nasty artefact possible with small screens.
	up_arrow(1);
	return; 
    }
    cur_up_flag = TRUE;

    new_topline  = topline - screenlines + 2;
    if(new_topline > topline){
	// artefact
	new_topline = topline;
    } else if(new_topline < 1){
	new_topline = 1;
    }

    freeze();
    fl_set_panel_topline    (panel, new_topline);
    fl_set_panel_current(panel, newline);
    unfreeze();
}

void browser_c::rescale(int force_flag){
    /* do some reasonable error checking, eg. wether the whole
     * baggage is already initialized.
     */
    if(dbg_flag){
	cerr << "rescale..." << endl;
    }

    if(form->w == win_w && form->h == win_h && !force_flag) return;
    if(dbg_flag)cerr << "new browser size\n";
    win_w = form->w;
    win_h = form->h;
    
    /* Now, rescale the bugger. 
     * Set width & height, fix position
     * box_height is the dividing point, the height of the top status strip.
     */

//    box_height = fl_get_panel_fontsize(panel) + 3;
    int asc, desc;
    box_height = fl_get_char_height(
        fl_get_panel_fontstyle(panel),
        fl_get_panel_fontsize(panel),
        &asc, &desc
    );
    box_height += desc + 2;  // fudge factor may change.
    //cerr << "a,d:" << asc << "," << desc << endl;


    /* we don't need to find the box, we have it's browser.
     * hehehehe :-) We can fix it straight away.
     */

    source_box->h = box_height;
    source_box->w = box_width;
    source_box->x = 0;
//    source_box->y = y_magic;
    source_box->y = 0;    

    target_box->h = box_height;    // same
    target_box->w = box_width; // same, + fudge.
    target_box->x = box_width;     // we went one box 2 the right.
    target_box->y = 0;       // same

    panel->h = win_h - box_height;
    panel->y = box_height;
    panel->x = 0;
    panel->w = win_w;

    text->h = box_height-text_h_fudge;          // fix height.
//    text->x = 2 * box_width+1;   // two boxes to the right, + fudge.
    text->x = 2 * box_width+text_x_fudge;   // two boxes to the right + fudge.
    text->y = 0;	        // same

    text_box->h = box_height;        // fix height.
    text_box->x = 2 * box_width;   // two boxes to the right, + fudge.
    text_box->y = 0;	               // same
    text_box->w = win_w-text_box->x;

    //underscores->h = text->h;
    underscores->h = box_height;   // better without the fudge, else the underscores are shown incorrectly.
    underscores->x = text->x;
    underscores->y = text->y;

    column_hidden_button->h = box_height;
    column_hidden_button->x = 2*box_width;
    column_hidden_button->y = 0;

/*
    menu_text->h = box_height;
    menu_text->w = 2*box_width-1;  // fudge.
    menu_text->x = 1;
    menu_text->y = 0;
*/

    menu_hidden_button->h = box_height;
    menu_hidden_button->w = 2*box_width;
    menu_hidden_button->x = 0;
    menu_hidden_button->y = 0;    

    const int search_text_width = 85;
    int search_text_x = form->w - search_text_width;
    if(search_text_x < 0){
	search_text_x = 0;
    }
    search_text->h = box_height;
    search_text->w = search_text_width;
    search_text->x = search_text_x;
    search_text->y = 0;

    // do not raise after rescaling. 
    set_complete_wintitle();  
}

void browser_c::sync_colour_opts (browser_c *browser){
    if(!browser) return;
    // init the colours.
    int SelCurLabelCol, SelCurBackCol, SelLabelCol, SelBackCol, 
        CurLabelCol, CurBackCol, LabelCol, BackCol
    ;
    fl_get_panel_colors(browser->panel, 
	&SelCurLabelCol, &SelCurBackCol, &SelLabelCol, &SelBackCol,
	&CurLabelCol, &CurBackCol, &LabelCol, &BackCol
    );
    fl_set_panel_colors(panel, 
	SelCurLabelCol, SelCurBackCol, SelLabelCol, SelBackCol,
	CurLabelCol, CurBackCol, LabelCol, BackCol
    );
}

void browser_c::sync_display_opts(browser_c *browser){
    if(!browser) return;
    show_inode_flag        = browser->show_inode_flag;
    show_filetype_flag     = browser->show_filetype_flag;
    show_perms_flag        = browser->show_perms_flag;
    show_nlinks_flag       = browser->show_nlinks_flag;
    show_owner_flag        = browser->show_owner_flag;
    show_symbolic_flag     = browser->show_symbolic_flag;
    show_group_flag        = browser->show_group_flag;
    show_size_flag         = browser->show_size_flag;
    show_modific_flag      = browser->show_modific_flag;
    show_access_flag       = browser->show_access_flag;
    show_time_seconds_flag = browser->show_time_seconds_flag;
    show_time_years_flag   = browser->show_time_years_flag;

    follow_links_flag      = browser->follow_links_flag;
    long_wintitle_flag     = browser->long_wintitle_flag;
    list_hidden_flag       = browser->list_hidden_flag;

    set_fontsize (fl_get_panel_fontsize (browser->panel));
    set_fontstyle(fl_get_panel_fontstyle(browser->panel));
}

void browser_c::sync_sort_opts(browser_c *browser){
    if(!browser) return;
    fst_sort = browser->fst_sort;
    snd_sort = browser->snd_sort;
    sort_fst_reverse_flag = browser->sort_fst_reverse_flag;
    sort_snd_reverse_flag = browser->sort_snd_reverse_flag;
}

/* return success */
int browser_c::clone(void){
    browser_c* new_browser, *old_last_browser;

    old_last_browser = rb_get_last_browser();
    new_browser = new browser_c();

    if(!new_browser){
	program_unstable("UC Internal: Unable to clone browser.");
	return FALSE;
    }

    /*
     * insert in ring, after the last browser.
     */

    old_last_browser->rb_insert_after(new_browser);
    
    // data that need copying.
    new_browser->sync_colour_opts(this);
    new_browser->sync_display_opts(this);
    new_browser->sync_sort_opts(this);

    /*
     * problematic issues. check wether constructor defaults are good.
     * dismiss ID's. They should be ok.
     */

    // char *cwd_buf;
    //int source_flag;
    //int target_flag;

    /*
     * init the window prefs, then update. Should be okay, we do that
     * after reading in the defaults database.
     */
    new_browser->win_w = form->w;
    new_browser->win_h = form->h;
    new_browser->win_x = form->x + 50;
    if(form->y + 50 < fl_scrh){
        new_browser->win_y = form->y + 50;
    } else {
        new_browser->win_y = fl_scrh;
    }
    //new_browser->change_dir(cwd_buf);
    //error_msg("not implemented: setting cloned browser to cwdbuf.");
    // works by default, via the '.' mechanism. good.
    
    new_browser->show(TRUE);
    return TRUE;
}

void browser_c::set_dark_yellow(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_YELLOW; 
    CSB = FL_COL1;
    CNF = FL_WHITE; 
    CNB = FL_COL1;
    NSF = FL_YELLOW;
    NSB = FL_BLACK;
    NNF = FL_WHITE;
    NNB = FL_BLACK;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_dark_yellow_bright(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_YELLOW; 
    CSB = FL_DEEPPINK;
    CNF = FL_WHITE; 
    CNB = FL_DOGERBLUE;
    NSF = FL_YELLOW;
    NSB = FL_BLACK;
    NNF = FL_WHITE;
    NNB = FL_BLACK;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_dark_red(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_RED; 
    CSB = FL_COL1;
    CNF = FL_WHITE; 
    CNB = FL_COL1;
    NSF = FL_RED;
    NSB = FL_BLACK;
    NNF = FL_WHITE;
    NNB = FL_BLACK;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_dark_green(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_GREEN; 
    CSB = FL_COL1;
    CNF = FL_WHITE; 
    CNB = FL_COL1;
    NSF = FL_GREEN;
    NSB = FL_BLACK;
    NNF = FL_WHITE;
    NNB = FL_BLACK;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_greyish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_YELLOW;
    CSB = FL_BOTTOM_BCOL;
    CNF = FL_BLACK; 
    CNB = CSB;
    NSF = CSF;
    NSB = FL_COL1;    // grey
    NNF = CNF;
    NNB = NSB;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_blueish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_YELLOW; 
    CSB = FL_SLATEBLUE;
    CNF = FL_WHITE; 
    CNB = CSB;
    NSF = CSF;
    NSB = FL_BLUE;
    NNF = FL_WHITE;
    NNB = FL_BLUE;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_reddish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_YELLOW; 
    CSB = FL_COL1;
    CNF = FL_WHITE; 
    CNB = FL_COL1;
    NSF = FL_YELLOW;
    NSB = FL_RED;
    NNF = FL_WHITE;
    NNB = FL_RED;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_greenish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_GREEN; 
    CSB = FL_BLACK;
    CNF = FL_WHITE; 
    CNB = FL_BLACK;
    NSF = FL_GREEN;
    NSB = FL_SLATEBLUE;  // was: 57
    NNF = FL_WHITE;
    NNB = FL_SLATEBLUE;  // was: 57
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_yellowish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_RED;
    CSB = FL_GREEN;
    CNF = FL_BLACK;
    CNB = FL_SLATEBLUE;
    NSF = CSF;
    NSB = FL_YELLOW;
    NNF = CNF;
    NNB = NSB;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::set_whitish(void){
    int CSF, CSB, NSF, NSB, CNF, CNB, NNF, NNB;
    CSF = FL_RED; 
    CSB = FL_BOTTOM_BCOL;
    CNF = FL_BLACK; 
    CNB = CSB;
    NSF = FL_RED;
    NSB = FL_WHITE;
    NNF = FL_BLACK;
    NNB = FL_WHITE;
    fl_set_panel_colors(panel,
        CSF,CSB,NSF,NSB,CNF,CNB,NNF,NNB
    );				     
}

void browser_c::forward_selection(void){
    if(dbg_flag) cerr << "forward selection called.\n";
    if(!xshell_visible_flag) return;

    char prefix_buf[MAXPATHLEN];
    if(this == source_browser){
	strcpy(prefix_buf, "");
    } else {
	strcpy(prefix_buf, get_cwd());
	strcat(prefix_buf, "/");
    }	
    const char *name_buf;
    int len        = 0,
	prefix_len = strlen(prefix_buf),
	max_len    = 232
    ;
    if(get_n_selected() <= 0){
	name_buf = get_current_fname();
	if(name_buf){
	    if(prefix_len) xshell_string(prefix_buf);
	    xshell_string(name_buf);
	    xshell_needs_enter_flag = TRUE;
	}
    } else {
	name_buf = get_first_selection();
	if(name_buf){
	    xshell_needs_enter_flag = TRUE;
	}
	while(name_buf){
	    len += strlen(name_buf) + prefix_len;
	    if(len < max_len){
		if(prefix_len) xshell_string(prefix_buf);
		xshell_string(name_buf);
		deselect_file(name_buf);
		name_buf = get_next_selection();
	    } else {
		name_buf = NULL;
		status_msg(1, "Selection partially transferred.");
		static int first = TRUE;
		if(first){
		    first = FALSE;
		}
		timed_notifier("Line too long: Selection was transferred only partially.",
			       notifier_secs
			       );
	    }
	    // flush on last :-)
	    xshell_key(' ', FALSE, (name_buf == NULL), TRUE);
	}
    }
}

int browser_c::key_event_callback(char key, KeySym keysym){
    if(IsModifierKey(keysym)){
	return FALSE;
    }
    if(search_key_event_callback(key, keysym)){
	return TRUE;
    }
    if(key == 10 || key == 13){
	if((xshell_visible_flag && xshell_needs_enter_flag) || tool_get_ctrl()){
	    return FALSE;
	} else {
	    NewEnterKeyCB(NULL, 0);
	    return TRUE;
	}
    }
    if(key <= 26){
	switch(key){
	case 9:   /* if(dbg_flag) cerr << "killing tab.\n"; */ return TRUE;
	default:  break;
	}

	// massage the key.
	key += 'a' - 1;
	//if(dbg_flag) cerr << "ctrl detected, fixing to " << (int) key << endl;
	int retval;
	tool_disable_ctrl();
	switch(key){
	case 'a': 
	    // simulate_click(200, 200);
	    HomeKeyCB      (NULL, TRUE);
	    retval = TRUE; 
	    break;
	case 'b': LeftArrowKeyCB (NULL, 0);                  retval = TRUE; break;
	case 'e': EndKeyCB       (NULL, 0);                  retval = TRUE; break;
	case 'f': RightArrowKeyCB(NULL, 0);                  retval = TRUE; break;
	case 'l': center_current_event_cb(NULL,0);           retval = TRUE; break;
	//case 'L': fl_qenter(CENTER_SELECTED_CMSG, 1);    retval = TRUE; break;
	case 'n': DownArrowKeyCB (NULL, 0);                  retval = TRUE; break;
	case 'p': UpArrowKeyCB   (NULL, 0);                  retval = TRUE; break;
	case 's': source_browser->search_event_cb();         retval = TRUE; break;
	case 't': SetTargetCB    (NULL, 0);                  retval = TRUE; break;
	case 'v': PageDownKeyCB  (NULL, 0);                  retval = TRUE; break;
	default :	                                     retval = FALSE;break;
	}
	tool_enable_ctrl();
	return retval;
    }
    if(tool_get_alt() || tool_get_esc()){
	if(dbg_flag) cerr << "alt detected, working on " << (int) key << endl;
	int retval = FALSE;
	tool_disable_meta();
	switch(key){
	case '<': HomeKeyCB   (NULL, TRUE); retval = TRUE; break;
	case '>': EndKeyCB    (NULL, 0); retval = TRUE; break;
	case 'v': PageUpKeyCB (NULL, 0); retval = TRUE; break;
	case 'F': target_browser->forward_selection();      retval = TRUE; break;
	case 'f': forward_selection();      retval = TRUE; break;
	case 'a':
	case 'A': 
	    move_focus_to_column(access_e);
	    browser_access_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'd':
	case 'D': dir_menu_cb(); retval = TRUE; break;
	case 'g':
	case 'G': 
	    move_focus_to_column(group_e);
	    browser_group_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'i':
	case 'I': 
	    move_focus_to_column(inode_e);
	    browser_inode_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'k':
	case 'K':
	    browser_make_menu_cb(NULL, 0); retval = TRUE; 
	    break;
	case 'l':
	case 'L': 
	    move_focus_to_column(nlinks_e);
	    browser_nlinks_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'm':
	case 'M': 
	    move_focus_to_column(modific_e);
	    browser_modific_menu_cb(NULL, 0); 
	    retval = TRUE; 	
	    break;
	case 'n':
	case 'N': 
	    move_focus_to_column(name_e);
	    browser_name_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'o':
	case 'O': 
	    move_focus_to_column(owner_e);
	    browser_owner_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'p':
	case 'P': 
	    move_focus_to_column(perms_e);
	    browser_perms_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'q':
	case 'Q': 
	    BrowserKillButtonCB(NULL, 0);              
	    retval = TRUE; 
	    break;
	case 't':
	case 'T': 
	    move_focus_to_column(filetype_e);
	    browser_filetype_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 's':
	case 'S': 
	    move_focus_to_column(size_e);
	    browser_size_menu_cb(NULL, 0); 
	    retval = TRUE; 
	    break;
	case 'w':
	case 'W': windows_menu_cb(); retval = TRUE; break; 
	default:  retval = FALSE; break;
	}
	tool_enable_meta();
	return retval;
    }
    return FALSE;
}


void browser_c::default_actions(void){
    if(search_text->visible){
	fl_hide_object(search_text);
    }
}

int browser_c::search_again_event_cb(void){
    if(dbg_flag) cerr << "search again event_cb\n";
    if(!search_text->visible) return FALSE;
    char *label_buf = search_text->label, keybuf[NAME_MAX+1];
    if(!label_buf) return FALSE;
    if(strlen(label_buf) == 0) return FALSE;

    strcpy(keybuf, label_buf);

    if(set_current_below_match(keybuf, search_again_line)){
	status_msg(1, "`%s' matches file `%s'.", keybuf, get_current_fname());
	fl_set_object_label(search_text, keybuf);
	search_again_line = fl_get_panel_current(panel);
	search_wrap_possible_flag = FALSE;
	strcpy(search_last_fname_buf, keybuf);
    } else {
	XBell(fl_display, 0);
	status_msg(1, "no new match for `%s'.", keybuf);
	search_wrap_possible_flag = TRUE;
    }
    return TRUE;
}

void browser_c::search_event_cb(void){
    if(dbg_flag) cerr << "search event_cb\n";
    if(search_text->visible){
	if(strlen(search_text->label) <= 0){
	    search_again_line = 0;
	    // research for previous search text only if valid.
	    if(strlen(search_last_fname_buf) > 0){
		fl_set_object_label(search_text, search_last_fname_buf);
	    }
	} else {
	    if(search_again_line > 0 && search_wrap_possible_flag){
		// found it at least once, so we can wrap.
		search_again_line = 0;
	    }
	}
	source_browser->search_again_event_cb();
	return;
    }
    search_again_line = 0;
    search_wrap_possible_flag = FALSE;
    search_last_line = fl_get_panel_current(panel);
    fl_set_object_label(search_text, "");
    fl_show_object(search_text);
}

int browser_c::search_key_event_callback(char key, KeySym keysym){
    if(IsModifierKey(keysym)){
	return FALSE;
    }
    char keybuf[NAME_MAX+4];
    if(!search_text->visible){
	return FALSE;
    }
    if(key == '\033'){
	// escape.
	//freeze();
	fl_hide_object(search_text);
	if(search_last_line > 0){
	    set_current(search_last_line, TRUE);
	}
	//unfreeze();
	return TRUE;
    }

    if(key == 10 || key == 13){
	fl_hide_object(search_text);
	fl_set_object_label(search_text, "");
	return TRUE;
    }
    if( !(isalnum(key) || ispunct(key)) && key != 8){
	// only want normal stuff here. (and no backspace)
	return FALSE;
    }
    
    char *label_buf = search_text->label;
    if(!label_buf) return FALSE;
    int len = strlen(label_buf);
    if(len == 0){
	if(key == 8) return TRUE;
	sprintf(keybuf, "%c*", key);
    } else {
	if(key == 8){
	    if(len <= 1){
		return TRUE;
	    }
	    label_buf[len-2] = '\0';
	} else {
	    // eliminate the trailing '*' in the label. 
	    // overwrite it with the new character, and append a '*'. 
	    label_buf[len-1] = key;
	}
	sprintf(keybuf, "%s*", label_buf);
    }
    if(set_current_first_match(keybuf)){
	status_msg(1, "`%s' matches file `%s'.", keybuf, get_current_fname());
	fl_set_object_label(search_text, keybuf);
	search_again_line = fl_get_panel_current(panel);
	search_wrap_possible_flag = FALSE;
	strcpy(search_last_fname_buf, keybuf);	
    } else {
	XBell(fl_display, 0);
	status_msg(1, "no match for `%s'.", keybuf);
	search_wrap_possible_flag = TRUE;
    }
    return TRUE;
}

static void column_hidden_button_cb(FL_OBJECT *column_hidden_button, long val){
    if(dbg_flag) cerr << "column hidden button.\n";
    tool_set_source(column_hidden_button->form);
    if(!column_hidden_button->belowmouse){
	cerr << "column not below mouse ?\n";
	return;
    }
    source_browser->column_hidden_button_cb(column_hidden_button, val);
}

FL_Coord browser_c::compute_column_x_offset(sorttype_et which){
    int blank_w = fl_get_string_width(text->lstyle, text->lsize, " ", 1);
    int blank_col = strlen(line_icon_indent); // what a kludge. Bah. 
    blank_col++;  // even worse! Duh.

    // this code is slightly wrong if the which-column is hidden.
    if(show_inode_flag)		blank_col += show_inode_len;
    if(which == inode_e) 	return blank_col * blank_w;

    if(show_filetype_flag)	blank_col += show_filetype_len;
    if(which == filetype_e) 	return blank_col * blank_w;

    if(show_perms_flag)		blank_col += show_perms_len;
    if(which == perms_e) 	return blank_col * blank_w;

    if(show_nlinks_flag)	blank_col += show_nlinks_len;
    if(which == nlinks_e) 	return blank_col * blank_w;

    if(show_owner_flag)		blank_col += show_owner_len;
    if(which == owner_e) 	return blank_col * blank_w;

    if(show_group_flag)		blank_col += show_group_len;
    if(which == group_e) 	return blank_col * blank_w;

    if(show_size_flag)		blank_col += show_size_len;
    if(which == size_e) 	return blank_col * blank_w;

    if(show_modific_flag)	blank_col += show_modific_len;
    if(which == modific_e) 	return blank_col * blank_w;

    if(show_access_flag)	blank_col += show_access_len;
    if(which == access_e) 	return blank_col * blank_w;

    return blank_col * blank_w;
}

/* potential mess-up if in different browser windows */
void browser_c::move_focus_to_column(sorttype_et which){
    FL_Coord x_offset = compute_column_x_offset(which);
    if (x_offset < 0) x_offset = 0;
    if (x_offset >= form->w) x_offset = form->w - 1;

    tool_focus_move(form, column_hidden_button, x_offset - column_hidden_button->w/2, 0);
}

void browser_c::column_hidden_button_cb(FL_OBJECT *, long){
    FL_Coord xx, yy;
    Window window;
    focus_get(xx, yy, window);
    xx -= source_browser->form->x;
    int blank_w = fl_get_string_width(text->lstyle, text->lsize, " ", 1);
    int blank_no = (int) floor( (xx - column_hidden_button->x)/blank_w);
    if(dbg_flag) cerr << "xx: " << xx << ", ->x: " << column_hidden_button->x << endl;
    if(dbg_flag) cerr << "blank: " << blank_no << endl;

    int blank_col = strlen(line_icon_indent); // what a kludge. Bah. 
    blank_col++;  // even worse! Duh.
    
    if(show_inode_flag){
	blank_col += show_inode_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "inode\n";
	    browser_inode_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_filetype_flag){
	blank_col += show_filetype_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "filetype\n";
	    browser_filetype_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_perms_flag){
	blank_col += show_perms_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "perm\n";
	    browser_perms_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_nlinks_flag){
	blank_col += show_nlinks_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "nlinks\n";
	    browser_nlinks_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_owner_flag){
	blank_col += show_owner_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "owner\n";
	    browser_owner_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_group_flag){
	blank_col += show_group_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "group\n";
	    browser_group_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_size_flag){
	blank_col += show_size_len;
	if(blank_no < blank_col){
	    if(dbg_flag) cerr << "size\n";
	    browser_size_menu_cb(NULL, 0);
	    return;
	}
    }
    if(show_modific_flag){
	blank_col += show_modific_len;
	if(blank_no < blank_col){
	    browser_modific_menu_cb(NULL, 0);
	    return;
	}
    }

    if(show_access_flag){
	blank_col += show_access_len;
	if(blank_no < blank_col){
	    browser_access_menu_cb(NULL, 0);
	    return;
	}
	// raw = 12
    }
    //cerr << "must be name\n";
    browser_name_menu_cb(NULL, 0);
}

static void menu_hidden_button_cb(FL_OBJECT *menu_hidden_button, long){
    if(dbg_flag) cerr << "hidden button.\n";
    tool_set_source(menu_hidden_button->form);
    browser_menu_event_cb(NULL, 0);
}

void browser_c::time_stamp(void){
    struct stat statbuf;
    static int failure_flag = FALSE;
    if(stat(cwd_buf, &statbuf) != 0){
	if(!failure_flag){
	    M_err("browser_c::time_stamp",
		"can't stat browser dir `%s': %s.",
		cwd_buf, strerror(errno)
	    );
	}
	failure_flag = TRUE;  
	return;
    }
    stamp_stat = statbuf;
    // if(dbg_flag) cerr << "making timestamp ...\n";
}

int browser_c::time_stamp_check(int update_flag){
    struct stat statbuf;
    static int failure_flag = FALSE;
    if(stat(cwd_buf, &statbuf) != 0){
	if(!failure_flag){
	    M_err("browser_c::time_stamp",
		"can't stat browser dir `%s': %s.",
		cwd_buf, strerror(errno)
	    );
	}
	failure_flag = TRUE;  
	return FALSE;
    }
    // if(dbg_flag) cerr << "checking timestamp ...\n";
    // how about mtime ?
    if(stamp_stat.st_ctime != statbuf.st_ctime ||
       stamp_stat.st_mtime != statbuf.st_mtime /*|| 
       stamp_stat.st_atime != statbuf.st_atime */
       ){
	if(dbg_flag) cerr << "timestamp says: update\n";
	if(update_flag){
	    if(dbg_flag) cerr << "do it..\n";
	    tool_update_matching_dirs(cwd_buf, TRUE);
	}
	return FALSE;
    }
    return TRUE;
}
