
#include <sys/types.h>
#include <sys/stat.h>

#include "cb.h"
#include "license_forms.h"
#include "global.h"
#include "client_msg.h"

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

// one global callback table.
callback_table_c callback_table;

// callbacks 
#include "cb_key.h"
#include "cb_fkey.h"
#include "browser.h"
#include "browser_menu.h"
#include "error_notifier_forms_aux.h"
#include "ghelp_forms_aux.h"
#include "help_forms.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "options_forms_aux.h"
#include "xshell.h"

void init_callbacks(void){
    /* kill source brwser */
    callback_table.register_callback(BrowserKillButtonCB, BROWSER_WINSHUT_CMSG);
    
    // menu requests
    callback_table.register_callback(browser_menu_event_cb, BROWSER_MENU_CMSG);

    // xshell
    //fl_qdevice(XSHELL_ERROR_HANDLER_CMSG);
    callback_table.register_callback(xshell_error_handler_cb, XSHELL_ERROR_HANDLER_CMSG);

    callback_table.register_callback(callback_loop_init_cb, CALLBACK_LOOP_INIT_CMSG);
}

void directory_update(int nap_ticks){
    if (browser_c::directory_update_count <= 0) return;
    static int update_count = 0;
    browser_c *tmp_browser;
    update_count+=nap_ticks;
    if(update_count >= (1000 / browser_c::directory_update_count)){
	/* default of 40 results in 25 ticks, equivalent to 4 per second */
	update_count = 0;
	static int first = TRUE;
	if(first){
	    first = FALSE;
	} else {
	    tmp_browser = source_browser;
	    do {
		tmp_browser->time_stamp_check();
		tmp_browser = tmp_browser->next_browser;
	    } while (tmp_browser != source_browser);
	}
	tmp_browser = source_browser;
	do {
	    tmp_browser->time_stamp();
	    tmp_browser = tmp_browser->next_browser;
	} while (tmp_browser != source_browser);
    }
}

static int xshell_update(int nap_ticks){
    if (xshell_update_count <= 0) return FALSE;
    static int xshell_count = 0;
    xshell_count += nap_ticks;
    if(xshell_count >= (1000 / xshell_update_count)){
	xshell_count = 0;
	int flag = FALSE;
	while(xshell_check_event()){
	    // eat them all.
	    tool_got_event_flag = TRUE;  // remember.
	    flag = TRUE;
	}
	return flag;
    } else {
	return FALSE;
    }
}

/* 
 * key routine that handles all interaction. Called by main.
 */

void callback_event_loop(void){
    int nap_ticks = 1;
    int pre_count = 100;
    //M_warn("callback_event_loop", "entry");
    while(1){
#ifdef MALLOC_DBG
	dmalloc_verify(0);
#endif
        //M_warn("callback_event_loop", "looping");
	FL_OBJECT *obj;
	XEvent xev;
	obj = fl_check_forms();
	if(dbg_flag) cerr << "check_forms\n";

	if(obj == NULL && tool_cmsg){
	    callback_table.execute_cmsg(tool_cmsg);
	    tool_cmsg = 0;
	    tool_got_event_flag = TRUE;
	}
	if(obj == FL_EVENT){
	    if(dbg_flag) cerr << "FL_EVENT...";
	    if(fl_XNextEvent(&xev)){
	        if(!xshell_check_event(&xev)){
		    callback_table.execute_xev(&xev);
		}
	    } else {
		cerr << "FL_EVENT returned but fl_XNextEvent failed !\n";
	    }
	    if(dbg_flag) cerr << "finished\n";
	}
	if(obj == NULL && !tool_got_event_flag /*&& !forms_active_flag*/ && FALSE){
	    if(dbg_flag) cerr << "obj == NULL\n";
	    // wait a while before shutting off:
	    if(pre_count){
		pre_count--;
	    } else {
		// 10: better for bugs; was 20: go from .6 %. 
		if(nap_ticks < 10){  
		    nap_ticks++;
		}
	    }
	} else {
	    nap_ticks = 1;   // not 0, much better load.
	    pre_count = 100; // we're active again.
	    if(dbg_flag) cerr << "obj != NULL\n";
	    tool_got_event_flag = FALSE;
	    //forms_active_flag = FALSE;
	}
	//nap_ticks = 1;   // force for a while.
	directory_update(nap_ticks);
	xshell_update(nap_ticks);
	if(nap_ticks){
	    // cerr << "n:" << nap_ticks << endl;
	    sginap(nap_ticks); // 1 means 100 ticks per minute.
	}
    }
}
