//#include <iostream.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "forms.h"
#include "global.h"
#include "browser.h"
#include "cb_fkey.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "fork_pipe.h"
#include "global.h"
#include "pathnames.h"
#include "utility.h"

#include "action_forms.h"
#include "action_forms_aux.h"
#include "options_forms.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"

/* all functions/callbacks associated with the copying op. */

static int copy_single_file(
    const char *fname,
    const char *source_dir, 
    const char *target_dir
);

static int copy_do_ask_overwrite   = TRUE;
static int copy_ask_directory_flag = TRUE;

static int do_fork_copy(
    const char *fname, 
    const char *source_dir, 
    const char *target_dir
){
    if(!fname || !source_dir || !target_dir){
	program_unstable("UC internal: do_fork_copy null ptrs.");
	return FALSE;
    }
    
    // we can do -r: on files, it doesn't matter; on dirs, we asked.
    static const char *cp_argv[5] = {"cp", "-r", NULL, NULL, NULL};
    
    cp_argv[1] = tool_cp_minusp_flag ? "-rp" : "-r";

    char buffer[2048];
    sprintf(buffer, "%s/%s", source_dir, fname);		    
    cp_argv[2] = buffer;
    cp_argv[3] = target_dir;
    return do_fork_cmd(cp_path, cp_argv);
}


static int
copy_many_files(const char *source_dir, const char *target_dir){
    if(!source_dir || !target_dir || !source_browser) return 0;
    /* just loop, until the first difficult one is hit. Then, show
     * the aux form with extra buttons. Cancel All stops the complete
     * operation, cancel file this file. Go confirms this file, Go all
     * confirms this particular operation/issue (e.g overwrite, directory)
     * for all items. Thus, Go all can be called for different ones, and
     * it's label needs to be set accordingly.
     * we will use a small actual copying function that returns
     * fail, cond1, cond2, success
     * cond1 could be 'is directory'.
     */
    const char *fname = source_browser->get_first_selection();
    if(!fname) return 0;
    
    int n_changes = 0;
    int do_file_flag;
    int ask_overwrite_flag = TRUE;
    int ask_directory_flag = copy_ask_directory_flag;
    int copy_first = TRUE;
	//    int skip_always_flag = FALSE;
    
    char buffer[1024];
    
    struct stat statbuf;
	
    while(fname){
        sprintf(buffer, "(%s ...)", fname);
	action_msg(buffer);
	// cerr << "copy many files works on " << fname << endl;
	// source_browser->set_current_fname(fname);
	do_file_flag = TRUE; /* reinit each time */
	/*
	 * test for existence
	 */
	if(stat(fname, &statbuf) != 0){
	    sprintf(buffer, "copyfile error: cannot get stat info for file `%s' !", fname);
	    timed_error_notifier(buffer, notifier_secs);
	    source_browser->update();
	    return n_changes;
	}

	/*
	 * test for recursivity - directory
	 */
        int n_files_in_dir = 0;
	if(ask_directory_flag && do_file_flag){
            if(S_ISDIR(statbuf.st_mode)){
                n_files_in_dir = files_in_dir(fname);
	        if(n_files_in_dir != 2){
                    sprintf(buffer, "File `%s' is a directory (%d files). Copy complete tree ?", fname, n_files_in_dir);
		    switch(ask_multi_question(FALSE, 2, buffer,
	                "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Copy Tree", "Dd", NoSymbol, "Copy ALL Trees", "Aa"
		    )){
 		    case -1:
 		    case 0:
		        status_msg(1, "all files cancelled.");
			return n_changes;    
		    case 1:
			status_msg(1, "this file cancelled.");
			do_file_flag = FALSE;
		        break;
		    case 2:
			status_msg(1, "copying this tree is ok.");
			break;
		    case 3:
			status_msg(1, "copying all trees is ok.");
			ask_directory_flag = FALSE;
			break;
		    default:
		        program_unstable("Strange return from switch(multi_q)");
		        return n_changes;
		    }
                }
	    }
	}
	
	/* THIS NEEDS DIR TEST, CANNOT OVERWRITE DIR's !
	 * test for overwriting.
	 */
	if(do_file_flag){
	    strcpy(buffer, target_dir);
	    strcat(buffer, "/");
	    strcat(buffer, fname);
	    if(stat(buffer, &statbuf) == 0){	
		if(S_ISDIR(statbuf.st_mode)){
		    timed_error_notifier("Target exists and is directory and cannot be overwritten.", notifier_secs);
		    do_file_flag = FALSE;
		} else if(ask_overwrite_flag){
		    sprintf(buffer, "file `%s' exists in target directory. Overwrite ?", fname);
		    switch(ask_multi_question(TRUE,1,buffer,
		        "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Overwrite", "Dd", NoSymbol, "Overwrite ALL", "Aa"
	            )){
		    case -1:
		    case 0:
			/* all over */
			status_msg(1, "all files cancelled.");
			return n_changes;
		    case 1:
			/* don't do this file */
			status_msg(1, "this file cancelled.");
			do_file_flag = FALSE;
			break;
		    case 2:
			/* else just do this file */
			status_msg(1, "overwriting this file is ok.");
			break;
		    case 3:
			/* do them all */
			status_msg(1, "overwriting all files is ok.");
			ask_overwrite_flag = FALSE;
			break;
		    default:
			program_unstable("Strange return from switch(action_ctrl)");
			return n_changes;
		    }
		}
	    }
	}
	if(do_file_flag){
	    //status_msg(2, "copying %s", fname);
	    if(do_fork_copy(fname, source_dir, target_dir)){
   	        if(copy_first){
		    remember_target_current(fname);
		    copy_first = FALSE;
		}
		source_browser->deselect_file(fname);
		n_changes++;
	    } else {
		status_msg(0, "copy failed.");
		/*
		if (!skip_always_flag) {
		    switch(ask_multi_question(FALSE, NULL, 0, 0, 0, buffer, 
		        "Skip file",   "f", NoSymbol,
		        "Skip always", "a", NoSymbol,
		        "Cancel", "Cc", NoSymbol
                    )){
		    case 0: break;
		    case 1: skip_always_flag = TRUE; break;
		    default: return n_changes;
	  	    }
		}
		*/
		return n_changes;
	    }
	} else {
	    // cerr << "keep select - not copying file " << fname << endl;
	}
        if(!action_step()){
	    status_msg(0, "copying cancelled.");
	    return n_changes;
	}  	    
	fname = source_browser->get_next_selection();
    }
    return n_changes;
}  


void CopyButtonCB(FL_OBJECT *, long){
	if (toolbox_tree_flag) {
		tree_copy();
		return;
	}
    char buffer[1024];
    if(!target_browser || !source_browser){
	program_unstable("UC internal: copy: no source/target browser (?)");
	return;
    }
    if(tool_get_meta()){
	CopyAndRenameCB(NULL, 0);
	return;
    }
    const char *source_dir = source_browser->get_cwd();
    char target_dir[MAXPATHLEN+1];
    int target_override_flag;
    
    strcpy(target_dir,target_browser->get_cwd());
    target_override_flag = target_dir_get_override(target_dir);

    if(!source_dir || !target_dir){
	program_unstable("UC internal: null ptrs in copy button.");
	return;
    }

    if(!fully_validate_source_target_dirs(source_dir, target_dir, "Copy")) return;

    int n_selected = source_browser->get_n_selected(),
        n_changes  = 0, // record how much happened.
        frozen_flag = FALSE
    ;
    reset_target_current();
    
    if(n_selected <= 0){
	/* 
	 * current file only. always ask for overwrite, directory, special.
	 */
	if(n_selected == 0){
	    const char *fname = source_browser->get_current_fname();
	    if(!fname) return;
	    sprintf(buffer, "Copy current file `%s'", fname);
	    //cerr << "Debug: " << buffer << endl;
	    int new_target_dir_flag = FALSE;
	    if(ask_action_work_question(buffer, source_dir, target_dir, 1, copy_help_text, "Copy",
	         TRUE, new_target_dir_flag, FALSE, FALSE
	    )){
		if(new_target_dir_flag)strcpy(target_dir, tree_new_target_dir);
		//sprintf(buffer, "Copying `%s' ...", fname);
		//indicator_simple_bar(buffer);
		source_browser->freeze();
		target_browser->freeze();
		frozen_flag = TRUE;
		n_changes = copy_single_file(fname, source_dir, target_dir);
		action_end();
		if(n_changes){
		    status_msg(2, "file copied.");
		}
	    }
	}
    } else {
	if(n_selected == 1){
	    const char *fname = source_browser->get_first_selection();
	    if(!fname) return;
	    sprintf(buffer, "Copy selected file `%s'", fname);
	} else {
	    sprintf(buffer, "Copy %d selected files", n_selected);
	}
	int new_target_dir_flag = FALSE;
	if(ask_action_work_question(buffer, source_dir, target_dir, n_selected, copy_help_text, "Copy",
            TRUE, new_target_dir_flag, FALSE, FALSE
	)){
	    if(new_target_dir_flag)strcpy(target_dir, tree_new_target_dir);
	    sprintf(buffer, "Copying %d files ...", n_selected);
	    //indicator_bar(n_selected, buffer);
            source_browser->freeze();
            target_browser->freeze();
            frozen_flag = TRUE;
	    n_changes = copy_many_files(source_dir, target_dir);
	    action_end();
	    if(n_changes == 1){
		status_msg(1, "1 file copied.");
	    } else {
		status_msg(1, "%d files copied.", n_changes);
	    }
	}
    }
    if(n_changes > 0 || toolop_cancelled_flag){
	//cerr << n_changes << " > 0, updating ..." << endl;
	source_browser->update(TRUE);
	if(source_browser != target_browser && !target_override_flag){
	    // redundant test in case of copy, but ...
	    target_browser->update(TRUE);
	    restore_target_current();
	}
    }
    if(frozen_flag){
        source_browser->unfreeze();
        target_browser->unfreeze();
    }
}

// does copying + extra interaction.
static int copy_single_file(
    const char *fname,
    const char *source_dir, 
    const char *target_dir
){

    if(!fname) return FALSE;
    char buffer[1024];
    buffer[0] = '\0';
    
    /*
     * checks go first.
     */
    
    if(!target_dir || !source_dir) return FALSE;
    
    struct stat statbuf;
    if(stat(fname, &statbuf) != 0){
	char lbuffer[1024];
	sprintf(lbuffer, "copyfile error: cannot get stat info for file `%s' !", fname);
	timed_error_notifier(lbuffer, notifier_secs);
	source_browser->update();
	return FALSE;
    }
    
    /* here we check wether we're dealing with a directory. If the respective flag is
     * on, we'll ask wether to copy the whole tree. Else, we'll do it by default.
     */
    
    if(S_ISDIR(statbuf.st_mode) && copy_ask_directory_flag){
	int n_files_in_dir = files_in_dir(fname);
	if(n_files_in_dir != 2){
	    sprintf(buffer, "File `%s' is a directory. Copy complete tree ?", fname);
	    if(!ask_question(FALSE, buffer)){
		return FALSE;
	    }
	}
    }

    /* 
     * does the target exist ? Overwrite ?
     * we construct the target fname and check.
     * we then ask for confirmation if the respective flag is set.
     */
    strcpy(buffer, target_dir);
    strcat(buffer, "/");
    strcat(buffer, fname);

    if(stat(buffer, &statbuf) == 0){
	if(S_ISDIR(statbuf.st_mode)){
	    timed_error_notifier("Target exists and is directory and cannot be overwritten.", notifier_secs);
	    return FALSE;
	}
	if(copy_do_ask_overwrite){
	    sprintf(buffer, "file `%s' exist in target directory. Try to overwrite ?", fname);
	    if(!ask_question(TRUE, buffer)){
		status_msg(1, "don't overwrite.");
		return FALSE;
	    } else {
		status_msg(1, "overwrite ok.");
	    }
	}   /* else just do it */
    }
    remember_target_current(fname);
    return do_fork_copy(fname, source_dir, target_dir);
}



