
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "action_forms_aux.h"
#include "browser.h"
#include "cb_fkey.h"
#include "error_notifier_forms_aux.h"
#include "edit_forms_aux.h"
#include "file_utility.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "pathnames.h"
#include "fork_pipe.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "xshell.h"

void HelpButtonCB(FL_OBJECT *, long){
    info_show_form();
}

void ExitButtonCB(FL_OBJECT *, long){
    if(!tool_get_meta()){
	// else unconditional bye, bye; no saving. 
	char fname_buf[3*MAXPATHLEN];
	int count = get_edit_changed_count(fname_buf, sizeof(fname_buf));
	if (count == 0) {
	    if(!ask_question(FALSE, "Really quit UC ?")) return;
	} else if (count == 1) {
	    if(!ask_question(TRUE, 
		"One UCEdit (%s) has unsaved changes. Really quit?",
		fname_buf
	    )) return;
	} else {
	    if(!ask_question(TRUE, 
		"%d UCEditors (%s) have unsaved changes. Really quit?", 
		count,
		fname_buf
	    )) return;
	}
	    int save_flag = FALSE;
	    switch(save_config_on_exit_state){
	    case always_e: save_flag = TRUE; break;
	    case ask_e:
		switch(ask_multi_question(
		    FALSE, 0, "Save Configuration to Database ?",
		    "Go!",        "GgyY",     NoSymbol,
		    "Cancel",     "CcnN\033", XK_Escape,
		    "Don't quit, oops", "Dd"
		)) {
		case 2:  return;
		case 1:  save_flag = FALSE; break;
		case 0:  save_flag = TRUE;  break;
		default: break;
		}
		break;
	    case dont_e:   
	    default:       save_flag = FALSE; break;
	    }
	    if (save_flag) tool_defaults_save(NULL);
	    if (!conditional_tree_dump()) return;
    }
    xshell_close();
#ifdef MALLOC_DBG
    dmalloc_shutdown();
#endif
    if (tool_focus_on_startup_flag) {
	focus_restore();	// OK.
    }
    exit(0);
}

int do_fork_mkdir(
    const char *path
){
    if(!path){
	program_unstable("UC internal: do_fork_path null path.");
	return FALSE;
    }
    // remember where we are
    
    // we do -p for the path.
    static const char *mkdir_argv[4] = {"mkdir", "-p", NULL, NULL};

    mkdir_argv[2] = path;
    return do_fork_cmd(mkdir_path, mkdir_argv);
}

void MkPathButtonCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;

    static const int bufsize = MAXPATHLEN+2;
    char buf[bufsize];
    getcwd(buf,bufsize);
    strcat(buf, "/");
    const char *result = OKCancel("Make Directory PATH", buf, mkpath_help_text, "Make Path");

    if(!result || !strlen(result)) return;

    /* first get it, then reset it .. */
    /* do it */
    action_work_noform();
    if(!do_fork_mkdir(result)){
	timed_error_notifier("couldn't make the path.", notifier_secs);
	return;
    }   /* else o.k. */

    if(result[0] == '/'){
	tree_construct_path(result);
    } else {
	char buf2[2*bufsize]; // make it oversize
	strcpy(buf2, source_browser->get_cwd());
	strcat(buf2, "/");
	strcat(buf2, result);
	tree_construct_path(buf2);
    }

    /* want to see the new dir, so update */
    source_browser->freeze();
    tool_update_matching_dirs(source_browser->get_cwd());

    if(strstr(result, source_browser->get_cwd())){
	result += strlen(source_browser->get_cwd()) + 1;  // go behind the dir and the first slash. 
	char *slash = strchr(result, '/');              // stop at the next slash. 
	if(slash){
	    *slash = '\0';
	}
	source_browser->set_current_fname(result);
    }
    source_browser->unfreeze();
}

void MkDirButtonCB(FL_OBJECT *, long){
	if (toolbox_tree_flag) {
		tree_mkdir();
		return;
	}
    if(tool_get_meta()){
	MkPathButtonCB(NULL, 0);
	return;
    }
    if(!source_browser){
	program_unstable("No source browser !");
	return;
    }
    static const int bufsize = MAXPATHLEN+2;
    char buf[bufsize];
    getcwd(buf,bufsize);
    strcat(buf, "/");
    const char *result = OKCancel("Make Directory", buf, mkdir_help_text, "Make Directory");
    if(!result || !strlen(result)) return;

    /* first get it, then reset it .. */
    /* do it */
    if(mkdir(result, mkdir_mode_mask) != 0){
	timed_strerror_notifier("Can't make directory:");
	return;
    }   /* else o.k. */

    if(result[0] == '/'){
	tree_construct_path(result);
    } else {
	char buf2[2*bufsize]; // make it oversize
	strcpy(buf2, source_browser->get_cwd());
	strcat(buf2, "/");
	strcat(buf2, result);
	tree_construct_path(buf2);
    }

    /* want to see the new dir, so update */
    source_browser->freeze();
    tool_update_matching_dirs(source_browser->get_cwd());
    char *last_slash = strrchr(result, '/');
    if(last_slash){
	source_browser->set_current_fname(last_slash+1);
    } else {
	source_browser->set_current_fname(result);
    }
    source_browser->unfreeze();
}

void TreeButtonCB(FL_OBJECT *, long){
    do_tree();
}

void QuickEditButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_view();
	return;
    }
    if(tool_get_meta()){
	EditHistoryCB();
	return;
    }

    if(!source_browser) return;
    const char *fname = source_browser->get_current_fname();
    if(!fname) return;
    quick_edit_file(fname, source_browser->get_cwd());
}

#include <sys/types.h>
#include <sys/wait.h>
static int edit_ok(const char *fname){
    if(!fname) return FALSE;
    // check what kind of thing the file is.
    char buffer[1024];
    struct stat statbuf;
    if(stat(fname, &statbuf) < 0){
	sprintf(buffer, "edit error: cannot stat file `%s' !", fname);
	timed_error_notifier(buffer, notifier_secs);
	return FALSE;
    } 
    if(S_ISDIR(statbuf.st_mode)){
	sprintf(buffer, "file `%s' is a directory. Really edit ?", fname);
	return ask_question(FALSE, buffer);
    }
    if(!S_ISREG(statbuf.st_mode) && S_ISLNK(statbuf.st_mode)){
	sprintf(buffer, "file `%s' is not a regular file. Cannot edit.", fname);
	timed_error_notifier(buffer, notifier_secs);
	return FALSE;
    }
    return TRUE;
}

int edit_file(const char *fname, int bBusyCursorFlag){
    if(!fname) return FALSE;
    if(!edit_ok(fname)) return FALSE;
    if(strlen(default_editor) == 0) strcpy(default_editor, "<builtin>");
    if (strcmp(default_editor, "<builtin>") == 0) {
	return quick_edit_file(fname);
    }
    char buffer[1024];
    int retval;
    status_msg(1, "starting %s for %s ...", get_basename(default_editor), fname);
    /* doesn't work with the session manager either: maybe now, after pwd. */
    sprintf(buffer, "%s %s &", default_editor, fname);

    if (bBusyCursorFlag) {
	fl_set_global_cursor(XC_watch);
    }
    //cerr << "buffer: '" << buffer << "'\n";
    safe_system(buffer);
    status_msg(1, "done.");
    retval = TRUE;
#if 0
    int status = system(buffer);
    if(WIFEXITED(status) && status != 0){
	sprintf(buffer, "Unable to execute the editor `%s', status %d !", default_editor, status);
	timed_error_notifier(buffer, "Check path, file names and arguments.", notifier_secs);
	retval = FALSE;
    } else {
	sprintf(buffer, "done(%d).", status);
	status_msg(1, buffer);
	retval = TRUE;
    }
#endif
    if (bBusyCursorFlag) {
	fl_reset_global_cursor();
    }
    return retval;
}

/*
 * simple: we assume that once one has the go to edit
 * them all, no cancel all necessary.
 */

static int edit_file_aux(const char *fname, int &nchanged){
    if(edit_file(fname)) nchanged++;
    return TRUE;
}

/* 
 * assume editing takes long, and in the background.
 * So, we don't update after calling edit. We wait 
 * for the time stamping to alert us of the changes.
 */

void EditButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_edit();
	return;
    }
    if (!tool_editor_info_shown_flag) {
	tool_editor_info_shown_flag = TRUE;
	switch(ask_multi_question(FALSE, 0, 
	    "Which editor would you like to use?", 
	    "UC's Built-In", "Bb", NoSymbol,
	    "Your Own",      "Oo", NoSymbol)
	){
	case 1:
	    OptionsFormEditorButtonCB(NULL, 0);
	    break;
	default:
	    strcpy(default_editor, "<builtin>");
	}
    }

    int n_selected = source_browser->get_n_selected();
    if(n_selected == 0){
	edit_file(source_browser->get_current_fname());
    } else if(n_selected == 1){
	edit_file(source_browser->get_first_selection());
    } else if(n_selected >1){
	char buf[256];
	// built-in editor can't do a special multi-edit, just
	// do it normally. 
	if(tool_get_meta() &&
	    (strcmp(default_editor, "<builtin>") != 0)
	){
	    char *name_buf = strdup(get_basename(default_editor));
	    char *pcSpace = strchr(name_buf, ' ');
	    if(pcSpace){
		*pcSpace = '\0';
	    }
	    sprintf(buf, "Ready to simultaneously edit %d files in one editor (%s) ?", n_selected, name_buf);
	    fl_free(name_buf);
	    name_buf = NULL;
	    if(ask_question(FALSE, buf)){
		status_msg(1, "starting editor on %d files ...", n_selected);
		const int linebuf_size = 2048;
		char linebuf[linebuf_size+16];    // 16 for luck
		const char *fname;
		sprintf(linebuf, "%s ", default_editor);
		fname = source_browser->get_first_selection();
		while(fname){
		    if(strlen(linebuf) + strlen(fname) + 1 < linebuf_size){
			strcat(linebuf, " ");
			strcat(linebuf, fname);
			fname = source_browser->get_next_selection();
		    } else {
			timed_notifier("Too many files to edit. Using only the first ones.");
			fname = NULL;
		    }
		}
		strcat(linebuf, " &");

		safe_system(linebuf);		
	    }
	} else {
	    sprintf(buf, "Ready to separately edit %d files ?", n_selected);
	    if(ask_question(FALSE, buf)){
		source_browser->selection_action_loop(edit_file_aux);
	    }
	}   /* else cancelled */
    }   /* else error */
}

void CommandButtonCB(FL_OBJECT *, long){
//    cmd_show_form();
}


