
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>   
#include <sys/time.h>   
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "browser.h"
#include "cb.h"
#include "cb_key.h"
#include "client_msg.h"
#include "edit_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "global.h"
#include "menu.h"
#include "panel.h"
#include "status_forms_aux.h"
#include "suffix.h"
#include "utility.h"

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

void PadAsteriskKeyCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;
    const char *new_filter;
    if(tool_get_meta()){
	const char *old_pat = "*";
	if(source_browser->filter_full_line_flag){
	    old_pat = source_browser->get_filter_pattern();
	}
        new_filter = OKCancel(
            "give full LINE filter expression (*, ?, [a-Z], [!]) (* to reset)", 
            old_pat, filter_line_help_text, "Filter by Full Line"
        );
	if(!new_filter){
	    return;  // LOCAL cancel
	} else if(strlen(new_filter) <= 0){
	    // safeguard.
	    source_browser->filter_full_line_flag = FALSE;
	    new_filter = "*";
	} else {
	    source_browser->filter_full_line_flag = TRUE;
	}
    } else {
        new_filter = OKCancel(
            "give file name filter expression (*, ?, [a-Z], [!]) (* to reset)", 
            source_browser->get_filter_pattern(), filter_filename_help_text, "Filter by File Name"			      
        );
	if(!new_filter){
	    return;  // LOCAL cancel
	} else if(strlen(new_filter) <= 0){
	    new_filter = "*";
	}
	source_browser->filter_full_line_flag = FALSE;
    }
    source_browser->set_filter_pattern_and_update(new_filter);
}

void PadVirguleKeyCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser ?");
	return;
    }
    if(tool_get_meta()){
	tool_update_all_dirs(TRUE);
    } else {    
	tool_update_matching_dirs(source_browser->get_cwd(), TRUE);
    }
}

void center_current_event_cb(FL_OBJECT *, long){
    if(!source_is_stable()) return;
    fl_center_panel_current(source_browser->panel);
}

void center_selected_event_cb (FL_OBJECT *, long){
    const char *fname = source_browser->check_first_selection();
    if(fname){
	source_browser->set_current_fname(fname, TRUE);
    }
}

void Pad5KeyCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;
    if(tool_get_meta()){
	SetTargetCB(NULL, 0);
    } else {
	center_current_event_cb(NULL, 0);
    }
}

void SetTargetCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;
    static browser_c *prev_target;
    if(source_browser != target_browser){
	prev_target = target_browser;
	tool_set_target(source_browser);
    } else {
	// search !
	browser_c *tmp_browser = source_browser;
	do {
	    if(tmp_browser == prev_target){
		prev_target = source_browser;
		tool_set_target(tmp_browser);
		return;
	    }
	    tmp_browser = tmp_browser->next_browser;
	} while (tmp_browser != source_browser);
	// not found, choose next.
	if(source_browser->next_browser != tmp_browser){
	    tool_set_target(source_browser->next_browser);
	}
    }
}

void scroll_home_event_cb(FL_OBJECT *, long){
    source_browser->home();
}

void cd_home_event_cb(FL_OBJECT *, long){
    char *home_buf = getenv("HOME");
    if(home_buf){
	tool_change_dir(home_buf, TRUE, TRUE, TRUE, TRUE, TRUE);
    } else {
	timed_error_notifier("Couldn't getenv HOME !");
    }
}

void HomeKeyCB(FL_OBJECT *, long bForceNormal){
    if(!source_browser){
	program_unstable("no source browser ?"); 
	return;
    }
    if(bForceNormal || !tool_get_meta()){
	scroll_home_event_cb(NULL, 0);
    } else {
	cd_home_event_cb(NULL, 0);
    }
}

void EndKeyCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser ?");
	return;
    }
    source_browser->end();
}

int down_arrow_key_verify(void){
    //cerr << "incomplete: down_arrow_key_verify\n";
    return TRUE;
}

/* slightly incomplete, because I don't query the keymap wether numlock is
 * pressed. What the hell :)
 */
static long max_msecs_diff = 100;
static int
    max_jump = 20,
    max_run_count = 2
;

static int keysym_event(KeySym k0, KeySym k1, KeySym k2){
    char keybuf[1];
    KeySym keysym;
    XEvent xev;
    if(QLength(fl_display)){
//    if(XPending(fl_display) > 0){
	if(dbg_flag)cerr << "is pending.\n";
	XPeekEvent(fl_display, &xev);
	if(xev.type == KeyPress){
	    if(dbg_flag)cerr << "found another key xev\n";
	    XLookupString((XKeyEvent *) &xev, keybuf, sizeof(keybuf), &keysym, 0);
	    if(keysym == k0 || keysym == k1 || keysym == k2){
		if(dbg_flag)cerr << "Is second desired keysym.\n";
		XNextEvent(fl_display, &xev);
		return TRUE;
	    } 
	} else {
	    if(dbg_flag)cerr << "wrong type (" << xev.type << ")\n";
	    if(xev.type == NoExpose || xev.type == KeyRelease){
		if(dbg_flag)cerr << "override type.\n";
		XNextEvent(fl_display, &xev);
		return keysym_event(k0, k1, k2);
	    }
	}
    }
    return FALSE;
}

void DownArrowKeyCB(FL_OBJECT *, long override){
    static struct timeval prev_down_time;
    static int down_lines = 1;
    int count;
    if(!source_is_stable()) return;
    if(!override){
	if(!down_arrow_key_verify()) return;
    }
    if(!tool_default_actions()) return;
    
    static int run_count = 0;
    struct timeval down_time;
    gettimeofday(&down_time, NULL);
    long diff = msecs_diff(down_time, prev_down_time);
    if(diff < max_msecs_diff){
	if(dbg_flag)cerr << "fast enough. " << down_lines << ", " << diff << endl;
	if(run_count == max_run_count){
	    run_count = 0;
	    if(down_lines < max_jump)
		down_lines++;
	} else {
	    run_count++;
	}
	// eliminate a few down arrow events, if they clog up the queue. Equiv.
	// the mask check below. (verification above).
	count = 1;
	while(count < down_lines){
	    count++;
#if defined(SUN) || defined(HP) || defined(LINUX) || defined(AIX) || defined(GENERICBSD) || defined(UNIXWARE)
	    if(keysym_event(XK_Down, XK_Down, XK_KP_2)){
#else
	    if(keysym_event(XK_KP_Down, XK_Down, XK_KP_2)){
#endif
		if(dbg_flag)cerr << "killed a down event from the queue.\n";
	    } else {
		if(dbg_flag)cerr << "no more down events.\n";
		count = down_lines;
	    }
	}
    } else {
	if(dbg_flag)cerr << "too slow. " << down_lines << ", " << diff << endl;
	down_lines = 1;
    }
    
    /* forms :-(
    XEvent xev;
    if(XCheckMaskEvent(fl_display, KeyReleaseMask, &xev)){
	//cerr << "found keyrelease event.\n";
    }
    */
    
    source_browser->down_arrow(down_lines);
    XSync(fl_display, False);
    // after execution/sync :-) 
    gettimeofday(&down_time, NULL);
    prev_down_time.tv_sec  = down_time.tv_sec;
    prev_down_time.tv_usec = down_time.tv_usec;
}

int up_arrow_key_verify(void){
    return TRUE;
}

void UpArrowKeyCB(FL_OBJECT *, long override){
    static struct timeval prev_up_time;
    static int up_lines = 1;
    int count;
    if(!source_is_stable()) return;
    if(!override){
	if(!up_arrow_key_verify()) return;
    }
    if(!tool_default_actions()) return;
    
    static int run_count = 0;
    struct timeval up_time;
    gettimeofday(&up_time, NULL);
    long diff = msecs_diff(up_time, prev_up_time);
    if(diff < max_msecs_diff){
	if(run_count == max_run_count){
	    run_count = 0;
	    if(up_lines < max_jump)
		up_lines++;
	} else {
	    run_count++;
	}
	// eliminate a few up arrow events, if they clog up the queue. Equiv.
	// the mask check below. (verification above).
	count = 1;
	while(count < up_lines){
	    count++;
#if defined(SUN) || defined(HP) || defined(LINUX) || defined(AIX) || defined(GENERICBSD) || defined(UNIXWARE)
	    if(!keysym_event(XK_Up, XK_Up, XK_KP_8)){
#else
	    if(!keysym_event(XK_KP_Up, XK_Up, XK_KP_8)){
#endif
		count = up_lines;
	    }
	}
    } else {
	up_lines = 1;
    }
    
    source_browser->up_arrow(up_lines);
    XSync(fl_display, False);
    gettimeofday(&up_time, NULL);
    prev_up_time.tv_sec  = up_time.tv_sec;
    prev_up_time.tv_usec = up_time.tv_usec;
}

void PageDownKeyCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser ?");
	return;
    }
    if(tool_get_meta()){
	source_browser->end();
    } else {
	source_browser->page_down();
    }
}

void ChangeDirUpCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser ?");
	return;
    }
    //cerr << "attempting change up \n";
    if(strcmp(source_browser->get_cwd(), "/") != 0){
	// only change up if not root. 
	tool_change_dir("..", TRUE, TRUE, TRUE, TRUE, TRUE);
    }
}

void PageUpKeyCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser ?");
	return;
    }
    if(tool_get_meta()){
	LeftArrowKeyCB(NULL, 0);
    } else {
	source_browser->page_up();
    }
}


void NewEnterKeyCB(FL_OBJECT *, long){
    if(!source_browser) program_unstable("no source browser ?");
    //cerr << "enter key detected ... ";
    const char *ffname = source_browser->get_current_fname();
    if(!ffname){
	//cerr << "no current line" << endl;
	return;
    } else {
	//cerr << "operating on " << ffname << endl;
    }
    char fname[NAME_MAX+1];
    strcpy(fname, ffname);

    /* check filetype here, eg is it a directory ? */
    struct stat statbuf;
    if(stat(fname, &statbuf) < 0){
	char buffer[1024];
	sprintf(buffer, "Launch error: cannot stat file %s !", fname);
	timed_error_notifier(buffer, notifier_secs);
	return;
    } 

    /*
     * note that the if's like to return !
     */

    if(S_ISDIR(statbuf.st_mode)){
	//cerr << "seems like a directory ... " << endl;
	tool_change_dir(fname, TRUE, TRUE, TRUE, TRUE, FALSE);
        return;
    }
    /* check for a rule */
    if(has_sfx_rule(fname)){
	if(dbg_flag) cerr << "Has a suffix rule: " << fname << endl;
        status_msg(2, "executing suffix rule for %s.", fname);
	execute_sfx_rule(fname);
	return;
    } else {
	if(dbg_flag) cerr << "Has no suffix rule: " << fname << endl;
    }
    
    if(S_ISREG(statbuf.st_mode)){
	/* is a good old file, let's see wether we can execute it... */
	int 
	    we_can = FALSE, 
	    someone_can = FALSE,
	    statex = FALSE
        ;
	someone_can = statex = (int) (statbuf.st_mode & S_IXUSR);
	if(statex && (statbuf.st_uid == getuid()) ){
	    we_can = TRUE;
	} else {
	    statex = (int) (statbuf.st_mode & S_IXGRP);
	    if(statex && (statbuf.st_gid == getgid()) ){
		we_can = TRUE;
	    } else {
		someone_can = (statex || someone_can);
		if(statbuf.st_mode & S_IXOTH){
		    we_can = TRUE;
		}
	    }
	}
	if(we_can){
	    launch_command(fname);
	} else if(someone_can){
	    char buffer[1024];
	    sprintf(buffer, "execute permission denied for file %s.", fname);
	    timed_error_notifier(buffer, notifier_secs);
	} else {
	    /*
	    char buffer[64];
	    sprintf(buffer, "View file `%s' ?", fname);
	    if(ask_question(FALSE, buffer)){
	    */
		//cerr << "viewing " << fname << endl;
	    quick_edit_file(fname, source_browser->get_cwd());
	    //p}
	}
    } else if(S_ISLNK(statbuf.st_mode)){
	timed_notifier("seems like a link - can't do that yet !", notifier_secs);
    } else {
	char buffer[1024];
	sprintf(buffer, "file `%s' is not dir, regular file or link.", fname);
	timed_notifier(buffer, notifier_secs);
    }
}

void TabKeyCB(FL_OBJECT *, long no_raise_flag){
    if(!source_is_stable()) return;
    if(tool_get_shift()){
	tool_set_source(source_browser->prev_browser, TRUE);
    } else {
	tool_set_source(source_browser->next_browser, TRUE);
    }
    if(!no_raise_flag){
	source_browser->win_raise();
    }
}

void LeftArrowKeyCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;
    ChangeDirUpCB(NULL, 0);
}

void RightArrowKeyCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;

    // search.
    int up_flag = TRUE;
    int current_line = fl_get_panel_current(source_browser->panel);
    int maxline = fl_get_panel_maxline(source_browser->panel);
    while(1){    
	// search up.
	int new_line = current_line;
	int check_file_flag = TRUE;
	while( (up_flag && new_line >= 1) || (!up_flag && new_line <= maxline) ){
	    const char *fname = source_browser->extract_filename(new_line);
	    check_file_flag = TRUE;
	    if(!fname) 	return;
	    if(!dot_dotdot_check(fname)){
		check_file_flag = FALSE;
	    }
		
	    // check for directory
	    if(source_browser->show_filetype_flag && check_file_flag){
		char dir_char = source_browser->extract_type(new_line);
		if(dir_char != 'd' && dir_char != 'l'){
		    check_file_flag = FALSE;
		}
	    }
	    if(check_file_flag){
		struct stat statbuf;
		if(stat(fname, &statbuf) < 0){
		    char buffer[1024];
		    sprintf(buffer, "right arrow key error: cannot stat file %s !", fname);
		    timed_strerror_notifier(buffer, notifier_secs);
		    return;
		} 
		if(S_ISDIR(statbuf.st_mode)){
		    if(new_line == current_line){
			tool_change_dir(fname, TRUE, TRUE, TRUE, TRUE, FALSE);
		    } else {
			source_browser->set_current_with_margin(new_line);
		    }
		    return;
		}
	    }
	    if(up_flag){
		new_line--;
	    } else {
		new_line++;
	    }
	}
	if(up_flag){
	    // now, try down.
	    up_flag = FALSE;
	} else {
	    return;
	}
    }
}

void update_selection_state(int status_msg_flag){
    static browser_c *last_browser = NULL;
    static int last_n_selected = -1;

    if(!source_browser){
	return;
    }

    int n_selected = source_browser->get_n_selected();
    
    if(last_browser != source_browser || n_selected != last_n_selected){
	source_browser->set_complete_wintitle();
	if(status_msg_flag){
	   status_msg(1, "%d files selected.", n_selected);
        }
    }
    last_browser    = source_browser;
    last_n_selected = n_selected;
}

void DeselectButtonCB(FL_OBJECT *, long){
    if(!source_browser) program_unstable("no source browser ?");
    static char *old_pat = strdup("*");
    const char *result;
    if(tool_get_meta()){
        result = OKCancel("give full LINE DEselection expression (*, ?, [a-Z], [!])",
			  old_pat, deselect_line_help_text, "Deselect by Line");
	if(!result) return;
	source_browser->deselect_all_line_matches(result);
    } else {
        result = OKCancel("give file name DEselection expression (*, ?, [a-Z], [!])",
			  old_pat, deselect_filename_help_text, "Deselect by File Name");
	if(!result) return;
	source_browser->deselect_all_name_matches(result);
    }
    if(old_pat){
	fl_free(old_pat);
    }
    old_pat = strdup(result);
    update_selection_state();
}

void SelectButtonCB(FL_OBJECT *, long){
    if(!source_browser) program_unstable("no source browser ?");
    static char *old_pat = strdup("*");
    const char *result;
    if(tool_get_meta()){
        result = OKCancel("give full LINE selection expression (*, ?, [a-Z], [!])",
			  old_pat, select_line_help_text, "Select by Line");
	if(!result) return;
	source_browser->select_all_line_matches(result);
    } else {
        result = OKCancel("give file name selection expression (*, ?, [a-Z], [!])",
			  old_pat, select_filename_help_text, "Select by File Name");
	if(!result) return;
	source_browser->select_all_name_matches(result);
    }
    if(old_pat){
	fl_free(old_pat);
    }
    old_pat = strdup(result);
    update_selection_state();
}

void InsertKeyCB(FL_OBJECT *, long){
    if(!source_is_stable()) return;

    /* old linux doesn't have XK_KP_Insert... */
#if defined(SUN) || defined(HP) || defined(AIX)  || defined(GENERICBSD) || defined(UNIXWARE)
    if(keysym_event(XK_Insert, XK_Insert, XK_KP_0)){
#else
    if(keysym_event(XK_KP_Insert, XK_Insert, XK_KP_0)){
#endif
	M_info2("InsertKeyCB", "found an insert");
    }
    if(tool_get_meta()){
	source_browser->mirror_selection();
	return;
    }
    const char *fname = source_browser->get_current_fname();
    if(!fname) return; // program_unstable("insert: no current line ?");

    int do_toggle_flag = TRUE;
    if(strcmp(fname, ".") == 0){
	//timed_error_notifier("You cannot select the current directory `.'", notifier_secs);
	do_toggle_flag = FALSE;
    } else if(strcmp(fname, "..") == 0){
	//timed_error_notifier("You cannot select the parent directory `..'", notifier_secs);
	do_toggle_flag = FALSE;
    }
    source_browser->freeze();
    if(do_toggle_flag){
	source_browser->toggle_select_current();
    }

    if(fl_get_panel_current(source_browser->panel) == 1){
	source_browser->cur_up_flag = FALSE;
    }
    if(source_browser->cur_up_flag){
	source_browser->up_arrow();
    } else {
	source_browser->down_arrow();
    }
    source_browser->unfreeze();
}

int direct_window_access_cb(XKeyEvent *xkev, KeySym keysym, int move_mouse_flag){
    if(!xkev) return FALSE;
    if(keysym >= XK_0 && keysym <= XK_9){
	int id = (int) (keysym - XK_0);
	if(controlkey_down(xkev->state)){
	    if(dbg_flag) cerr << "is direct browser access" << ", id: " << id << endl;
	    if(id > 0 && id < browser_c::next_ID){
		// focus_save();
		tool_set_source(id, move_mouse_flag);
		source_browser->win_raise();
		return TRUE;
	    }
	} else if(metakey_down(xkev->state) || tool_get_esc()){
	    if(dbg_flag) cerr << "is direct editor access" << ", id: " << id << endl;
	    return direct_editor_access_cb(id, move_mouse_flag);
	}
    }
    return FALSE;
}


static int s_fsize, s_fstyle;
static FL_OBJECT *s_browser = NULL;

static void fstyle_cb(int style){
    if(style > 0 && s_browser){
	s_fstyle = style;
	fl_set_browser_fontstyle(s_browser, s_fstyle);
    }
}

static void fsize_cb(int size){
    if(size > 0 && s_browser){
	s_fsize = size;
	fl_set_browser_fontsize(s_browser, s_fsize);
    }
}

void standard_browser_font_cb(
    FL_OBJECT *browser,
    int &fsize,
    int &fstyle,
    const char *title
){
    long fsize_menu, fstyle_menu, options_menu;
    
    s_fsize   = fsize;
    s_fstyle  = fstyle;
    s_browser = browser;
    
    if(!title){
	title = "Font Options";
    }

    fsize_menu  = make_fsize_menu((long) fsize_cb);
    fstyle_menu = make_fstyle_menu((long) fstyle_cb);
    
    options_menu = new_pup();
    char buffer[MAXPATHLEN];
    sprintf(buffer, "%s%s", title, "%t");
    add_to_pup(options_menu, buffer, 0);
    add_to_pup(options_menu, "Font Size %m",   fsize_menu,  "zZ");
    add_to_pup(options_menu, "Font Style %m",  fstyle_menu, "tT");

    do_pup(options_menu);
    
    free_pup(fsize_menu);
    free_pup(fstyle_menu);
    free_pup(options_menu);

    // if changed, new. Else just restore.
    fsize   = s_fsize;
    fstyle  = s_fstyle;
}
