//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "browser.h"
#include "cb_fkey.h"
#include "error_notifier_forms_aux.h"
#include "help_forms_aux.h"
#include "file_utility.h"
#include "fork_pipe.h"
#include "global.h"
#include "help_text.h"
#include "panel.h"
#include "pathnames.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "utility.h"

#include "action_forms.h"
#include "action_forms_aux.h"
#include "options_forms.h"

/*
 * all functions/callbacks associated with the move
 * operation/F7 key.
 *
 * note that we use lstat, not stat, for move.
 */

static int move_single_file(const char *fname,             const char *source_dir, const char *target_dir);
static int move_many_files (const char *source_dir,        const char *target_dir);
static int do_rename_wmove (const char *full_source_fname, const char *full_target_fname);

static int 
    move_verbose_flag             = TRUE,
    move_do_ask_one_file_flag     = TRUE,
    move_do_ask_directory_flag    = FALSE,
    move_do_ask_overwrite_flag    = TRUE,
    move_do_ask_writeprotect_flag = TRUE
;

static int new_source_current_flag = FALSE;

void MoveButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_move();
	return;
    }
    if(!target_browser || !source_browser){
	program_unstable("UC internal: move: no source/target browser (?)");
	return;
    }
    // asap:
    if(tool_get_meta()){
	MoveAndRenameCB(NULL, 0);
	return;
    }

    int do_move_flag, frozen_flag = FALSE, target_override_flag;
    char 
        buffer[1024],
	target_dir[MAXPATHLEN+1]
    ;
    strcpy(target_dir,target_browser->get_cwd());
    target_override_flag = target_dir_get_override(target_dir);

    const char
	*fname,
	*source_dir = source_browser->get_cwd()
    ;	

    if(!fully_validate_source_target_dirs(source_dir, target_dir, "Move")) return;

    int n_selected = source_browser->get_n_selected();
    int n_changes = 0; // record how much happened.

    // reset current buffers.
    reset_source_current();
    reset_target_current();
    new_source_current_flag = FALSE;
    
    /* both the case of no selection (= current) and only one selection
     * warrant a bit of special attention - we don't want option overkill.
     */

    if(n_selected <= 1){
	/* 
	 * current file only. always ask for overwrite, directory, special.
	 */
	if(n_selected <= 0){
	    fname = source_browser->get_current_fname();
	    if(!fname){
		//program_unstable("Couldn't extract current filename from source browser.");
		return;
	    }
	    sprintf(buffer, "Move current file `%s'", fname);
            new_source_current_flag = TRUE;
	    //cerr << "Debug: " << buffer << endl;
	} else {
            int selection_line;
	    fname = source_browser->get_first_selection(selection_line);
            int current_line = fl_get_panel_current(source_browser->panel);
            if(current_line == selection_line){   // catch the special case.
	        new_source_current_flag = TRUE;
            }
	    if(!fname){
		program_unstable("Couldn't extract first selection filename from source browser.");
		return;
	    }
	    sprintf(buffer, "Move selected file `%s'", fname);
	}
	if(move_do_ask_one_file_flag){
            int new_target_dir_flag = FALSE;
	    do_move_flag = ask_action_work_question(buffer, source_dir, target_dir, 1, move_help_text, "Move",
	         TRUE, new_target_dir_flag, FALSE, FALSE
	    );
    	    if(new_target_dir_flag)strcpy(target_dir, tree_new_target_dir);
	} else {
	    do_move_flag = TRUE;
	}
	if(do_move_flag){
            source_browser->freeze();
            target_browser->freeze();
            frozen_flag = TRUE;
	    if(move_single_file(fname, source_dir, target_dir)){
		//cerr << "move of current/single selected file succeeded." << endl;
		n_changes = 1;
	    } else {
		//cerr << "move failed." << endl;
	    }
	    action_end();
	}
    } else {
        int new_target_dir_flag = FALSE;
	sprintf(buffer, "Move %d selected files", n_selected);
	if(ask_action_work_question(buffer, source_dir, target_dir, n_selected, move_help_text, "Move",
            TRUE, new_target_dir_flag, FALSE, FALSE
	)){
    	    if(new_target_dir_flag)strcpy(target_dir, tree_new_target_dir);
            source_browser->freeze();
            target_browser->freeze();
            frozen_flag = TRUE;
	    n_changes = move_many_files(source_dir, target_dir);
	    action_end();
	    //cerr << n_changes << "files were moved." << endl;
	}
    }
    if(n_changes > 0 || toolop_cancelled_flag){
	//cerr << n_changes << " > 0, updating ..." << endl;
	// always ok to update the source browser, if n_changes > 0. Either
	// something failed (no stat, etc), or we moved something. Both are ok.
	// updating the target browser is not strictly necessary if the first
	// lstat on a selection list failed, but hey, that's very rare and
	// possibly a good precaution anyway.
	tool_update_matching_dirs(source_dir, TRUE);
        restore_source_current();
    
	tool_update_matching_dirs(target_dir, TRUE);
	if (!target_override_flag) {
	    // meaningless if target directory was different from
	    // target_browser.
	    restore_target_current();
	}
    }
    if(frozen_flag){
        source_browser->unfreeze();
        target_browser->unfreeze();
    }
}

static int n_dir_changes = 0;

static void move_tree_sync(void){
    if(n_dir_changes > 0){
	tree_reshape();
    }
}

static int
move_many_files(
    const char *source_dir, const char *target_dir
){
    if(!source_dir || !target_dir || !source_browser) return 0;

    int selection_line = 0;
    int first_selection_line = 0;
    const char *fname = source_browser->get_first_selection(selection_line);
    first_selection_line = selection_line;
    
    if(!fname) return 0;
    /*
     * just loop, until the first difficult one is hit. Then, show
     * the aux form with extra buttons. Cancel All stops the complete
     * operation, cancel file this file. Go confirms this file, Go all
     * confirms this particular operation/issue (e.g overwrite, directory)
     * for all items. Thus, Go all can be called for different ones, and
     * it's label needs to be set accordingly.
     * we will use a small actual moveing function that returns
     * fail, cond1, cond2, success
     * cond1 could be 'is directory'.
     */
    
    char buffer[1024], full_source_fname[1024], full_target_fname[1024];
    
    int n_changes = 0,
        do_file_flag,
	move_first = TRUE,         // first successful move - for kludge.
        loop_ask_overwrite_flag    = TRUE,   // always.
        loop_ask_directory_flag    = move_do_ask_directory_flag,
        loop_ask_writeprotect_flag = move_do_ask_writeprotect_flag
    ;
    n_dir_changes = 0;  // reset.

    struct stat source_file_statbuf, target_file_statbuf;

    while(fname){	
	sprintf(buffer, "(%s ...)", fname);
	action_msg(buffer);
        //cerr << "move many files works on " << fname << endl;
	// source_browser->set_current_fname(fname);
	do_file_flag = TRUE; /* reinit each time */

	// can't move '.' or '..' :
	do_file_flag = dot_dotdot_test(fname);
	
	if(do_file_flag){
	    /* does the source file exist ?
	     * if it doesn't, it is too important an error to
	     * continue blindly, without user interaction.
	     */
	    strcpy(full_source_fname, source_dir);
	    strcat(full_source_fname, "/");
	    strcat(full_source_fname, fname);
#ifdef SUN	    
	    if(lstat((char*)full_source_fname, &source_file_statbuf) != 0){
#else
	    if(lstat(full_source_fname, &source_file_statbuf) != 0){
#endif
		sprintf(buffer, "Couldn't get lstat info for file `%s'.", fname);
		timed_strerror_notifier(buffer, notifier_secs);
		n_changes++;     // use this to trigger an update unconditionally.
		move_tree_sync();
		return n_changes;
	    }
	}

	/* directory check */

	if(do_file_flag && move_do_ask_directory_flag){
	    if(S_ISDIR(source_file_statbuf.st_mode)){
		sprintf(buffer, "File `%s' is a directory. Move complete tree ?", fname);
		switch(ask_multi_question(FALSE, 2, buffer,
		    "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Move Tree", "Dd", NoSymbol, "Move ALL Trees", "Aa" 
                )){
		case 3:  loop_ask_directory_flag = FALSE; break;
		case 2:  break;
		case 1:  do_file_flag = FALSE; break;
  	        case 0:  
		case -1: 
		default:
		    move_tree_sync();
		    return n_changes;    // fall-through
		}
	    }
	}

	/* are we still on ? */

	if(do_file_flag){
	    /* 
	     * does the target exist ? Overwrite ? Are they the same ?
	     * we construct the target fname and check.
	     * we then ask for confirmation if the respective flag is set.
	     */
	    strcpy(full_target_fname, target_dir);
	    strcat(full_target_fname, "/");
	    strcat(full_target_fname, fname);
	    
#ifdef SUN	    
	    if(lstat((char*)full_target_fname, &target_file_statbuf) == 0){
#else
	    if(lstat(full_target_fname, &target_file_statbuf) == 0){
#endif
		if(source_file_statbuf.st_dev == target_file_statbuf.st_dev &&
		   source_file_statbuf.st_ino == target_file_statbuf.st_ino){
		    sprintf(buffer, "Source and target files are identical - not moved.");
		    timed_error_notifier(buffer, notifier_secs);
		    do_file_flag = FALSE;
		} else if(S_ISDIR(target_file_statbuf.st_mode)){
		    sprintf(buffer, "Target file is a directory - not removed.");
		    timed_error_notifier(buffer, notifier_secs);
		    do_file_flag = FALSE;
		} else if(loop_ask_overwrite_flag){
		    sprintf(buffer, "file `%s' exist in target directory. Try to overwrite ?", fname);
		    switch(ask_multi_question(TRUE, 1, buffer,
		        "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Overwrite", "Dd", NoSymbol, "Overwrite ALL", "Aa"
		    )){
		    case 3: loop_ask_overwrite_flag = FALSE; break;
		    case 2: break;
		    case 1: do_file_flag = FALSE; break;
		    case 0:
		    case -1:
		    default:
			move_tree_sync();
			return n_changes;
		    }
		}
		if(loop_ask_writeprotect_flag && do_file_flag){
		    if(!try_access(&target_file_statbuf, W_OK) &&
		       !S_ISLNK(target_file_statbuf.st_mode)){
			sprintf(buffer, "file `%s' in target directory is write protected. Force overwrite ?", fname);
			switch(ask_multi_question(TRUE, 1, buffer,
		            "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Force", "Dd", NoSymbol, "Force ALL", "Aa"
		        )){
			case 3: loop_ask_writeprotect_flag = FALSE; break;
			case 2: break;
			case 1: do_file_flag = FALSE; break;
			case 0:
			case -1:
			default:
			    move_tree_sync();
			    return n_changes;
			}
		    }
		}
	    } else if(errno != ENOENT){
		// lstat failed not because target didn't exist
		sprintf(buffer, "lstat(%s) syscall failed.", full_target_fname);
		timed_strerror_notifier(buffer, notifier_secs);
		do_file_flag = FALSE;
	    }
	}
	
	if(do_file_flag){
	    // cerr << "deselect .. moving file " << fname << endl;
	    if(do_rename_wmove(full_source_fname, full_target_fname)){
	        if(S_ISDIR(source_file_statbuf.st_mode)){
		    if(tree_move_path2dir(full_source_fname, target_dir, FALSE)){
			n_dir_changes++;
		    }
		}
	        if(move_verbose_flag){
		    //cerr << "Move of file " << fname << " succeeded." << endl;
		}
		source_browser->deselect_file(fname);
		if(move_first){
	            remember_target_current(fname);
	            move_first = FALSE;
		}

	        int current_line = fl_get_panel_current(source_browser->panel);
   	        if(current_line == selection_line){
	            new_source_current_flag = TRUE;
         	}
 	        if(new_source_current_flag){
	            int max_line = fl_get_panel_maxline(source_browser->panel);
	            int new_line = selection_line;
	            if(new_line == max_line){
	                new_line = first_selection_line -1;
		    } else {
	                new_line++;
                    }
	            remember_source_current(source_browser->extract_filename(new_line));
        	}
		n_changes++;
	    } else {
		//cerr << "do_rename_wmove failed." << endl;
		move_tree_sync();
	        return n_changes;
	    }
	} else {
	    //cerr << "keep select - not moving file " << fname << endl;
	}
        if(!action_step()){
	    //cerr << "move cancelled." << endl;
	    move_tree_sync();
	    return n_changes;
	}  	    
	fname = source_browser->get_next_selection(selection_line);
    }
    move_tree_sync();
    return n_changes;
}  

// does moveing + extra interaction.
static int move_single_file(const char *fname, const char *source_dir, const char *target_dir){
    //action_mini_step(FALSE);  // for the fun of it.
    action_mini_step(FALSE);  

    if(!fname || !source_dir || !target_dir) return FALSE;
    char buffer[1024], full_source_fname[1024], full_target_fname[1024];
    //full_source_fname[0] = '\0';
    //full_target_fname[0] = '\0';
    struct stat source_file_statbuf, target_file_statbuf;

    // can't move '.' or '..' :
    if(!dot_dotdot_test(fname))	return FALSE;
    
    strcpy(full_source_fname, source_dir);
    strcat(full_source_fname, "/");
    strcat(full_source_fname, fname);

    /* does the source file exist ? */

#ifdef SUN	    
    if(lstat((char*)full_source_fname, &source_file_statbuf) != 0){
#else
    if(lstat(full_source_fname, &source_file_statbuf) != 0){
#endif

	sprintf(buffer, "Couldn't get lstat info for file `%s'.", fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    }
 
    /*
     * here we check wether we're dealing with a directory. If the respective flag is
     * on, we'll ask wether to move the whole tree. Else, we'll do it by default.
     */
    
    if(S_ISDIR(source_file_statbuf.st_mode)){
	if(move_do_ask_directory_flag){
	    sprintf(buffer, "File `%s' is a directory. Move complete tree ?", fname);
	    if(!ask_question(FALSE, buffer)){
		return FALSE;
	    }
	}
    }

    /* 
     * does the target exist ? Overwrite ? Are they the same ?
     * we construct the target fname and check.
     * we then ask for confirmation if the respective flag is set.
     */
    strcpy(full_target_fname, target_dir);
    strcat(full_target_fname, "/");
    strcat(full_target_fname, fname);
#ifdef SUN	    
    if(lstat((char*)full_target_fname, &target_file_statbuf) == 0){
#else
    if(lstat(full_target_fname, &target_file_statbuf) == 0){
#endif
	if(source_file_statbuf.st_dev == target_file_statbuf.st_dev &&
	   source_file_statbuf.st_ino == target_file_statbuf.st_ino){
	    sprintf(buffer, "Source and target files are identical - not moved.");
	    timed_error_notifier(buffer, notifier_secs);
	    return FALSE;
	} else if(S_ISDIR(target_file_statbuf.st_mode)){
	    sprintf(buffer, "Target file is a directory - not removed.");
	    timed_error_notifier(buffer, notifier_secs);
	    return FALSE;
	}
	if(move_do_ask_overwrite_flag){
	    sprintf(buffer, "file `%s' exist in target directory. Try to overwrite ?", fname);
	    if(!ask_question(TRUE, buffer)){
		//cerr << "don't overwrite" << endl;
		return FALSE;
	    } else {
		//cerr << "overwrite" << endl;
	    }
	}
	if(move_do_ask_writeprotect_flag){
	    if(!try_access(&target_file_statbuf, W_OK) &&
	       !S_ISLNK(target_file_statbuf.st_mode)){
		sprintf(buffer, "file `%s' in target directory is write protected. Force overwrite ?", fname);
		if(!ask_question(TRUE, buffer)){
		    //cerr << "don't force overwrite" << endl;
		    return FALSE;
		} else {
		    //cerr << "force overwrite" << endl;
		}
	    }
	}
	/* else just do it */
    } else if(errno != ENOENT){
	// lstat failed not because target didn't exist
	sprintf(buffer, "lstat(`%s') syscall failed.", full_target_fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    }

    /* if we're here, it's
     * - ok to overwrite, 
     * - ok to move directory,
     * - source exists,
     * - full target's no directory, etc.
     *
     * some nice munging around with the current lines, so that they are
     * in good default positions.
     */
    
    int new_line = fl_get_panel_current(source_browser->panel);  // dangerous call
    if(do_rename_wmove(full_source_fname, full_target_fname)){
	if(S_ISDIR(source_file_statbuf.st_mode)){
	    tree_move_path2dir(full_source_fname, target_dir);
	}
	remember_target_current(fname);
	int max_line = fl_get_panel_maxline(source_browser->panel);
	if(new_line == max_line){
	    new_line--;
	} else {
	    new_line++;
	}
	if(new_source_current_flag){
	    remember_source_current(source_browser->extract_filename(new_line));
	}
	return TRUE;
    } else {
	return FALSE;
    }
}

static int do_rename_wmove(const char *full_source_fname, const char *full_target_fname){
    char buffer[2048];
    action_mini_step(FALSE);
    if(rename(full_source_fname, full_target_fname) == 0){
	action_mini_step(FALSE);
	return TRUE;
    }
    if(errno == EXDEV){
	struct stat source_file_statbuf;
#ifdef SUN	    
	if(lstat((char*)full_source_fname, &source_file_statbuf) != 0){
#else
	if(lstat(full_source_fname, &source_file_statbuf) != 0){
#endif	
	    cerr << "couldn't get stat info for " << full_source_fname << endl;
	    return FALSE;
	} else {
	    if(S_ISDIR(source_file_statbuf.st_mode)){
		timed_notifier("Sorry: cannot move directories between filesystems.", notifier_secs);
		return FALSE;
	    }
	}
	// this takes long anyway, because it involves a copy. 
	static const char *mv_argv[4] = {"mv", NULL, NULL, NULL};
	mv_argv[0] = "mv";
        mv_argv[1] = full_source_fname;
        mv_argv[2] = full_target_fname;
	mv_argv[3] = NULL;
	if(do_fork_cmd(mv_path, mv_argv)){
	    return TRUE;
	} else {
	    timed_strerror_notifier("cross-filesystem move failed.", notifier_secs);
	    return FALSE;
	}
    } else {
	sprintf(buffer, "Cannot move `%s' to `%s'.", full_source_fname, full_target_fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    } 
}


