//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "action_forms_aux.h"
#include "cb_remove.h"
#include "forms.h"
#include "action_forms.h"
#include "browser.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "global.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "options_forms.h"
#include "panel.h"
#include "pathnames.h"
#include "fork_pipe.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "utility.h"

/* all functions/callbacks associated with the delete operation.
 * note that we use lstat, not stat, for delete. ??
 */

static int delete_single_file(const char *fname, const char *source_dir);
static int delete_many_files (const char *source_dir);
static int delete_do_each    (const char *fname, int &counter);
static int delete_file       (const char *full_source_fname);

int delete_dir(const char *full_source_fname);

//    loop_ask_always_flag       = TRUE,    // in loop, always ask for removal
// note that the source_dir_action already asks for a single file, and for
// multiple files, we also ask via a panel. Could need a single step option.

static int
    loop_flag                  = FALSE,   // are we in a loop ?
    loop_ask_directory_flag    = TRUE,    // ask for directories ?
    loop_ask_writeprotect_flag = TRUE,    // ask for overriding write protect
    n_dir_changes              = 0        // for tree sync
;

static int new_source_current_flag = FALSE;

void DeleteButtonCB(FL_OBJECT *, long){
	if (toolbox_tree_flag) {
		tree_delete();
		return;
	}
    source_dir_action(delete_single_file, delete_many_files, "DELETE", TRUE, FALSE, delete_help_text,
		      FALSE, FALSE, TRUE);
}
static const char
    *loop_source_dir = NULL
;

static int delete_many_files(const char *source_dir){
    if(!source_dir || !source_browser) return 0;
    loop_flag                  = TRUE;
    loop_ask_directory_flag    = TRUE,  
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    new_source_current_flag    = FALSE;
    n_dir_changes              = 0;

    fl_freeze_form(TreeForm);
    source_browser->freeze();
    int retval = source_browser->selection_action_loop(delete_do_each);
    source_browser->unfreeze();
    if(n_dir_changes>0){
	tree_reshape();
    }
    fl_unfreeze_form(TreeForm);
    return retval;
}

static int delete_single_file(const char *fname, const char *source_dir){
    if(!fname || !source_dir) return FALSE;
    int counter = 0;          
    loop_flag = FALSE;
    loop_ask_directory_flag    = TRUE,  
    loop_ask_writeprotect_flag = TRUE;       // always
    loop_source_dir            = source_dir;
    new_source_current_flag    = FALSE;
    n_dir_changes              = 0;
    fl_freeze_form(TreeForm);
    delete_do_each(fname, counter);
    if(n_dir_changes>0){
	tree_reshape();
    }
    fl_unfreeze_form(TreeForm);
    return counter;
}

// if we want to cancel all, ret false. Cancel file: ret true, don't
// inc counter.

static int delete_do_each(const char *fname, int &counter){
    if(!fname || !loop_source_dir) return FALSE;

    char buffer[1024], full_source_fname[1024];
    struct stat source_file_statbuf;
    static char cfname[MAXPATHLEN+1] = "";
    static int first_selection_line;
    
    int retval = TRUE, n_files_in_dir = 0;
    
    // can't delete '.' or '..' :
    if(!dot_dotdot_test(fname)){
	timed_error_notifier("Can't delete `.' or `..'.");
	return FALSE;
    }
    if(loop_flag){
	sprintf(buffer, "(%s ...)", fname);
	action_msg(buffer);
    }

    //cerr << "Fname before: " << fname << endl;

    // construct full fname
    strcpy(full_source_fname, loop_source_dir);
    strcat(full_source_fname, "/");
    strcat(full_source_fname, fname);

    // does the source file exist ? 
    if(lstat(full_source_fname, &source_file_statbuf) != 0){
	sprintf(buffer, "Couldn't get lstat info for file `%s'.", fname);
	timed_strerror_notifier(buffer, notifier_secs);
	return FALSE;
    }
    if(counter == 0){
	const char *tmp_fname = source_browser->get_current_fname(); // only do this once.
	if(!tmp_fname){
	    tmp_fname = "error:-)";
	}
	strcpy(cfname, tmp_fname);
	first_selection_line = source_browser->get_selection_workline();  // re-init.
    }
    if(!new_source_current_flag){
	//cerr << "fname, cfname: " << fname << ", " << cfname << endl;
	if(strcmp(fname, cfname) == 0){
	    new_source_current_flag = TRUE;
	}
    }
    //cerr << "Fname after: " << fname << endl;

    // write protection. delete links without asking for w.p.
    if(loop_ask_writeprotect_flag){
	if(!try_access(&source_file_statbuf, W_OK) &&
	   !S_ISLNK(source_file_statbuf.st_mode)){
	    sprintf(buffer, "File `%s' is write protected. Force delete ?", fname);
	    if(loop_flag){
		switch(ask_multi_question(TRUE, 1, buffer,
                    "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Force", "Dd", NoSymbol, "Force ALL", "Aa"
                )){
		case -1:
		case 0:
		    status_msg(1, "all files cancelled.");
		    return FALSE;
		case 1:
		    status_msg(1, "this file cancelled.");
		    action_step(FALSE);
		    return TRUE;   // don't increment counter !
		case 2:
		    status_msg(1, "overwriting this file is ok.");
		    break;
		case 3:
		    status_msg(1, "overwriting all files is ok.");
		    loop_ask_writeprotect_flag = FALSE;
		    break;
		default:
		    program_unstable("Strange return from switch(action_ctrl)");
		    return FALSE;
		}
	    } else {
		if(!ask_question(TRUE, buffer)){
		    //cerr << "don't force w.p. delete" << endl;
		    return TRUE;
		} else {
		    //cerr << "force w.p delete" << endl;
		}
	    }
	}
    }
    /* here we check wether we're dealing with a directory. If the respective flag is
     * on, we'll ask wether to delete the whole tree. Else, we'll do it by default.
     */
    
    if(S_ISDIR(source_file_statbuf.st_mode)){
	n_files_in_dir = files_in_dir(full_source_fname);
	if(loop_ask_directory_flag && n_files_in_dir != 2){
	    sprintf(buffer, "File `%s' is a non-empty directory (%d files). Really delete complete tree ?", fname, n_files_in_dir);
	    if(loop_flag){
		switch(ask_multi_question(TRUE, 1, buffer, 
                    "Cancel All", "Cc", NoSymbol, "Cancel File", "Ff", NoSymbol, "Do Delete Tree", "Dd", NoSymbol, "D. ALL Trees", "Aa"
                )){
		case -1:
		case 0:
		    status_msg(1, "all files cancelled.");
		    return FALSE;
		case 1:
		    status_msg(1, "this file cancelled.");
		    action_step(FALSE);
		    return TRUE;   
		case 2:
		    status_msg(1, "deleting this tree is ok.");
		    break;
		case 3:
		    status_msg(1, "deleting all trees is ok.");
		    loop_ask_directory_flag = FALSE;
		    break;
		default:
		    program_unstable("Strange return from switch(action_ctrl)");
		    return FALSE;
		}
	    } else {
		if(!ask_question(TRUE, buffer)){
		    //cerr << "don't delete tree" << endl;
		    return TRUE;
		} else {
		    //cerr << "do delete tree" << endl;
		}	
	    }
	}
	if(!loop_flag){
	    sprintf(buffer, "Removing directory `%s' ...", fname);
	    //indicator_simple_bar(buffer);
	}
	retval = delete_dir(full_source_fname);
	if(retval){
	    if(tree_delete_path(full_source_fname, FALSE)){
		n_dir_changes++;
	    }
	}
    } else {
	if(!loop_flag){
	    sprintf(buffer, "Removing file `%s' ...", fname);
	    //indicator_bar(1, buffer);
	}	
	retval = delete_file(full_source_fname);
    }
    if(!loop_flag){
        action_step(FALSE);
	action_end();
    }

    
    if(retval){
	//cerr << "deleting " << full_source_fname << " succeeded.\n";
	if(loop_flag){
	    source_browser->deselect_file(fname);
	    retval = action_step();   // only in loops.
	}
	counter++;
	// have a nicely positioned current line.
	if(new_source_current_flag){
	    int max_line = fl_get_panel_maxline(source_browser->panel);
	    int new_line = source_browser->get_selection_workline();
	    if(new_line == max_line){
		new_line = first_selection_line -1;
	    } else {
		new_line++;
	    }
	    remember_source_current(source_browser->extract_filename(new_line));
	}
    } else {
	//cerr << "deleting " << full_source_fname << " failed.\n";
    }
    return retval;
}

/* delete a file */
static int delete_file (const char *full_source_fname){
    char buffer[1024];
    if(!action_mini_step(FALSE)) return FALSE;
    if(!full_source_fname) return FALSE;
    if(unlink(full_source_fname)){
	if(errno != ENOENT){
	    sprintf(buffer, "Deleting `%s' failed:", full_source_fname);
	    timed_error_notifier(buffer, strerror(errno), notifier_secs);
	    return FALSE;
	}
    }
    if(!action_mini_step()) return FALSE;
    return TRUE;
}

int delete_dir(const char *full_source_fname){
    if(!full_source_fname){
	program_unstable("UC internal: delete_dir null ptrs.");
	return FALSE;
    }
    // we can do -rf: we asked for dirs
    static const char *rm_argv[4] = {"rm", "-rf", NULL, NULL};
    rm_argv[2] = full_source_fname;
    return do_fork_cmd(rm_path, rm_argv);
}

