#include "iostream.h"

#include "cb_table.h"
#include "global.h"
#include "utility.h"

static int dbg_flag = tool_debug_flag;

void callback_table_c::execute_xev(XEvent *xev){
    if(!xev){
	cerr << "NULL X event in callback_table_c::execute_xev.\n";
	return;
    }
    if(xev->type == ClientMessage){
	XClientMessageEvent *xcm = (XClientMessageEvent *) xev;
	if(xcm->message_type == XInternAtom(fl_display, "UC_MESSAGES", 0)){
	    cerr << "cb table: is uc message.\n";
	    int type = (int) xcm->data.l[0];
	    execute_cmsg(type);
	} else {
	    cerr << "cb table: is NOT uc message.\n";
	}
    } else {
	if(non_client_callback && !x_restricted_flag){
	    (*non_client_callback)(xev);
	}
    }
}

void callback_table_c::execute_cmsg(int cmsg){
    int index = lookup_index(cmsg);
    fl_callback_t a_callback = NULL;
    if(index >= 0){
	if(restricted_flags[index]){
	    if(dbg_flag) cerr << "cannot execute: ";
	    if(dbg_flag) cerr << "callback no. " << index << " for client msg no " << cmsg << " is restricted.\n";
	    return;
	} else {
	    a_callback = callbacks[index];
	}
    }
    if(a_callback){
	(*a_callback)(NULL,0);
    } else {
	if(dbg_flag){
	    error_msg("Warning: no callback for %d cmsg. discarded.",cmsg);
	}
    }
}

fl_callback_t callback_table_c::lookup(int cmsg){
    for(int i=0; i<n_callbacks; i++){
	if(cmsg_no[i] == cmsg){			// linear search, I know :-)
	    return callbacks[i];
	}
    }
    return NULL;
}

int callback_table_c::lookup_index(int cmsg){
    for(int i=0; i<n_callbacks; i++){
	if(cmsg_no[i] == cmsg){			// linear search, I know :-)
	    return i;
	}
    }
    return -1;
}

callback_table_c::callback_table_c(void){
    non_client_callback = NULL;
    callbacks = NULL;
    cmsg_no = NULL;
    restricted_flags = NULL;
    n_callbacks = n_slots = 0;
    x_restricted_flag = FALSE;
}

void callback_table_c::register_callback(fl_callback_t a_callback, int cmsg){  	
    if(n_callbacks >= n_slots){
	const int increment = 20;
	
	fl_callback_t 	*tmp_callbacks = callbacks;
	int		*tmp_cmsg_no = cmsg_no;
	int             *tmp_r_flags   = restricted_flags;
	
	callbacks        = new fl_callback_t	[n_slots + increment];
	cmsg_no          = new int		[n_slots + increment];
	restricted_flags = new int              [n_slots + increment];
	
	for(int i=0; i<n_slots; i++){
	    callbacks[i]        = tmp_callbacks[i];
	    cmsg_no[i]          = tmp_cmsg_no[i];
	    restricted_flags[i] = tmp_r_flags[i];
	}
	
	if(tmp_callbacks){
	    delete [] tmp_callbacks;
	    tmp_callbacks = NULL;
	}
	if(tmp_cmsg_no){
	    delete [] tmp_cmsg_no;
	    tmp_cmsg_no = NULL;
	}
	if(tmp_r_flags){
	    delete [] tmp_r_flags;
	    tmp_r_flags = NULL;
	}
	
	n_slots += increment;
	register_callback(a_callback, cmsg);
    } else {
	if(lookup(cmsg)){
	    error_exit("attempt to register a 2nd callback on cmsg %d", cmsg);
	} else {
	    callbacks       [n_callbacks] = a_callback;
	    cmsg_no       [n_callbacks] = cmsg;
	    restricted_flags[n_callbacks] = FALSE;
	    n_callbacks++;
	    return;
	}
    }
}

/* idea: Find callback. Move last callback over this callback.
 * decrease n_callbacks by one.
 */

void callback_table_c::deregister_callback(int cmsg){
    int i=0, found_flag = FALSE;
    while(i<n_callbacks && !found_flag){
	if(cmsg_no[i] == cmsg){
	    found_flag = TRUE;
	} else {
	    i++;
	}
    }

    if(found_flag){
	int last_cb = n_callbacks-1;
	if(i == last_cb){
	    // it's the very last one.
	    callbacks[i] = NULL;
	    cmsg_no[i] = 0;
	} else {
	    // overwrite it by the last.
	    callbacks[i] = callbacks[last_cb];
	    cmsg_no[i] = cmsg_no[last_cb];
	    callbacks[last_cb] = NULL;
	    cmsg_no[last_cb] = 0;
	}
	n_callbacks--;
    }
}

void callback_table_c::restrict(int cmsg){
    int index = lookup_index(cmsg);
    if(index >= 0){
	restricted_flags[index] = TRUE;
    }
}

void callback_table_c::unrestrict(int cmsg){
    int index = lookup_index(cmsg);
    if(index >= 0){
	restricted_flags[index] = FALSE;
    }
}

void callback_table_c::restrict_all(void){
    for(int i=0; i<n_slots; i++){
	restricted_flags[i] = TRUE;
    }
    x_restricted_flag = TRUE;
}

void callback_table_c::unrestrict_all(void){
    for(int i=0; i<n_slots; i++){
	restricted_flags[i] = FALSE;
    }
    x_restricted_flag = FALSE;
}

