/*
 * callback_table.h:
 * collect callback/client message pairs and allow to call the
 * callback via its client message number. 
 * main support for client messages as defined in client_msg.h.
 * call execute_xev with an XEvent. If everything is restricted,
 * just return. If it is a client message, check wether it is
 * restricted, and conditionally execute, return. If not, execute
 * the non_client_callback on the XEvent. 
 */

#ifndef CALLBACK_TABLE_H
#define CALLBACK_TABLE_H

#include "forms.h"

typedef void (*fl_callback_t)(FL_OBJECT *, long);
typedef void (*non_client_callback_t)(XEvent *xev);

class callback_table_c {
private:
    fl_callback_t *callbacks;		// function array
    int	*cmsg_no;			// associated cmsg's.
    int *restricted_flags;              // which cb is on/off
    int n_callbacks;			// how many callbacks in array 
    int n_slots;			// how many slots in array
    int lookup_index(int cmsg);	        // for internal arrays.
    int x_restricted_flag;              // set by restrict_all.
public:
    non_client_callback_t *non_client_callback;
    void
        execute_cmsg(int cmsg),         // args are NULL, 0.
        execute_xev(XEvent *xev),
        deregister_callback(int cmsg),                  
        register_callback(fl_callback_t aFunc, int cmsg),
	restrict(int cmsg),
        restrict_all(void),             // must be paired with unrestrict_all.
	unrestrict(int cmsg),
        unrestrict_all(void)
    ;
    fl_callback_t lookup(int cmsg);	        // supply your own args.

    callback_table_c(void);    
};

#endif // CALLBACK_TABLE_H

