/*
 * functionality for colour panel.
 * basically, it can be popped up anytime, and always
 * applies to the source browser.
 * thus, on every callback, the source browser could have 
 * changed. Similarly to the tool_change_dir function on the
 * current working directory panel, the colours here need to
 * be updated if we switch to a different source browser. Let's
 * assume for one moment that this is done elsewhere. Note that
 * the 'set_' functions also need to update this information -
 * not necessarily from within the browser object. We can get 
 * the colours and set them for the form here. Much better.
 */

#include <sys/param.h>
#include "browser.h"
#include "colour_forms.h"
#include "colour_forms_aux.h"
#include "forms_aux.h"
#include "ghelp_forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "menu.h"
#include "panel.h"

void ColoursFormHelpButtonCB(FL_OBJECT *, long){
    ghelp_colours_cb(NULL, 0);
}

void ColoursFormDefaultsButtonCB(FL_OBJECT *, long){
    source_browser->set_dark_yellow_bright();
    colours_form_sync();
}

void ColoursFormSyncAllButtonCB(FL_OBJECT *, long){
    browser_c *tmp_browser = source_browser->next_browser;
    while(tmp_browser != source_browser){
	tmp_browser->freeze();
	tmp_browser->sync_colour_opts(source_browser);
	tmp_browser->update(TRUE);
	tmp_browser->unfreeze();
	tmp_browser = tmp_browser->next_browser;
    }
}


static void colours_help_func(void){
    ColoursFormHelpButtonCB(NULL,0);
}

static long predefined_menu = 0;

static int colours_form_raw_cb(FL_FORM *form, void *xev){
    XEvent *xevent = (XEvent *)xev;
    if(form != ColoursForm){
	return FALSE;
    }
    if(adjust_return_button(ColoursForm, xevent)){
	return TRUE;
    }
    return standard_form_event_cb(ColoursForm, xevent, NULL, colours_help_func);
}

static void colours_set_wintitle(void){
    char buffer[MAXPATHLEN];
    sprintf(buffer, "Adjust Browser #%d Colours:", source_browser->get_ID());
    XStoreName(fl_display, ColoursForm->window, buffer);
}

void colours_show_form(void){
    /* set up the menu */
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(ColoursForm, FL_ALL_EVENT, colours_form_raw_cb);
	
	predefined_menu = new_pup();
	
	add_to_pup(predefined_menu, "Colour Schemes%t", 0);
	add_to_pup(predefined_menu, "Dark & Yellow%x1", 0,"Yy");
	add_to_pup(predefined_menu, "Dark & Red%x2",    0,"Rr");
	add_to_pup(predefined_menu, "Dark & Green%x3",  0,"Gg");
	add_to_pup(predefined_menu, "Reddish%x4",       0,"dD");
	add_to_pup(predefined_menu, "Greenish%x5",      0,"nN");
	add_to_pup(predefined_menu, "Blueish%x6",       0,"Bb");
	add_to_pup(predefined_menu, "Yellowish%x7",     0,"lL");
	add_to_pup(predefined_menu, "Whitish%x8",       0,"Ww");
	add_to_pup(predefined_menu, "Greyish%x9",       0,"eE");

	fl_set_object_shortcut(ColoursFormCloseReturnButton, "CcH^M", TRUE);
	fl_set_button_shortcut(ColoursFormPredefinedButton,  "Pp", TRUE);
	fl_set_button_shortcut(ColoursFormDefaultsButton,    "Dd", TRUE);
	fl_set_button_shortcut(ColoursFormSyncAllButton,     "yY^y", TRUE);
    }
    colours_form_sync();
    set_to_ret_button(ColoursFormCloseReturnButton);
    focus_save();
    if(!ColoursForm->visible){
	fl_show_form(ColoursForm, FL_PLACE_CENTER, TRUE, "Adjust Browser Colours:");
	colours_set_wintitle();
    } else {
	XMapRaised(fl_display, ColoursForm->window);
    }
    tool_focus_move(ColoursForm, ColoursFormCloseReturnButton);
}

/*
 * sync displayed colours with the source browser.
 */

void colours_form_sync(void){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      
	
    fl_set_object_color(ColoursFormCSFButton, CSF, 47);
    fl_set_object_color(ColoursFormCSBButton, CSB, 47);
    fl_set_object_color(ColoursFormCNFButton, CNF, 47);
    fl_set_object_color(ColoursFormCNBButton, CNB, 47);
    fl_set_object_color(ColoursFormNSFButton, NSF, 47);
    fl_set_object_color(ColoursFormNSBButton, NSB, 47);
    fl_set_object_color(ColoursFormNNFButton, NNF, 47);
    fl_set_object_color(ColoursFormNNBButton, NNB, 47);    

    fl_set_object_color(ColoursFormSourceColourButton, tool_source_colour, 47);
    fl_set_object_color(ColoursFormTargetColourButton, tool_target_colour, 47);

    colours_set_wintitle();
}

static void browser_sync_st_colours(void){
    source_browser->sync_source_target_colour();
    if(source_browser != target_browser){
	target_browser->sync_source_target_colour();
    }
}

void ColoursFormSourceTargetButtonCB(FL_OBJECT *, long val){
    int &col = val ? tool_target_colour : tool_source_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? ColoursFormTargetColourButton : ColoursFormSourceColourButton, col, 47);
	browser_sync_st_colours();
    }
}

void ColoursFormCSFButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(CSF);
    if(new_col != CSF){
	CSF = new_col;
	fl_set_object_color(ColoursFormCSFButton, CSF, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormCNFButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(CNF);
    if(new_col != CNF){
	CNF = new_col;
	fl_set_object_color(ColoursFormCNFButton, CNF, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormNNFButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(NNF);
    if(new_col != NNF){
	NNF = new_col;
	fl_set_object_color(ColoursFormNNFButton, NNF, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormNSFButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(NSF);
    if(new_col != NSF){
	NSF = new_col;
	fl_set_object_color(ColoursFormNSFButton, NSF, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormCSBButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(CSB);
    if(new_col != CSB){
	CSB = new_col;
	fl_set_object_color(ColoursFormCSBButton, CSB, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormCNBButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(CNB);
    if(new_col != CNB){
	CNB = new_col;
	fl_set_object_color(ColoursFormCNBButton, CNB, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormNSBButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(NSB);
    if(new_col != NSB){
	NSB = new_col;
	fl_set_object_color(ColoursFormNSBButton, NSB, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormNNBButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int CSF, CSB, CNF, CNB, NSF, NSB, NNF, NNB;
    fl_get_panel_colors(source_browser->panel,
        &CSF, &CSB, &NSF, &NSB,
        &CNF, &CNB, &NNF, &NNB
    );				      

    int new_col = fl_show_colormap(NNB);
    if(new_col != NNB){
	NNB = new_col;
	fl_set_object_color(ColoursFormNNBButton, NNB, 47);    
	fl_set_panel_colors(source_browser->panel,
            CSF, CSB, NSF, NSB,
            CNF, CNB, NNF, NNB
        );				     
    }
}

void ColoursFormCloseReturnButtonCB(FL_OBJECT *, long){
    if(ColoursForm->visible){
	focus_restore_start();
	fl_hide_form(ColoursForm);
	focus_restore_finish();
    }
}

void ColoursFormPredefinedButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    switch(do_pup(predefined_menu)){
    case 1: source_browser->set_dark_yellow(); break;
    case 2: source_browser->set_dark_red();    break;
    case 3: source_browser->set_dark_green();  break;
    case 4: source_browser->set_reddish();     break;
    case 5: source_browser->set_greenish();    break;
    case 6: source_browser->set_blueish();     break;
    case 7: source_browser->set_yellowish();   break;
    case 8: source_browser->set_whitish();     break;
    case 9: source_browser->set_greyish();     break;
    default: return;
    }
    colours_form_sync();
}

