
#include <iostream.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>

#include "cstring.h"

//************************************************************
// CVarString Implementation
//************************************************************

CVarString& CVarString::operator=(const char *s)
{
    int nl= strlen(s);
    if(nl+1 >= allocated) grow((nl-allocated)+allocinc);
    assert(allocated > nl+1);
    strcpy(a, s);
    len= nl;
    return *this;
}

CVarString& CVarString::operator=(const CVarString& n)
{
    if(this != &n){
	if(n.len+1 >= allocated){ // if it is not big enough
#	    ifdef	DEBUG
	    fprintf(stderr, "~operator=(CVarString&) a= %p\n", a);
#	    endif
	    delete [] a; // get rid of old one
	    allocated= n.allocated;
	    allocinc= n.allocinc;
	    a= new char[allocated];
#	    ifdef	DEBUG
	    fprintf(stderr, "operator=(CVarString&) a= %p, source= %p\n", a, n.a);
#	    endif
	}
        len= n.len;
	strcpy(a, n.a);
    }
    return *this;
}

void CVarString::grow(int n)
{
    if(n == 0) n= allocinc;
    allocated += n;
    char *tmp= new char[allocated];
    strcpy(tmp, a);
#ifdef	DEBUG
    fprintf(stderr, "CVarString::grow() a= %p, old= %p, allocinc= %d\n", tmp, a, allocinc);
    fprintf(stderr, "~CVarString::grow() a= %p\n", a);
#endif
    delete [] a;
    a= tmp;
}

void CVarString::add(char c)
{
    if(len+1 >= allocated) grow();
    assert(allocated > len+1);
    a[len++]= c; a[len]= '\0';
}

void CVarString::add(const char *s)
{
    int nl= strlen(s);
    if(len+nl >= allocated) grow(((len+nl)-allocated)+allocinc);
    assert(allocated > len+nl);
    strcat(a, s);
    len+=nl;
}

void CVarString::add(int ip, const char *s)
{
    int nl= strlen(s);
    if(len+nl >= allocated) grow(((len+nl)-allocated)+allocinc);
    assert(allocated > len+nl);
    memmove(&a[ip+nl], &a[ip], (len-ip)+1); // shuffle up
    memcpy(&a[ip], s, nl);
    len+=nl;
    assert(a[len] == '\0');
}

void CVarString::remove(int ip, int n)
{
    assert(ip+n <= len);
    memmove(&a[ip], &a[ip+n], (len-ip)-n+1); // shuffle down
    len-=n;
    assert(a[len] == '\0');    
}

//************************************************************
// CVarString Implementation
//************************************************************

ostream& operator<<(ostream& os,  const CVarString& arr)
{
#ifdef TEST
    os << "(" << arr.length() << ")" << (const char *)arr;
#else
    os << (const char *)arr;
#endif
	
    return os;    
}

CVarString::CVarString(int n)
{
    a= new char[n];
    *a= '\0';
    len= 0;
    allocated= n;
    allocinc= n;
#   ifdef	DEBUG
    fprintf(stderr, "CVarString(int %d) a= %p\n", allocinc, a);
#   endif
}

CVarString::CVarString(const char* s)
{
    int n= strlen(s) + 1;
    a= new char[n];
    strcpy(a, s);
    len= n-1;
    allocated= n;
    allocinc= ALLOCINC;
#   ifdef	DEBUG
    fprintf(stderr, "CVarString(const char *(%d)) a= %p\n", allocinc, a);
#   endif
}

CVarString::CVarString(const char* s, int n)
{
    a= new char[n+1];
    if(n) strncpy(a, s, n);
    a[n]= '\0';
    len= n;
    allocated= n+1;
    allocinc= ALLOCINC;
#   ifdef	DEBUG
    fprintf(stderr, "CVarString(const char *, int(%d)) a= %p\n", allocinc, a);
#   endif
}

CVarString::CVarString(char c)
{
    int n= 2;
    a= new char[n];
    a[0]= c; a[1]= '\0';
    len= 1;
    allocated= n;
    allocinc= ALLOCINC;
#   ifdef	DEBUG
    fprintf(stderr, "CVarString(char (%d)) a= %p\n", allocinc, a);
#   endif
}


/*
const char CVarString::operator[](const int i) const
{
     assert((i >= 0) && (i < len) && (a[len] == '\0'));
     return a[i];
}
*/

char& CVarString::operator[](const int i) const
{
     assert((i >= 0) && (i < len) && (a[len] == '\0'));
     return a[i];
}

CVarString::CVarString(const CVarString& n)
{
    allocated= n.allocated;
    allocinc= n.allocinc;
    len= n.len;
    a= new char[allocated];
    strcpy(a, n.a);
#ifdef	DEBUG
    fprintf(stderr, "CVarString(CVarString&) a= %p, source= %p\n", a, n.a);
#endif

}

//************************************************************
// CString stuff
//************************************************************

// assignments
CString& CString::operator=(const CString& n)
{
    if(this == &n) return *this;
    pstr= n.pstr;
    return *this;
}

CString& CString::operator=(const CSubString& sb)
{
    CVarString tmp(sb.pt, sb.len);
    pstr= tmp;
    return *this;
}

// concatenations
CString CString::operator+(const CString& s) const
{
    CString ts(*this);
    ts.pstr.add(s);    
    return ts; 
}

CString CString::operator+(const char *s) const
{
    CString ts(*this);
    ts.pstr.add(s);
    return ts; 
}

CString CString::operator+(char c) const
{
    CString ts(*this);
    ts.pstr.add(c);
    return ts; 
}

CString operator+(const char *s1, const CString& s2)
{
    CString ts(s1);
    ts = ts + s2;
//    cout << "s2[0] = " << s2[0] << endl; // gives incorrect error
    return ts; 
}

//************************************************************
// other stuff
//************************************************************

char CString::chop(void)
{
    int n= length();
    if(n <= 0) return '\0'; // empty
    char tmp= pstr[n-1];
    pstr.remove(n-1);
    return tmp;
}

int CString::index(const CString& s, int offset)
{
    if(offset < 0) offset= 0;
    for(int i=offset;i<length();i++){
	if(strncmp(&pstr[i], s, s.length()) == 0) return i;
    }

    return -1;
}

int CString::rindex(const CString& s, int offset)
{
    if(offset == -1) offset= length()-s.length();
    else offset= offset-s.length()+1;
    if(offset > length()-s.length()) offset= length()-s.length();
      
    for(int i=offset;i>=0;i--){
	if(strncmp(&pstr[i], s, s.length()) == 0) return i;
    }
    return -1;
}

CSubString CString::substr(int offset, int len)
{
    if(len == -1) len= length() - offset; // default use rest of string
    if(offset < 0){
	offset= length() + offset;  // count from end of string
	if(offset < 0) offset= 0;   // went too far, adjust to start
    }
    return CSubString(*this, offset, len);
}

// this is private
// it shrinks or expands string as required
void CString::insert(int pos, int len, const char *s, int nlen)
{
    if(pos < length()){ // nothing to delete if not true
	if((len+pos) > length()) len= length() - pos;
	pstr.remove(pos, len);  // first remove subrange
    }else pos= length();

    CVarString tmp(s, nlen);
    pstr.add(pos, tmp);	    // then insert new CSubString
}

//
// I know! This is not fast, but it works!!
//
int CString::tr(const char *sl, const char *rl, const char *opts)
{
    if(length() == 0 || strlen(sl) == 0) return 0;

    int cflg= strchr(opts, 'c') != NULL; // thanks Michael
    int dflg= strchr(opts, 'd') != NULL;
    int sflg= strchr(opts, 's') != NULL;

    int i, cnt=0, flen=0;
    CString t;
    unsigned char lstc= '\0', fr[256];
    
    // build search array, which is a 256 byte array that stores the index+1
    // in the search string for each character found, == 0 if not in search
    memset(fr, 0, 256);
    for(i=0;i< (int) strlen(sl);i++){
	if(i && sl[i] == '-'){ // got a range
	    assert(i+1 < (int) strlen(sl) && lstc <= sl[i+1]); // sanity check
	    for(unsigned char c=lstc+1;c<=sl[i+1];c++){
		fr[c]= ++flen;
	    }
	    i++; lstc= '\0';
	}else{
	    lstc= sl[i];
	    fr[sl[i]]= ++flen;
	}
    }

    int rlen;
    // build replacement list
    if((rlen=strlen(rl)) != 0){
	for(i=0;i<rlen;i++){
	    if(i && rl[i] == '-'){ // got a range
		assert(i+1 < rlen && t[t.length()-1] <= rl[i+1]); // sanity check
		for(char c=t[i-1]+1;c<=rl[i+1];c++) t += c;
		i++;
	    }else t += rl[i];
	}
    }

    // replacement string that is shorter uses last character for rest of string
    // unless the delete option is in effect or it is empty
    while(!dflg && rlen && flen > t.length()){
	t += t[t.length()-1]; // duplicate last character
    }

    rlen= t.length(); // length of translation string   
   
    // do translation, and deletion if dflg (actually falls out of length of t)
    // also squeeze translated characters if sflg

    CString tmp; // need this in case dflg, and string changes size
    for(i=0;i<length();i++){
	int off;
	if(cflg){ // complement, ie if NOT in f
	    char rc= !dflg ? t[t.length()-1] : '\0'; // always use last character for replacement
	    if((off=fr[(*this)[i]]) == 0){ // not in map
		cnt++;
		if(!dflg && (!sflg || tmp.length() == 0 || tmp[tmp.length()-1] != rc))
		    tmp += rc;
	    }else tmp += (*this)[i]; // just stays the same
	}else{ // in fr so substitute with t, if no equiv in t then delete
	    if((off=fr[(*this)[i]]) > 0){
		off--; cnt++;
		if(rlen==0 && !dflg && (!sflg || tmp.length() == 0 || tmp[tmp.length()-1] != (*this)[i])) tmp += (*this)[i]; // stays the same
		else if(off < rlen && (!sflg || tmp.length() == 0 || tmp[tmp.length()-1] != t[off]))
		    tmp += t[off]; // substitute
	    }else tmp += (*this)[i]; // just stays the same
	}
    }

    *this= tmp;
    return cnt;
}

//************************************************************
// streams stuff
//************************************************************

istream& operator>>(istream& ifs, CString& s){
    char c = '\0';
#if 0
    char buf[40];
#else
    char buf[132];
#endif

    s= ""; // empty string
    ifs.get(buf, sizeof buf); 
    // This is tricky because a line teminated by end of file that is not terminated
    // with a '\n' first is considered an OK line, but ifs.good() will fail.
    // This will correctly return the last line if it is terminated by eof with the
    // stream still in a non-fail condition, but at eof, so next call will fail as
    // expected
    if(ifs){ 		// previous operation was ok
        s += buf; 	// append buffer to string
//	cout << "<" << buf << ">" << endl;
	// if its a long line continue appending to string
	while(ifs.good() && (c=ifs.get()) != '\n'){
//	    cout << "eof1= " << ifs.eof() << endl;
	    ifs.putback(c);
//          cout << "eof2= " << ifs.eof() << endl;
	    if(ifs.get(buf, sizeof buf)) s += buf; // append to line
	}
    }
    return ifs;    
}

ostream& operator<<(ostream& os,  const CString& arr)
{
#ifdef TEST
    os << "(" << arr.length() << ")" << "\"";
    os << (const char *)arr;
    os << "\"";
#else
    os << (const char *)arr;
#endif
    return os;   
}

void CSubString::operator=(const char *s) {
    str.insert(pos, len, s, strlen(s));
}

void CSubString::operator=(const CString& s) {
    if (&str == &s) { // potentially overlapping
	CVarString tmp(s);
	str.insert(pos, len, tmp, strlen(tmp));
    } else {
	str.insert(pos, len, s, s.length());
    }
}

void CSubString::operator=(const CSubString& s) {
    if(&str == &s.str){ // potentially overlapping
	CVarString tmp(s.pt, s.len);
	str.insert(pos, len, tmp, strlen(tmp));
    }else str.insert(pos, len, s.pt, s.len);
}

CSubString::CSubString(CString& os, int p, int l) : str(os) {
    if(p > os.length()) p= os.length();
    if((p+l) > os.length()) l= os.length() - p;
    pos= p; len= l;
    if(p == os.length()){
	pt= 0; // append to end of string
    } else {
	pt= &os.pstr[p];       
        // +++ WARNING this may be illegal as nested classes
	// can't access its enclosing classes privates!
    }
}



