
#ifndef	CSTRING_H
#define	CSTRING_H

#include <assert.h>
#include <string.h>
#include <stdio.h>

//****************************************************************
// just a mechanism for self deleteing strings which can be hacked
//****************************************************************

const int ALLOCINC = 32;

class CTempString
{
private:
    char *str;
public:
    CTempString(const char *s)    
    {
	str= new char[strlen(s) + 1];
        strcpy(str, s);
    }
    
    CTempString(const char *s, int len)    
    {
        str= new char[len + 1];
        if(len) strncpy(str, s, len);
        str[len]= '\0';
    }

    ~CTempString(){ delete [] str; }

    operator char*() const { return str; }
};

//************************************************************
// This class takes care of the mechanism behind variable
// length strings
//************************************************************

class CVarString
{
private:
//    enum{ALLOCINC=32};
    char *a;
    int len;
    int allocated;
    int allocinc;
    inline void grow(int n= 0);

public:
    CVarString(int n= 32);
    CVarString(const CVarString& n);
    CVarString(const char *);
    CVarString(const char* s, int n);
    CVarString(char);

    ~CVarString(){
#       ifdef	DEBUG
	fprintf(stderr, "~CVarString() a= %p, allocinc= %d\n", a, allocinc);
#       endif
	delete [] a;
    }

    CVarString& operator=(const CVarString& n);
    CVarString& operator=(const char *);

    //inline const char operator[](const int i) const;
    char& operator[](const int i) const;

    operator const char *() const{ return a; }

    int length(void) const{ return len; }

    void add(char);
    void add(const char *);
    void add(int, const char *);
    void remove(int, int= 1);

    void erase(void){ len= 0; }
};

//************************************************************
// Implements the perl specific string functionality 
//************************************************************

// This idea lifted from NIH class library -
// to handle CSubString LHS assignment
// Note if subclasses can't be used then take external and make
// the constructors private, and specify friend CString

class CString;

class CSubString {
public:	
    int pos, len;
    CString& str;
    char *pt;

    CSubString(CString& os, int p, int l);
    void operator=(const CString& s);
    void operator=(const CSubString& s);
    void operator=(const char *s);
};

class CString
{
private:
    CVarString pstr;  // variable length string mechanism
    friend class CSubString;
    
public:
    
    CString():pstr(){}
    CString(const CString& n) : pstr(n.pstr){}     
    CString(const char *s) : pstr(s){}
    CString(char *s) : pstr(s){}
    CString(const char c) : pstr(c){}
    CString(const CSubString& sb) : pstr(sb.pt, sb.len){}
    
    CString& operator=(const char *s){pstr= s; return *this;}        
    CString& operator=(const CString& n); 
    CString& operator=(const CSubString& sb);

    operator const char*() const{return pstr;}
    const char* GetBuffer() const{return pstr;}
    char operator[](int n) const{ return pstr[n]; }

    int length(void) const{ return pstr.length(); }
    
    char chop(void);
    
    int index(const CString& s, int offset= 0);    
    int rindex(const CString& s, int offset= -1);
    CSubString substr(int offset, int len= -1);
    //CSubString substr(const Range& r){ return substr(r.start(), r.length());}
        
    int m(const char *, const char *opts=""); // the regexp match m/.../ equiv
   
    int tr(const char *, const char *, const char *opts="");
    //int s(const char *, const char *, const char *opts="");

    //CStringList split(const char *pat= "[ \t\n]+", int limit= -1);
    
    int operator<(const CString& s) const { return (strcmp(pstr, s) < 0); }
    int operator>(const CString& s) const { return (strcmp(pstr, s) > 0); }
    int operator<=(const CString& s) const { return (strcmp(pstr, s) <= 0); }
    int operator>=(const CString& s) const { return (strcmp(pstr, s) >= 0); }
    int operator==(const CString& s) const { return (strcmp(pstr, s) == 0); }
    int operator!=(const CString& s) const { return (strcmp(pstr, s) != 0); }

    int operator<(const char *s) const { return (strcmp(pstr, s) < 0); }
    int operator>(const char *s) const { return (strcmp(pstr, s) > 0); }
    int operator<=(const char *s) const { return (strcmp(pstr, s) <= 0); }
    int operator>=(const char *s) const { return (strcmp(pstr, s) >= 0); }
    int operator==(const char *s) const { return (strcmp(pstr, s) == 0); }
    int operator!=(const char *s) const { return (strcmp(pstr, s) != 0); }

    friend int operator<(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) < 0); }
    friend int operator>(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) > 0); }
    friend int operator<=(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) <= 0); }
    friend int operator>=(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) >= 0); }
    friend int operator==(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) == 0); }
    friend int operator!=(const char *s, const CString& sp)  { return (strcmp(s, sp.pstr) != 0); }

    CString operator+(const CString& s) const;
    CString operator+(const char *s) const;
    CString operator+(char c) const;
    friend CString operator+(const char *s1, const CString& s2);

    CString& operator+=(const CString& s){pstr.add(s); return *this;}
    CString& operator+=(const char *s){pstr.add(s); return *this;}
    CString& operator+=(char c){pstr.add(c); return *this;}
 
private:
    void insert(int pos, int len, const char *pt, int nlen);
};

extern ostream& operator<<(ostream& os,  const CString& arr);
extern ostream& operator<<(ostream& os,  const CVarString& arr);



#endif



