
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

#include "browser.h"
//#include "cwd_forms_aux.h"
//#include "cmd_forms_aux.h"
#include "dir_menu.h"
#include "edit_forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "fkeybar_forms_aux.h"
#include "flspecial.h"
#include "help_forms_aux.h"
#include "license_forms_aux.h"
#include "make.h"
#include "options_forms_aux.h"
#include "pathnames.h"
#include "resource.h"
#include "status_forms_aux.h"
#include "suffix.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "wm_forms_aux.h"
#include "xshell.h"

const char *tool_ID_string = "tool";

//    {"",              int_e,    (void *) &}, 

prm_entry_t tool_prm_table[] = {
    {"automouse_flag",              int_e,    (void *) &auto_mouse_flag}, 
    {"browser_autoraise_flag",      int_e,    (void *) &browser_c::auto_raise_flag}, 
    {"browser_dir_update_count",    int_e,    (void *) &browser_c::directory_update_count}, 
    {"cp_minusp_flag",              int_e,    (void *) &tool_cp_minusp_flag},
    {"dir_timeout_secs",            int_e,    (void *) &tool_dir_timeout_secs}, 
    {"du_auto_flag",                int_e,    (void *) &tool_du_auto_flag}, 
    {"du_sort_flag",                int_e,    (void *) &tool_du_sort_flag}, 
    {"editor_info_shown",           int_e,    (void *) &tool_editor_info_shown_flag},
    {"file_timeout_secs",           int_e,    (void *) &tool_file_timeout_secs}, 
    {"focus_on_startup_flag",       int_e,    (void *) &tool_focus_on_startup_flag},
    {"number_of_browsers",          int_e,    (void *) &tool_number_of_browsers}, 
    {"save_config_on_exit_state",   int_e,    (void *) &save_config_on_exit_state},
    {"save_tree_on_exit_state",     int_e,    (void *) &save_tree_on_exit_state},
    {"startup_msg_flag",            int_e,    (void *) &tool_startup_msg_flag},
    {"fontsize",                    int_e,    (void *) &tool_fontsize}, 
    {"fontstyle",                   int_e,    (void *) &tool_fontstyle}, 
    {"menu_fontsize",               int_e,    (void *) &tool_menu_fontsize}, 
    {"menu_fontstyle",              int_e,    (void *) &tool_menu_fontstyle}, 
    {"source_colour",               int_e,    (void *) &tool_source_colour}, 
    {"target_colour",               int_e,    (void *) &tool_target_colour}, 
    {"default_editor",              char_p_e, (void *) default_editor},
    {"",                            int_e,  NULL}
};

void tool_defaults_backup(void) {
    const char *fname = tool_default_fname;
    char buffer[2*MAXPATHLEN+64];
    sprintf(buffer, "%s %s %s.old", cp_path, fname, fname);
    safe_system(buffer);  
}

int tool_defaults_save(const char *fname){

    if(!fname) fname = tool_default_fname;

    status_msg(1, "saving defaults to %s", fname);

    // make a copy of defaults file if it exists. Only do this for ucdb.
    if(strcmp(fname, "ucdb") == 0 && access(fname, F_OK) == 0){
	tool_defaults_backup();
    }
    
    ofstream ofile;
    ofile.open(fname);
    if(!ofile){
	error_msg("UC internal: unable to open database file %s.", fname);
	return FALSE;
    } else {
	ofile << "# Unix Cockpit version " << version_major << ".";
	ofile << version_minor << " resource data base." << endl;
	ofile << "# Better don't edit :-)" << endl << endl << endl;

	tool_number_of_browsers = browser_c::next_ID-1;    // kludge
	resource_save_defaults(ofile, tool_prm_table, tool_ID_string);
	
	fkeybar_save_defaults(ofile);
	help_save_defaults(ofile);
	status_save_defaults(ofile);
	toolbox_save_defaults(ofile);
	tree_save_defaults(ofile);
	edit_save_defaults(ofile);
	wm_save_defaults(ofile);
	xshell_save_defaults(ofile);
	dir_menu_save_defaults(ofile);
//	cmd_save_cmd_list(ofile, "cmd");
//	cwd_save_defaults(ofile);
	
	ofile << endl;

	// ring - find first browser, work your way through the ring 
	browser_c *tmp_browser, *tmp_browser_0;
        tmp_browser = tmp_browser_0 = source_browser->rb_get_last_browser()->next_browser;
	do {
	    tmp_browser->defaults_save(ofile);
	    tmp_browser = tmp_browser->next_browser;
	} while (tmp_browser != tmp_browser_0);

	ofile << "# End of UC defaults." << endl;
	if(!ofile) M_err("tool_save_defaults","error saving database file");
	ofile.close();
    }

    status_msg(1, "done.");
    return TRUE;
}

/* return success */

int tool_defaults_load(const char *fname, int do_sync_flag){
    M_info("tool_defaults_load", "start");
    tool_number_of_browsers = 2; // safety first.

    makefile_load();
    suffix_file_load();

    // input file stream operations
    ifstream ifile;
    if(!fname) fname = tool_default_fname;
    ifile.open(fname);

    if(!ifile){
	tool_looser_flag = TRUE;
	M_trace("tool_defaults_load", "no database");
	return FALSE;
    }

    // announce new defaults to all the components who need it:
    dir_menu_new_defaults();
    
    // default configuration: 2 browsers.
    // can be overridden by default data base.

    const int bufsize = 2*MAXPATHLEN;
    char bbuf[bufsize];
    while(ifile.getline(bbuf, bufsize) ){
	M_trace("tool_defaults_load:", "processing %s", bbuf);
	if(bbuf[0] == '#' || bbuf[0] == '\0' || bbuf[0] == ' '){
	    M_trace("tool_defaults_load:", "line empty, comment, or blank");
	    continue;
	    // it is a comment, or empty line, or starts with a blank.
	}
	// look for object ID token.
	char *obj_tok = strtok(bbuf, ".");
	if(!obj_tok){
	    cerr << "UC: strange line in default database file - no object.\n";
	    continue;
	}

	// look for name token.
	char *name_tok = strtok(NULL, ":");
	if(!name_tok){
	    cerr << "UC: strange line in default database file - no item for `" << obj_tok << "'.\n";
	    continue;
	}

	// look for value (use the remaining part of the line, to allow spaces.
	// this works wonders for string defaults, such as the default editor.
	//char *val_tok = strtok(NULL, " ");
	char *val_tok = strtok(NULL, "");
	if(!val_tok){
	    cerr << "UC: strange line in default database file - no value given for `"
		 << obj_tok << "." << name_tok << "'.\n";
	    continue;
	} 
	// strip leading whitespace.
	while(*val_tok == ' ' || *val_tok == '\t'){
	    val_tok++;
	}
	
	//cerr << "got valtok: `" << val_tok << "'\n";

	if(help_load_default        (obj_tok, name_tok, val_tok)) continue;
	if(dir_menu_load_default    (obj_tok, name_tok, val_tok)) continue;
	if(edit_load_default        (obj_tok, name_tok, val_tok)) continue;
	if(fkeybar_load_default     (obj_tok, name_tok, val_tok)) continue;
	if(status_load_default      (obj_tok, name_tok, val_tok)) continue;
	if(toolbox_load_default     (obj_tok, name_tok, val_tok)) continue;
	if(tree_load_default        (obj_tok, name_tok, val_tok)) continue;
	if(wm_load_default          (obj_tok, name_tok, val_tok)) continue;
	if(xshell_load_default      (obj_tok, name_tok, val_tok)) continue;
	if(resource_load_default    (tool_prm_table,  tool_ID_string,     obj_tok, name_tok, val_tok)) continue;
	// else it's something different, e.g. a browser parameter.
    }
    ifile.close();
    bound_value(tool_number_of_browsers, 1, 20);  // safety
    if(do_sync_flag){
	tool_sync_defaults();
    }
    M_info("tool_defaults_load", "returning true.");
    return TRUE;
}
