/*
 * all extra functions needed besides automatic stuff 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#include "browser.h"
#include "forms.h"
#include "dir_menu.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "linklist.h"
#include "menu.h"
#include "resource.h"
#include "status_forms_aux.h"

static linklist_c dirlist, last_dirlist;

const char *dir_menu_ID_string = "dir_menu";
//static int defaults_loaded_flag = FALSE;

static int needs_remake_flag = TRUE;

static void dir_menu_remember(const char *dir);

void dir_menu_setup_default_dirs(void){
    dirlist.add(strdup("/"));
    char *home_string = getenv("HOME");
    if(!home_string){
	program_unstable("Could get home directory !");
    } else {
	dirlist.add(strdup(home_string));
    }
    needs_remake_flag = TRUE;
}

/*
    static int first_time = TRUE;
    if(first_time){
	first_time = FALSE;
	if(!defaults_loaded_flag){
	    dir_menu_setup_default_dirs();
	}
	*/

static void dir_menu_clear_menus(void){
    while(dirlist.getsize()){
    	char *val = (char *) dirlist.getval();
		if (val) fl_free(val);
	dirlist.remove();
    }
    while(last_dirlist.getsize()){
    	char *val = (char *) last_dirlist.getval();
		if (val) fl_free(val);
		last_dirlist.remove();
    }
    //dir_menu_setup_default_dirs();
    needs_remake_flag = TRUE;
}
    
void dir_menu_save_defaults(ofstream &ofile){
    char *buf;
    ofile << "# dirlist resources.\n";
    ofile << "#\n";
    int n_items = dirlist.getsize();
    for(int i=0; i< n_items; i++){
	buf = (char *) dirlist[n_items - (i+1)];
	if(buf){
	    ofile << dir_menu_ID_string << ".dir_list_entry:      " << buf << endl;
	}
    }
}

void dir_menu_new_defaults(void){
//    defaults_loaded_flag = TRUE;
    dir_menu_clear_menus();
}


int dir_menu_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    if(!obj_tok || !name_tok || !val_tok) return FALSE;
    
    if(strcmp(obj_tok, dir_menu_ID_string) == 0){
	if(strcmp(name_tok, "dir_list_entry") == 0){
	    dir_menu_remember(val_tok);
	    return TRUE;
	}
    }
    return FALSE;
}

static long make_sub_menu(long cb_val, const char *title = NULL){
    char buffer[MAXPATHLEN + 64], *val;
    long menu;
    menu = new_pup();

    if(title){
	sprintf(buffer, "%s%s", title, "%t");
	add_to_pup(menu, buffer, 0);
    }

    int size = dirlist.getsize();
    if(size <= 0){
	dir_menu_setup_default_dirs();
	size = dirlist.getsize();
    }
    for(int i=0; i<size; i++){
	val = (char *) dirlist[i];
	if(val){
	    if(i<10){
		sprintf(buffer, "%d:  %s%cx%d", i, val, '%', i);
		char digit[8];
		sprintf(digit, "%d", i);
		add_to_pup(menu, buffer,   cb_val, digit);
	    } else {
		sprintf(buffer, "     %s%cx%d", val, '%', i);
		add_to_pup(menu, buffer,   cb_val, "");
	    }
	} else {
	    cerr << "illegal val in DirMenu::make_sub_menu. \n";
	}
    }
    return menu;
}

static void last_change_cb(int no){
    if(!source_is_stable()) return;
    if(no < 0 || no >= last_dirlist.getsize()){
	cerr << "Invalid no(" << no << ") in last_change_cb.\n";
	return;
    }
    char *val = (char *) last_dirlist[no];
    if(val){  
	tool_change_dir(val, TRUE, TRUE, TRUE, TRUE, FALSE);
    }
}

static long make_last_menu(void){
    char buffer[MAXPATHLEN + 64], *val;
    long menu;
    menu = new_pup();

    add_to_pup(menu, "Directory History%t", 0);

    int size = last_dirlist.getsize();
    if(size <= 0){
	dir_menu_setup_default_dirs();
	size = last_dirlist.getsize();
    }
    for(int i=0; i<size; i++){
	val = (char *) last_dirlist[i];
	if(val){
	    if(i<10){
		sprintf(buffer, "%d:  %s%cx%d", i, val, '%', i);
		char digit[8];
		sprintf(digit, "%d", i);
		add_to_pup(menu, buffer,   (long) last_change_cb, digit);
	    } else {
		sprintf(buffer, "     %s%cx%d", val, '%', i);
		add_to_pup(menu, buffer,   (long) last_change_cb, "");
	    }
	} else {
	    cerr << "illegal val in DirMenu::make_last_menu. \n";
	}
    }
    return menu;
}

static void forget_all_cb(int){
    if(ask_question(FALSE, "Really clear directory list ?")){
	dir_menu_clear_menus();
    }
}

static void dir_menu_remember(const char *dir){
    if(!dir) return;
    dirlist.headend();
    dirlist.insert(strdup(dir));
    needs_remake_flag = TRUE;
}

static void remember_cb(int){
    dir_menu_remember(source_browser->get_cwd());
}

static void forget_cb(int no){
    if(!source_is_stable()) return;
    if(no < 0 || no >= dirlist.getsize()){
	cerr << "Invalid no(" << no << ") in forget_cb.\n";
	return;
    }
    char *val = (char *) dirlist[no];
    if(val){  
	dirlist.remove();
	fl_free(val);
    }
    needs_remake_flag = TRUE;
}

static void change_cb(int no){
    if(!source_is_stable()) return;
    if(no < 0 || no >= dirlist.getsize()){
	cerr << "Invalid no(" << no << ") in change_cb.\n";
	return;
    }
    char *val = (char *) dirlist[no];
    if(val){  
	tool_change_dir(val, TRUE, TRUE, TRUE, TRUE, FALSE);
    }
}

static void sync_cb(int){
    if(!source_is_stable()) return;
    if(source_browser != target_browser){
	tool_change_dir(target_browser->get_cwd(), TRUE, TRUE, TRUE, TRUE, TRUE);
    }
}

static void sync_reverse_cb(int){
    if(!source_is_stable()) return;
    if(source_browser != target_browser){
	tool_change_target_dir(source_browser->get_cwd(), TRUE, TRUE);
    }
}

static void sync_all_cb(int){
    if (!source_is_stable()) return;
    const char *cwd_buf = source_browser->get_cwd();
    if (!cwd_buf) return;
    browser_c *tmp_browser = source_browser->next_browser;
    while (tmp_browser != source_browser) {
	if(tmp_browser == target_browser){
	    tool_change_target_dir(cwd_buf, TRUE, TRUE);
	} else {
	    tmp_browser->set_cwd_buf(cwd_buf);
	    tmp_browser->update();
	}
	tmp_browser = tmp_browser->next_browser;
    }
}

static long dir_menu=0, change_menu=0, forget_menu=0, last_menu=0;

static void free_dir_menu(void){
    if(change_menu)free_pup(change_menu);
    if(dir_menu)   free_pup(dir_menu);
    if(forget_menu)free_pup(forget_menu);
    if(last_menu)  free_pup(last_menu);

    change_menu = dir_menu = forget_menu = last_menu = 0;
}

static void make_dir_menu(int set_all_flag) {
    free_dir_menu();
    
    change_menu = make_sub_menu((long) change_cb, "Change Dir");
    forget_menu = make_sub_menu((long) forget_cb, "Forget Dir");
    last_menu   = make_last_menu();
    
    dir_menu = new_pup();
    add_to_pup(dir_menu, "Directory Stack (Alt-d, Esc-d)%t", 0);
    add_to_pup(dir_menu, "History ...%m",       last_menu,               "Hh");
    add_to_pup(dir_menu, "Change to ... %m",    change_menu,             "Cc");
    add_to_pup(dir_menu, "Forget ...%m",        forget_menu,             "Ff");
    add_to_pup(dir_menu, "Forget All%f",        (long) forget_all_cb,    "Aa");
    add_to_pup(dir_menu, "Remember Current%f",  (long) remember_cb,      "Rr");
    add_to_pup(dir_menu, "set Source:=Target%f", (long) sync_cb,         "Ss");
    add_to_pup(dir_menu, "set Target:=Source%f", (long) sync_reverse_cb, "Tt");

    if (set_all_flag) {
	add_to_pup(dir_menu, "set All:=Source%f", (long) sync_all_cb, "Aa");
    }
    needs_remake_flag = FALSE;
}

long get_dir_menu(void){
    static int set_all_flag = FALSE;
    int new_set_all_flag = (browser_c::next_ID > 2); // 2, because sb can be tb, one passive..
    if(needs_remake_flag || new_set_all_flag != set_all_flag){
	set_all_flag = new_set_all_flag;
	make_dir_menu(set_all_flag);
    }
    return dir_menu;
}

void dir_menu_cb(void){
    long menu = get_dir_menu();
    if(menu){
	do_pup(menu);
    }
}

void dir_menu_remember_last(const char *dir){
    char *val;
    if(!dir) return;
    int size = last_dirlist.getsize();
    // remove duplicates before insert at top.
    if(size > 0){
	for(int i=0; i<size; i++){
	    val = (char *) last_dirlist[i];
	    if(val && (strcmp(val, dir) == 0)){
		fl_free(val);
		last_dirlist.remove();
		size--;
	    }
	}
    }
    if (size > 30) {
    	last_dirlist.tailend();
    	char *val = (char *) last_dirlist.getval();
	if (val) fl_free(val);
	last_dirlist.remove();
    }
    last_dirlist.headend();
    last_dirlist.insert(strdup(dir));
    needs_remake_flag = TRUE;
}

/*
main(int argc, char *argv[]){
    fl_initialize("uc", "UC", NULL, 0, &argc, argv);
    do_pup(make_dir_menu());
    free_dir_menu();
}
*/
