#include <sys/param.h>
#include <sys/stat.h>

#ifdef LINUX_GLIBC2
#       include <errno.h>
#endif

#include "cb_fkey.h"
#include "dirinfo_forms.h"
#include "dirinfo_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "scroll_events.h"
#include "superls.h"

class dirinfo_c {
    char fullpath[MAXPATHLEN+16];
    void fs_command(FL_OBJECT *browser);
public:
    FD_DirinfoForm *DirinfoFD;
    dirinfo_c(void);
    int show(const char *dir);
    ~dirinfo_c();
    void auto_sort_sync(void);
    void du(void);
    void update(void);
    void freeze  (void){fl_freeze_form  (DirinfoFD->DirinfoForm);}
    void unfreeze(void){fl_unfreeze_form(DirinfoFD->DirinfoForm);}
};

static int dirinfo_atclose_cb(FL_FORM* form, void*){
    if(form && form->u_vdata){
	delete (dirinfo_c *) form->u_vdata;
    } else {
	M_err("dirinfo_atclose","invalid ptrs");
    }
    return FL_IGNORE;
}

void dirinfo_context_func(void){
    load_help_browser(dirinfo_help_text, "Dirinfo");
}

static int dirinfo_raw_cb(FL_FORM *form, void *xev){
    if(!form || !form->u_vdata) return FALSE;
    dirinfo_c *dirinfo = (dirinfo_c *) form->u_vdata;
    tool_got_event_flag = TRUE;

    XEvent *xevent = (XEvent *) xev;
    KeySym keysym;
    char keybuf[1];
    
    switch(xevent->type){
    case KeyPress:
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	case XK_F1:
	    help_event_cb(dirinfo_context_func);
	    return TRUE;
	case XK_F2:
	    if(tool_get_meta()){
		OptionsButtonCB(NULL,0);
	    } else {
		//DirinfoMenuButtonCB(dirinfo->DirinfoFD->MenuButton, 0);
	    }
	    return TRUE;
	case SunXK_F37: if(!sun_keyboard_flag) return FALSE;
	case XK_F12:
	    ExitButtonCB(NULL, 0);
	    return TRUE;
	case XK_Escape:
	    delete dirinfo;
	    return TRUE;
	default:
	    break;
	}
    }
    if(adjust_return_button(form, xevent)) return TRUE;
    if(scroll_browser_event_cb(dirinfo->DirinfoFD->Browser, xevent)) return TRUE;
    return FALSE;
}

dirinfo_c::dirinfo_c(void){
    DirinfoFD = create_form_DirinfoForm();
    DirinfoFD->DirinfoForm->u_vdata = this;
    fl_set_form_atclose(DirinfoFD->DirinfoForm, dirinfo_atclose_cb, NULL);
    fl_register_call_back(DirinfoFD->DirinfoForm, FL_ALL_EVENT, dirinfo_raw_cb);

    fl_set_object_shortcut(DirinfoFD->CloseReturnButton, "Cc^M", TRUE);
    fl_set_button_shortcut(DirinfoFD->UpdateButton, "Uu", TRUE);
    fl_set_button_shortcut(DirinfoFD->AutoButton, "Aa", TRUE);
    fl_set_button_shortcut(DirinfoFD->SortButton, "Ss", TRUE);
    fl_set_button_shortcut(DirinfoFD->DUButton, "Dd", TRUE);
    fl_set_button_shortcut(DirinfoFD->HelpButton, "Hh", TRUE);

    fl_set_browser_fontstyle(DirinfoFD->Browser, FL_FIXED_STYLE);
    fl_set_browser_fontsize (DirinfoFD->Browser, FL_SMALL_SIZE);
}

dirinfo_c::~dirinfo_c(){
    if(DirinfoFD->DirinfoForm->visible){
	focus_restore_start();
	fl_hide_form(DirinfoFD->DirinfoForm);
	focus_restore_finish();
    }
    fl_free_form(DirinfoFD->DirinfoForm);
}

void dirinfo_c::auto_sort_sync(void){
    fl_set_button(DirinfoFD->AutoButton, tool_du_auto_flag);
    fl_set_button(DirinfoFD->SortButton, tool_du_sort_flag);
}

// linux, dec, freebsd, ibm, solaris, sgi & unixware have -k.
// sunos, hp don't.

#if defined(AIX) || defined(GENERICBSD) || defined(SOLARIS) || defined(SGI) || defined(DEC_ALPHA) || defined(LINUX) || defined(UNIXWARE) || defined(SINIX)
const char * du_opt = "-k", *du_title = "k";
#else
const char * du_opt = "", *du_title = "blocks";
#endif

void dirinfo_c::du(void){
    char buf[2*MAXPATHLEN];
    sprintf(buf, "%s\"%sDisc Use (%s) of %s\" -e sh -c 'cd %s; du %s %s; echo \"press Enter to continue\"; read x' & ",
	"xterm -bg black -fg white -sb -sl 640 -geometry 80x40+300+200 -title ",
	tool_du_sort_flag ? "Please wait for sorted " : "",
	du_title,
	fullpath,
	fullpath,
	du_opt,
	(tool_du_sort_flag ? "| sort -n" : "")
    );
    safe_system(buf);
}

int dirinfo_c::show(const char *dir){
    // only absolute dirnames. Why ?
    if(!dir || dir[0] != '/') return FALSE;
    char rbuf[MAXPATHLEN];
    if(!real_path(dir, rbuf)){
	timed_error_notifier("Invalid directory passed to dirinfo_c::show", rbuf);
	return FALSE;
    }
    sprintf(fullpath, "UCDirInfo: %s", get_basename(rbuf));
    focus_save();
    auto_sort_sync();
    fl_set_browser_leftslider(DirinfoFD->Browser, TRUE);
    fl_show_form(DirinfoFD->DirinfoForm,FL_PLACE_FREE_CENTER,FL_FULLBORDER, fullpath);
    //fl_set_form_size(DirinfoFD->DirinfoForm, 450, 300);
    tool_focus_move(DirinfoFD->DirinfoForm, DirinfoFD->CloseReturnButton);
    strcpy(fullpath, rbuf);
    update();
    return TRUE;
}

#ifdef SGINOTYET
#include <sys/types.h>
#include <sys/statvfs.h>

void dirinfo_c::fs_command(FL_OBJECT *browser){
    char buf[MAXPATHLEN];
    if(!browser) return;
    struct statvfs vfsbuf;
    if(statvfs(fullpath, &vfsbuf) < 0){
	sprintf(buf, "Cannot stat directory: %s", strerror(errno));
	fl_add_browser_line(browser, buf);
	return;
    }
    sprintf(buf, "Filesystem   : %s %s %s, %s avail",
	    ((vfsbuf.f_flag & ST_LOCAL  ) ? "local" : "remote"),
	    ((vfsbuf.f_flag & ST_RDONLY ) ? "read-only" : "read-write"),
	    vfsbuf.f_basetype, blocks_to_string(vfsbuf.f_bavail, vfsbuf.f_bsize)
    );
    fl_add_browser_line(browser, buf);

    ulong blocks_used;
    if(vfsbuf.f_blocks < vfsbuf.f_bfree){
	blocks_used = vfsbuf.f_blocks;     // proc file systems.
    } else {
	blocks_used = vfsbuf.f_blocks - vfsbuf.f_bfree;
    }
    if(blocks_used + vfsbuf.f_bavail == 0) blocks_used++;
    ulong blocks_pused = (blocks_used * 100.0 / (blocks_used + vfsbuf.f_bavail) + 0.5);
    sprintf(buf, "Blocks used  : %lu (%lu%%) of %lu total, %ld bytes each", blocks_used, blocks_pused , vfsbuf.f_blocks, vfsbuf.f_bsize);
    fl_add_browser_line(browser, buf);
    
    // cerr << vfsbuf.f_frsize << endl;
    // cerr << vfsbuf.f_bavail << endl;

    if(!vfsbuf.f_files) vfsbuf.f_files = 1; // safeguard.
    ulong inodes_used  = vfsbuf.f_files - vfsbuf.f_ffree;
    ulong inodes_pused = (inodes_used * 100.0 / vfsbuf.f_files + 0.5);
    sprintf(buf, "Inodes used  : %lu (%lu%%) of %lu total", inodes_used, inodes_pused , vfsbuf.f_files);
    fl_add_browser_line(browser, buf);

//    cerr << vfsbuf.f_favail << endl;
//    cerr << vfsbuf.f_fsid << endl;  // device ? map. 
//    cerr << vfsbuf.f_flag << endl;
//    cerr << vfsbuf.f_namemax << endl; // useless
//    cerr << vfsbuf.f_fstr << endl;    // empty
//    cerr << vfsbuf.f_ << endl;
    /*
    cerr << ((vfsbuf.f_flag & ST_NOSUID ) ? "doesn't support suid.\n" : "does support suid\n");
    cerr << ((vfsbuf.f_flag & ST_NOTRUNC) ? "doesn't truncate long names.\n" : "does truncate long names\n");
    cerr << ((vfsbuf.f_flag & ST_NODEV  ) ? "doesn't allow opening of device files.\n" : "allows opening of device files.\n");
    cerr << ((vfsbuf.f_flag & ST_GRPID  ) ? "assigns GID from directory.\n" : "doesn't assign GID from directory\n");
    */
}

#endif

#include <stdio.h>
#include <stdlib.h>
#include <iomanip.h>
#include <iostream.h>
#include <strstream.h>

void dirinfo_c::fs_command(FL_OBJECT *browser){
    if(!browser) return;
    //char buf[MAXPATHLEN];
#ifdef HPUX
    // fl_add_browser_line(browser, "No file system stats avail on HPUX);
    return;
#endif
    
#if defined(GENERICBSD) || defined(SOLARIS) || defined(SGI) || defined(DEC_ALPHA) || defined(LINUX) || defined(UNIXWARE) || defined(SINIX)
    char *cmd = "df -k .";
#elif defined(HPUX)
    char *cmd = "df ";  // hp-sux
#else
    char *cmd = "df .";  // aix, sunos, unixware.
#endif
    char buf0[BUFSIZ], buf1[BUFSIZ], tmp_buf0[BUFSIZ], tmp_buf1[BUFSIZ], tmp_buf2[BUFSIZ];
    FILE *pipe_p;

    if ((pipe_p = popen(cmd, "r")) != NULL){
	if(fgets(buf0, BUFSIZ, pipe_p) && fgets(buf1, BUFSIZ, pipe_p)){
	    istrstream str0(buf0), str1(buf1);
	    strcpy(tmp_buf1, "");
	    int count = 0;
	    while(str0 && str1 && strcmp(tmp_buf0, "Mounted") != 0){
#if (defined(SUN) && !defined(SOLARIS)) || defined(LINUX)	
		str0 >> tmp_buf0;
		str1 >> tmp_buf1;
#else
		str0 >> setw(BUFSIZ) >> tmp_buf0;
		str1 >> setw(BUFSIZ) >> tmp_buf1;
#endif
		//cerr << "'" << tmp_buf0 << "'" << tmp_buf1 << "'" << endl;
#ifdef AIX
		// exception for `Total KB'
		if(count == 1){
		  str0 >> tmp_buf0;
		  strcpy(tmp_buf0, "Total KB");
		}
#endif
		if(tmp_buf0 && tmp_buf1){
		    if(strcmp(tmp_buf0, "Mounted") == 0){
			sprintf(tmp_buf0, "Mounted on");
		    }
		    if(strcmp(tmp_buf0, "on") != 0){
			sprintf(tmp_buf2, "%-13s: %s", tmp_buf0, tmp_buf1);
			fl_add_browser_line(browser, tmp_buf2);
		    }
		}
		count++;
	    }
	} else {
	    sprintf(buf0, "Unable to obtain disc free stats: %s", strerror(errno));
	    fl_add_browser_line(browser, buf0);
	}
	pclose(pipe_p);
    }
}

void dirinfo_c::update(void){
    char buf[2*MAXPATHLEN];
    freeze();
    FL_OBJECT *browser = DirinfoFD->Browser;
    fl_clear_browser(browser);
    sprintf(buf, "Full path    : %s", fullpath);
    fl_add_browser_line(browser, buf);
    struct stat statbuf;
    if(stat(fullpath, &statbuf) < 0){
	sprintf(buf, "Can' stat file (%s) !", strerror(errno));
	fl_add_browser_line(browser, buf);
	unfreeze();
	return;
    }
    if(!S_ISDIR(statbuf.st_mode)){
	unfreeze();
	return;
    }
    int dummy = 0;
    char buf2[MAXPATHLEN];
    sprintf(buf2, getfmtstr ('o', 3, 0, 1), statbuf.st_mode & (~S_IFMT));
    
    sprintf(buf, "Permissions  : %s (%s)", aperms(statbuf.st_mode, "%p",&dummy), buf2);
    fl_add_browser_line(browser, buf);

    const char *usr_name = getpwname(statbuf.st_uid);
    if(!usr_name) usr_name = "<unknown>";
    sprintf(buf, "Owner        : %s (OID %d)", usr_name, statbuf.st_uid);
    fl_add_browser_line(browser, buf);
    
    const char *grp_name = getgrname(statbuf.st_gid);
    if(!grp_name) grp_name = "<unknown>";
    sprintf(buf, "Group        : %s (GID %d)", grp_name, statbuf.st_gid);
    fl_add_browser_line(browser, buf);

    sprintf(buf, "Size,Links   : %s, %d", bytes_to_string(statbuf.st_size), statbuf.st_nlink);
    fl_add_browser_line(browser, buf);

    dummy = 0;
    sprintf(buf, "Access       : %s", fmtdate (statbuf.st_atime, "%h %d 19%y %T", &dummy));
    fl_add_browser_line(browser, buf);

    dummy = 0;
    sprintf(buf, "Status Change: %s", fmtdate (statbuf.st_ctime, "%h %d 19%y %T", &dummy));
    fl_add_browser_line(browser, buf);

    dummy = 0;
    sprintf(buf, "Modification : %s", fmtdate (statbuf.st_mtime, "%h %d 19%y %T", &dummy));
    fl_add_browser_line(browser, buf);

    fs_command(browser);
        
    unfreeze();
    if(tool_du_auto_flag){
	du();
    }
}

void dirinfo(const char *dir){
    if(!dir) return;
    dirinfo_c *dirinfo = new dirinfo_c;
    if(!dirinfo->show(dir)){
	delete dirinfo;
    }
}

static int vob_valid(FL_OBJECT *ob, const char *msg = ""){
    if(!ob || !ob->form || !ob->form->u_vdata){
	M_err("Dirinfo", "Invalid ptrs (%s)", msg);
	return FALSE;
    } else {
	return TRUE;
    }
}

void DirinfoCloseReturnButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Close")) return;
    delete (dirinfo_c *) ob->form->u_vdata;
}

void DirinfoUpdateButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Update")) return;
    ((dirinfo_c *) ob->form->u_vdata)->update();
}

void DirinfoHelpButtonCB(FL_OBJECT *, long){
    dirinfo_context_func();
}

void DirinfoAutoButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Auto")) return;
    toggle_flag(tool_du_auto_flag);
    ((dirinfo_c *) ob->form->u_vdata)->auto_sort_sync();
}

void DirinfoSortButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Sort")) return;
    toggle_flag(tool_du_sort_flag);
    ((dirinfo_c *) ob->form->u_vdata)->auto_sort_sync();
}

void DirinfoDUButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Du")) return;
    ((dirinfo_c *) ob->form->u_vdata)->du();    
}
