
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

#include "browser.h"
#include "display_forms.h"
#include "display_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "menu.h"
#include "panel.h"

void display_form_help_cb(void){
    DisplayHelpButtonCB(NULL,0);
}

void DisplayHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(display_help_text, "Display Options");
}

int display_form_raw_cb(FL_FORM *form, void *xev){
    XEvent *xevent = (XEvent *)xev;
    if(adjust_return_button(form, xevent)) return TRUE;
    if(standard_form_event_cb(form, xevent, NULL, display_form_help_cb)) return TRUE;
    if(xevent->type == KeyPress){
	KeySym keysym;
	char keybuf[1];
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	case XK_n:
	case XK_N:
	    source_browser->show_symbolic_flag = FALSE; break;
	case XK_b:
	case XK_B:
	    source_browser->show_symbolic_flag = TRUE; break;
	default:  return FALSE;
	}
	display_form_sync(TRUE);
	source_browser->update(TRUE);
	return TRUE;
    }    
    return FALSE;
}

static long fsize_menu, fstyle_menu;
	
static void display_set_wintitle(void){
    char buffer[MAXPATHLEN];
    sprintf(buffer, "Adjust Browser #%d Display Options:", source_browser->get_ID());
    XStoreName(fl_display, DisplayForm->window, buffer);
}

void display_show_form(void){
    static int first = TRUE;
    if(first){
	fl_register_call_back(DisplayForm, FL_ALL_EVENT, display_form_raw_cb);

	first = FALSE;
	fl_set_button_shortcut(DisplayListHiddenButton,  "^Ll", TRUE);
	fl_set_button_shortcut(DisplayLongButton,        "^oO", TRUE);
	fl_set_button_shortcut(DisplaySyncAllButton,   "^yY", TRUE);
	
	fl_set_button_shortcut(DisplayOwnerButton,       "Oo", TRUE);
	fl_set_button_shortcut(DisplayGroupButton,       "Gg", TRUE);
	fl_set_button_shortcut(DisplaySizeButton,        "Ss", TRUE);
	fl_set_button_shortcut(DisplayPermissionsButton, "Pp", TRUE);
	fl_set_button_shortcut(DisplayNLinksButton,      "Ll", TRUE);
	fl_set_button_shortcut(DisplayAccessButton,      "Aa", TRUE);
	fl_set_button_shortcut(DisplayModificationButton,"Mm", TRUE);
	fl_set_button_shortcut(DisplayInodeButton,       "Ii", TRUE);
	fl_set_button_shortcut(DisplayFiletypeButton,    "Ff", TRUE);
	fl_set_button_shortcut(DisplayTypeCharButton,    "hH", TRUE);

	fl_set_object_shortcut(DisplaySymbolicIDButton,  "bB", TRUE);
	fl_set_object_shortcut(DisplayNumericIDButton,   "Nn", TRUE);
	
	fl_set_button_shortcut(DisplayYearsButton,       "Yy", TRUE);
	fl_set_button_shortcut(DisplaySecondsButton,     "eE", TRUE);

	fl_set_button_shortcut(DisplayFollowLinksButton, "wW", TRUE);

	fl_set_button_shortcut(DisplayDefaultsButton,    "Dd", TRUE);
	fl_set_object_shortcut(DisplayCloseReturnButton,  "Cc^M", TRUE);

	fl_set_button_shortcut(DisplayFontStyleButton, "tT", TRUE);
	fl_set_button_shortcut(DisplayFontSizeButton,  "zZ", TRUE);

	fsize_menu = make_fsize_menu();
	fstyle_menu = new_pup();
	add_to_pup(fstyle_menu, "Style%t", 0);
	char buffer[256];
	sprintf(buffer, "Normal%%x%d", 4);
	//cerr << buffer << endl;
	add_to_pup(fstyle_menu, buffer, 0, "Nn");
	sprintf(buffer, "Bold%%x%d", 5);
	//cerr << buffer << endl;
	add_to_pup(fstyle_menu, buffer, 0, "Bb");
    }

    set_to_ret_button(DisplayCloseReturnButton);
    display_form_sync();
    focus_save();
    if(!DisplayForm->visible){
	fl_show_form(DisplayForm, FL_PLACE_CENTER, TRUE, "Adjust Display Options:");
	display_set_wintitle();
    } else {
	XMapRaised(fl_display, DisplayForm->window);
    }
    tool_focus_move(DisplayForm, DisplayCloseReturnButton);
}

void display_form_sync(int sym_num_only_flag){
    M_trace("display_form_sync", "start");
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    fl_freeze_form(DisplayForm);
    fl_set_button(DisplaySymbolicIDButton,   source_browser->show_symbolic_flag);
    fl_set_button(DisplayNumericIDButton,   !source_browser->show_symbolic_flag);

    if(!sym_num_only_flag){
	fl_set_button(DisplayListHiddenButton,   source_browser->list_hidden_flag);
	fl_set_button(DisplayOwnerButton,        source_browser->show_owner_flag);
	fl_set_button(DisplayGroupButton,        source_browser->show_group_flag);
	fl_set_button(DisplaySizeButton,         source_browser->show_size_flag);
	fl_set_button(DisplayPermissionsButton,  source_browser->show_perms_flag);
	fl_set_button(DisplayNLinksButton,       source_browser->show_nlinks_flag);
	fl_set_button(DisplayAccessButton,       source_browser->show_access_flag);
	fl_set_button(DisplayModificationButton, source_browser->show_modific_flag);
	fl_set_button(DisplayInodeButton,        source_browser->show_inode_flag);
	fl_set_button(DisplayFiletypeButton,     source_browser->show_filetype_flag);
	fl_set_button(DisplayTypeCharButton,     source_browser->show_type_char_flag);
	fl_set_button(DisplayFollowLinksButton,  source_browser->follow_links_flag);
	fl_set_button(DisplayLongButton,         source_browser->long_wintitle_flag);
	
	fl_set_button(DisplayYearsButton,        source_browser->show_time_years_flag);
	fl_set_button(DisplaySecondsButton,      source_browser->show_time_seconds_flag);
	
	int fontsize = fl_get_panel_fontsize(source_browser->panel);
	char buffer[64];
	sprintf(buffer, "Size %.2d", fontsize);
	fl_set_object_label(DisplayFontSizeButton, buffer);
	int fontstyle  = fl_get_panel_fontstyle(source_browser->panel);
	switch(fontstyle){
	case 4: sprintf(buffer, "Style Normal");  break;
	case 5:   sprintf(buffer, "Style Bold");    break;
	default:sprintf(buffer, "Style Unknown"); break;
	}
	fl_set_object_label(DisplayFontStyleButton, buffer);
	// resync underlining.
	fl_set_button_shortcut(DisplayFontStyleButton, "tT", TRUE);
	fl_set_button_shortcut(DisplayFontSizeButton,  "zZ", TRUE);
	display_set_wintitle();
    }
    fl_unfreeze_form(DisplayForm);
}

void DisplayYearsButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    toggle_flag(source_browser->show_time_years_flag);
    source_browser->update(TRUE);
}

void DisplaySecondsButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    toggle_flag(source_browser->show_time_seconds_flag);
    source_browser->update(TRUE);
}

void DisplayDefaultsButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    source_browser->freeze();
    source_browser->set_show_defaults();
    source_browser->set_font_defaults();
    display_form_sync();
    source_browser->update(TRUE);
    source_browser->unfreeze();
}

void DisplaySymbolicNumericIDButtonCB(FL_OBJECT *, long val){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    if(val){
	source_browser->show_symbolic_flag = FALSE;
    } else {
	source_browser->show_symbolic_flag = TRUE;
    }
    source_browser->update(TRUE);
}

void DisplayFollowLinksButtonCB(FL_OBJECT *, long){
    toggle_flag(source_browser->follow_links_flag);
    source_browser->update(TRUE);
}

void DisplayLongButtonCB(FL_OBJECT *, long){
    toggle_flag(source_browser->long_wintitle_flag);
    source_browser->set_complete_wintitle();
}

void DisplaySyncAllButtonCB(FL_OBJECT *, long){
    browser_c *tmp_browser = source_browser->next_browser;
    while(tmp_browser != source_browser){
	tmp_browser->freeze();
	tmp_browser->sync_display_opts(source_browser);
	tmp_browser->update(TRUE);
	tmp_browser->unfreeze();
	tmp_browser = tmp_browser->next_browser;
    }
}

void DisplayCloseReturnButtonCB(FL_OBJECT *, long){
    if(DisplayForm->visible){
	focus_restore_start();
	fl_hide_form(DisplayForm);
	focus_restore_finish();
    }
}

void DisplayCB(FL_OBJECT *, long option){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    switch(option){
    case access_e:   toggle_flag(source_browser->show_access_flag);    break; 
    case filetype_e: toggle_flag(source_browser->show_filetype_flag);  break; 
    case group_e:    toggle_flag(source_browser->show_group_flag);     break; 
    case inode_e:    toggle_flag(source_browser->show_inode_flag);     break; 
    case modific_e:  toggle_flag(source_browser->show_modific_flag);   break; 
    case nlinks_e:   toggle_flag(source_browser->show_nlinks_flag);    break; 
    case owner_e:    toggle_flag(source_browser->show_owner_flag);     break; 
    case perms_e:    toggle_flag(source_browser->show_perms_flag);     break; 
    case size_e:     toggle_flag(source_browser->show_size_flag);      break;
    case typechar_e: toggle_flag(source_browser->show_type_char_flag); break; 
    default:    return;   /* no action on name_e, always display. */
    }
    source_browser->update(TRUE);
}

void DisplayListHiddenButtonCB(FL_OBJECT *, long){
    toggle_flag(source_browser->list_hidden_flag);
    source_browser->update(TRUE);
}

void DisplayFontStyleButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
  
    int newstyle = (int) do_pup(fstyle_menu);
    if(newstyle >= 0){
	// cerr << "new style:" << newstyle << endl;
	source_browser->set_fontstyle(newstyle);
	display_form_sync();
    }
}

void DisplayFontSizeButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int newsize = (int) do_pup(fsize_menu);
    if(newsize > 0){
	// cerr << "new size:" << newsize << endl;
	source_browser->set_fontsize(newsize);
	display_form_sync();
    }
}

