
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <strstream.h>

#ifdef LINUX_GLIBC2
#       include <errno.h>
#endif

#include "browser.h"
#include "cb_key.h"
#include "cb_fkey.h"
#include "cstring.h"
#include "dirinfo_forms_aux.h"
#include "edit_colour_forms.h"
#include "edit_forms.h"
#include "edit_forms_aux.h"
#include "bitmaps/editor.bm"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "fl_edit.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "linklist.h"
#include "menu.h"
#include "pathnames.h"
#include "resource.h"
#include "status_forms_aux.h"
#include "wildcard.h"
#include "wm_forms_aux.h"

static int dbg_flag = FALSE;

static int edit_raw_cb(FL_FORM *form, void *xev);
static int edit_atclose_cb(FL_FORM* form, void*);

class edit_c;
edit_c *colour_edit = NULL;

static void edit_colour_close_form(void);
static void edit_colour_form_sync(int bSyncWintitle = FALSE);
static void edit_colour_show_form(void);
static void edit_remove_other_selections_cb(FL_OBJECT *, long);

// history menu
static void edit_history_save_defaults(ofstream &ofile);
static void edit_history_remember_last(const char *file);
static void edit_history_remove_duplicates(const char *pcFile);
static long make_edit_history_menu(void);
static int edit_history_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
);

// safeguard:
static int too_many_editors(void); 

class edef_c;
static void clone_edefs(edef_c *dst, const edef_c *src);

/**** CLASS EDIT_C ****/

class edit_c {
private:
    struct stat statbuf;
    char fullpath[MAXPATHLEN+16];
    int bNeedsBackup;

    int allready_open_ok(void);
    int load(const char *fname, int iAlreadyOpenOK = FALSE);

    void close(void);
    char *last_wintitle;
    int create_backup(const char * pcPath);
    int canonicalize_path(const char *fname);
    void set_icon(void);
    void sync_tool_dirs(void);
public:
    FD_EditForm *EditFD;
    int bSearchMode;

    edit_c();
    ~edit_c();
    int change_ok	(void),
	has_changed	(void),
	paste		(void)
    ;
    void 
	clone		(void),
	insert		(void),
	load_conditional(void),
	cut	    	(void),
	copy    	(void),
	goto_line    	(void),
	win_raise	(void),
	search		(void),
	select_all	(void)
    ;

    void set_wintitle(void);
    void show_search_text(const char *pText = NULL);

    int new_file(void);
    int new_file_path(char *new_path, int bDoCompare = FALSE);
    int new_window(void);
    int save(int bErrorMsg = TRUE);	// bark if fail
    int save_as(void);
    int show(const char *fname, int focus_flag = TRUE, int iAlreadyOpenOK = FALSE);
    const char *get_fname(void);
    const char *get_path(void);
    
    void sync_colours(void);
    void sync_fonts(void);
    void var_sync(void);

    void freeze(void);
    void unfreeze(void);
    
    edef_c* edef;
};	

/**** CLASS EDEF_C ****/

class edef_c {
private:
    static int ID_count;
public:
    int ID,
	fontsize, fontstyle,
	bg_colour, fg_colour, hi_colour,
	wx, wy, ww, wh
	;
    edef_c(void);
    edit_c *edit;
    void default_colours(void);
};

/**** EDEF_C METHODS ****/

int edef_c::ID_count = 0;

edef_c::edef_c(void){
    edit    = NULL;
    ID        = ++ID_count;
    fontsize  = FL_NORMAL_SIZE;
    fontstyle = FL_FIXED_STYLE;
    default_colours();
    // place center as default.
    ww = 660;
    wh = 330;
    if(ww >= 0.9 * tool_max_ww) ww = (int) (0.9 * tool_max_ww);
    wx = (int) (0.5 * (tool_max_ww - ww));
    wy = (int) (0.5 * (tool_max_wh - wh));
}
void edef_c::default_colours(void){
    fg_colour = FL_BLACK;
    bg_colour = FL_COL1;
    hi_colour = FL_GREEN;
}

static linklist_c busy_edef, free_edef, file_list;



/*
static void dump_free(void){
    cerr << "free list (" << free_edef.getsize() << ")" << endl;
    edef_c *edef;
    for(int i=0; i<free_edef.getsize(); i++){
	edef = (edef_c *) free_edef[i];
	cerr << "Entry: " << i << ", ptr: " << (long) edef << ", ID: " << edef->ID << endl;
    }
}*/

edef_c *get_edef(edit_c *edit){
    if(!edit){
	M_err("get_edef","NULL edit");
    }
    edef_c *edef;
    //dump_free();
    if(free_edef.getsize() <= 0){
	//cerr << "making new model ...";
	edef = new edef_c;
	if(busy_edef.getsize()){
	    //cerr << "and cloning";
	    // if new, use last busy def as model.
	    const edef_c *vtmp0, *vtmp1;
	    vtmp0 = (edef_c *) busy_edef[0];
	    for(int i=1; i<busy_edef.getsize(); i++){
		vtmp1 = (edef_c *) busy_edef[i];
		if(vtmp1->ID > vtmp0->ID){
		    vtmp0 = vtmp1;
		}
	    }
	    if(!vtmp0->edit){
		M_err("get_edef","no edit in vtmp");
	    } else {
		vtmp0->edit->var_sync();
	    }
	    clone_edefs(edef, vtmp0);
	}
	//cerr << "\n";
    } else {
	// we have a model.
	//cerr << "searching for a new model\n";
	edef_c *tmp_edef;
	edef = (edef_c *) free_edef[0];
	//cerr << "base ID is " << edef->ID << endl;
	for(int i=1; i<free_edef.getsize(); i++){
	    tmp_edef = (edef_c *) free_edef[i];
	    //cerr << "search ID is " << tmp_edef->ID << ", i:" << i << endl;
	    if(tmp_edef->ID < edef->ID){
		edef = tmp_edef;
	    }
	}
	free_edef.find(edef);
	free_edef.remove();
    }
    edef->edit = edit;
    //cerr << "getting edef ID" << edef->ID << endl;

    // insert the new edef at the correct position given by its ID.
    edef_c* pEdefSearch;

    for(int i=0; i<busy_edef.getsize(); i++){
      pEdefSearch = (edef_c *) busy_edef[i];
      if (!pEdefSearch) {
	cerr << "bad edef list!\n"; 
	// add anyway.
	busy_edef.add(edef);
	return edef;
      }

      if (pEdefSearch->ID > edef->ID) {
	busy_edef.insert(edef);
	return edef;
      }
    }
    // empty list, or add at tail.
    busy_edef.add(edef);

    return edef;
}

void release_edef(edef_c *edef){
    if(busy_edef.find(edef)){
	busy_edef.remove();
	free_edef.add(edef);
	//cerr << "releasing edef ID" << edef->ID << endl;
	//dump_free();
    } else {
	M_err("release_edef","edef not in busy list");
	delete edef;
    }
}

static void clone_edefs(edef_c *dst, const edef_c *src){
    dst->wx = src->wx + 50;
    dst->wy = src->wy + 50;
    dst->ww = src->ww;
    dst->wh = src->wh;
    dst->fontsize  = src->fontsize;
    dst->fontstyle = src->fontstyle;
    dst->fg_colour = src->fg_colour;
    dst->bg_colour = src->bg_colour;
    dst->hi_colour = src->hi_colour;
}

/**** EDIT_C METHODS ****/

const char *edit_c::get_fname(void){
    return get_basename(fullpath);
}

const char *edit_c::get_path(void){
    return fullpath;
}

void edit_c::clone(void){
    if (has_changed()) {
	if (ask_question(FALSE, 
	    "Save file `%s' before cloning?", get_fname())
	) {
	    if (!save(FALSE)) {
		timed_error_notifier("Save failed, cannot clone.");
		return;
	    }
	}
    }
    if (too_many_editors()) return;
	
    edit_c *pNewEdit = new edit_c;
    var_sync();
    clone_edefs(pNewEdit->edef, edef);
    pNewEdit->edef->edit = pNewEdit;
    pNewEdit->sync_colours();
    pNewEdit->sync_fonts();

    if(!pNewEdit->show(fullpath, TRUE, TRUE)){
	delete pNewEdit;
    } else {
	pNewEdit->freeze();
	//fl_set_textedit_topline(pNewEdit->EditFD->Editor, fl_get_textedit_topline(EditFD->Editor));
	pNewEdit->set_wintitle();
	pNewEdit->unfreeze();
    }
}

int edit_c::has_changed(void){
    return fl_get_textedit_bufchanged(EditFD->Editor);
}

int edit_c::change_ok(void){
    if (!has_changed()) return TRUE;
    CString strQuestion = "File `";
    strQuestion += get_basename(fullpath);
    strQuestion += "' was changed. Save edits?"; 
    int retval = ask_multi_question(
	TRUE, 1, strQuestion, 
	"Discard", "Dd", NoSymbol, 
	"Save",    "Ss", NoSymbol, 
	"Save As", "Aa", NoSymbol, 
	"Cancel",  "Cc", XK_Escape
    );
    switch(retval) {
    case -1:
	// esc/cancel
	return FALSE;
    case 0:
	// discard
	return TRUE;
    case 1:
	// save
	return save();
    case 2:
	// save as
	if (!save_as()) {
	    timed_error_notifier("Save failed.");
	    return FALSE;
	}
	return TRUE;
    default:
	return FALSE;
    }
}

void edit_c::close(void){
    if(EditFD->EditForm->visible){
	var_sync();
	focus_restore_start();
	fl_hide_form(EditFD->EditForm);
	focus_restore_finish();
    }
    fl_clear_textedit(EditFD->Editor);
}

void edit_c::copy(void){
    fl_textedit_copy(EditFD->Editor);
}

void edit_c::cut(void){
    fl_textedit_cut(EditFD->Editor);
    set_wintitle();
}

void edit_c::goto_line(void){
    fl_textedit_goto_line(EditFD->Editor);
}

int edit_c::paste(void){
    int retval = fl_textedit_paste(EditFD->Editor);
    set_wintitle();
    return retval;
}

void edit_c::win_raise(void){
    FL_FORM *form = EditFD->EditForm;
    if (form && form->visible) {
	tool_focus_move(form);
	XMapRaised(fl_display, form->window);
    }
}

void edit_c::select_all(void){
    fl_select_textedit_all(EditFD->Editor);
}

static CString strEditSearchExpr = "";

void edit_c::search(void){
    if (!EditFD) return;

    FL_OBJECT *pEditor = EditFD->Editor;
    if (!pEditor) return;

    SPEC *pSpec = (SPEC *) pEditor->spec;
    if (!pSpec) return;
    if (!bSearchMode || !pSpec->exp) {
	// the flag is set to false, and set true only after
	// the checkes turn out good. 
	bSearchMode = FALSE;
	const char *pText = OKCancel("Enter search text:", 
	    pSpec->exp ? pSpec->exp : strEditSearchExpr.GetBuffer()
	);
	if (!pText) {
	    return;
	}
	if (pSpec->exp) {
	    free(pSpec->exp);
	}
	pSpec->exp = strdup(pText);
	if (!pSpec->exp) {
	    malloc_unstable("edit_c::search");
	    return;
	}
	strEditSearchExpr = pText;
	bSearchMode = TRUE;
    }

    // trick: allways set the again-flag, to override the editors internal
    // search mechanism with it's input. 
    int nRetVal = fl_textedit_search(pEditor, TRUE);

    CString strTemp = "`";
    strTemp += pSpec->exp;
    strTemp += "'";

    if (nRetVal > 0) {
	// looks good
	show_search_text(strTemp);
    } else if (nRetVal < 0) {
	// wrap-around
	fl_textedit_home_file(pEditor);
	fl_textedit_search(pEditor, TRUE);
	strTemp += " [wrapped]";
	show_search_text(strTemp);
	XBell(fl_display, 0);
    } else {
	bSearchMode = FALSE;
	show_search_text();
	XBell(fl_display, 0);
	timed_notifier("Search string not found");
    }	
}


edit_c::edit_c(void){
    last_wintitle = NULL;
    bSearchMode = FALSE;
    bNeedsBackup = TRUE;
    EditFD = create_form_EditForm();
    EditFD->EditForm->u_vdata = this;
    fl_set_form_atclose(EditFD->EditForm, edit_atclose_cb, NULL);
    fl_register_call_back(EditFD->EditForm, FL_ALL_EVENT, edit_raw_cb);

    edef = get_edef(this);
    sync_colours();
    sync_fonts();
    fl_textedit_remove_other_selections_cb = edit_remove_other_selections_cb;
}

edit_c::~edit_c(){
    if(this == colour_edit){
	colour_edit = NULL;
	edit_colour_close_form();
    }
    if (last_wintitle) {
	fl_free(last_wintitle);
	last_wintitle = NULL;
    }

    edef->edit = NULL;  // just in case .. 
    release_edef(edef);
    close();
    fl_free_form(EditFD->EditForm);
}

int edit_c::canonicalize_path(const char *fname){
    if(!fname) return FALSE;
    char path_buf[MAXPATHLEN+1];
    if (fname[0] != '/') {
	const int cwd_bufsize = MAXPATHLEN+1;
	char cwd_buf[cwd_bufsize];
	if(!getcwd(cwd_buf,cwd_bufsize)){
	    return FALSE;
	}
	strcpy(path_buf, cwd_buf);
	strcat(path_buf, "/");
	strcat(path_buf, fname);
    } else {
	strcpy(path_buf, fname);
    }
    if (!real_path(path_buf, fullpath)) {
	timed_strerror_notifier(path_buf);
	return FALSE;
    }
    return TRUE;
}

int edit_c::load(const char *fname, int iAlreadyOpenOK){
    if(!fname) return FALSE;
    if (!canonicalize_path(fname)) return FALSE;

    // check what kind of thing the file is.
    char buffer[MAXPATHLEN];
    if(stat(fullpath, &statbuf) < 0){
	sprintf(buffer, "edit error: cannot stat file %s !", fname);
	timed_error_notifier(buffer, notifier_secs);
	edit_history_remove_duplicates(fullpath);
	return FALSE;
    } 
    if(S_ISDIR(statbuf.st_mode)){
	dirinfo(fname);
	return FALSE;
    }
    if(!S_ISREG(statbuf.st_mode)){
	timed_error_notifier("sorry, can't edit irregular files");
	return FALSE;
    }

    if (!iAlreadyOpenOK && !allready_open_ok()) return FALSE;

    if(!fl_load_textedit(EditFD->Editor, fullpath)){
        sprintf(buffer, "Can't edit file %s !", fname);
	timed_strerror_notifier(buffer);
	fl_clear_textedit(EditFD->Editor);
	fl_add_textedit_line(EditFD->Editor, "unable to load (?)");
	edit_history_remove_duplicates(fullpath);
	return FALSE;
    } else {
	// new file, so no backup yet.
	edit_history_remember_last(fullpath);
	bNeedsBackup = TRUE;
        return TRUE;
    }
}

void edit_c::insert(void){
    const char *fname = fl_show_fselector("Insert file:", 
	get_dirname(fullpath), "",""
    );
    if (fname) {
	fl_insert_textedit_file(EditFD->Editor, (char *) fname);
    }
}

int edit_c::new_file(void){
    char new_path[2*MAXPATHLEN];
    if (!change_ok()) return TRUE;

    if (!new_file_path (new_path, TRUE)) return FALSE;
    return load(new_path);
}

// returns wether a new path was set.
int edit_c::new_file_path(char *new_path, int bDoCompare){
    char buf[2*MAXPATHLEN], fbuf[2*MAXPATHLEN];
    const char *result = OKCancel("Give name for new file: ", fullpath);
    if (!result) return FALSE;

    strip_trailing_slashes(result);
    if (bDoCompare && strcmp(result, fullpath) == 0) return FALSE;

    // make absolute path with directory where old file was:
    if (result[0] != '/') {
	strcpy(fbuf, get_dirname(fullpath));
	strcat(fbuf, "/");
	strcat(fbuf, result);
	result = fbuf;
    }
    if (!real_path(result, new_path)) {
	timed_strerror_notifier(result);
	return FALSE;
    }

    // does it exist?
    struct stat statbuf;
    if (stat(new_path, &statbuf) == 0) {
	return ask_question(FALSE, "File `%s' exists. Open it?", new_path);
    }
    // other errors?
    if (errno != ENOENT) {
	sprintf(buf, "Cannot create new file `%s'.", new_path);
	timed_strerror_notifier(buf);
	return FALSE;
    }
    // create it...
    FILE *fp = fopen(new_path, "w");
    if (!fp) {
	sprintf(buf, "Cannot create new file `%s'.", new_path);
	timed_strerror_notifier(buf);
	return FALSE;
    }
    fclose(fp);
    return TRUE;
}

int edit_c::new_window(void){
    const char *fname = NULL;
    char new_path[2*MAXPATHLEN];
    // load existing, create new, cancel. no cloning. 
    int retval = ask_multi_question(
	FALSE, 0, "Open new editor window and...",
	"create New file",   	"Nn", NoSymbol, 
	"Load existing file", 	"Ll", NoSymbol, 
	"Cancel",  		"Cc", XK_Escape
    );
    if (retval < 0) return TRUE;
    switch (retval) {
    case 0:
	if (!new_file_path (new_path, TRUE)) return FALSE;
	return quick_edit_file(new_path);
    case 1:
	fname = fl_show_fselector("Load file:", 
	    get_dirname(fullpath), "",""
	);
	if (fname) {
	    //sleep(5);
	    return quick_edit_file(fname);
	}
	return FALSE;
    default:
	return FALSE;
    }
}

void edit_c::load_conditional(void){
    if (change_ok()) {
	const char *fname = fl_show_fselector("Open file:", 
	    get_dirname(fullpath), "",""
	);
	if (fname) {
	    load(fname);
	}
    }
}

int edit_c::create_backup(const char * pcPath){
    if (!pcPath) return FALSE;
    // use same backup file convention (~) as Emacs does. 
    CString strBupExtension = "~";
    CString strBupFName = pcPath;
    strBupFName += strBupExtension;
    CString strCmd = cp_path;
    strCmd += " ";
    strCmd += fullpath;
    strCmd += " ";
    strCmd += strBupFName;
    status_msg(1, "Creating backup file %s...", strCmd.GetBuffer());
    safe_system(strCmd);
    //cerr << strCmd << endl;
    status_msg(1, "done.");

    // verify...
    struct stat src_statbuf, bup_statbuf;
    char buffer[MAXPATHLEN*2];

    if(stat(fullpath, &src_statbuf) < 0){
	sprintf(buffer, "backup failed: can't stat file `%s'!", fullpath);
	timed_strerror_notifier(buffer);
	return FALSE;
    } 
    if(stat(strBupFName, &bup_statbuf) < 0){
	sprintf(buffer, "backup failed: can't stat backup copy `%s'!", strBupFName.GetBuffer());
	timed_strerror_notifier(buffer);
	return FALSE;
    } 
    if (src_statbuf.st_size != bup_statbuf.st_size) {
	timed_error_notifier("backup failed: original and backup files have differing sizes.!");
	return FALSE;
    } 
	
    return TRUE;
}

void edit_c::sync_tool_dirs(void){
    const char *pDirName = get_dirname(fullpath);
    //cerr << pDirName;
    if (pDirName && (pDirName[0] == '/')) {
	tool_update_matching_dirs(pDirName, TRUE);
    }
}

int edit_c::save(int bErrorMsg) {
    if (!create_backup(fullpath)) {
	timed_error_notifier("Save failed: Cannot create backup. Try `Save As'.");
	return FALSE;
    } else {
	bNeedsBackup = FALSE;
    }
    int retval = fl_save_textedit(EditFD->Editor, fullpath);
    set_wintitle();
    sync_tool_dirs();
    if (bErrorMsg && !retval) {
	timed_error_notifier("Save failed.");
    }
    return retval;
}

int edit_c::save_as(void){
    // be careful with directory and fname.
    const char *fname = fl_show_fselector("Save As:", 
	get_dirname(fullpath), "*", get_basename(fullpath)
    );
    if (!fname) return TRUE;	// op was cancelled.
    if (!canonicalize_path(fname)) return FALSE;

    // file may exist, overwrite, create backup, cancel?
    struct stat statbuf;
    if(stat(fullpath, &statbuf) != 0){
	if (errno != ENOENT) {
	    timed_strerror_notifier("Cannot save file. Please choose a different name.");
	    return FALSE;
	}   // else it doesn't exist, fine.
    } else {
	CString strQuestion;
	strQuestion = "File `";
	strQuestion += fullpath;
	strQuestion += "' exists. ";
	int retval = ask_multi_question(
	    TRUE, 0, strQuestion.GetBuffer(),
	    "Create Backup",   	"Bb", NoSymbol, 
	    "Overwrite", 	"Oo", NoSymbol, 
	    "Cancel",  		"Cc", XK_Escape
	);
	switch(retval) {
	case -1:
	    // esc/cancel
	    status_msg(1, "File `%s' not saved.", fullpath);
	    timed_notifier("File not saved.");
	    return FALSE;
	case 0:
	    // backup
	    if (!create_backup(fullpath)) {
		timed_error_notifier("Cannot create backup file.");
		return FALSE;    
	    }
	    break;
	case 1:
	    // overwrite ok;
	    break;
	default:
	    return FALSE;
	}
    }
    int retval = fl_save_textedit(EditFD->Editor, fullpath);
    if (retval) {
	// a successfull save as should require no backup.
	// if the file existed, the backup was created above.
	// if not, no backup anyway. 
	bNeedsBackup = FALSE;
    }
    set_wintitle();
    sync_tool_dirs();
    fl_invalidate_fselector_cache();
    return retval;
}

void edit_c::set_icon(void){
    static Pixmap pixmap =0;
    FL_FORM *form = EditFD->EditForm;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, form->window, editor_bits, editor_width,
	    editor_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    if (pixmap > 0) {
	set_window_icon(form->window, pixmap);
    } else {
	cerr << "Failed to create pixmap for Editor.\n";
    }
}


int edit_c::show(const char *fname, int focus_flag, int iAlreadyOpenOK){
    char buffer[MAXPATHLEN];
    if(!fname || !fname[0]) return FALSE;
    if(access(fname, R_OK) != 0) {
	sprintf(buffer, "Cannot access file `%s'.", fname);
	timed_strerror_notifier(buffer);
	edit_history_remove_duplicates(fname);  // _not_ fullpath here. 
	return FALSE;
    }

    if(load(fname, iAlreadyOpenOK)){
	FL_FORM *form = EditFD->EditForm;
	if(focus_flag) focus_save();
	fl_set_form_minsize(form, 350, 60);
	fl_set_form_maxsize(form, tool_max_ww, tool_max_wh);
	fl_set_initial_placement(form, edef->wx, edef->wy, edef->ww, edef->wh);	
	fl_show_form(form,FL_PLACE_FREE,FL_FULLBORDER,"Edit");
	set_wintitle();
	set_icon();
	if(focus_flag) tool_focus_move(form, EditFD->FileMenuButton);
	return TRUE;
    }
    return FALSE;    
}

void edit_c::show_search_text(const char *pText) {
    CString strLabel = "";
    fl_freeze_form(EditFD->EditForm);
    if (pText) {
	strLabel = "search for: ";     
	strLabel += pText;
	//strLabel += " (use `Ctrl-s' to search again)";
	fl_set_object_color(EditFD->Text, FL_BLACK, FL_COL1);
	fl_set_object_lcol(EditFD->Text, FL_WHITE);
    } else {
	fl_set_object_color(EditFD->Text, FL_COL1, FL_BLACK);
	fl_set_object_lcol(EditFD->Text, FL_BLACK);
	bSearchMode = FALSE;	// extra safeguard.
    }
    fl_set_object_label(EditFD->Text, strLabel);
    fl_unfreeze_form(EditFD->EditForm);
}

void edit_c::set_wintitle(void){
    FL_FORM *form  = EditFD->EditForm;
    FL_OBJECT *obj = EditFD->Editor;
    if(!form->visible) return;
    
    char buf[MAXPATHLEN], tmpbuf[256];
    int maxline = fl_get_textedit_maxline(obj);

    int row = 0, col = 0;
    fl_get_textedit_cursorpos(obj, &col, &row, NULL);

    unsigned long len = fl_get_textedit_buflen(obj);

    if (row == 0 && maxline == 0 && len == 0) {
	strcpy(tmpbuf, "empty");
    } else {
	sprintf(tmpbuf, "line %d of %d, %s",
	    row+1, maxline, bytes_to_string(len)
	);
    }	

    sprintf(buf, "QuickEdit %d: %s (%s, %s]",
	edef->ID, get_basename(fullpath), tmpbuf, get_dirname(fullpath)
    );
    if (fl_get_textedit_bufchanged(EditFD->Editor)) {
	strcat(buf, "(modified)");
    }

    if (last_wintitle) {
	if (strcmp(last_wintitle, buf) == 0) return;
	fl_free(last_wintitle);
    }
    last_wintitle = strdup(buf);

    XStoreName(fl_display, form->window, buf);
    sprintf(buf, "QEd%d: %s", edef->ID, get_basename(fullpath)); 
    XSetIconName(fl_display, form->window, buf); // no WM yet.
}

void edit_c::sync_colours(void){
    fl_set_textedit_color(EditFD->Editor, 
	edef->bg_colour, 
	edef->fg_colour, 
	edef->hi_colour,
	TRUE
    );
}

void edit_c::sync_fonts(void){
    fl_set_textedit_fontstyle(EditFD->Editor, edef->fontstyle);
    fl_set_textedit_fontsize (EditFD->Editor, edef->fontsize);
}

void edit_c::freeze(void){
    fl_freeze_form(EditFD->EditForm);
}

void edit_c::unfreeze(void){
    fl_unfreeze_form(EditFD->EditForm);
}

void edit_c::var_sync(void){
    FL_FORM* form = EditFD->EditForm;
    if(form->visible){
	edef->ww = form->w;
	edef->wh = form->h;
	edef->wx = form->x;
	edef->wy = form->y;
    }
}

/**** FORMS CALLBACKS ****/

static int vob_valid(FL_OBJECT *ob, const char *msg = ""){
    if(!ob || !ob->form || !ob->form->u_vdata){
	M_err("Edit", "Invalid ptrs (%s)", msg);
	return FALSE;
    } else {
	return TRUE;
    }
}

void EditEditorCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Textedit")) return;
    //cerr << "EECB\n";
    edit_c *pEdit = (edit_c *) ob->form->u_vdata;
    if (!pEdit) return;
    pEdit->set_wintitle();
}

void EditCloseReturnButtonCB(FL_OBJECT *ob, long){
    if(!vob_valid(ob, "Close")) return;
    //    delete (edit_c *) ob->form->u_vdata;
}

/**** MENU SUPPORT ****/

static edit_c *menu_edit = NULL;

static void fstyle_cb(int style){
    if(style > 0 && menu_edit){
	menu_edit->edef->fontstyle = style;
	menu_edit->sync_fonts();
    }
}

static void fsize_cb(int size){
    if(size > 0 && menu_edit){
	menu_edit->edef->fontsize = size;
	menu_edit->sync_fonts();
    }
}

static void edit_close_cb(int){
    if (!menu_edit) return;
    if (menu_edit->change_ok()) {
	delete menu_edit;
	menu_edit = NULL;
    }
}

static void edit_colours_cb(int){
    colour_edit = menu_edit;
    edit_colour_show_form();
}

static void edit_clone_cb(int){
    if(menu_edit) menu_edit->clone();
}

static void edit_insert_cb(int){
    if(menu_edit) menu_edit->insert();
}

static void edit_new_file_cb(int){
    if(menu_edit) menu_edit->new_file();
}

static void edit_new_window_cb(int){
    if(menu_edit) menu_edit->new_window();
}

static void edit_load_cb(int){
    if(menu_edit) menu_edit->load_conditional();
}

static void edit_save_cb(int){
    if(menu_edit) menu_edit->save();
}

static void edit_save_as_cb(int){
    if(menu_edit) menu_edit->save_as();
}

void EditFileMenuButtonCB(FL_OBJECT *ob, long bFocus){
    if(!vob_valid(ob, "Menu")) return;
    menu_edit = (edit_c *) ob->form->u_vdata;
    if (!menu_edit) return;
    
    long file_menu = new_pup();

    long history_menu = make_edit_history_menu();
    if (history_menu <= 0) return;
    
    add_to_pup(file_menu, "File (Alt-f, Esc-f)%t",  0);
    add_to_pup(file_menu, "History%m",       history_menu,              "Hh");
    add_to_pup(file_menu, "Load%f",          (long) edit_load_cb,       "Ll");
    add_to_pup(file_menu, "Insert%f",        (long) edit_insert_cb,     "Ii");
    add_to_pup(file_menu, "New File%f",	     (long) edit_new_file_cb,   "Nn");
    add_to_pup(file_menu, "New Window%f",    (long) edit_new_window_cb, "Ww");
    add_to_pup(file_menu, "Save (Alt-s, Esc-s)%f",  (long) edit_save_cb,       "Ss");
    add_to_pup(file_menu, "Save As%f", 	     (long) edit_save_as_cb,    "Aa");
    add_to_pup(file_menu, "Clone%f",   	     (long) edit_clone_cb,      "Cc");
    add_to_pup(file_menu, "Quit (Alt-q, Esc-q)%f",  (long) edit_close_cb,      "Qq");
    
    if (bFocus) {
	/* focus saving is dangerous in this context if a menu cb does
	 * a focus operation itself. So, only move the ptr if a keyboard
	 * shortcut was used, and don't save and restore focus.
	 */
	tool_focus_move(ob->form, menu_edit->EditFD->FileMenuButton);
    }
    do_pup(file_menu);
    
    free_pup(file_menu);
    menu_edit = NULL;
}

static void edit_cut_cb(int){
    if(menu_edit) menu_edit->cut();
}
static void edit_copy_cb(int){
    if(menu_edit) menu_edit->copy();
}
static void edit_goto_cb(int){
    if(menu_edit) menu_edit->goto_line();
}
static void edit_ghelp_cb(int){
    info_show_form();
}
static void edit_help_cb(int){
    edit_context_func();
}
static void edit_paste_cb(int){
    if(menu_edit) menu_edit->paste();
}
static void edit_select_all_cb(int){
    if(menu_edit) menu_edit->select_all();
}
static void edit_search_cb(int){
    if(menu_edit) menu_edit->search();
}

static void do_global(int){
    OptionsButtonCB(NULL,0);   
}

static long fsize_menu, fstyle_menu, windows_menu;

static long
make_options_menu(void) {
    long options_menu;
    
    fsize_menu  = make_fsize_menu((long) fsize_cb);
    fstyle_menu = make_fstyle_menu((long) fstyle_cb);
    windows_menu = get_windows_menu(); 
    
    options_menu = new_pup();
    add_to_pup(options_menu, "Options (Alt-o, Esc-o, F2)%t", 0);
    add_to_pup(options_menu, "Font Size %m",                 fsize_menu,   "zZ");
    add_to_pup(options_menu, "Font Style %m",                fstyle_menu,  "tT");
    add_to_pup(options_menu, "Windows (Alt-w, Esc-w)%m",     windows_menu, "Ww");
    add_to_pup(options_menu, "Colours %f",                  (long) edit_colours_cb, "Cc");
    add_to_pup(options_menu, "Global Options (Meta-F2)%f",  (long) do_global,  "Gg");

    return options_menu;
}

static void free_options_menu(long options_menu) {
    free_pup(fsize_menu);
    free_pup(fstyle_menu);
    free_pup(options_menu);
}

static void do_options_menu_direct(edit_c *edit, int bFocus = TRUE){
    if (!edit) return;
    menu_edit = edit;
    long options_menu = make_options_menu();

    if (bFocus) {
	tool_focus_move(menu_edit->EditFD->EditForm, menu_edit->EditFD->OptionsMenuButton);
    }
    do_pup(options_menu);

    free_options_menu(options_menu);
    menu_edit = NULL;
}
	
int XSelectionIsValid(void) {
    if (XGetSelectionOwner(fl_display,XA_PRIMARY) != None) return TRUE;
    // need to query the darn cutbuffers!
    char *pcBuf;
    int iBufLen;
    pcBuf = XFetchBytes(fl_display, &iBufLen);
    if (!pcBuf || !iBufLen) return FALSE;
    if (pcBuf) XFree(pcBuf);
    return TRUE;
}

void EditEditMenuButtonCB(FL_OBJECT *ob, long bFocus){
    if(!vob_valid(ob, "Menu")) return;
    menu_edit = (edit_c *) ob->form->u_vdata;
    if (!menu_edit) return;
    
    long edit_menu = new_pup();

    int bHasSelection = (fl_get_textedit_selsize(menu_edit->EditFD->Editor) >= 0);
    
    add_to_pup(edit_menu, "Edit (Alt-e, Esc-e)%t", 0);
    if (bHasSelection) {
	add_to_pup(edit_menu, "Cut (Alt-x, Ctrl-x, Esc-x)%f",   (long) edit_cut_cb,        "xX");
	add_to_pup(edit_menu, "Copy (Alt-c, Ctrl-c, Esc-c)%f",  (long) edit_copy_cb,       "Cc");
    } else {
	add_to_pup(edit_menu, "[Cut]%x",      0,                         "xX");
	add_to_pup(edit_menu, "[Copy]%x",     0,                         "Cc");
    }
    if (XSelectionIsValid()) {
	add_to_pup(edit_menu, "Paste/Yank (Alt-y, Ctrl-y, Esc-y)%f",  (long) edit_paste_cb,      "Yy");
    } else {
	add_to_pup(edit_menu, "[Paste/Yank]%x",    0,                         "Yy");
    }
    add_to_pup(edit_menu, "Select All (Alt-a, Esc-a)%f", (long) edit_select_all_cb, "Aa");
    add_to_pup(edit_menu, "Search (Ctrl-s)%f",	         (long) edit_search_cb,     "Ss");
    add_to_pup(edit_menu, "Go to Line(Alt-g, Esc-g)%f",  (long) edit_goto_cb,       "Gg");
    add_to_pup(edit_menu, "Help (F1)%f",	         (long) edit_help_cb,       "Hh");
    add_to_pup(edit_menu, "Global Help (Meta-F1)%f",	 (long) edit_ghelp_cb,      "lL");

    if (bFocus) {
	tool_focus_move(ob->form, menu_edit->EditFD->EditMenuButton);
    }
    do_pup(edit_menu);
    
    free_pup(edit_menu);
    menu_edit = NULL;
}

void EditOptionsMenuButtonCB(FL_OBJECT *ob, long bFocus){
    if (!ob || !ob->form) return;
    edit_c *edit = (edit_c *) ob->form->u_vdata;
    if (!edit) return;
    do_options_menu_direct(edit, (int) bFocus);
}

void WindowsMenuHiddenButtonCB(FL_OBJECT *, long){
    long windows_menu = get_windows_menu(); 
    do_pup(windows_menu);
}

/**** FORMS SUPPORT FUNCTIONS ****/

static int edit_atclose_cb(FL_FORM* form, void*){
    if(form && form->u_vdata){
	edit_c *pEdit = (edit_c *)form->u_vdata;
	if (pEdit->change_ok()) {
	    delete pEdit;
	}
    } else {
	M_err("edit_atclose","invalid ptrs");
    }
    return FL_IGNORE;
}

void edit_context_func(void){
    load_help_browser(quick_edit_help_text, "F5 QuickEdit");
}

static void selection_clear_other_editors(edit_c *edit){
    const edef_c *pEdef;
    int nEdefs = busy_edef.getsize();
    //cerr << "clear other. " << nEdefs << endl;
    if (nEdefs <= 1) return;
    for (int i=0; i<nEdefs; i++) {
	pEdef = (edef_c *) busy_edef[i];
	if (pEdef->edit != edit && 
	    pEdef->edit && 
	    pEdef->edit->EditFD &&
	    pEdef->edit->EditFD->Editor
	) {
	    fl_textedit_remove_selection_backend(pEdef->edit->EditFD->Editor);
	}
    }
}

static void edit_remove_other_selections_cb(FL_OBJECT *pTextEdit, long){
    if (!pTextEdit || !pTextEdit->form) return;
    selection_clear_other_editors((edit_c *) pTextEdit->form->u_vdata);
}

static int edit_raw_cb(FL_FORM *form, void *xev){
    if(!form || !form->u_vdata) return FALSE;
    edit_c *edit = (edit_c *) form->u_vdata;
    if (!edit) return FALSE;
    tool_got_event_flag = TRUE;
    
    XButtonEvent *x_butt = (XButtonEvent*) xev;
    XEvent *xevent = (XEvent *) xev;
    XKeyEvent *xkev= (XKeyEvent *) xev;
    KeySym keysym;
    char keybuf[1];

    /* early check for cancelling of search mode. */
    int bCancelSearch = FALSE;
    if (xevent->type == ButtonPress || 
	xevent->type == ButtonRelease
    ) {
	bCancelSearch = TRUE;
    }

    if (xevent->type == KeyPress) {
	XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
	if (((keysym == XK_S || keysym == XK_s) && tool_get_ctrl()) ||
	    (keysym == XK_Control_L || keysym == XK_Control_R)
	){
	    // do nothing, it's search again.
	} else {
	    //cerr << "Keysym: " << hex << keysym << endl;
	    bCancelSearch = TRUE;
	}
    }

    if (bCancelSearch && edit->bSearchMode) {
	edit->show_search_text();
	if (keysym == XK_Escape) {
	    // cancelling is enough, don't close window. 
	    // switch on the flag, user may have pressed esc-q or similar.
	    tool_escape_flag = TRUE;
	    return TRUE;
	}
    }

    switch(xevent->type){
    case FocusIn:	
	colour_edit = edit;
	edit_colour_form_sync(TRUE);
	return FALSE;
    case ButtonPress:
	tool_got_event_flag = TRUE;
	if(x_butt->button == Button3){
	    // gobble for menu.
	    return TRUE;
	}
	return FALSE;
    case ButtonRelease:
	tool_got_event_flag = TRUE;
	if(x_butt->button == Button3){
	    do_options_menu_direct(edit, FALSE);
	    return TRUE;
	}
	return FALSE;
    case SelectionClear:
	selection_clear_other_editors(edit);
	return FALSE;
    case KeyRelease:
	XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
	//cerr << "Keysym: " << hex << keysym << endl;
	if (keysym == XK_Escape) {
	    //cerr << "escape FALSE.\n";
	    tool_escape_flag = FALSE;
	    return TRUE;
	}
	return FALSE;
    case KeyPress:
	tool_got_event_flag = TRUE;
	switch(keysym){
	case XK_F1:
	    help_event_cb(edit_context_func);
	    return TRUE;
	case XK_F2:
	    if(tool_get_meta()){
		OptionsButtonCB(NULL,0);
	    } else {
		do_options_menu_direct(edit);
	    }
	    return TRUE;
	case XK_F4:
	    TreeButtonCB(NULL,0);
	    return TRUE;
	case XK_F5:
	    if(tool_get_meta()){
		EditHistoryCB();
	    }
	    return TRUE;
	case SunXK_F37: if(!sun_keyboard_flag) return FALSE;
	case XK_F12:
	    ExitButtonCB(NULL, 0);
	    return TRUE;
	case XK_A:
	case XK_a:
	    if(tool_get_alt() || tool_get_esc()){
		edit->select_all();
		return TRUE;
	    }
	    break;
	case XK_C:
	case XK_c:
	    if(tool_get_alt() || tool_get_ctrl() || tool_get_esc()){
		edit->copy();
		return TRUE;
	    }
	    break;
	case XK_E:
	case XK_e:
	    if(tool_get_alt() || tool_get_esc()){
		EditEditMenuButtonCB(edit->EditFD->EditMenuButton, TRUE);
		return TRUE;
	    }
	    break;
	case XK_F:
	case XK_f:
	    if(tool_get_alt() || tool_get_esc()){
		EditFileMenuButtonCB(edit->EditFD->FileMenuButton, TRUE);
		return TRUE;
	    }
	    break;
	case XK_O:
	case XK_o:
	    if(tool_get_alt() || tool_get_esc()){
		do_options_menu_direct(edit);
		return TRUE;
	    }
	    break;
	case XK_Escape:
	    //cerr << "escape true.\n";
	    tool_escape_flag = TRUE;
	    return TRUE; 
	case XK_Q:
	case XK_q:
	    if(tool_get_alt() || tool_get_esc()){
		menu_edit = edit;
		edit_close_cb(0);
		return TRUE;
	    }
	    break;
	case XK_S:
	case XK_s:
	    if(tool_get_ctrl()){
		menu_edit = edit;
		edit_search_cb(0);
		edit->set_wintitle();
		// search mode.
		return TRUE;
	    } else if (tool_get_alt() || tool_get_esc()) {
		edit->save();
		return TRUE;
	    }
	    break;
	case XK_W:
	case XK_w:
	    if (tool_get_alt() || tool_get_esc()) {
	        WindowsMenuHiddenButtonCB(NULL, 0);
	        return TRUE; 
	    }
	    break;
	case XK_X:
	case XK_x:
	    if(tool_get_alt() || tool_get_ctrl() || tool_get_esc()){
		edit->cut();
		return TRUE;
	    }
	    break;
/*
	case XK_P:
	case XK_p:
 */
	case XK_Y:
	case XK_y:
	    if(tool_get_alt() || tool_get_ctrl() || tool_get_esc()){
		edit->paste();
		return TRUE;
	    }
	    break;
	default:
	    if(direct_window_access_cb(xkev, keysym, TRUE)) return TRUE;
	    break;
	}
    }
    return FALSE;
}

//extern int edit_handle_free(FL_OBJECT *obj, int event, FL_Coord, FL_Coord, int, void *);
//fl_add_free(FL_INACTIVE_FREE,0,0,1,1,"",edit_handle_free);

int edit_handle_free(FL_OBJECT *ob, int event, FL_Coord, FL_Coord, int, void *){
    if(event == FL_DRAW){
	if(!vob_valid(ob, "Free")) return 0;
	edit_c *edit = (edit_c *) ob->form->u_vdata;
	// make a geometry sanity check, also save the geometry.
	edit->var_sync();
    }
    return 0;
}

/**** RESOURCE SUPPORT ****/

static int n_edefs = 0;

static prm_entry_t global_prm_table[] = {
    {"n_edefs",      int_e, &n_edefs}, 
    {"",             int_e, NULL}
};

static prm_entry_t item_prm_table[] = {
    {"ww",           int_e, NULL}, 
    {"wh",           int_e, NULL},   
    {"wx",           int_e, NULL},   
    {"wy",           int_e, NULL},   
    {"fontsize",     int_e, NULL}, 
    {"fontstyle",    int_e, NULL}, 
    {"fg_colour",    int_e, NULL}, 
    {"bg_colour",    int_e, NULL}, 
    {"hi_colour",    int_e, NULL}, 
    {"",             int_e, NULL}
};

static void make_table(edef_c *edef, char *ID_string){
    if(!edef || !ID_string){
	M_err("make_table","Null ptrs");
	return;
    }
    sprintf(ID_string, "edit%d", edef->ID);
    item_prm_table[0].data_p = &edef->ww;
    item_prm_table[1].data_p = &edef->wh;
    item_prm_table[2].data_p = &edef->wx;
    item_prm_table[3].data_p = &edef->wy;
    item_prm_table[4].data_p = &edef->fontsize;
    item_prm_table[5].data_p = &edef->fontstyle;
    item_prm_table[6].data_p = &edef->fg_colour;
    item_prm_table[7].data_p = &edef->bg_colour;
    item_prm_table[8].data_p = &edef->hi_colour;
}

int edit_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    // check wether it is a entry in the history list.
    if (edit_history_load_default(obj_tok, name_tok, val_tok)) {
	return TRUE;
    }
    // when loading defaults, make sure no edits are left open.
    while(busy_edef.getsize()){
	busy_edef.headend();
	edef_c *edef = (edef_c *)busy_edef.getval();
	if(edef && edef->edit){
	    delete edef->edit;
	}
    }
    if(resource_load_default(global_prm_table, "edit", obj_tok, name_tok, val_tok)){
	return TRUE;
    }

    // make new edefs if necessary. 
    while(free_edef.getsize() < n_edefs){
	edef_c *edef = new edef_c;
	free_edef.add(edef);
    }
    for(int i=0; i<n_edefs; i++){
	char ID_string[10];
	make_table((edef_c *)free_edef[i], ID_string);
	if(resource_load_default(item_prm_table, ID_string, obj_tok, name_tok, val_tok)){
	    return TRUE;
	}
    }
    return FALSE;
}

void edit_save_defaults(ofstream &ofile){
    int i;
    //var_sync();
    n_edefs = free_edef.getsize() + busy_edef.getsize();
    resource_save_defaults(ofile, global_prm_table, "edit");
    char ID_string[10];
    for(i=0; i< free_edef.getsize(); i++){
	make_table((edef_c *)free_edef[i], ID_string);
	resource_save_defaults(ofile, item_prm_table, ID_string);
    }
    for(i=0; i< busy_edef.getsize(); i++){
	edef_c *edef = (edef_c *)busy_edef[i];
	if(edef){
	    if(edef->edit) edef->edit->var_sync();
	    make_table(edef, ID_string);
	    resource_save_defaults(ofile, item_prm_table, ID_string);
	}
    }
    edit_history_save_defaults(ofile);
}

static int edef_pointers_ok(edef_c *pEdef) {
  if (pEdef && pEdef->edit && pEdef->edit->EditFD && pEdef->edit->EditFD->EditForm){
    return TRUE;
  }
  return FALSE;
}

int is_editor_window(Window window) {
  for(int i=0; i< busy_edef.getsize(); i++){
    edef_c *pEdef = (edef_c *)busy_edef[i];
	
    if(edef_pointers_ok(pEdef)) {
      if(pEdef->edit->EditFD->EditForm->window == window) {
	return TRUE;
      }
    }
  }
  return FALSE;
}

void edit_sync_defaults(void){
    // after loading. don't do shit, don't affect the visible textedits.??
    // only affect the way the next textedit will look.
    /*
      fl_freeze_form(EditFileForm);
      fl_set_textedit_fontstyle(EditFileTextedit, edit_fontstyle);
      fl_set_textedit_fontsize(EditFileTextedit, edit_fontsize);
      edit_sync_colours();
      fl_unfreeze_form(EditFileForm);
    
      if(EditFileForm->visible){
      fl_set_form_position(EditFileForm, edit_wx, edit_wy);
      fl_set_form_size    (EditFileForm, edit_ww, edit_wh);
      }
      */
}

/**** INTERFACE ****/

int quick_edit_file(const char *fname, const char *dir){
    char buffer[MAXPATHLEN+16];
    if(!fname){
	program_unstable("No directory or file name in quick_edit_file.");
	return FALSE;
    } 
    if(dir){
	sprintf(buffer, "%s/%s", dir, fname);
    } else {
	strcpy(buffer, fname);
    }
    return quick_edit_file(buffer);
}

int quick_edit_file(const char *fname){
    if (too_many_editors()) return FALSE;
    edit_c *edit = new edit_c;
    if(!edit->show(fname)){
	delete edit;
	return FALSE;
    }
    return TRUE;
}

/**** Colour Support ****/
static int cv_valid(void){
    if(!colour_edit)edit_colour_close_form();
    return(colour_edit != NULL);
}

void EditFBColourButtonCB(FL_OBJECT *, long val){
    if(!cv_valid()) return;
    int &col = val ? colour_edit->edef->bg_colour : colour_edit->edef->fg_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? EditBgColourButton : EditFgColourButton, col, 47);    
	colour_edit->sync_colours();
    }
}

void EditCursorColourButtonCB(FL_OBJECT *, long){
    if(!cv_valid()) return;
    int new_col = fl_show_colormap(colour_edit->edef->hi_colour);
    if(new_col != colour_edit->edef->hi_colour){
	colour_edit->edef->hi_colour = new_col;
	fl_set_object_color(EditCursorColourButton, colour_edit->edef->hi_colour, 47);    
	colour_edit->sync_colours();
    }
}

static void edit_colour_close_form(void){
    if(EditColourForm->visible){
	focus_restore_start();
	fl_hide_form(EditColourForm);
	focus_restore_finish();
    }
}

void EditColourCloseReturnButtonCB(FL_OBJECT *, long){
    edit_colour_close_form();
}

void EditColourHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(quick_edit_colour_help_text, "Edit Colour Editing");
}

void EditColourDefaultsButtonCB(FL_OBJECT *, long){
    if(!cv_valid()) return;
    colour_edit->edef->default_colours();
    edit_colour_form_sync();
    colour_edit->sync_colours();
}

static void edit_colour_help_func(void){
    EditColourHelpButtonCB(NULL,0);
}

static int edit_colour_form_raw_cb(FL_FORM *form, void *xev){
    XEvent *xevent = (XEvent *)xev;
    if(form != EditColourForm){
	return FALSE;
    }
    if(adjust_return_button(EditColourForm, xevent)){
	return TRUE;
    }
    return standard_form_event_cb(EditColourForm, xevent, NULL, edit_colour_help_func);
}

static void edit_colour_form_wintitle(void){
    if(!colour_edit) return;
    if(!EditColourForm->visible) return;
    char buffer[MAXPATHLEN];
    sprintf(buffer, "Adjust Edit #%d Colours:", colour_edit->edef->ID);
    XStoreName(fl_display, EditColourForm->window, buffer);
}

static void edit_colour_form_sync(int bSyncWintitle){
    if (!colour_edit) return;
    fl_set_object_color(EditFgColourButton,     colour_edit->edef->fg_colour, 47);
    fl_set_object_color(EditBgColourButton,     colour_edit->edef->bg_colour, 47);
    fl_set_object_color(EditCursorColourButton, colour_edit->edef->hi_colour, 47);

    if (bSyncWintitle){
	edit_colour_form_wintitle();
    }
}

static void edit_colour_show_form(void){
    if(!colour_edit) return;
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(EditColourForm, FL_ALL_EVENT, edit_colour_form_raw_cb);
	fl_set_object_shortcut(EditColourCloseReturnButton, "Cc^M", TRUE);
	fl_set_button_shortcut(EditColourDefaultsButton, "Dd", TRUE);
    }
    edit_colour_form_sync();
    set_to_ret_button(EditColourCloseReturnButton);
    focus_save();
    char buffer[MAXPATHLEN];
    sprintf(buffer, "Adjust Edit #%d Colours:", colour_edit->edef->ID);
    if(!EditColourForm->visible){
	fl_show_form(EditColourForm, FL_PLACE_CENTER, TRUE, buffer);
    } else {
	XMapRaised(fl_display, EditColourForm->window);
	XStoreName(fl_display, EditColourForm->window, buffer);
    }
    tool_focus_move(EditColourForm, EditColourCloseReturnButton);
}
 
int edit_c::allready_open_ok(){
    for (int i=0; i<busy_edef.getsize(); i++){
	edef_c *pEdef = (edef_c *) busy_edef[i];
	if (pEdef && 
	    pEdef->edit && 
	    pEdef->edit != this &&
	    (strcmp(fullpath, pEdef->edit->fullpath) == 0)
	) {
	    ostrstream streamQuestion;
	    streamQuestion << "File `";
	    streamQuestion << fullpath;
	    streamQuestion << "' is allready open in Edit ";
	    streamQuestion << pEdef->ID;
	    streamQuestion << ". Really open again?";
	    streamQuestion << '\0';

	    ostrstream streamSwitchTo;
	    streamSwitchTo << "Switch to Edit ";
	    streamSwitchTo << pEdef->ID;
	    streamSwitchTo << '\0';

	    int retval = ask_multi_question(
		FALSE, 0, streamQuestion.str(),
		streamSwitchTo.str(), 	"Ss", NoSymbol, 
		"Open Again",   	"Oo", NoSymbol, 
		"Cancel",  		"Cc", XK_Escape
	    );
	    switch(retval) {
	    case -1:
		// esc/cancel
		return FALSE;
	    case 0:
		// switch (raise & focus)
		pEdef->edit->win_raise();
		return FALSE;
	    case 1:
		// open again
		return TRUE;
	    default:
		return FALSE;
	    }
	}
    }
    return TRUE;
}

int get_edit_changed_count(char *fname_buf, unsigned int fname_buf_size){
    edef_c *pEdef;
    int count = 0;
    if (fname_buf_size > 0) {
	strcpy(fname_buf, "");
    }
    for (int i=0; i<busy_edef.getsize(); i++){
	pEdef = (edef_c *) busy_edef[i];
	if (pEdef && pEdef->edit) {
	    if (pEdef->edit->has_changed()) {
		const char *fname = pEdef->edit->get_fname();
		if (fname && 
		    fname_buf &&
		    (strlen(fname_buf) + strlen(fname) + 3 < fname_buf_size)
		){
		    if (count) strcat(fname_buf, ", ");
		    strcat(fname_buf, fname);
		}
		count++;
		
	    }
	}
    }
    if (dbg_flag) cerr << "count: " << count << endl;
    return count;
}

int direct_editor_access_cb(int nId, int){
    const edef_c *pEdef;
    int nEdefs = busy_edef.getsize();
    if (nEdefs <= 0) return FALSE;
    for (int i=0; i<nEdefs; i++) {
	pEdef = (edef_c *) busy_edef[i];
	if (pEdef->ID == nId && pEdef->edit) {
	    pEdef->edit->win_raise();
	    return TRUE;
	}
    }
    return FALSE;
}

static int lValueOffset = 0;

static void do_editor(int IndirectID){
  // subtract browser offset.
  IndirectID -= lValueOffset;

  const edef_c *pEdef = (edef_c *) busy_edef[IndirectID];
  if (!pEdef) {
    cerr << "bad ID (" << IndirectID << ") in do_editor.\n";
    return;
  }
  sginap(10);
  direct_editor_access_cb(pEdef->ID, TRUE);
}

int AddEditorsToMenu(long window_list_menu, int lValueOffsetIV){
  char buffer[MAXPATHLEN + 64];
  const edef_c *pEdef;
  if (window_list_menu <= 0) return -1;
  lValueOffset = lValueOffsetIV;

  int nEdefs = busy_edef.getsize();
  if (nEdefs <= 0) return 0;

  for (int i=0; i<nEdefs; i++) {
    pEdef = (edef_c *) busy_edef[i];
    if (!pEdef || !pEdef->edit) {
      cerr << "Bad edef (" << i << ") in AddEditorsToMenu.\n";
      return -1;
    }
    int nID = pEdef->ID;

    int nShortcut = i+lValueOffset;
    if (nShortcut < 10) {
      sprintf(buffer, "%d:  Editor #%d: %s (Alt-%d, Esc-%d)    %cf", 
	      nShortcut, nID, pEdef->edit->get_path(), nID, nID, '%');
      char digit[8];
      sprintf(digit, "%d", nShortcut);
      add_to_pup(window_list_menu, buffer, (long) do_editor, digit);
    } else {
      sprintf(buffer, "     Editor #%d: %s (Alt-%d, Esc-%d)    %cf", 
	      nID, pEdef->edit->get_path(), nID, nID, '%');
      add_to_pup(window_list_menu, buffer, (long) do_editor, "");
    }
  }
  return nEdefs;
}


// no more than 10 files in history, and remember abs path.

static int needs_remake_flag = TRUE;
static long lEditHistoryMenu = 0;
static const char *edit_history_menu_ID_string = "edit_history";

static void last_edit_cb(int no){
    if(no < 0 || no >= file_list.getsize()){
	cerr << "Invalid no(" << no << ") in last_edit_cb.\n";
	return;
    }
    char *val = (char *) file_list[no];
    if(val){  
	if (!quick_edit_file(val)) {
	  // cannot concentrate dup removal here, because we
	  // also need it if we try loading via a different mechanism.
	  //edit_history_remove_duplicates(val);
	}
    }
}

static long make_edit_history_menu(void){
    char buffer[MAXPATHLEN + 64], *val;
    if (!needs_remake_flag){
	return lEditHistoryMenu;
    } else {
	if (lEditHistoryMenu > 0) {
	    free_pup(lEditHistoryMenu);
	}
	needs_remake_flag = FALSE;
    }

    lEditHistoryMenu = new_pup();

    add_to_pup(lEditHistoryMenu, "Edit History (Meta-F5)%t", 0);

    int size = file_list.getsize();
    if(size <= 0){
	size = file_list.getsize();
    }
    for(int i=0; i<size; i++){
	val = (char *) file_list[i];
	if(val){
	    if(i<10){
		sprintf(buffer, "%d:  %s%cx%d", i, val, '%', i);
		char digit[8];
		sprintf(digit, "%d", i);
		add_to_pup(lEditHistoryMenu, buffer,   (long) last_edit_cb, digit);
	    } else {
		sprintf(buffer, "     %s%cx%d", val, '%', i);
		add_to_pup(lEditHistoryMenu, buffer,   (long) last_edit_cb, "");
	    }
	} else {
	    cerr << "illegal val in make_edit_history_menu. \n";
	}
    }
    return lEditHistoryMenu;
}

static void edit_history_remove_duplicates(const char *pcFile){
    char *val;
    if(!pcFile) return;
    int size = file_list.getsize();
    if(size > 0){
	for(int i=0; i<size; i++){
	    val = (char *) file_list[i];
	    if(val && (strcmp(val, pcFile) == 0)){
		fl_free(val);
		file_list.remove();
		size--;
		needs_remake_flag = TRUE;
	    }
	}
    }
}

static void edit_history_remember_last(const char *pcFile){
    if(!pcFile) return;
    // remove duplicates before insert at top.
    edit_history_remove_duplicates(pcFile);

    int size = file_list.getsize();
    if (size > 9) {
    	file_list.tailend();
    	char *val = (char *) file_list.getval();
	if (val) fl_free(val);
	file_list.remove();
    }
    file_list.headend();
    file_list.insert(strdup(pcFile));
    needs_remake_flag = TRUE;
}

void EditHistoryCB(void){
    int size = file_list.getsize();
    if (size == 0) {
	timed_notifier("The edit history list is empty at the moment.");
	return;
    }
    long history_menu = make_edit_history_menu();
    if (history_menu <= 0) return;
    do_pup(history_menu);
}

static int edit_history_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
) {
    if(!obj_tok || !name_tok || !val_tok) return FALSE;
    
    if(strcmp(obj_tok, edit_history_menu_ID_string) == 0){
	if(strcmp(name_tok, "history_list_entry") == 0){
	    edit_history_remember_last(val_tok);
	    return TRUE;
	}
    }
    return FALSE;
}

static void edit_history_save_defaults(ofstream &ofile){
    char *buf;
    ofile << "# edit history list resources.\n";
    ofile << "#\n";
    int n_items = file_list.getsize();
    for(int i=0; i< n_items; i++){
	buf = (char *) file_list[n_items - (i+1)];
	if(buf){
	    ofile << edit_history_menu_ID_string << ".history_list_entry:      " << buf << endl;
	}
    }
}

static int too_many_editors(void){
    int nSize = busy_edef.getsize();
    if (nSize > 30 && nSize < 35) {
	if(!ask_question(TRUE, 
	    "More than 30 editors are not recommended. Continue?")
	){
	    return TRUE;
	}
    }
    if (nSize >= 35) {
	timed_error_notifier(FALSE, "Not supported: more than 35 editors.\n");
	return TRUE;
    }
    return FALSE;
}


