
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "forms.h"
#include "error_notifier_forms.h"
#include "error_notifier_forms_aux.h"
#include "gl_utility.h"
#include "global.h"

double notifier_secs = 8.0;

void timed_notifier(const char *text){
    timed_notifier(text, notifier_secs);
}

static void context_func(void){
    // empty, cannot context nor show error message.
}

static void t_close_func(void){
    if(TimedErrorNotifierForm->visible){
	focus_restore_start();
	fl_hide_form(TimedErrorNotifierForm);
	focus_restore_finish();
    }
}

static int t_form_raw_cb(FL_FORM *form, void *xev){
    return standard_form_event_cb(form, (XEvent *) xev, t_close_func, context_func);
}

void t_notifier(const char *text, double secs){
    static int first = TRUE;
    if(first){
	first = FALSE;
	make_standard_cancel_shortcut(TimedErrorNotifierCloseReturnButton);
	fl_register_call_back(TimedErrorNotifierForm, FL_ALL_EVENT, t_form_raw_cb);
    }
    if(text == NULL) return;

    fl_set_object_label(TimedErrorNotifierText, text);
    if(secs > 0.0){
	fl_set_timer(TimedErrorNotifierTimer, secs);
    }

    fl_deactivate_all_forms();

    int text_width = fl_get_string_width(TimedErrorNotifierText->lstyle, TimedErrorNotifierText->lsize, text, strlen(text));
    int tmp_x = TimedErrorNotifierCloseReturnButton->x;
    int tmp_w = TimedErrorNotifierCloseReturnButton->w;
    int new_w = tmp_x + tmp_w + text_width + 20 + tmp_w + 10;
    fl_set_form_size(TimedErrorNotifierForm, new_w, TimedErrorNotifierForm->h);

    TimedErrorNotifierCloseReturnButton->x = tmp_x;
    TimedErrorNotifierCloseReturnButton->w = tmp_w;
    TimedErrorNotifierText->x  = tmp_x + tmp_w;
    TimedErrorNotifierText->w  = text_width + 20;
    TimedErrorNotifierTimer->x = TimedErrorNotifierText->x + TimedErrorNotifierText->w;
    TimedErrorNotifierTimer->w = tmp_w;

    fl_activate_form(TimedErrorNotifierForm);
    focus_save();
    fl_show_form(TimedErrorNotifierForm, FL_PLACE_CENTER, FALSE, NULL);
    tool_focus_move(TimedErrorNotifierForm, TimedErrorNotifierCloseReturnButton);
    while(TimedErrorNotifierForm->visible){
	if(!tool_mung_forms()){
	    t_close_func();
	}
    }
    fl_activate_all_forms(); 
}			

void TimedErrorNotifierCloseReturnButtonCB(FL_OBJECT*, long){
    t_close_func();
}

static void t2_close_func(void){
    if(TimedErrorNotifier2Form->visible){
	focus_restore_start();
	fl_hide_form(TimedErrorNotifier2Form);
	focus_restore_finish();
    }
}

void TimedErrorNotifier2CloseReturnButtonCB(FL_OBJECT*, long){
    t2_close_func();
}

void timed_notifier(const char *text, double secs){
    if(text == NULL) return;
    fl_set_object_lcol(TimedErrorNotifierTimer, FL_BLACK);
    fl_set_object_label(TimedErrorNotifierTimer, "Notifier");
    t_notifier(text, secs);
}

void timed_error_notifier(const char *text, double secs){
    if(text == NULL) return;
    fl_set_object_lcol(TimedErrorNotifierTimer, FL_RED);
    fl_set_object_label(TimedErrorNotifierTimer, "Error !");
    t_notifier(text, secs);
}

void timed_error_notifier(const char *text){
    if(text == NULL) return;
    fl_set_object_lcol(TimedErrorNotifierTimer, FL_RED);
    fl_set_object_label(TimedErrorNotifierTimer, "Error !");
    t_notifier(text, notifier_secs);
}

static int t2_form_raw_cb(FL_FORM *form, void *xev){
    return standard_form_event_cb(form, (XEvent *) xev, NULL, context_func);
}

void t2_notifier(
    const char *text_line_0, 
    const char *text_line_1, 
    double secs
){
    static int first = TRUE;
    if(first){
	first = FALSE;
	make_standard_cancel_shortcut(TimedErrorNotifier2CloseReturnButton);
	fl_register_call_back(TimedErrorNotifier2Form, FL_ALL_EVENT, t2_form_raw_cb);
    }

    if(text_line_0 == NULL) return;
    if(text_line_1 == NULL){
	timed_error_notifier(text_line_0, secs);
    }

    fl_set_object_label(TimedErrorNotifier2TextLine0, text_line_0);
    fl_set_object_label(TimedErrorNotifier2TextLine1, text_line_1);
    if(secs > 0.0){
	fl_set_timer(TimedErrorNotifier2Timer, secs);
    }

    fl_deactivate_all_forms();

    int text_width = fl_get_string_width(TimedErrorNotifier2TextLine0->lstyle, TimedErrorNotifier2TextLine0->lsize, text_line_0, strlen(text_line_0));
    int tmp_width  = fl_get_string_width(TimedErrorNotifier2TextLine1->lstyle, TimedErrorNotifier2TextLine1->lsize, text_line_1, strlen(text_line_1));
    if(tmp_width > text_width) text_width = tmp_width;

    int tmp_x = TimedErrorNotifier2CloseReturnButton->x;
    int tmp_w = TimedErrorNotifier2CloseReturnButton->w;
    int new_w = tmp_x + tmp_w + text_width + 20 + 5;

    fl_set_form_size(TimedErrorNotifier2Form, new_w, TimedErrorNotifier2Form->h);
    
    TimedErrorNotifier2CloseReturnButton->x = tmp_x;
    TimedErrorNotifier2CloseReturnButton->w = tmp_w;
    TimedErrorNotifier2Timer->x = tmp_x;
    TimedErrorNotifier2Timer->w = tmp_w;
    TimedErrorNotifier2TextLine0->w = text_width + 20;
    TimedErrorNotifier2TextLine1->w = text_width + 20;
    TimedErrorNotifier2TextLine0->x = tmp_x + tmp_w;
    TimedErrorNotifier2TextLine1->x = tmp_x + tmp_w;

    fl_activate_form(TimedErrorNotifier2Form);
    focus_save();
    fl_show_form(TimedErrorNotifier2Form, FL_PLACE_CENTER, FALSE, NULL);
    tool_focus_move(TimedErrorNotifier2Form, TimedErrorNotifier2CloseReturnButton);
    while(TimedErrorNotifier2Form->visible){
	if(!tool_mung_forms()){
	    t2_close_func();
	}
    }
    fl_activate_all_forms(); 
}


void timed_notifier(
    const char *text_line_0, 
    const char *text_line_1 
){
    timed_notifier(text_line_0, text_line_1, notifier_secs);
}

void timed_notifier(
    const char *text_line_0, 
    const char *text_line_1, 
    double secs
){
    fl_set_object_lcol(TimedErrorNotifier2Timer, FL_BLACK);
    fl_set_object_label(TimedErrorNotifier2Timer, "Notifier");
    t2_notifier(text_line_0, text_line_1, secs);
}

void timed_error_notifier(
    const char *text_line_0, 
    const char *text_line_1,
    double secs
){
    fl_set_object_lcol(TimedErrorNotifier2Timer, FL_RED);
    fl_set_object_label(TimedErrorNotifier2Timer, "Error !");
    t2_notifier(text_line_0, text_line_1, secs);
}

void timed_error_notifier(
    const char *text_line_0, 
    const char *text_line_1
){
    fl_set_object_lcol(TimedErrorNotifier2Timer, FL_RED);
    fl_set_object_label(TimedErrorNotifier2Timer, "Error !");
    t2_notifier(text_line_0, text_line_1, notifier_secs);
}

/* add a line containing strerror(errno) */
void timed_strerror_notifier(const char *text, double secs){
    char buffer[1024];
    sprintf(buffer, "Reason: %s", strerror(errno));
    timed_error_notifier(text, buffer, secs);
}

void timed_strerror_notifier(const char *text){
    char buffer[1024];
    sprintf(buffer, "Reason: %s", strerror(errno));
    timed_error_notifier(text, buffer, notifier_secs);
}

/* don't use the timer */
void error_notifier(const char *text){
    timed_error_notifier(text, (double) 0.0);
}

void fl_error_msg(const char *fmt, ...){
    va_list arg;
    char msg[8192];	/* extremely dirty, I don't know how long fmt
			   will be after insertion of arguments!!! */

    va_start(arg, fmt);
    vsprintf(msg, fmt, arg);
    va_end(arg);

    //fl_show_alert(msg, "", "", TRUE);
    error_notifier(msg);
    
    return;
}
