/*
 *  file_utility.h:
 *  Define some always used functions.
 */

#ifndef FILE_UTILITY_H
#define FILE_UTILITY_H

#include <sys/types.h>

/* try wether we can access a file, and how */
int try_access(const struct stat *statbuf_p, mode_t mode);

/* do low-level checks wether we're dealing with valid directories. 
 * if allow_equality_flag is set, don't barf if the directories are
 * the same (device/inode-wise).
 */

typedef enum {error_e, identical_e, ok_e} dir_validate_et;

dir_validate_et validate_source_target_dirs(
    const char *source_dir, 
    const char *target_dir
);

/* can we stat the dir, and is it a dir ? 
 * also, have a copy with `source' error msg.
 */
int validate_dir(const char *dir);
int validate_source_dir(const char *source_dir);

// how many files in the directory ? 
int files_in_dir(const char *dir);

/* is it '.' or '..' ? */
int dot_dotdot_check(const char *fname);
int dot_dotdot_test(const char *fname, const char *msg = "Action");

const char *get_basename(const char *path);
const char *get_dirname (const char *path);
const char *strip_trailing_slashes(const char *path);

/* meaningful chdir (need r-x, not only --x, for browsers) */
int safe_chdir(const char *path);

/* for avoiding infinite loops */
int dir_is_child(const char *child, const char *ancestor);

char *real_path(const char *path, char *resolved);

#endif	// FILE_UTILITY_H
