/*
 * Allow for variations between differing X-Releases.
 */


#ifndef FIX_KEYSYMS_H
#define FIX_KEYSYMS_H

#include <X11/keysym.h>

/* two strategies are possible. 
 * a) if the keysym isn't defined, define it to a useful value that
 *    may be correct on the system it is compiled on. Scenario: Run 
 *    a non-Sun UC on a Sun display. Possible big fuck-up if keysyms
 *    go wild.
 * b) if it isn't defined, alias it to a known keysym. Safer, but
 *    less functionality. 
 * c) both are pretty bad, as they lead to duplicate case values.
 */

#ifndef SunXK_F36
#define SunXK_F36 0x1005FF10	/* F11 */
#endif

#ifndef SunXK_F37
#define SunXK_F37 0x1005FF11	/* F12 */
#endif

#ifndef SunXK_PageUp
#define SunXK_PageUp		0x0000FF55 	/* Same as XK_Prior */
#endif

#ifndef SunXK_PageDown
#define SunXK_PageDown		0x0000FF56	/* Same as XK_Next */
#endif

/* taken from Linux 2.0 X11R6 
 * can lead to duplicate case values. Better to have global is_{delete, bla}
 * function that does the #ifdef once and for all. 
 */
#if 0
#ifndef XK_KP_Home
#define XK_KP_Home		0xFF95
#endif
#ifndef XK_KP_Left
#define XK_KP_Left		0xFF96
#endif
#ifndef XK_KP_Up
#define XK_KP_Up		0xFF97
#endif
#ifndef XK_KP_Right
#define XK_KP_Right		0xFF98
#endif
#ifndef XK_KP_Down
#define XK_KP_Down		0xFF99
#endif
#ifndef XK_KP_Prior
#define XK_KP_Prior		0xFF9A
#endif
#ifndef XK_KP_Page_Up
#define XK_KP_Page_Up		0xFF9A
#endif
#ifndef XK_KP_Next
#define XK_KP_Next		0xFF9B
#endif
#ifndef XK_KP_Page_Down
#define XK_KP_Page_Down		0xFF9B
#endif
#ifndef XK_KP_End
#define XK_KP_End		0xFF9C
#endif
#ifndef XK_KP_Begin
#define XK_KP_Begin		0xFF9D
#endif
#ifndef XK_KP_Insert
#define XK_KP_Insert		0xFF9E
#endif
#ifndef XK_KP_Delete
#define XK_KP_Delete		0xFF9F
#endif
#ifndef XK_KP_Equal
#define XK_KP_Equal		0xFFBD	/* equals */
#endif
#ifndef XK_KP_Multiply
#define XK_KP_Multiply		0xFFAA
#endif
#ifndef XK_KP_Add
#define XK_KP_Add		0xFFAB
#endif
#ifndef XK_KP_Separator
#define XK_KP_Separator		0xFFAC	/* separator, often comma */
#endif
#ifndef XK_KP_Subtract
#define XK_KP_Subtract		0xFFAD
#endif
#ifndef XK_KP_Decimal
#define XK_KP_Decimal		0xFFAE
#endif
#ifndef XK_KP_Divide 
#define XK_KP_Divide		0xFFAF
#endif
#endif	// 0

#endif  // FIX_KEYSYMS_H
