
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

#include "browser.h"
#include "browser_menu.h"
#include "cb_key.h"
#include "cb_fkey.h"
#include "client_msg.h"
#include "error_notifier_forms_aux.h"
#include "bitmaps/fkeybar.bm"
#include "fkeybar_forms.h"
#include "fkeybar_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "status_forms.h"
#include "status_forms_aux.h"
#include "toolbox_forms.h"
#include "toolbox_forms_aux.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "xshell.h"

int
    fkeybar_ww = 600,
    fkeybar_wh = 50,
    fkeybar_wx = 10,
    fkeybar_wy = 10
    /*
    fkeybar_ww = default_fkeybar_width, 
    fkeybar_wh = 2*default_browser_height + 2*xshell_real_bwidth + xshell_real_topbar_height, 
    fkeybar_wx = default_browser_width + 3*xshell_real_bwidth,
    fkeybar_wy = 0
    */
;

int         fkeybar_visible_flag = TRUE;
const char *fkeybar_ID_string = "fkeybar";

prm_entry_t fkeybar_prm_table[] = {
    {"ww",           int_e,    (void *) &fkeybar_ww}, 
    {"wh",           int_e,    (void *) &fkeybar_wh}, 
    {"wx",           int_e,    (void *) &fkeybar_wx}, 
    {"wy",           int_e,    (void *) &fkeybar_wy}, 
    {"visible_flag", int_e,    (void *) &fkeybar_visible_flag}, 
    {"",             int_e,    NULL}
};

static void var_sync(void){
    fkeybar_ww = FKeyBarForm->w;
    fkeybar_wh = FKeyBarForm->h;
    fkeybar_wx = FKeyBarForm->x;
    fkeybar_wy = FKeyBarForm->y;
}

void fkeybar_save_defaults(ofstream &ofile){
    if(FKeyBarForm->visible){
	var_sync();
    }
    resource_save_defaults(ofile, fkeybar_prm_table, fkeybar_ID_string);
}    

int fkeybar_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    return resource_load_default(
        fkeybar_prm_table,  
        fkeybar_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

void fkeybar_set_fontsize(int size){
    fl_freeze_form(FKeyBarForm);
    fl_set_object_lsize(FKeyBarFormF1Text,size);
    fl_set_object_lsize(FKeyBarFormF2Text,size);
    fl_set_object_lsize(FKeyBarFormF3Text,size);
    fl_set_object_lsize(FKeyBarFormF4Text,size);
    fl_set_object_lsize(FKeyBarFormF5Text,size);
    fl_set_object_lsize(FKeyBarFormF6Text,size);
    fl_set_object_lsize(FKeyBarFormF7Text,size);
    fl_set_object_lsize(FKeyBarFormF8Text,size);
    fl_set_object_lsize(FKeyBarFormF9Text,size);
    fl_set_object_lsize(FKeyBarFormF10Text,size);
    fl_set_object_lsize(FKeyBarFormF11Text,size);
    fl_set_object_lsize(FKeyBarFormF12Text,size);

    fl_set_object_lsize(FKeyBarFormF1Button,size);
    fl_set_object_lsize(FKeyBarFormF2Button,size);
    fl_set_object_lsize(FKeyBarFormF3Button,size);
    fl_set_object_lsize(FKeyBarFormF4Button,size);
    fl_set_object_lsize(FKeyBarFormF5Button,size);
    fl_set_object_lsize(FKeyBarFormF6Button,size);
    fl_set_object_lsize(FKeyBarFormF7Button,size);
    fl_set_object_lsize(FKeyBarFormF8Button,size);
    fl_set_object_lsize(FKeyBarFormF9Button,size);
    fl_set_object_lsize(FKeyBarFormF10Button,size);
    fl_set_object_lsize(FKeyBarFormF11Button,size);
    fl_set_object_lsize(FKeyBarFormF12Button,size);

    fl_set_object_lsize(FKeyBarFormMetaButton,size);
    fl_unfreeze_form(FKeyBarForm);
}

void fkeybar_close_form(void){
    static int first = TRUE;
    if(FKeyBarForm->visible){
	if(first){
	    first = FALSE;
	    if(!OptionsForm->visible && tool_looser_flag){
		timed_notifier("use Ctrl-F2 Options to get the fkeybar back.", notifier_secs);
	    }
	}
	var_sync();
	fl_hide_form(FKeyBarForm);
    }
    fkeybar_visible_flag = FALSE;
    first = FALSE;
    options_sync_defaults();
}

void fkeybar_iconify(void){    
    if(FKeyBarForm->visible) XIconifyWindow(fl_display, FKeyBarForm->window, fl_screen);
}

void fkeybar_uniconify(void){    
    if(FKeyBarForm->visible) XMapRaised(fl_display, FKeyBarForm->window);
}

int fkeybar_raw_cb(FL_FORM *form, void *xev){
    // a few are special: F1, F2.
    XEvent *xevent = (XEvent *) xev;
    KeySym keysym;
    char keybuf[1];

    switch(xevent->type){
    case KeyPress:
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	    /*
	      case XK_Control_L:
	      case XK_Control_R:
	      tool_ctrl_on_flag = TRUE;
	      fkeybar_meta_sync();
	      break;
	      */
	default:
	    break;
	}
	break;
    case KeyRelease:
	//cerr << "release\n";
	/*
	  XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	  switch(keysym){
	  case XK_Control_L:
	  case XK_Control_R:
	  tool_ctrl_on_flag = FALSE;
	  fkeybar_meta_sync();
	  break;
	  default:
	  tool_ctrl_on_flag = controlkey_down(fl_keymask);
	  fkeybar_meta_sync();
	  break;
	  }
	  // special trick so that ctrl-key up/down works.
	  return TRUE;
	  */
    default:
	break;
    }
    if(standard_form_event_cb(form, xevent, fkeybar_close_form, fkeybar_help_func)){
	return TRUE;
    } else {
	if (toolbox_tree_flag) {
	    return tree_handle_free(TreeFormFreeObj, FL_OTHER, 0, 0, 0, xevent);
	} else {
	    return browser_handle_free(source_browser->free_obj, FL_OTHER, 0, 0, 0, xevent);
	}
    }
}

void fkeybar_help_func(void){
    load_help_browser(fkeybar_help_text, "Function Keybar");
}

void FKeyBarFormF1ButtonCB(FL_OBJECT *, long){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	HelpButtonCB(NULL, 0);
    } else {
	fkeybar_help_func();
    }
}

void FKeyBarFormF2ButtonCB(FL_OBJECT *, long){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	OptionsButtonCB(NULL, 0);
    } else {
	if (toolbox_tree_flag) {
	    TreeFormOptionsButtonCB(NULL, 0);
	} else {
	    browser_menu_event_cb(NULL, 0);
    	}
    }
}

void fkeybar_show_form(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(FKeyBarForm, FL_ALL_EVENT, fkeybar_raw_cb);
        fl_set_form_minsize(FKeyBarForm, 500, 40);
        fl_set_form_maxsize(FKeyBarForm, 1280, 100);
#ifdef DEC_ALPHA	
	fl_set_object_label(FKeyBarFormF11Button,"F13");
#endif
    }
    M_info("fkeybar::show_form", "x %d,y %d,w %d,h %d", fkeybar_wx, fkeybar_wy, fkeybar_ww, fkeybar_wh);
    fl_set_initial_placement(FKeyBarForm, fkeybar_wx, fkeybar_wy, fkeybar_ww, fkeybar_wh);
    fl_show_form(FKeyBarForm, FL_PLACE_FREE, TRUE, "UCFkeys");
    static Pixmap pixmap =0;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, FKeyBarForm->window, fkeybar_bits, fkeybar_width,
	    fkeybar_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    set_window_icon(FKeyBarForm->window, pixmap);
    fkeybar_set_wintitle();
    
    fkeybar_visible_flag = TRUE;
    options_sync_defaults();
    var_sync();
}


void fkeybar_metach_sync(void){
    //cerr << "fk-metach: " << tool_ctrl_on_flag << ", " << toolbox_tree_flag << endl;
    fl_freeze_form(FKeyBarForm);
    if(tool_ctrl_on_flag){
	fl_set_object_label(FKeyBarFormMetaButton, "M");
	fl_set_button(FKeyBarFormMetaButton, TRUE);
	
	//fl_set_object_label(FKeyBarFormFText,      "");	
	fl_set_object_label(FKeyBarFormF1Text,      "GlblHelp");
	fl_set_object_label(FKeyBarFormF2Text,      "GlblOpt");	
	fl_set_object_label(FKeyBarFormF3Text,      "MkPath");
	//fl_set_object_label(FKeyBarFormF4Text,      "(Tree)");
	fl_set_object_label(FKeyBarFormF5Text,      "EdHstry");
	//fl_set_object_label(FKeyBarFormF7Text,      "(Rename)");
	fl_set_object_label(FKeyBarFormF8Text,      "MvRenm");
	//fl_set_object_label(FKeyBarFormF9Text,      "(Duplicate)");
	fl_set_object_label(FKeyBarFormF10Text,     "CpRenm");
	//fl_set_object_label(FKeyBarFormF11Text,     "(Delete)");
	fl_set_object_label(FKeyBarFormF12Text,     "FastExit");

	if (toolbox_tree_flag) {
	    fl_hide_object(FKeyBarFormF6Button);
	    fl_hide_object(FKeyBarFormF6Text);
	    fl_hide_object(FKeyBarFormF8Button);
	    fl_hide_object(FKeyBarFormF8Text);
	    fl_hide_object(FKeyBarFormF10Button);
	    fl_hide_object(FKeyBarFormF10Text);
	} else {
	    fl_set_object_label(FKeyBarFormF6Text,      "MultiEdit");	
	}

	fl_hide_object(FKeyBarFormF4Button);
//	fl_hide_object(FKeyBarFormF5Button);
	fl_hide_object(FKeyBarFormF7Button);
	fl_hide_object(FKeyBarFormF9Button);
	fl_hide_object(FKeyBarFormF11Button);

	fl_hide_object(FKeyBarFormF4Text);
//	fl_hide_object(FKeyBarFormF5Text);
	fl_hide_object(FKeyBarFormF7Text);
	fl_hide_object(FKeyBarFormF9Text);
	fl_hide_object(FKeyBarFormF11Text);
    } else {
	fl_set_object_label(FKeyBarFormMetaButton, "m");
	fl_set_button(FKeyBarFormMetaButton, FALSE);
	fl_set_object_label(FKeyBarFormF1Text,      "Help");
	fl_set_object_label(FKeyBarFormF2Text,      "Menu");
	fl_set_object_label(FKeyBarFormF3Text,      "MkDir");
	fl_set_object_label(FKeyBarFormF4Text,      "Dir Tree");
	//fl_set_object_label(FKeyBarFormF7Text,      "Rename");
	fl_set_object_label(FKeyBarFormF8Text,      "Move");
	//fl_set_object_label(FKeyBarFormF9Text,      "Duplicate");
	fl_set_object_label(FKeyBarFormF10Text,     "Copy");
	//fl_set_object_label(FKeyBarFormF11Text,     "Delete");
	fl_set_object_label(FKeyBarFormF12Text,     "Exit");

	if (toolbox_tree_flag) {
	    fl_set_object_label(FKeyBarFormF5Text,      "Info");
	    fl_set_object_label(FKeyBarFormF6Text,      "Browser");
	} else {
	    fl_set_object_label(FKeyBarFormF5Text,      "QEdit");
	    fl_set_object_label(FKeyBarFormF6Text,      "Edit");
	}
	
	fl_show_object(FKeyBarFormF4Button);
	fl_show_object(FKeyBarFormF5Button);
	fl_show_object(FKeyBarFormF7Button);
	fl_show_object(FKeyBarFormF9Button);
	fl_show_object(FKeyBarFormF11Button);
	
	fl_show_object(FKeyBarFormF4Text);
	fl_show_object(FKeyBarFormF5Text);
	fl_show_object(FKeyBarFormF7Text);
	fl_show_object(FKeyBarFormF9Text);
	fl_show_object(FKeyBarFormF11Text);

	// possibly hidden via tree:
	fl_show_object(FKeyBarFormF6Button);
	fl_show_object(FKeyBarFormF6Text);
	fl_show_object(FKeyBarFormF8Button);
	fl_show_object(FKeyBarFormF8Text);
	fl_show_object(FKeyBarFormF10Button);
	fl_show_object(FKeyBarFormF10Text);
    }
    fl_redraw_form(FKeyBarForm);
    fl_unfreeze_form(FKeyBarForm);
}


void FKeyBarFormJoinButtonCB(FL_OBJECT *, long){
    toolbox_split();
    if (!toolbox_visible_flag || !ToolBoxForm->visible) {
	toolbox_show_form();
    }
}

static void fkeybar_wintitle_backend(const char *new_title) {
    static char *title = NULL;
    if (!new_title) return;
    if (title) {
	if (strcmp(title, new_title) == 0) return;
	fl_free(title);
    }
    title = strdup(new_title);
    assert(title);
    XStoreName(fl_display, FKeyBarForm->window, title);
}

void fkeybar_set_wintitle(void){
    if (!source_browser || !FKeyBarForm->visible) return;
    char title_buf[2*MAXPATHLEN];
    XSetIconName(fl_display, FKeyBarForm->window, "FKeys");
    if (toolbox_tree_flag) {
	sprintf(title_buf, "FKeys->Tree: %s", tree_get_path());
    } else {
	sprintf(title_buf, "FKeys->#%d: %s", source_browser->get_ID(),
	    source_browser->get_cwd()
	);
    }
    fkeybar_wintitle_backend(title_buf);
}
