
#include <iostream.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
//#include <sys/param.h>
#include <sys/errno.h>
//#include <sys/prctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "action_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "fork_pipe.h"
#include "status_forms_aux.h"
#include "status_forms.h"

//static int dbg_flag = TRUE;
static int dbg_flag = tool_debug_flag;


static void kill_fork_proc(pid_t cmd_pid){
    int wait_status;
    status_msg(1, "operation cancelled.");
    toolop_cancelled_flag = TRUE;
    if(dbg_flag) cerr << "SIGKILL killing cmd pipe (" << cmd_pid << ") group... ";
    kill(-1 * cmd_pid, SIGKILL);
    // not much of a point waiting here, actually ... 
#if defined(SUN) || defined(LINUX) || defined(GENERICBSD) || defined(UNIXWARE)
    waitpid(cmd_pid, &wait_status, WNOHANG);
#else
    wait(&wait_status);
#endif
    if(dbg_flag) cerr << "done.\n";
}

static void cleanup(int cmd_pipe[2]){
    // now close the stuff.
    close(cmd_pipe[0]);  
    close(cmd_pipe[1]);
}

static void display_pipebuf(char *pipebuf){
    if(!pipebuf) return;
    if(strchr(pipebuf, '\n')){
	timed_strerror_notifier("A child process has written data to stderr. Check status messages.");
	if(!status_visible_flag){
	    status_visible_flag = TRUE;
	    status_sync_defaults();
	} else {
	    if(StatusForm->visible){
		XMapRaised(fl_display, StatusForm->window);
	    }
	}
    } else {
	timed_strerror_notifier(pipebuf);
    }
    char *line = strtok(pipebuf, "\n");
    while(line){
	status_msg(0, line);
	line = strtok(NULL, "\n");
    }
}

int do_fork_cmd(
    const char *file,    // not full pathname, only file name. we're not SUID.
    const char **argv    // full ! argument vector., with argv[0] and trailing NULL.
){
    if(!file || !argv){
	program_unstable("UC internal: do_fork null ptrs.");
	return FALSE;
    }
    if(!strlen(file)){
	M_err("UC: do_fork_cmd","empty command");
	return FALSE;
    }
    toolop_cancelled_flag = FALSE;

    if(dbg_flag) cerr << "do_fork_cmd called.\n";
	
    pid_t cmd_pid = -1,	pid;
    
    int cmd_pipe[2],
        continue_flag = TRUE,
	wait_status
    ;
    off_t n_chars;

    if(pipe(cmd_pipe) < 0){
	timed_strerror_notifier("Can't open pipe:");
	return FALSE;
    }

    cmd_pid = fork();
    switch(cmd_pid){
    case -1:
	// error
	timed_strerror_notifier("Can't fork:");
	cleanup(cmd_pipe);
	return FALSE;
    case 0:
	// child
	// this is the section that executes the working command
	/*
	if(setpgid(0,0) < 0){
	    timed_strerror_notifier("Can't set new process group ID:");
	    cleanup(cmd_pipe);
	    return FALSE;
	}
	*/
#if defined(GENERICBSD)
	setpgrp(0,0);
#else
	setpgrp(); // should be equivalent to setpgid above.
#endif
	if(dbg_flag) cerr << "Child pgroup: " << getpgrp() << endl;
	// close our standard error:
	close(2);
	// make our standard error go TO the pipe:
	if(dup(cmd_pipe[1]) < 0){
	    timed_strerror_notifier("Can't duplicate fdescriptor:");
	    cleanup(cmd_pipe);
	    return FALSE;
	}
	close(cmd_pipe[0]);  // save fdescriptors
	close(cmd_pipe[1]);
	execvp(file, (char **) argv);
	perror("UC: return from execv: fatal");
	// no cleanup. no returning !
	_exit(-1);
    default:
	// parent
	// this is the controlling UC section
	if(dbg_flag){
	    cerr << "parent pgroup: " << getpgrp() << endl;
	    cerr << "parent cmd_pid: " << cmd_pid << endl;
	}

	// don't close anything !
	// cmd_pipe[0] is our read connection to read possible error data back.
	// first we spin, update the display and wait for completion of the child,
	// then we read all data.
	while(continue_flag){
	    sginap(10);
	    if(!action_mini_step()){
		kill_fork_proc(cmd_pid);
		cleanup(cmd_pipe);
		return FALSE;
	    }
	    pid = waitpid(-1 * cmd_pid, &wait_status, WNOHANG);
	    switch(pid){
	    case -1:
		if(errno != ECHILD){
		    timed_strerror_notifier("parent process control: couldn't waitpid.");
		    cleanup(cmd_pipe);
		    return FALSE;
		}   // else we were very fast :-)
		continue_flag = FALSE;
		break;
	    case 0:
		// children exist, but are not stopped and haven't exited either.
		if(dbg_flag) cerr << "Child is still running.\n";
		break;
	    default:
		if(WIFEXITED(wait_status)){
		    if(dbg_flag) cerr << "Child has exited.\n";
		    if(dbg_flag) status_msg(1, "status: %d", wait_status);
		    if(wait_status != 0){
			status_msg(0, "child exited with return value != 0 (status: %d).", WEXITSTATUS(wait_status));
		    }
		    continue_flag = FALSE;
		} else if(WIFSTOPPED(wait_status)){
		    if(dbg_flag) cerr << "Child is stopped.\n";
		    if(!ask_question(TRUE, "Child process group (ID: %d) was stopped. Continue operation ?", cmd_pid)){
			status_msg(0, "child was stopped.");
			while(ask_question(TRUE, "UC: It is necessary to kill the child. Go ahead now ?")){}
			continue_flag = FALSE;
			kill_fork_proc(cmd_pid);
		    }
		} else {
		    if(dbg_flag) cerr << "Child is neither stopped nor has exited.\n";
		}		    
		break;
		    
	    }
	}

	struct stat statbuf;

	if(fstat(cmd_pipe[0], &statbuf) < 0){
	    timed_strerror_notifier("Fstat cmd pipe failed:");
	    cleanup(cmd_pipe);
	    return FALSE;
	}
	n_chars = statbuf.st_size;
	if(n_chars > 0){
	    // seems we have an error
	    if(dbg_flag) cerr << "statbuf says pipe has data :-( " << endl;
	    char *pipebuf = new char[n_chars + 2];
	    if(!read(cmd_pipe[0], pipebuf, (unsigned) n_chars)){
		timed_strerror_notifier("can't read cmd pipe: ");
	    } else {
		pipebuf[n_chars]   = '\0';  // just in case.(no Null)
		pipebuf[n_chars-1] = '\0';  // just in case (no trailing newline, please)
		if(dbg_flag) cerr << pipebuf;
		display_pipebuf(pipebuf);
	    }
	    cleanup(cmd_pipe);
	    delete [] pipebuf;
	    return FALSE;
	}   // else allright, no err msges in pipe.
	cleanup(cmd_pipe);
	return TRUE;
    }
}
