
#include <stdlib.h>
#include <iostream.h>
#include <string.h>

#include "flspecial.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "menu.h"
#include "wm_forms_aux.h"

//static int dbg_flag = TRUE;
static int dbg_flag = tool_debug_flag;

static void
    make_to_normal_button(FL_OBJECT *button),
    make_to_ret_button   (FL_OBJECT *button)
;

int fl_default_curname = FL_DEFAULT_CURSOR;


void set_to_ret_button(FL_OBJECT *button){
    if(!button) return;
    if(button->objclass != FL_BUTTON/* || button->objclass != FL_LIGHTBUTTON*/) return;
    FL_OBJECT *obj = button->form->first;

    fl_freeze_form(button->form);
    while(obj){
	if(obj->objclass == FL_BUTTON/* || obj->objclass == FL_LIGHTBUTTON*/){
	    if(obj == button){
		make_to_ret_button(obj);
	    } else {
		if(obj->type == FL_RETURN_BUTTON){
		    make_to_normal_button(obj);
		}
	    }
	}
	obj = obj->next;
    }
    fl_unfreeze_form(button->form);
}    

void make_to_ret_button(FL_OBJECT *button){
    /* sometimes necessary to force the shortcut correct even when
     * the type field is already fixed. Do that externally.
     */
    if(button->type == FL_RETURN_BUTTON){
	// cerr << "already is a return button.\n";
	return;
    }
    if(button->shortcut == NULL){
	fl_set_button_shortcut(button, "^M", TRUE);
    } else {
        int count = 0;
        while(button->shortcut[count] != '\0') {
	    count++;
	}
	button->shortcut = (long *)fl_realloc(
            button->shortcut, 
            (count + 2) * sizeof(*(button->shortcut))
        );
        button->shortcut[count] = 'M' - 'A' + 1;  // ctrl-M.
        button->shortcut[count+1] = '\0';
    }
    button->type = FL_RETURN_BUTTON;
    fl_redraw_object(button);
}

void make_to_normal_button(FL_OBJECT *button){
    if(button->type == FL_NORMAL_BUTTON){
	// cerr << "already is a normal button.\n";
	return;
    }
    button->type = FL_NORMAL_BUTTON;
    if(button->shortcut == NULL){
	return;
    } else {
        int count = 0, where = -1;
        while(button->shortcut[count] != '\0') {
	    if (button->shortcut[count] == 'M' - 'A' + 1) {
	        where = count;
	    }
	    count++;
	}
	if (where >= 0 && count > 0){
	    button->shortcut[where] = button->shortcut[count-1];
	    button->shortcut[count-1] = '\0';
	    button->shortcut = (long *)fl_realloc(button->shortcut, count * sizeof(*(button->shortcut))
        );
	}
    }
    fl_redraw_object(button);
}

FL_OBJECT *find_return_button(FL_FORM *form){
    if(!form) return NULL;
    FL_OBJECT *obj = form->first;
    while(obj){
	if(obj->objclass == FL_BUTTON/* || obj->objclass == FL_LIGHTBUTTON*/){
	    if(obj->type == FL_RETURN_BUTTON){
		return obj;
	    }
	}
	obj = obj->next;
    }
    return NULL;
}

/* search for the next right button on the same y height. If none exist,
 * return the left most one.
 */

static FL_OBJECT *find_next_right_button(FL_OBJECT *current_button, int &wrap_flag){
    if(!current_button) return NULL;

    FL_OBJECT
		*obj        = current_button->form->first,
		*best_match = NULL,
		*wrap_obj   = NULL
    ;
    
    while(obj){
		if(obj != current_button && obj->visible){
		    if(obj->objclass == FL_BUTTON/* || obj->objclass == FL_LIGHTBUTTON*/){
				if(obj->y == current_button->y){
				    if(obj->x > current_button->x){
						if(!best_match){
						    best_match = obj;
						} else {
			    			if(obj->x < best_match->x){
								best_match = obj;
			    			}
						}
		    		} else {
						if(!wrap_obj){
						    wrap_obj = obj;
						} else {
			    			if(obj->x < wrap_obj->x){
								wrap_obj = obj;
			    			}
						}
				    }
				}
	    	}
		}
		obj = obj->next;	
    }
    if(best_match){
		// cerr << "right: best match\n";
		wrap_flag = FALSE;
		return best_match;
    } else {
		// cerr<< "right: wrap obj\n";
		// if(!wrap_obj) cerr<< "wrap == NULL\n";
		wrap_flag = TRUE;
		return wrap_obj;
    }
}

static FL_OBJECT *find_next_left_button(FL_OBJECT *current_button, int &wrap_flag){
	int tmp_flag = FALSE, rwrapped_flag = FALSE;
    if(!current_button) return NULL;
    
    FL_OBJECT *old_button_0, *old_button_1;
    old_button_0 = current_button;
    do {
		if (tmp_flag) {
			rwrapped_flag = TRUE;
		}
		old_button_1 = old_button_0;
		old_button_0 = find_next_right_button(old_button_0, tmp_flag);
    } while (old_button_0 && old_button_0 != current_button);
	if(old_button_1 == current_button){
		old_button_1 = NULL;
    }
	wrap_flag = !rwrapped_flag;
    return old_button_1;
}

static FL_OBJECT *find_next_higher_button(FL_OBJECT *current_button){
    if(!current_button) return NULL;

    FL_OBJECT
	*obj        = current_button->form->first,
	*best_match = NULL,
	*wrap_obj   = NULL
    ;
    
    while(obj){
	if(obj != current_button && obj->visible){
	    if(obj->objclass == FL_BUTTON/* || obj->objclass == FL_LIGHTBUTTON*/){
		if(obj->x == current_button->x){
		    if(obj->y > current_button->y){
			if(!best_match){
			    best_match = obj;
			} else {
			    if(obj->y < best_match->y){
				best_match = obj;
			    }
			}
		    } else {
			if(!wrap_obj){
			    wrap_obj = obj;
			} else {
			    if(obj->y < wrap_obj->y){
				wrap_obj = obj;
			    }
			}
		    }
		}
	    }
	}
	obj = obj->next;	
    }
    if(best_match){
	// cerr<< "higher: best match\n";
	return best_match;
    } else {
	// cerr<< "higher: wrap obj.\n";
	// if(!wrap_obj) cerr<< "wrap == NULL\n";
	return wrap_obj;
    }
}

static FL_OBJECT *find_next_lower_button(FL_OBJECT *current_button){
    if(!current_button) return NULL;
    
    FL_OBJECT *old_button_0, *old_button_1;
    old_button_0 = current_button;
    do {
	old_button_1 = old_button_0;
	old_button_0 = find_next_higher_button(old_button_0);
    } while (old_button_0 && old_button_0 != current_button);
    if(old_button_1 == current_button){
	old_button_1 = NULL;
    }
    return old_button_1;
}

static int find_input(FL_FORM *form){
    if(!form) return FALSE;
    FL_OBJECT *obj = form->first;

    while(obj){
	if(obj->objclass == FL_INPUT){
	    return TRUE;
	}
	obj = obj->next;
    }
    return FALSE;
}

int adjust_return_button(FL_FORM *form, XEvent *xev, int override_flag){
	int wrap_flag;
    if(dbg_flag) cerr << "adjust return button.\n";
    if(!form || !xev){
        //cerr << "NULL form or xev in adjust return button.\n";
		return FALSE;
    }

    FL_OBJECT
		*new_ret_button = NULL,
		*old_ret_button = find_return_button(form)
#if 0
, 
		*tmp_ret_button = NULL
#endif
    ;
    if(!old_ret_button) return FALSE;
    int has_input_flag = override_flag ? FALSE : find_input(form);
    if(dbg_flag) cerr << "has_input_flag: " << has_input_flag << ", override: " << override_flag << endl;
	
    //cerr << "xport: adjust_ret_b\n";
    XKeyEvent *xkev= (XKeyEvent *) xev;
    KeySym keysym;
    char keybuf[1];

    switch(xev->type){
    case KeyPress:
		if(dbg_flag) cerr << "found an X key press event.\n";
		XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);

		if(is_left(keysym, xkev)){
		    if(dbg_flag) cerr << "left in adjust return button.\n";
		    if(has_input_flag) return FALSE;
		    new_ret_button = find_next_left_button(old_ret_button, wrap_flag);
		} else if(is_right(keysym, xkev)){
		    if(dbg_flag) cerr << "right in adjust return button.\n";
	    	if(has_input_flag) return FALSE;
		    new_ret_button = find_next_right_button(old_ret_button, wrap_flag);
		} else if(is_up(keysym, xkev)){
	    	if(dbg_flag) cerr << "up in adjust return button.\n";
	   		new_ret_button = find_next_lower_button(old_ret_button);
		} else if(is_down(keysym, xkev)){
		    if(dbg_flag) cerr << "down in adjust return button.\n";
	    	new_ret_button = find_next_higher_button(old_ret_button);
		} else if(keysym == XK_Tab || keysym == XK_KP_Tab) {
	    	if(has_input_flag) return FALSE;
			if (shiftkey_down(xkev->state)) {
		    	if(dbg_flag) cerr << "shift tab in adjust return button.\n";
		    	new_ret_button = find_next_left_button(old_ret_button, wrap_flag);
			} else {
		    	if(dbg_flag) cerr << "tab in adjust return button.\n";
		    	new_ret_button = find_next_right_button(old_ret_button, wrap_flag);
			}
#if 0
		    	new_ret_button = find_next_left_button(old_ret_button, wrap_flag);
				//cerr << "left wrap: " << wrap_flag << endl;
				if(wrap_flag || !new_ret_button) {
					// Coords are upside down...
		    		tmp_ret_button = find_next_lower_button(old_ret_button);
					if (tmp_ret_button) {
	    				new_ret_button = find_next_left_button(tmp_ret_button, wrap_flag);
						if (!new_ret_button) {
							new_ret_button = tmp_ret_button;
						}
					}
				}
			} else {
				//cerr << "right wrap: " << wrap_flag << endl;
		    	new_ret_button = find_next_right_button(old_ret_button, wrap_flag);
				if(!new_ret_button){
	    			new_ret_button = find_next_higher_button(old_ret_button);
				} else if (wrap_flag){
	    			tmp_ret_button = find_next_higher_button(old_ret_button);
					if (tmp_ret_button) {
		    			new_ret_button = find_next_right_button(tmp_ret_button);
						if (!new_ret_button) {
							new_ret_button = tmp_ret_button;
						}
					}
				}	
			}
#endif
		} else {
		    if(dbg_flag) cerr << "ignoring keysym in adjust return button.\n";
	    	return FALSE;
		}
		break;
    default:
		if(dbg_flag) cerr << "unprocessed xevent in adjust return button.\n";
		return FALSE;
    }
    if(new_ret_button && new_ret_button != old_ret_button){
		fl_freeze_form(form);
		make_to_normal_button(old_ret_button);
		make_to_ret_button(new_ret_button);
		fl_unfreeze_form(form);
		tool_got_event_flag = TRUE;
		return TRUE;
    } else {
		return FALSE;
    }
    /*
    switch(dev){
    case KEYBD:
	if(val == 9){
	    // kill all tabs in such a situation. 
	    return TRUE;
	}
	val += 'a'-1;
	switch(val){
	case 'f':
	    new_ret_button = find_next_right_button (old_ret_button);
	    if(has_input_flag) return FALSE;
	    break;
	case 'b':
	    new_ret_button = find_next_left_button  (old_ret_button);
	    if(has_input_flag) return FALSE;
	    break;
	case 'p': new_ret_button = find_next_higher_button(old_ret_button); break;
	case 'n': new_ret_button = find_next_lower_button (old_ret_button); break;
	default:  return FALSE;
	}
	break;
    case TABKEY:
	if(val){ 
	    if(tool_get_shift()){
		// cerr<< "shift tab\n";
		new_ret_button = find_next_left_button(old_ret_button);
		if(!new_ret_button){
		    new_ret_button = find_next_higher_button(old_ret_button);
		}
	    } else {
		// cerr<< "normal tab\n";
		new_ret_button = find_next_right_button(old_ret_button);
		if(!new_ret_button){
		    new_ret_button = find_next_lower_button(old_ret_button);
		}
	    }
	}
	break;
    default:
	return FALSE;
    }
    
    */
}

void correct_slider_size(FL_OBJECT *ob, double &slsize){
    double org_slsize;
    /* HAK: min pixels for browser slider:  */
    if(slsize > 1.0){
	slsize = 1.0;
	return;
    }
    if(slsize * ob->h < 14.0){
	if(ob->h > 0.0){
	    org_slsize = slsize;
	    slsize = 14.0 / ob->h;
	    if(slsize > 1.0){
		slsize = org_slsize;
	    }
	}
    }
}

static void
local_move_window(FL_FORM * form)
{
    if (form->visible > 0){
	XMoveWindow(fl_display, form->window, form->x, form->y);
    }
}

void fl_set_form_position_direct(FL_FORM *form, FL_Coord x, FL_Coord y){
    int oldx, oldy;
    if (form == NULL){
	fl_error("fl_set_form_position_direct", "NULL form.");
	return;
    }

    oldx = form->x;
    oldy = form->y;
    form->x = x;
    form->y = y;
    //form->x = (x >= 0) ? x : (fl_scrw + x);
    //form->y = (y >= 0) ? y : (fl_scrh + y);
    if (oldx != form->x || oldy != form->y){
	local_move_window(form);
    }
}

long make_fsize_menu(long user_cb){
    long menu = new_pup();
    //if(!menu) return menu;
    char buffer[256];
    add_to_pup(menu, "Font Size%t", 0);
//    add_to_pup(menu, "6%x6", 0, "6");
    sprintf(buffer, "Tiny%%x%d", FL_TINY_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Tt");
    sprintf(buffer, "Small%%x%d", FL_SMALL_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Ss");
    sprintf(buffer, "Normal%%x%d", FL_NORMAL_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Nn");
    sprintf(buffer, "Medium%%x%d", FL_MEDIUM_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Mm");
    sprintf(buffer, "Large%%x%d", FL_LARGE_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Ll");
    sprintf(buffer, "Huge%%x%d", FL_HUGE_SIZE);
    //cerr << buffer << endl;
    add_to_pup(menu, buffer,   user_cb, "Hh");
    return menu;
}

long make_fstyle_menu(long user_cb){
    long menu = new_pup();
    add_to_pup(menu, "Style%t", 0);
    add_to_pup(menu, "Fixed-Normal%x4", (long) user_cb, "Nn");
    add_to_pup(menu, "Fixed-Bold%x5",   (long) user_cb, "Bb");
    add_to_pup(menu, "Prop-Normal%x8",  (long) user_cb, "rR");
    add_to_pup(menu, "Prop-Bold%x9",    (long) user_cb, "lL");
    return menu;
}

void set_window_icon(Window window, Pixmap tree_pixmap){
    XWMHints *hints = XAllocWMHints();
    if(hints){
	hints->flags = IconPixmapHint;
	hints->icon_pixmap = tree_pixmap;
	XSetWMHints(fl_display, window, hints);
	XFree((char *)hints);
    } else {
	M_err("set_window_icon:","couldn't alloc wmhints.");
    }
}

/* True if event of type event mask etc is in queue. Doesn't change the queue
 * at all. Necessary because put_back_event is out of sync;
 */

static int pred_retval = FALSE;
static long pred_event_type = ButtonRelease;
static Window pred_window = 0;

#ifdef SGI_IRIX4
int predicate(Display *, XEvent *xev, char *){
#else
Bool predicate(Display *, XEvent *xev, XPointer){
#endif
    if(xev->type == pred_event_type && xev->xany.window == pred_window){
	pred_retval = TRUE;
    }
    return False;
}

int CheckPeekWindowEvent(Display *display, Window window, long event_type){
    XEvent xevent;
    pred_retval = FALSE;
    pred_event_type = event_type;
    pred_window = window;
    XCheckIfEvent(display, &xevent, predicate, 0);
    return pred_retval;
}    

FL_FORM *fl_find_pointer_form(void){
    FL_Coord x, y;
    unsigned int keymask;
    return fl_find_pointer_form(&x, &y, &keymask);
}

FL_FORM *fl_find_pointer_form(FL_Coord *x, FL_Coord *y, unsigned int *keymask){
    FL_FORM *tmp_form  = NULL;
    Window wid         = fl_root;
    unsigned int level = 0;
    while (!tmp_form && wid != None && level<4) {
	wid = fl_get_win_mouse(wid, x, y, keymask);
	tmp_form = fl_win_to_form(wid);
	level++;
    }
    // necessary (try OW3.0 vs Linux fvwm!)
    *x += wm_forms_fudge_dx;
    *y += wm_forms_fudge_dy;
    return tmp_form;
}

void fl_set_form_label(FL_FORM *form, const char *label){
    if (!label || !form) return;
    if (form->label) {
	if (strcmp(label, form->label) == 0) return;
	fl_free(form->label);
    }
    form->label = strdup(label);
    if (form->visible) {
	XStoreName(fl_display, form->window, label);
    }
}

void
fl_activate_form_always(FL_FORM *form) {
    if (!form) return;
    while (form->deactivated) {
	fl_activate_form(form);
    }
}

void fl_set_global_cursor(int cursor){
#ifdef UNUSED_CODE
    register FL_FORM **fws, **fw;
    for (fw = forms, fws = fw + formnumb; fw < fws; fw++){
    fl_set_cursor((*fw)->window, cursor);
    }
#endif
}

void fl_reset_global_cursor(void){
#ifdef UNUSED_CODE
    register FL_FORM **fws, **fw;
    for (fw = forms, fws = fw + formnumb; fw < fws; fw++){
    fl_reset_cursor((*fw)->window);
    }
#endif
}

void fl_set_browser_leftslider (FL_OBJECT *, int){}
