
#include <unistd.h>

#include "forms_aux.h"
#include "edit_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "ghelp_forms_aux.h"
#include "ghelp_forms.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms.h"
#include "options_forms_aux.h"
#include "pathnames.h"
#include "scroll_events.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "xshell.h"

// browser form
static void ghelp_browser_close_form(void){
    if(GHelpBrowserForm->visible){
	focus_restore_start();
	fl_hide_form(GHelpBrowserForm);
	focus_restore_finish();
    }
}

static int ghelp_browser_form_raw_cb(FL_FORM *form, void *xev){
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, ghelp_browser_close_form);
}

void ghelp_browser_cb(FL_OBJECT *, long){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_browser_leftslider(GHelpBrowserFormInfoBrowser, TRUE);
	fl_set_browser_fontstyle(GHelpBrowserFormInfoBrowser, FL_FIXED_STYLE);
	fl_set_browser_fontsize (GHelpBrowserFormInfoBrowser, FL_NORMAL_FONT);
	fl_set_object_shortcut(GHelpBrowserFormCloseReturnButton, "Cc^M", TRUE);
	fl_set_button_shortcut(GHelpBrowserFormMoreButton,        "Mm", TRUE);
	fl_register_call_back(GHelpBrowserForm, FL_ALL_EVENT, ghelp_browser_form_raw_cb);
	load_text_in_browser(GHelpBrowserFormInfoBrowser, browser_help_text);
    }
    set_to_ret_button(GHelpBrowserFormCloseReturnButton);
    fl_set_browser_topline(GHelpBrowserFormInfoBrowser, 1);
    focus_save();
    if(GHelpBrowserForm->visible){
	XMapRaised(fl_display, GHelpBrowserForm->window);	
    } else {
	fl_show_form(GHelpBrowserForm, FL_PLACE_CENTER, TRUE, "BrowserHelp:");
    }
    tool_focus_move(GHelpBrowserForm, GHelpBrowserFormCloseReturnButton);
}

void GHelpBrowserFormCloseReturnButtonCB(FL_OBJECT *, long){
    ghelp_browser_close_form();
}

void GHelpBrowserFormMoreButtonCB(FL_OBJECT *, long){
    info_show_form();
    // load_help_browser(info_browser_handling, "Browser");
}

// colours form

static void ghelp_colours_close_form(void){
    if(GHelpColoursForm->visible){
	focus_restore_start();
	fl_hide_form(GHelpColoursForm);
	focus_restore_finish();
    }
}

static int ghelp_colours_form_raw_cb(FL_FORM *form, void *xev){
    return standard_form_event_cb(form, (XEvent *) xev, ghelp_colours_close_form);
}


void ghelp_colours_cb(FL_OBJECT *, long){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_object_shortcut(GHelpColoursFormCloseReturnButton, "Cc^M", TRUE);
	fl_set_browser_fontstyle(GHelpColoursFormInfoBrowser, FL_FIXED_STYLE);
	fl_set_browser_fontsize (GHelpColoursFormInfoBrowser, FL_NORMAL_FONT);
	fl_register_call_back(GHelpColoursForm, FL_ALL_EVENT, ghelp_colours_form_raw_cb);
	fl_set_browser_leftslider(GHelpColoursFormInfoBrowser, TRUE);
	load_text_in_browser(GHelpColoursFormInfoBrowser, colours_help_text);
    }
    focus_save();
    if(GHelpColoursForm->visible){
	XMapRaised(fl_display, GHelpBrowserForm->window);	
    } else {
	fl_show_form(GHelpColoursForm, FL_PLACE_CENTER, TRUE, "ColoursHelp:");
    }
    tool_focus_move(GHelpColoursForm, GHelpColoursFormCloseReturnButton);
}

void GHelpColoursFormCloseReturnButtonCB(FL_OBJECT *, long){
    ghelp_colours_close_form();
}

///////

void GlobalHelpToolBoxButtonCB(FL_OBJECT *, long){
    toolbox_help_func();
}

void GlobalHelpStatusButtonCB(FL_OBJECT *, long){
    status_context_func();
}

void GlobalHelpShellButtonCB(FL_OBJECT *, long){
    xshell_help_func();
}

void GlobalHelpTreeButtonCB(FL_OBJECT *, long){
    tree_form_help_func();
}

void GlobalHelpBrowserButtonCB(FL_OBJECT *, long){
    ghelp_browser_cb(NULL, 0);
}

void GlobalHelpAboutUCButtonCB(FL_OBJECT *, long){
    load_help_browser(other_acks, "About UC ...");  
}

void GlobalHelpIntroductionButtonCB(FL_OBJECT *, long){
    load_help_browser(new_user_info, "New User Introduction");  
}

void GlobalHelpOverviewButtonCB(FL_OBJECT *, long){
    load_help_browser(info_overview, "Overview");     
}

void GlobalHelpSelectionsButtonCB(FL_OBJECT *, long){
    load_help_browser(info_managing_selections, "Managing Selections"); 
}

void GlobalHelpGlobalOptionsButtonCB(FL_OBJECT *, long){
    options_help_func();
}

void GlobalHelpQuickEditButtonCB(FL_OBJECT *, long){
    edit_context_func();
}

void GlobalHelpEditButtonCB(FL_OBJECT *, long){
    load_help_browser(edit_help_text, "Edit");
}

void GlobalHelpRenameButtonCB(FL_OBJECT *, long){
    load_help_browser(rename_help_text, "Rename");
}

void GlobalHelpDuplicateButtonCB(FL_OBJECT *, long){
    load_help_browser(duplicate_help_text, "Duplicate");
}

void GlobalHelpDeleteButtonCB(FL_OBJECT *, long){
    load_help_browser(delete_help_text, "Delete");
}

void GlobalHelpMoveButtonCB(FL_OBJECT *, long){
    load_help_browser(move_help_text, "Move");
}

void GlobalHelpCopyButtonCB(FL_OBJECT *, long){
    load_help_browser(copy_help_text, "Copy");
}

void GlobalHelpMoveRenameButtonCB(FL_OBJECT *, long){
    load_help_browser(move_rename_help_text, "Move and Rename");
}

void GlobalHelpCopyRenameButtonCB(FL_OBJECT *, long){
    load_help_browser(copy_rename_help_text, "Copy and Rename");
}

void GlobalHelpFileSuffixRulesButtonCB(FL_OBJECT *, long){
    load_help_browser(suffix_rules_help_text, "File Suffix Rules");
}

void GlobalHelpMakeInterfaceButtonCB(FL_OBJECT *, long){
    load_help_browser(make_help_text, "Make Interface");
}

void GlobalHelpLicenseLegaleseButtonCB(FL_OBJECT *, long){
    load_help_browser(license_help_text, "Lame License Legalese");
}

/*
void GlobalHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(, "");
}*/

void global_help_hide_func(void){
    if(GlobalHelpForm->visible){
	focus_restore_start();
	fl_hide_form(GlobalHelpForm);
	focus_restore_finish();
    }
}

void GlobalHelpCloseReturnButtonCB(FL_OBJECT *, long){
    global_help_hide_func();    
}

static int global_help_form_raw_cb(FL_FORM *form, void *xev){
    if(standard_form_event_cb(form, (XEvent *) xev, global_help_hide_func)) return TRUE;
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    XEvent *xevent = (XEvent *) xev;
    KeySym keysym;
    char keybuf[1];
    
    switch(xevent->type){
    case KeyPress:
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	case XK_F5:   GlobalHelpQuickEditButtonCB(NULL, 0); return TRUE;
	case XK_F6:   GlobalHelpEditButtonCB(NULL, 0); return TRUE;
	case XK_F7:   GlobalHelpRenameButtonCB(NULL, 0); return TRUE;
	case XK_F8:
	    if(tool_get_meta()){
		GlobalHelpMoveRenameButtonCB(NULL, 0);
	    } else {
		GlobalHelpMoveButtonCB(NULL, 0);
	    }
	    return TRUE;
	case XK_F9:   GlobalHelpDuplicateButtonCB(NULL, 0); return TRUE;
	case XK_F10:
	    if(tool_get_meta()){
		GlobalHelpCopyRenameButtonCB(NULL, 0);
	    } else {
		GlobalHelpCopyButtonCB(NULL, 0);
	    }
	    return TRUE;
	case SunXK_F36: if(!sun_keyboard_flag) return FALSE;
#ifdef DEC_ALPHA
	case XK_F13:
#endif
	case XK_F11:  GlobalHelpDeleteButtonCB(NULL, 0); return TRUE;
	default:
	    break;
	}
    }
    return FALSE;
}

void global_help_show_form(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_button_shortcut(GlobalHelpAboutUCButton,         "Aa", TRUE);
	fl_set_button_shortcut(GlobalHelpToolBoxButton,         "Bb", TRUE);
	fl_set_object_shortcut(GlobalHelpCloseReturnButton,     "Cc^M", TRUE);
	fl_set_button_shortcut(GlobalHelpSelectionsButton,      "eE", TRUE);
	fl_set_button_shortcut(GlobalHelpIntroductionButton,    "Ii", TRUE);
	fl_set_button_shortcut(GlobalHelpLicenseLegaleseButton, "Ll", TRUE);
	fl_set_button_shortcut(GlobalHelpMakeInterfaceButton,   "Mm", TRUE);
	fl_set_button_shortcut(GlobalHelpIntroScreenButton,     "nN", TRUE);
	fl_set_button_shortcut(GlobalHelpGlobalOptionsButton,   "Oo", TRUE);
	fl_set_button_shortcut(GlobalHelpTutorialButton,        "rR", TRUE);
	fl_set_button_shortcut(GlobalHelpShellButton,           "Ss", TRUE);
	fl_set_button_shortcut(GlobalHelpTreeButton,            "Tt", TRUE);
	fl_set_button_shortcut(GlobalHelpStatusButton,          "uU", TRUE);
	fl_set_button_shortcut(GlobalHelpOverviewButton,        "vV", TRUE);
	fl_set_button_shortcut(GlobalHelpBrowserButton,         "wW", TRUE);
	fl_set_button_shortcut(GlobalHelpFileSuffixRulesButton, "xX", TRUE);
	fl_register_call_back(GlobalHelpForm, FL_ALL_EVENT, global_help_form_raw_cb);
    }
    set_to_ret_button(GlobalHelpCloseReturnButton);
    focus_save();
    if(!GlobalHelpForm->visible){
	char title[1024];
	sprintf(title, "Global help (UC %s version %s.%s Release %s):", version_os,  
		version_major, version_minor, version_date);
	fl_show_form(GlobalHelpForm, FL_PLACE_CENTER, TRUE, title);
    } else {
	XMapRaised(fl_display, GlobalHelpForm->window);	
    }
    if(GlobalHelpForm->deactivated){      // we are extra careful. 
	fl_activate_form(GlobalHelpForm);
    }
    tool_focus_move(GlobalHelpForm, GlobalHelpCloseReturnButton);
}    

void GlobalHelpTutorialButtonCB(FL_OBJECT *, long){
  do_tutorial();
}

void do_tutorial(void){
    static int first = TRUE;
    static char *name_buf = NULL;
    if(first){
	first = FALSE;
	ofstream ofile;
	name_buf = tempnam(NULL, "tutor");
	if(!name_buf){
	    cerr << "couldn't generate temporary name.\n";
	    return;
	}
	ofile.open(name_buf);
	int i=0;
	while(tutorial_help_text[i] != 0){
	    ofile << tutorial_help_text[i] << endl;
	    i++;
	}
	ofile.close();
    }
    if (!name_buf) return;
    quick_edit_file(name_buf);
}

