
#include <iostream.h>
#include <stdlib.h>

#include "flspecial.h"
#include "global.h"
#include "gl_utility.h"
#include "linklist.h"

#include "stack.h"
#include "fix_keysyms.h"


#ifdef UNUSED_CODE
static void ungrab_focus(void);
static void grab_focus(Window window);
#endif

// saves the move-to coordinates.
// auto-disables restore if current-coords. are not the
// same as the move-to coordinates. 


static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

class ptr_info_c {
public:
    FL_Coord x,y, move2x, move2y;
    Window window, move2window;
	/*
    static int ID;
    int my_ID;
    ptr_info_c(void){my_ID = ID++; cerr << "making " << my_ID << endl;}
    ~ptr_info_c(){cerr << "destroying " << my_ID << endl;}
    ptr_info_c& operator=(ptr_info_c& info){
	x = info.x;
	y = info.y;
	move2x = info.move2x;
	move2y = info.move2y;
	window = info.window;
	move2window = info.move2window;
	return info;
	*/
    void print(void){
	cerr << "(" << x << "," << y << "," << window << ") -> (";
	cerr << move2x << "," << move2y << "," << move2window << ");\n";
    }
};

// int ptr_info_c::ID = 0;
    
static stack_c info_stack(5);

int auto_mouse_flag = TRUE;

void focus_clear(void){
    while(info_stack.size()>0){
	delete (ptr_info_c *) info_stack.pop();
    }
}

void focus_save(void){
    // (storing the same position twice is no error)

    ptr_info_c info;
    focus_get(info.x, info.y, info.window);
    
    // use focus_move to set these values correctly.
    info.move2x = 0;
    info.move2y = 0;
    info.move2window = 0;
    if(dbg_flag){
	//cerr << "focus_save: ";
	//info.print();
    }
    
    ptr_info_c *tmp_info = new ptr_info_c;                      // new
    *tmp_info = info;                                           // store
    tmp_info = (ptr_info_c *) info_stack.force_push(tmp_info);  // push
    if(tmp_info){                                               // if overflow, freemem
	//cerr << "stack overflow, force push.\n";
	delete tmp_info;
    }
}

static int bDontWait = FALSE;

void focus_restore(int override_focus_flag){
    bDontWait = TRUE;
    if (dbg_flag) cerr << "classical restore.\n";
    focus_restore_start (override_focus_flag);
    focus_restore_finish();
    bDontWait = FALSE;
}

/* always use these pairwise. 
 * start sets an internal flag indicating wether the focus restore 
 * is o.k., while the original window is still visible.
 * Then, typically a hide takes place.
 * restore_finish WAITS! until the original window ID is no longer
 * the focus window, i.e. it was actively hidden. Then it restores
 * the focus and thus gets around window manager/X-Server bugs.
 *
 * Don't use this function if you don't hide. It is clever enough
 * not to wait if the from and to windows are the same, though. 
 */

// if this is Null, no restore.
// start gets it, finish deletes it.
static ptr_info_c *pInfo = NULL;

void focus_restore_start (int override_focus_flag) {
    pInfo = NULL;
    static int fudge_diff = 3;
    if(info_stack.size() == 0){
	if(dbg_flag) cerr << "Discarding restore, stack empty.\n";
	return;
    }
    pInfo = (ptr_info_c *) info_stack.pop();
    if (!pInfo) {
	cerr << "Bad pointer stack in restore_start!\n";
	return;
    }
    FL_Coord x,y;
    Window window;
    focus_get(x,y,window);
    if(abs(x - pInfo->move2x) <= fudge_diff &&
	abs(y - pInfo->move2y) <= fudge_diff &&
	(window == pInfo->move2window || override_focus_flag)
    ){
	// no changes have happened, ok to restore.
	if(dbg_flag) cerr << "focus_restore: ok (f,t)=" << pInfo->window << "," << window << "\n";
    } else {
	if(dbg_flag) cerr << "Discarding ptr restore, win mismatch or mouse has moved too much.\n";
	if(dbg_flag) cerr << "from, to, actual)=" << pInfo->window << "," << pInfo->move2window << "," << window << "\n";
	// this is a hack, e.g. when closing an ask_question with the mouse in a click-to-focus
	// wm. But: The focus could be in a completely different (non-Form) window. So, add a window check.
	if(window == pInfo->move2window){
	    int flag = auto_mouse_flag;
	    auto_mouse_flag = FALSE;
	    focus_set(pInfo->x, pInfo->y, pInfo->window);
	    auto_mouse_flag = flag;
	}
	// signal the failure:
	delete pInfo;
	pInfo = NULL;

	// discard stack:
	focus_clear();
    }
}

void focus_restore_finish(void){
    FL_Coord x,y;
    Window window;
    if (!pInfo) {
	if (dbg_flag) cerr << "restore finish: discard.\n";
	return;
    }

    if (dbg_flag) cerr << "restore finish:";

    // strategy: hide other, wait for wm to fuck up the focus,
    // raise target, move ptr there, set focus.
    // we assume the window is going to be hidden (or bDontWait
    // should be true).
    
    if (!bDontWait && pInfo->window != pInfo->move2window) {
	if (dbg_flag) cerr << " waiting...";
	// wait until input focus is away from move2window:
	// half-decimal seconds.
	int hdecs_secs=0; 
	do {
	    hdecs_secs++;
	    focus_get(x,y,window);
	    if (window == None || window == pInfo->move2window) {
		sginap(5);
		if (dbg_flag) cerr << "napping...\n";
	    } else {
		if (dbg_flag) cerr << "windows are not the same: " << window << "\n";
	    }

	} while (hdecs_secs<10 && window == pInfo->move2window);
    }
    //the raising isn't allways correct (source_browser != focus window)
    //XMapRaised(fl_display, pInfo->window);
    focus_set(pInfo->x, pInfo->y, pInfo->window);

    delete pInfo;
    pInfo = NULL;
    if (dbg_flag) cerr << " done.\n";
}

void focus_move(void){
    FL_Coord x, y;
    Window window;
    focus_get(x,y,window);
    focus_move(x,y,window);
}

void focus_move(FL_Coord x, FL_Coord y, Window window){
//    if(!auto_mouse_flag) return;
    //if(dbg_flag) cerr << "\nmouse move.\n";
    if(x>=0 && x<fl_scrw && y>=0 && y<fl_scrh){

	if(info_stack.size() > 0) {
	    ptr_info_c *info = (ptr_info_c*) info_stack.get();
	    if(info) {
		info->move2x      = x;
		info->move2y      = y;
		info->move2window = window;
		if (dbg_flag) {
		    cerr << "focus_move: ";
		    info->print();
		}
	    }
	}
	focus_set(x,y,window);
    }
}

void focus_get(FL_Coord &x, FL_Coord &y, Window &window){
    unsigned int key;
    fl_get_mouse(&x, &y, &key);
    int dummy;
    XGetInputFocus(fl_display, &window, &dummy);
}

#ifdef UNUSED_CODE
static void focus_wait(FL_Coord x, FL_Coord y, Window window){
    Window wFocus;
    FL_Coord xn, yn;
    // after we set focus to window, query the XServer for a second or 
    // so to verify the focus. 
    for (int hdecs_ecs=0; hdecs_ecs<10; hdecs_ecs++) {
	focus_get(xn, yn, wFocus);
	if (wFocus == window) {
	    if (dbg_flag) cerr << "Focus correct.\n";
	    if(auto_mouse_flag){
		if (xn == x && yn == y) {
		    if (dbg_flag) cerr << "Mouse correct.\n";
		    return;
		} else {
		    cerr << "Mouse incorrect: (" << x << "," << y 
			 << ") != (" << xn << "," << yn << ")\n";
		    XWarpPointer(fl_display, None, fl_root, 0,0,0,0,x,y);
		}
	    } else {
		return;
	    }
	}
	sginap(5);
	if (dbg_flag) cerr << "waiting " << hdecs_ecs << endl;
	//cerr << "waiting " << hdecs_ecs << endl;
    }
    cerr << "UC: Failed to set focus quickly. Do you have a slow X-Server?\n";
}
#endif

void focus_set(FL_Coord x, FL_Coord y, Window window){
    int bWait = FALSE;
    if(x>=0 && x<fl_scrw && y>=0 && y<fl_scrh){
	char *pName = NULL;
	if(dbg_flag) {
	    cerr << "setting focus to " << window << "(" << x << "," << y << ")";
	}
	if (XFetchName(fl_display, window, &pName)) {
	    if (dbg_flag) cerr << "[" << pName << "]\n";
	    XFree(pName);
	} else {
	    if (dbg_flag) cerr << "[unknown]\n";
	}
        /*
	simulate_click(x, y, window);
	return;
	*/
	if(window != None){
	    // want no intermediary window being raised. (has problems with click2focus WM's.
            // XSetInputFocus(fl_display, window, RevertToPointerRoot, CurrentTime);
	    // how about RevertToParent? don't know. 

	    // was sort of good in UC1.5:
	    if (auto_mouse_flag) {
		XWarpPointer(fl_display, None, fl_root, 0,0,0,0,x,y);
	    }
	    XSetInputFocus(fl_display, window, RevertToNone, CurrentTime);
	    bWait = TRUE;
	} else {
	    if (dbg_flag) cerr << "set mouse only, window == none.\n";
	}
	if (bWait){
	    // good for debugging, bad for running. 
	    // focus_wait(x, y, window);
	}
    } else {
	cerr << "invalid x,y (" << x << "," << y << ") in focus_set.\n";
    }
}

int is_delete_window(Window window, XClientMessageEvent *xcm){
    Atom atom_type, atom_data;
    //cerr << "is delete win.\n";
    atom_type = XInternAtom(xcm->display, "WM_PROTOCOLS", 0);
    atom_data = XInternAtom(xcm->display, "WM_DELETE_WINDOW", 0);
    return (xcm->message_type == atom_type &&
	    xcm->data.l[0] == (long) atom_data &&
	    xcm->window == window);
}

int is_iconify_window(Window window, XClientMessageEvent *xcm){
    Atom atom_type;
    cerr << "func::is iconify win.\n";
    atom_type = XInternAtom(xcm->display, "WM_CHANGE_STATE", 0);
    return (xcm->message_type == atom_type &&
	    xcm->data.l[0] == IconicState &&
	    xcm->window == window);
}

int keypad_fucked_flag = FALSE;

int good_state(XKeyEvent *xkey){
    if(!xkey) return FALSE;
    if(keypad_fucked_flag) return TRUE;
    if(shiftkey_down(xkey->state)) return FALSE;
    if(mod_down(xkey->state)) return FALSE;
    return TRUE;
}

int is_delete(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
#ifdef XK_KP_Separator
    case XK_KP_Separator:
	return good_state(xkey);
#endif
#ifdef XK_KP_Delete
    case XK_KP_Delete:
#endif
    case XK_Delete:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_end(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_1:
	return good_state(xkey);
    case XK_F33: if(!sun_keyboard_flag) return FALSE;
#if !defined(SUNOS)	// else duplicate labels.
    case XK_KP_End:
#endif
    case XK_End:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_down(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_2:
	if(dbg_flag) cerr << "KP2.\n";
	return good_state(xkey);
#if !defined(SUNOS)
    case XK_KP_Down:
	if(dbg_flag) cerr << "down arrow keypad keysym .\n";
#endif
    case XK_Down:
	if(dbg_flag) cerr << "down arrow keysym .\n";
	return TRUE;
    default:
	return FALSE;
    }
}
    
int is_page_down(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_3:
	return good_state(xkey);
    case XK_F35: if(!sun_keyboard_flag) return FALSE;

// On SGI 6.2, KP_Next equals KP_Page_Down.
#ifdef XK_KP_Next
#   if !defined(SGI) && !defined(DEC_ALPHA)
    case XK_KP_Next:
#   endif 
#endif

#if defined(SUN) || defined(SGI_IRIX4) || defined(HP) || defined(LINUX) || defined(AIX) || defined(GENERICBSD) || defined(UNIXWARE) || defined(SCO) || defined(SINIX)
    case XK_Next:
#else
    case XK_KP_Page_Down:
    case XK_Page_Down:
#endif
	return TRUE;
    default:
	return FALSE;
    }
}

int is_left(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_4:
	return good_state(xkey);
#if !defined(SUNOS)
    case XK_KP_Left:
#endif
    case XK_Left:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_begin    (KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_5:
	return good_state(xkey);
    case XK_F31: if(!sun_keyboard_flag) return FALSE;
#if !defined(SUNOS)
    case XK_KP_Begin:
#endif    
    case XK_Begin:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_right(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_6:
	return good_state(xkey);
#if !defined(SUNOS)
    case XK_KP_Right:
#endif
    case XK_Right:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_home(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_7:
	return good_state(xkey);
    case XK_F27: if(!sun_keyboard_flag) return FALSE;
#if !defined(SUNOS)
    case XK_KP_Home:
#endif
    case XK_Home:
	return TRUE;
    default:
	return FALSE;
    }
}

int is_up(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_8:
	if(dbg_flag) cerr << "KP8 .\n";
	return good_state(xkey);
#if !defined(SUNOS)
    case XK_KP_Up:
	if(dbg_flag) cerr << "up arrow keypad keysym .\n";
#endif
    case XK_Up:
	if(dbg_flag) cerr << "up arrow keysym .\n";
	return TRUE;
    default:
	return FALSE;
    }
}

int is_page_up(KeySym keysym, XKeyEvent *xkey){
    switch(keysym){
    case XK_KP_9:
	if(dbg_flag) cerr << "KP9 .\n";
	return good_state(xkey);
    case XK_F29: if(!sun_keyboard_flag) return FALSE;
// On SGI 6.2, KP_Prior equals KP_Page_Up.
#ifdef XK_KP_Prior
#   if !defined(SGI) && !defined(DEC_ALPHA)
    case XK_KP_Prior:
#   endif 
#endif

#if defined(SUN) || defined(SGI_IRIX4) || defined(HP) || defined(LINUX) || defined(AIX) || defined(GENERICBSD) || defined(UNIXWARE) || defined(SCO) || defined(SINIX)
    case XK_Prior:
	if(dbg_flag) cerr << "xk prior keysym .\n";
#else
    case XK_KP_Page_Up:
    case XK_Page_Up:
#endif
	return TRUE;
    default:
	return FALSE;
    }
}

static linklist_c xerr_list;

//typedef int (*xerr_handler_pt)(Display*, XErrorEvent*);

static char *xerr_msg = NULL;
static int xerr_exit_flag = FALSE, xerr_errval = FALSE;

static int xerr_handler(Display *dpy, XErrorEvent *err){
    xerr_errval = TRUE;
    if(!xerr_msg) xerr_msg = "no data.";
    M_err("UC: X error:", "%s\n", xerr_msg);
    if(dpy != fl_display){
	M_err("UC: xerror", "displays out of sync");
	return 0;
    }
    if(!err){
	M_err("UC: xerror", "Null error event.\n");
	return 0;
    }
    char error_buf[1024];
    XGetErrorText(dpy, err->error_code, error_buf, sizeof(error_buf));
    M_err("UC: X error", "(%s)\n  serial %d\n  major opcode %d\n  minor opcode %d\n  resource %d\n",
	error_buf,
	err->serial,
	(int) err->request_code,
	(int) err->minor_code,
	err->resourceid
    );
    if(xerr_exit_flag){
	exit(-1);
    }
    return 0;
}

class xerr_node_c {
public:
    char *msg;
    int exit_flag, errval;
    XErrorHandler handler;
    xerr_node_c(const char *msg_iv, int exit_flag_iv, int errval_iv, XErrorHandler handler_iv);
    ~xerr_node_c();
    xerr_node_c& operator=(const xerr_node_c& xerr_node);
};

xerr_node_c::xerr_node_c(const char *msg_iv, int exit_flag_iv, int errval_iv, XErrorHandler handler_iv){
    if(!handler_iv){
	cerr << "Warning: Null error handler in constructor.\n";
	handler_iv = xerr_handler;
    }
    if(!msg_iv) msg_iv = "local mem error.\n";
    msg = strdup(msg_iv);
    exit_flag = exit_flag_iv;
    handler = handler_iv;
    errval = errval_iv;
}

xerr_node_c::~xerr_node_c(){
    if(msg){
	fl_free(msg);
	msg = NULL;
    }
}

void xerrstack_push(const char *msg, int exit_flag){
    if(!msg) return;
    XErrorHandler tmp_handler;
    tmp_handler = XSetErrorHandler(xerr_handler);
    if(tmp_handler){
	// store old
	xerr_node_c *tmp_node = new xerr_node_c(xerr_msg, xerr_exit_flag, xerr_errval, tmp_handler);
	xerr_list.add((void *)tmp_node);
	if(xerr_msg){
	    delete xerr_msg;
	    xerr_msg = NULL;
	}
    }
    xerr_errval = FALSE;
    xerr_msg = strdup(msg);
    xerr_exit_flag = exit_flag;
}

int xerrstack_pop (void){
    if(xerr_list.getsize() > 0){
	xerr_list.tailend();
	xerr_node_c *tmp_node = (xerr_node_c *) xerr_list.getval();
	if(tmp_node){
	    xerr_errval = tmp_node->errval;
	    xerr_exit_flag = tmp_node->exit_flag;
	    if(xerr_msg){
		fl_free(xerr_msg);
		xerr_msg = NULL;
	    }
	    xerr_msg = strdup(tmp_node->msg);
	    xerr_list.remove();
	    delete tmp_node;
	}
    }
    return xerr_errval;
}

#ifdef UNUSED_CODE
static void grab_focus(Window window) {
    fl_set_cursor_color(XC_sb_right_arrow, FL_RED, FL_WHITE);
    if(XGrabPointer(
	fl_display,
	window,
	False,
	ButtonPressMask | ButtonReleaseMask | LeaveWindowMask | EnterWindowMask | PointerMotionMask,
	GrabModeSync,
	GrabModeSync,
	None,
	None,	// hour glass
	CurrentTime
    ) != GrabSuccess){
	cerr << "couldn't grab pointer. Uggh.\n";
    }
} 

static void ungrab_focus(void) {
    XUngrabPointer(fl_display, CurrentTime);
}
#endif

void simulate_click(int x, int y, Window wTarget) {
    if (!auto_mouse_flag) return;

    XWarpPointer(fl_display, None, fl_root, 0,0,0,0,x,y);

    XButtonEvent xbev;
    memset(&xbev, 0, sizeof(XButtonEvent));   // just in case.

    xbev.type        = ButtonPress;	
    xbev.display     = fl_display;
    xbev.window      = wTarget;
    xbev.root        = RootWindow(fl_display, fl_screen);
    xbev.button      = Button4;    
    xbev.state       = 0;
    xbev.time        = CurrentTime;
    xbev.same_screen = True;
    xbev.subwindow   = None;
    xbev.x = x;	// WRONG
    xbev.y = y;
    xbev.x = 10;	// WRONG
    xbev.y = 10;
    xbev.x_root = x;
    xbev.y_root = y;

    if (XSendEvent(fl_display,wTarget,True,ButtonPressMask,(XEvent *) &xbev)
== 0){
	cerr << "XSendEvent failed.\n";
    }

    xbev.type = ButtonRelease;
    if (XSendEvent(fl_display,wTarget,True,ButtonReleaseMask,(XEvent *) &xbev) == 0) {
	cerr << "XSendEvent failed.\n";
    }
    cerr << "send event ok.\n";
}
