/* gl_utility.h:
 */

#ifndef GL_UTILITY_H
#define GL_UTILITY_H

#include "global.h" 

/* handy routines that manipulate the mouse position, and 
 * therefore the input focus. Mouse center saves the current
 * position before centering, so that it can be restored.
 * you can also explicitly save the mouse position.
 * These functions can be dis/enabled via the auto_mouse_flag.
 */
extern int auto_mouse_flag;

void focus_clear  (void);                                       // clear stack
void focus_save   (void);					// save x,y,win
void focus_set    (FL_Coord fx, FL_Coord fy, Window fwindow);	// move ptr, set new focus, no remember
void focus_move   (FL_Coord fx, FL_Coord fy, Window fwindow);	// move ptr, set new focus, remember
void focus_move   (void);					// stack op, get and set move2 values.
void focus_restore(int override_focus_flag = FALSE);	   	// restore current stack pos/focus

/* always use these pairwise. 
 * start sets an internal flag indicating wether the focus restore 
 * is o.k., while the original window is still visible.
 * Then, typically a hide takes place.
 * restore_finish WAITS! until the original window ID is no longer
 * the focus window, i.e. it was actively hidden. Then it restores
 * the focus and thus gets around window manager/X-Server bugs
 *
 * Don't use this function if you don't hide. It is clever enough
 * not to wait if the from and to windows are the same, though. But
 * for a focus_restore where the original window stays open, use
 * the old restore function.
 */
void focus_restore_start (int override_focus_flag = FALSE);	
void focus_restore_finish(void);

/* query function */
void focus_get(FL_Coord &x, FL_Coord &y, Window &window);       // get focus window (can be different)

/* X convenience functions */
int is_delete_window (Window window, XClientMessageEvent *xcm);
int is_iconify_window(Window window, XClientMessageEvent *xcm);

int is_delete   (KeySym keysym, XKeyEvent *xkey);
int is_end      (KeySym keysym, XKeyEvent *xkey);
int is_down     (KeySym keysym, XKeyEvent *xkey);
int is_page_down(KeySym keysym, XKeyEvent *xkey);
int is_left     (KeySym keysym, XKeyEvent *xkey);
int is_begin    (KeySym keysym, XKeyEvent *xkey);
int is_right    (KeySym keysym, XKeyEvent *xkey);
int is_home     (KeySym keysym, XKeyEvent *xkey);
int is_up       (KeySym keysym, XKeyEvent *xkey);
int is_page_up  (KeySym keysym, XKeyEvent *xkey);

#define mod_down(mask)    ((mask) & Mod1Mask || \
			   (mask) & Mod2Mask || \
			   (mask) & Mod3Mask || \
			   (mask) & Mod4Mask || \
			   (mask) & Mod5Mask)

void xerrstack_push(const char *msg, int exit_flag = FALSE);
int xerrstack_pop (void);

void simulate_click(int x, int y, Window wTarget);

#endif // GL_UTILITY_H
