#ifndef GLOBAL_H
#define GLOBAL_H

#include "system.h"
#include "forms.h"
#include <assert.h>
#include <sys/types.h>

/* necessary for sunkeyboard hack */
#define SunXK_F36               0x1005FF10      /* Labeled F11 */
#define SunXK_F37               0x1005FF11      /* Labeled F12 */

/* type definitions.
 * - to remember whether to save config on exit. 
 * - ls options
 */

typedef enum {always_e, ask_e, dont_e} save_on_exit_state_et;
typedef enum {
    access_e, filetype_e, group_e, inode_e, modific_e, name_e,
    nlinks_e, owner_e, perms_e, size_e, typechar_e, extension_e
} sorttype_et;

// long long if we have it (handle with care) 
#if defined(DEC_ALPHA) || defined(LINUX)
typedef long long llong_t;
#else
typedef long llong_t;
#endif

#ifdef UNIXWARE
#define S_ISLNK(m)      (((m) & S_IFMT) == S_IFLNK)
#endif

// globally visible variables.
// start with a forward class declaration of the all-important browser_c.
extern class browser_c *source_browser, *target_browser;   

extern long double_click_msecs_max;

extern int 
    tool_bw_flag,
    tool_ctrl_on_flag,
    tool_cp_minusp_flag,
    tool_debug_flag,
    tool_du_auto_flag,
    tool_du_sort_flag,
    tool_editor_info_shown_flag,
    // wether we're in escape mode (escape keypress received,
    // and no keyrelease yet. Verify with keysympressed.
    tool_escape_flag,	
    tool_focus_on_startup_flag,
    tool_got_event_flag,
    tool_group_windows_flag,
    tool_looser_flag,
    tool_startup_msg_flag,
    tool_cmsg,
    tool_number_of_browsers,
    tool_dir_timeout_secs,  
    tool_file_timeout_secs,
    tool_fontsize,
    tool_fontstyle,
    tool_menu_fontsize,
    tool_menu_fontstyle,
    tool_max_ww,
    tool_max_wh,
    tool_source_colour,
    tool_target_colour,

    sun_keyboard_flag
;

extern mode_t mkdir_mode_mask;

extern save_on_exit_state_et 
    save_config_on_exit_state, 
    save_tree_on_exit_state
;

extern const char *tool_dnd_op_str;

extern char 
    default_editor[],
    *tool_gdefault_dir,   // global defaults. 
    *tool_ldefault_dir,   // local defaults.
    *tool_host_string,    // for wintitles
    *tool_default_fname,
    *version_date, 
    *version_major, 
    *version_minor, 
    *version_os
;

/* set to false at the beginning of each toolop operation, 
 * set to true if the operation was cancelled. Allows for 
 * the necessary updates when an atomic operation was cancelled,
 * but still could have left effects.
 */
extern int toolop_cancelled_flag;

/* tool-level global actions. */
/* if any global actions need to be done first, eg. answering a
 * question or so. This belongs into every callback in the whole
 * tool... Only continue locally if this function returns TRUE.
 */
int tool_default_actions(void);

int tool_change_dir(
    const char *fname, 
    int set_current_flag,
    int update_browser_bufs_flag,
    int update_tree_flag,
    int update_xshell_flag,
    int attempt_root_flag
);

int tool_change_target_dir(
    const char *fname, 
    int update_browser_bufs_flag,
    int update_tree_flag
);

/* default database operations (returns success) */
int  tool_defaults_save(const char *fname);
int  tool_defaults_load(const char *fname, int do_sync_flag = TRUE);
void tool_sync_defaults(void);
void tool_defaults_backup(void);

// check wether ctrl is pressed, or the ctrl flag is on.
// also check for shifts and alt.
// meta function checks for any of shift, ctrl, alt.
// some X-Servers translate shift-F1 into F11, shift-F2 into F12 etc.
// this makes using shift dangerous. 
int tool_get_alt  (void);
int tool_get_ctrl (XEvent *xev = NULL);
int tool_get_shift(void);
int tool_get_meta (XEvent *xev = NULL);
// get_escape looks for the global escape flag, and does a 
// keysym_pressed if the flag is true. This should be efficient
// and reliable.
int tool_get_esc(void);

// disable/enable ctrl or meta locking (via flag):
void tool_disable_ctrl(void);
void tool_enable_ctrl (void);
void tool_set_ctrl (void); 

void tool_disable_meta(void);
void tool_enable_meta (void);

/* update all browsers whose directory is the same as the dir_buf.
 * The keep-flag is passed on to browser_c::update().
 * 
 * Also have a version that updates all directories. 
 */
void tool_update_matching_dirs(
    const char *dir_buf, 
    int keep_selection_current_flag = FALSE
);

void tool_update_all_dirs(int keep_selection_current_flag = FALSE);

void tool_do_colours(void);

/* check/set a new source browser.
 * return wether a new source browser was set.
 * note that two versions are supported, one that 
 * directly addresses a browser_c pointer, one that 
 * indirectly uses the fl_browser object contained
 * by a browser_c object.
 */

int tool_set_source(FL_FORM   *a_form,    int move_focus_flag = FALSE);
int tool_set_source(browser_c *a_browser, int move_focus_flag = FALSE);
int tool_set_source(int new_id,           int move_focus_flag = FALSE);
int tool_set_target(browser_c *a_browser);
browser_c* tool_find_browser(FL_FORM *a_form);
browser_c* tool_find_browser_id(int id);
int is_browser_window(Window window);

/* translate an event into an object. */
//void event2obj(FL_OBJECT * &retobj, Device device, FL_OBJECT *target_obj);

// bind escape, (del and padperiod) to object.
void make_standard_shortcut(FL_OBJECT *);

// bind escape, cCnN to object. If return button, also Enter.
void make_standard_cancel_shortcut(FL_OBJECT *obj);

// bind return, enter, gGyY to object. 
void make_standard_ok_shortcut(FL_OBJECT *obj);


// here is the standard callback, so it can be called by other event cb which
// closes on escape and cancel device
// also supports F1 help, Ctrl-F1 context sensitive help, and F12/Ctrl-F12 exit.
// basically need its functionality plus a little bit (eg, browser processing).
typedef void (*void_func_pt)(void);
int standard_form_event_cb(
    FL_FORM *form, 
    XEvent *xev, 
    void_func_pt close_func = NULL, 
    void_func_pt help_func = NULL
);

int standard_options_event_cb(
    FL_FORM *form,
    XEvent *xev,
    void_func_pt options_func = NULL,
    int allow_rmouse_flag     = FALSE			      
);

void add_free_obj(FL_FORM* form, FL_HANDLEPTR func, int wantall = FALSE);

// extra F1/Ctrl-F1 checker for customized event functions. (limited functionality).
void help_event_cb(void_func_pt help_func);

/* perform restricted intelligent check_forms, with gobble-up and nap.
 * returns FALSE if cancel (padperiod) was pressed.
 * also, have a version that only eats events.
 * the callback version is `handle with care'. You need to disable all event
 * callbacks that you want off, and enable those that you want on !
 */
void tool_mung_callbacks(void);
void tool_mung_events(void);
int  tool_mung_forms (int do_nap_flag = TRUE);

/* issue a 'program unstable message', ask for program exit. */
void program_unstable(char *msg = NULL);
void malloc_unstable(char *msg = NULL);
/*
 * have a standard test for the source browser, returns TRUE
 * if the source browser pointer is valid. Otherwise, barfs
 * and returns false.
 */
int source_is_stable(void);
int source_target_are_stable(void);

/* toggle a flag */
void toggle_flag(int &flag);

// center mouse over object in a form, with save.
// don't use default values here, else the FL focus function
// won't work. 
void tool_focus_move(
    FL_FORM *form, 
    FL_OBJECT *obj, 
    FL_Coord x_offset, 
    FL_Coord y_offset
);
void tool_focus_move(FL_FORM *form);
void tool_focus_move(FL_FORM *form, FL_OBJECT *obj);
void tool_focus_set (FL_FORM *form, FL_OBJECT *obj);

// on screen. works great with initial placement. (see help_forms_aux).
void center_form(FL_FORM *form);

// conversion routine
const char *bytes_to_string(llong_t bytes);
const char *blocks_to_string(ulong n_blocks, ulong bsize);

// for startup event delay nerving.
void callback_loop_init_cb(FL_OBJECT *, long);

// uses own process and process group for cmd:
int safe_system(const char *cmd);

#endif	// GLOBAL_H
