//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>   // gettimeofday
#include <unistd.h>

#include "browser.h"
#include "cb_fkey.h"
#include "cb_key.h"
#include "client_msg.h"
#include "edit_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "ghelp_forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "pathnames.h"
#include "scroll_events.h"
#include "status_forms_aux.h"
#include "utility.h"
#include "wildcard.h"

//static int dbg_flag = TRUE;
static int dbg_flag = tool_debug_flag;

int help_fontsize  = FL_NORMAL_SIZE,
    help_fontstyle = FL_FIXED_STYLE
;

const char *help_ID_string = "help";

prm_entry_t help_prm_table[] = {
    /*
    {"ww",           int_e,   (void *) &help_ww}, 
    {"wh",           int_e,   (void *) &help_wh}, 
    {"wx",           int_e,   (void *) &help_wx}, 
    {"wy",           int_e,   (void *) &help_wy},
    */
    {"fontsize",     int_e,   (void *) &help_fontsize}, 
    {"fontstyle",    int_e,   (void *) &help_fontstyle}, 
    {"",             int_e,   NULL}
};

void help_save_defaults(ofstream &ofile){
    /*
    if(HelpForm->visible){
        var_sync();
    }
    */
    resource_save_defaults(ofile, help_prm_table, help_ID_string);
}    

int help_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    M_trace("help_load_default", "start");
    return resource_load_default(
        help_prm_table,  
        help_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

void load_text_in_browser(FL_OBJECT *browser, const char *text[]){
    if(!browser || !text) return;
    fl_freeze_form(browser->form);
    fl_clear_browser(browser);
    int i=0;
    while(text[i] != NULL){
	if(text[i][0]){
	    fl_addto_browser(browser, text[i]);
	} else {
	    fl_addto_browser(browser, "\n");
	}
	i++;
    }
    fl_set_browser_topline(browser, 1);
    fl_unfreeze_form(browser->form);
}

void HelpContextFormCloseReturnButtonCB(FL_OBJECT *, long){
    if(HelpContextForm->visible){
	focus_restore_start();
	fl_hide_form(HelpContextForm);
	focus_restore_finish();
    }
}

static int help_context_form_raw_cb(FL_FORM *form, void *xev){
    return standard_form_event_cb(form, (XEvent *) xev);
}

static void help_form_help(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_object_shortcut(HelpContextFormCloseReturnButton, "Cc^M", TRUE);
	fl_register_call_back(HelpContextForm, FL_ALL_EVENT, help_context_form_raw_cb);
    }
    focus_save();
    if(!HelpContextForm->visible){
	fl_show_form(HelpContextForm, FL_PLACE_CENTER, TRUE, "Help: Help");
    } else {
	XMapRaised(fl_display, HelpContextForm->window);
    }
    tool_focus_move(HelpContextForm, HelpContextFormCloseReturnButton);
}

void help_hide_func(void){
    HelpFormCloseReturnButtonCB(NULL, 0);
}

void HelpOptionsButtonCB(FL_OBJECT *, long){    
    standard_browser_font_cb(HelpFormBrowser, help_fontsize, help_fontstyle, "Help Options");
}

void help_options_func(void){
    HelpOptionsButtonCB(NULL, 0);
}
    
static int help_form_raw_cb(FL_FORM *form, void *xev){
   if(scroll_browser_event_cb(HelpFormBrowser, (XEvent *) xev))return TRUE;
   if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
   if(standard_options_event_cb(form, (XEvent *) xev, help_options_func, TRUE)) return TRUE;
   return standard_form_event_cb(form, (XEvent *) xev, help_hide_func, help_form_help);
}

static void_func_pt help_more_func = 0;

void load_help_browser(const char **text,
		       const char *title,
		       const char *more_title,
		       void_func_pt more_func
		       ){
    char buffer[1024];
    sprintf(buffer, "Help: %s", title);
    // display the text in a help browser. open browser if it was closed.
    // also winpop it.
    static int first = TRUE;
    if(!text) return;
    if(first){
	first = FALSE;
	fl_set_browser_leftslider(HelpFormBrowser, TRUE);
	fl_set_object_shortcut(HelpFormCloseReturnButton, "Cc^M", TRUE);
	fl_set_button_shortcut(HelpFormMoreButton, "Mm", TRUE);

	fl_set_object_resize(HelpFormCloseReturnButton, FL_RESIZE_NONE);
	fl_set_object_resize(HelpFormMoreButton,        FL_RESIZE_NONE);
	fl_set_object_resize(HelpFormText0,             FL_RESIZE_NONE);
	fl_set_object_resize(HelpFormText1,             FL_RESIZE_NONE);
	fl_set_object_resize(HelpFormText2,             FL_RESIZE_NONE);
	fl_set_object_resize(HelpFormBrowser,           FL_RESIZE_ALL);	

	fl_set_object_gravity(HelpFormCloseReturnButton, SouthWestGravity, SouthGravity);
	fl_set_object_gravity(HelpFormMoreButton,        SouthWestGravity, SouthGravity);
	fl_set_object_gravity(HelpFormText0,             SouthWestGravity, SouthGravity);
	fl_set_object_gravity(HelpFormText1,             SouthWestGravity, SouthGravity);
	fl_set_object_gravity(HelpFormText2,             SouthWestGravity, SouthGravity);
	fl_set_object_gravity(HelpFormBrowser,           NorthGravity, SouthEastGravity);	

	fl_register_call_back(HelpForm, FL_ALL_EVENT, help_form_raw_cb);

	// allow for a little extra space.
	fl_set_browser_fontstyle(HelpFormBrowser, help_fontstyle); // courier-normal ?
	fl_set_browser_fontsize (HelpFormBrowser, help_fontsize);
        // FL_Coord tmp_w = (FL_Coord) (8.5*fl_get_string_width(help_fontstyle, help_fontsize, "          ", 10));
	FL_Coord tmp_w = (FL_Coord) (8.5*fl_get_string_width(help_fontstyle, help_fontsize, "aFiMrVzEhL", 10));
	fl_set_form_size(HelpForm, tmp_w, HelpForm->h);
	center_form(HelpForm);
    }

    if(!more_title){
      fl_hide_object(HelpFormMoreButton);
      help_more_func = 0;
    } else {
      help_more_func = more_func;
      fl_set_object_label(HelpFormMoreButton, more_title);
      fl_show_object(HelpFormMoreButton);
    }
      
    set_to_ret_button(HelpFormCloseReturnButton);
    load_text_in_browser(HelpFormBrowser, text);
    focus_save();
    if(!HelpForm->visible){
	fl_set_initial_placement(HelpForm,
	    HelpForm->x,
	    HelpForm->y,
	    HelpForm->w,
	    HelpForm->h
	);
	fl_set_form_minsize(HelpForm, 40, 40);
	fl_set_form_maxsize(HelpForm, tool_max_ww, tool_max_wh);
	fl_show_form(HelpForm, FL_PLACE_FREE, TRUE, buffer);
    } else {
	XStoreName(fl_display, HelpForm->window, buffer);
	XMapRaised(fl_display, HelpForm->window);
    }
    tool_focus_move(HelpForm, HelpFormCloseReturnButton);

    if(HelpForm->deactivated){      // we are extra careful. 
	fl_activate_form(HelpForm);
    }
}

void HelpFormMoreButtonCB(FL_OBJECT *, long){
    if(help_more_func){
	(*help_more_func)();
    } else {
	cerr << "MoreButton called with NULL func to execute.\n";
    }
}

void HelpFormCloseReturnButtonCB(FL_OBJECT *, long){
    if(HelpForm->visible){
	focus_restore_start();
	fl_hide_form(HelpForm);
	focus_restore_finish();
    }
}

/*
 * show instructions, put default_text into input field.
 * We have one small version, one bigger version. Auto-resizing
 * would be cool.
 * also have a help text that works when F1 is pressed. 
 */

static int
    local_continue_flag  = TRUE,    // keep on munging
    local_cancelled_flag = FALSE    // if true, cancelled -> return NULL.
;

static const char
    **local_help_text = NULL,
    *local_help_title = NULL
;

static void local_help_func(void){
    if(local_help_text){
	load_help_browser(local_help_text, local_help_title, NULL);
    }
}

static void local_close_func(void){
    OKFormCancelButtonCB(NULL, 0);  // close is side-effect. 
}

static int ok_cancel_raw_cb(FL_FORM *form, void *xev){
    if(form != OKForm){
	cerr << "invalid form in raw cb.\n";
	return FALSE;
    }
    if(adjust_return_button(OKForm, (XEvent *) xev)){
	//cerr << "adjust says true.\n";
	return TRUE;
    }
    if(standard_form_event_cb(OKForm, (XEvent *) xev, local_close_func, local_help_func)){
	//cerr << "standard says true.\n";
	return TRUE;
    } else {
	//cerr << "standard says false.\n";
	return FALSE;
    }	
}

const char *OKCancel(
    const char *instructions, 
    const char *default_text, 
    const char **help_text,
    const char *help_title,
    int no_close_flag, 
    int special_range_flag,
    int begin_range,
    int end_range
){
    static char 
	no_instructions[] = "<uhoh>",
	no_default_text[] = "<uhoh>"
    ;

    static int first = TRUE;
    if(first){
	first = FALSE;
	// make_standard_cancel_shortcut(OKFormCancelButton);
	fl_set_input_color(OKFormInput, FL_WHITE, FL_RED);
	fl_register_call_back(OKForm, FL_ALL_EVENT, ok_cancel_raw_cb);
    }

    if(!instructions) instructions = no_instructions;
    fl_set_object_label(OKFormText, instructions);

    if(!default_text) default_text = no_default_text;
    fl_set_input(OKFormInput, default_text);

    if(help_text){
	local_help_text = help_text;
	local_help_title = help_title;
    } else {
	local_help_text = NULL;
	local_help_title = NULL;
    }
	
    if(special_range_flag){
	fl_set_input_selected_range(OKFormInput, begin_range, end_range);
    } else {
	fl_set_input_selected(OKFormInput, TRUE);
    }

    fl_deactivate_all_forms();

    local_continue_flag  = TRUE;  // re-init. 
    local_cancelled_flag = FALSE;
    
    set_to_ret_button(OKFormGoReturnButton);
    if(!OKForm->visible){
	// i.e., not the noclose option is being used .. 
	focus_save();
    }
    if(auto_mouse_flag){
	fl_show_form(OKForm, FL_PLACE_CENTER, FL_TRANSIENT, "Question:");
    } else {
	fl_show_form(OKForm, FL_PLACE_MOUSE, FL_TRANSIENT, "Question:");
    }
    tool_focus_move(OKForm, OKFormGoReturnButton);

    if(OKForm->deactivated){      // we are extra careful. 
	fl_activate_form(OKForm);
    }

    while(local_continue_flag){
	if(!tool_mung_forms()){
	    local_continue_flag = FALSE;
	}
    }
    if(!no_close_flag){
	OKCancelClose();
    }
    fl_activate_all_forms();
    if(local_cancelled_flag){
	return NULL;
    } else {
	return fl_get_input(OKFormInput);
    }
}

void OKFormCancelButtonCB(FL_OBJECT *, long){
    local_continue_flag  = FALSE;
    local_cancelled_flag = TRUE;
}

void OKFormGoReturnButtonCB(FL_OBJECT *, long){
    local_continue_flag  = FALSE;
    local_cancelled_flag = FALSE; // extra.
}

void BigOKCancelClose(void){
    if(BigOKForm->visible){
	focus_restore_start();
	fl_hide_form(BigOKForm);
	focus_restore_finish();
    }
}

void OKCancelClose(void){
    if(OKForm->visible){
	focus_restore_start();
	fl_hide_form(OKForm);
	focus_restore_finish();
    }
}

static int big_ok_cancel_raw_cb(FL_FORM *form, void *xev){
    if(form != BigOKForm){
	cerr << "invalid form in raw cb.\n";
	return FALSE;
    }
    if(adjust_return_button(BigOKForm, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(BigOKForm, (XEvent *) xev, local_close_func, local_help_func);
}

const char *BigOKCancel(
    const char *instructions, 
    const char *default_text, 
    const char **help_text,
    const char *help_title,
    int no_close_flag, 
    int special_range_flag,
    int begin_range,
    int end_range
){
    static char 
	no_instructions[] = "no instructions",
	no_default_text[] = "no default text"
    ;
    static int first = TRUE;
    if(first){
	first = FALSE;
	// make_standard_cancel_shortcut(BigOKFormCancelButton); NO
	fl_set_input_color(BigOKFormInput, FL_WHITE, FL_RED);

	fl_register_call_back(BigOKForm, FL_ALL_EVENT, big_ok_cancel_raw_cb);
    }

    if(!instructions) instructions = no_instructions;
    fl_set_object_label(BigOKFormText, instructions);

    if(!default_text) default_text = no_default_text;
    fl_set_input(BigOKFormInput, default_text);

    if(help_text){
	local_help_text = help_text;
	local_help_title = help_title;
	//BigOKForm->help_func = local_help_func;
    } else {
	local_help_text = NULL;
	local_help_title = NULL;
	//BigOKForm->help_func = NULL;
    }

    if(special_range_flag){
	fl_set_input_selected_range(BigOKFormInput, begin_range, end_range);
    } else {
	fl_set_input_selected(BigOKFormInput, TRUE);
    }

    local_continue_flag  = TRUE;  // re-init. 
    local_cancelled_flag = FALSE;

    fl_deactivate_all_forms();
    set_to_ret_button(BigOKFormGoReturnButton);
    if(!BigOKForm->visible){
	focus_save();
    }
    if(auto_mouse_flag){
	fl_show_form(BigOKForm, FL_PLACE_CENTER, FL_TRANSIENT, "Question");
    } else {
	fl_show_form(BigOKForm, FL_PLACE_MOUSE, FL_TRANSIENT, "Question");
    }
    tool_focus_move(BigOKForm, BigOKFormGoReturnButton);

    if(BigOKForm->deactivated){
	fl_activate_form(BigOKForm);
    }

    while(local_continue_flag){
	if(!tool_mung_forms()){
	    local_continue_flag = FALSE;
	}
    }

    if(!no_close_flag){
	BigOKCancelClose();
    }
    fl_activate_all_forms();
    if(local_cancelled_flag){
	return NULL;
    } else {
	return fl_get_input(BigOKFormInput);
    }
}

/*** MultiQForm ***/

static int multiq_retval = -1;

void MultiQFormButtonCB(FL_OBJECT *, long val){
    multiq_retval = (int) val;
    if(MultiQForm->visible){
	focus_restore_start();
	fl_hide_form(MultiQForm);
	focus_restore_finish();
    }
}

static void multiq_close_func(void){
    MultiQFormButtonCB(NULL, -1);
}

void multiq_help_func(void){
    // load_help_browser(action_multi_help_text, "Do All, Do File, Cancel File, Cancel All");
}

static KeySym mq_keysym_0,mq_keysym_1,mq_keysym_2,mq_keysym_3,mq_keysym_4;

static int multiq_raw_cb(FL_FORM *form, void *xev){
    if(form != MultiQForm){
	cerr << "invalid form in multi raw cb.\n";
	return FALSE;
    }
    if(((XEvent *) xev)->type == KeyPress){
	int retval = FALSE, tmp_retval = -1;
	XKeyEvent *xkev = (XKeyEvent *)xev;
	KeySym keysym;
	char keybuf[2];
	XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
	if(keysym != NoSymbol){
	    if(keysym == mq_keysym_0){
		tmp_retval = 0;
		retval = TRUE;
	    } else if(keysym == mq_keysym_1){
		tmp_retval = 1;
		retval = TRUE;
	    } else if(keysym == mq_keysym_2){
		tmp_retval = 2;
		retval = TRUE;
	    } else if(keysym == mq_keysym_3){
		tmp_retval = 3;
		retval = TRUE;
	    } else if(keysym == mq_keysym_4){
		tmp_retval = 4;
		retval = TRUE;
	    } else if(keysym == XK_F12) {
		// ignore F12 exit's here.
		return TRUE;
		}
	}
	if(retval){
	    multiq_close_func();
	    multiq_retval = tmp_retval;
	    return TRUE;
	}
    }   
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, multiq_close_func, multiq_help_func);
}

int ask_multi_question(
    int danger_flag, 
    int default_button,
    const char *msg, 
    const char *label_0, const char *shortcut_0, KeySym keysym_0,
    const char *label_1, const char *shortcut_1, KeySym keysym_1,
    const char *label_2, const char *shortcut_2, KeySym keysym_2,
    const char *label_3, const char *shortcut_3, KeySym keysym_3,
    const char *label_4, const char *shortcut_4, KeySym keysym_4
){
    return ask_multi_question(danger_flag, NULL, 0, 0, default_button, msg, 
			      label_0,shortcut_0,keysym_0,
			      label_1,shortcut_1,keysym_1,
			      label_2,shortcut_2,keysym_2,
			      label_3,shortcut_3,keysym_3,
			      label_4,shortcut_4,keysym_4);
}

int ask_multi_question(
    int danger_flag, 
    unsigned char *bitmap_bits, int bitmap_w, int bitmap_h, 
    int default_button,
    const char *msg, 
    const char *label_0, const char *shortcut_0, KeySym keysym_0,
    const char *label_1, const char *shortcut_1, KeySym keysym_1,
    const char *label_2, const char *shortcut_2, KeySym keysym_2,
    const char *label_3, const char *shortcut_3, KeySym keysym_3,
    const char *label_4, const char *shortcut_4, KeySym keysym_4
){
    static int first = TRUE;
    int n_buttons = 2;
    
    multiq_retval = -1;
    mq_keysym_0 = keysym_0;
    mq_keysym_1 = keysym_1;
    mq_keysym_2 = keysym_2;
    mq_keysym_3 = keysym_3;
    mq_keysym_4 = keysym_4;
    
    if(first){
	first = FALSE;
	// no shortcut, I don't know which one should be the cancel button.
	// on escape, we return -1, not the cancel button number. 
	fl_register_call_back(MultiQForm, FL_ALL_EVENT, multiq_raw_cb);
    }
    fl_freeze_form(MultiQForm);

    // labels and hiding
    fl_set_object_label(MultiQFormText,msg);
    if(label_0){
	fl_set_object_label(MultiQForm0Button,label_0);
    } else {
	cerr << "need at least 2 labels in multi_q.\n";
	return -1;
    }
    if(label_1){
	fl_set_object_label(MultiQForm1Button,label_1);
    } else {
	cerr << "need at least 2 labels in multi_q.\n";
	return -1;
    }
    if(label_2){
	n_buttons++;
	fl_set_object_label(MultiQForm2Button,label_2);
	fl_show_object(MultiQForm2Button);
    } else {
	fl_hide_object(MultiQForm2Button);
    }
    if(label_3){
	n_buttons++;
	fl_set_object_label(MultiQForm3Button,label_3);
	fl_show_object(MultiQForm3Button);
    } else {
	fl_hide_object(MultiQForm3Button);
    }
    if(label_4){
	n_buttons++;
	fl_set_object_label(MultiQForm4Button,label_4);
	fl_show_object(MultiQForm4Button);
    } else {
	fl_hide_object(MultiQForm4Button);
    }

    const int form_bitmap_w = 60;
    if(bitmap_bits){
      //fl_set_object_color(MultiQFormBitmap, FL_WHITE, FL_BLACK);
      fl_set_object_color(MultiQFormBitmap, FL_BLACK, FL_WHITE); 
      fl_set_bitmap_data(MultiQFormBitmap, bitmap_w, bitmap_h, bitmap_bits);
      fl_show_object(MultiQFormBitmap);
    } else {
      fl_hide_object(MultiQFormBitmap);
    }

    // shortcuts (reset type before setting to retbutton !)
    MultiQForm0Button->type = FL_NORMAL_BUTTON;
    MultiQForm1Button->type = FL_NORMAL_BUTTON;
    MultiQForm2Button->type = FL_NORMAL_BUTTON;
    MultiQForm3Button->type = FL_NORMAL_BUTTON;
    MultiQForm4Button->type = FL_NORMAL_BUTTON;

    if (!shortcut_0) shortcut_0 = "";
    if (!shortcut_1) shortcut_1 = "";
    if (!shortcut_2) shortcut_2 = "";
    if (!shortcut_3) shortcut_3 = "";
    if (!shortcut_4) shortcut_4 = "";

    fl_set_object_shortcut(MultiQForm0Button, shortcut_0, TRUE);
    fl_set_object_shortcut(MultiQForm1Button, shortcut_1, TRUE);
    fl_set_object_shortcut(MultiQForm2Button, shortcut_2, TRUE);
    fl_set_object_shortcut(MultiQForm3Button, shortcut_3, TRUE);
    fl_set_object_shortcut(MultiQForm4Button, shortcut_4, TRUE);

    // return button (remember for mouse moving).
    FL_OBJECT *default_obj = NULL;
    switch(default_button){
    case 0:    set_to_ret_button(MultiQForm0Button); default_obj = MultiQForm0Button; break;
    case 1:    set_to_ret_button(MultiQForm1Button); default_obj = MultiQForm1Button; break;
    case 2:    set_to_ret_button(MultiQForm2Button); default_obj = MultiQForm2Button; break;
    case 3:    set_to_ret_button(MultiQForm3Button); default_obj = MultiQForm3Button; break;
    case 4:    set_to_ret_button(MultiQForm4Button); default_obj = MultiQForm4Button; break;
    default:   set_to_ret_button(MultiQForm0Button); default_obj = MultiQForm0Button; break;
    }

    // raw size (no margin)
    int msg_w = fl_get_string_width(
	MultiQFormText->lstyle,
	MultiQFormText->lsize, 
	MultiQFormText->label,
	strlen(MultiQFormText->label)
    );

    // button size
    int max_button_w = 0;
    for(int i=0; i<n_buttons; i++){
	int tmp_w;
	switch(i){
	case 0:
	    tmp_w = fl_get_string_width(
		MultiQForm0Button->lstyle, MultiQForm0Button->lsize,
		MultiQForm0Button->label,  strlen(MultiQForm0Button->label)
	    );
	    break;
	case 1:
	    tmp_w = fl_get_string_width(
		MultiQForm1Button->lstyle, MultiQForm1Button->lsize,
		MultiQForm1Button->label,  strlen(MultiQForm1Button->label)
	    );
	    break;
	case 2:
	    tmp_w = fl_get_string_width(
		MultiQForm2Button->lstyle, MultiQForm2Button->lsize,
		MultiQForm2Button->label,  strlen(MultiQForm2Button->label)
	    );
	    break;
	case 3:
	    tmp_w = fl_get_string_width(
		MultiQForm3Button->lstyle, MultiQForm3Button->lsize,
		MultiQForm3Button->label,  strlen(MultiQForm3Button->label)
	    );
	    break;
	case 4:
	    tmp_w = fl_get_string_width(
		MultiQForm4Button->lstyle, MultiQForm4Button->lsize,
		MultiQForm4Button->label,  strlen(MultiQForm4Button->label)
	    );
	    break;
	default:
	    tmp_w = 0;
	    break;
	}
	if(tmp_w > max_button_w){
	    max_button_w = tmp_w;
	}
    }

    // fix total size. Count retarrow twice cause of text centering.
    const int border_margin = 5, text_margin = 10, retarrow_margin = 10;
    max_button_w += 2 * (text_margin + border_margin + retarrow_margin);
    
    int total_text_w   = 2 * text_margin + msg_w,
	total_button_w = n_buttons * max_button_w
    ;
    // make it divisible by n_buttons, for good fit.
    while(total_text_w % n_buttons != 0){
	total_text_w++;
    }

    // adjust button_w if text is bigger.
    if(total_text_w > total_button_w){
	max_button_w = total_text_w / n_buttons;
    }
			  
    // compute total w of all objects and set.
    int total_obj_w = max(total_text_w, total_button_w);
    if(bitmap_bits){
        total_obj_w += form_bitmap_w;
    }
    fl_set_form_size(MultiQForm, total_obj_w + 2 * border_margin, MultiQForm->h);

    int obj_margin = border_margin;
    if(bitmap_bits){
      // can't do that earlier
      MultiQFormBitmap->w = form_bitmap_w;
      obj_margin += form_bitmap_w;
      MultiQFormBitmap->x = border_margin;
    }
    // fix obj size and position (visibility doesn't matter here).
    MultiQFormText->x = obj_margin;
    MultiQFormText->w = total_obj_w;
    if(bitmap_bits){
      MultiQFormText->w -= form_bitmap_w;
    }
    MultiQForm0Button->x = obj_margin;
    MultiQForm0Button->w = MultiQForm1Button->w = MultiQForm2Button->w = MultiQForm3Button->w =
	MultiQForm4Button->w = max_button_w;

    MultiQForm1Button->x = obj_margin + max_button_w;
    MultiQForm2Button->x = obj_margin + 2*max_button_w;
    MultiQForm3Button->x = obj_margin + 3*max_button_w;
    MultiQForm4Button->x = obj_margin + 4*max_button_w;

    fl_unfreeze_form(MultiQForm);
    fl_deactivate_all_forms();

    int below_focus_flag =
	MultiQForm0Button->belowmouse || MultiQForm1Button->belowmouse ||
	MultiQForm2Button->belowmouse || MultiQForm3Button->belowmouse ||
	MultiQForm4Button->belowmouse
    ;	
    if(!below_focus_flag) focus_save();

    if(auto_mouse_flag){
	fl_show_form(MultiQForm, FL_PLACE_CENTER, FL_TRANSIENT, "Multi-Question:");
    } else {
	fl_show_form(MultiQForm, FL_PLACE_MOUSE, FL_TRANSIENT, "Multi-Question:");
    }
    if(!below_focus_flag) tool_focus_move(MultiQForm, default_obj);

    if(danger_flag){
	fl_set_cursor_color(XC_pirate, FL_BLACK, FL_RED);
	fl_set_cursor(MultiQForm->window, XC_pirate);
    }
    
    while(MultiQForm->visible){
	if(!tool_mung_forms()){
	    multiq_close_func();
	}
    }
    fl_activate_all_forms();
    if(danger_flag){
	fl_reset_cursor(MultiQForm->window);
    }
    return multiq_retval;
}

/*** OKQuestionForm ***/

static int question_yes_flag = FALSE;

int ask_question_raw_cb(FL_FORM *form, void *xev){
    //cerr << "q raw cb.\n";
    if(form != OKQuestionForm){
	cerr << "invalid form in raw cb.\n";
	return FALSE;
    }
    if(adjust_return_button(OKQuestionForm, (XEvent *) xev)){
	return TRUE;
    } 
    return FALSE;
}

int ask_question(int danger_flag, const char *format_string, ...){
    static int first = TRUE;
    if(first){
	first = FALSE;
	make_standard_cancel_shortcut(OKQuestionFormCancelButton);
	make_standard_ok_shortcut(OKQuestionFormGoReturnButton);
	fl_register_call_back(OKQuestionForm, FL_ALL_EVENT, ask_question_raw_cb);
	/*
	fl_set_object_resize(OKQuestionFormText, FL_RESIZE_NONE, WestGravity, EastGravity);
	fl_set_object_resize(OKQuestionFormGoReturnButton, FL_RESIZE_NONE, WestGravity, EastGravity);
	fl_set_object_resize(OKQuestionFormCancelButton, FL_RESIZE_NONE, WestGravity, EastGravity);
	*/
    }
    va_list arg_pointer;
    char question[1024];

    
    if(!format_string) return FALSE;
    va_start(arg_pointer, format_string);
    vsprintf(question, format_string, arg_pointer);
    va_end(arg_pointer);

    
    fl_set_object_label(OKQuestionFormText, question);
    int string_width = fl_get_string_width(OKQuestionFormText->lstyle, OKQuestionFormText->lsize, question, strlen(question)) + 10;  // extra padding :-)
    int text_width = 240;
    while(text_width < string_width){
	text_width += 80;
    }
    // necessary.
    fl_set_form_size(OKQuestionForm, text_width+10, OKQuestionForm->h);
    OKQuestionForm->w = text_width+10;
    OKQuestionFormText->x = 5;
    OKQuestionFormText->w = text_width;
    OKQuestionFormGoReturnButton->x = 5;
    OKQuestionFormGoReturnButton->w = (int) floor((double) (text_width/2));
    OKQuestionFormCancelButton->x   = OKQuestionFormGoReturnButton->x + OKQuestionFormGoReturnButton->w;
    OKQuestionFormCancelButton->w   = text_width - OKQuestionFormCancelButton->x + 5;

    fl_deactivate_all_forms();
    set_to_ret_button(danger_flag ? OKQuestionFormCancelButton : OKQuestionFormGoReturnButton);

    focus_save();
    if(auto_mouse_flag){
      fl_show_form(OKQuestionForm, FL_PLACE_CENTER, FL_TRANSIENT, "Question:");
    } else {
      fl_show_form(OKQuestionForm, FL_PLACE_MOUSE, FL_TRANSIENT, "Question:");
    }
    tool_focus_move(OKQuestionForm, danger_flag ? OKQuestionFormCancelButton : OKQuestionFormGoReturnButton);
    if(danger_flag){
	fl_set_cursor_color(XC_pirate, FL_BLACK, FL_RED);
	fl_set_cursor(OKQuestionForm->window, XC_pirate);
    }
    question_yes_flag = FALSE;
    while(OKQuestionForm->visible){
	if(!tool_mung_forms()){
	    OKQuestionFormCancelButtonCB(NULL, 0);
	}
    }
    fl_activate_all_forms();

    if(danger_flag){
	fl_reset_cursor(OKQuestionForm->window);
    }
    return question_yes_flag;
}

void OKQuestionFormCancelButtonCB(FL_OBJECT *, long){
    if(OKQuestionForm->visible){
	//cerr << "hiding OKQ.\n";
	focus_restore_start();
	fl_hide_form(OKQuestionForm);
	focus_restore_finish();
	question_yes_flag = FALSE;
    } else {
	cerr << "OKQ: already hidden.\n";
    }
}

void OKQuestionFormGoReturnButtonCB(FL_OBJECT *, long){
    if(OKQuestionForm->visible){
	//cerr << "hiding OKQ.\n";
	focus_restore_start();
	fl_hide_form(OKQuestionForm);
	focus_restore_finish();
    }
    question_yes_flag = TRUE;
}


/* launching */

static int
    launch_in_xwsh_flag   = FALSE,
    launch_with_args_flag = FALSE
;

int launch_form_raw_cb(FL_FORM *form, void *xev){
    if(form != LaunchForm) return FALSE;
    XEvent *xevent = (XEvent *) xev;
    KeySym keysym;
    char keybuf[1];
    
    switch(xevent->type){
    case KeyPress:
	if(dbg_flag) cerr << "browser: found an X key press event.\n";
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	case XK_F5:
	    LaunchFormQuickEditButtonCB(NULL, 0);
	    return TRUE;
	default:
	    break;
	}
    default:
	break;
    }
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev);
}

int launch_command(const char *cmd){
    if(!cmd){
	program_unstable("Trying to launch NULL command.");
	return FALSE;
    }
    static int first = TRUE;
    if(first){
	first = FALSE;
	make_standard_cancel_shortcut(LaunchFormCancelButton);
	make_standard_ok_shortcut(LaunchFormGoReturnButton);
	fl_set_button_shortcut(LaunchFormInXwshButton, "Xx", TRUE);
	fl_set_button_shortcut(LaunchFormQuickEditButton, "Ee", TRUE);
	fl_set_button_shortcut(LaunchFormWithArgsButton, "Aa", TRUE);
	fl_register_call_back(LaunchForm, FL_ALL_EVENT, launch_form_raw_cb);
    }
    fl_deactivate_all_forms();
    fl_set_object_label(LaunchFormNameText, cmd);
    set_to_ret_button(LaunchFormGoReturnButton);
    focus_save();
    if(auto_mouse_flag){
      fl_show_form(LaunchForm,FL_PLACE_CENTER, FL_TRANSIENT, "Launch Executable:");
    } else {
      fl_show_form(LaunchForm,FL_PLACE_MOUSE, FL_TRANSIENT, "Launch Executable:");
    }
    tool_focus_move(LaunchForm, LaunchFormGoReturnButton);
    while(LaunchForm->visible){
	if(!tool_mung_forms()){
	    LaunchFormCancelButtonCB(NULL, 0);
	}
    }
    fl_activate_all_forms();
    return TRUE;
}

void LaunchFormCancelButtonCB(FL_OBJECT *, long){
    if(LaunchForm->visible){
	focus_restore_start();
	fl_hide_form(LaunchForm);    
	focus_restore_finish();
    }
}

void LaunchFormWithArgsButtonCB(FL_OBJECT *, long){
    toggle_bool(launch_with_args_flag);
}

void LaunchFormInXwshButtonCB(FL_OBJECT *, long){
    toggle_bool(launch_in_xwsh_flag);
}

void LaunchFormGoReturnButtonCB(FL_OBJECT *, long){
    char cmdbuffer[1024];
    char *current_cmdbuf = LaunchFormNameText->label;
    if(!current_cmdbuf){
	error_msg("Tried to launch a NULL current cmdbuf.");
    }
    if(launch_with_args_flag){
	if(launch_in_xwsh_flag){
#ifndef SGI
	    sprintf(cmdbuffer, "%s -title %s -e ./%s &", xterm_path, current_cmdbuf, current_cmdbuf);
#else
	    sprintf(cmdbuffer, "%s -title %s -e ./%s &", winterm_path, current_cmdbuf, current_cmdbuf);
#endif
	} else {
	    sprintf(cmdbuffer, "./%s &", current_cmdbuf);
	}
	const char *result = OKCancel("Complete the command line:", cmdbuffer);
	if(!result){
	    return;
	} else {
	    strcpy(cmdbuffer, result);
	}
    } else {
	if(launch_in_xwsh_flag){
#ifndef SGI
	    sprintf(cmdbuffer,
	        "%s -title %s -e ./%s ",
	        xterm_path,
		current_cmdbuf,
		current_cmdbuf
	    );
#else
	    sprintf(cmdbuffer,
	        "%s -title %s -e ./%s ",
		winterm_path,    
		current_cmdbuf,
		current_cmdbuf
	    );
#endif
	} else {
	    sprintf(cmdbuffer, "./%s &", current_cmdbuf);
	}
    }
    if(dbg_flag) cerr << "cmdbuf: " << cmdbuffer << endl;
    fl_set_global_cursor(XC_watch);
    safe_system(cmdbuffer);
    fl_reset_global_cursor();
    LaunchFormCancelButtonCB(NULL, 0);
}

void LaunchFormQuickEditButtonCB(FL_OBJECT *, long){
    char *fname = LaunchFormNameText->label;
    LaunchFormCancelButtonCB(NULL, 0);
    if(!fname) return;
    quick_edit_file(fname, source_browser->get_cwd());
}

static int indicator_simple_flag = FALSE;

// return true on catch. 
static int indicator_time_trap(int update_time_flag = TRUE){
    static int first = TRUE;
    static struct timeval prev_click_time;

    if(first){
	first = FALSE;
	gettimeofday(&prev_click_time, NULL);
	    
	return FALSE;
    }
    struct timeval click_time;
    if(gettimeofday(&click_time, NULL) != 0){
	return FALSE;
    }
    long diff = msecs_diff(click_time, prev_click_time);

    if(diff < 100){
	// don't remember time, we do not want to continuously kill.
	//cerr << "time trap true.\n";
	return TRUE;
    } else {
	if(update_time_flag){
	    // remember new time., double click flag stays true.
	    prev_click_time.tv_sec  = click_time.tv_sec;
	    prev_click_time.tv_usec = click_time.tv_usec;
	}
	//cerr << "time trap false.\n";
	return FALSE;
    }
}

void indicator_bar (int n_steps, char *msg){
    indicator_simple_flag = FALSE;
    fl_set_object_lcol     (IndicatorFormText, FL_BLACK);
    fl_set_object_label    (IndicatorFormText, msg);
    fl_set_slider_bounds   (IndicatorFormValSlider, 0.00, (double) n_steps);
    fl_set_slider_precision(IndicatorFormValSlider, 1);
    fl_set_slider_value    (IndicatorFormValSlider, 0.00);
    fl_hide_object         (IndicatorFormMiniMsgText);
    fl_hide_object         (IndicatorFormProgressText);

    make_standard_cancel_shortcut (IndicatorFormCancelButton);

    fl_deactivate_all_forms();
    fl_activate_form(IndicatorForm);
    fl_set_timer(IndicatorFormTimer, 0.0);
    focus_save();
    // transient bombs under old fvwm, fullborder doesn't show.
    fl_show_form(IndicatorForm, FL_PLACE_CENTER, FL_TRANSIENT, "Progress Indicator:");
    tool_focus_move(IndicatorForm, IndicatorFormCancelButton);
}

void indicator_simple_bar (char *msg){
    indicator_simple_flag = TRUE;
    fl_set_object_lcol            (IndicatorSimpleFormText, FL_BLACK);
    fl_set_object_label           (IndicatorSimpleFormText, msg);
    fl_hide_object                (IndicatorSimpleFormMiniMsgText);
    make_standard_cancel_shortcut (IndicatorSimpleFormCancelButton);
    fl_deactivate_all_forms();
    fl_activate_form(IndicatorSimpleForm);
    fl_set_timer(IndicatorSimpleFormTimer, 0.0);
    focus_save();
    fl_show_form(IndicatorSimpleForm, FL_PLACE_CENTER, FL_TRANSIENT, "Progress Indicator:");
    tool_focus_move(IndicatorSimpleForm, IndicatorSimpleFormCancelButton);
}

void IndicatorSimpleFormTimerCB(FL_OBJECT *, long){
    if(IndicatorSimpleForm->visible){
	focus_restore_start();
	fl_hide_form(IndicatorSimpleForm);
	focus_restore_finish();
    }
}

void IndicatorFormTimerCB(FL_OBJECT *, long){
    if(IndicatorForm->visible){
	focus_restore_start();
	fl_hide_form(IndicatorForm);
	focus_restore_finish();
    }
}

static int indicator_simple_check(char *msg = NULL){
    if(!indicator_simple_flag){
	return indicator_check(msg);
    }
    if(!IndicatorSimpleForm->visible) return TRUE;
    if(indicator_time_trap()) return TRUE;

    FL_OBJECT *retobj;
    if(msg){
	fl_set_object_label(IndicatorSimpleFormMiniMsgText, msg);
	fl_show_object     (IndicatorSimpleFormMiniMsgText);	
    }
    retobj = fl_check_forms();
    //event2obj(retobj, PADPERIOD, IndicatorSimpleFormCancelButton);
    if(retobj == IndicatorSimpleFormCancelButton){
	fl_set_object_label(IndicatorSimpleFormText, "O P E R A T I O N   C A N C E L L E D !");
	fl_set_object_lcol (IndicatorSimpleFormText, FL_RED);
	fl_set_timer(IndicatorSimpleFormTimer, 2.0);
	fl_activate_all_forms();
	return FALSE;
    } else {
	XEvent xev;
	if(retobj == FL_EVENT){
	    fl_XNextEvent(&xev);
	}
	return TRUE;
    }
}

int indicator_check(char *msg){
    if(indicator_simple_flag){
	return indicator_simple_check(msg);
    }
    if(!IndicatorForm->visible) return TRUE;
    if(indicator_time_trap()) return TRUE;
    FL_OBJECT *retobj;
    if(msg){
	fl_set_object_label(IndicatorFormMiniMsgText, msg);
	fl_show_object     (IndicatorFormMiniMsgText);	
    }
    retobj = fl_check_forms();
    //event2obj(retobj, PADPERIOD, IndicatorFormCancelButton);
    if(retobj == IndicatorFormCancelButton){
	fl_set_object_label(IndicatorFormText, "O P E R A T I O N   C A N C E L L E D !");
	fl_set_object_lcol (IndicatorFormText, FL_RED);
	fl_set_timer(IndicatorFormTimer, 2.0);
	fl_activate_all_forms();
	return FALSE;
    } else {
	XEvent xev;
	if(retobj == FL_EVENT){
	    fl_XNextEvent(&xev);
	}
	return TRUE;
    }
}

int indicator_simple_step(int check_forms_flag = TRUE){
    if(!indicator_simple_flag){
	return indicator_step(check_forms_flag);
    }
    if(!IndicatorSimpleForm->visible) return TRUE;
    if(check_forms_flag){
	if(!indicator_simple_check()){
	    // that's ok... window stays open for 2 seconds.
	    //focus_restore();
	    return FALSE;
	}
    }
    return TRUE;
}

int indicator_step(int check_forms_flag){
    if(indicator_simple_flag){
	return indicator_simple_step(check_forms_flag);
    }
    if(!IndicatorForm->visible) return TRUE;
    static double step_size = 1.0;
    if(indicator_time_trap(FALSE)){
	step_size += 1.0;
	return TRUE;
    } 
    double val = fl_get_slider_value(IndicatorFormValSlider);
    fl_set_slider_value(IndicatorFormValSlider, val + step_size);
    step_size = 1.0;

    if(check_forms_flag){
	if(!indicator_check()){
	    //focus_restore();
	    return FALSE;
	}
    }
    return TRUE;
}

int indicator_simple_ministep(int check_forms_flag = TRUE){
    if(!indicator_simple_flag){
	return indicator_ministep(check_forms_flag);
    }
    if(!IndicatorSimpleForm->visible) return TRUE;
    if(indicator_time_trap()) return TRUE;
    static int state = 0;
    static char newstate[2];
    newstate[1] = '\0';
    switch(state){
    case 0: newstate[0] = '|'; break;
    case 1: newstate[0] = '/'; break;
    case 2: newstate[0] = '-'; break;
    case 3: newstate[0] = '\\'; break;
    }
    fl_set_object_label(IndicatorSimpleFormProgressText, newstate);
    fl_show_object     (IndicatorSimpleFormProgressText);	
    state++;
    if(state>3) state = 0;

    if(check_forms_flag){
	if(!indicator_simple_check()){
	    //focus_restore();
	    return FALSE;
	}
    }
    return TRUE;
}

int indicator_ministep(int check_forms_flag){
    if(indicator_simple_flag){
	return indicator_simple_ministep(check_forms_flag);
    }
    if(!IndicatorForm->visible) return TRUE;
    if(indicator_time_trap()) return TRUE;
    static int state = 0;
    static char newstate[2];
    newstate[1] = '\0';
    switch(state){
    case 0: newstate[0] = '|'; break;
    case 1: newstate[0] = '/'; break;
    case 2: newstate[0] = '-'; break;
    case 3: newstate[0] = '\\'; break;
    }
    fl_set_object_label(IndicatorFormProgressText, newstate);
    fl_show_object     (IndicatorFormProgressText);	
    state++;
    if(state>3) state = 0;

    if(check_forms_flag){
	if(!indicator_check()){
	    //focus_restore();
	    return FALSE;
	}
    }
    return TRUE;
}

void indicator_end (void){
    if(IndicatorForm->visible){
	focus_restore_start();
	fl_hide_form(IndicatorForm);
	focus_restore_finish();
    }
    if(IndicatorSimpleForm->visible){
	focus_restore_start();
	fl_hide_form(IndicatorSimpleForm);
	focus_restore_finish();
    }
    fl_activate_all_forms();
}

void info_show_form(void){
    global_help_show_form();
}

