/*
 * interface functions and objects.
 * extra functionality must be explicitly added/included via cwdForms.h.
 */

#ifndef HELPFORMSAUX_H
#define HELPFORMSAUX_H

#include <fstream.h>
#include "forms.h"

extern FL_FORM *HelpForm, *MessageForm;

extern FL_OBJECT
        *MessageFormText;

extern void info_show_form(void);

extern int help_fontsize, help_fontstyle;

void help_save_defaults(ofstream &ofile);

int help_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
);

void load_text_in_browser(FL_OBJECT *browser, const char *text[]);
// more up2date: 
// set more_title to NULL if you want no more button, or supply your own.
void load_help_browser(const char **text, const char *title = "",
		       const char *more_title = "More ...",
		       void_func_pt more_func = info_show_form
		       );

/* display the default text and instructions.
 * ask for confirmation. returns NULL on cancel, 
 * data string on OK. 
 * BigOK is a version with a larger window, for longer text.
 */
const char *OKCancel(
    const char *instructions, 
    const char *default_text, 
    const char **help_text  = NULL,
    const char *help_title  = NULL,
    int no_close_flag = FALSE,
    int special_range_flag = FALSE,
    int begin_range = 0,
    int end_range   = 0
);

const char *BigOKCancel(
    const char *instructions, 
    const char *default_text, 
    const char **help_text  = NULL,
    const char *help_title  = NULL,
    int no_close_flag = FALSE, 
    int special_range_flag = FALSE,
    int begin_range = 0,
    int end_range   = 0
);

void    OKCancelClose(void);
void BigOKCancelClose(void);

// display a question only. return FALSE on cancel.
int ask_question(int danger_flag, const char *format_string, ...);

// question with multiple (at least 2) buttons and shortcuts. 
// return number of button pressed. -1 on escape/failure.
// potentially add help text, and button no for default/return button.

int ask_multi_question(
    int danger_flag, 
    unsigned char *bitmap_bits, int bitmap_w, int bitmap_h, 
    int default_button,
    const char *msg, 
    const char *label_0,        const char *shortcut_0,      KeySym keysym_0,
    const char *label_1,        const char *shortcut_1,      KeySym keysym_1 = NoSymbol,
    const char *label_2 = NULL, const char *shortcut_2 = "", KeySym keysym_2 = NoSymbol,
    const char *label_3 = NULL, const char *shortcut_3 = "", KeySym keysym_3 = NoSymbol,
    const char *label_4 = NULL, const char *shortcut_4 = "", KeySym keysym_4 = NoSymbol
);

int ask_multi_question(
    int danger_flag, 
    int default_button,
    const char *msg, 
    const char *label_0,        const char *shortcut_0,      KeySym keysym_0,
    const char *label_1,        const char *shortcut_1,      KeySym keysym_1 = NoSymbol,
    const char *label_2 = NULL, const char *shortcut_2 = "", KeySym keysym_2 = NoSymbol,
    const char *label_3 = NULL, const char *shortcut_3 = "", KeySym keysym_3 = NoSymbol,
    const char *label_4 = NULL, const char *shortcut_4 = "", KeySym keysym_4 = NoSymbol
);

// launch a command.
int launch_command(const char *cmd);

/* control a `work in progress'-indicator bar. 
 * all forms a deactivated while the indicator bar is visible.
 * The underlying process shows it's progress, and queries the
 * indicator bars 'cancel' button. The step function returns
 * TRUE as long as cancel wasn't pressed.
 * the ministep function shows & advances a little rotating bar |/-\ 
 * and also returns TRUE.
 * the simple bar has no slider, and is used for atomic operations.
 *
 * the application says how many steps it is planning to do 
 * until completion. At present, there should be at most one
 * indicator active in the system.
 */

int  indicator_check      (char *msg = NULL);
int  indicator_step       (int check_forms_flag = TRUE);
int  indicator_ministep   (int check_forms_flag = TRUE);
void indicator_bar        (int n_steps, char *msg = "Hang on ...");
void indicator_simple_bar (char *msg = "Hang on ...");
void indicator_end        (void);

// creation routine
extern void help_create_the_forms(void);

#endif // HELPFORMSAUX_H

