
#include <iostream.h>
#include "linklist.h"
#include "utility.h"

linklist_c::linklist_c() { 
    listsize = 0;

    listError = 0;       // Error Return Value.

    head = new node_c; 
    tail = new node_c; 

    head->previous = 0; 
    head->next = tail; 

    tail->next = 0; 
    tail->previous = head; 

    cnode = head; 
    cnode_index = -1;
}

linklist_c::~linklist_c(){ 	// Zap list when leaving.
    headend(); 
    while(listsize) remove();
    delete head;
    delete tail; 
}

void* linklist_c::operator[](int index){	// main access op
    int distance, i;
    if(index>= 0 && index < listsize){
	distance = index - cnode_index;
	if(distance < 0){
	    for(i=0; i > distance; i--){
		backstep();
	    }
	} else {
	    for(i=0; i< distance; i++){
		step();
	    }
	}
	return(getval());
    } else {
	error_msg("linklist: Illegal index %d. size %d", index, listsize);
	return NULL;
    }
}

void linklist_c::add(void* E){
    cnode = tail; 
    cnode_index = listsize;
    linklist_c::insert(E); 
}	

void linklist_c::backstep(void){ 
    // cnode = ((cnode->previous) ? cnode->previous : cnode); translates to
    if(cnode->previous != NULL){
	// that is: If not head, step back.
	cnode = cnode->previous;
	cnode_index--;
    } 	// else stay

    // cnode = ((cnode->previous) ? cnode : cnode->next); translates to
    if(cnode->previous == NULL){
	// If head, advance
	cnode = cnode->next;
	cnode_index++;
    }	// else stay
}

int linklist_c::find(void* E){
    int retval;
    int index;

    node_c *nptr = head;
    index = -1;
    while(nptr->next && nptr->data != E){
	index++;
	nptr = nptr->next;
    }
    retval = (nptr->next) ? 1 : 0;
    if(retval){
	cnode = nptr;
	cnode_index = index;
    }
    return retval;
}	

int linklist_c::getsize(void){
    return listsize;
}

void* linklist_c::getval(void){
    listError = 0;

    //  Error messages had been used for testing.
    if(!cnode->previous){
	cerr << "\nNo List?! Pointer is at home position! ";
	return NULL;
    } else if(!cnode->next){
	cerr << "\nPointer off the end of List! ";
	return NULL;
    }

    if((!cnode->previous) || (!cnode->next)) listError = 1;
    return cnode->data;
}

void linklist_c::headend(void){
    if(listsize){
	cnode = head->next; 
	cnode_index = 0;
    } else { 
	cnode = head; 
	cnode_index = -1;
    }
} 

void linklist_c::insert(void* E){
    if (!cnode) {
	cerr << "No cnode in linklist_c::insert.\n";
	return;
    }
    node_c *newptr = new node_c;
    node_c *pptr = cnode->previous;

    // If at the head of an empty list, fake an add(). 
    if(!listsize && !cnode->previous) { cnode = tail; pptr = head; }

    // build new node
    newptr->data = E;
    newptr->next = cnode;
    newptr->previous = cnode->previous;

    // Update Adjacent Nodes
    pptr->next = newptr;
    cnode->previous = newptr;

    cnode = newptr;	// insert occurs on the old position. cnode_index not changed.
    listsize++;
}

void linklist_c::remove(void){
    node_c *pptr = cnode->previous;
    node_c *nptr = cnode->next;

    listError = 0;
    
    // These Error messages aren't needed and are
    //  better handled in main() or elsewhere.
    //  So they've been commented out.

    if(!listsize){
	listError = 1;
	cerr << "\nError!  You can't remove from an Empty List!"; 
    } else if(!pptr){
	listError = 1;
        cerr << "\nError!  You can't remove the Head Node!"; 
    } else if(!nptr){
	listError = 1;
        cerr << "\nError!  You can't remove the Tail Node!";
    } else {
        // let the previous and next nodes reference each other.
        pptr->next = cnode->next;
        nptr->previous = cnode->previous;
	
        // remove the current node.
        delete cnode; 

        // If the next pointer is not tail, make it current,
        // else make the previous pointer the current node.
        // cnode = (nptr->next) ? nptr : pptr;
	if(nptr->next == NULL){
	    // next is tail
	    cnode = pptr;
	    cnode_index--;
	} else {
	    // next is not tail (index remains unchanged)
	    cnode = nptr;
	}
        listsize--;
    }
} 

void linklist_c::step(void){ 
    // cnode = (cnode->next) ? cnode->next : cnode; 
    if(cnode->next != NULL){
	// if not tail, step forward
	cnode = cnode->next;
	cnode_index++;
    }

    // cnode = (cnode->next) ? cnode : cnode->previous; 
    if(cnode->next == NULL){
	// if tail, step back
	cnode =  cnode->previous;
	cnode_index--;
    }
}   

void linklist_c::tailend(void){ 
    if(listsize){
	cnode = tail->previous; 
	cnode_index = listsize - 1;
    } else {
	cnode = tail; 
	cnode_index = listsize;
    }
} 


//  Output the list, head -> tail. 
void list_traverse(linklist_c &intList)
{
    // cerr << "\nTraversing linked_list. GetSize = " << intList.getsize();

    intList.headend();
    if(intList.getsize()){
        // cerr << "\n" << intList.getval();
        intList.step();
        for(int i = intList.getsize() - 1; i > 0; --i){
            // cerr << ", " << intList.getval();
            intList.step();
        }
    } else {
	// cerr << "\nList is Empty!!\n\n";
    }
    // cerr << "\n" << flush;
}	
