
//  linked_list.h
//  inspired by a (non-copyrighted posting) to comp.lang.c++:
//  A linked-list object:  May 22, 1990, Brian L. Donat
//  Thank you.
//  
//  I added an operator[].
//
//  The following classes compose a very general doubly linked list, 
//  which may be added to, inserted into, deleted from, and read from.
//  This size of list is maintained, allowing external algorithms to
//  perform finite operations on the list data.
//
// If you like or hate it, find bugs, etc, mail to:
// S: Stettener Str. 50, 821 Prien, FRG;
// E: UH311AE@DM0LRZ01.BITNET & uh311ae@sun7.lrz-muenchen.de
//
// Feel free to use it non-commercially, but leave a note. 
// No guarantees !


#ifndef LINKLIST_H
#define LINKLIST_H

//#include <iostream.h>

class linklist_c;						

//  NODE is an abstract class which allows for 
//  the creation of a doubly linked list.

class node_c {

    node_c *next;
    node_c *previous;
    void* data;

    friend linklist_c;
};



// LIST is the actual doubly linked list class
// 
//   listsize allows for finite list operations.
// 
//   head and tail delimit the list.
//   cnode is the current node.
//
//   list() constructs a minimum list with only
//       a head and a tail.  cnode is head to
//       begin, but with listsize > 0, cnode
//       will avoid equality with head or tail.
//       
//       The two functions headend() and 
//       tailend() demonstrate cnode's aversion 
//       for this equality.  cnode will only 
//       assume equality with either if 
//       listsize = 0.  Similar rules apply
//       when cnode is the victim of the 
//       last record deleted w. remove(). 
//
//   One may move forward or backward through
//   a list once it's formed or find instances
//   of given data w. step() and backstep().
//
//   Note how both head and tail are
//   accessed directly for methods find()
//   and add().
//

class linklist_c {                          
private:
    int 	cnode_index;	// maintain an index of the cnode (head=-1, tail=listsize)
    int		listsize;		// speeds up operator[]
    node_c *head;
    node_c *tail;
    node_c *cnode;   // current node or copy node.
public:
    linklist_c();
    ~linklist_c();
    // main access op(cnode_index = index):
    int 		listError;
    void*		operator[](int index);	
    void 		add	(void* E);	// insert before Tail
    void		backstep(void);
    int		 	find	(void* E);	// 1 yup, 0 nope
    int 		getsize	(void);
    void* 	getval	(void);			// get c_node
    void 		headend	(void);		// head->next (?==tail:->head)
    void 		insert	(void* E);	// previous->E->c_node. E= c_node.
    void 		remove	(void);			// next!=tail?c_node=next:cnode=previous
    void 		step	(void);
    void 		tailend	(void);			// tail->previous (?==head:->tail)
};	

void list_traverse(linklist_c &);	// traverse the list, output elements

#endif // LINKLIST_H

