#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

#include "action_forms_aux.h"
#include "browser.h"
#include "cstring.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "global.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "linklist.h"
#include "menu.h"
#include "fork_pipe.h"
#include "pathnames.h"
#include "status_forms_aux.h"

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

int new_makefile_save(const char *fname){
    ofstream ofile;
    if(!fname) return FALSE;
    ofile.open(fname, ios::app);
    if(!ofile) return FALSE;
    ofile << "# Unix Cockpit makefile \n";
    ofile << "# \n";
    ofile << "# The targets in this makefile are used as menu entries in the F2\n";
    ofile << "# browser menu. You can run them with UC's selection as argument\n";
    ofile << "# via the special macro ${UC_FILES}. You can also have UC prompt\n";
    ofile << "# you for arguments (like ${UC_Tarfile}, see below) to be used in\n";
    ofile << "# the make. In this fashion, you can wire most of your custom Unix\n";
    ofile << "# commands into the UC menu structure, and have them, structured and\n";
    ofile << "# with shortcuts, at your fingertips.\n";
    ofile << "# You can also use any normal make target, and edit this file at\n";
    ofile << "# will. After editing, reload the configuration database via global\n";
    ofile << "# options (Ctrl-F2 load), or restart UC.\n";
    ofile << "# Please note that this is a standard makefile that needs TABS in\n";
    ofile << "# the rule lines following the targets. Make a backup copy of the\n";
    ofile << "# makefile before changing it - makefiles are tricky ! If you get\n";
    ofile << "# lost, remove it, and UC will regenerate its default makefile.\n";
    ofile << "# \n";
    ofile << "\n";
    ofile << "# The following fake target results in silent execution. Comment\n";
    ofile << "# out for debugging, or if you want to see the command executed.\n";
    ofile << ".SILENT:\n";
    ofile << "\n";
    ofile << "# A nifty macro to xterm commands (note the -title at the end,\n",
    ofile << "# which needs an argument!). We also switch on the scrollbar.\n";
    ofile << "XTERM=xterm -fg white -bg black -sb -geometry 80x40+300+200 -title \n";
    ofile << "\n";
    ofile << "# This target is a bit special - it allows to load this file from\n";
    ofile << "# within UC's menu into an editor. Note the `#h' shortcut specification\n";
    ofile << "# following the target. It is ignored by make, but used as shortcut in\n";
    ofile << "# the F2 menu:\n";
    ofile << "Help:  #Hh\n";
#ifdef SGI
    ofile << "\t/usr/sbin/jot ${HOME}/.uc/ucmk\n";
#else
    ofile << "\t/usr/bin/X11/xedit ${HOME}/.uc/ucmk\n";
#endif
    ofile << "\n";

#if 0
    // UC has no bugs :-)
    ofile << "ReportBug:  #Bb\n";
#ifdef SGI
    ofile << "\t/usr/sbin/winterm -title 'Send email to henrik@klagges.com (kill window to cancel) ...' -e /usr/sbin/Mail henrik@klagges.com\n";
#else
    ofile << "\t${XTERM} 'Send email to henrik@klagges.com (kill window to cancel) ...' -e mail henrik@klagges.com\n";
#endif
    ofile << "\n";
#endif

    ofile << "# Here are a few complex examples:\n";
    ofile << "# (they are put here because of alphabetical order, easier examples can\n";
    ofile << "# be found below).\n";
    ofile << "# \n";
    ofile << "# Here's a handy macro that will be used below, it's used to make your terminal\n";
    ofile << "# emulator wait for you to press enter:\n";
    ofile << "ENTER=echo \"press Enter to continue\"; read x \n";
    ofile << "\n";
    ofile << "# First note that the `find_file' target contains an underscore (`_').\n";
    ofile << "# This makes UC display it as an entry `file' in a submenu `find'. Next,\n";
    ofile << "# note that two shortcuts are needed, separated by a blank (one for the\n";
    ofile << "# menu and one for the entry).\n";
    ofile << "# `find_file' fires up a xterm emulator with the XTERM macro defined above.\n";
    ofile << "# xterm executes (-e) a bourne shell ('sh') that executes the actual find.\n";
    ofile << "# (A bourne shell is necessary so that the 'Press Enter' at the end works).\n";
    ofile << "# The shell first prints a 'Working ...' status message to the xterm, and\n";
    ofile << "# then uses 'find' to search for a file names containing the UC_Name string\n";
    ofile << "# given by the user. It prints the result in a list, and waits for you to \n";
    ofile << "# press 'Enter'. Try out the command from UC to understand it's execution.\n";
    ofile << "\n";
    ofile << "Find_File:   #Ff Ff\n";
    ofile << "\t${XTERM} \"FindFile (*${UC_Name}*)\" -e sh -c \\\n";
    ofile << "\t'echo \"Working ...\"; find . -name \"*${UC_Name}*\" -print; ${ENTER}' & \n";
    ofile << "\n";
    ofile << "# FindAndGrep is even more complicated. It shows that you can easily interface\n";
    ofile << "# complex commands to UC. ('xargs' allows 'fgrep' to work with any number of\n";
    ofile << "# command line arguments. Adding '/dev/null' fixes a peculiar 'fgrep' bug.)\n";
    ofile << "# FindAndGrep makes a list of all files with a certain name (UC_Name) that\n";
    ofile << "# contain a certain pattern (UC_Pattern). \n";
    ofile << "# \n";
    ofile << "Find_Pattern:  #Ff Pp\n";
    ofile << "\t${XTERM} \"FindFile (*${UC_FileNames}*) and grep for (${UC_Pattern})\" -e sh -c \\\n";
    ofile << "\t'echo \"Working ...\"; find . -name \"*${UC_FileNames}*\" -print | xargs fgrep \"${UC_Pattern}\" /dev/null; ${ENTER}' & \n";

    ofile << "# Here are a few examples that use UC's selection:\n";
    ofile << "# (All are backgrounded with `&', so that you don't have to wait).\n";
    ofile << "Packer_Compress:  #Pp Cc\n";
    ofile << "\tcompress ${UC_FILES} &\n";
    ofile << "\n";
    ofile << "Packer_Uncompress:   #Pp Uu\n";
    ofile << "\tuncompress ${UC_FILES} &\n";    
    ofile << "\n";
    ofile << "# gzip/gunzip must be installed for this to work:\n";
    ofile << "# (note the capitalized Z shortcut for gunzip).\n";
    ofile << "Packer_Gzip:  #Pp Gg\n";
    ofile << "\tgzip ${UC_FILES} &\n";
    ofile << "\n";
    ofile << "Packer_Gzip9:  #Pp 9\n";
    ofile << "\tgzip -9 ${UC_FILES} &\n";
    ofile << "\n";
    ofile << "Packer_gunZip:   #Pp Zz\n";
    ofile << "\tgunzip ${UC_FILES} &\n";    
    ofile << "\n";
    ofile << "# Here is a plain target (note the backgrounding `&' - you don't\n";
#ifndef SGI
    ofile << "# want UC to wait for your xterm to finish !):\n";
    ofile << "Shell_Xterm:   #Ss Xx\n";
    ofile << "\t${XTERM} Hello &\n";
#else
    ofile << "# want UC to wait for your winterm to finish !):\n";
    ofile << "Shell_Hello:    #Ss Hh\n";
    ofile << "\twinterm -title Hello &\n";
#endif
    ofile << "\n";
    ofile << "# This example will prompt for the argument `OutputTarfile' when called\n";
    ofile << "# from UC. It then uses the argument as the name of the tarfile to\n";
    ofile << "# be created. Always use descriptive argument names: It is much nicer\n";
    ofile << "# to be prompted for a `Tarfile' than for a `ARG0'.\n";
    ofile << "Tar_Create:   #Tt Cc\n";
    ofile << "\t${XTERM} \"Creating Tarfile (${UC_OutputTarFile}) from ${UC_FILES}\" -e sh -c \\\n";
    ofile << "\t'tar cvf ${UC_OutputTarFile} ${UC_FILES}; ${ENTER}' & \n";
    ofile << "\n";
    ofile << "Tar_CreateAndGzip:   #Tt Gg\n";
    ofile << "\t${XTERM} \"Creating TGZ-file (${UC_OutputTGZFile}) from ${UC_FILES}\" -e sh -c \\\n";
    ofile << "\t'tar cvf - ${UC_FILES} | gzip - > ${UC_OutputTGZFile}; ${ENTER}' & \n";
    ofile << "\n";
    ofile << "Tar_List:   #Tt Ll\n";
    ofile << "\t${XTERM} \"List of ${UC_FILES}\" -e sh -c \\\n";
    ofile << "\t'tar tf ${UC_FILES}; ${ENTER}' & \n";
    ofile << "\n";
    ofile << "Tar_Extract:   #Tt xX\n";
    ofile << "\t${XTERM} \"Extracting archive ${UC_FILES}\" -e sh -c \\\n";
    ofile << "\t'tar xovf ${UC_FILES}; ${ENTER}' & \n";
    ofile << "\n";
    ofile.close();
    return TRUE;
}

/*** CLASS TO DESCRIBE TARGET & IT'S ARGUMENTS ***/

class target_c {
public:
    char *name, *tail;
    int uc_files_flag;
    linklist_c arg_list;
    target_c(const char *name_iv, const char *tail_iv, int uc_files_flag_iv = FALSE);
    ~target_c();
    int search_arg(const char *arg_name);
};

int target_c::search_arg(const char *arg_name){
    if(!arg_name) return FALSE;
    int size = arg_list.getsize();
    for(int i=0; i<size; i++){
	char *buf = (char *) arg_list[i];
	if(!buf){
	    return FALSE;
	} else {
	    if(strcmp(buf, arg_name) == 0){
		if(dbg_flag) cerr << "search_arg success for: " << arg_name << endl;
		return TRUE;
	    }
	}
    }
    return FALSE;
}

target_c::~target_c(){
    while(arg_list.getsize()){
	char *buf = (char *)arg_list.getval();
	if(buf)fl_free(buf);
	arg_list.remove();
    }
    if(name){
	fl_free(name);
	name = NULL;
    }
}

target_c::target_c(const char *name_iv, const char *tail_iv, int uc_files_flag_iv){
    if(!name_iv || !tail_iv){
	cerr << "NULL name in target_c constructor.\n";
	return;
    } else {
	name          = strdup(name_iv);
	tail          = strdup(tail_iv);
	uc_files_flag = uc_files_flag_iv;
    }
}

/*** ASSOCIATIVE MAP FOR MENU HIERARCHY ***/
// long menu ID, char * name. Use full pathname menu_sub_sub1 for name.

class entry_c {
public:
    char *name;
    long id;
    entry_c () { name = 0; id = 0; }
    ~entry_c() { if(name){fl_free(name); name = NULL;}}
};

class map_c {
public:
    linklist_c menu_list;
    long search(const char *menu_name);   // 0 if not found.
    void insert(const char *menu_name, long id);
    void clear();
};

long map_c::search(const char *menu_name){
    if(!menu_name) return 0;
    int size = menu_list.getsize();
    for(int i=0; i<size; i++){
	entry_c *ebuf = (entry_c *) menu_list[i];
	if(!ebuf) return 0;
	if(!ebuf->name) return 0;
	if(strcmp(ebuf->name, menu_name) == 0){
	    if(dbg_flag) cerr << "search_menu success for: " << menu_name << endl;
	    return ebuf->id;
	}
    }
    return 0;
}

void map_c::insert(const char *menu_name, long id){
    if(!menu_name) return;
    if(search(menu_name) > 0){
	cerr << "cannot insert, menu exists.\n";
	return;
    }
    entry_c *ebuf = new entry_c;
    ebuf->name = strdup(menu_name);
    ebuf->id   = id;
    menu_list.add(ebuf);
}
    
void map_c::clear(){
    while(menu_list.getsize()){
	entry_c *ebuf = (entry_c *)menu_list.getval();
	if(ebuf) delete ebuf;
	menu_list.remove();
    }
}

/*** VARS ***/

static int menu_made_flag = FALSE;
static long makefile_menu = 0;
static linklist_c target_list;
static int uc_make_the_target(void);
static target_c *ucmake_target;
static map_c menu_map;

static void ucmake_cb(int which){
    if(dbg_flag) cerr << "ucmake_cb: " << which << "\n";
    if(which >= 0 && which < target_list.getsize()){
	ucmake_target = (target_c *) target_list[which];
	if(ucmake_target){
	    uc_make_the_target();
	} else {
	    cerr << "invalid NULL target at index: " << which << endl;
	}
    } else {
	cerr << "invalid index into target list: " << which << endl;
    }
}

static char ucmake_fname[MAXPATHLEN+1];

static void empty_menu_map_list(void){
    menu_made_flag = FALSE;
    //M_err("empty", "%d", makefile_menu);
    //cerr << hex << makefile_menu << endl;
    free_pup(makefile_menu);
    while(target_list.getsize()){
	target_c *buf = (target_c *)target_list.getval();
	if(buf) delete buf;
	target_list.remove();
    }
    menu_map.clear();
}

static int makefile_open(ifstream &ifile){
    if(tool_ldefault_dir){
	sprintf(ucmake_fname, "%s/ucmk", tool_ldefault_dir);
    } else {
	cerr << "UC: No tool default dir.\n";
	return FALSE;
    }
    ifile.open(ucmake_fname);
    if(!ifile){
	// no makefile in home. Attempt global.
	sprintf(ucmake_fname, "%s/ucmk", tool_gdefault_dir);
	ifile.open(ucmake_fname);
	if(!ifile){
	    // no global. Make local.
	    sprintf(ucmake_fname, "%s/ucmk", tool_ldefault_dir);
	    if(new_makefile_save(ucmake_fname)){
		// re-attempt local.
		ifile.open(ucmake_fname);
		if(!ifile) return FALSE;	    
	    } else {
		return FALSE;
	    }
	}
    }
    return TRUE;
}

static void process_taction_line(const char *mbuf, target_c *current_target){
    if(strstr(mbuf, "${UC_FILES}")){
	if(dbg_flag) cerr << "found UC_FILES in context for " << current_target->name << endl;
	current_target->uc_files_flag = TRUE;
    }
    const char *where = strstr(mbuf, "${UC_");
    while(where){
	if(dbg_flag) cerr << "found UC_ in context for " << current_target->name << endl;
	char *right_par = strchr(where, '}');
	if(!right_par){
	    cerr << "Malformed line in ucmk: No `}' for UC_\n";
	    where = NULL;
	} else {
	    *right_par = '\0';
	    where+=strlen("${UC_");
	    if(dbg_flag) cerr << "Argname: UC_" << where << endl;
	    if(strcmp(where, "FILES") == 0 ||
	       strcmp(where, "SRC_DIR") == 0 ||
	       strcmp(where, "TGT_DIR") == 0
	       ){
		if(dbg_flag) cerr << "arg is UC_FILES. No action.\n";
	    } else {
		if(!current_target->search_arg(where)){
		    if(dbg_flag) cerr << "adding it.\n";
		    current_target->arg_list.add((void *) strdup(where));
		} else {
		    if(dbg_flag) cerr << "not adding it, we have it already.\n";
		}
	    }
	    where = strstr(right_par+1, "${UC_");
	}
    }
}

#if defined(SGI_IRIX4) || defined(HP) || defined(SUN) || defined(LINUX) || defined(DEC_ALPHA) || defined(AIX) || defined(GENERICBSD) || defined(SINIX)
// incomplete, uses only first separator.
char *strtok_r(char *buf, const char *sep, char **lasts){
  if(!sep || !lasts)return NULL;
  if(*sep == '\0')return NULL;
  char *ztmp;
  if(buf){
    // strip leading stuff.
    while(*buf == *sep){
      buf++;
    }
    ztmp = buf;
    while(*ztmp != *sep && *ztmp != '\0'){
      ztmp++;
    }
    if(ztmp == '\0'){
      // string exhausted.
      return NULL;
    } else {
      *ztmp = '\0';   // terminate token.
      *lasts = ztmp;  // remember the location.
      return buf;
    }
  } else {
    // check lasts. I know lasts+1 should be still valid.
    char *buf_tmp = *lasts + 1;
    // strip leading stuff.
    while(*buf_tmp == *sep){
      buf_tmp++;
    }
    ztmp = buf_tmp;
    while(*ztmp != *sep && *ztmp != '\0'){
      ztmp++;
    }
    if(ztmp == '\0'){
      // string exhausted.
      return NULL;
    } else {
      *ztmp = '\0';   // terminate token.
      *lasts = ztmp;  // remember the location.
      return buf_tmp;
    }
  }
}
#endif

static target_c* insert_target(char *tail_buf, const char *obj_tok){

    long parent_menu = makefile_menu;
    char obj_buf[1024], menu_buf[1024], shortcuts[1024], tmp_buf[1024];
    const char *name_tok = obj_tok;
    
    if(dbg_flag)cerr << "insert target: processing " << obj_tok << "\n";

    strcpy(shortcuts, "");
    char *null_p = NULL;
    char **lasts = &null_p;
    char *shortcut_tok = strtok_r(tail_buf, " ", lasts);
    if(shortcut_tok){
	if(dbg_flag)cerr << "got shortcuttok: `" << shortcut_tok << "'\n";
	if(shortcut_tok[0] == '#'){
	    strcpy(shortcuts, shortcut_tok+1);
	}
    }
	
    strcpy(obj_buf, obj_tok);
    const char *last_uscore = strrchr(obj_buf, '_');
    if(last_uscore && obj_buf[0] != '_' && !strstr(obj_buf, "__")){
	// no pathologic cases, please.
	char *tok = strtok(obj_buf, "_");
	strcpy(menu_buf, "");
	parent_menu = makefile_menu;
	int first = TRUE;
	while(tok){
	    if(tok < last_uscore){
		// still a menu.
		if(first){
		    first = FALSE;
		} else {
		    strcat(menu_buf, "_");
		}
		strcat(menu_buf, tok);
		if(dbg_flag)cerr << "menubuf: " << menu_buf << endl;
		long menu = menu_map.search(menu_buf);
		if(menu){
		    parent_menu = menu;
		    if(dbg_flag)cerr << "found parent.\n";
		} else {
		    if(dbg_flag)cerr << "not found, creating.\n";
		    menu = new_pup();
		    if(menu){
			// title
			sprintf(tmp_buf, "%s%%t", tok);
			add_to_pup(menu, tmp_buf, 0);
			
			// link into parent !
			sprintf(tmp_buf, "%s%%m", tok);
			add_to_pup(parent_menu, tmp_buf, menu, shortcuts);

			menu_map.insert(menu_buf, menu);
			parent_menu = menu;  // remember
		    } else {
			cerr << "cannot create more menues.\n";
			return NULL;
		    }
		}
	    } else {
		// not a menu, it's the name component.
		name_tok = tok;
	    }
	    tok = strtok(0, "_");
	    if(tok){
		shortcut_tok = strtok_r(0, " ", lasts);
		if(!shortcut_tok){
		    strcpy(shortcuts, "");
		    if(dbg_flag)cerr << "no extra shortcuttok, resetting.\n";
		} else {
		    if(dbg_flag)cerr << "got extra shortcuttok: `" << shortcut_tok << "'\n";
		    strcpy(shortcuts, shortcut_tok);
		}
	    }
	}
    }
    
    if(dbg_flag)cerr << "name: " << name_tok << "\n";

    sprintf(tmp_buf, "%s%%x%d", name_tok, target_list.getsize());
    add_to_pup(parent_menu, tmp_buf, (long) ucmake_cb, shortcuts);
    target_c *tmp_target = new target_c(obj_tok, name_tok);
    target_list.add(tmp_target);
    return tmp_target;
}

int makefile_load(void){
    M_info("makefile_load", "start");
    if(menu_made_flag){
	empty_menu_map_list();
    }
    
    // input file stream operations. If none, make one, reopen.
    ifstream ifile;
    if(!makefile_open(ifile)){
	return FALSE;
    }

    // construct menu
    const int bufsize = 2*MAXPATHLEN;
    char mbuf[bufsize];
    makefile_menu = new_pup();
    if(!makefile_menu){
	return FALSE;
    } else {
	menu_made_flag = TRUE;
    }

    add_to_pup(makefile_menu, "Make (Alt-k, Esc-k)%t", 0);

    int in_target_block_flag = FALSE;
    target_c *current_target = NULL;
    
    while(ifile.getline(mbuf, bufsize) ){
	if(mbuf[0] == '#'){
	    continue;
	    // it is a comment
	}
	if(mbuf[0] == '\t' && in_target_block_flag && current_target){
	    process_taction_line(mbuf, current_target);
	    continue;
	}
	   
	if(mbuf[0] == '.' || mbuf[0] == ' ' || mbuf[0] == '\0'){
	    in_target_block_flag = FALSE;
	    continue;
	    // it is a special rule, or empty line, or starts with a blank.
	}
	if(!strchr(mbuf, ':')){
	    in_target_block_flag = FALSE;
	    continue;
	    // it is not a target
	}

	// it is target, look for ID token.
	char *obj_tok = strtok(mbuf, ":");
	if(!obj_tok){
	    cerr << "UC: strange line ucmk: empty target.\n";
	    continue;
	}
	current_target = insert_target(obj_tok + strlen(obj_tok) + 1, obj_tok);
	if(current_target){
	    in_target_block_flag = TRUE;
	} else {
	    cerr << "insert target failed.\n";
	}
    }
    ifile.close();
    if(target_list.getsize() <= 0){
	char buf[1024];
	sprintf(buf, "UC makefile (%s) contains no targets !", ucmake_fname);
	add_to_pup(makefile_menu, "<empty !> %x-1", (long) ucmake_cb, "");
	timed_error_notifier(buf, "Make submenu will be non-functional.");
	return FALSE;
    } else {
	return TRUE;
    }
}

long get_makefile_menu(void){
    if(!menu_made_flag){
	if(!makefile_load()){
	    return 0;
	}
    }
    return makefile_menu;
}

static int make_single_many_files(const char *fname, const char *source_dir){
    const int nCmdLenMax = 5120;
    int i;
    if(!source_dir || !make_path) return FALSE;
    CString strCmd = "make -f ";
    int n_target_args = ucmake_target->arg_list.getsize();
    
    strCmd += ucmake_fname;
    strCmd += " ";

    const char *dir = source_browser->get_cwd();
    if(dir){
	strCmd += " \"UC_SRC_DIR=";
	strCmd += dir;		    
	strCmd += "\" ";
    }
    dir = target_browser->get_cwd();
    if(dir){
	strCmd += " \"UC_TGT_DIR=";
	strCmd += dir;		    
	strCmd += "\" ";
    }
    strCmd += " \"UC_FILES=";
    if(fname){
	if(strchr(fname, '#')){
	    timed_notifier("Sorry, make module can't process filenames with the `#'-character.");
	    return FALSE;
	}
	strCmd += fname;		    
    } else {
	source_browser->freeze();
	const char *name_buf = source_browser->get_first_selection();
	while(name_buf){
	    if(strchr(name_buf, '#')){
		timed_notifier("Sorry, make module can't process filenames with the `#'-character.");
		return FALSE;
	    }
	    if(strlen(name_buf) + 64 + strCmd.length() < nCmdLenMax){
		strCmd += name_buf;
		strCmd += " ";
		source_browser->deselect_file(name_buf);
		name_buf = source_browser->get_next_selection();
	    } else {
		name_buf = NULL;
		status_msg(1, "Selection partially transferred to `make'.");
		timed_notifier(
		    "Unix command line limit:", 
		    "Selection too long: only partially to `make'."
		);
	    }
	}
	source_browser->unfreeze();
    }
    // finish up UC_FILES
    strCmd += "\" ";


    for(i=0; i<n_target_args; i++){
	char buffer[2048];
	sprintf(buffer, "Argument `%s' required for making `%s':", 
	    (char *) ucmake_target->arg_list[i], 
	    ucmake_target->name
	);
	const char *result = OKCancel(
	    buffer, "", 
	    arg_required_help_text, 
	    "Argument Required", 
	    TRUE
	);
	if(result){
	    strCmd += " \"";
	    strCmd += "UC_";
	    strCmd += (char *) ucmake_target->arg_list[i]; 
	    strCmd += "=";
	    strCmd += result;
	    strCmd += "\" ";
	} else {
	    OKCancelClose();
	    return FALSE;
	}
    }
    OKCancelClose();
    strCmd += ucmake_target->name;

    //strCmd += " >/dev/null";

    if(dbg_flag){
	cerr << "make cmd line: " << strCmd << endl;
    }
	
    safe_system(strCmd);
    return TRUE;
}

static int make_many_files(const char *source_dir){
    return make_single_many_files(NULL, source_dir);
}

static int make_single_file(const char *fname, const char *source_dir){
    return make_single_many_files(fname, source_dir);
}

static int uc_make_the_target(void){
    if(!ucmake_target) return FALSE;
    if(dbg_flag) cerr << "would execute make target: " << ucmake_target->name;
    if(dbg_flag) cerr << ", n_args: " << ucmake_target->arg_list.getsize() << endl;
    char tbuf[256];
    sprintf(tbuf, "Make `%s'", ucmake_target->tail);
    if(ucmake_target->uc_files_flag){
	strcat(tbuf, " on");
    }
    source_dir_action(make_single_file, make_many_files, tbuf, TRUE, TRUE, make_help_text, TRUE, !ucmake_target->uc_files_flag);
    return TRUE;
}


