#include <ctype.h>
#include <iostream.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

//#include "client_msg.h"
#include "flspecial.h"
#include "forms.h"
#include "gl_utility.h"
#include "global.h"
#include "menu.h"

typedef void (*cb_fpt)(int);

//static int dbg_flag = TRUE;
static int dbg_flag = tool_debug_flag;

// left over after closing of a submenu. 
static XEvent *parent_event = NULL;  // also used as flag (if NULL);
static XEvent parent_event_data;

static void pup_ungrab(void);
static void set_timer(int secs);
void PupTimerCB(FL_OBJECT *, long);

/*** LINE_C ***/

class line_c {
public:
    FL_OBJECT *line;
    static const int align_margin;
    static int text_box_h, lstyle, lsize;
    char *shortcuts, *string;
    cb_fpt cb_fp;
    int menu_flag, // are we a menu ?
	val_flag   // are we a val ?
    ;
    long val;
    
    line_c(void);
    ~line_c(void);
    
    int get_height(void);
    int get_width(void);
    void make(int form_w, int start_h);
    int check_shortcut(char c);
};

const int
    line_c::align_margin = 7  // used slightly unclean. better: text and form margins.
;    

int line_c::lstyle = FL_TIMES_STYLE;
int line_c::lsize = FL_NORMAL_SIZE;
int line_c::text_box_h = (int) ((double) line_c::lsize * 1.8);

int
line_c::check_shortcut(char c){
    if(!isprint(c))return FALSE;
    if(shortcuts){
	return strchr(shortcuts, c) != NULL;
    } else {
	return FALSE;
    }
}

line_c::line_c(void){
    line = NULL;
    shortcuts = string = NULL;
    cb_fp = NULL;
    menu_flag = FALSE;
    val_flag = FALSE;
    val = 0;
}

line_c::~line_c(void){
    if(shortcuts){
	fl_free(shortcuts);
	shortcuts = NULL;
    }
    if(string){
	fl_free(string);
	string = NULL;
    }
}

int line_c::get_height(void){
    // maybe a special line (a separator) will have a different height
    if(line){
        return line->h;
    } else {
        return text_box_h;  
    }
}

int line_c::get_width(void){
    if(!string) return 0;
    char *real_text = string;
    if(real_text[0] == '@'){
	real_text++;
	if(real_text[0] != '\0'){
	    real_text++;
	}
    }
    int tmp_w = fl_get_string_width(lstyle, lsize, real_text, strlen(real_text));
    // adjust for left/right margins
    tmp_w += 2 * align_margin;
    // adjust for menu, etc.
    return tmp_w;
}

void line_c::make(int form_w, int start_h){
    FL_OBJECT *obj = line = fl_add_text(FL_NORMAL_TEXT,0,start_h,form_w,text_box_h,string);
    fl_set_object_align(obj,FL_ALIGN_LEFT);
    fl_set_object_boxtype(obj, FL_NO_BOX);
    fl_set_object_lsize(obj,lsize);
    fl_set_object_lstyle(obj,lstyle);
    if(shortcuts){
	obj->active = TRUE;
	fl_set_object_shortcut(obj, shortcuts, TRUE);
	obj->active = FALSE;
    }
    // make fl_find_object work:
    obj->active = TRUE;
    // menu
    if(menu_flag){
	int arrow_w = 20, arrow_h = text_box_h;
	obj = fl_add_text(FL_NORMAL_TEXT,form_w-arrow_w,start_h,arrow_w,arrow_h,"@>");
	fl_set_object_align(obj,FL_ALIGN_LEFT);
	fl_set_object_boxtype(obj, FL_NO_BOX);
	fl_set_object_lsize(obj,lsize);
	fl_set_object_lstyle(obj,lstyle);
    }
}

/*** PUP_C ***/

static const int maxlines = 50;
static cb_fpt exit_func = 0;  // if non-Null, we have to call a function on exit, else ret a val.
static long exit_val = -1;
static int real_exit_flag = 0;
static const int menu_margin = 10;

const int
    title_gap_h = 5     // between title and options. 
;

int title_box_h = 10 + (int) (1.2 * (double) line_c::lsize);   // for the title.

class pup_c {
private:
    static int timeout_secs;
    int lower_box_h;
    line_c lines[maxlines+1];
    int maxline,
        menu_flag,
        active_line
    ;
    char *title;  // NULL if none;
    
    FL_FORM *PupForm;
    FL_OBJECT *TitleText, *ActiveBox, *TitleBox;
    static FL_OBJECT *Timer; 

    int check_title(const char *str);  // true if is title and register.
    int check_cb   (const char *str, long arg, const char *shortcuts);  // true if cb, and register func.
    int check_menu (const char *str, long arg, const char *shortcuts);  // true if menu, and register m.
    int check_val  (const char *str, long arg, const char *shortcuts);  // true if val, and register val.

    void create       (void);
    void close        (void);
    void show         (void);

    void down_arrow_cb(void);
    void up_arrow_cb  (void);
    void end_cb       (void);
    void home_cb      (void);

    void do_line      (int line);
    void draw_active  (void);
    void set_active   (int line);
    int set_active    (int mx, int my, int menu_always_flag = FALSE); 
    // return TRUE if a submenu was fired up, otherwise false. strange ret val.

    int key_event_cb(XEvent *xev);
    int mouse_event_cb(XEvent *xev, FL_Coord dx, FL_Coord dy);  // correction factors.
    
    int mouse_outside(int x, int y);
    int mouse_over_line(int line, int real_x, int real_y);
    int find_line(int real_x, int real_y);
public:
    int my_stack_pos;
    inline int get_active_line(void){return active_line;}
    inline Window get_window(void){return PupForm->window;}
    pup_c(void);
    ~pup_c();
    void add(const char *str, long arg, const char *shortcuts = NULL);
    void close_all    (void);
    void set_timer(int secs);
    long do_pup(void);
    int event_cb(FL_FORM *form, XEvent *xev);
    //  int handle_free(FL_OBJECT *obj, int event, int mx, int my, char key);
};

FL_OBJECT *pup_c::Timer = NULL;
int pup_c::timeout_secs = 20;  // was 20.

/*** STACK ***/

static const int stack_maxsize = 20;
static int stackpos = -1;
static pup_c *pup_stack[stack_maxsize];

static void stack_push(pup_c *pup){
    stackpos++;
    if(stackpos >= stack_maxsize){
	cerr << "popup stack too big !\n";
	exit(-1);
    }
    pup_stack[stackpos] = pup;
    pup->my_stack_pos = stackpos;
}

static void stack_pop(void){
    if(stackpos < 0){
	cerr << "popup stack underflow !\n";
	exit(-1);
    }
    stackpos--;
}

/*** PUP_C METHODS ***/
int
pup_c::key_event_cb(XEvent *xev){
    if(!xev) return FALSE;
    int i;
    // init.
    if(xev->type != KeyPress) return FALSE;

    KeySym keysym;
    char keybuf[1];
    XKeyEvent *xkey= (XKeyEvent *) xev;
    XLookupString(xkey, keybuf, sizeof(keybuf), &keysym, 0);

    if(controlkey_down(xkey->state)){
	switch(keysym){
	case XK_a: case XK_A: home_cb      (); return TRUE;
	case XK_b: case XK_B: keysym = XK_Left; break;        // trigger prev_menu.
	case XK_e: case XK_E: end_cb       (); return TRUE;
	case XK_f: case XK_F: keysym = XK_Right; break;       // trigger next menu
	case XK_n: case XK_N: down_arrow_cb(); return TRUE;
	case XK_p: case XK_P: up_arrow_cb  (); return TRUE;
	case XK_v: case XK_V: end_cb       (); return TRUE;
	default:                               return FALSE;
	}
    }
    if(metakey_down(xkey->state)){
	switch(keysym){
	case XK_less:         home_cb   (); return TRUE;
	case XK_greater:      end_cb    (); return TRUE;
	case XK_v: case XK_V: home_cb   (); return TRUE;
	default:                            return FALSE;
	}
    }

    if(is_end(keysym, xkey)){
	end_cb();
	return TRUE;
    }
    if(is_down(keysym, xkey)){
	down_arrow_cb();
	return TRUE;
    }
    if(is_page_down(keysym, xkey)){
	end_cb();
	return TRUE;
    }
    if(is_left(keysym, xkey)){
	exit_val = -1;
	exit_func= 0;
	close();
	return TRUE;
    }
    if(is_right(keysym, xkey)){
	// search for next menu below us. if fail, search for next menu above us.
	i=active_line;
	if(i<0) i=0;
	while(i<=maxline){  // search down
	    if(lines[i].menu_flag){
		set_active(i);
		do_line(i);
		return TRUE;
	    }
	    i++;
	}
	i = active_line-1;
	while(i>=0){        // search up
	    if(lines[i].menu_flag){
		set_active(i);
		do_line(i);
		return TRUE;
	    }
	    i--;
	}
	return TRUE;
    }
    if(is_home(keysym, xkey)){
	home_cb(); 
	return TRUE;
    }
    if(is_up(keysym, xkey)){
	up_arrow_cb();
	return TRUE;
    }
    if(is_page_up(keysym, xkey)){
	home_cb();
	return TRUE;
    }

    switch(keysym){
    case XK_Escape:
	if(dbg_flag) cerr << "escape.\n";
	exit_val = -1;
	exit_func= 0;
	real_exit_flag = TRUE;
	close();
	return TRUE;	    
#ifndef SUN
    case XK_KP_Enter:	
#endif
    case XK_Return:
	// enter ?
	if(active_line >=0 && active_line <= maxline){
	    do_line(active_line);
	}
	return TRUE;
    default:
	// shortcut ?
	for(i=0; i<=maxline; i++){
	    if(lines[i].check_shortcut(keybuf[0])){
		set_active(i);
		do_line(i);
		return TRUE;
	    }
	}
	return FALSE;
    }
}

void
pup_c::do_line(int line){
    if(line < 0 || line > maxline){
	close_all();
	return;
    }
    
    if(lines[line].menu_flag){
	if(dbg_flag) cerr << "do submenu\n";
	if(lines[line].cb_fp){
	    if(dbg_flag) cerr << "calling menu ...";
	    exit_val = ((pup_c *)(lines[line].cb_fp))->do_pup();
	    if(exit_val == -1 && !real_exit_flag){
		if(dbg_flag) cerr << "continuing, no selection was made.\n";
		if(parent_event){
		    // possible exit event ?
		    event_cb(pup_stack[0]->PupForm, parent_event);
		}
	    } else {
		if(dbg_flag) cerr << "subselection was made, done.\n";
		close();
	    }
	} else {
	    if(dbg_flag) cerr << "cannot execute a null ptr submenu.\n";
	}
    } else {
	if(lines[line].cb_fp){
	    if(dbg_flag) cerr << "registering an execute ... \n";
	    exit_func= lines[line].cb_fp;
	} else {
	    if(dbg_flag) cerr << "cannot execute a null ptr, returning line.\n";
	    exit_func= NULL;
	}
	if(lines[line].val_flag){
	    exit_val = lines[line].val;
	} else {
	    exit_val = line+1;
	}
	real_exit_flag = TRUE;
	close();
    }
    // execute the callback ! AFTER hiding.
    // don\'t close if a sub menu
    // close();
}    

void
pup_c::set_active(int line){
    if(!PupForm){
	cerr << "PupForm not valid anymore for set_active.\n";
	return;
    }
    if(line!=active_line){
	active_line = line;
	draw_active();
    }
}

int pup_c::mouse_over_line(int line, int real_x, int real_y){
    if(!PupForm) return FALSE;
    int x = real_x - PupForm->x,
        y = real_y - PupForm->y
    ;
    FL_OBJECT *mobj = fl_find_last(PupForm, FL_FIND_MOUSE, x, y);
    return (lines[line].line == mobj);
}

int pup_c::find_line(int real_x, int real_y){
    if(!PupForm) return FALSE;
    FL_OBJECT *mobj = fl_find_last(PupForm, FL_FIND_MOUSE, real_x, real_y);
    if (!mobj) return -1;
    for(int i=0; i<=maxline; i++){
        if (lines[i].line == mobj) return i;
    }
    return -1;
}
    
int
pup_c::set_active(int mx, int my, int menu_always_flag){
    if(!PupForm){
	cerr << "PupForm not valid anymore for set_active (mouse).\n";
	return FALSE;
    }
    int line = find_line(mx, my);
    if (line >= 0) {
        set_active(line);
	if (lines[line].menu_flag && 
	    (mx > PupForm->w - 5 - menu_margin || menu_always_flag)
        ){
	    do_line(line);
	    return TRUE;
	} else {
	    return FALSE;
	}
    }
    set_active(-1);
    return FALSE;
}

int
pup_c::event_cb(FL_FORM *form, XEvent *xev){
    parent_event = NULL;
    FL_Coord dx = 0, dy = 0;
    if(form != PupForm){
	if(dbg_flag) cerr << "different window - searching stack...";
	dx = form->x - PupForm->x;
	dy = form->y - PupForm->y;
	if(dbg_flag) cerr << "correcting: dx: " << dx << ", dy: " << dy << endl;
    }
    if(!xev || !form)return FALSE;
    if(mouse_event_cb(xev,dx,dy)){
	set_timer(timeout_secs);
	return TRUE;
    }
    if(key_event_cb(xev)){
	set_timer(timeout_secs);
	return TRUE;
    }
    return FALSE;
}

void
pup_c::close(void){
    if(PupForm->visible){
	fl_hide_form(PupForm);
	if(dbg_flag) cerr << "hidden\n";
    } else {
	// cerr << "already hidden !\n";
    }
}

int pup_raw_cb(FL_FORM *form, void *xev){
    pup_c *pup = pup_stack[stackpos];
    if(pup){
	return pup->event_cb(form, (XEvent *)xev);
    } else {
	cerr << "Invalid pup for form event cb.\n";
	if(!form){
	    cerr << "form is zero.\n";
	}
	return 0;
    }
}

int pup_c::mouse_outside(int x, int y){
    if(x < 0 || x >= PupForm->w ||
       y < 0 || y >= PupForm->h
    ){
	return TRUE;
    } else {
	return FALSE;
    }
}

void pup_c::close_all(void){
    real_exit_flag = TRUE;
    exit_val = -1;
    exit_func= 0;
    close();
}

int pup_c::mouse_event_cb(XEvent *xev, FL_Coord dx, FL_Coord dy){
    static int oldx=0, oldy=0;
    int x = dx, y = dy;
    switch(xev->type){
    case ButtonPress:
	x += xev->xbutton.x;
	y += xev->xbutton.y;
	if(dbg_flag) cerr << "push mouse at " << x << "," << y << endl;
	return TRUE;
    case ButtonRelease:
	x += xev->xbutton.x;
	y += xev->xbutton.y;
	if(dbg_flag) cerr << "release mouse at " << x << "," << y << endl;
	if(mouse_outside(x,y)){
	    if(my_stack_pos >0){
			pup_c *parent = pup_stack[my_stack_pos-1];
			if(parent->mouse_over_line(parent->get_active_line(), x + PupForm->x, y+PupForm->y)){
			   	if(dbg_flag) cerr << "release exception: don't close: we're on parent line.\n";
				return TRUE;
			} 
		}
	    exit_val = -1;
	    exit_func= 0;
	    close();
	    parent_event_data = (*xev);
	    parent_event = &parent_event_data;
	} else {
	    if(!set_active(x, y)){
			do_line(active_line);
	    }
	}
	// release
	return TRUE;
    case MotionNotify:
	x += xev->xmotion.x;
	y += xev->xmotion.y;
	// get the mouse pos
	if(oldx != x || oldy != y){
	    if(dbg_flag) cerr << "position mouse at " << x << "," << y << endl;
	} else {
	    if(dbg_flag) cerr << "mouse unchanged.\n";
	    // no change.
	    //return FALSE;
	}
	oldx = x;
	oldy = y;
//	if(fl_keymask & Button1Mask){
	if(xev->xmotion.state & Button1Mask){	
	    if(dbg_flag) cerr << "mouse button 1 is down" << endl;
	    if(!set_active(x, y, TRUE)){
		if(mouse_outside(x,y)){
		    // potentially, check wether we're in the parent.
		    if(my_stack_pos >0){
				// have parent. Is the mouse still on its active line ?
				pup_c *parent = pup_stack[my_stack_pos-1];
				if(parent->mouse_over_line(parent->get_active_line(), x + PupForm->x, y+PupForm->y)){
			    	if(dbg_flag) cerr << "exception: don't close: we're on parent line.\n";
				} else {
			    	if(dbg_flag) cerr << "should do close: we're not on parent line.\n";
			    	exit_val = -1;
			    	exit_func= 0;
			    	close();
			    	parent_event_data = (*xev);
			    	parent_event = &parent_event_data;
				}
		    } else {
				if(dbg_flag) cerr << "have no parent, just hang on.\n";
		    }
		}
	    } else {
		if(dbg_flag) cerr << "mouse button is up, no motion action." << endl;
	    }
	}	    
	return TRUE;
    default:
	if(dbg_flag) cerr << "default menu event.\n";
	return FALSE;
    }
}

pup_c::pup_c(void){
    PupForm   = NULL;
    TitleBox  = NULL;
    TitleText = NULL;
    ActiveBox = NULL;
    title = NULL;      // no title
    maxline = -1;      // no line
    menu_flag = FALSE; // no menues yet.
    active_line = -1;    // no by default (try)
    //active_line = 0;   // first by default
    my_stack_pos = 0;  // not built-in yet. 
}

pup_c::~pup_c(void){
    // note: it's legal to have a pup with no form,
    //       if it wasn't shown/created.
    if(Timer && Timer->form == PupForm){
	M_warn("~pup_c","Deleting form containing pup timer.");
	Timer = NULL;
    }
    if(title){
	fl_free(title);
	title = NULL;
    }
    if(PupForm){
	if(PupForm->visible){
	    fl_hide_form(PupForm);
	}
	fl_free_form(PupForm);
	PupForm = NULL;
    }
}

int
pup_c::check_title(const char *str){
    if(!str) return FALSE;
    if(strstr(str, "%t")){
	title = strdup(str);
	char *where = strstr(title, "%t");
	if(where){
	    *where = '\0';                   // don\'t want the %t.
	}
	return TRUE;
    } else {
	return FALSE;
    }
}

int
pup_c::check_cb(const char *str, long arg, const char *shortcuts){
    if(!str) return FALSE;
    if(strstr(str, "%f")){
	if(maxline == maxlines-1){
	    cerr << "popup: cannot add more entries.\n";
	    return TRUE;
	} else {
	    maxline++;
	}
	lines[maxline].string    = strdup(str);
	char *where = strstr(lines[maxline].string, "%f");
	if(where){
	    *where = '\0';                   // don\'t want the %f.
	}
	if(shortcuts){
	    lines[maxline].shortcuts = strdup(shortcuts);
	}
	lines[maxline].cb_fp = (cb_fpt) arg;
	return TRUE;
    } else {
	return FALSE;
    }
}

int
pup_c::check_menu(const char *str, long arg, const char *shortcuts){
    if(!str) return FALSE;
    if(strstr(str, "%m")){
	if(maxline == maxlines-1){
	    cerr << "popup: cannot add more entries.\n";
	    return TRUE;
	} else {
	    maxline++;
	}
	lines[maxline].string    = strdup(str);
	char *where = strstr(lines[maxline].string, "%m");
	if(where){
	    *where = '\0';                   // don\'t want the %m.
	}
	if(shortcuts){
	    lines[maxline].shortcuts = strdup(shortcuts);
	}
	lines[maxline].cb_fp = (cb_fpt) arg;
	lines[maxline].menu_flag = TRUE;
	menu_flag = TRUE;
	return TRUE;
    } else {
	return FALSE;
    }
}

int
pup_c::check_val(const char *str, long arg, const char *shortcuts){
    if(!str) return FALSE;
    if(strstr(str, "%x")){
	if(maxline == maxlines-1){
	    cerr << "popup: cannot add more entries.\n";
	    return TRUE;
	} else {
	    maxline++;
	}
	lines[maxline].string    = strdup(str);
	char *where = strstr(lines[maxline].string, "%x");
	if(where){
	    *where = '\0';                   // don\'t want the %m.
	    where += 2;    // go past the %x.
	    if(*where != '\0'){
		lines[maxline].val = atol(where);
	    }
	}
	if(shortcuts){
	    lines[maxline].shortcuts = strdup(shortcuts);
	}
	lines[maxline].cb_fp = (cb_fpt) arg;
	lines[maxline].val_flag = TRUE;
	return TRUE;
    } else {
	return FALSE;
    }
}

// support %t title and %f, callbacks (void (*)(int)), and values %xn.
// more menus %m done. potentially text lines, and dividers %l and

void
pup_c::add(const char *str, long arg, const char *shortcuts){
    if(!str) return;
    if(PupForm){
	cerr << "menu already created. Cannot add more entries.\n";
	return;
    }
    if(check_title(str)) return;
    if(check_cb   (str, arg, shortcuts)) return;
    if(check_menu (str, arg, shortcuts)) return;
    if(check_val  (str, arg, shortcuts)) return;
    cerr << "pup_c::not recognized line: " << str << endl;
}

long
pup_c::do_pup(void){
    if(maxline < 0) return -1;   // straight away.
    stack_push(this);
    if(stackpos == 0){
	real_exit_flag = FALSE;  // reset. global flags. 
	parent_event = NULL;
	focus_save();
    }
    exit_func = 0;
    create();
    show();
    if(stackpos == 0){
	focus_move();
    }
    while(PupForm->visible){
	tool_mung_events();
    }
    
    stack_pop();
    if(dbg_flag) cerr << "in do_pup, after munging events. exit_func: " << (long) exit_func << ", val: " << exit_val << "\n";

    if(stackpos == -1){
	// call last only, after hiding the hierarchy.
	pup_ungrab();
	fl_activate_all_forms();
	if(dbg_flag) cerr << "last in do_pup.\n";
	focus_restore(TRUE);
	if(exit_func){
	    if(dbg_flag) cerr << "calling final cb.\n";
	    (*exit_func)((int) exit_val);
	}
    }
    return exit_val;
}

void
pup_c::create(void){
	int i=0;
    if(PupForm){
	// reset active box to 0.
	set_active(0);
	return;
    }
    if(!fl_display){
	M_err("pup_c::create","fl not initialized");
	exit(-1);
    }
    lower_box_h = 10; // for the upper and lower 5 margins.

    // determine size:
    int form_w = 0, form_h = 0, tmp_w;

    for(i=0; i<=maxline; i++){
	lower_box_h += lines[i].get_height();
	tmp_w = lines[i].get_width();
	//cerr << tmp_w << endl;
	if(tmp_w > form_w) form_w = tmp_w;
    }

    // extra margin for menus.
    if(menu_flag){
	form_w += menu_margin;
    }
    
    form_h = lower_box_h;
    if(title){
	form_h += title_gap_h + title_box_h;
	tmp_w = fl_get_string_width(line_c::lstyle, line_c::lsize, title, strlen(title)) + 2 * line_c::align_margin;
	//cerr << tmp_w << endl;
	if(tmp_w > form_w) form_w = tmp_w;
    }

    // make the form of the size. 
    FL_OBJECT *obj;
    PupForm = fl_bgn_form(FL_NO_BOX,form_w,form_h);
    obj = fl_add_box(FL_FLAT_BOX,0,0,form_w,form_h,"");
    fl_set_object_color(obj,FL_BLACK,FL_COL1);

    // optional timer
    if(stackpos == 0){
	Timer = obj = fl_add_timer(FL_HIDDEN_TIMER,0,0,1,1,"1");
	fl_set_object_callback(obj,PupTimerCB,0);
    }

    // lower box
    fl_add_box(FL_UP_BOX,0,title_gap_h+title_box_h,form_w,lower_box_h,"");

    // title box
    TitleBox = obj = fl_add_box(FL_UP_BOX,0,0,form_w,title_box_h,"");
    fl_set_object_align(obj,FL_ALIGN_LEFT);

    if(title){
	// removed a +5 in y.
	obj = fl_add_text(FL_NORMAL_TEXT,0,0,form_w,title_box_h,title);
	fl_set_object_boxtype(obj, FL_NO_BOX);
	fl_set_object_lsize(obj,line_c::lsize);
	fl_set_object_lstyle(obj,line_c::lstyle);
    }

    int start_h = title_gap_h + title_box_h + 5;
    for(i=0; i<=maxline; i++){
	if(i==0){
	    obj = ActiveBox = fl_add_box(
		tool_bw_flag ? FL_BORDER_BOX : FL_FLAT_BOX,
		5,start_h,form_w-10,lines[i].get_height(),""
	    );
//	    fl_set_object_color(obj,FL_BOTTOM_BCOL,FL_COL1);	    
	    fl_set_object_color(obj,tool_bw_flag ? FL_WHITE : FL_TOP_BCOL,FL_COL1);	    
	}
	lines[i].make(form_w, start_h);   // start_x == 0, height = text_box_h.
	start_h += lines[i].get_height();
    }    

    fl_end_form();
	fl_set_form_dblbuffer(PupForm, TRUE);
    fl_register_call_back(PupForm, FL_ALL_EVENT ,pup_raw_cb);    
    set_active(0);
}
	

void
pup_c::down_arrow_cb(void){
    if(active_line == maxline){
	set_active(0);
    } else {
	set_active(active_line+1);
    }
}

void
pup_c::up_arrow_cb(void){
    if(active_line <= 0){
	set_active(maxline);
    } else {
	set_active(active_line-1);
    }
}

void
pup_c::home_cb(void){
    set_active(0);
}

void
pup_c::end_cb(void){
    set_active(maxline);
}

void
pup_c::draw_active(void){
    fl_freeze_form(PupForm);
    if(active_line < 0 || active_line > maxline){
	if(ActiveBox->visible){
	    fl_hide_object(ActiveBox);
	}
    } else {
	ActiveBox->x = lines[active_line].line->x+5;
	ActiveBox->y = lines[active_line].line->y;
	if(!ActiveBox->visible){
	    fl_show_object(ActiveBox);
	}
    }
    if(PupForm->visible){
	fl_redraw_form(PupForm);
    }
    fl_unfreeze_form(PupForm);
}

static void pup_ungrab(void){
    if(dbg_flag) cerr << "ungrabbing.\n";
    XUngrabKeyboard(fl_display, CurrentTime);
    XUngrabPointer(fl_display, CurrentTime);
    set_timer(0); // we just did it ..
}

static void set_timer(int secs){
    if(stackpos >= 0){
	pup_stack[0]->set_timer(secs);
    }
}

void pup_c::set_timer(int secs){
    if(Timer && secs >= 0){
	fl_set_timer(Timer, secs);
    }
    if(!Timer){
	M_warn("set_timer","setting NULL timer to %d secs", secs);
    }
}

void PupTimerCB(FL_OBJECT *, long){
    if(dbg_flag)cerr << "pup_c: timeout.\n";
    if(stackpos >= 0){
	if(dbg_flag)cerr << "pup_c: closing ...\n";
	pup_stack[stackpos]->close_all();
    }
    /*
    pup_ungrab();  // a.s.a.p.
    cerr << "pup_c: ungrab.\n";
    if(stackpos >= 0){
	pup_stack[stackpos]->close_all();
    }
    */
    if(dbg_flag)cerr << "pup_c: timeout: all closed.\n";
}

void
pup_c::show(void){
    if(!PupForm){
	M_err("pup_c::show", "cannot show null PupForm");
	return;
    }
    FL_Coord x1,y1;
    if(stackpos == 0){
	fl_deactivate_all_forms();
	x1 = PupForm->w/2;

	if(x1 > PupForm->w - 5 - menu_margin){
	    x1 = PupForm->w - 5 - menu_margin;
	}
	y1 = lines[0].get_height()/2;
	/*
	if(title){
	    y1 += TitleBox->h + title_gap_h;
	}
	*/
	//cerr << x1 << "," << y1 << endl;
		if(x1<0) x1 = 0;
	if(y1<0) y1 = 0;
    
	fl_set_form_hotspot(PupForm, x1, y1);
	fl_show_form(PupForm, FL_PLACE_HOTSPOT, FL_NOBORDER, "");
    } else {
	// we're a submenu. position ourselves relative to the active line of our parent.
	// weird y correction so that I continue the submenu on the same y-height.
	// 5 overlap correction.
	pup_c *pup = pup_stack[stackpos-1];
	FL_OBJECT *obj = pup->lines[pup->active_line].line;
	x1 = pup->PupForm->x + pup->PupForm->w;
	y1 = pup->PupForm->y + obj->y - 1;
	if(title){
	    y1 -= TitleBox->h + title_gap_h;
	}
/*
#ifdef SUN
	if(title){
	    y1 -= TitleBox->h + title_gap_h;
	}
#else
	if(!title){
	    y1 += TitleBox->h + title_gap_h;
	}
#endif
*/
	if(y1 + PupForm->h > fl_scrh){
	    y1 = fl_scrh - PupForm->h;
	    if(y1<0){
		y1 = 0;
	    }
	}
	if(x1 + PupForm->w > fl_scrw){
	    x1 = fl_scrw - PupForm->w;
	    if(x1<0){
		x1 = 0;
	    }
	}
	fl_set_form_position(PupForm, x1, y1);
	fl_show_form(PupForm, FL_PLACE_POSITION, FL_NOBORDER, "");
    }
    if(stackpos == 0){
	if(dbg_flag) cerr << "grabbing.\n";
	// we want all events. Else, events in frozen windows get lost.
	if(XGrabKeyboard(fl_display, PupForm->window, False, GrabModeAsync, GrabModeAsync, CurrentTime)
	   != GrabSuccess){
	    cerr << "couldn't grab keyboard. Uggh.\n";
	}
	//XChangeActivePointerGrab(fl_display, 0, XC_watch, CurrentTime);
	fl_set_cursor_color(XC_sb_right_arrow, FL_RED, FL_WHITE);
	if(XGrabPointer
	   (
	    fl_display,
	    PupForm->window,
	    False,
	    ButtonPressMask | ButtonReleaseMask | LeaveWindowMask | EnterWindowMask | PointerMotionMask,
	    GrabModeAsync,
	    GrabModeAsync,
	    None,
	    fl_get_cursor_byname(XC_sb_right_arrow),
	    CurrentTime
	    )
	   != GrabSuccess){
	    cerr << "couldn't grab pointer. Uggh.\n";
	}
	set_timer(timeout_secs);
    }
}

void free_pup  (long& pup){
    if(pup>0){
	delete (pup_c *) pup;
	for(int i=0; i<= stackpos; i++){
	    if(pup_stack[i] == (pup_c *) pup){
		M_err("free_pup","freeing a pup that's still in the active stack!");
		pup_stack[i] = NULL;
	    }
	}
    } else {
	cerr << "freeing NULL pup.\n";
    }
    pup = 0;
}

long new_pup (void){
    line_c::lsize  = tool_menu_fontsize;
    line_c::lstyle = tool_menu_fontstyle;
    line_c::text_box_h = (int) ((double) line_c::lsize * 1.8);
    title_box_h = 10 + (int) (1.2 * (double) line_c::lsize);   
    return (long) new pup_c;
}

// show the menu, return the value of the item selected. Else return -1;
long do_pup(long pup){
    long retval = -1;
    if(pup>0){
	retval = ((pup_c *)pup)->do_pup();
    }
    return retval;
}

// register another menu line in a popup, 
// include a text with specifiers, callback or value, and possibly, shortcuts.
void add_to_pup(long pup, const char *str, long arg, const char *shortcuts){
    if(pup>0){
	((pup_c *)pup)->add(str, arg, shortcuts);
    }
}

//#define UNDER_DEVELOPMENT
#ifdef UNDER_DEVELOPMENT

void kill_cb(int){
    exit(1);
}

void m0_cb(int){
    cerr << "m0_cb\n";
}

void cb_b(int){
    cerr << "sub cb_b\n";
}

void cb_d(int){
    cerr << "sub cb_d\n";
}

main(int argc, char *argv[]){
    fl_initialize("uc", "UC", NULL, 0, &argc, argv);
    long sub = new_pup();
    add_to_pup(sub, "SubTitle %t", 0);
    add_to_pup(sub, "cb %f", (long) cb_b, "cC");
    add_to_pup(sub, "db %f", (long) cb_d, "dD");
    long retval = do_pup(sub);

    /*
    long sub2 = new_pup();
    add_to_pup(sub2, "SubTitle 2%t", 0);
    add_to_pup(sub2, "@__cb 2 %f", (long) cb_b, "cC");
    add_to_pup(sub2, "@__db 2 %f", (long) cb_d, "dD");
    
    long pup = new_pup();
    add_to_pup(pup, "Title %t", 0);
    add_to_pup(pup, "@__m0 %f", (long) m0_cb, "mM");
    add_to_pup(pup, "@__Kill %f", (long) kill_cb, "kK");
    add_to_pup(pup, "@__sub %m", sub, "sS");
    add_to_pup(pup, "@_men_u1 %f", 0, "uU");
    add_to_pup(pup, "menu333 %f", 0);
    add_to_pup(pup, "menu333verylonglong%m", sub2);
    add_to_pup(pup, "menu333verylonglong%f", 0);
    long retval = do_pup(pup);
    */
    cerr << "final retval: " << retval << endl;
    return 0;
}

#endif
