/* options stuff */

//#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h> // MAXPATHLEN
#include <unistd.h>

#include "action_forms_aux.h"
#include "browser.h"
#include "cb_fkey.h"
#include "display_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "fkeybar_forms_aux.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "gl_utility.h"
#include "global.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "linklist.h"
#include "menu.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "pathnames.h"
#include "sort_forms_aux.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "wm_forms_aux.h"
#include "xshell.h"

// piggyback on options_sync_defaults.
static void timeout_sync_defaults(void); 
static void special_ops_show_form(void);

static void AlwaysAskDontSync(void);

void options_close_form(void){
    if(OptionsForm->visible){
	focus_restore_start();
	fl_hide_form(OptionsForm);
	focus_restore_finish();
    }
}

void OptionsFormEditorButtonCB(FL_OBJECT*, long){
    if(strlen(default_editor) == 0){
	cerr << "UC: no default editor ?\n";
	strcpy(default_editor, "<builtin>");
    }
    const char *new_ed = OKCancel(
        "Give F6 default editor (for IMPORTANT help, press F1, or use <builtin>)",
	default_editor,
	default_editor_help_text,
	"Default Editor"
    );
    if(!new_ed) return;
    if(new_ed[0] == '\0') return;
    if (strcmp(new_ed, default_editor) == 0) return;
    if (strcmp(new_ed, "<builtin>") == 0) {
	strcpy(default_editor, "<builtin>");
	return;
    }
    if (strstr(new_ed, "builtin")) {
	if(!ask_question(TRUE,
	    "Are you sure you don't want <builtin> instead?"
	)){
	    strcpy(default_editor, "<builtin>");
	    tool_editor_info_shown_flag = TRUE;
	    return;
	}
    }
	
    if(new_ed[0] == '/'){
	// abs path:
	char new_ed_buf[MAXPATHLEN+1];
	strcpy(new_ed_buf, new_ed);
	char *where = strchr(new_ed_buf, ' ');
	if(where){
	    //strip trailing blanks and arguments.
	    *where = '\0';
	}
	    
	if(access(new_ed_buf, X_OK) != 0){
	    timed_strerror_notifier("Cannot access the new default editor !", notifier_secs);
	    return;
	}
    }
    if(ask_question(FALSE, "Are you sure `%s' is a valid X window editor, such as `emacs' or `xterm -e vi' ?", new_ed)){
	strcpy(default_editor, new_ed);
    }
}

void OptionsFormAutoMouseButtonCB(FL_OBJECT*, long){
    toggle_flag(auto_mouse_flag);
}

void OptionsFormSpecialOpsButtonCB(FL_OBJECT*, long){
    special_ops_show_form();
}

/** special options **/
void special_ops_help_func(void){
    SpecialOpsHelpButtonCB(NULL,0);
}

static int special_ops_raw_cb(FL_FORM *form, void *xev){
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, NULL, special_ops_help_func);
}

void special_ops_sync_defaults(void){
    fl_set_button(SpecialOpsPFlagButton, tool_cp_minusp_flag);
    fl_set_button(SpecialOpsFocusFlagButton, tool_focus_on_startup_flag);
    fl_set_counter_value(SpecialOpsCtrlTimeoutCounter, toolbox_ctrl_timeout);
    fl_set_counter_value(SpecialOpsBrowserFrequencyCounter, browser_c::directory_update_count);
    fl_set_counter_value(SpecialOpsXShellFrequencyCounter, xshell_update_count);
}

void SpecialOpsCtrlTimeoutCounterCB(FL_OBJECT *ob, long){
    //cerr << "dir:" << fl_get_counter_value(ob);
    toolbox_ctrl_timeout = fl_get_counter_value(ob);
}

void SpecialOpsBrowserFrequencyCounterCB(FL_OBJECT *ob, long){
    browser_c::directory_update_count = (int)fl_get_counter_value(ob);
}

void SpecialOpsXShellFrequencyCounterCB(FL_OBJECT *ob, long){
    xshell_update_count = (int)fl_get_counter_value(ob);
}

void SpecialOpsCloseReturnButtonCB(FL_OBJECT *, long){
    if(SpecialOpsForm->visible){
	focus_restore_start();
	fl_hide_form(SpecialOpsForm);
	focus_restore_finish();
    }
}

void SpecialOpsHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(special_ops_help_text, "Special Options");
}

void SpecialOpsDefaultsButtonCB(FL_OBJECT *, long){
    browser_c::directory_update_count = 40;
    xshell_update_count = 50;
    toolbox_ctrl_timeout = 2.0;
    tool_cp_minusp_flag  = FALSE;
    tool_focus_on_startup_flag  = FALSE;
    tool_menu_fontsize   = tool_fontsize;  // ?
    tool_menu_fontstyle  = 8;      // times bold.
    special_ops_sync_defaults();
}

static void menu_restart_msg(void){
    static int first = TRUE;
    if(first){
        first = FALSE;
	timed_notifier("Your change will be effective in some menus immediately,", 
		       "but in all of them only after a restart.");
    }
}

void SpecialOpsMenuFSizeButtonCB(FL_OBJECT *, long){
  long fsize_menu = make_fsize_menu();
  int newsize = (int) do_pup(fsize_menu);
  if(newsize > 0){
    tool_menu_fontsize = newsize;
    menu_restart_msg();
  }
  free_pup(fsize_menu); 
}

void SpecialOpsMenuFStyleButtonCB(FL_OBJECT *, long){
  long fstyle_menu = make_fstyle_menu();
  int newstyle = (int) do_pup(fstyle_menu);
  if(newstyle > 0){
    tool_menu_fontstyle = newstyle;
    menu_restart_msg();
  }
  free_pup(fstyle_menu); 
}

void SpecialOpsToolFSizeButtonCB(FL_OBJECT *, long){
    static int first = TRUE;
    long fsize_menu = make_fsize_menu();
    int newsize = (int) do_pup(fsize_menu);
    if(newsize > 0){
        tool_fontsize = newsize;
	if(first){
	    first = FALSE;
	    timed_notifier("Experimental: tool font size. Manual resize or restart may be necessary",
			   "Effects can be seen in toolbox.");
	}
	action_sync_defaults();
	toolbox_sync_defaults();
    }
	free_pup(fsize_menu); 
}

void SpecialOpsToolFStyleButtonCB(FL_OBJECT *, long){
  long fstyle_menu = make_fstyle_menu();
  int newstyle = (int) do_pup(fstyle_menu);
  if(newstyle > 0){
    tool_fontstyle = newstyle;
    timed_notifier("not yet implemented: tool font style. Sorry !");
    //action_sync_defaults();
    //toolbox_sync_defaults();
  }
  free_pup(fstyle_menu); 
}

static void special_ops_show_form(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
        fl_set_object_shortcut(SpecialOpsCloseReturnButton,      "Cc^M", TRUE);
        fl_set_object_shortcut(SpecialOpsTimeoutsButton,         "Tt", TRUE);
        fl_set_object_shortcut(SpecialOpsDefaultsButton,         "Dd", TRUE);
        //fl_set_object_shortcut(SpecialOpsHelpButton,             "Hh", TRUE);
	fl_set_button_shortcut(SpecialOpsFocusFlagButton,        "Oo", TRUE);
	fl_set_button_shortcut(SpecialOpsPFlagButton,            "nN", TRUE);
	fl_register_call_back(SpecialOpsForm, FL_ALL_EVENT, special_ops_raw_cb);
	fl_set_counter_bounds   (SpecialOpsCtrlTimeoutCounter, 1.0, 30.0);
	fl_set_counter_step     (SpecialOpsCtrlTimeoutCounter, .25, 2.5);
	fl_set_counter_precision(SpecialOpsCtrlTimeoutCounter, 2);

	fl_set_counter_bounds   (SpecialOpsBrowserFrequencyCounter, 0.0, 100.0);
	fl_set_counter_step     (SpecialOpsBrowserFrequencyCounter, 1.0, 5.0);
	fl_set_counter_precision(SpecialOpsBrowserFrequencyCounter, 1);

	fl_set_counter_bounds   (SpecialOpsXShellFrequencyCounter, 0.0, 100.0);
	fl_set_counter_step     (SpecialOpsXShellFrequencyCounter, 1.0, 5.0);
	fl_set_counter_precision(SpecialOpsXShellFrequencyCounter, 1);

    }
    set_to_ret_button(SpecialOpsCloseReturnButton);
    special_ops_sync_defaults();
    
    focus_save();
    if(!SpecialOpsForm->visible){
	fl_show_form(SpecialOpsForm, FL_PLACE_CENTER, TRUE, "Special Options:");
    } else {
	XMapRaised(fl_display, SpecialOpsForm->window);
    }
    tool_focus_move(SpecialOpsForm, SpecialOpsCloseReturnButton);
}

void SpecialOpsFocusFlagButtonCB(FL_OBJECT*, long){
    toggle_flag(tool_focus_on_startup_flag);
}

void SpecialOpsPFlagButtonCB(FL_OBJECT*, long){
    toggle_flag(tool_cp_minusp_flag);
}

void SpecialOpsTimeoutsButtonCB(FL_OBJECT*, long){
    timeout_adjust();
}

void OptionsFormFKeyBarButtonCB(FL_OBJECT*, long){
	if (!toolbox_split_flag) {
		if (!ask_question(FALSE, "This option requires splitting the ToolBox.")) {
			fl_set_button(OptionsFormFKeyBarButton, FALSE);
			return;
		}
		OptionsFormToolBoxSplitButtonCB(NULL, 0);
	}

    toggle_flag(fkeybar_visible_flag);
    if(fkeybar_visible_flag){
		fkeybar_show_form();
		XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
    } else {
		fkeybar_close_form();
    }
}

void OptionsFormToolBoxSplitButtonCB(FL_OBJECT*, long){
	toolbox_split();
	XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
}

void OptionsFormToolBoxVisibleButtonCB(FL_OBJECT*, long){
    toggle_flag(toolbox_visible_flag);
    if(toolbox_visible_flag){
	toolbox_show_form();
	XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
    } else {
	toolbox_close_form();
    }
}

void OptionsFormWinManagerButtonCB(FL_OBJECT*, long){
    toggle_flag(wm_visible_flag);
    if(wm_visible_flag){
	wm_show_form();
	XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
    } else {
	wm_close_form();
    }
}

void OptionsFormShellOpsButtonCB(FL_OBJECT*, long){
    xshell_show_form();
}

void OptionsFormUCShellButtonCB(FL_OBJECT*, long){
    toggle_flag(xshell_visible_flag);
    if(xshell_visible_flag){
	fl_set_cursor_color(XC_watch, FL_RED, FL_WHITE);
	fl_set_global_cursor(XC_watch);
	xshell_show();
	XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
	fl_reset_global_cursor();
    } else {
	// kludge 
	xshell_visible_flag = TRUE;
	xshell_close();
	xshell_visible_flag = FALSE;
    }
}

void OptionsFormStartupMsgButtonCB(FL_OBJECT*, long){
    toggle_flag(tool_startup_msg_flag);
}

void OptionsFormStatusButtonCB(FL_OBJECT*, long){
    toggle_flag(status_visible_flag);
    if(status_visible_flag){
	status_show_form();
	XSetInputFocus(fl_display, OptionsForm->window, RevertToPointerRoot, CurrentTime);
    } else {
	status_close_form();
    }
}

void options_sync_defaults(void){
    M_info("options_sync_defaults", "start");
    special_ops_sync_defaults();
    timeout_sync_defaults();
    fl_freeze_form(OptionsForm);
    fl_set_button(OptionsFormAutoRaiseButton,       browser_c::auto_raise_flag);
    fl_set_button(OptionsFormAutoMouseButton,       auto_mouse_flag);
	fl_set_button(OptionsFormFKeyBarButton,         fkeybar_visible_flag);
    fl_set_button(OptionsFormToolBoxVisibleButton,  toolbox_visible_flag);
    fl_set_button(OptionsFormToolBoxSplitButton,    toolbox_split_flag);
    fl_set_button(OptionsFormWinManagerButton,      wm_visible_flag);
    fl_set_button(OptionsFormUCShellButton,         xshell_visible_flag);
    fl_set_button(OptionsFormStartupMsgButton,      tool_startup_msg_flag);
    fl_set_button(OptionsFormStatusButton,          status_visible_flag);

    AlwaysAskDontSync();
    fl_unfreeze_form(OptionsForm);
    M_info("options_sync_defaults", "done");
}

void options_help_func(void){
    OptionsFormHelpButtonCB(NULL,0);
}

static int options_form_raw_cb(FL_FORM *form, void *xev){
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, options_close_form, options_help_func);
}

void OptionsButtonCB(FL_OBJECT *, long){
    static int first = TRUE;
    if(first){
	first = FALSE;
    
	fl_set_button_shortcut(OptionsFormTreeAlwaysButton,      "A", TRUE);
	fl_set_button_shortcut(OptionsFormAlwaysButton,          "a", TRUE);
	fl_set_button_shortcut(OptionsFormToolBoxSplitButton,    "Bb", TRUE);
    fl_set_object_shortcut(OptionsFormCloseReturnButton,     "Cc^M", TRUE);
    fl_set_object_shortcut(OptionsFormDefaultsButton,        "Dd", TRUE);
	fl_set_button_shortcut(OptionsFormEditorButton,          "Ee", TRUE);
	fl_set_button_shortcut(OptionsFormFactoryDefaultsButton, "Ff", TRUE);
	fl_set_button_shortcut(OptionsFormToolBoxVisibleButton,  "iI", TRUE);
	fl_set_button_shortcut(OptionsFormTreeAskButton,         "K", TRUE);
	fl_set_button_shortcut(OptionsFormAskButton,             "k", TRUE);
	fl_set_button_shortcut(OptionsFormLoadConfigButton,      "Ll", TRUE);
	fl_set_button_shortcut(OptionsFormAutoMouseButton,       "Mm", TRUE);
	fl_set_button_shortcut(OptionsFormTreeDontButton,        "N", TRUE);
	fl_set_button_shortcut(OptionsFormDontButton,            "n", TRUE);
	fl_set_button_shortcut(OptionsFormSpecialOpsButton,      "Oo", TRUE);
	fl_set_button_shortcut(OptionsFormShellOpsButton,        "pP", TRUE);
	fl_set_button_shortcut(OptionsFormAutoRaiseButton,       "Rr", TRUE);
	fl_set_button_shortcut(OptionsFormUCShellButton,         "Ss", TRUE);
	fl_set_button_shortcut(OptionsFormStartupMsgButton,      "uU", TRUE);
	fl_set_button_shortcut(OptionsFormStatusButton,          "tT", TRUE);
	fl_set_button_shortcut(OptionsFormSaveConfigButton,      "vV", TRUE);
	fl_set_button_shortcut(OptionsFormWinManagerButton,      "Ww", TRUE);
	fl_set_button_shortcut(OptionsFormFKeyBarButton,         "yY", TRUE);
	
/*	fl_addto_menu(OptionsMetaCharMenu,          "Alt");
	fl_addto_menu(OptionsMetaCharMenu,          "Ctrl");
	fl_show_menu_symbol(OptionsMetaCharMenu,          TRUE);
*/	
	fl_register_call_back(OptionsForm, FL_ALL_EVENT, options_form_raw_cb);
    }
    set_to_ret_button(OptionsFormCloseReturnButton);

    options_sync_defaults();

    focus_save();
    if(!OptionsForm->visible){
//	fl_show_form(OptionsForm, FL_PLACE_CENTER, TRUE, "Global Options (Meta-F2):");
	fl_show_form(OptionsForm, FL_PLACE_CENTER, TRUE, "Global Options:");
    } else {
	XMapRaised(fl_display, OptionsForm->window);
    }
    tool_focus_move(OptionsForm, OptionsFormCloseReturnButton);
}

void OptionsFormAutoRaiseButtonCB(FL_OBJECT *, long){
    browser_c::auto_raise_flag = !browser_c::auto_raise_flag;
}

void OptionsFormCloseReturnButtonCB(FL_OBJECT *, long){
    options_close_form();
}

void OptionsFormDefaultsButtonCB(FL_OBJECT *, long){
    browser_c::auto_raise_flag = TRUE;
    auto_mouse_flag = TRUE;
    save_config_on_exit_state = ask_e;
    save_tree_on_exit_state = ask_e;
    options_sync_defaults();
}

void OptionsFormHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(options_help_text, "Global Options");
}

static void conditional_restart(void) {
    if(!ask_question(FALSE, "A restart is needed for the changes to become effective. Restart now?")) {
	return;
    }
    tool_ctrl_on_flag = TRUE;	    
    safe_system("uc -f");
    sleep(2);
    ExitButtonCB(NULL, 0);
}

void OptionsFormFactoryDefaultsButtonCB(FL_OBJECT *, long){
    if(ask_question(TRUE, "To restore all defaults to original, the database will be removed. OK ?")){
	char buf[1024];
	sprintf(buf, "%s -f %s &", rm_path, tool_default_fname);
	safe_system(buf);
	conditional_restart();
    }
}

void OptionsFormSaveConfigButtonCB(FL_OBJECT *, long){
    if(!tool_ldefault_dir){
	timed_notifier("Your local database directory `~/.uc' seems to be missing.\n");
	return;
    }
    if(access(tool_ldefault_dir, X_OK) != 0){
	char buffer[MAXPATHLEN*2];
	sprintf(buffer, "Cannot access your local database directory `%s':", tool_ldefault_dir);
	timed_strerror_notifier(buffer);
	return;
    }

    linklist_c db_list;
    int n_db_files = 0, n_raw_files = 0;
    const FL_Dirlist *raw_db_list = NULL;
    raw_db_list = fl_get_dirlist(tool_ldefault_dir, "ucdb*", &n_raw_files, TRUE);

    // special case: for save, always have ucdb in list.
    db_list.add((void *) "ucdb");
    if(raw_db_list){
	for(int i=0; i<n_raw_files; i++){
	    if(raw_db_list[i].type == FT_FILE){
		n_db_files++;
		if(strcmp(raw_db_list[i].name, "ucdb") != 0){
		    // avoid special case.
		    db_list.add((void *) raw_db_list[i].name);
		}
	    }
	}
    }
    db_list.add((void *) "<new file>");
    
    long pup = new_pup();
    add_to_pup(pup, "Databases%t", 0);
    for(int i=0; i<db_list.getsize(); i++){
	char buffer[MAXPATHLEN];
	sprintf(buffer, "%s%%x%d", (const char *) db_list[i], i);
	add_to_pup(pup, buffer, 0);
    }
    int retval = (int) do_pup(pup);
    free_pup(pup);

    char db_file_name[MAXPATHLEN];
    if(retval < 0){
	return;
    } else if(retval == db_list.getsize()-1){
	const char *new_name = OKCancel("Give a new database name starting with `ucdb':", "ucdb",
	    NULL, NULL, FALSE, TRUE, 5, 5
	);
	if(new_name){
	    if(strncmp(new_name, "ucdb", strlen("ucdb")) != 0){
		timed_notifier("The new name you supplied didn't start with `ucdb'.");
		return;
	    }
	    sprintf(db_file_name, "%s/%s", tool_ldefault_dir, new_name);
	} else {
	    return;
	}
    } else {
	sprintf(db_file_name, "%s/%s", tool_ldefault_dir, (const char *) db_list[retval]);
    }
    tool_defaults_save(db_file_name);
}

void OptionsFormLoadConfigButtonCB(FL_OBJECT *, long){
    if (!tool_default_fname) return;

    linklist_c db_list;
	
    int n_db_files = 0, n_raw_files = 0;
    const FL_Dirlist *raw_db_list = NULL;
    const char *db_dir = NULL;
    if(tool_ldefault_dir){
	raw_db_list = fl_get_dirlist(tool_ldefault_dir, "ucdb*", &n_raw_files, TRUE);
	db_dir = tool_ldefault_dir;
	if(raw_db_list){
	    for(int i=0; i<n_raw_files; i++){
		if(raw_db_list[i].type == FT_FILE){
		    n_db_files++;
		    db_list.add((void *) raw_db_list[i].name);
		}
	    }
	}
    }
    if(!n_db_files && tool_gdefault_dir){
	raw_db_list = fl_get_dirlist(tool_gdefault_dir, "ucdb*", &n_raw_files, TRUE);
	db_dir = tool_gdefault_dir;
	if(raw_db_list){
	    n_db_files=0;
	    for(int i=0; i<n_raw_files; i++){
		if(raw_db_list[i].type == FT_FILE){
		    n_db_files++;
		    db_list.add((void *) raw_db_list[i].name);
		}
	    }
	}
    }
    if(!raw_db_list || !n_db_files || !db_dir){
	timed_notifier("No configuration files available for loading.");
	return;
	/*
    } else {
	for(int i=0; i<n_db_files; i++){
	    cerr << "file: " << (char *) db_list[i] << endl;
	}
	*/
    }
    const char * db_base_name = tool_default_fname;
    if(n_db_files == 1){
	db_base_name = (const char *) db_list[0];
    } else {
	long pup = new_pup();
	add_to_pup(pup, "Databases%t", 0);
	for(int i=0; i<n_db_files; i++){
	    char buffer[MAXPATHLEN];
	    sprintf(buffer, "%s%%x%d", (const char *) db_list[i], i);
	    add_to_pup(pup, buffer, 0);
	}
	int retval = (int) do_pup(pup);
	free_pup(pup);
	if(retval < 0){
	    return;
	} else {
	    db_base_name = (const char *) db_list[retval];
	}
    }
    if(!db_base_name) return;

    char db_file_name[MAXPATHLEN];
    sprintf(db_file_name, "%s/%s", db_dir, db_base_name);

    // make backup copy of old file, cp db_file to ucdb,
    // restart; 
    tool_defaults_backup();

    const char *fname = tool_default_fname;
    char buffer[2*MAXPATHLEN+64];
    sprintf(buffer, "%s %s %s", cp_path, db_file_name, fname);
    safe_system(buffer);
    
    conditional_restart();
}

static void AlwaysAskDontSync(void){
    fl_freeze_form(OptionsForm);
    fl_set_button(OptionsFormAlwaysButton, 0);
    fl_set_button(OptionsFormAskButton,    0);
    fl_set_button(OptionsFormDontButton,   0);

    fl_set_button(OptionsFormTreeAlwaysButton, 0);
    fl_set_button(OptionsFormTreeAskButton,    0);
    fl_set_button(OptionsFormTreeDontButton,   0);

    FL_OBJECT *Button;
    switch(save_config_on_exit_state){
    case always_e: Button = OptionsFormAlwaysButton; break;
    case ask_e:    Button = OptionsFormAskButton; break;
    case dont_e:   Button = OptionsFormDontButton; break;
    default:       Button = OptionsFormDontButton; break;
    }
    fl_set_button(Button, 1);

    switch(save_tree_on_exit_state){
    case always_e: Button = OptionsFormTreeAlwaysButton; break;
    case ask_e:    Button = OptionsFormTreeAskButton; break;
    case dont_e:   Button = OptionsFormTreeDontButton; break;
    default:       Button = OptionsFormTreeDontButton; break;
    }
    fl_set_button(Button, 1);
    
    fl_unfreeze_form(OptionsForm);
}    

void OptionsFormAlwaysAskDontButtonCB(FL_OBJECT *, long val){
    static int first = TRUE;
    save_config_on_exit_state = (save_on_exit_state_et) val;
    if(save_config_on_exit_state != always_e && first){
      first = FALSE;
      status_show_form();
      status_msg(1, "!!!!!!");
      status_msg(1, "In order for this new value to be remembered");
      status_msg(1, "and used next time, make sure that _this_ ");
      status_msg(1, "configuration is saved now or at the end.");
      status_msg(1, "!!!!!!");
    }
    AlwaysAskDontSync();
}

void OptionsFormTreeAlwaysAskDontButtonCB(FL_OBJECT *, long val){
    save_tree_on_exit_state = (save_on_exit_state_et) val;
    AlwaysAskDontSync();
}

void OptionsBrowserColoursButtonCB(FL_OBJECT*, long){
    tool_do_colours();
}

void OptionsBrowserDisplayButtonCB(FL_OBJECT*, long){
    display_show_form();
}

void OptionsBrowserSortingButtonCB(FL_OBJECT*, long){
    sort_show_form();
}

void BrowserCloneButtonCB(FL_OBJECT*, long){
    if(browser_c::next_ID > 10 && browser_c::next_ID <= 20){
	if(!ask_question(TRUE, "More than 10 browsers not recommended. Continue cloning ?")){
	    return;
	}
    }
    if(browser_c::next_ID > 20){
	timed_error_notifier(FALSE, "Not supported: more than 20 browsers.\n");
	return;
    }
    if(source_browser->clone()){
	// set new one source.
	tool_set_source(source_browser->rb_get_last_browser()); 
    }
}

void BrowserKillButtonCB(FL_OBJECT*, long){
    if(!source_browser){
	program_unstable("kill: no source browser ?");
	return;
    }
    browser_c *tmp_browser = source_browser;
    if(tmp_browser->next_browser == source_browser){
	// there's only one left - exit the tool ! No, don\'t it is stupid.
	// ask first. 
	if(ask_question(FALSE, "Kill the last remaining browser and exit ?")){
	    tool_ctrl_on_flag = TRUE;	    
	    ExitButtonCB(NULL, 0);
	}
	return;
    } else {
	int my_ID = tmp_browser->get_ID();
//	if(ask_question(FALSE, "Really kill this browser (#%d)?", my_ID)){
	    // set next one source.
	    // also check wether we were target.
	    // decrease next_ID by one.
	    tool_set_source(tmp_browser->next_browser);
	    if(target_browser == tmp_browser){
		tool_set_target(source_browser);
	    }
	    browser_c::next_ID--;
	    delete tmp_browser;
	    // renumber:
	    //cerr << "renumbering ID's. " << endl;
	    tmp_browser = source_browser;
	    do {
		int tmp_ID = tmp_browser->get_ID();
		if(tmp_ID > my_ID){
		    tmp_browser->set_ID(tmp_ID - 1);
		}
		tmp_browser = tmp_browser->next_browser;
	    } while (tmp_browser != source_browser);
//	}
    }
}

void TimeoutFormSingleFileCounterCB(FL_OBJECT *ob, long){
    //cerr << "singlefile: " << fl_get_counter_value(ob) << endl;
    tool_file_timeout_secs = (int) fl_get_counter_value(ob);
}

void TimeoutFormCumulativeDirCounterCB(FL_OBJECT *ob, long){
    //cerr << "dir:" << fl_get_counter_value(ob);
    tool_dir_timeout_secs = (int) fl_get_counter_value(ob);
}

void TimeoutFormCloseReturnButtonCB(FL_OBJECT *, long){
    if(TimeoutForm->visible){
	focus_restore_start();
	fl_hide_form(TimeoutForm);
	focus_restore_finish();
    }
}

void TimeoutFormHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(timeout_help_text, "Operation Timeouts");
}

void TimeoutFormDefaultsButtonCB(FL_OBJECT *, long){
    tool_dir_timeout_secs   = 5;  
    tool_file_timeout_secs  = 1;
    timeout_sync_defaults();
}

void timeout_sync_defaults(void){
    if(tool_dir_timeout_secs < tool_file_timeout_secs){
	tool_file_timeout_secs = tool_dir_timeout_secs;
    }
    fl_set_counter_value(TimeoutFormSingleFileCounter, tool_file_timeout_secs);
    fl_set_counter_value(TimeoutFormCumulativeDirCounter, tool_dir_timeout_secs);
}

void timeout_help_func(void){
    TimeoutFormHelpButtonCB(NULL, 0);
}

static int timeout_form_raw_cb(FL_FORM *form, void *xev){
    if(adjust_return_button(form, (XEvent *) xev)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, NULL, timeout_help_func);
}

void timeout_adjust(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_counter_bounds   (TimeoutFormSingleFileCounter, 1.0, 19.0);
	fl_set_counter_bounds   (TimeoutFormCumulativeDirCounter, 1.0, 299.0);
	fl_set_counter_step     (TimeoutFormSingleFileCounter, 1.0, 5.0);
	fl_set_counter_step     (TimeoutFormCumulativeDirCounter, 1.0, 10.0);
	fl_set_counter_precision(TimeoutFormSingleFileCounter, 0);
	fl_set_counter_precision(TimeoutFormCumulativeDirCounter, 0);
        fl_set_object_shortcut(TimeoutFormDefaultsButton,         "Dd", TRUE);
	fl_register_call_back   (TimeoutForm, FL_ALL_EVENT, timeout_form_raw_cb);
	make_standard_cancel_shortcut(TimeoutFormCloseReturnButton);	
    }
    timeout_sync_defaults();

    focus_save();
    fl_show_form(TimeoutForm, FL_PLACE_CENTER, FALSE, "");
    tool_focus_move(TimeoutForm, TimeoutFormCloseReturnButton);

    while(TimeoutForm->visible){
	tool_mung_events();
    }
}
	
#ifdef OLDCODE

/*
void OptionsMetaCharMenuCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    int retval = fl_get_menu(OptionsMetaCharMenu);
    if(retval != -1){
	if(retval == 1){
            tool_meta_char = alt_e;
	} else {	
            tool_meta_char = ctrl_e;
	}
	options_sync_defaults();
	tool_meta_sync();
    }
}
*/

#endif
