
/*
 * $Id: panel.c,v 1.2 1997/03/17 10:35:24 henrik Exp henrik $
 *.
 * This file is part of Unix Cockpit package
 * Copyright (c) 1994  Henrik Klagges
 *.
 *
 *  Object class panel - Unix Cockpit only.
 *
 * BW support: 
 *       F      B
 * NN: white, black
 * NC: black, white
 * SN: white, black, white frame 
 * SC: black, white, black frame
 */
#if defined(F_ID) || defined(DEBUG)
char *fl_id_brw = "$Id: panel.c,v 1.2 1997/03/17 10:35:24 henrik Exp henrik $";
#endif

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include "forms.h"
#include "flspecial.h"
#include "panel.h"

/* file type pixmaps */
#include "bitmaps/block10x10.xbm"
#include "bitmaps/char10x10.xbm"
#include "bitmaps/dir10x10.xbm"
#include "bitmaps/link10x10.xbm"
#include "bitmaps/pipe10x10.xbm"
#include "bitmaps/reg10x10.xbm"
#include "bitmaps/socket10x10.xbm"
#include "bitmaps/unknown10x10.xbm"

#include "bitmaps/block14x14.xbm"
#include "bitmaps/char14x14.xbm"
#include "bitmaps/dir14x14.xbm"
#include "bitmaps/link14x14.xbm"
#include "bitmaps/pipe14x14.xbm"
#include "bitmaps/reg14x14.xbm"
#include "bitmaps/socket14x14.xbm"
#include "bitmaps/unknown14x14.xbm"

#include "bitmaps/block18x18.xbm"
#include "bitmaps/char18x18.xbm"
#include "bitmaps/dir18x18.xbm"
#include "bitmaps/link18x18.xbm"
#include "bitmaps/pipe18x18.xbm"
#include "bitmaps/reg18x18.xbm"
#include "bitmaps/socket18x18.xbm"
#include "bitmaps/unknown18x18.xbm"

/* create pixmap for a given bitmap name including size: */
#define PIXCREATESPEC(name) \
    name##_px = XCreateBitmapFromData(fl_display, \
    win, (char *) name##_bits, name##_width, name##_height);

/* create all pixmaps for all sizes of a given bitmap basename: */
#define PIXCREATE(name) \
    PIXCREATESPEC(name##10x10); \
    PIXCREATESPEC(name##14x14); \
    PIXCREATESPEC(name##18x18); 

/* show a specific pixmap in the line's colour */
#define PIXSHOWSPEC(name) \
    if (name##_px != None) { \
    fl_color(bitmap_lcol); \
    fl_bk_color(bitmap_bgcol); \
    XCopyPlane(fl_display, name##_px, FL_ObjWin(ob), \
        fl_get_gc(), 0, 0, name##_width, name##_height, xx_bitmap, yy_bitmap, 1 \
    ); }

/* show the bitmap depending on size: */
#define PIXSHOW(name) \
    if (bitmap_size >= 18) { \
	PIXSHOWSPEC(name##18x18); \
    } else if (bitmap_size >= 14) { \
	PIXSHOWSPEC(name##14x14); \
    } else { \
	PIXSHOWSPEC(name##10x10); \
    }


Pixmap  
    block10x10_px   = None,
    char10x10_px    = None,
    dir10x10_px     = None,
    link10x10_px    = None,
    pipe10x10_px    = None,
    reg10x10_px     = None,
    socket10x10_px  = None,
    unknown10x10_px = None,

    block14x14_px   = None,
    char14x14_px    = None,
    dir14x14_px     = None,
    link14x14_px    = None,
    pipe14x14_px    = None,
    reg14x14_px     = None,
    socket14x14_px  = None,
    unknown14x14_px = None,

    block18x18_px   = None,
    char18x18_px    = None,
    dir18x18_px     = None,
    link18x18_px    = None,
    pipe18x18_px    = None,
    reg18x18_px     = None,
    socket18x18_px  = None,
    unknown18x18_px = None
;
static void create_pixmaps(FL_OBJECT *ob);

static cleanup_func_pt
    cleanup_func_fp       = NULL
;
static exec_func_pt
    exec_func_fp          = NULL
;
static external_func_pt
    track_external_func_fp= NULL
;
static verify_func_pt 
    source_verify_func_fp = NULL
;
static show_func_pt
    track_show_func_fp    = NULL
;

#if 0
#define STYPE  FL_PANEL_SLIDER
#else
#define STYPE  FL_VERT_BROWSER_SLIDER
#endif

#define NL    10
#define BSW   24
#define CHARHEIGHT_MUNG 2  /* works nicely with all font sizes.  */
#define ASCEND_MUNG 3      /* same here. */


/* Object specific information */

typedef struct
{
    char selected;		/* Whether selected */
    char needs_redraw;          /* wether this line needs redraw (HAK) */
    void *client_data;          /* HAK */
    char *txt;
}

LINE;

typedef struct
{
    LINE *text[FL_PANEL_MAXLINE + 1];	/* lines of text(0 not used) */
    int topline;		/* Current topline */
    int lines;			/* Number of lines */
    int selectline;		/* Last selected line */
    int fontstyle;		/* Style of font */
    int fontsize;		/* The character size */
    char bitmapkey;		/* Key that indicates a filetype bitmap */
    char specialkey;		/* Key that indicates a special symbol  */

    int current_is_hidden;      /* flag wether current line is hidden   */
    int currentline;		/* current current line HAK             */
    int partial;	        /* 0=not, 1=no back, 2=no back, slider. */
    int xcopy_flag;
    int cursor;
    int dnd_sensor_width,       /* depends on the font, width is "   "+1*/
        dnd_deselect_flag,      /* if dnd file was selected by dnd, reset it on fail. */
        dnd_source_line,        /* the line for the above               */
        dnd_flag,               /* wether dnd is active, used for drawing */
        dnd_valid_tracking_flag,/* wether we've a non-source valid tracking */
        xdnd_valid_flag         /* indicates slave mode (external) tracking */
    ;

    int
	SelectedCurrentLabelColor,	/* initialized via #define */
	SelectedCurrentBackColor,
	SelectedLabelColor,
	CurrentLabelColor,
	CurrentBackColor
    ;

    int oldtopline;		/* change mark              */
    int charheight;		/* base font height         */
    int chardesc;		/* base font descent        */
    int attrib;			/* set if there is attrib change */
    int vmode;			/* vmode GCs are valid for  */
    int screenlines;		/* no. of visible lines     */
    int pushed;                 /* 0=none, 1=upper, 2=lower */
    GC primaryGC;		/* text drawing GC          */
    FL_Coord x, y, w, h;		/* browser drawing area     */
}

SPEC;


/***************** DATA STRUCTURE MAINTENANCE ************************/

static void set_redraw(SPEC *sp, int line, int flag){
    if(line > 0 && line <= sp->lines){
	sp->text[line]->needs_redraw = flag;
    }
}

/* Deletes a line from the panel */
static void
delete_line(FL_OBJECT * ob, int linenumb)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    LINE *ttt;
    int i;
    ttt = sp->text[linenumb];
    for (i = linenumb; i < sp->lines; i++)
	sp->text[i] = sp->text[i + 1];
    sp->text[sp->lines] = ttt;
    if(ttt->txt){
	fl_free(ttt->txt);
	ttt->txt = NULL;
    }
    sp->lines--;
    if (sp->selectline == linenumb)
	sp->selectline = 0;
    else if (sp->selectline > linenumb)
	sp->selectline--;

    if(sp->currentline == linenumb){  	/* HAK >> */
  	if(sp->lines == 0){
	    sp->currentline = -1;
  	} else {
	    sp->currentline = 1;
  	}
    } else if(sp->currentline > linenumb){
  	sp->currentline--;
    }					/* << HAK */
}

/* Inserts a line of text to a panel */
static int
insert_line(FL_OBJECT * ob, int linenumb, const char *newtext)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    LINE *ttt;
    int i;
    if (sp->lines == FL_PANEL_MAXLINE)
    {
	delete_line(ob, 0);
	linenumb--;
    }
    sp->lines++;
    /* Create new line if required */
    if (sp->text[sp->lines] == 0){
	sp->text[sp->lines] = (LINE *) fl_calloc(1, sizeof(LINE));
	sp->text[sp->lines]->client_data = NULL;
	sp->text[sp->lines]->txt = NULL;
    }
    /* Shift lines */
    ttt = sp->text[sp->lines];
    for (i = sp->lines - 1; i >= linenumb; i--)
	sp->text[i + 1] = sp->text[i];
    sp->text[linenumb] = ttt;
    /* Fill in line */
    sp->text[linenumb]->selected = 0;

    if(sp->text[linenumb]->txt){
	fl_free(sp->text[linenumb]->txt);
    }
    sp->text[linenumb]->txt = (char *)fl_calloc(strlen(newtext)+1, sizeof(char));
    strcpy(sp->text[linenumb]->txt, newtext);

    if (sp->selectline >= linenumb) sp->selectline++;
    /* HAK */
    if (sp->currentline >= linenumb) sp->currentline++;
    return linenumb;
}

/* Replaces a line of text in a panel */
static void
replace_line(FL_OBJECT * ob, int linenumb, const char *newtext)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    /* Fill in line */
    if(sp->text[linenumb]->txt){
	fl_free(sp->text[linenumb]->txt);
    }
    sp->text[linenumb]->txt = (char *)fl_calloc(strlen(newtext)+1, sizeof(char));
    strcpy(sp->text[linenumb]->txt, newtext);
}

/******************************** DRAWING ********************************/

/* HAK : = 3: repaint marked lines only. */
/* note that we are clever: partials are cumulative. You can only set partial
 * to a lower value than you started with. The redraw functions then reset it.
 * note that -1 means uninitialized. Thus, external expose/draw events redraw
 * the whole thing, as we'd like it, but internal requests can be more 
 * efficient. 
 * Extra kludge: For a possible XCopyArea when partial == 0, we have a flag.
 * all these flags need to be spec flags, because of freezing interference.
 */

static void set_partial(FL_OBJECT *ob, int val){
    SPEC *sp = (SPEC *) (ob->spec);
    if((val < sp->partial) || (sp->partial == -1)){
	sp->partial = val;
    }
    if(val >= 0 && val < 3){
	/* for cumulativeness of xcopy (use val, not partial) */
	/* fprintf(stderr, "set_partial(%d): setting xcopy to FALSE.\n", val);*/
	sp->xcopy_flag = FALSE;
    } else {
	/* fprintf(stderr, "set_partial(%d): keeping xcopy.\n", val);*/
    }
}

/* Calculates the area in which the text is drawn */
static void
calc_textarea(FL_OBJECT * ob, FL_Coord * xx, FL_Coord * yy,
	      FL_Coord * ww, FL_Coord * hh, int *screenlines)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    FL_Coord charheight = sp->charheight,     /* character height */
        eff_bw = FL_PANEL_BW+1;
    
    /* total browser drawing area, to be adjusted for scroll bar */
/* changed from new browser.c */
    *xx = ob->x + eff_bw;
    *yy = ob->y + eff_bw;
    *ww = ob->w - 2 * eff_bw;
    *hh = ob->h - 2 * eff_bw;

/* HAK ORG
    *xx = ob->x + 1.5 * FL_PANEL_BW;
    *yy = ob->y + 1.0 * FL_PANEL_BW;
    *ww = ob->w - 3.0 * FL_PANEL_BW;
    *hh = ob->h - 2.0 * FL_PANEL_BW;
*/

    *screenlines = (int) (*hh / charheight);

    /* Check whether a slider will be shown */
    /* always - HAK */
    /* if (sp->lines > *screenlines){ */
    *ww -= BSW;
    *xx += BSW;
}

/* Corrects the position of the topline inside the panel */
static void
correct_topline(FL_OBJECT * ob)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    int screenlines;		/* lines on screen */
    FL_Coord xx, yy, ww, hh;	/* text box */

    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);
    if (sp->lines > screenlines)
    {
	if (sp->lines - sp->topline + 1 < screenlines)
	    sp->topline = sp->lines - screenlines + 1;
	if (sp->topline < 1)
	    sp->topline = 1;
    }
    else
	sp->topline = 1;
}

/* mini triangular symbol for selection marking */
#define AddPoint(p, xp, yp) p->x = xp; p->y = yp; p++

static void
draw_triangle(FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
	    int face_left, FL_COLOR lcol, FL_COLOR bcol)
{
    int xc = x + w / 2, yc = y + h / 2;
    XPoint xpoint[5], *xp;
    float wm = w * 0.5, hm = h * 0.5;
    float pl = 1.0, ps = 1.0;
    int pswm = (int) (ps * wm + 0.1), plhm = (int) (pl * hm + 0.1);

    xp = xpoint;

    if(face_left){
	AddPoint(xp, xc + pswm, yc - plhm);
	AddPoint(xp, (short) (xc - 0.5 * wm), yc);
	AddPoint(xp, xc + pswm, yc + plhm);
    } else {
	AddPoint(xp, xc - pswm, yc - plhm);
	AddPoint(xp, (short) (xc + 0.5 * wm), yc);
	AddPoint(xp, xc - pswm, yc + plhm);
    }
    fl_rectf(x, y, w, h, bcol);
    fl_polyf(xpoint, 3, lcol);
/*    fl_polyl(xpoint, 3, FL_BLACK); */
}

/* Draws the line on the panel at postion xx,yy, with maximal width ww.
 * back indicates whether the background should be drawn */
static void
draw_textline(FL_OBJECT * ob, 
    int line, FL_Coord xx, FL_Coord yy, FL_Coord ww, 
    int back, int bitmap_only_flag
){
    SPEC *sp = ((SPEC *) (ob->spec));
    char *str;			/* Start of actual string */
    int style = sp->fontstyle;	/* Actual font style used */
    FL_Coord size = sp->fontsize;	/* Actual font size used */
    int lcol = ob->lcol;	/* Actual font color used */
    int align = FL_ALIGN_LEFT | FL_ALIGN_INSIDE;	/* Actual alignment used */
    FL_Coord ascend = sp->charheight - sp->chardesc + ASCEND_MUNG;	/* character height */
    FL_Coord hh_bitmap, xx_bitmap, yy_bitmap; /* HAK */
    int tmp_col     = -1,            /* HAK */
        bitmap_size = 10,         /* HAK */
        bitmap_lcol,
        bitmap_bgcol
    ;

    int text_margin = FL_PANEL_TEXTMARGIN;
    int bw_flag = (fl_depth(fl_vmode) <= 1);
    if(bw_flag){
	text_margin *= 2;
    }

    /* make sure it's off now */
    if (!bitmap_only_flag) {
	set_redraw(sp, line, FALSE);        
    }
    /* fprintf(stderr, "drawing line %d (back: %d).\n", line, back);   */

    /* HAK: draw the current/ cursel box if required */
    /* HAK: Add new colour names */
    /* also: take care of the current_is_hidden flag */
    if(line < 1 || line > sp->lines){
	return;
    }
    lcol = bw_flag ? FL_WHITE : lcol;
    if (sp->text[line]->selected){
	if(line == sp->currentline && !sp->current_is_hidden){
	    /* selected and current */
	    lcol    = bw_flag ? FL_BLACK : sp->SelectedCurrentLabelColor;
	    tmp_col = bw_flag ? FL_WHITE : sp->SelectedCurrentBackColor;
	} else {
	    /* selected only */
	    lcol    = bw_flag ? FL_WHITE : sp->SelectedLabelColor;
	    tmp_col = bw_flag ? FL_BLACK : (int) ob->col2;
	}
    } else if(line == sp->currentline && !sp->current_is_hidden){
	/* not selected, but current */
	lcol    = bw_flag ? FL_BLACK : sp->CurrentLabelColor;
	tmp_col = bw_flag ? FL_WHITE : sp->CurrentBackColor;
    } else if(back) {
	tmp_col = bw_flag ? FL_BLACK : (int) ob->col1;
    }
    if(tmp_col >= 0 && !bitmap_only_flag){
	fl_rectf(xx, (FL_Coord) (yy - 0.5 * ascend), ww, sp->charheight, tmp_col);
    }

    xx += text_margin;  /* new */

    /* Check for special lines */
    str = sp->text[line]->txt;
    while (str[0] != '\0' && str[0] == sp->specialkey)
    {
	switch (str[1])
	{
	case 'l':
	    size = FL_LARGE_FONT;
	    yy -= 3;
	    break;
	case 'm':
	    size = FL_MEDIUM_FONT;
	    break;
	case 's':
	    size = FL_SMALL_FONT;
	    break;
	case 'b':
	    style = FL_BOLD_STYLE;
	    break;
	case 'i':
	    style = FL_ITALIC_STYLE;
	    break;
	case 'f':
	    style = FL_FIXED_STYLE;
	    break;
	case 'c':
	    align = FL_ALIGN_CENTER;
	    break;
	case 'r':
	    align = FL_ALIGN_RIGHT;
	    break;
	case 'C':
	    /* here we really should change the way color is used: TODO */
	    lcol = 0;
	    while (str[2] >= '0' && str[2] <= '9')
	    {
		lcol = 10 * lcol + str[2] - '0';
		str++;
	    }
	    if (lcol == 0)
		lcol = FL_BLACK;
	    else if (lcol == 1)
		lcol = FL_RED;
	    else if (lcol == 2)
		lcol = FL_GREEN;
	    else if (lcol == 3)
		lcol = FL_YELLOW;
	    else if (lcol == 4)
		lcol = FL_BLUE;
	    else if (lcol == 5)
		lcol = FL_MAGENTA;
	    else if (lcol == 6)
		lcol = FL_CYAN;
	    else if (lcol == 7)
		lcol = FL_WHITE;
	    break;
	case '_':		/* line under it */
	    fl_line(xx, (FL_Coord) (yy + 0.5 * sp->charheight), ww - 2*text_margin, 1, FL_BLACK);
	    break;
	}
	str++;
	str++;
    }
    /* the 8/bitmap_size is font dependent! Idea: Use bitmap plus one line
     * above and one line below bitmap as size criterion. 
     * experimentally, that looks very ugly. So, switch to +2 extra lines.
     * that's about good. But try 2 and extra +4 now:
     * plus, try new bitmaps that have extra border.
     */
    hh_bitmap = sp->charheight;
    bitmap_size = 10;
    if (hh_bitmap >= 18) bitmap_size = 14;
    if (hh_bitmap >= 22) bitmap_size = 18;
    xx_bitmap = xx;
    yy_bitmap = (FL_Coord) (yy - 0.5 * ascend + (hh_bitmap - bitmap_size)/2.0);

    if (tmp_col < 0){
	tmp_col = bw_flag ? FL_BLACK : (int) ob->col1;
    }
    bitmap_lcol = lcol;
    bitmap_bgcol= tmp_col;
    if (sp->dnd_flag) {
	if (sp->text[line]->selected || 
	    (line == sp->currentline && fl_get_panel_dnd_track_valid(ob))
        ){
	    bitmap_lcol = tmp_col;
	    bitmap_bgcol= lcol;
	}
    } else if (sp->xdnd_valid_flag && line == sp->currentline) {
	bitmap_lcol = tmp_col;
	bitmap_bgcol= lcol;
    }

    if (str[0] == sp->bitmapkey && str[1] != '\0') {
	switch(str[1]) {
	case 'B': PIXSHOW(block);   break;
	case 'C': PIXSHOW(char);    break;
	case 'D': PIXSHOW(dir);     break;
	case 'L': PIXSHOW(link);    break;
	case 'P': PIXSHOW(pipe);    break;
	case 'R': PIXSHOW(reg);     break;
	case 'S': PIXSHOW(socket);  break;
	default:  PIXSHOW(unknown); break;
	}
	str+=2;
    }
    if (bitmap_only_flag){
	return;
    }

    /* Draw the string, centered on y */
    /* HAK side effect: lcol may have changed in the multi-if above */
    fl_drw_text(align, xx, yy, ww, 0, lcol, style, size, str); 

    if(bw_flag && sp->text[line]->selected){
	/* tmp_col = (tmp_col == FL_BLACK) ? FL_WHITE : FL_BLACK; */
	FL_Coord tw = text_margin, yc = (int) (0.5 * ascend);
	draw_triangle(xx-tw,       yy - yc, tw, sp->charheight, 0, lcol, tmp_col);
	draw_triangle(xx-tw+ww-tw, yy - yc, tw, sp->charheight, 1, lcol, tmp_col);
	/* fl_line(xx, (FL_Coord) (yy + 0.5 * sp->charheight), ww, 1, lcol); ugly */
    }
}

static void correct_slsize(FL_OBJECT *ob, float *slsize){
    const float minsize = 20.0;
    float org_slsize;
    /* HAK: min pixels for panel slider:  */
    if(*slsize > 1.0){
	*slsize = 1.0;
	return;
    }
    if(*slsize * ob->h < minsize){
	if(ob->h > 0.0){
	    org_slsize = *slsize;
	    *slsize = minsize / ob->h;
	    if(*slsize > 1.0){
		*slsize = org_slsize;
	    }
	}
    }
}

/* Draws the slider if required. */
static void
draw_browslider(FL_OBJECT * ob, int screenlines)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    float slsize, slpos;	/* slider size and position */

    if(sp->lines > 0){		/* HAK */
	slsize = (1.0 * screenlines) / sp->lines;
    } else {
	slsize = 1.0;
    }
    if(sp->lines <= screenlines){		/* HAK */
  	slsize = 1.0;
	slpos = 0.0;
    } else {
	slpos = (1.0 * sp->topline - 1.0) / (sp->lines - screenlines);
    }

    correct_slsize(ob, &slsize);

    fl_drw_slider(ob->boxtype,
		  ob->x, ob->y+BSW, BSW, ob->h-2*BSW,
		  FL_BROWSER_SLCOL, FL_BROWSER_SLCOL, STYPE,
		  slsize, slpos, "",
		  FL_SLIDER_ALL, 0, ob->bw);
}

/* 0=don't draw, 1=draw normal, 2=draw pushed */
static void
draw_browbuttons(FL_OBJECT * ob, int upper_state, int lower_state){
    SPEC *sp = ((SPEC *) (ob->spec));
    FL_Coord xx = ob->x, boxb = 1;

    if(upper_state == 2){
	sp->pushed = 1;
    } else if(lower_state == 2){
	sp->pushed = 2;
    } else {
	sp->pushed = 0;
    }

    if(upper_state > 0){
	fl_drw_box((ob->boxtype == FL_UP_BOX && upper_state == 2) ? FL_DOWN_BOX : ob->boxtype,
		   xx, ob->y,               BSW, BSW, FL_BROWSER_SLCOL, FL_BUTTON_BW);
	fl_drw_text_beside(FL_ALIGN_CENTER, xx+boxb, ob->y+boxb, BSW-2*boxb, BSW-2*boxb,
			   FL_BLACK, FL_FIXEDBOLD_STYLE, FL_MEDIUM_FONT, "@8");
    }
    if(lower_state > 0){
	fl_drw_box((ob->boxtype == FL_UP_BOX && lower_state == 2) ? FL_DOWN_BOX : ob->boxtype,
		   xx, ob->y + ob->h - BSW, BSW, BSW, FL_BROWSER_SLCOL, FL_BUTTON_BW);
	fl_drw_text_beside(FL_ALIGN_CENTER, xx+boxb, ob->y + ob->h - BSW -boxb, BSW-2*boxb, BSW-2*boxb,
			   FL_BLACK, FL_FIXEDBOLD_STYLE, FL_MEDIUM_FONT, "@2");
    }
}

static void
draw_slider_motion(FL_OBJECT * ob);

/* Draws the panel */
static void
draw_panel(FL_OBJECT * ob)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    FL_Coord charheight = sp->charheight;
    FL_Coord ascend = charheight - sp->chardesc + ASCEND_MUNG;	/* character height */
    /* character height */
    int screenlines;		/* lines on screen */
    int i;			/* counters, etc. */
    FL_Coord xx, yy, ww, hh;	/* text box */

    /* manually increase sensor width by 1 char. Don't want to miss it. This is a hack.*/
    sp->dnd_sensor_width = fl_get_string_width(sp->fontstyle, sp->fontsize,"   ", 3);
    create_pixmaps(ob);         /* now, so that ob's win is valid */

    /* fprintf(stderr, "full panel draw.\n");  */
    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);
    correct_topline(ob);
    /* always draw slider */
    draw_browbuttons(ob, 1, 1);
    draw_browslider(ob, screenlines);
    fl_drw_box(ob->boxtype,
	       ob->x + BSW,
	       ob->y, ob->w - BSW,
	       ob->h, ob->col1, FL_PANEL_BW);

    fl_drw_text_beside(ob->align, ob->x, ob->y, ob->w, ob->h,
		       ob->lcol, ob->lstyle, ob->lsize, ob->label);

    fl_set_clipping(xx, yy, ww, hh - 1); 
    fl_set_text_clipping(xx, yy, ww, hh - 1); 
    yy += (int) (0.5 * ascend);
    for (i = 0; i < screenlines; i++, yy += charheight){
	draw_textline(ob, i + sp->topline, xx, yy, ww, FALSE, FALSE);
    }
    fl_unset_clipping();
    fl_unset_text_clipping();
}

/* Draws the contents of the panel only. slid indicated whether the
 * slider should be redrawn. */
static void
draw_panel_partial(FL_OBJECT * ob, int slid)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    int screenlines;		/* lines on screen */
    int i;			/* counters, etc. */
    FL_Coord xx, yy, ww, hh;	/* text box */
    FL_Coord charheight = sp->charheight;
    FL_Coord ascend = charheight - sp->chardesc + ASCEND_MUNG;	/* character height */

    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);
    correct_topline(ob);

    /* always. HAK */
    if (slid) draw_browslider(ob, screenlines);

    fl_set_clipping(xx, yy, ww, hh - 1);
    fl_set_text_clipping(xx, yy, ww, hh - 1); 

    yy += (int) (0.5 * ascend);
    for (i = 0; i < screenlines; i++, yy += charheight){
	draw_textline(ob, i + sp->topline, xx, yy, ww, TRUE, FALSE);
    }
    fl_unset_clipping();
    fl_unset_text_clipping();
}

static void draw_panel_partial3(FL_OBJECT *ob)
/* Draws the contents of the panel only. Each line has a flag indicating
 * wether it should be redrawn. Draw those and reset the flag of ALL others.
 */
{
    SPEC *sp = ((SPEC *) (ob->spec));
    int screenlines;		/* lines on screen */
    int i,i2;			/* counters, etc. */
    FL_Coord xx, yy, ww, hh;	/* text box */
    FL_Coord charheight = sp->charheight;
    FL_Coord ascend = charheight - sp->chardesc + ASCEND_MUNG;	/* character height */

    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);
    correct_topline(ob);

    fl_set_clipping(xx, yy, ww, hh - 1);
    fl_set_text_clipping(xx, yy, ww, hh - 1); 
    yy += (int) (0.5 * ascend);
    for (i = 0; i < screenlines; i++, yy += charheight){
	i2 = i + sp->topline;
	if(sp->text[i2]){
	    if(sp->text[i2]->needs_redraw){
		draw_textline(ob, i2, xx, yy, ww, TRUE, FALSE);
		set_redraw(sp, i2, FALSE);        
	    }
	}
    }
    fl_unset_clipping();
    fl_unset_text_clipping();
}

/* draw only the bitmaps of current line and selected lines. */
static void draw_panel_bitmaps(FL_OBJECT *ob, int current_only_flag){
    SPEC *sp = ((SPEC *) (ob->spec));
    int screenlines;		/* lines on screen */
    int i,i2;			/* counters, etc. */
    FL_Coord xx, yy, ww, hh;	/* text box */
    FL_Coord charheight = sp->charheight;
    FL_Coord ascend = charheight - sp->chardesc + ASCEND_MUNG;	/* character height */

    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);
    correct_topline(ob);

    fl_set_clipping(xx, yy, ww, hh - 1);
    fl_set_text_clipping(xx, yy, ww, hh - 1); 
    yy += (int) (0.5 * ascend);
    for (i = 0; i < screenlines; i++, yy += charheight){
	i2 = i + sp->topline;
	if(sp->text[i2]){
	    if((sp->text[i2]->selected && !current_only_flag) ||
	        i2 == sp->currentline
	    ){
		draw_textline(ob, i2, xx, yy, ww, FALSE, TRUE);
	    }
	}
    }
    fl_unset_clipping();
    fl_unset_text_clipping();
}

/*
 * Before we start drawing, find out some basic info about fontsize,
 * charactor height etc and most importantly setup the GCs we will
 * be using to reduce GC changes on the fly.
 *
 * Bugs: If user screws around with obj color field, this won't work
 *       correctly. Hate to stick more stuff into SPEC
 */
static void
prepare_redraw(FL_OBJECT * ob, SPEC *sp)
{
    FL_Coord w,h;
    int junk;
    sp->charheight = fl_get_char_height(sp->fontstyle, sp->fontsize, &junk, &sp->chardesc) + CHARHEIGHT_MUNG;

    /* from no slider to slider, clip might be wrong. Check that */
    w = sp->w;
    h = sp->h;
    calc_textarea(ob, &(sp->x), &(sp->y), &(sp->w), &(sp->h),
		  &(sp->screenlines));

    M_info("DrawPrepare", "x=%d y=%d w=%d h=%d", sp->x, sp->y, sp->w, sp->h);
    if (!sp->primaryGC || sp->vmode != fl_vmode || sp->attrib || w != sp->w || h != sp->h){
	XGCValues xgcv;
	unsigned long gcvm;
	Screen *scr = ScreenOfDisplay(fl_display, fl_screen);

	M_warn("Browser", "Creating new GCs");

	if (sp->primaryGC){
	    XFreeGC(fl_display, sp->primaryGC);
	    sp->primaryGC = 0;
	    /*
	    XFreeGC(fl_display, sp->specialGC);
	    XFreeGC(fl_display, sp->bkGC);
	    XFreeGC(fl_display, sp->selectGC);
	    */
	}

	/* solicit graphics_exposure only if no backing store */
	xgcv.graphics_exposures = !DoesBackingStore(scr) || !fl_cntl.backingStore;
	gcvm = GCGraphicsExposures | GCForeground;

	/* primaryGC. text drawing */
	xgcv.foreground = fl_get_flcolor(ob->lcol);
	sp->primaryGC = XCreateGC(fl_display, FL_ObjWin(ob), gcvm, &xgcv);
	XSetFont(fl_display, sp->primaryGC, fl_state[fl_vmode].cur_fnt->fid);
	fl_set_gc_clipping(sp->primaryGC, sp->x, sp->y, sp->w, sp->h - 1);
#if 0
	/* background GC is used to redraw deselected lines */
	xgcv.foreground = fl_get_flcolor(ob->col1);
	sp->bkGC = XCreateGC(fl_display, FL_ObjWin(ob), gcvm, &xgcv);
	fl_set_gc_clipping(sp->bkGC, sp->x, sp->y, sp->w, sp->h - 1);

	/* selectGC is used to mark the current selection */
	xgcv.foreground = fl_get_flcolor(ob->col2);
	sp->selectGC = XCreateGC(fl_display, FL_ObjWin(ob), gcvm, &xgcv);
	fl_set_gc_clipping(sp->selectGC, sp->x, sp->y, sp->w, sp->h - 1);

	/* specialGC is used to handle font/color change on the fly */
	sp->specialGC = XCreateGC(fl_display, FL_ObjWin(ob), gcvm, &xgcv);
	fl_set_gc_clipping(sp->specialGC, sp->x, sp->y, sp->w, sp->h - 1);
#endif
	sp->attrib = 0;
	sp->vmode = fl_vmode;
    }
}

/*
 * Slider moved. We simple blit the "scroll region" to its new position
 * and redraw only the lines that just come into view. However, special
 * provision should be made where blit might not perform well as this
 * routine is callable by application program via set_topline.
 */
static void
draw_slider_motion(FL_OBJECT * ob){
    SPEC *sp = ((SPEC *) (ob->spec));
    FL_Coord charheight, ascend;	/* character height */
    int screenlines;		/* lines on screen */
    int i;			/* counters, etc. */
    FL_Coord xx, yy, ww, hh;	/* text box */
    int delta, bh, tmp;

    /* fprintf(stderr, "motion/xcopy draw.\n");  */

    /* do all redraws but the new currentline, the first is necessary 
     * and the second an optimization 
     */
    tmp = sp->topline;
    sp->topline = sp->oldtopline;
    set_redraw(sp, sp->currentline, FALSE);
    draw_panel_partial3(ob);  
    sp->topline = tmp;

    charheight = sp->charheight;
    ascend = charheight - sp->chardesc + ASCEND_MUNG;
    xx = sp->x;
    yy = sp->y;
    ww = sp->w;
    hh = sp->h;

    screenlines = sp->screenlines;
    correct_topline(ob);


    delta = (sp->oldtopline - sp->topline);
    M_info("SliderMotion", "Delta=%d", delta);

    if (sp->lines > screenlines && delta)
	draw_browslider(ob, screenlines);

    /* since user can influence the topline, blit is not always faster or
       appropriate. Set threshold to about 2/3 of the visible lines */


    if (FL_abs(delta) > (2 * screenlines / 3))
    {
	fl_set_clipping(xx, yy, ww, hh - 1);
	fl_set_text_clipping(xx, yy, ww, hh - 1); 
	for (yy += (int) (0.5 * ascend), i = 0; i < screenlines; i++, yy += charheight){
	    draw_textline(ob, i + sp->topline, xx, yy, ww, TRUE, FALSE);
	}
	fl_unset_clipping();
	fl_unset_text_clipping();
	return;
    }

    if (sp->oldtopline > sp->topline)
    {
	int sl;
	/* calculate the area needs to be blited */
	bh = (screenlines - delta) * charheight;

	XCopyArea(fl_display, FL_ObjWin(ob), FL_ObjWin(ob), sp->primaryGC,
		  xx, sp->y, ww, bh, xx, sp->y + delta * charheight);

	/* draw the new lines */
	fl_set_clipping(xx, yy, ww, hh - 1);
	fl_set_text_clipping(xx, yy, ww, hh - 1); 
	for (yy += (int) (0.5 * ascend), i = 0; i < delta; yy += charheight, i++){
	    draw_textline(ob, i + sp->topline, xx, yy, ww, TRUE, FALSE);
	}

	/* Need to take care of special lines that use larger fonts */
	sl = sp->topline + 1;
	if (sl < sp->lines && sp->text[sl]->txt[0] == sp->specialkey){
	    draw_textline(ob, sl, xx, sp->y + ascend + charheight, ww, FALSE, FALSE);
	}
	fl_unset_clipping();
	fl_unset_text_clipping();
    } else if (sp->oldtopline < sp->topline){
	int start;

	delta = (sp->topline - sp->oldtopline);

	/* calculate the area needs to be blited */
	bh = (screenlines - delta) * charheight;

	XCopyArea(fl_display, FL_ObjWin(ob), FL_ObjWin(ob), sp->primaryGC,
		  xx, sp->y + delta * charheight, ww, bh, xx, sp->y);

	fl_set_clipping(xx, yy, ww, hh - 1);
	fl_set_text_clipping(xx, yy, ww, hh - 1); 
	yy += (int) (0.5 * ascend);
	yy += (screenlines - delta) * charheight;
	start = screenlines - delta + sp->topline;
	for (i = 0; i < delta; yy += charheight, i++){
	    draw_textline(ob, i + start, xx, yy, ww, TRUE, FALSE);
	}
	fl_unset_clipping();
	fl_unset_text_clipping();
    }
    /* clean up currentline - optimization */
    set_redraw(sp, sp->currentline, TRUE);
    draw_panel_partial3(ob);  
}

/***************** HANDLING EVENTS **************************/

#define NOEVENT		 0
#define SELECTEVENT	 1
#define DESELECTEVENT	 2
#define SLIDEREVENT	 3
#define PAGEEVENT	 4
#define MOVECURRENTEVENT 5
#define DNDSTARTEVENT    6
#define DNDTRACKEVENT    7

static int eventtype = NOEVENT;	/* Type of interaction taking place */
static int pagesize;		/* Amount to scroll for page events */

static int sweep_from_line = 1; 
static int page_event_count = 0;

/* handles a mouse change. returns whether a selection change has occured */
static int
handle_mouse(FL_OBJECT * ob, FL_Coord mx, FL_Coord my, int key, void *xev)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    float charheight = sp->charheight;
    XEvent *xevent = (XEvent *) xev;

    int screenlines;		      /* lines on screen */
    float slsize, slpos=0.0, slval;   /* slider size,pos and value */
    int line;		              /* new number of line */
    int oldtopline = sp->topline;     /* old value of topline */
    int newtopline = -1;	      /* New topline */
    int slaction = 0;		      /* Type of slider action */
    int in_dnd_xregion_flag = FALSE,  /* horizontally */
        in_dnd_yregion_flag = FALSE,  /* vertically */
        old_valid_flag      = sp->dnd_valid_tracking_flag
    ;

    /* cursor gizmo (HAK) */ 
    int new_cursor  = XC_sb_v_double_arrow,
	maxline = fl_get_panel_maxline(ob)
    ;
    FL_Coord xx = ob->x;

    /* fprintf(stderr, "handle mouse: (int) key is: %d\n", (int) key); */

    /* Check whether ther are any lines */
    if (sp->lines == 0)
	return FALSE;

    /* Compute possible slider position change */
    correct_topline(ob);
    screenlines = (int) ((ob->h - 2.0 * FL_PANEL_BW) / charheight);

    if (sp->lines > screenlines)  /* HAK NO CHANGE !! (overrides nicely) */
    {
	slsize = (1.0 * screenlines) / sp->lines;
	slpos = (1.0 * sp->topline - 1.0) / (sp->lines - screenlines);
	slaction = fl_get_pos_in_slider(
            ob->x, ob->y+BSW, BSW, ob->h-2*BSW,
	    STYPE, slsize, mx, my, slpos, &slval
        );

	newtopline = (int) (slval * (sp->lines - screenlines) + 1.0);
    }

    if(key>0 && xevent->type != MotionNotify){
	/* need to disambiguate this from cont. slider op */
	if(mx>=xx && mx < xx+BSW){
	    if(my >= ob->y && my < ob->y + BSW){
		slaction = 1;
		draw_browbuttons(ob, 2, 0);
	    } else if(my >= ob->y + ob->h -BSW && my < ob->y + ob->h){
		slaction = -1;
		draw_browbuttons(ob, 0, 2);
	    }  
	}
    }

    /* compute and normalize line, see where we are. */
    line = (int) (sp->topline + (my - ob->y - FL_PANEL_BW) / charheight);
    if(line > sp->lines){
	line = sp->lines;
    } else if (line < 1){
	line = 1;
    }
    if ((mx >= ob->x + BSW) && 
	(mx < ob->x + BSW + sp->dnd_sensor_width)
    ){
	in_dnd_xregion_flag = TRUE;
    }
    if ((my >= ob->y + FL_PANEL_BW) &&
	(my <  ob->y + ob->h - FL_PANEL_BW)
    ){
	in_dnd_yregion_flag = TRUE;
    }

    /* Determine the type of event */
    if (eventtype == NOEVENT) {
	/* reset dnd here, just in case: */
	sp->dnd_flag = FALSE;   
	sp->dnd_valid_tracking_flag = FALSE;

	if (sp->lines > screenlines && (mx < ob->x + BSW) ){
	    if (slaction == 0) {
		eventtype = SLIDEREVENT;
	    } else {
		eventtype = PAGEEVENT;
		if (slaction == -2)
		    pagesize = screenlines;
		else if (slaction == -1)
		    pagesize = 1;
		else if (slaction == 1)
		    pagesize = -1;
		else if (slaction == 2)
		    pagesize = -screenlines;
	    }
	} else {
	    eventtype = SELECTEVENT;
	    if (key > 0) {
		/*
		fprintf(stderr, "panel: meta, copy: %d, %d\n",
		    (fl_keymask & ControlMask),
		    ((fl_keymask & Button2Mask) || (fl_keymask & Button3Mask))
		);
		*/
		if (key == 1) {
		    eventtype = MOVECURRENTEVENT;
		    if (controlkey_down(fl_keymask) || 
			metakey_down   (fl_keymask) || 
			shiftkey_down  (fl_keymask)
		    ) {
			/* emulate selection with shift-leftmouse */
			eventtype = SELECTEVENT;
			sweep_from_line = line;
		    }
		} else if (key == 2) {
		    /* middle mouse, select */
		    sweep_from_line = line;
		} else if (key == 3) {
		    /* menu event, don't do a thing. */
		}

		if (eventtype == SELECTEVENT) {
		    if (line >= 1 && line <= sp->lines &&
			line < sp->topline + screenlines &&
			sp->text[line]->selected){
			eventtype = DESELECTEVENT;
		    }
		}

		if (in_dnd_xregion_flag) {
		    if (!source_verify_func_fp || 
			source_verify_func_fp(ob, line, (XEvent *) xev)
		    ){
			/* is valid as dnd argument (typically dot/dotdot) */
			eventtype = DNDSTARTEVENT;
			sp->dnd_flag = TRUE;
		    }
		}
	    }
	}
    }

    /* mung the event-specific cursor */
    /* default comes first. Override below if necessary. */
    if(sp->currentline <= 1){
	new_cursor = XC_sb_down_arrow;
    } else if(sp->currentline >= maxline){
	new_cursor = XC_sb_up_arrow;
    }

    switch (eventtype){
    case DNDSTARTEVENT:  
	/* unconditionally on START */
	new_cursor = XC_fleur;
	break;
    case DNDTRACKEVENT:
	sp->dnd_valid_tracking_flag = FALSE;
	if (in_dnd_xregion_flag && in_dnd_yregion_flag) {
	    new_cursor = XC_fleur;
	    if (line != sp->dnd_source_line) {
		/* may still be changed below! */
		sp->dnd_valid_tracking_flag = TRUE;
	    }
	}
	break;
    case PAGEEVENT:
	/* fprintf(stderr, "page\n"); */
	new_cursor = XC_left_ptr;
	break;
    case SLIDEREVENT:
	/* fprintf(stderr, "slider (%f).\n", slpos); */
	if(slpos <= 0.0){
	    new_cursor = XC_sb_down_arrow;
	} else if(slpos >= 1.0){
	    new_cursor = XC_sb_up_arrow;
	}
	break;
    default:
	/* fprintf(stderr, "default\n");*/
	break;
    }
    fl_set_panel_cursor(ob, new_cursor);

    if (old_valid_flag != sp->dnd_valid_tracking_flag) {
	draw_panel_bitmaps(ob, TRUE);
    }

    /* fprintf(stderr, "Panel event: %d (dndflag %d).\n", eventtype, sp->dnd_flag); */

    /* Handle the event */
    switch (eventtype)
    {
    case PAGEEVENT:
	if(page_event_count <= 0 || page_event_count >= 20){
	    int mult, tmp_size = pagesize;
	    if(abs(pagesize) == 1 && page_event_count >= 20){
		mult = (page_event_count-19)/2;
		if (mult > 20) mult = 20;
		if (mult <= 0) mult = 1;
		/* fprintf(stderr, "mult: %d", mult); */
		tmp_size *= mult;
	    }
	    /* fprintf(stderr, "pecount: %d, pagesize %d\n", page_event_count, pagesize);*/
	    sp->topline += tmp_size;
	    correct_topline(ob);
	    if (sp->topline == oldtopline)
		return 0;
#ifdef UNUSED_CODE	    
	    forms_active_flag = TRUE;  /* superkludge :-) */
#endif
	    set_partial(ob, 1);
	    fl_redraw_object(ob);
	}
	page_event_count++;
	return 0;  /* to feed back geometry change */
    case SLIDEREVENT:
	/* fprintf(stderr, "slider event: ");*/
	page_event_count=0;
	sp->topline = newtopline;
	correct_topline(ob);
	if (sp->topline == oldtopline)
	    return 0;
	set_partial(ob, 1);
	fl_redraw_object(ob);
	return 0;
    default:
	page_event_count=0;
	if (line < sp->topline)
	    line = sp->topline;
	if (line >= sp->topline + screenlines)
	    line = sp->topline + screenlines - 1;
	if (line > sp->lines)
	    line = sp->lines;
	/* fprintf(stderr, "default event: %d", eventtype); */

	/* HAK */
	if (eventtype == MOVECURRENTEVENT) {
	    /* perror(" move c ... \n"); */
	    sp->selectline  = 0;    /* HAK - we moved, no selection happened */
	    if(line == sp->currentline) return 1;
	    set_redraw(sp, sp->currentline, TRUE);
	    set_redraw(sp, line, TRUE);
	    sp->currentline = line;
	} else if (eventtype == DNDSTARTEVENT || eventtype == DNDTRACKEVENT) {
	    /* so far, similar to move current... */
	    sp->selectline  = 0;    /* don't know this yet */

	    /* special code. Do not track if we are not in dnd mode. safeguard.
	     */
	    if (!sp->dnd_flag) {
		return 1;
	    }
	    /* in track mode, optimize the redraw away. Also, don't
	     * track onto a non-track line.
	     */
	    if (eventtype == DNDTRACKEVENT) {
		/* redraw optimization: (bitmaps are handled elsewhere) */
		if (line == sp->currentline) {
		    if (track_show_func_fp) track_show_func_fp(ob, line);
		    return 1;
		}
	    }
	    set_redraw(sp, sp->currentline, TRUE);
	    set_redraw(sp, line, TRUE);
	    sp->currentline = line;

	    if (eventtype == DNDSTARTEVENT) {
		/* do this on start only, and switchover type right now, */
		/* to distinguish between the click and the moving.      */
		eventtype = DNDTRACKEVENT;

		/* remember the original line */
		sp->dnd_source_line = line;
		if (sp->text[line]->selected) {
		    /* dnd multi. Indicate dnd-ness with bitmap. */
		    sp->dnd_deselect_flag = FALSE;
		    draw_panel_bitmaps(ob, FALSE);
		} else {
		    /* dnd single. deselect rest, select this. */
		    fl_deselect_panel(ob);
		    fl_select_panel_line(ob, line);
		    sp->dnd_deselect_flag = TRUE;
		}
	    }
	    if (track_show_func_fp) track_show_func_fp(ob, line);
	} else if (eventtype == SELECTEVENT){
	    /* Not in my version: (interferes with sweep)
	      if (sp->text[line]->selected)
	      return 0;
	      */

	    sp->text[line]->selected = 1;
	    sp->selectline = line;
	    /* fprintf(stderr, "move currentline in select event\n"); */
	    set_redraw(sp, line, TRUE);
	    set_redraw(sp, sp->currentline, TRUE);
	    sp->currentline = line;
	    /* fill the gaps (mark selected and redraw, not current, not selectline) */
	    if(line > sweep_from_line){    
		int i;
		for(i=sweep_from_line+1; i<line; i++){
		    if(!(sp->text[i]->selected)){
			sp->text[i]->selected = 1;
			set_redraw(sp, i, TRUE);        
		    }
		}
	    } else if(line < sweep_from_line){    
		int i;
		for(i=sweep_from_line-1; i>line; i--){
		    if(!(sp->text[i]->selected)){
			sp->text[i]->selected = 1;
			set_redraw(sp, i, TRUE);      
		    }  
		}
	    }   /* else same, no move, or no gap > 0 */  /* << HAK */
	} else {
	    /* eventtype==DESELECTEVENT */
	    /* interferes with sweep.
	    if (!sp->text[line]->selected)
		return 0;
	     */

	    set_redraw(sp, line, TRUE);
	    set_redraw(sp, sp->currentline, TRUE);
	    sp->currentline = line;
	    /* fill the gaps (mark selected and redraw, not current, not selectline) */
	    if(line > sweep_from_line){    
		int i;
		for(i=sweep_from_line+1; i<line; i++){
		    if(sp->text[i]->selected){
			sp->text[i]->selected = 0;
			set_redraw(sp, i, TRUE); 
		    }
		}
	    } else if(line < sweep_from_line){    
		int i;
		for(i=sweep_from_line-1; i>line; i--){
		    if(sp->text[i]->selected){
			sp->text[i]->selected = 0;
			set_redraw(sp, i, TRUE);
		    }  
		}
	    }   /* else same, no move, or no gap > 0 */
	    sp->text[line]->selected = 0;
	    sp->selectline = -line;
	}
	set_partial(ob, 3); 
	set_redraw(sp, line, TRUE);        
	fl_redraw_object(ob);
	return 1;
    }
}

/* HAK: Key handling */

#ifdef UNUSED_CODE
static void fl_panel_uparrow(FL_OBJECT *ob){
    int newline = fl_get_panel_current(ob) -1;
    if(newline>=1){
	fl_set_panel_current(ob, newline);
	fl_redraw_object(ob);
    }
}

static void fl_panel_downarrow(FL_OBJECT *ob){
    int newline = fl_get_panel_current(ob) +1;
    if(newline<= fl_get_panel_maxline(ob) && newline > 0){
	fl_set_panel_current(ob, newline);
	fl_redraw_object(ob);
    }
}

/* Handles a key press, returns whether something has changed (always 0 ?) */
static int
handle_key(FL_OBJECT * ob, int key)
{
    switch (key){
    case 3:			/* Up key */
	if (shiftkey_down(fl_keymask)){
	} else {
	    fl_panel_uparrow(ob); 
	    return 1;
	}
	break;
    case 4:			/* Down key */
	if (shiftkey_down(fl_keymask)){
	} else {
	    fl_panel_downarrow(ob);   
	    return 1;
	}
	break;
    default:
	break;
    }
    return 0;
}
#endif

/*-----------------------------------------------*/

static int statuschanged = 0;

/* Handles the panel */
static int
handle_panel(FL_OBJECT * ob, int event, FL_Coord mx, FL_Coord my, int key, void *xev)
{
    int i;
    SPEC *sp = ((SPEC *) (ob->spec));

    switch (event)
    {
    case FL_DRAW:
	/* fprintf(stderr, "FL_DRAW in browser.\n"); */
	/* was partial == 0, but then no redraw after rescaling */
	if (sp->partial <= 0 || ob->form->use_pixmap){
	    prepare_redraw(ob, sp);
	    if(sp->xcopy_flag){
		/* fprintf(stderr, "could attempt Xblit.\n"); */
		draw_slider_motion(ob);
		sp->xcopy_flag = FALSE;
	    } else {
		/* fprintf(stderr, "would not Xblit.\n");*/
		draw_panel(ob);
	    }
	} else if (sp->partial == 1){
	    draw_panel_partial(ob, TRUE);
	} else if (sp->partial == 2){
	    draw_panel_partial(ob, FALSE);
	} else if (sp->partial == 3){ 
	    draw_panel_partial3(ob);
	} 
	sp->partial = -1; /* reset partial value - important */
	return 0;
    case FL_PUSH:
	eventtype = NOEVENT; /* reset global event flag */
	statuschanged = 0;
	/*fprintf(stderr, "is push, key = %d", key);*/
	/* FALL-THROUGH */
    case FL_MOUSE:
	/*fprintf(stderr, "is mouse, key = %d", key);*/
	/* freezing, so that the current line hiding in xcopyarea works. 
	 * no speed penalty, apparently.
	 * DNDSTARTEVENT cannot occur here. 
	 */
	/* capture dnd processing early */
	if (eventtype == DNDTRACKEVENT) {
	    if (mx < 0 || my < 0 || mx > ob->form->w || my > ob->form->h) {
		if (track_external_func_fp && track_external_func_fp(ob)){
		    sp->dnd_valid_tracking_flag = FALSE;
		    return 0;
		}
	    }
	}

	fl_freeze_form(ob->form);
	if (eventtype == SELECTEVENT || eventtype == DESELECTEVENT || 
	    eventtype == MOVECURRENTEVENT || eventtype == DNDTRACKEVENT
	){
	    int diff = 0;
	    if(my < ob->y){
		diff = (int) ( (ob->y - my) / 20.0);
		if(diff <= 0) diff = 1;
		fl_set_panel_topline(ob,sp->topline-diff);
	    } else if(my > ob->y+ob->h){
		diff = (int) ( (my - (ob->y + ob->h)) / 20.0);
		if(diff <= 0) diff = 1;
		fl_set_panel_topline(ob,sp->topline+diff);
	    }
	}

	if (handle_mouse(ob, mx, my, key, xev)) statuschanged = 1;
	fl_unfreeze_form(ob->form);

	if (statuschanged){
	    statuschanged = 0;
	    return 1;
	}
	return 0;
/*
    case FL_KEYBOARD:
	return handle_key(ob, key);
*/
    case FL_ENTER:
    case FL_LEAVE:
	return 0;
    case FL_RELEASE:
	/* fprintf(stderr, "button release.\n"); */
	page_event_count=0;
    fl_set_panel_cursor(ob, XC_left_ptr);

	/* only if none down of 1 & 2*/
	/*
	if(!(
	     ((XButtonEvent*) xev)->state & Button1Mask || 
	     ((XButtonEvent*) xev)->state & Button2Mask
        )){      
	    fl_reset_cursor(ob->form->window);
	}
	*/
	if(sp->pushed == 1){
	    draw_browbuttons(ob, 1, 0);
	} else if(sp->pushed == 2){
	    draw_browbuttons(ob, 0, 1);
	}
	if (sp->dnd_flag) {
	    if (exec_func_fp) {
		exec_func_fp(ob);
	    }
	    
	    sp->dnd_flag = FALSE;
	    draw_panel_bitmaps(ob, FALSE);
	    if (sp->dnd_deselect_flag) {
		sp->dnd_deselect_flag = FALSE;
		fl_deselect_panel_line(ob, sp->dnd_source_line);
	    }
	    if (cleanup_func_fp) cleanup_func_fp(ob);

	    statuschanged=1;
	}
	return statuschanged;
    case FL_FREEMEM:
	for (i = 1; i <= FL_PANEL_MAXLINE; i++){
	    if (sp->text[i] != 0){
		if (sp->text[i]->txt){
		    fl_free(sp->text[i]->txt);
		}
		fl_free(sp->text[i]);
	    }
	}
	if(sp->primaryGC){
	    XFreeGC(fl_display, sp->primaryGC);
	    sp->primaryGC = 0;
	}
	/* 
	  XFreeGC(fl_display, sp->specialGC);
	  XFreeGC(fl_display, sp->bkGC);
	  XFreeGC(fl_display, sp->selectGC);
	  sp->primaryGC = 0;
	 */

	free(ob->spec);
	return 0;
    case FL_OTHER:
	switch( ((XEvent *) xev)->type){
	case GraphicsExpose:
	    {
		XGraphicsExposeEvent *xge = (XGraphicsExposeEvent *)xev;
		if (xge->count == 0 && xge->drawable == FL_ObjWin(ob)){
		    set_partial(ob, 0);
		    fl_redraw_object(ob);
		    M_warn("Browser", "GraphicsExposeRedraw");
		}
	    }
	    break;
	default:
	    break;
	}
    }
    return 0;
}

/*-----------------------------------------------*/

static void create_pixmaps(FL_OBJECT *ob){
    static int first = TRUE;
    Window win = FL_ObjWin(ob) ? FL_ObjWin(ob) : fl_root;
    if (first) {
	first = FALSE;
/*
	dir10x10_px = XCreateBitmapFromData(
            fl_display, win, (char *) dir10x10_bits, 
            dir10x10_width, dir10x10_height
        );
*/
	PIXCREATE(block);
	PIXCREATE(char);   
	PIXCREATE(dir);    
	PIXCREATE(link);   
	PIXCREATE(pipe);   
	PIXCREATE(socket); 
	PIXCREATE(reg);    
	PIXCREATE(unknown);
    }
}

/* Creates a panel */
FL_OBJECT *
fl_create_panel(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
		  const char *label)
{
    int i;
    int junk;
    FL_OBJECT *ob = NULL;
    SPEC *sp;

    ob = fl_make_object(FL_PANEL, type, x, y, w, h, label, handle_panel);
    ob->boxtype = FL_PANEL_BOXTYPE;
    ob->align = FL_PANEL_ALIGN;

    /* fl_set_object_dblbuffer(ob, 1); */

    ob->spec = fl_calloc(1, sizeof(SPEC));
    sp = ((SPEC *) (ob->spec));

    sp->fontsize = FL_NORMAL_SIZE;
    sp->fontstyle = FL_FIXED_STYLE;
    sp->lines = 0;
    sp->topline = 1;
    sp->selectline = 0;
    sp->specialkey = '\0';  /* default to none */
    sp->bitmapkey = '@';
    sp->primaryGC = 0;

    for (i = 0; i <= FL_PANEL_MAXLINE; i++){
	sp->text[i] = 0;
    }
    ob->input = 0; /* 1; for keys */
    sp->currentline = -1;
    sp->current_is_hidden = FALSE;

    sp->SelectedCurrentLabelColor = FL_PANEL_SELCUR_LCOL;
    sp->SelectedCurrentBackColor  = FL_PANEL_SELCUR_BCOL;
    sp->SelectedLabelColor	  = FL_PANEL_SEL_LCOL;
    ob->col2			= FL_PANEL_SEL_BCOL;
    sp->CurrentLabelColor	= FL_PANEL_CUR_LCOL;
    sp->CurrentBackColor	= FL_PANEL_CUR_BCOL;
    ob->lcol 			= FL_PANEL_LCOL;
    ob->col1 			= FL_PANEL_BCOL;

    sp->charheight = fl_get_char_height(sp->fontstyle, sp->fontsize, &junk, &sp->chardesc) + CHARHEIGHT_MUNG;

    sp->partial = -1;	
    sp->xcopy_flag = FALSE;
    sp->cursor = None;
    sp->dnd_sensor_width = 10; /* sensible initialization */
    sp->dnd_deselect_flag= FALSE;
    sp->dnd_source_line = 0;
    sp->dnd_flag = FALSE;
    sp->xdnd_valid_flag = FALSE;

    return ob;
}

/* Adds an object */
FL_OBJECT *
fl_add_panel(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h, const char *label)
{
    FL_OBJECT *ob;
    ob = fl_create_panel(type, x, y, w, h, label);
    fl_add_object(fl_current_form, ob);
    return ob;
}

/*-----------------------------------------------*/

/* Sets the topline for the panel. */
void
fl_set_panel_topline(FL_OBJECT * ob, int line)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    int old_partial = sp->partial;
    int old_flag = sp->xcopy_flag;
    if (line < 1)
	line = 1;
    if (line > sp->lines)
	line = sp->lines;
    if (line == sp->topline)
	return;
    sp->oldtopline = sp->topline;
    sp->topline = line;
    set_partial(ob, 0);
    if(old_partial >= 3 || old_partial < 0 || old_flag){
	/* only if no redraw loss: */
	/* fprintf(stderr, "set_topline: setting xcopy to TRUE.\n"); */
	sp->xcopy_flag = TRUE;
    }
    fl_redraw_object(ob);
}

/*-----------------------------------------------*/

/* Clears the panel. */
void
fl_clear_panel(FL_OBJECT * ob)
{
    ((SPEC *) (ob->spec))->lines = 0;
    ((SPEC *) (ob->spec))->topline = 1;
    ((SPEC *) (ob->spec))->selectline = 0;
    ((SPEC *) (ob->spec))->currentline = -1; /* HAK */
    set_partial(ob, 0); 
    fl_redraw_object(ob);
}

/* Adds a line of text to a panel */
int
fl_add_panel_line(FL_OBJECT * ob, const char *newtext)
{
    int newline = insert_line(ob, ((SPEC *) (ob->spec))->lines + 1, newtext);
    set_partial(ob, 0); 
    fl_redraw_object(ob);
    return newline;
}

/* Adds a line of text to a panel, and changes focus */
int
fl_addto_panel(FL_OBJECT * ob, const char *newtext)
{
    int newline = insert_line(ob, ((SPEC *) (ob->spec))->lines + 1, newtext);
    ((SPEC *) (ob->spec))->topline = ((SPEC *) (ob->spec))->lines;
    set_partial(ob, 0); 
    fl_redraw_object(ob);
    return newline;
}

/* Inserts a line of text in a panel */
int 
fl_insert_panel_line(FL_OBJECT * ob, int linenumb, const char *newtext)
{
    int new_line;
    if (linenumb < 1 || linenumb > ((SPEC *) (ob->spec))->lines)
	return -1;
    new_line = insert_line(ob, linenumb, newtext);
    set_partial(ob, 0);
    fl_redraw_object(ob);
    return new_line;
}

/* Deletes a line from a panel. */
void
fl_delete_panel_line(FL_OBJECT * ob, int linenumb)
{
    if (linenumb < 1 || linenumb > ((SPEC *) (ob->spec))->lines)
	return;
    delete_line(ob, linenumb);
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/* Replaces a line of text in a panel */
void
fl_replace_panel_line(FL_OBJECT * ob, int linenumb, const char *newtext)
{
    if (linenumb < 1 || linenumb > ((SPEC *) (ob->spec))->lines)
	return;
    replace_line(ob, linenumb, newtext);
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/* Returns a pointer to a particular line in the panel. */
const char *
fl_get_panel_line(FL_OBJECT * ob, int linenumb)
{
    if (linenumb < 1 || linenumb > ((SPEC *) (ob->spec))->lines)
	return 0;
    return ((SPEC *) (ob->spec))->text[linenumb]->txt;
}

/* Sets the panel to a particular file */
int
fl_load_panel(FL_OBJECT * ob, const char *filename)
{
    SPEC *sp;
    FILE *fl;
    char newtext[FL_PANEL_LINELENGTH];
    int c, i;

    if (ob == 0 || ob->objclass != FL_PANEL)
	return 0;

    sp = ((SPEC *) (ob->spec));
    fl_clear_panel(ob);
    if (filename == 0 || filename[0] == 0)
    {
	set_partial(ob, 0);
	fl_redraw_object(ob);
	return 1;
    }
    /* LOAD THE FILE */
    fl = fopen(filename, "r");
    if (fl == 0)
	return 0;
    i = 0;
    do
    {
	c = getc(fl);
	if (c == NL || c == EOF)
	{
	    newtext[i] = 0;
	    insert_line(ob, sp->lines + 1, newtext);
	    i = 0;
	}
	else if (c == 9)	/* Tab */
	    do
		if (i < FL_PANEL_LINELENGTH - 1)
		    newtext[i++] = ' ';
	    while (i % 8 != 0 && i < FL_PANEL_LINELENGTH - 1) ;
	else if (i < FL_PANEL_LINELENGTH - 1)
	    newtext[i++] = c;
    }
    while (c != EOF);
    fclose(fl);
    if(sp->lines > 0){
	fl_set_panel_current(ob, 1);
    }
    set_partial(ob, 0);
    fl_redraw_object(ob);
    return 1;
}

/* Returns the number of lines in the panel. */
int
fl_get_panel_maxline(FL_OBJECT * ob)
{
    return ((SPEC *) (ob->spec))->lines;
}

/*-----------------------------------------------*/
/* Selects a line in the panel. */
void
fl_select_panel_line(FL_OBJECT * ob, int line)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    if (line < 1 || line > sp->lines) return;
    if(sp->text[line]->selected) return;      /* don't overwork yourself, HAK. */

    sp->text[line]->selected = 1;

    set_redraw(sp, abs(sp->selectline), TRUE);        
    set_redraw(sp, line, TRUE);        
    sp->selectline = line;

    set_partial(ob, 3);
    fl_redraw_object(ob);
}

/* Deselects a line in the panel. */
void
fl_deselect_panel_line(FL_OBJECT * ob, int line)
{
    SPEC *sp = ((SPEC *) (ob->spec));

    if (line < 1 || line > sp->lines)return;

    if(!sp->text[line]->selected) return;  /* don't overwork yourself, HAK. */

    sp->text[line]->selected = 0;
    set_redraw(sp, abs(sp->selectline), TRUE);        
    set_redraw(sp, line, TRUE);        
    sp->selectline = -line;

    set_partial(ob, 3);
    fl_redraw_object(ob);
}

/* Deselects all lines in the panel. */
void
fl_deselect_panel(FL_OBJECT * ob)
{
    SPEC *sp = ((SPEC *) (ob->spec));
    int i, n_visible_selected = 0, maxline;

    maxline = sp->lines;
    if (maxline > sp->topline + sp->screenlines -1) {
	maxline = sp->topline + sp->screenlines -1;
    }
    for (i=sp->topline; i<=maxline; i++){
	if (sp->text[i]->selected) {
	    n_visible_selected++;
	}
    }
    if (n_visible_selected < sp->screenlines/3 ) {
	/* try to be efficient */
	fl_freeze_form(ob->form);
	for (i = 1; i <= sp->lines; i++){
	    fl_deselect_panel_line(ob, i);
	}
	fl_unfreeze_form(ob->form);
	return;
    }
    for (i = 1; i <= sp->lines; i++){
	sp->text[i]->selected = 0;
    }
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/* Returns whether a line in the panel is selected. */
int
fl_isselected_panel_line(FL_OBJECT * ob, int line)
{
    if (line < 1 || line > ((SPEC *) (ob->spec))->lines)
	return 0;
    return ((SPEC *) (ob->spec))->text[line]->selected;
}

/* Returns the last selection in the panel. */
int
fl_get_panel(FL_OBJECT * ob)
{
    return ((SPEC *) (ob->spec))->selectline;
}

int
fl_get_panel_n_selected(FL_OBJECT *ob){
    int i, n_selected = 0;
    int maxline = fl_get_panel_maxline(ob);
    
    for(i=1; i <= maxline; i++){
	if(fl_isselected_panel_line(ob,i)) n_selected++;
    }
    return n_selected;
}

/*-----------------------------------------------*/

/* Sets the font size inside the panel. */
void
fl_set_panel_fontsize(FL_OBJECT * ob, int size)
{
    ((SPEC *) (ob->spec))->fontsize = size;
    ((SPEC *) (ob->spec))->attrib = 1;
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/*TCZ */
int
fl_get_panel_topline(FL_OBJECT * ob)
{
    return ((SPEC *) ob->spec)->topline;
}

/* Sets the font style inside the panel. */
void
fl_set_panel_fontstyle(FL_OBJECT * ob, int style)
{
    ((SPEC *) (ob->spec))->fontstyle = style;
    ((SPEC *) (ob->spec))->attrib = 1;
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/* Sets the escape key used in the text */
void
fl_set_panel_specialkey(FL_OBJECT * ob, char specialkey)
{
    ((SPEC *) (ob->spec))->specialkey = specialkey;
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/* Sets the bitmap key that triggers a file type bitmap */
void
fl_set_panel_bitmapkey(FL_OBJECT * ob, char bitmapkey)
{
    ((SPEC *) (ob->spec))->bitmapkey = bitmapkey;
    set_partial(ob, 0);
    fl_redraw_object(ob);
}

/*-----------------------------------------------*/
/* Gets the number of visible lines for the panel. */

int fl_get_panel_screenlines(FL_OBJECT *ob){
    int screenlines;		/* lines on screen */
    FL_Coord xx, yy, ww, hh;	/* text box */
    calc_textarea(ob, &xx, &yy, &ww, &hh, &screenlines);

    return screenlines;
}

void *fl_get_panel_line_client_data(FL_OBJECT *ob, int linenumb)
/* Returns a pointer to a particular line in the panel. 
 * Return even if the linenumb is > than lines but smaller than MAXLINES.
 * This is so that a destructor can access all potential lines. 
 */
{
  if (linenumb < 1 || linenumb > FL_PANEL_MAXLINE) return NULL; 
  if (((SPEC *)(ob->spec))->text[linenumb]){
      return ((SPEC *)(ob->spec))->text[linenumb]->client_data;
  } else { 
      return NULL;      
  }
}

void fl_set_panel_line_client_data(FL_OBJECT *ob, int linenumb, void *client_data)
/* Returns a pointer to a particular line's client data ptr in the panel. */
{
  if (linenumb < 1 || linenumb > ((SPEC *)(ob->spec))->lines) return;
  ((SPEC *)(ob->spec))->text[linenumb]->client_data = client_data;
}

int fl_get_panel_current(FL_OBJECT *ob)
/* HAK : Returns the current line in the panel. can be invalid (-1) */
{
    return ((SPEC *)(ob->spec))->currentline; 
}

int fl_get_panel_dnd_flag(FL_OBJECT *ob){
    return ((SPEC *)(ob->spec))->dnd_flag;
}

int fl_get_panel_dnd_track_valid(FL_OBJECT *ob){
    return ((SPEC *)(ob->spec))->dnd_valid_tracking_flag;
}

/* adjust currentline always, 
 * if in sensorregion: draw bitmap and return the line to indicate
 * valid tracking. 
 * ignore selected lines for xdnd-tracking. Don't change cursor
 * for the time being.
 * problem: reseting xdnd (i.e., state clear and redraw).
 * might be doable because current line is redrawn or hidden anyway,
 * and it is the only line affected in slave mode.
 */
int fl_get_panel_xdnd_track(FL_OBJECT *ob, FL_Coord mx, FL_Coord my){
    SPEC *sp = ((SPEC *) (ob->spec));
    float charheight = sp->charheight;
    int line, in_dnd_xregion_flag = FALSE;

    /* area check */
    if (my < ob->y || my >= ob->y+ob->h || 
	mx < ob->x+BSW || mx >= ob->x+ob->w
    ) {
	return 0;
    }

    /* compute line. If normalize necessary, no valid dnd possible. */
    line = (int) (sp->topline + (my - ob->y - FL_PANEL_BW) / charheight);
    if(line > sp->lines){
	line = sp->lines;
    } else if (line < 1){
	line = 1;
    } else {
	/* valtrack check */
	if ((mx >= ob->x + BSW) && 
	    (mx < ob->x + BSW + sp->dnd_sensor_width)
	){
	    in_dnd_xregion_flag = TRUE;
	}
    }
    sp->selectline = 0;
    sp->xdnd_valid_flag = in_dnd_xregion_flag;

    if (line == sp->currentline) { 
	draw_panel_bitmaps(ob, TRUE);    /* bitmap-only change is possible */
    } else {
	set_redraw(sp, sp->currentline, TRUE);
	set_redraw(sp, line, TRUE);
	sp->currentline = line;
	set_partial(ob, 3); 
	fl_redraw_object(ob);
    }
    sp->xdnd_valid_flag = FALSE;
    return in_dnd_xregion_flag;
}

int fl_get_panel_fontsize(FL_OBJECT *ob){
    /* Sets the font size inside the panel. */
    if(!ob) return 0;
    return ((SPEC *)(ob->spec))->fontsize;
}

int fl_get_panel_fontstyle(FL_OBJECT *ob)
/* Sets the font style inside the panel. */
{
    if(!ob) return 0;
    return  ((SPEC *)(ob->spec))->fontstyle;
}

void fl_set_panel_dnd_track_show_cb(show_func_pt tsfunc){
    track_show_func_fp = tsfunc;
}

void fl_set_panel_dnd_source_verify_cb(verify_func_pt svfunc){
    source_verify_func_fp = svfunc;
}

void fl_set_panel_dnd_track_external_cb(external_func_pt tefunc){
    track_external_func_fp = tefunc;
}

void fl_set_panel_dnd_execute_cb(exec_func_pt efunc){
    exec_func_fp = efunc;
}

void fl_set_panel_dnd_cleanup_cb(cleanup_func_pt cfunc){
    cleanup_func_fp = cfunc;
}

void fl_set_panel_cursor(FL_OBJECT *ob, int new_cursor){
    SPEC *sp = ((SPEC *) (ob->spec));
    if(ob && ob->form->window && new_cursor != sp->cursor && new_cursor != None) {
	fl_set_cursor_color(new_cursor, FL_RED, FL_WHITE);
	fl_set_cursor(ob->form->window, new_cursor);
	sp->cursor = new_cursor;
    }
}

void fl_set_newmulti_panel_colors( /* HAK >> */
	FL_OBJECT *ob,
	int SelCurLabelCol,
	int SelCurBackCol,
	int SelLabelCol,
	int SelBackCol,
	int CurLabelCol,
	int CurBackCol,
	int LabelCol,
	int BackCol
){
    SPEC *sp; if(!ob) return; sp = (SPEC *)(ob->spec);
    sp->SelectedCurrentLabelColor = SelCurLabelCol;
    sp->SelectedCurrentBackColor	= SelCurBackCol;
    sp->SelectedLabelColor	= SelLabelCol;
    ob->col2 				= SelBackCol;
    sp->CurrentLabelColor		= CurLabelCol;
    sp->CurrentBackColor		= CurBackCol;
    ob->lcol				= LabelCol;
    ob->col1 				= BackCol;
    set_partial(ob, 2);
    fl_redraw_object(ob);
}

void fl_hide_panel_current(FL_OBJECT *ob){
    SPEC *sp; 
    if(!ob) return; 
    sp = (SPEC *)(ob->spec);
    if(sp->current_is_hidden) return;
    sp->current_is_hidden = TRUE;
    set_partial(ob, 3); 
    set_redraw(sp, sp->currentline, TRUE);
    fl_redraw_object(ob);
}

void fl_unhide_panel_current(FL_OBJECT *ob){
    SPEC *sp; if(!ob) return; sp = (SPEC *)(ob->spec);
    if(!sp->current_is_hidden) return;
    sp->current_is_hidden = FALSE;
    set_partial(ob, 3); 
    set_redraw(sp, sp->currentline, TRUE);
    fl_redraw_object(ob);
}

void fl_center_panel_current(FL_OBJECT *ob){
  int maxtopline, newtopline, screenlines;
       
  SPEC *sp = ((SPEC *)(ob->spec));
  screenlines = fl_get_panel_screenlines(ob);
  maxtopline  = sp->lines - screenlines + 1;

  newtopline = sp->currentline - screenlines/2;
  if(newtopline < 1) newtopline = 1;
  if(newtopline > maxtopline) newtopline = maxtopline;
  fl_set_panel_topline(ob, newtopline);
}

/*-----------------------------------------------*/

void fl_set_panel_current(FL_OBJECT *ob, int line)
/* HAK:
 * Sets a 'current' line in the panel.
 * leaves selection status unchanged.
 * allows smooth scrolling if we just moved up/down from the last
 * current line.
 * is clever enough to redraw the panel partially if possible.
 */
{
    const int smoothskip = 2;  
    int screenlines;
    int smoothflag=0;   /* we're moving by one  line */
    int smoothtopline;  /* corrected topline (mind top/bottom ! */
    int oldtopline;     /* need for redraw check */
    int maxtopline;     /* maximally allowed line */
    int direction=1;    /* 1:down, -1: up */
    
    SPEC *sp = ((SPEC *)(ob->spec));
    if (line < 1 || line > sp->lines) return;
    if(line == sp->currentline) return;			/* already are*/
    
/*    fl_freeze_form(ob->form); */
    
    set_redraw(sp, line, TRUE);
    set_redraw(sp, sp->currentline, TRUE);
    
    screenlines = fl_get_panel_screenlines(ob);
    oldtopline = sp->topline;		
    maxtopline  = sp->lines - screenlines + 1;
    
    /* the following test is necessary if we don't have a previous currentline.*/
    if (sp->currentline >= 1 && sp->currentline <=  sp->lines){
	
	direction =  line - sp->currentline;              /* (can't if no current) */
	
	if(direction == 1){                               /* down */
	    if(sp->currentline != sp->topline + screenlines -1){   /* not the last line */
		smoothflag = 1;
	    }
	} else if(direction == -1){                       /* up */
	    if(sp->currentline != sp->topline){           /* not the first line */
		smoothflag = 1;
	    }
	}
    } else {
	/* no or invalid old current line */
    }
    sp->currentline = line;				/* remember   */
    
    if(smoothflag){
	/* fprintf(stderr, "smooth.\n"); */
	/* 
	 * are we in the upper margin, and are we going up ?
	 * also check wether we are at the top of the panel.
	 */
	if( (line < sp->topline + smoothskip) && direction == -1) {
	    smoothtopline = line - smoothskip;
	    /* this kludge ensures extra redraw if topline stays the
	     * same, and only current line moves.
	     */
	    if(oldtopline != smoothtopline && smoothtopline >= 1){  /* mind top border */
		fl_set_panel_topline(ob, smoothtopline);
		set_partial(ob, 3); /* new - want to redraw the current line and old current line. */
	    } else {
		/* else we're in the _very_ top lines, can't scroll further up.
		 * just redraw 
		 */
		set_partial(ob, 3);
	    }
	    
	} else if( (line >= sp->topline + screenlines - smoothskip) && direction == 1) {
	    /* are we in the lower margin, and going down ? */
	    smoothtopline = line - screenlines + smoothskip + 1;
	    /* are we in the _very_ bottom lines ? */
	    if(smoothtopline <= maxtopline){
		fl_set_panel_topline(ob, smoothtopline);
		set_partial(ob, 3); /* new - want to redraw the current line and old current line. */
	    } else if(oldtopline != maxtopline){   
		/* guard against pathologic case */
		fl_set_panel_topline(ob, maxtopline);		  
		set_partial(ob, 3); /* new - want to redraw the current line and old current line. */
	    } else {
		/* else nothing special, just redraw */
		set_partial(ob, 3);
	    }
	} else {
	    /* else nothing special, just redraw */
	    set_partial(ob, 3);
	}
    } else {
	/* attempt to be nice, and center the new current line ? */
	/* that could conflict a lot with moving up by one step !*/
	if(line < sp->topline){
	    fl_set_panel_topline(ob, line);
	} else if(line >= sp->topline + screenlines){
	    fl_set_panel_topline(ob, line-screenlines+1);
	} else {
	    /* else nothing special, we're just moving smoothly. just redraw */
	    set_partial(ob, 3);
	}
    }
    fl_redraw_object(ob);
/*    fl_unfreeze_form(ob->form);  */
}

void fl_set_panel_colors( /* HAK >> */
	FL_OBJECT *ob,
	int SelCurLabelCol,
	int SelCurBackCol,
	int SelLabelCol,
	int SelBackCol,
	int CurLabelCol,
	int CurBackCol,
	int LabelCol,
	int BackCol
){
    SPEC *sp; if(!ob) return; sp = (SPEC *)(ob->spec);
    sp->SelectedCurrentLabelColor = SelCurLabelCol;
    sp->SelectedCurrentBackColor	= SelCurBackCol;
    sp->SelectedLabelColor	= SelLabelCol;
    ob->col2 				= SelBackCol;
    sp->CurrentLabelColor		= CurLabelCol;
    sp->CurrentBackColor		= CurBackCol;
    ob->lcol				= LabelCol;
    ob->col1 				= BackCol;
    set_partial(ob, 0);
    sp->attrib = 1;
    fl_redraw_object(ob);
}

void fl_get_panel_colors(FL_OBJECT *ob, 
	int *SelCurLabelCol,
	int *SelCurBackCol,
	int *SelLabelCol,
	int *SelBackCol,
	int *CurLabelCol,
	int *CurBackCol,
	int *LabelCol,
	int *BackCol
){
    SPEC *sp; if(!ob) return; sp = (SPEC *)(ob->spec);
    *SelCurLabelCol= sp->SelectedCurrentLabelColor ; 
    *SelCurBackCol = sp->SelectedCurrentBackColor	; 
    *SelLabelCol   = sp->SelectedLabelColor	; 
    *SelBackCol    = ob->col2 				; 
    *CurLabelCol   = sp->CurrentLabelColor	;
    *CurBackCol    = sp->CurrentBackColor	; 
    *LabelCol      = ob->lcol				; 
    *BackCol       = ob->col1 				; 
}
