/*
 * $Id: panel.h,v 1.1 1996/10/26 21:17:19 henrik Exp henrik $
 *
 *  Object class Panel
 */
#ifndef FL_PANEL_H
#define FL_PANEL_H		/* { */

#if defined(__cplusplus)
extern "C"
{
#endif

/***** Class ID *****/
#define FL_PANEL 1464   // hopefully random between 1k and 10k :)

/***** Types    *****/
enum panels
{
    FL_NORMAL_PANEL
};

/***** Defaults *****/

#define FL_PANEL_BOXTYPE	FL_DOWN_BOX
/*
#define FL_PANEL_COL1	FL_COL1
#define FL_PANEL_COL2	FL_YELLOW
#define FL_PANEL_LCOL	FL_LCOL 
*/
#define FL_PANEL_ALIGN	FL_ALIGN_BOTTOM

#define FL_PANEL_SELCUR_LCOL	FL_YELLOW	/* yellow HAK */
#define FL_PANEL_SELCUR_BCOL	FL_CYAN 	/* lblue  HAK */
#define FL_PANEL_SEL_LCOL	FL_YELLOW	/* yellow HAK */
#define FL_PANEL_SEL_BCOL	FL_BLUE	        /* blue   HAK */
#define FL_PANEL_CUR_LCOL	FL_WHITE	/* white  HAK */
#define FL_PANEL_CUR_BCOL	FL_CYAN 	/* lblue   HAK */
#define FL_PANEL_LCOL	        FL_WHITE	/* white  HAK */
#define FL_PANEL_BCOL	        FL_BLUE	        /* blue   HAK */

/***** Others   *****/

#define FL_PANEL_SLCOL		FL_COL1
#define FL_PANEL_BW		FL_BOUND_WIDTH
#define FL_PANEL_LINELENGTH	8192  /* for loading only, storage is dynamic. HAK */
#define FL_PANEL_MAXLINE	8192
#define FL_PANEL_TEXTMARGIN	5

/* DND functions.
 * external:  track externally. (implicit outside flag)
 * verify:    verify source target (i.e., no dot/dotdot). 
 * exec:      on release, query master/slave state and checkexec.
 * show:      track internally, feedback to state display.
 */
typedef int  (*external_func_pt) (FL_OBJECT *ob);
typedef int  (*verify_func_pt)   (FL_OBJECT *ob, int line, XEvent *xev);  
typedef void (*cleanup_func_pt)  (FL_OBJECT *ob);
typedef void (*exec_func_pt)     (FL_OBJECT *ob);
typedef void (*show_func_pt)     (FL_OBJECT *ob, int line);  

/***** Routines *****/
FL_OBJECT *
fl_create_panel(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
		  const char *label);

extern FL_OBJECT *fl_add_panel(int, FL_Coord, FL_Coord, FL_Coord, FL_Coord,
				 const char *);

extern void fl_set_panel_topline(FL_OBJECT *, int);
extern int fl_get_panel_topline(FL_OBJECT *);
extern void fl_clear_panel(FL_OBJECT *);
extern int fl_add_panel_line(FL_OBJECT *, const char *);    /* return new line no */
extern int fl_addto_panel(FL_OBJECT *, const char *);
extern int fl_insert_panel_line(FL_OBJECT *, int, const char *);
extern void fl_delete_panel_line(FL_OBJECT *, int);
extern void fl_replace_panel_line(FL_OBJECT *, int, const char *);
extern const char *fl_get_panel_line(FL_OBJECT *, int);
extern int fl_load_panel(FL_OBJECT *, const char *);
extern int fl_get_panel_maxline(FL_OBJECT *);
extern void fl_select_panel_line(FL_OBJECT *, int);
extern void fl_deselect_panel_line(FL_OBJECT *, int);
extern void fl_deselect_panel(FL_OBJECT *);
extern int fl_isselected_panel_line(FL_OBJECT *, int);
extern int fl_get_panel(FL_OBJECT *);
extern void fl_set_panel_fontsize(FL_OBJECT *, int);
extern void fl_set_panel_fontstyle(FL_OBJECT *, int);
extern void fl_set_panel_specialkey(FL_OBJECT *, char);
extern void fl_set_panel_bitmapkey(FL_OBJECT *, char);

/* HAK */
extern void fl_hide_panel_current(FL_OBJECT *);
extern void fl_unhide_panel_current(FL_OBJECT *);
extern void fl_center_panel_current(FL_OBJECT *);


extern int fl_get_panel_screenlines    (FL_OBJECT *ob);
extern int fl_get_panel_current        (FL_OBJECT *ob);
extern int fl_get_panel_fontsize       (FL_OBJECT *ob);
extern int fl_get_panel_fontstyle      (FL_OBJECT *ob);
extern int fl_get_panel_n_selected     (FL_OBJECT *ob);
extern int fl_get_panel_dnd_flag       (FL_OBJECT *ob);
extern int fl_get_panel_dnd_track_valid(FL_OBJECT *ob);
extern int fl_get_panel_xdnd_track     (FL_OBJECT *ob, FL_Coord mx, FL_Coord my);

extern void fl_set_panel_current              (FL_OBJECT *ob, int line);
extern void fl_set_panel_cursor               (FL_OBJECT *ob, int new_cursor);
extern void fl_set_panel_dnd_cleanup_cb       (cleanup_func_pt cfunc);
extern void fl_set_panel_dnd_execute_cb       (exec_func_pt efunc);
extern void fl_set_panel_dnd_source_verify_cb (verify_func_pt svfunc);
extern void fl_set_panel_dnd_track_show_cb    (show_func_pt stfunc);
extern void fl_set_panel_dnd_track_external_cb(external_func_pt tefunc);


/* each line can have client data (e.g., a struct stat). You are
 * responsible to manage its allocation/deallocation.
 * client data can persist even if its line number is invalid, so
 * check wether the ptr is NULL. 
 */

extern void *fl_get_panel_line_client_data(FL_OBJECT *ob, int linenumb);
extern void fl_set_panel_line_client_data(FL_OBJECT *ob, int linenumb, void *client_data);


/* the following is rather `on foot'. */

extern void fl_set_panel_colors(
	FL_OBJECT *ob,
	int SelCurLabelCol,
	int SelCurBackCol,
	int SelLabelCol,
	int SelBackCol,
	int CurLabelCol,
	int CurBackCol,
	int LabelCol,
	int BackCol
);

extern void fl_get_panel_colors(FL_OBJECT *ob, 
	int *SelCurLabelCol,
	int *SelCurBackCol,
	int *SelLabelCol,
	int *SelBackCol,
	int *CurLabelCol,
	int *CurBackCol,
	int *LabelCol,
	int *BackCol
);

#if defined(__cplusplus)
}
#endif

#endif /* PANEL_H } */

