
/* pathnames.C:
 * try to get the best of both worlds, fixed path names as long
 * as possible, searching the best ones first, allowing system
 * specific defaults, and fullblown shell searching.
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include "global.h"
#include "pathnames.h"

const char 
    *chgrp_path    = "/bin/chgrp",
    *chmod_path    = "/usr/bin/chmod",
    *chown_path    = "/bin/chown",
    *cp_path       = "/bin/cp",
    *mkdir_path    = "/bin/mkdir",
    *mv_path       = "/bin/mv",
    *rm_path       = "/bin/rm",
    *touch_path    = "/bin/touch",
#ifdef SGI
    *xwsh_path     = "/usr/sbin/xwsh",
    *winterm_path  = "/usr/bin/X11/winterm",
#endif
    *xterm_path    = "/usr/bin/X11/xterm"    
//    *xterm_path  = "/usr/bin/X11/hpterm"    
;

#if defined(LINUX) || defined(GENERICBSD)
const char *make_path   = "/usr/bin/make";
#elif defined(UNIXWARE)
const char *make_path   = "/usr/ccs/bin/make";
#else
const char *make_path   = "/bin/make";
#endif

//static int dbg_flag = TRUE;
static int dbg_flag = FALSE;

// /usr/ccs/bin is very important on SVR4 (make)

char *standard_base_dirs[] = {
    "/bin", "/usr/bin", "/etc", "/usr/etc", "/sbin", "/usr/sbin", 
    "/usr/bin/X11", "/usr/X11/bin", "/usr/share/bin", "/usr/bsd/", 
    "/usr/local/bin", "/opt/bin", "/usr/ucb", "/usr/ccs/bin",
    "/usr/openwin/bin",
    "/usr/X11R6/bin", "/usr/X11R5/bin", "/usr/X11R4/bin",
    "/usr/X386/bin", "/usr/x386/bin", 
#ifdef GENERICBSD
    "/usr/X11R6/bin",
#endif
    NULL 
};

char **base_dirs = NULL;

static int construct_path(const char *basename, const char * &full_path, int fatal_flag = TRUE){
    if(!basename){
	cerr << "UC: Null basename in construct_path.\n";
	return FALSE;
    }
    if(full_path){
	// is the default good ?
#if defined(SGI) && !defined(SGI_IRIX4)	
	if(access(full_path, EX_OK) == 0){
#else
	if(access(full_path, X_OK) == 0){
#endif
	    if(dbg_flag) cerr << "fullpath is good: " << full_path << endl;
	    return TRUE;
	}
    }
    int i=0;
    while(base_dirs[i] != NULL){	
	char buffer[MAXPATHLEN+1];
	sprintf(buffer, "%s/%s", base_dirs[i], basename);
	if(dbg_flag) cerr << "    checking buffer: " << buffer << endl;
#if defined(SGI) && !defined(SGI_IRIX4)	
	if(access(buffer, EX_OK) == 0){
#else
	if(access(buffer, X_OK) == 0){
#endif
	    full_path = strdup(buffer);
	    if(dbg_flag) cerr << "new fullpath is good: " << full_path << endl;
	    return TRUE;
	} else {
	    i++;
	}
    }
    if(fatal_flag){
        cerr << "UC: Fatal error: No regular path found for Unix command `" << basename << "'.\n";
        cerr << "    Unix Cockpit can't run without this auxiliary program.\n";
	cerr << "    Please check your path and Unix installation.\n";
	exit(-1);
    } else {
        cerr << "UC: Warning: No regular path found for Unix command `" << basename << "'.\n";
        cerr << "    Some functionality won't be there without this auxiliary command.\n";
	cerr << "    Please check your path and Unix installation.\n";
	full_path = NULL;
    }
    return FALSE;
}

void init_pathnames(void){
    const char *path = getenv("PATH");
    if(!path){
        base_dirs = standard_base_dirs;
    } else {
	int count = 0;
	const char *tmp = path;
	while((tmp = strchr(tmp, ':'))){
	    count++;
	    tmp++;  // jump past `:'
	}
	int i=0;
	while(standard_base_dirs[i]){
	    count++;
	    i++;
	}
	count+=3; // for extra leading or trailing or missing colon's.
	base_dirs = new char*[count];
	for(i=0;i<count;i++){
	    base_dirs[i] = NULL;  // just in case.
	}
	i=0;
	while(standard_base_dirs[i]){
	    base_dirs[i] = standard_base_dirs[i];
	    i++;
	}
	char *pathbuf = strdup(path);
	char *tok = strtok(pathbuf, ":");
	while(tok && i<count-1){
	    if(strlen(tok) > 0 && tok[0] == '/'){  // abs paths only
		if(dbg_flag){
		    cerr << "path component: " << tok << endl;
		}
		base_dirs[i] = strdup(tok);
		i++;
	    } else {
		base_dirs[i] = NULL;
	    }		
	    tok = strtok(NULL, ":");
	}
	//cerr << "i,count:" << i << "," << count << endl;
    }
    construct_path("chgrp",    chgrp_path, FALSE);
    construct_path("chmod",    chmod_path, FALSE);
    construct_path("chown",    chown_path, FALSE);
    construct_path("cp",       cp_path);  
    construct_path("make",     make_path, FALSE);
    construct_path("mkdir",    mkdir_path);
    construct_path("mv",       mv_path);
    construct_path("rm",       rm_path);
    construct_path("touch",    touch_path, FALSE);
//    construct_path("hpterm", xterm_path);
    construct_path("xterm",    xterm_path);
#ifdef SGI
    construct_path("xwsh",     xwsh_path);
    construct_path("winterm",  winterm_path);
#endif
    
//    construct_path("",  _path);
}

/*
main(){
    init_pathnames();
}
*/



