
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "resource.h"
#include "utility.h"

void prm_read_val(
    prm_entry_t *entry,
    const char *val_tok
){
    if(!val_tok){
	error_msg("Invalid value token in resource parameter read.");
	return;
    }
    switch(entry->type){
    case char_p_e:
	//if(entry->data_p){
	//    fl_free(entry->data_p); entry->data_p = NULL; }
	//entry->data_p = (void *) strdup(val_tok);
	strcpy((char *) entry->data_p, val_tok);
	break;
    case double_e:
	*((double*) entry->data_p) = atof(val_tok);
	break;
    case int_e:
	*((int*)    entry->data_p) = atoi(val_tok);
	break;
    case long_e:
	*((long*)   entry->data_p) = atoi(val_tok);
	break;
    case short_e:
	*((short*)  entry->data_p) = atoi(val_tok);
	break;
    }
}

void prm_write_val(ofstream &ofile, prm_entry_t *entry){
    if(!entry) return;
    if(!entry->data_p) return;
    switch(entry->type){
    case char_p_e:
	ofile << (char *)entry->data_p;
	break;
    case double_e:
	ofile << *((double *)entry->data_p);
	break;
    case int_e:
	ofile << *((int *)entry->data_p);
	break;
    case long_e:
	ofile << *((long *)entry->data_p);
	break;
    case short_e:
	ofile << *((short *)entry->data_p);
	break;
    }
}    

int prm_read_name(
    prm_entry_t prm_table[],
    const char *name_tok,
    const char *val_tok
){
    if(!prm_table || !name_tok || !val_tok) return FALSE;
    prm_entry_t *entry = prm_table;
    while(entry->name[0] != '\0'){
	if(strcmp(entry->name, name_tok) == 0){
	    prm_read_val(entry, val_tok);
	    return TRUE;
	}
	entry++;
    }
    return FALSE;
}

int resource_load_default(
    prm_entry_t prm_table[],  
    const char *ID_string,    
    const char *obj_tok,      
    const char *name_tok,     
    const char *val_tok       
){
    if(!ID_string || !prm_table || !obj_tok || !name_tok || !val_tok){
	//M_trace("resource::load_default", "null pointers input.");
	return FALSE;
    }
    if(strcmp(obj_tok, ID_string) != 0){
	//M_trace("resource::load_default", "ID string failure for %s != %s", obj_tok, ID_string);
	return FALSE;
    }
    
    if(!prm_read_name(prm_table, name_tok, val_tok)){
	error_msg("Unrecognized database entry: %s.%s", obj_tok, name_tok);
	return FALSE;
    } else {
	//M_trace("resource::load_default", "success for %s.%s:%s", obj_tok, name_tok, val_tok);
	return TRUE;
    }
}

const int n_blanks = 30;
void write_blanks(ofstream &ofile, int n){
    for(int i=0; i<n; i++){
	ofile << " ";
    }
}

void resource_save_defaults(
    ofstream &ofile,
    prm_entry_t prm_table[],  // supplied `keypad_prm_table'
    const char *ID_string     // supplied `keypad_ID_string'
){
    ofile << "# " << ID_string << " resources" << endl;
    ofile << "#" << endl; 
    prm_entry_t *entry = prm_table;
    while(entry->name[0] != '\0'){
	ofile << ID_string << ".";
	ofile << entry->name << ": ";
	write_blanks(ofile, n_blanks - strlen(entry->name));
	prm_write_val(ofile, entry);
	ofile << endl;
	entry++;
    } 
    ofile << endl; 
}
