
/*
 * Scan the directory dirname calling select to make a list of selected
 * directory entries then sort using qsort and compare routine dcomp. Returns
 * the number of entries and a pointer to a list of pointers to struct direct
 * (through namelist). Returns -1 if there were any errors. 
 *
 *  SCANDIR
 *  Scan a directory, collecting all (selected) items into a an array.
 *  for the totally hosed solaris 5.3 BSD lib.
 */

#include "global.h"

#if defined(SOLARIS) || defined(UNIXWARE) || defined(SCO) || defined(SINIX)

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/file.h>
#include <unistd.h>

#include "global.h"

/* Initial guess at directory size. */
# define INITIAL_SIZE	20

/* A convenient shorthand. */
typedef struct dirent	 ENTRY;
	    
# ifndef DIRSIZ
#  define DIRSIZ(d) (sizeof(struct dirent) + strlen(d->d_name) + 1) 
# endif


int scandir(
    const char *Name, 
    ENTRY **List[], 
    int (*Selector)(ENTRY*), 
    int (*Sorter)(ENTRY**, ENTRY**) 
){
    register ENTRY	 **names;
    register ENTRY	  *E;
    register DIR	  *Dp;
    register int	   i;
    register int	   size;

    /* Get initial list space and open directory. */
    size = INITIAL_SIZE;
    if ((names = (ENTRY **) fl_malloc(size * sizeof names[0])) == NULL
     || access(Name, R_OK | X_OK)
     || (Dp = opendir(Name)) == NULL)
	return(-1);

    /* Read entries in the directory. */
    for (i = 0; E = readdir(Dp); )
	if (Selector == NULL || (*Selector)(E)) {
	    /* User wants them all, or he wants this one. */
	    if (++i >= size) {
		size <<= 1;
		names = (ENTRY **) fl_realloc((char *)names, size * sizeof names[0]);
		if (names == NULL) {
		    closedir(Dp);
		    return(-1);
		}
	    }

	    /* Copy the entry. */
	    if ((names[i - 1] = (ENTRY *) fl_malloc(DIRSIZ(E))) == NULL) { 
		closedir(Dp);
		return(-1);
	    }
	    names[i - 1]->d_ino = E->d_ino;
	    names[i - 1]->d_reclen = E->d_reclen;
	 /*   names[i - 1]->d_namlen = E->d_namlen; */
	    (void)strcpy(names[i - 1]->d_name, E->d_name);
	}

    /* Close things off. */
    names[i] = NULL;
    *List = names;
    closedir(Dp);

    /* Sort? */
    if (i && Sorter)
	qsort((char *)names, i, sizeof names[0], 
	      (int (*)(const void*,const void*)) Sorter
	      );

    return(i);
}

/*
 * Alphabetic order comparison routine for those who want it. 
 */
int alphasort(ENTRY **d1, ENTRY **d2){
    return (strcmp((*d1)->d_name, (*d2)->d_name));
}

#ifdef UNDER_DEVELOPMENT

#include <iostream.h>

int scandir_select(struct dirent *entry){
  cout << "select on: " << entry->d_name << endl;
  return 1;
}

main(){
  struct dirent **namelist;
  int n_files = scandir(".", &namelist, scandir_select, alphasort);
  for (int i=0; i<n_files; i++){
    cout << "output namelist: " << namelist[i]->d_name << endl;
  }

  return 0;
}

#endif

#endif
