
#include <iostream.h>

#include "cb_key.h"
#include "forms.h"
#include "global.h"
#include "gl_utility.h"

static void scroll_center(FL_OBJECT *browser){
    if(browser->type == FL_NORMAL_BROWSER) return;
    int maxline, n_lines, newline;
    maxline = fl_get_browser_maxline(browser);
    n_lines = fl_get_browser_screenlines(browser);
    int line = fl_get_browser(browser);
    if(line > 0){
	newline = line - n_lines/2;
	fl_set_browser_topline(browser, newline);
    }
}

static void scroll_end(FL_OBJECT *browser){
    int maxline, n_lines, newline;
    maxline = fl_get_browser_maxline(browser);
    n_lines = fl_get_browser_screenlines(browser);
    newline = maxline - n_lines + 1;
    if(newline <= 0) newline = 1;
    fl_set_browser_topline(browser, newline);
}

static void scroll_home(FL_OBJECT *browser){
    fl_set_browser_topline(browser, 1);
}

static void scroll_down_arrow(FL_OBJECT *browser){
    if(!down_arrow_key_verify()) return;    
    int maxline, topline, n_lines, newline;
    maxline = fl_get_browser_maxline(browser);
    topline = fl_get_browser_topline(browser);
    n_lines = fl_get_browser_screenlines(browser);
    
    newline = topline + 1;
    if(maxline >= newline + n_lines -1){
	fl_set_browser_topline(browser, newline);
    }
}

static void scroll_up_arrow(FL_OBJECT *browser){
    if(!up_arrow_key_verify()) return;    
    int topline, newline;
    topline = fl_get_browser_topline(browser);
    newline = topline - 1;
    if(newline > 0){
	fl_set_browser_topline(browser, newline);
    }
}

static void scroll_page_down(FL_OBJECT *browser){
    int maxline, topline, n_lines, newline;
    maxline = fl_get_browser_maxline(browser);
    topline = fl_get_browser_topline(browser);
    n_lines = fl_get_browser_screenlines(browser);
    
    newline = topline + n_lines;
    if(maxline < newline + n_lines -1){
	newline = maxline - n_lines + 1;
    }
    if(newline < 0) newline = 1;
    fl_set_browser_topline(browser, newline);
}

static void scroll_page_up(FL_OBJECT *browser){
    int topline, newline, n_lines;
    topline = fl_get_browser_topline(browser);
    n_lines = fl_get_browser_screenlines(browser);
    newline = topline - n_lines;
    if(newline <= 0) newline = 1;
    fl_set_browser_topline(browser, newline);
}

static int scroll_key_event(FL_OBJECT *browser, XKeyEvent *, KeySym keysym){
    if(tool_get_ctrl()){
	switch(keysym){
	case XK_a: case XK_A: scroll_home        (browser); return TRUE;
	case XK_e: case XK_E: scroll_end         (browser); return TRUE;
	case XK_l: case XK_L: scroll_center      (browser); return TRUE;
	case XK_n: case XK_N: scroll_down_arrow  (browser); return TRUE;
	case XK_p: case XK_P: scroll_up_arrow    (browser); return TRUE;
	case XK_v: case XK_V: scroll_page_down   (browser); return TRUE;
	default:                                            return FALSE;
	}
    }
    if(tool_get_alt()){
	switch(keysym){
	case XK_less:         scroll_home      (browser); return TRUE;
	case XK_greater:      scroll_end       (browser); return TRUE;
	case XK_v: case XK_V: scroll_page_up   (browser); return TRUE;
	default:                                          return FALSE;
	}
    }
    return FALSE;
}

int scroll_browser_event_cb(FL_OBJECT *browser, XEvent *xevent, int use_emacs_flag){
    if(!browser) return FALSE;
    // distinguish between pad events, which can be shifted and thus disabled, 
    // or normal `home' or `end' etc. events, which are always valid.
    // we want a browser * as input, manipulate it if we can, and say wether
    // we did.
    KeySym keysym;
    char keybuf[1];
    XKeyEvent *xkev= (XKeyEvent *) xevent;
    
    switch(xevent->type){
    case KeyPress:
	XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);

	// compressed switch :)
	if(is_end(keysym, xkev)){
	    scroll_end(browser);
	} else if(is_down(keysym, xkev)){
	    scroll_down_arrow(browser);
	} else if(is_page_down(keysym, xkev)){
	    scroll_page_down(browser);
	} else if(is_begin(keysym, xkev)){
	    scroll_center(browser);
	} else if(is_home(keysym, xkev)){
	    scroll_home(browser);
	} else if(is_up(keysym, xkev)){
	    scroll_up_arrow(browser);
	} else if(is_page_up(keysym, xkev)){
	    scroll_page_up(browser);    
	} else if(use_emacs_flag){
	    return scroll_key_event(browser, xkev, keysym);
	} else {
	    return FALSE;
	}
	return TRUE;
    default:
	return FALSE;
    }
}


