#include <sys/param.h>

#include "browser.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "sort_forms.h"
#include "sort_forms_aux.h"

void sort_form_help_cb(void){
    SortHelpButtonCB(NULL, 0);
}

void SortHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(sort_help_text, "Sort Options");
}

int sort_form_raw_cb(FL_FORM *form, void *xev){
    sorttype_et new_sort;
    XEvent *xevent = (XEvent *)xev;
    if(adjust_return_button(form, xevent)) return TRUE;
    if(standard_form_event_cb(form, xevent, NULL, sort_form_help_cb)) return TRUE;
    if(xevent->type == KeyPress){
	KeySym keysym;
	char keybuf[1];
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	case XK_a:
	case XK_A: new_sort = access_e; break; 	    
	case XK_f:
	case XK_F: new_sort = filetype_e; break; 	    
	case XK_g:
	case XK_G: new_sort = group_e; break; 	    
	case XK_i:
	case XK_I: new_sort = inode_e; break; 	    
	case XK_m:
	case XK_M: new_sort = modific_e; break;
	case XK_n:
	case XK_N: new_sort = name_e; break;
	case XK_l:
	case XK_L: new_sort = nlinks_e; break;
	case XK_o:
	case XK_O: new_sort = owner_e; break;
	case XK_s:
	case XK_S: new_sort = size_e; break; 	    
	case XK_x:
	case XK_X: new_sort = extension_e; break; 	    
	default:  return FALSE;
	}
	if(tool_get_ctrl()){
	    source_browser->snd_sort = new_sort;
	} else {
	    source_browser->fst_sort = new_sort;
	}
	sort_form_sync();
	source_browser->update(TRUE);
	return TRUE;
    }
    return FALSE;
}

static void sort_set_wintitle(void){
    char buffer[MAXPATHLEN];
    sprintf(buffer, "Adjust Browser #%d Sort Options:", source_browser->get_ID());
    XStoreName(fl_display, SortForm->window, buffer);
}

void sort_show_form(void){
    static int first = TRUE;
    if(first){
	fl_set_object_shortcut(SortAccessButton,       "Aa", TRUE);
	fl_set_object_shortcut(SortFileTypeButton,     "Ff", TRUE);
	fl_set_object_shortcut(SortGroupButton,        "Gg", TRUE);
	fl_set_object_shortcut(SortInodeButton,        "Ii", TRUE);
	fl_set_object_shortcut(SortModificationButton, "Mm", TRUE);
	fl_set_object_shortcut(SortNameButton,         "Nn", TRUE);
	fl_set_object_shortcut(SortNLinksButton,       "Ll", TRUE);
	fl_set_object_shortcut(SortOwnerButton,        "Oo", TRUE);
	fl_set_object_shortcut(SortSizeButton,         "Ss", TRUE);
	fl_set_object_shortcut(SortExtensionButton,    "xX", TRUE);

	fl_set_object_shortcut(AuxSortAccessButton,       "^A^a", TRUE);
	fl_set_object_shortcut(AuxSortFileTypeButton,     "^F^f", TRUE);
	fl_set_object_shortcut(AuxSortGroupButton,        "^G^g", TRUE);
	fl_set_object_shortcut(AuxSortInodeButton,        "^I^i", TRUE);
	fl_set_object_shortcut(AuxSortModificationButton, "^M^m", TRUE);
	fl_set_object_shortcut(AuxSortNameButton,         "^N^n", TRUE);
	fl_set_object_shortcut(AuxSortNLinksButton,       "^L^l", TRUE);
	fl_set_object_shortcut(AuxSortOwnerButton,        "^O^o", TRUE);
	fl_set_object_shortcut(AuxSortSizeButton,         "^S^s", TRUE);
	fl_set_object_shortcut(AuxSortExtensionButton,    "^x^X", TRUE);
	
	fl_set_button_shortcut(SortReverseButton,     "Rr", TRUE);
	fl_set_button_shortcut(AuxSortReverseButton,  "^R^r", TRUE);
	fl_set_button_shortcut(SortSyncAllButton,   "^yY", TRUE);
	fl_set_button_shortcut(SortDefaultsButton,    "Dd", TRUE);
	fl_set_object_shortcut(SortCloseReturnButton,  "Cc^M", TRUE);
	fl_register_call_back(SortForm, FL_ALL_EVENT, sort_form_raw_cb);
	first = FALSE;
    }
    set_to_ret_button(SortCloseReturnButton);
    sort_form_sync();
    focus_save();
    if(!SortForm->visible){
	fl_show_form(SortForm, FL_PLACE_CENTER, TRUE, "Adjust Sort Options:");
	sort_set_wintitle();
    } else {
	XMapRaised(fl_display, SortForm->window);
    }
    tool_focus_move(SortForm, SortCloseReturnButton);
}

void sort_form_sync(void){
    M_trace("sort_form_sync", "start");
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    fl_freeze_form(SortForm);
    fl_set_button(SortAccessButton, 0);
    fl_set_button(SortFileTypeButton, 0);
    fl_set_button(SortGroupButton, 0);
    fl_set_button(SortInodeButton, 0);
    fl_set_button(SortModificationButton, 0);
    fl_set_button(SortNameButton, 0);
    fl_set_button(SortNLinksButton, 0);
    fl_set_button(SortOwnerButton, 0);
    fl_set_button(SortSizeButton, 0);
    fl_set_button(SortExtensionButton, 0);

    fl_set_button(AuxSortAccessButton, 0);
    fl_set_button(AuxSortFileTypeButton, 0);
    fl_set_button(AuxSortGroupButton, 0);
    fl_set_button(AuxSortInodeButton, 0);
    fl_set_button(AuxSortModificationButton, 0);
    fl_set_button(AuxSortNameButton, 0);
    fl_set_button(AuxSortNLinksButton, 0);
    fl_set_button(AuxSortOwnerButton, 0);
    fl_set_button(AuxSortSizeButton, 0);
    fl_set_button(AuxSortExtensionButton, 0);

    fl_set_button(SortReverseButton,    source_browser->sort_fst_reverse_flag);
    fl_set_button(AuxSortReverseButton, source_browser->sort_snd_reverse_flag);
	    
    switch(source_browser->fst_sort){
    case access_e		: fl_set_button(SortAccessButton, 1);       break; 
    case filetype_e     : fl_set_button(SortFileTypeButton, 1);     break; 
    case group_e 		: fl_set_button(SortGroupButton, 1);        break; 
    case inode_e 		: fl_set_button(SortInodeButton, 1);        break; 
    case modific_e		: fl_set_button(SortModificationButton, 1); break; 
    case name_e  		: fl_set_button(SortNameButton, 1);         break; 
    case nlinks_e		: fl_set_button(SortNLinksButton, 1);       break; 
    case owner_e		: fl_set_button(SortOwnerButton, 1);        break; 
    case size_e 		: fl_set_button(SortSizeButton, 1);         break; 
    case extension_e 	: fl_set_button(SortExtensionButton, 1);    break; 
    default             : program_unstable("UC internal: browser_c:: unknown sort option.");
    }
    
    switch(source_browser->snd_sort){
    case access_e	: fl_set_button(AuxSortAccessButton, 1);   break; 
    case filetype_e     : fl_set_button(AuxSortFileTypeButton, 1); break; 
    case group_e 	: fl_set_button(AuxSortGroupButton, 1);    break; 
    case inode_e 	: fl_set_button(AuxSortInodeButton, 1);    break; 
    case modific_e	: fl_set_button(AuxSortModificationButton, 1); break; 
    case name_e 	: fl_set_button(AuxSortNameButton, 1);     break; 
    case nlinks_e	: fl_set_button(AuxSortNLinksButton, 1);   break; 
    case owner_e	: fl_set_button(AuxSortOwnerButton, 1);    break; 
    case size_e 	: fl_set_button(AuxSortSizeButton, 1);     break; 
    case extension_e: fl_set_button(AuxSortExtensionButton, 1);    break; 
    default             : program_unstable("UC internal: browser_c:: unknown sort option.");
    }
    fl_unfreeze_form(SortForm);
    sort_set_wintitle();
}

void SortCloseReturnButtonCB(FL_OBJECT *, long){
    if(SortForm->visible){
	focus_restore_start();
	fl_hide_form(SortForm);
	focus_restore_finish();
	source_browser->update(TRUE);
    }
}

void SortReverseButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    toggle_flag(source_browser->sort_fst_reverse_flag);
    source_browser->update(TRUE);
}

void AuxSortReverseButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    toggle_flag(source_browser->sort_snd_reverse_flag);
    source_browser->update(TRUE);
}

void SortSyncAllButtonCB(FL_OBJECT *, long){
    browser_c *tmp_browser = source_browser->next_browser;
    while(tmp_browser != source_browser){
	tmp_browser->freeze();
	tmp_browser->sync_sort_opts(source_browser);
	tmp_browser->update(TRUE);
	tmp_browser->unfreeze();
	tmp_browser = tmp_browser->next_browser;
    }
}

void SortDefaultsButtonCB(FL_OBJECT *, long){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    source_browser->set_sort_defaults();
    source_browser->update(TRUE);
    sort_form_sync();
}

void AuxSortCB(FL_OBJECT *, long option){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    switch(option){
    case access_e:   source_browser->snd_sort = access_e;  break; 
    case filetype_e: source_browser->snd_sort = filetype_e;break; 
    case group_e:    source_browser->snd_sort = group_e;   break; 
    case inode_e:    source_browser->snd_sort = inode_e;   break; 
    case modific_e:  source_browser->snd_sort = modific_e; break; 
    case nlinks_e:   source_browser->snd_sort = nlinks_e;  break; 
    case owner_e:    source_browser->snd_sort = owner_e;   break; 
    case name_e:     source_browser->snd_sort = name_e;   break; 
    case size_e:     source_browser->snd_sort = size_e;    break;
    case extension_e:   source_browser->snd_sort = extension_e;    break;
    default:    return;   /* no action on perms_e, doesn't work. */
    }
    source_browser->update(TRUE);
}

void SortCB(FL_OBJECT *, long option){
    if(!source_browser){
	program_unstable("no source browser.");
	return;
    }
    switch(option){
    case access_e:   source_browser->fst_sort = access_e;  break; 
    case filetype_e: source_browser->fst_sort = filetype_e;break; 
    case group_e:    source_browser->fst_sort = group_e;   break; 
    case inode_e:    source_browser->fst_sort = inode_e;   break; 
    case modific_e:  source_browser->fst_sort = modific_e; break; 
    case nlinks_e:   source_browser->fst_sort = nlinks_e;  break; 
    case owner_e:    source_browser->fst_sort = owner_e;   break; 
    case name_e:     source_browser->fst_sort = name_e;   break; 
    case size_e:     source_browser->fst_sort = size_e;    break;
    case extension_e:   source_browser->fst_sort = extension_e;    break;
    default:    return;   /* no action on perms_e, doesn't work. */
    }
    source_browser->update(TRUE);
}

