
#ifndef STACK_H
#define STACK_H

#include "system.h"
#include <iostream.h>

// for the time being, always no_template.
#define NO_TEMPLATE
#ifdef NO_TEMPLATE

typedef void * vptr_t;

class stack_c {
    vptr_t *v, *p;
    int sz;
public:
    stack_c(int s) { v = p = new vptr_t [sz=s]; }
    ~stack_c() { delete[] v; }

    int size() const { return p-v; }

    void push(vptr_t a){
	if(size() < sz){
	    *p++ = a;
	} else {
	    cerr << "stack overflow!\n";
	}
    }

    // return force_pushed element, if any, so that we can delete it.
    vptr_t force_push(vptr_t a);

    vptr_t pop(void){
	if(size() > 0){
	    return *--p;
	} else {
	    cerr << "stack underflow!\n";
	    return NULL;
	}
    }
    
    void clear(void){
	p = v;
    }

    vptr_t operator[](int i){
	if(i<0 || i>=size()){
	    cerr << "invalid stack access op [" << i << "]\n";
	    return NULL;
	} else {
	    return v[i];
	}
    }

    vptr_t get(void){
	if(size() > 0){
	    return *(p-1);
	} else {
	    cerr << "stack underflow!\n";
	    return NULL;
	}
    }
};

#else

template<class T>
class stack_c {
    T *v, *p;
    int sz;
public:
    stack_c(int s) { v = p = new T[sz=s]; }
    ~stack_c() { delete[] v; }

    int size() const { return p-v; }

    void push(T a){
	if(size() < sz){
	    *p++ = a;
	} else {
	    cerr << "stack overflow!\n";
	}
    }

    void force_push(T a){
	if(size() < sz){
	    *p++ = a;
	} else {
	    for(int i=0; i<sz-1; i++){
		v[i] = v[i+1];
	    }
	    *p = a;
	}
    }

    T pop(void){
	if(size() > 0){
	    return *--p;
	} else {
	    cerr << "stack underflow!\n";
	    return *v;
	}
    }
    
    void clear(void){
	p = v;
    }

    T operator[](int i){
	if(i<0 || i>=size()){
	    cerr << "invalid stack access op [" << i << "]\n";
	    return *v;
	} else {
	    return v[i];
	}
    }

    T get(void){
	if(size() > 0){
	    return *(p-1);
	} else {
	    cerr << "stack underflow!\n";
	    return *v;
	}
    }
};

#endif

#endif
