
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cb_fkey.h"
#include "cb_key.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_text.h"
#include "help_forms_aux.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "scroll_events.h"
#include "bitmaps/status.bm"
#include "status_forms.h"
#include "status_forms_aux.h"


int status_fontsize  = FL_NORMAL_SIZE,
    status_fontstyle = FL_TIMES_STYLE,
    status_ww = 300,
    status_wh = 90,
    status_wx = 10,
    status_wy = 10
;

int status_visible_flag     = TRUE,
    status_importance_level = 2
;

const char *status_ID_string = "status";

prm_entry_t status_prm_table[] = {
    {"ww",           int_e,   (void *) &status_ww}, 
    {"wh",           int_e,   (void *) &status_wh}, 
    {"wx",           int_e,   (void *) &status_wx}, 
    {"wy",           int_e,   (void *) &status_wy}, 
    {"fontsize",     int_e,   (void *) &status_fontsize}, 
    {"fontstyle",    int_e,   (void *) &status_fontstyle}, 
    {"visible_flag", int_e,   (void *) &status_visible_flag}, 
    {"",             int_e,   NULL}
};

static void var_sync(void){
    //M_info("Status", "var_sync");
    status_ww = StatusForm->w;
    status_wh = StatusForm->h;
    status_wx = StatusForm->x;
    status_wy = StatusForm->y;
}

void status_save_defaults(ofstream &ofile){
    if(StatusForm->visible){
        var_sync();
    }
    resource_save_defaults(ofile, status_prm_table, status_ID_string);
}    

int status_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    M_trace("status_load_default", "start");
    return resource_load_default(
        status_prm_table,  
        status_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

void status_sync_defaults(void){
    // after loading. 
    if(status_visible_flag){
	status_show_form();
    } else {
	status_close_form();
    }
}

void status_close_form(void){
    if(StatusForm->visible){
	var_sync();
	fl_hide_form(StatusForm);
	static int first = TRUE;
	if(first){
	    first = FALSE;
	    if(!OptionsForm->visible && tool_looser_flag){
		timed_notifier("Use global options (`Ctrl-F2') to switch the message panel back on.", notifier_secs);
	    }
	}
    }
    status_visible_flag = FALSE;
    options_sync_defaults();
}

void status_iconify(void){    
    if(StatusForm->visible)XIconifyWindow(fl_display, StatusForm->window, fl_screen);
}

void status_uniconify(void){    
    if(StatusForm->visible)  XMapRaised(fl_display, StatusForm->window);
}

void status_context_func(void){
    load_help_browser(status_help_text, "Status Messages");
}

void status_options_func(void){
    standard_browser_font_cb(StatusFormBrowser, status_fontsize, status_fontstyle, "Status Options");
}

static int status_raw_cb(FL_FORM *form, void *xev){
    if(standard_options_event_cb(form, (XEvent *) xev, status_options_func, TRUE)) return TRUE;
    return standard_form_event_cb(form, (XEvent *) xev, status_close_form, status_context_func);
}

void status_show_form(void){
    static int first = TRUE;
    M_info("status_show_form", "x %d, y %d, w %d, h%d ", status_wx, status_wy, status_ww, status_wh);
    if(first){
	first = FALSE;
        // times-roman, space is premium here. 
	fl_set_browser_fontstyle (StatusFormBrowser, status_fontstyle);
	fl_set_browser_fontsize  (StatusFormBrowser, status_fontsize);
	fl_set_browser_leftslider(StatusFormBrowser, TRUE);
	fl_register_call_back(StatusForm, FL_ALL_EVENT, status_raw_cb);
    }
    if(!StatusForm->visible){
        fl_set_initial_placement(StatusForm, status_wx, status_wy, status_ww, status_wh);
	fl_set_form_minsize(StatusForm, 40, 40);
	fl_set_form_maxsize(StatusForm, tool_max_ww, tool_max_wh);
        fl_show_form(StatusForm, FL_PLACE_FREE, TRUE, "Messages:");
	static Pixmap pixmap =0;
	if(pixmap == 0){
	    pixmap = XCreatePixmapFromBitmapData(
	        fl_display, StatusForm->window, status_bits, status_width,
	        status_height,
	        WhitePixel(fl_display, fl_screen),
	        BlackPixel(fl_display, fl_screen),
		1
	    );
	}
	set_window_icon(StatusForm->window, pixmap);
        XSetIconName(fl_display, StatusForm->window, "UCStatus");
    }
    
    status_visible_flag = TRUE;
    options_sync_defaults();
    var_sync();
    // no good, conflicts with options form:
    // tool_focus_move(StatusForm, StatusFormCloseReturnButton);
    M_info("status_show_form", "done");
}

// importance levels 0 = always, 1 = mostly, 2 = maybe ...
void status_msg(int importance, const char *format_string, ...){
    va_list arg_pointer;
    static int n_lines = 0;
    const int max_lines = 200;
    char buffer[FL_BROWSER_LINELENGTH + 1];
    
    if(!format_string) return;
    if(status_importance_level < importance) return;

    va_start(arg_pointer, format_string);
    vsprintf(buffer, format_string, arg_pointer);
    va_end(arg_pointer);
    
    fl_freeze_form(StatusForm);
    fl_addto_browser(StatusFormBrowser, buffer);
    n_lines++;
    if(n_lines >= max_lines){
	fl_delete_browser_line(StatusFormBrowser, 1);
    }
    fl_unfreeze_form(StatusForm);
}

void StatusFormCloseReturnButtonCB(FL_OBJECT *, long){
    status_close_form();
}
