#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>

#include "action_forms_aux.h"
#include "browser.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "global.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "linklist.h"
#include "pathnames.h"
#include "suffix.h"
#include "status_forms_aux.h"

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

int new_sfx_file_save(const char *fname){
    ofstream ofile;
    if(!fname) return FALSE;
    ofile.open(fname, ios::app);
    if(!ofile) return FALSE;
    ofile << "# Unix Cockpit file name suffix makefile\n";
    ofile << "#\n";
    ofile << "# This special makefile is used to bind user actions to file \n";
    ofile << "# suffixes such as `.rgb' or `.tiff'. For each new suffix, you can\n";
    ofile << "# write new rules, change existing ones, or leave it unused.\n";
    ofile << "#\n";
    ofile << "# Note that makefiles need TABS, not spaces, in the rule lines\n";
    ofile << "# following the targets. Watch out when X-copy-pasting into this\n";
    ofile << "# file - sometimes the tabs are expanded to spaces :-).\n";
    ofile << "#\n";
    ofile << "# DON'T use this makefile for anything else but new suffix rules.\n";
    ofile << "# The rules must be independent and not rely on other rules. Each\n";
    ofile << "# rule can contain only one suffix. The commands following a rule \n";
	ofile << "# must be ended by an empty line. If you want to recreate this file,\n";
    ofile << "# remove it and run UC again.\n";
    ofile << "#\n";
    ofile << "# You can define multiple rules for the same suffix. For example,\n";
    ofile << "# for a .tar file, you could write a rule for listing and one for\n";
    ofile << "# extracting the archive. If you write multiple rules, make sure\n";
    ofile << "# to include a #-comment in the same line as the .sfx:-specification.\n";
    ofile << "#\n";
    ofile << "# This would look like `.tar:     #List Archive'\n";
    ofile << "#                            ^TAB ^here is the comment.\n";
    ofile << "#\n";
    ofile << "# followed by the commands to run. Note the TAB between `:' and `#'.\n";
    ofile << "#\n";
    ofile << "# Before executing the rule, UC will prompt you to choose between\n";
    ofile << "# the alternatives. You can also define shortcuts for the different\n";
    ofile << "# options (also look at the rules for .tar two pages below):\n";
    ofile << "#\n";
    ofile << "# This looks like `.tar:     #List Archive#Ll'\n";
    ofile << "#                                         ^here are the shortcuts (L,l).\n";
    ofile << "#\n";
    ofile << "# Shortcuts only make sense if you have many rules for a suffix.\n";
    ofile << "#\n";
    ofile << "# If you specify a #-comment for suffix for which there is only one \n";
    ofile << "# rule, you will be prompted for confirmation before execution of\n";
    ofile << "# that rule. If there is no #-comment, it will be executed directly.\n";
    ofile << "# In this way, you can control wether you want confirmation or not.\n";
    ofile << "#\n";
    ofile << "#\n";
    ofile << "#\n";
    ofile << "#\n";
    ofile << "# Don't remove this line, and don't redefine it either:\n";
    ofile << ".SUFFIXES:   ${UC_SFX_LIST}\n";
    ofile << "#\n";
    ofile << "# The following fake target results in silent execution. Comment\n";
    ofile << "# out for debugging, or if you want to see the command executed.\n";
    ofile << ".SILENT:\n";
    ofile << "\n";
    ofile << "# Here's a handy macro that will be used below, it's used to make your terminal\n";
    ofile << "# emulator wait for you to press enter:\n";
    ofile << "ENTER=echo \"press Enter to continue\"; read x \n";
    ofile << "\n";
    ofile << "# Another macro to xterm commands (note the -title at the end,\n",
    ofile << "# which needs an argument!):\n";
    ofile << "XTERM=xterm -fg white -bg black -sb -geometry 80x40+300+200 -title \n";
    ofile << "\n";
    ofile << "# A few example rules (change or add yours):\n";
    ofile << "# (Within a rule, use the standard make macro `$<' for the file that\n";
    ofile << "# will be worked upon, see below).\n";
    ofile << "\n";

    ofile << "# arj archives:\n";
    ofile << ".arj:\t#Extract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting: $<\" -e sh -c 'unarj x $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".arj:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"Listing: $<\" -e sh -c 'unarj l $<; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# compressed files:\n";
    ofile << ".Z:\t#Uncompress\n";
    ofile << "\tuncompress $< &\n";
    ofile << "\n";
    ofile << ".Z:\t#Uncompress to More#Mm\n";
    ofile << "\t${XTERM} \"Uncompressing to More: $<\" -e sh -c 'uncompress -c $< | more; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# gnuzipped files:\n";
    ofile << ".gz:\t#Gunzip#Gg\n";
    ofile << "\tgunzip $< &\n";
    ofile << "\n";
    ofile << ".gz:\t#Gunzip to More#Mm\n";
    ofile << "\t${XTERM} \"Gunzipping to More: $<\" -e sh -c 'gunzip -c $< | more; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# gnuzipped tar archives:\n";
    ofile << ".tgz:\t#Gunzip#gG\n";
    ofile << "\tgunzip $< &\n";
    ofile << "\n";
    ofile << ".tgz:\t#Gunzip to tar#Tt\n";
    ofile << "\t${XTERM} \"Gunzipping to Tar: $<\" -e sh -c 'gunzip -dc $< | tar xvf - ; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".tgz:\t#Gunzip and tar List#Ll\n";
    ofile << "\t${XTERM} \"Gunzip and tar listing: $<\" -e sh -c 'gunzip -c $< | tar tvf - ; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# lha archives:\n";
    ofile << ".lha:\t#Extract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting: $<\" -e sh -c 'lha x $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".lha:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"Listing: $<\" -e sh -c 'lha l $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".lzh:\t#Extract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting: $<\" -e sh -c 'lha x $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".lzh:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"Listing: $<\" -e sh -c 'lha l $<; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# rar archives:\n";
    ofile << ".rar:\t#Extract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting: $<\" -e sh -c 'unrar x $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".rar:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"Listing: $<\" -e sh -c 'unrar l $<; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# zip archives:\n";
    ofile << ".zip:\t#Extract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting: $<\" -e sh -c 'unzip $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".zip:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"Listing: $<\" -e sh -c 'unzip -l $<; ${ENTER}' & \n";
    ofile << "\n";


    ofile << "# aiff/c sound:\n";
    ofile << ".aiff:\t#Soundeditor#Ss\n";
    ofile << "\tsoundeditor $< &\n";
    ofile << "\n";
    ofile << ".aiff:\t#Playaiff#Pp\n";
    ofile << "\tplayaiff $< &\n";
    ofile << "\n";
    ofile << ".aifc:\t#Soundeditor#Ss\n";
    ofile << "\tsoundeditor $< &\n";
    ofile << "\n";

    ofile << "# X bitmap files:\n";
    ofile << ".bm:\t#Bitmap\n";
    ofile << "\tbitmap $< &\n";
    ofile << "\n";
    ofile << ".bmp:\t#Bitmap\n";
    ofile << "\tbitmap $< &\n";
    ofile << "\n";
    ofile << ".xbm:\t#Bitmap\n";
    ofile << "\tbitmap $< &\n";
    ofile << "\n";
    ofile << "# tex previewables:\n";
    ofile << ".dvi:\t#xdvi\n";
    ofile << "\txdvi $< &\n";
    ofile << "\n";
    ofile << "# gif images:\n";
    ofile << ".gif:\t#xv\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << "# jpeg images:\n";
    ofile << ".jpeg:\t#xv\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << ".jpg:\t#xv\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << "# pgm images:\n";
    ofile << ".pgm:\t#xv\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << "# X pixmaps:\n";
    ofile << ".xpm:\t#xv\n";
    ofile << "\txv $< &\n";
    ofile << "\n";

    ofile << "# WWW docs:\n";
    ofile << ".html:\t#Mosaic#Mm\n";
    ofile << "\tmosaic $< &\n";
    ofile << "\n";
    ofile << ".html:\t#Netscape#Nn\n";
    ofile << "\tnetscape $< &\n";
    ofile << "\n";
    ofile << ".htm:\t#Mosaic#Mm\n";
    ofile << "\tmosaic $< &\n";
    ofile << "\n";
    ofile << ".htm:\t#Netscape#Nn\n";
    ofile << "\tnetscape $< &\n";
    ofile << "\n";
    ofile << ".shtml:\t#Mosaic#Mm\n";
    ofile << "\tmosaic $< &\n";
    ofile << "\n";
    ofile << ".shtml:\t#Netscape#Nn\n";
    ofile << "\tnetscape $< &\n";
    ofile << "\n";

    ofile << "# post/ghostscript:\n";
    ofile << ".ps:\t#Ghostview#Gg\n";
    ofile << "\tghostview $< &\n";
    ofile << "\n";

    ofile << "# portable document format:\n";
    ofile << ".pdf:\t#Acrobat Reader#Aa\n";
    ofile << "\tacroread $< &\n";
    ofile << "\n";

    ofile << "# SGI images:\n";
    ofile << ".bw:\t#Xv#Xx\n";
    ofile << "\tipaste $< &\n";
    ofile << "\n";
    ofile << ".bw:\t#Ipaste#Ii\n";
    ofile << "\tipaste $< &\n";
    ofile << "\n";
    ofile << ".rgb:\t#Ipaste#Ii\n";
    ofile << "\tipaste $< &\n";
    ofile << "\n";
    ofile << ".rgb:\t#Xv#Xx\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << "# SGI showcase:\n";
    ofile << ".shw:\t#Showcase\n";
    ofile << "\tshowcase $< &\n";
    ofile << "\n";
    ofile << ".sav:\t#Showcase\n";
    ofile << "\tshowcase $< &\n";
    ofile << "\n";

    ofile << "# RedHat Package Modules:\n";
    ofile << ".rpm:\t#Get info#Gg\n";
    ofile << "\t${XTERM} \"Package info: $<\" -e sh -c \\\n";
    ofile << "\t'rpm -q -i -p $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".rpm:\t#List package#Ll\n";
    ofile << "\t${XTERM} \"Package contents: $<\" -e sh -c \\\n";
    ofile << "\t'rpm -q -l -p $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".rpm:\t#Install package#Ii\n";
    ofile << "\t${XTERM} \"Installing package: $<\" -e sh -c \\\n";
    ofile << "\t'rpm -i $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".rpm:\t#Force install#Ff\n";
    ofile << "\t${XTERM} \"Force installing package: $<\" -e sh -c \\\n";
    ofile << "\t'rpm -i --force $<; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# tar archives:\n";
    ofile << ".tar:\t#List archive#Ll\n";
    ofile << "\t${XTERM} \"List of $<\" -e sh -c \\\n";
    ofile << "\t'tar tf $<; ${ENTER}' & \n";
    ofile << "\n";
    ofile << ".tar:\t#eXtract archive#Xx\n";
    ofile << "\t${XTERM} \"Extracting $<\" -e sh -c \\\n";
    ofile << "\t'tar xvf $<; ${ENTER}' & \n";
    ofile << "\n";

    ofile << "# TIFF Images:\n";
    ofile << ".tiff:\t#Xv#Xx\n";
    ofile << "\txv $< &\n";
    ofile << "\n";
    ofile << ".tiff:\t#Imgview#Ii\n";
    ofile << "\timgview $< &\n";
    ofile << "\n";

    ofile << "# UU-encoded files:\n";
    ofile << ".uu:\t#uudecode\n";
    ofile << "\tuudecode $< &\n";
    ofile << "\n";
    ofile << "\n";

    ofile << "# Don't remove this last target, it cleans up temporary files of the\n";
    ofile << "# UC suffix execution procedure.\n";
    ofile << "uc_cleanup_tgt:\n";
    ofile << "\t/bin/rm ${UC_CLEANUP_FNAME}\n";
    ofile << "\n";
    ofile.close();
    return TRUE;
}

static char uc_sfx_fname[MAXPATHLEN+1];
static linklist_c sfx_list;

class sfx_c {
public:
    char *m_pComment, *m_pShortcuts, *m_pSuffix;
    sfx_c(const char *pSuffix, const char *pComment, const char *pShortcuts);
    ~sfx_c();
    int has_valid_comment(void);
private:
    inline void zero_ptrs(void) {
	m_pComment = m_pShortcuts = m_pSuffix = NULL;
    }
};

sfx_c::sfx_c(const char *pSuffix, const char *pComment, const char *pShortcuts){
    zero_ptrs();
    if (!pSuffix) {
	M_err("sfx_c", "NULL ptr");
	return;
    }

    m_pSuffix = strdup(pSuffix);
    if (pComment)	m_pComment	= strdup(pComment);
    if (pShortcuts)	m_pShortcuts	= strdup(pShortcuts);
}

sfx_c::~sfx_c(){
    if (m_pComment)	fl_free (m_pComment);
    if (m_pShortcuts)	fl_free (m_pShortcuts);
    if (m_pSuffix)	fl_free (m_pSuffix);
    zero_ptrs();
}

int sfx_c::has_valid_comment(void){
    return (m_pComment && strlen(m_pComment));
}

static void empty_list(void){
    while(sfx_list.getsize()){
	sfx_c *pSfx = (sfx_c *)sfx_list.getval();
	if(pSfx) delete pSfx;
	sfx_list.remove();
    }
}

static int is_special_rule(const char *sfx_tok){
    if(!sfx_tok) return FALSE;
    if(strcmp(sfx_tok, ".DEFAULT") == 0) return TRUE;
    if(strcmp(sfx_tok, ".IGNORE") == 0) return TRUE;
    if(strcmp(sfx_tok, ".PRECIOUS") == 0) return TRUE;
    if(strcmp(sfx_tok, ".SILENT") == 0) return TRUE;
    if(strcmp(sfx_tok, ".SUFFIXES") == 0) return TRUE;
    return FALSE;
}


static int sfx_file_loaded_flag = FALSE;

int suffix_file_load(void){
    M_info("suffix_file_load", "start");
    empty_list();

    if(tool_ldefault_dir){
	sprintf(uc_sfx_fname, "%s/ucsfx", tool_ldefault_dir);
    } else {
	cerr << "UC: no tool default dir.\n";
	return FALSE;
    }
    
    // input file stream operations
    ifstream ifile;
    ifile.open(uc_sfx_fname);
    if(!ifile){
	// no suffix file in home/.uc. Attempt global.
	sprintf(uc_sfx_fname, "%s/ucsfx", tool_gdefault_dir);
	ifile.open(uc_sfx_fname);
	if(!ifile){
	    // no global. Make local.
	    sprintf(uc_sfx_fname, "%s/ucsfx", tool_ldefault_dir);
	    if(new_sfx_file_save(uc_sfx_fname)){
		ifile.open(uc_sfx_fname);
		if(!ifile) return FALSE;	    
	    } else {
		return FALSE;
	    }
	}
    }

    const int bufsize = 2*MAXPATHLEN;
    char mbuf[bufsize];

    int n_sfxs = 0;
    while(ifile.getline(mbuf, bufsize) ){
	if(mbuf[0] != '.' || mbuf[0] == ' ' || mbuf[0] == '\0' || mbuf[0] == '#'){
	    continue;
	    // it is not a special rule, or empty line, or starts with a blank, or
	    // is a comment.
	}
	char *sfx_tok = strtok(mbuf, ":");
	if(!sfx_tok || strchr(sfx_tok, '#')){
	    // guard against pathologic cases (no token or #-malformed makefile)
	    continue;
	}
	if(dbg_flag) cerr << "got sfx_tok: `" << sfx_tok << "'\n";
	if(is_special_rule(sfx_tok)){
	    if(dbg_flag) cerr << "sfx_tok is special: `" << sfx_tok << "'\n";
	} else {
	    // get the rest of the line and parse it.
	    char
		*pLine      = strtok(NULL, ""), 
		*pComment   = NULL, 
		*pShortcuts = NULL
	    ;

	    if (pLine && strlen(pLine)) {
		pComment = strchr(pLine, '#');
		if (pComment) {
		    if (strlen(pComment) > 1) {
			pComment++;	// step over '#'.
			if (dbg_flag) cerr << "Comment: " << pComment << endl;
			pShortcuts = strchr(pComment, '#');
			if (pShortcuts) {
			    // separate comment and shortcuts.
			    *pShortcuts = '\0';
			    // step over.
			    pShortcuts++;
			    if (dbg_flag) cerr << "Shortcuts: " << pShortcuts << endl;
			}
		    } else {
			// Null it to to make it really clear there's no comment.
			pComment = NULL;
		    }
		}
	    }
	    sfx_c *pSfx = new sfx_c(sfx_tok, pComment, pShortcuts);
	    sfx_list.add(pSfx);
	    n_sfxs++;
	}
    }
    if(dbg_flag) cerr << "read " << n_sfxs << " suffixes.\n";
    ifile.close();
    sfx_file_loaded_flag = TRUE;    
    if(n_sfxs == 0){
	timed_error_notifier("Suffix file contains no suffixes ?");
	return FALSE;
    }
    return TRUE;
}

/* return the number of rules for fname */
int has_sfx_rule(const char *fname){
    if(!fname){
	cerr << "NULL fname in has_sfx_rule.\n";
	return 0;
    }
    if(!sfx_file_loaded_flag){
	if(!suffix_file_load()){
	    return 0;
	}
    }
    const char *sfx_p = strrchr(fname, '.');
    if(!sfx_p){
	if(dbg_flag) cerr << "file name has no suffix.\n";
	return 0;
    }
    int size = sfx_list.getsize();
    int count = 0;
    for(int i=0; i<size; i++){
	sfx_c *pSfx = (sfx_c *)sfx_list[i];
	if(!pSfx){
	    cerr << "NULL buf in suffix list!\n";
	    return 0;
	}
	if(strcmp(sfx_p, pSfx->m_pSuffix) == 0){
	    if(dbg_flag) cerr << "File " << fname << " has a suffix rule match.\n";
	    count++;
	}
    }
    if(dbg_flag) cerr << "File " << fname << " has NO suffix rule match.\n";
    return count;
}

static int mung_makefile(const char *sfx_p, int which, char sfx_fname_buf[]){
    if (!sfx_p) return FALSE;
    char *tmp_nam = tempnam(NULL, "ucmk");
    if (!tmp_nam) return FALSE;
    strcpy(sfx_fname_buf, tmp_nam);

    ofstream ofile(tmp_nam);
    if (!ofile) return FALSE;

    ifstream ifile;
    ifile.open(uc_sfx_fname);
    if(!ifile) return FALSE;

    const int bufsize = 2*MAXPATHLEN;
    char real_buf[bufsize], work_buf[bufsize];

    int count = 0;
    while(ifile.getline(real_buf, bufsize) ){
	strcpy(work_buf, real_buf);
	char *sfx_tok = strtok(work_buf, ":");
	if (real_buf[0] != '.' || real_buf[0] == ' ' || 
	    real_buf[0] == '\0' || real_buf[0] == '#'
	){
	    // ok. 
	} else if (!sfx_tok || strchr(sfx_tok, '#')){
	    // guard against pathologic cases 
	} else if (is_special_rule(sfx_tok)){
	    // ok.
	} else {
	    // kill all suffix rules except the one matching in
	    // suffix and count.
	    int equal_flag = (strcmp(sfx_tok, sfx_p) == 0);
	    // count is pre-inc'ed.
	    if (equal_flag) count++;

	    if (!equal_flag || (count != which+1)) { 
		// remove all following lines that are not empty.
		while (ifile.getline(real_buf, bufsize) && strlen(real_buf));
		// add a newline to the output file.
		ofile << endl;
		continue;
	    }   // else just output it.
	}
	ofile << real_buf << endl;
    }
    ifile.close();
    ofile.close();
    return TRUE;
}

static sfx_c *get_sfx_info(const char *p_sfx, int which) {
    if (!p_sfx) return NULL;
    if (which < 0) return NULL;

    int size = sfx_list.getsize();
    int count = 0;
    for(int i=0; i<size; i++){
	sfx_c *pSfx = (sfx_c *)sfx_list[i];
	if(!pSfx){
	    cerr << "NULL buf in suffix list!\n";
	    return NULL;
	}
	if(strcmp(p_sfx, pSfx->m_pSuffix) == 0){
	    if (count == which) return pSfx;
	    count++;
	}
    }
    return NULL;
}

// get comments from makefile, ask_question, mung makefile, return;
static int ask_multi_and_mung(
    const char *fname, 
    const char *sfx_p, 
    char sfx_fname_buf[]
){
    // no default makefile, use munged only, and autoremove them.
    strcpy(sfx_fname_buf, "");
    
    if(!sfx_p) return FALSE;
    int count = has_sfx_rule(fname);
    if (dbg_flag) cerr << count << " rules for " << fname << endl;
    if(count == 0){
	cerr << "No rule for " << fname << " in execute_sfx_rule.\n";
	return FALSE;
    } else if (count > 1) {
	// ask question, determine which, mung temporary makefile.
	if (count > 4) {
	    timed_error_notifier("Warning: Can handle only up to 4 rules for the same suffix.");
	    count = 4;
	}
	const char *labels[5], *shortcuts[5];
	sfx_c *pSfx;
	// 5 so that we can add a cancel button.
	for (int i=0; i<5; i++) {
	    if (i<count) {
		pSfx = get_sfx_info(sfx_p, i);
		if (!pSfx) return FALSE;
		labels   [i] = pSfx->m_pComment;
		shortcuts[i] = pSfx->m_pShortcuts;
	    } else {
		if (i == count) {
		    labels   [i] = "Cancel";
		    shortcuts[i] = "Cc";
		} else {
		    labels   [i] = NULL;
		    shortcuts[i] = NULL;
		}
	    }
	}
	char tmp_buf[MAXPATHLEN];
	sprintf(tmp_buf, "Choose an action on file `%s':", fname);
	int retval = ask_multi_question(FALSE, 0, tmp_buf, 
	    labels[0], shortcuts[0], NoSymbol,
	    labels[1], shortcuts[1], NoSymbol,
	    labels[2], shortcuts[2], NoSymbol,
	    labels[3], shortcuts[3], NoSymbol,
	    labels[4], shortcuts[4], NoSymbol
	);
	if (dbg_flag) cerr << retval << endl;
	if (retval < 0 || retval >= count) {
	    return FALSE;
	} else {
	    return mung_makefile(sfx_p, retval, sfx_fname_buf);
	}
    } else {
	sfx_c *pSfx = get_sfx_info(sfx_p, 0);
	if (!pSfx) return FALSE;
	if (pSfx->has_valid_comment()) {
	    if (!ask_question(FALSE, "Do make `%s' on file `%s' ?", pSfx->m_pComment, fname)) {
		return FALSE;
	    }
	}
	// if no comment, simply execute.
    }
    return mung_makefile(sfx_p, 0, sfx_fname_buf);
}

int execute_sfx_rule(const char *fname){
    if(!fname || !make_path){
	cerr << "NULL fname/make_path in execute_sfx_rule.\n";
	return FALSE;
    }
    if(strchr(fname, '#')){
	timed_notifier("Sorry, suffix module can't process filenames with the `#'-character.");
	return FALSE;
    }
    char buffer[MAXPATHLEN*2], sfx_fname_buf[MAXPATHLEN];
    const char *sfx_p = strrchr(fname, '.');
    if(!sfx_p){
	if(dbg_flag) cerr << "file name has no suffix.\n";
	return FALSE;
    }
    char fname_buf[MAXPATHLEN+1];
    int len = sfx_p-fname;
    if(len <= 0){
	if(dbg_flag) cerr << "file name has no prefix.\n";
	return FALSE;
    }
    strncpy(fname_buf, fname, len);
    fname_buf[len] = '\0';

    if (!ask_multi_and_mung(fname, sfx_p, sfx_fname_buf)) return FALSE;

    char sfx_macro_buf[NAME_MAX+1];
    sprintf(sfx_macro_buf, "UC_SFX_LIST=%s", sfx_p);

    char sfx_cleanup_buf[MAXPATHLEN];
    sprintf(sfx_cleanup_buf, "UC_CLEANUP_FNAME=%s", sfx_fname_buf);
    /* make flags:
     * -u for unconditional.  (doesn't work on SunOS)
     * -f for file.
     * -r for killing of built-in rules.
     */
/*
 * -u is a bit dangerous, for example if people use gmake on SGI.
 *  -u for unconditional. (only supported on SGI, not sun, linux. dec different meaning)
 *  sprintf(buffer, "make -u -r -f %s '%s' %s &", sfx_fname_buf, sfx_macro_buf, fname_buf);
 */
#ifdef SGI
    sprintf(buffer, "%s -u -r -f %s '%s' '%s' %s uc_cleanup_tgt &", 
#else			       
    sprintf(buffer, "%s -r -f %s '%s' '%s' %s uc_cleanup_tgt &",
#endif			       
		make_path, 
		sfx_fname_buf, 
		sfx_macro_buf, 
		sfx_cleanup_buf,
		fname_buf
	);
    fl_set_global_cursor(XC_watch);
    if (dbg_flag) cerr << buffer << endl;
    safe_system(buffer);
    fl_reset_global_cursor();
    return TRUE;       
}
