
#include <stdlib.h>
#include <string.h>

#include "browser.h"
#include "browser_menu.h"
#include "cb_key.h"
#include "cb_fkey.h"
#include "client_msg.h"
#include "error_notifier_forms_aux.h"
#include "fkeybar_forms.h"
#include "fkeybar_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "status_forms.h"
#include "status_forms_aux.h"
#include "bitmaps/toolbox.bm"
#include "toolbox_forms.h"
#include "toolbox_forms_aux.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "wm_forms_aux.h"
#include "xshell.h"


int
    toolbox_tree_flag	= FALSE,
    toolbox_split_flag	= FALSE,
    toolbox_ww = 124,
    toolbox_wh = 560,
    toolbox_wx = 10,
    toolbox_wy = 10
    /*
    toolbox_ww = default_toolbox_width, 
    toolbox_wh = 2*default_browser_height + 2*xshell_real_bwidth + xshell_real_topbar_height, 
    toolbox_wx = default_browser_width + 3*xshell_real_bwidth,
    toolbox_wy = 0
    */
;

double toolbox_ctrl_timeout = 2.0;

int         toolbox_visible_flag = TRUE;
const char *toolbox_ID_string = "toolbox";

prm_entry_t toolbox_prm_table[] = {
    {"ww",           int_e,    (void *) &toolbox_ww}, 
    {"wh",           int_e,    (void *) &toolbox_wh}, 
    {"wx",           int_e,    (void *) &toolbox_wx}, 
    {"wy",           int_e,    (void *) &toolbox_wy}, 
    {"split_flag",   int_e,    (void *) &toolbox_split_flag}, 
    {"visible_flag", int_e,    (void *) &toolbox_visible_flag}, 
    {"ctrl_timeout", double_e, (void *) &toolbox_ctrl_timeout}, 
    {"",             int_e,    NULL}
};

static void toolbox_metach_sync(void);

static void var_sync(void){
    toolbox_ww = ToolBoxForm->w;
    toolbox_wh = ToolBoxForm->h;
    toolbox_wx = ToolBoxForm->x;
    toolbox_wy = ToolBoxForm->y;
}

void toolbox_save_defaults(ofstream &ofile){
    if(ToolBoxForm->visible){
	var_sync();
    }
    resource_save_defaults(ofile, toolbox_prm_table, toolbox_ID_string);
}    

int toolbox_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    return resource_load_default(
        toolbox_prm_table,  
        toolbox_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

static void toolbox_set_fontsize(int size){
    fl_freeze_form(ToolBoxForm);
/*
  fl_set_object_lsize(ToolBoxFormF1Text,size);
  fl_set_object_lsize(ToolBoxFormF2Text,size);
  fl_set_object_lsize(ToolBoxFormF3Text,size);
  fl_set_object_lsize(ToolBoxFormF4Text,size);
  fl_set_object_lsize(ToolBoxFormF5Text,size);
  fl_set_object_lsize(ToolBoxFormF6Text,size);
  fl_set_object_lsize(ToolBoxFormF7Text,size);
  fl_set_object_lsize(ToolBoxFormF8Text,size);
  fl_set_object_lsize(ToolBoxFormF9Text,size);
  fl_set_object_lsize(ToolBoxFormF10Text,size);
  fl_set_object_lsize(ToolBoxFormF11Text,size);
  fl_set_object_lsize(ToolBoxFormF12Text,size);

  fl_set_object_lsize(ToolBoxFormF1Button,size);
  fl_set_object_lsize(ToolBoxFormF2Button,size);
  fl_set_object_lsize(ToolBoxFormF3Button,size);
  fl_set_object_lsize(ToolBoxFormF4Button,size);
  fl_set_object_lsize(ToolBoxFormF5Button,size);
  fl_set_object_lsize(ToolBoxFormF6Button,size);
  fl_set_object_lsize(ToolBoxFormF7Button,size);
  fl_set_object_lsize(ToolBoxFormF8Button,size);
  fl_set_object_lsize(ToolBoxFormF9Button,size);
  fl_set_object_lsize(ToolBoxFormF10Button,size);
  fl_set_object_lsize(ToolBoxFormF11Button,size);
  fl_set_object_lsize(ToolBoxFormF12Button,size);

  fl_set_object_lsize(ToolBoxFormMetaButton,size);
  */
    fl_set_object_lsize(ToolBoxFormGroup,size);
    fl_unfreeze_form(ToolBoxForm);
}

void toolbox_sync_defaults(void){
    // this function is called after loading. 
    fkeybar_set_fontsize(tool_fontsize);
    toolbox_set_fontsize(tool_fontsize);

    // remember _wh...
    FL_Coord tmp_wh = toolbox_wh;

    if (toolbox_split_flag) {
	// flag gets flipped.
	toolbox_split_flag = !toolbox_split_flag;
	toolbox_split();
	// do vertical scaling, the splitting did reset wh.
	fl_scale_form(ToolBoxForm, 1.0, (double) tmp_wh / (double) toolbox_wh);
	toolbox_wh = tmp_wh;
    } else {
	if(fkeybar_visible_flag) { 
	    fkeybar_close_form();
            fkeybar_visible_flag = TRUE;    // Mung it here, too.
	}
    }

    if(toolbox_visible_flag){
	toolbox_show_form();
    } else {
	toolbox_close_form();
    }
}

void toolbox_close_form(void){
    static int first = TRUE;
    if(ToolBoxForm->visible){
	if(first){
	    first = FALSE;
	    if(!OptionsForm->visible && tool_looser_flag){
		timed_notifier("use Ctrl-F2 Options to get the toolbox back.", notifier_secs);
	    }
	}
	var_sync();
	fl_hide_form(ToolBoxForm);
    }
    toolbox_visible_flag = FALSE;
    first = FALSE;
    options_sync_defaults();
}

void toolbox_iconify(void){    
    if(ToolBoxForm->visible) XIconifyWindow(fl_display, ToolBoxForm->window, fl_screen);
}

void toolbox_uniconify(void){    
    if(ToolBoxForm->visible) XMapRaised(fl_display, ToolBoxForm->window);
}

int toolbox_raw_cb(FL_FORM *form, void *xev){
    // a few are special: F1, F2.
    XEvent *xevent = (XEvent *) xev;
    KeySym keysym;
    char keybuf[1];

    switch(xevent->type){
    case KeyPress:
	XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	switch(keysym){
	    /*
	      case XK_Control_L:
	      case XK_Control_R:
	      tool_ctrl_on_flag = TRUE;
	      toolbox_metach_sync();
	      break;
	      */
	default:
	    break;
	}
	break;
    case KeyRelease:
	//cerr << "release\n";
	/*
	  XLookupString((XKeyEvent *) xevent, keybuf, sizeof(keybuf), &keysym, 0);
	  switch(keysym){
	  case XK_Control_L:
	  case XK_Control_R:
	  tool_ctrl_on_flag = FALSE;
	  toolbox_metach_sync();
	  break;
	  default:
	  tool_ctrl_on_flag = controlkey_down(fl_keymask);
	  toolbox_metach_sync();
	  break;
	  }
	  // special trick so that ctrl-key up/down works.
	  return TRUE;
	  */
    default:
	break;
    }
    if(standard_form_event_cb(form, xevent, toolbox_close_form, toolbox_help_func)){
	return TRUE;
    } else {
	if (toolbox_tree_flag) {
	    return tree_handle_free(TreeFormFreeObj, FL_OTHER, 0, 0, 0, xevent);
	} else {
	    return browser_handle_free(source_browser->free_obj, FL_OTHER, 0, 0, 0, xevent);
	}
    }
}

void toolbox_help_func(void){
    load_help_browser(toolbox_help_text, "Tool Box");
}

void ToolBoxFormF1ButtonCB(FL_OBJECT *, long){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	HelpButtonCB(NULL, 0);
    } else {
	toolbox_help_func();
    }
}

void ToolBoxFormF2ButtonCB(FL_OBJECT *, long){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	OptionsButtonCB(NULL, 0);
    } else {
	if (toolbox_tree_flag) {
	    TreeFormOptionsButtonCB(NULL, 0);
	} else {
	    browser_menu_event_cb(NULL, 0);
	}
    }
}

static void 
    toolbox_set_wintitle(void){
	if (!ToolBoxForm->visible) return;
	const char *new_title = toolbox_tree_flag ? "->Tree" : "ToolBox";
	static char *title = NULL;
	if (title) {
	    if (strcmp(title, new_title) == 0) return;
	    fl_free(title);
	}
	title = strdup(new_title);
	assert(title);
	XStoreName(fl_display, ToolBoxForm->window, title);
	XSetIconName(fl_display, ToolBoxForm->window, "UCToolBox");
    }

static void set_form_constraints(void){
    if(toolbox_split_flag){
        fl_set_form_minsize(ToolBoxForm, 60, 100);
        fl_set_form_maxsize(ToolBoxForm, 300, 500);
    } else {
        fl_set_form_minsize(ToolBoxForm, 60, 300);
        fl_set_form_maxsize(ToolBoxForm, 300, 800);
    }
}

void toolbox_show_form(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(ToolBoxForm, FL_ALL_EVENT, toolbox_raw_cb);
	set_form_constraints();
#ifdef DEC_ALPHA	
	fl_set_object_label(ToolBoxFormF11Button,"F13");
#endif
    }
    M_info("toolbox::show_form", "x %d,y %d,w %d,h %d", toolbox_wx, toolbox_wy, toolbox_ww, toolbox_wh);
    fl_set_initial_placement(ToolBoxForm, toolbox_wx, toolbox_wy, toolbox_ww, toolbox_wh);
    fl_show_form(ToolBoxForm, FL_PLACE_FREE, TRUE, "ToolBox");
    static Pixmap pixmap =0;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, ToolBoxForm->window, toolbox_bits, toolbox_width,
	    toolbox_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    set_window_icon(ToolBoxForm->window, pixmap);
    toolbox_set_wintitle();
    
    toolbox_visible_flag = TRUE;
    options_sync_defaults();
    var_sync();
}

void ToolBoxFormPad0ButtonCB(FL_OBJECT *, long){
    InsertKeyCB(NULL, 1);
}
void ToolBoxFormPad1ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_end();
    } else {
    	EndKeyCB(NULL, 0);
    }
}
void ToolBoxFormPad2ButtonCB(FL_OBJECT *, long){
    // cerr <<"toolbox downarrow.\n";
    if (toolbox_tree_flag) {
	tree_down_arrow(TRUE);
    } else {
    	DownArrowKeyCB(NULL, TRUE);
    }
}
void ToolBoxFormPad3ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_page_down();
    } else {
    	PageDownKeyCB(NULL, 0);
    }
}
void ToolBoxFormPad4ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_left_arrow();
    } else {
    	LeftArrowKeyCB(NULL, 0);
    }
}
void ToolBoxFormPad5ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	TreeCenterCB(NULL, 0);
    } else {
    	Pad5KeyCB(NULL, 0);
    }
}
void ToolBoxFormPad6ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_right_arrow();
    } else {
    	RightArrowKeyCB(NULL, 0);
    }
}
void ToolBoxFormPad7ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	TreeHomeCB(NULL, 0);
    } else {
    	HomeKeyCB(NULL, 0);
    }
}
void ToolBoxFormPad8ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_up_arrow(TRUE);
    } else {
    	UpArrowKeyCB(NULL, TRUE);
    }
}
void ToolBoxFormPad9ButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
	tree_page_up();
    } else {
    	PageUpKeyCB(NULL, 0);
    }
}
void ToolBoxFormHiddenReturnButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
    	TreeEnterKeyCB(NULL, 0);
    } else {
    	NewEnterKeyCB(NULL, 0);
    }
}
void ToolBoxFormPadVirguleButtonCB(FL_OBJECT *, long){
    PadVirguleKeyCB(NULL, 0);
}    
void ToolBoxFormPadPlusButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
    	TreeExpandCB(NULL, 0);
    } else {
    	SelectButtonCB(NULL, 0);
    }
}
void ToolBoxFormPadMinusButtonCB(FL_OBJECT *, long){
    if (toolbox_tree_flag) {
    	TreeCollapseCB(NULL, 0);
    } else {
    	DeselectButtonCB(NULL, 0);
    }
}
void ToolBoxFormPadAsteriskButtonCB(FL_OBJECT *, long){
    PadAsteriskKeyCB(NULL, 0);
}

void ToolBoxFormMetaTimerCB(FL_OBJECT *, long){
    if ((ToolBoxFormMetaButton->belowmouse && ToolBoxFormMetaButton->pushed) ||
    	(FKeyBarFormMetaButton->belowmouse && FKeyBarFormMetaButton->pushed)
    ){
	// continue ... but, build in fast release. 
	fl_set_timer(ToolBoxFormMetaTimer, 0.1);
	fl_set_timer(FKeyBarFormMetaTimer, 0.1);
    } else {
	tool_ctrl_on_flag = FALSE;    
	toolbox_metach_sync();
    }
}

void ToolBoxFormMetaButtonCB(FL_OBJECT *, long){
    double timeout;
    if (fl_get_button(ToolBoxFormMetaButton) ||
	fl_get_button(FKeyBarFormMetaButton)
    ){
	timeout = toolbox_ctrl_timeout;
	tool_ctrl_on_flag = TRUE;
    } else {
	timeout = 0.0;
	tool_ctrl_on_flag = FALSE;    
    }
    fl_set_timer(FKeyBarFormMetaTimer, timeout);
    fl_set_timer(ToolBoxFormMetaTimer, timeout);
    toolbox_metach_sync();
}

static void toolbox_metach_sync(void){
    fkeybar_metach_sync();

    fl_freeze_form(ToolBoxForm);
    // unconditional tree stuff 
    if (toolbox_tree_flag) {
	fl_hide_object(ToolBoxFormPad0Button);
	fl_hide_object(ToolBoxFormPadVirguleButton);
	fl_hide_object(ToolBoxFormPadAsteriskButton);
    } else {
	fl_show_object(ToolBoxFormPad0Button);
	fl_show_object(ToolBoxFormPadVirguleButton);
	fl_show_object(ToolBoxFormPadAsteriskButton);
    }

    if(tool_ctrl_on_flag){
	fl_set_object_label(ToolBoxFormMetaButton, "Meta ON");
	fl_set_button(ToolBoxFormMetaButton, TRUE);
	
	//fl_set_object_label(ToolBoxFormFText,      "");	
	fl_set_object_label(ToolBoxFormF1Text,      "Global Hlp");
	fl_set_object_label(ToolBoxFormF2Text,      "Global Opt");
	fl_set_object_label(ToolBoxFormF3Text,      "Make Path");
	//fl_set_object_label(ToolBoxFormF4Text,      "(Dir Tree)");
	fl_set_object_label(ToolBoxFormF5Text,      "EditHistory");
	//fl_set_object_label(ToolBoxFormF7Text,      "(Rename)");
	fl_set_object_label(ToolBoxFormF8Text,      "Mv&Rnme");	
	//fl_set_object_label(ToolBoxFormF9Text,      "(Duplicate)");
	fl_set_object_label(ToolBoxFormF10Text,     "Cp&Rnme");	
	//fl_set_object_label(ToolBoxFormF11Text,     "(Delete)");
	fl_set_object_label(ToolBoxFormF12Text,     "Fast Exit");

	if (toolbox_tree_flag) {
	    fl_hide_object(ToolBoxFormF6Button);
	    fl_hide_object(ToolBoxFormF6Text);
	    fl_hide_object(ToolBoxFormF8Button);
	    fl_hide_object(ToolBoxFormF8Text);
	    fl_hide_object(ToolBoxFormF10Button);
	    fl_hide_object(ToolBoxFormF10Text);
	} else {
	    fl_set_object_label(ToolBoxFormF6Text,      "MultiEdit");	
	}
	fl_set_object_label(ToolBoxFormPad5Button,      "T");	
	fl_set_object_label(ToolBoxFormPad7Button,      "H");	
		
	fl_hide_object(ToolBoxFormF4Button);
	//fl_hide_object(ToolBoxFormF5Button);
	fl_hide_object(ToolBoxFormF7Button);
	fl_hide_object(ToolBoxFormF9Button);
	fl_hide_object(ToolBoxFormF11Button);

	fl_hide_object(ToolBoxFormF4Text);
	//fl_hide_object(ToolBoxFormF5Text);
	fl_hide_object(ToolBoxFormF7Text);
	fl_hide_object(ToolBoxFormF9Text);
	fl_hide_object(ToolBoxFormF11Text);

	fl_hide_object(ToolBoxFormPad1Button);
	fl_hide_object(ToolBoxFormPad2Button);
	fl_hide_object(ToolBoxFormPad3Button);
	fl_hide_object(ToolBoxFormPad4Button);
	fl_hide_object(ToolBoxFormPad6Button);
	fl_hide_object(ToolBoxFormPad8Button);
	fl_hide_object(ToolBoxFormPad9Button);
	 
    } else {
	fl_set_object_label(ToolBoxFormMetaButton, "Meta off");
	fl_set_button(ToolBoxFormMetaButton, FALSE);

	fl_set_object_label(ToolBoxFormF1Text,      "Local Help");	
	fl_set_object_label(ToolBoxFormF2Text,      "Menu");	
	fl_set_object_label(ToolBoxFormF3Text,      "Make Dir");	
	//fl_set_object_label(ToolBoxFormF4Text,      "Dir Tree");	
	//fl_set_object_label(ToolBoxFormF7Text,      "Rename");	
	fl_set_object_label(ToolBoxFormF8Text,      "Move");	
	//fl_set_object_label(ToolBoxFormF9Text,      "Duplicate");	
	fl_set_object_label(ToolBoxFormF10Text,     "Copy");	
	//fl_set_object_label(ToolBoxFormF11Text,     "Delete");	
	fl_set_object_label(ToolBoxFormF12Text,     "Exit");	
	
	fl_set_object_label(ToolBoxFormPad5Button,      "@circle");	
	fl_set_object_label(ToolBoxFormPad7Button,      "@8->|"); 
	
	if (toolbox_tree_flag) {
	    fl_set_object_label(ToolBoxFormF5Text,      "Info");	
	    fl_set_object_label(ToolBoxFormF6Text,      "Browser");	
	} else {
	    fl_set_object_label(ToolBoxFormF5Text,      "QuickEdit");	
	    fl_set_object_label(ToolBoxFormF6Text,      "Edit");	
	}
	fl_show_object(ToolBoxFormF4Button);
	fl_show_object(ToolBoxFormF5Button);
	fl_show_object(ToolBoxFormF7Button);
	fl_show_object(ToolBoxFormF9Button);
	fl_show_object(ToolBoxFormF11Button);
	
	fl_show_object(ToolBoxFormF4Text);
	fl_show_object(ToolBoxFormF5Text);
	fl_show_object(ToolBoxFormF7Text);
	fl_show_object(ToolBoxFormF9Text);
	fl_show_object(ToolBoxFormF11Text);
		
	// possibly hidden via tree:
	fl_show_object(ToolBoxFormF6Button);
	fl_show_object(ToolBoxFormF6Text);
	fl_show_object(ToolBoxFormF8Button);
	fl_show_object(ToolBoxFormF8Text);
	fl_show_object(ToolBoxFormF10Button);
	fl_show_object(ToolBoxFormF10Text);
	
	fl_show_object(ToolBoxFormPad1Button);
	fl_show_object(ToolBoxFormPad2Button);
	fl_show_object(ToolBoxFormPad3Button);
	fl_show_object(ToolBoxFormPad4Button);
	fl_show_object(ToolBoxFormPad6Button);
	fl_show_object(ToolBoxFormPad8Button);
	fl_show_object(ToolBoxFormPad9Button);
    }
    fl_redraw_form(ToolBoxForm);
    fl_unfreeze_form(ToolBoxForm);
}

void ToolBoxFormSplitButtonCB(FL_OBJECT *, long){
    // the sync here insures good operation while the window is
    // visible, but doesn't mess up the coords on initial loading.
    var_sync();
    toolbox_split();
}

static void resize_fudge(int none_flag = TRUE) {
    return;
    fl_set_object_resize(ToolBoxFormGroup, 
	none_flag ? FL_RESIZE_NONE : FL_RESIZE_ALL
    );
    fl_set_object_gravity(ToolBoxFormGroup, 
	none_flag ? NorthWestGravity : ForgetGravity,
	none_flag ? NorthWestGravity : ForgetGravity
    );
}

static void toolbox_reshape(int big_flag = TRUE){
    FL_Coord 	new_h;
    resize_fudge();
    fl_freeze_form(ToolBoxForm);
    if (big_flag) {
	fl_set_object_label(ToolBoxFormSplitButton, "@2");
	fl_show_object(ToolBoxFormGroup);
	new_h = ToolBoxFormF12Button->y + ToolBoxFormF12Button->h + 5;
    } else {
	fl_set_object_label(ToolBoxFormSplitButton, "@8");
	fl_hide_object(ToolBoxFormBox0);
	fl_hide_object(ToolBoxFormBox1);
	fl_hide_object(ToolBoxFormBox2);
	fl_hide_object(ToolBoxFormF1Button);
	fl_hide_object(ToolBoxFormF2Button);
	fl_hide_object(ToolBoxFormF3Button);
	fl_hide_object(ToolBoxFormF4Button);
	fl_hide_object(ToolBoxFormF5Button);
	fl_hide_object(ToolBoxFormF6Button);
	fl_hide_object(ToolBoxFormF7Button);
	fl_hide_object(ToolBoxFormF8Button);
	fl_hide_object(ToolBoxFormF9Button);
	fl_hide_object(ToolBoxFormF10Button);
	fl_hide_object(ToolBoxFormF11Button);
	fl_hide_object(ToolBoxFormF12Button);
	
	fl_hide_object(ToolBoxFormF1Text);
	fl_hide_object(ToolBoxFormF2Text);
	fl_hide_object(ToolBoxFormF3Text);
	fl_hide_object(ToolBoxFormF4Text);
	fl_hide_object(ToolBoxFormF5Text);
	fl_hide_object(ToolBoxFormF6Text);
	fl_hide_object(ToolBoxFormF7Text);
	fl_hide_object(ToolBoxFormF8Text);
	fl_hide_object(ToolBoxFormF9Text);
	fl_hide_object(ToolBoxFormF10Text);
	fl_hide_object(ToolBoxFormF11Text);
	fl_hide_object(ToolBoxFormF12Text);
	new_h = ToolBoxFormPad0Button->y + ToolBoxFormPad0Button->h + 5;
    }
    ToolBoxForm->h = new_h;
    ToolBoxForm->first->h = ToolBoxForm->first->next->h = new_h;
    toolbox_wh = new_h;
    fl_unfreeze_form(ToolBoxForm);
    resize_fudge(FALSE);
}

void toolbox_split(void){
    int visible_flag = ToolBoxForm->visible;
    toolbox_split_flag = !toolbox_split_flag;
    if (visible_flag) {
	fl_hide_form(ToolBoxForm);
    }
    if (toolbox_split_flag) {
        toolbox_reshape(FALSE);
        if (fkeybar_visible_flag) {
            fkeybar_show_form();
        }
    } else {
        toolbox_reshape();
        if (fkeybar_visible_flag) {
            fkeybar_close_form();
            fkeybar_visible_flag = TRUE;    // Mung it here.
        }
    }
    //var_sync();
    if (visible_flag) {
	toolbox_show_form();
    }
}

void toolbox_attach(int tree_flag){
    assert(source_browser);
    if (!source_browser) return;
    static int last_ID = 0;
    int new_ID = source_browser->get_ID();
    //cerr << "attach(" << tree_flag << "), status " << toolbox_tree_flag << endl;
    if (toolbox_tree_flag == tree_flag) {
	// no change, optimize away, except for
	// special case for changing browsers only:
	if (!tree_flag && last_ID != new_ID) {
	    last_ID = new_ID;
	    fkeybar_set_wintitle();
	    toolbox_set_wintitle();
	}
	return;
    }
    toolbox_tree_flag = tree_flag;
    last_ID = new_ID;
    toolbox_attach_title();
    toolbox_metach_sync();
}

void toolbox_attach_title(void){
    fkeybar_set_wintitle();
    toolbox_set_wintitle();
}
