/*
 * functionality for tree colour panel.
 * basically, it can be popped up anytime, and needs sync with the tree.
 */

#include "browser.h"
#include "tree_colour_forms.h"
#include "tree_colour_forms_aux.h"
#include "tree_forms_aux.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "menu.h"

void TreeColourCloseReturnButtonCB(FL_OBJECT *, long){
    if(TreeColourForm->visible){
	focus_restore_start();
	fl_hide_form(TreeColourForm);
	focus_restore_finish();
    }
}

void TreeColourHelpButtonCB(FL_OBJECT *, long){
    load_help_browser(tree_colour_help_text, "Tree Colour Editing");
}

static void tree_colour_help_func(void){
    TreeColourHelpButtonCB(NULL,0);
}

static long predefined_menu = 0;

static int tree_colour_form_raw_cb(FL_FORM *form, void *xev){
    XEvent *xevent = (XEvent *)xev;
    if(form != TreeColourForm){
	return FALSE;
    }
    if(adjust_return_button(TreeColourForm, xevent)){
	return TRUE;
    }
    return standard_form_event_cb(TreeColourForm, xevent, NULL, tree_colour_help_func);
}

extern int 
    tree_lines_fg_colour, tree_lines_bg_colour, tree_normal_bg_colour, tree_normal_fg_colour,
    tree_current_bg_colour, tree_current_fg_colour, tree_link_bg_colour, tree_link_fg_colour
;
extern void tree_sync_colours(void);

static void tree_colour_form_sync(void){
    fl_set_object_color(TreeColourNormalFButton,  tree_normal_fg_colour,  47);
    fl_set_object_color(TreeColourNormalBButton,  tree_normal_bg_colour,  47);
    fl_set_object_color(TreeColourLinkFButton,    tree_link_fg_colour,    47);
    fl_set_object_color(TreeColourLinkBButton,    tree_link_bg_colour,    47);
    fl_set_object_color(TreeColourCurrentFButton, tree_current_fg_colour, 47);
    fl_set_object_color(TreeColourCurrentBButton, tree_current_bg_colour, 47);
    fl_set_object_color(TreeColourLinesFButton,   tree_lines_fg_colour,   47);
    fl_set_object_color(TreeColourLinesBButton,   tree_lines_bg_colour,   47);

    fl_set_object_color(TreeColourSourceColourButton, tree_source_colour, 47);
    fl_set_object_color(TreeColourTargetColourButton, tree_target_colour, 47);
}

void tree_colour_sync_defaults(void){
    tree_colour_form_sync();
}


void tree_colour_show_form(void){
    /* set up the menu */
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(TreeColourForm, FL_ALL_EVENT, tree_colour_form_raw_cb);
	
	predefined_menu = new_pup();
	add_to_pup(predefined_menu, "Colour Schemes%t", 0);
	add_to_pup(predefined_menu, "Dark & Yellow%x1", 0,"Yy");
	add_to_pup(predefined_menu, "Dark & Green%x2",  0,"Gg");
	add_to_pup(predefined_menu, "Dark & Cyan%x3",   0,"Cc");
	add_to_pup(predefined_menu, "Dark & White%x4",  0,"Ww");
	add_to_pup(predefined_menu, "Slate & Grey%x6",  0,"Ss");
	add_to_pup(predefined_menu, "Blue & Slate%x7",  0,"Bb");
	add_to_pup(predefined_menu, "Black & Blue%x8",  0,"kK");
	add_to_pup(predefined_menu, "Grey%x9",          0,"eE");
	add_to_pup(predefined_menu, "Grey & Blue%x10",  0,"uU");
	/*
	add_to_pup(predefined_menu, "Dark & Red%x2",    0,"Rr");
	add_to_pup(predefined_menu, "Reddish%x4",       0,"dD");
	add_to_pup(predefined_menu, "Greenish%x5",      0,"nN");
	add_to_pup(predefined_menu, "Blueish%x6",       0,"Bb");
	add_to_pup(predefined_menu, "Yellowish%x7",     0,"lL");
	add_to_pup(predefined_menu, "Whitish%x8",       0,"Ww");
	add_to_pup(predefined_menu, "Greyish%x9",       0,"eE");
	*/
	fl_set_object_shortcut(TreeColourCloseReturnButton, "Cc^M", TRUE);
	fl_set_button_shortcut(TreeColourPredefinedButton, "Pp", TRUE);
	fl_set_button_shortcut(TreeColourDefaultsButton, "Dd", TRUE);
    }
    tree_colour_form_sync();
    set_to_ret_button(TreeColourCloseReturnButton);
    focus_save();
    if(!TreeColourForm->visible){
	fl_show_form(TreeColourForm, FL_PLACE_CENTER, TRUE, "Adjust Tree Colours:");
    } else {
	XMapRaised(fl_display, TreeColourForm->window);
    }
    tool_focus_move(TreeColourForm, TreeColourCloseReturnButton);
}

void TreeColourNormalFBButtonCB(FL_OBJECT *, long val){
    int &col = val ? tree_normal_bg_colour : tree_normal_fg_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? TreeColourNormalBButton : TreeColourNormalFButton, col, 47);    
	tree_sync_colours();
    }
}

void TreeColourLinkFBButtonCB(FL_OBJECT *, long val){
    int &col = val ? tree_link_bg_colour : tree_link_fg_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? TreeColourLinkBButton : TreeColourLinkFButton, col, 47);    
	tree_sync_colours();
    }
}

void TreeColourCurrentFBButtonCB(FL_OBJECT *, long val){
    int &col = val ? tree_current_bg_colour : tree_current_fg_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? TreeColourCurrentBButton : TreeColourCurrentFButton, col, 47);    
	tree_sync_colours();
    }
}

void TreeColourLinesFBButtonCB(FL_OBJECT *, long val){
    int &col = val ? tree_lines_bg_colour : tree_lines_fg_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? TreeColourLinesBButton : TreeColourLinesFButton, col, 47);    
	tree_sync_colours();
    }
}

void TreeColourSourceTargetButtonCB(FL_OBJECT *, long val){
    int &col = val ? tree_target_colour : tree_source_colour;
    int new_col = fl_show_colormap(col);
    if(new_col != col){
	col = new_col;
	fl_set_object_color(val ? TreeColourTargetColourButton : TreeColourSourceColourButton, col, 47);
	tree_sync_colours();
    }
}

void tree_dark_white_colours_cb(void){
    tree_normal_fg_colour  = FL_WHITE;
    tree_normal_bg_colour  = FL_BOTTOM_BCOL;
    tree_link_fg_colour    = FL_YELLOW;
    tree_link_bg_colour    = FL_BLACK;
    tree_current_fg_colour = FL_BLACK;
    tree_current_bg_colour = FL_WHITE;
    tree_lines_fg_colour   = FL_WHITE;
    tree_lines_bg_colour   = FL_BLACK;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_YELLOW;
}

void tree_dark_white_blue_colours_cb(void){
    tree_dark_white_colours_cb();
    tree_current_fg_colour = FL_YELLOW;
    tree_current_bg_colour = FL_DOGERBLUE;
    tree_lines_fg_colour   = FL_YELLOW;
}

void tree_dark_yellow_colours_cb(void){
    tree_dark_white_colours_cb();
    tree_normal_fg_colour  = FL_YELLOW;
    tree_link_fg_colour    = FL_GREEN;
}

void tree_dark_green_colours_cb(void){
    tree_dark_white_colours_cb();
    tree_normal_fg_colour  = FL_GREEN;
    tree_link_fg_colour    = FL_CYAN;
}

void tree_dark_cyan_colours_cb(void){
    tree_dark_white_colours_cb();
    tree_normal_fg_colour  = FL_CYAN;
    tree_link_fg_colour    = FL_GREEN;
}

void tree_black_blue_colours_cb(void){
    tree_normal_fg_colour  = FL_WHITE;
    tree_normal_bg_colour  = FL_BLUE;
    tree_link_fg_colour    = FL_YELLOW;
    tree_link_bg_colour    = FL_BLACK;
    tree_current_fg_colour = FL_BLUE;
    tree_current_bg_colour = FL_WHITE;
    tree_lines_fg_colour   = FL_WHITE;
    tree_lines_bg_colour   = FL_BLACK;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_YELLOW;
}

void tree_grey_colours_cb(void){
    tree_normal_fg_colour  = FL_BLACK;
    tree_normal_bg_colour  = FL_COL1;
    tree_link_fg_colour    = FL_BLACK;
    tree_link_bg_colour    = FL_COL1;
    tree_current_fg_colour = FL_BLACK;
    tree_current_bg_colour = FL_WHITE;
    tree_lines_fg_colour   = FL_BLACK;
    tree_lines_bg_colour   = FL_COL1;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_RED;
}

void tree_grey_blue_colours_cb(void){
    tree_normal_fg_colour  = FL_BLACK;
    tree_normal_bg_colour  = FL_SLATEBLUE;
    tree_link_fg_colour    = FL_YELLOW;
    tree_link_bg_colour    = FL_BOTTOM_BCOL;
    tree_current_fg_colour = FL_WHITE;
    tree_current_bg_colour = FL_BLUE;
    tree_lines_fg_colour   = FL_GREEN;
    tree_lines_bg_colour   = FL_BOTTOM_BCOL;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_CYAN;
}

void tree_blue_slate_colours_cb(void){
    tree_normal_fg_colour  = FL_YELLOW;
    tree_normal_bg_colour  = FL_SLATEBLUE;
    tree_link_fg_colour    = FL_WHITE;
    tree_link_bg_colour    = FL_BLUE;
    tree_current_fg_colour = FL_BLACK;
    tree_current_bg_colour = FL_YELLOW;
    tree_lines_fg_colour   = FL_WHITE;
    tree_lines_bg_colour   = FL_BLUE;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_YELLOW;
}

void tree_slate_grey_colours_cb(void){
    tree_normal_fg_colour  = FL_BLACK;
    tree_normal_bg_colour  = FL_COL1;
    tree_link_fg_colour    = FL_BLACK;
    tree_link_bg_colour    = FL_SLATEBLUE;
    tree_current_fg_colour = FL_BLACK;
    tree_current_bg_colour = FL_WHITE;
    tree_lines_fg_colour   = FL_YELLOW;
    tree_lines_bg_colour   = FL_SLATEBLUE;

    tree_source_colour     = FL_GREEN;
    tree_target_colour     = FL_YELLOW;
}

/*
void tree__colours_cb(void){
    tree_normal_fg_colour  = FL_;
    tree_normal_bg_colour  = FL_;
    tree_link_fg_colour    = FL_;
    tree_link_bg_colour    = FL_;
    tree_current_fg_colour = FL_;
    tree_current_bg_colour = FL_;
    tree_lines_fg_colour   = FL_;
    tree_lines_bg_colour   = FL_;

    tree_source_colour     = FL_;
    tree_target_colour     = FL_;
}
*/

void TreeColourPredefinedButtonCB(FL_OBJECT *, long){
    switch(do_pup(predefined_menu)){
    case 1: tree_dark_yellow_colours_cb(); break; 
    case 2: tree_dark_green_colours_cb(); break; 
    case 3: tree_dark_cyan_colours_cb(); break; 
    case 4: tree_dark_white_colours_cb(); break; 
    case 5:  break; 
    case 6: tree_slate_grey_colours_cb(); break; 
    case 7: tree_blue_slate_colours_cb(); break; 
    case 8: tree_black_blue_colours_cb(); break; 
    case 9: tree_grey_colours_cb(); break; 
    case 10:tree_grey_blue_colours_cb(); break; 
    default: return;
    }
    tree_colour_form_sync();
    tree_sync_colours();
}

void TreeColourDefaultsButtonCB(FL_OBJECT *, long){
    tree_dark_yellow_colours_cb();
    tree_colour_form_sync();
    tree_sync_colours();
}


