/*
 * the UC Tree - the coolest directory tree in the Unix world !
 */

#include <ctype.h>
#include <errno.h>
#include <iostream.h>
#include <limits.h>     // for NAME_MAX
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include "global.h"
#include "action_forms_aux.h"
#include "browser.h"
#include "cb.h"
#include "cb_duplicate.h"
#include "cb_key.h"
#include "cb_fkey.h"
#include "cb_remove.h"
#include "cb_table.h"
#include "client_msg.h"
#include "dirinfo_forms_aux.h"
#include "dir_menu.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "menu.h"
#include "license_forms_aux.h"
#include "options_forms_aux.h"
#include "pathnames.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "bitmaps/tree.bm"
#include "tree_colour_forms_aux.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "wildcard.h"
#include "wm_forms_aux.h"
#include "xshell.h"

#ifdef LINUX_GLIBC2
#include <errno.h>
#endif

#if !defined(HP) && !defined(DEC_ALPHA) && !defined(AIX) && !defined(LINUX) && !defined(SOLARIS) && !defined(GENERICBSD) && !defined(UNIXWARE) &&!defined(SINIX) && !defined(SUNOS) && !defined(SCO) && !defined(LINUX_GLIBC2)
#       include <bstring.h>
#endif

#if defined(HP) || defined(SGI) || defined(DEC_ALPHA) || defined(SOLARIS)
#	include <dirent.h>
#else 
#	include <sys/dir.h> // SUN, AIX4.1
#endif

#if defined(SUNOS)
// no alphasort prototype 
extern "C" int alphasort(struct direct **,struct direct **);
#endif

#if defined(UNIXWARE) || defined(SINIX) || defined(SCO)
#include "scandir.h"
#endif

int tree_source_colour      = FL_GREEN,
    tree_target_colour      = FL_YELLOW,
    tree_visible_flag       = FALSE,
    tree_ww = 600,
    tree_wh = 400,
    tree_wx = 10,
    tree_wy = 40
;

static void new_root_rel(void);

/* use new X form coordinates as coordinate system. */
static const int
    box_unit     = 20,                    // also arrow button height.
    frame_margin = 5,                     // generic
    bottom_margin= box_unit + frame_margin,
    tool_width   = 0,
    left_margin  = tool_width + box_unit + 2 * frame_margin,  
    right_margin = frame_margin,
    top_margin   = frame_margin + box_unit
;

static int
    bIsSearchMode   = FALSE,
    button_w          = 50,                  // width
    button_h          = box_unit,            // height
    button_x_gap      = (int) floor(button_w / 4),        // between buttons.
    button_y_gap      = (int) floor(button_h / 5),        // between buttons.(tested)

    dnd_notarget_flag = FALSE,                            // true mouse not on target_node
    tree_char_height  = 15,                               // pot. dangerous
    tree_font_size    = FL_NORMAL_SIZE,
    tree_font_style   = FL_TIMES_STYLE,                                 

    root_rel_x = 0,
    root_rel_y = 0,

    tree_has_changed_flag             = FALSE,
    tree_packing_level                = 1,
    tree_name_length                  = 6, 
    tree_target_mode_flag             = FALSE,
    tree_target_mode_was_visible_flag = FALSE    
;

int 
    tree_normal_fg_colour  = FL_YELLOW,
    tree_normal_bg_colour  = FL_BOTTOM_BCOL,  
    tree_link_fg_colour    = FL_GREEN,
    tree_link_bg_colour    = FL_BLACK,
    tree_current_fg_colour = FL_BLACK,
    tree_current_bg_colour = FL_WHITE,
    tree_lines_fg_colour   = FL_WHITE, 
    tree_lines_bg_colour   = FL_BLACK 
;

static char
    *last_wintitle           = NULL, 
    path_buf[2*MAXPATHLEN+1] = ""
;

const char *tree_ID_string = "tree";

int dnd_copy_flag = FALSE;
tree_node_c *dnd_old_target_node = NULL;

prm_entry_t tree_prm_table[] = {
    {"source_colour", int_e, (void *) &tree_source_colour}, 
    {"target_colour", int_e, (void *) &tree_target_colour}, 
    {"ww",            int_e, (void *) &tree_ww}, 
    {"wh",            int_e, (void *) &tree_wh}, 
    {"wx",            int_e, (void *) &tree_wx}, 
    {"wy",            int_e, (void *) &tree_wy}, 
    {"packing_level", int_e, (void *) &tree_packing_level}, 
    {"font_size",     int_e, (void *) &tree_font_size}, 
    {"font_style",    int_e, (void *) &tree_font_style}, 
    {"name_length",   int_e, (void *) &tree_name_length}, 
    {"visible_flag",  int_e, (void *) &tree_visible_flag},
    {"NormalFgCol",   int_e, (void *) &tree_normal_fg_colour},
    {"NormalBgCol",   int_e, (void *) &tree_normal_bg_colour},
    {"LinkFgCol",     int_e, (void *) &tree_link_fg_colour},
    {"LinkBGCol",     int_e, (void *) &tree_link_bg_colour},
    {"CurrentFgCol",  int_e, (void *) &tree_current_fg_colour},
    {"CurrentBgCol",  int_e, (void *) &tree_current_bg_colour},
    {"LinesFgCol",    int_e, (void *) &tree_lines_fg_colour},
    {"LinesBgCol",    int_e, (void *) &tree_lines_bg_colour},
    {"",              int_e, NULL}
};


class tree_node_c;
class tree_node_c {
private:
    void _get_path(void);   // side effect function.
    void clip_line(
        int x,int y,int w,int h,
	int col, int up_arrow = FALSE, int down_arrow = FALSE, int left_arrow = FALSE, int right_arrow = FALSE
    );
    int my_ID;
    void make_link_button(void);
public:
    static int next_ID;
    char *name;
    char *link_name;
//    int collapsed;      // if collapsed, don't show
    FL_OBJECT *button, *link_button;
    unsigned char hidden_flag, no_perms_flag, symlink_flag;
    tree_node_c
        *child0,  // 0th child. Other childs are child0->next's.
        *parent,
	*next,    // in list. for same  on level: if null, parent->prev->child0->next until0.
	*prev     
    ;
    // tree_node_c(void);
    ~tree_node_c(void);
    
    tree_node_c(const char *name_iv); 
    tree_node_c
        *attempt_change_to_first(int level_count),
        *attempt_change_to_last(int level_count),
	*cut_yourself(void),                    // extract (not delete) self from tree if possible.
	*find_child  (const char *name),        // look for child with matching name
	*find_y      (tree_node_c *best_node, int exact_x, int target_y),
	*find_deepest(int my_level, int &max_level),
	*find_geometrical_xmax (int &xmax),
	*find_last   (void),                    // bottom of tree
	*find_node   (tree_node_c *node),       // verify a pointer (return node itself on s)
	*find_object (FL_OBJECT *button),       // look for pointer by object
	*find_pattern(const char *pattern, int &current_index, int pattern_index),
	*find_path   (const char *path),        // look for node by path if avail
	*insert_child(tree_node_c *new_child),  // take existing node and splice it in
	*insert_new  (const char *new_name),    // make a new one and insert among childs.
        *down_arrow  (void),
        *left_arrow  (void),
        *right_arrow (void),
        *up_arrow    (void)
    ;
    void
        delete_sub_trees_below(int cutoff_level),      // delete all subtrees below a certain level.
        delete_sub_tree       (tree_node_c *sub_root), // delete the subtree at sub_root if exist
        realign               (void),
        recolour              (void),
	reconnect             (tree_node_c *new_parent),    // first cut, then paste yourself
        retrunc               (int recursive_flag = TRUE),  // name. 
        draw_lines            (void),
        set_current           (int flag)
    ;
    int dump_tree       (ofstream &ofile),
	read_link        (void),              // if we're link, get the name
        restat          (int quiet_flag, int subtree_flag = TRUE)
    ;
    const char *get_path(void);
};

int tree_node_c::next_ID = 0;

static tree_node_c
    *root = NULL,
    *current_node = NULL,
    *source_node = NULL,
    *target_node = NULL
;


static tree_node_c *search_last_node = NULL;
static int tree_auto_sync_dir_flag = FALSE;
static int tree_new_target_done_flag = FALSE;
static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

static const int SCAN_ALL_LEVELS = -2;
static int tree_scanning_flag = FALSE, scan_continue_flag = TRUE;

static int
    cnode_is_stable   (void),  // query/verify function
    tree_dump         (void),
    tree_restat       (tree_node_c *node, int quiet_flag = FALSE),
    tree_restat_below (tree_node_c *node, int recurse_level, int quiet_flag = FALSE),
    tree_scan         (
        tree_node_c *parent,
        int max_levels = 2,
        int max_nodes = 500,
        int rescan_flag = FALSE
    ),
    tree_scan_below   (
        tree_node_c *parent,
	int recurse_level = 0,
	int max_nodes = 500,
	int rescan_flag = FALSE
    ),
    tree_search_key_cb(XKeyEvent *xkey, KeySym keysym, char key)
;


static void
    tree_adjust_sliders         (void),
	tree_cd_home_event_cb		(FL_OBJECT *, long),
    tree_form_scale_cb          (void),
    tree_collapse               (void),
    tree_collapse_subtree_all_cb(int),
    tree_default_actions        (void),
    tree_expand                 (int allow_iteration_flag = FALSE),
    tree_expand_subtree_all_cb  (int),
    tree_hide_form              (void),   // comp. close_form!
    tree_search_clear_mode      (int bRestoreLastNode = FALSE),
    tree_search_set_mode        (void),
    tree_search_event_cb        (FL_OBJECT *, long),
    tree_search_again_event_cb  (FL_OBJECT *, long),
    tree_set_current            (
	tree_node_c *node,
	int always_center_flag = FALSE,
	int tool_sync_flag     = TRUE
    ),
    tree_set_target_cb          (FL_OBJECT *, long),
	tree_center					(int always_center_flag = FALSE),
    tree_destroy_node           (tree_node_c *tmp_node),
    tree_remove_invalid_node    (tree_node_c *node),
    tree_refit                  (void),
    tree_retrunc                (void),
    tree_set_wintitle_backend   (const char *title),
    tree_set_source             (tree_node_c *new_node),
    tree_set_target             (tree_node_c *new_node)
;

// operation functions
void
    tree_copy     (void),
    tree_delete   (void),
    tree_duplicate(void),
    tree_edit     (void),
    tree_mkdir    (void),
    tree_move     (void),
    tree_rename   (void),
    tree_view     (void)
;

void tree_reshape(void);



static void DirButtonCB(FL_OBJECT *, long val);
static void LinkButtonCB(FL_OBJECT *, long val);

static void var_sync(void){
    tree_ww = TreeForm->w;
    tree_wh = TreeForm->h;
    tree_wx = TreeForm->x;
    tree_wy = TreeForm->y;
}

const char *tree_get_path(void){
    if(!cnode_is_stable()){
		return "<empty>";
	}
    const char *cwd_buf = current_node->get_path();
    if (!cwd_buf) {
	return "<empty>";
    }
    return cwd_buf;
}

void tree_save_defaults(ofstream &ofile){
    if(TreeForm->visible){
	var_sync();
    }
    resource_save_defaults(ofile, tree_prm_table, tree_ID_string);
    // tree_dump();
}    

int tree_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    return resource_load_default(
        tree_prm_table,  
        tree_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

void tree_sync_colours(void){
    fl_freeze_form(TreeForm);
    fl_set_object_color(TreeFormBox, tree_lines_bg_colour, FL_COL1);
    fl_set_object_color(TreeSourceBox, tree_source_colour, FL_COL1);
    fl_set_object_color(TreeTargetBox, tree_target_colour, FL_COL1);
    if(root){
	root->recolour();
    }
    fl_redraw_form(TreeForm);
    fl_unfreeze_form(TreeForm);
}

void tree_sync_defaults(void){
    // after loading.
    tree_retrunc();
    tree_sync_colours();
    if (tree_visible_flag) {
	do_tree();
    } else {
	tree_hide_form();
    }
}

void tree_sync_to_target_browser(void){
    if(!root) return;
    tree_node_c *tmp_node = root->find_path(target_browser->get_cwd());
    if(!tmp_node){
	tmp_node = tree_construct_path(target_browser->get_cwd());
	if(!tmp_node) return;
    }
    tree_set_target(tmp_node);
    if(tree_target_mode_flag){
	tree_set_current(tmp_node, FALSE, FALSE);   // external sync, don't update again.
    }
}

void tree_close_form(void){
    int hide_flag = TRUE;
    if(TreeForm->visible){
	var_sync();
	if(tree_target_mode_flag){
	    tree_new_target_done_flag = TRUE;
	    tree_target_mode_flag = FALSE;
	    if(tree_target_mode_was_visible_flag){
		XLowerWindow(fl_display, TreeForm->window);
		tree_set_complete_wintitle();
		hide_flag = FALSE;
	    }
	}
	tree_default_actions();
	if(hide_flag){
	    focus_restore_start();
	    tree_hide_form();
	    focus_restore_finish();
	    toolbox_attach(FALSE);
	} else {
	    focus_restore();	// Probably OK.
	}
    }
    tree_visible_flag = FALSE;
}

static void tree_raise_form(void){
    XMapRaised(fl_display, TreeForm->window);
}

void tree_iconify(void){    
    if(TreeForm->visible) XIconifyWindow(fl_display, TreeForm->window, fl_screen);
}

void tree_uniconify(void){    
    if(TreeForm->visible) XMapRaised(fl_display, TreeForm->window);
}

static int shove_flag = FALSE, dnd_flag = FALSE, xdnd_flag = FALSE;
static tree_node_c *deact_node = NULL;

int tree_raw_cb(FL_FORM *form, void *xev){
    //cerr << "raw cb called. \n";
    if(!form){
	cerr << "NULL form in raw cb.\n";
	return FALSE;
    }
    /* strategy: mini-timeout. If no button release event in queue,
     * assume its a shove. Otherwise, I would need to generate a fake
     * press or release for forms - which is difficult.
     */
    XEvent *xevent = (XEvent *)xev;
    if(xevent->xany.window != TreeForm->window){
	cerr << "tree form raw cb with wrong window ID !";
    }
    
    switch(xevent->type){
	case FocusIn:	
		//cerr << "Tree FocusIn.\n";
		toolbox_attach();
		return FALSE;
	case FocusOut:	
		//cerr << "Tree FocusOut.\n";
		//toolbox_attach(FALSE);
		return FALSE;
 	case EnterNotify:
		//cerr << "Tree EnterNotify.\n";
		if(xevent->xcrossing.focus){
			toolbox_attach();
		}
		return FALSE;
 	case LeaveNotify:
		//cerr << "Tree LeaveNotify.\n";
		if(xevent->xcrossing.focus){
			toolbox_attach();
		}
		return FALSE;
    case ButtonPress:
	if(dbg_flag) cerr << "tree press.\n";
	shove_flag = FALSE;  // reset, just in case.
	dnd_flag   = FALSE;
	if(!shove_flag){
	    // if not in shove mode (buttons pressed), it may be a slider or button.
	    if (xevent->xbutton.x < left_margin || 
		xevent->xbutton.y < top_margin ||
		xevent->xbutton.y > TreeForm->h - bottom_margin
	    ){
		return FALSE;
	    }
	}

	if(xevent->xbutton.button != Button1 ||    // only on 1.
	   button_down(xevent->xbutton.state)   // no complications, please
	   ){
	    // only shove on button 1.
	    return FALSE;
	}
	if(controlkey_down(xevent->xbutton.state)){  // k.i.s.s.
	    return FALSE;
	}

	sginap(15);  // 10 was a little fast.
	if(CheckPeekWindowEvent(fl_display, TreeForm->window, ButtonRelease)){
	    if(dbg_flag)cerr << "is real click.\n";
	    shove_flag = FALSE;
	    dnd_flag   = FALSE;
	    return FALSE;
	} else {
	    if(dbg_flag)cerr << "is HOLD.\n";
	    // is it on a button ? (obj under the mouse)
	    FL_OBJECT *obj;
	    obj = fl_find_last(TreeForm, FL_FIND_MOUSE, xevent->xbutton.x, xevent->xbutton.y);
	    if(obj == TreeFormFreeObj){
		shove_flag = TRUE;
	    } else if(root && !tree_target_mode_flag){
		deact_node = root->find_object(obj);
		if(deact_node){
		    if(!deact_node->button){
			cerr << "warning: node without button.\n";
			deact_node = NULL;
			return FALSE;
		    }
		    if(deact_node == root){
			status_msg(1, "cannot move root directory.\n");
			deact_node = NULL;
			return FALSE;
		    } else {
			if(dbg_flag)cerr << "deact_node: " << deact_node->name << endl;
			fl_deactivate_object(deact_node->button);
			dnd_flag = TRUE;
			return FALSE;
		    }
		} else {
		    // internal error
		    if(dbg_flag)cerr << "null node\n";
		    return FALSE;
		}
	    } else {
		// internal error
	    }
	    return FALSE;
	}
    case KeyPress:
	// necessary because FL_KEY_ALL is fucked.
	tree_handle_free(TreeFormFreeObj, FL_KEYBOARD, 0, 0, 0, xevent);
	return TRUE;
    case KeyRelease:
	return TRUE;
    default:
	break;
    }
    return FALSE;
}

static int tree_atclose_cb(FL_FORM*, void*){
    tree_close_form();
    return FL_IGNORE;
}

static void tree_hide_form(void){
    if(TreeForm->visible){
	fl_hide_form(TreeForm);
    }
    tree_visible_flag = FALSE;
}

void tree_show_form(void){
    tree_visible_flag = TRUE;
    static int first = TRUE;
    if(first){
	first = FALSE;
	// fl_set_form_dblbuffer(TreeForm, TRUE); // too slow & buggy
	/* done by hand in tree_forms.c and .h:
	*TreeFormFreeObj,
	extern int tree_handle_free(FL_OBJECT *obj, int event, FL_Coord x, FL_Coord y, int key, void *xev);
	fl_addto_form(TreeForm);
	TreeFormFreeObj = obj = fl_add_free(FL_INPUT_FREE,0,0,TreeForm->w,TreeForm->h,"",tree_handle_free);
	obj->input = TRUE;
	obj->wantkey = FL_KEY_ALL; // doesn't work yet.
	fl_end_form();
	*/
	fl_register_call_back(TreeForm, FL_ALL_EVENT, tree_raw_cb);
	//fl_register_call_back(TreeForm, ButtonPressMask|ButtonReleaseMask, tree_raw_cb);
	fl_set_form_atclose(TreeForm, tree_atclose_cb, NULL);
	
	fl_set_object_resize(TreeFormBox,            FL_RESIZE_ALL);
	fl_set_object_resize(TreeFormHorSlider,      FL_RESIZE_X);
	fl_set_object_resize(TreeFormVertSlider,     FL_RESIZE_Y);
	fl_set_object_resize(TreeFormHorLeftButton,  FL_RESIZE_NONE);
	fl_set_object_resize(TreeFormHorRightButton, FL_RESIZE_NONE);
	fl_set_object_resize(TreeFormVertDownButton, FL_RESIZE_NONE);
	fl_set_object_resize(TreeFormVertUpButton,   FL_RESIZE_NONE);
	fl_set_object_resize(TreeSourceBox,          FL_RESIZE_NONE);
	fl_set_object_resize(TreeTargetBox,          FL_RESIZE_NONE);
	//fl_set_object_resize(TreeForm, FL_RESIZE_NONE);

	fl_set_object_gravity(TreeFormBox,            NorthWestGravity, SouthEastGravity);
	fl_set_object_gravity(TreeFormHorSlider,      SouthWestGravity, SouthEastGravity);
	fl_set_object_gravity(TreeFormVertSlider,     NorthWestGravity, SouthWestGravity);
	fl_set_object_gravity(TreeFormHorLeftButton,  SouthWestGravity, SouthWestGravity);
	fl_set_object_gravity(TreeFormHorRightButton, SouthEastGravity, SouthEastGravity);
	fl_set_object_gravity(TreeFormVertDownButton, SouthWestGravity, SouthWestGravity);
	fl_set_object_gravity(TreeFormVertUpButton,   NorthWestGravity, NorthWestGravity);
	fl_set_object_gravity(TreeSourceBox,          ForgetGravity,    ForgetGravity);
	fl_set_object_gravity(TreeTargetBox,          ForgetGravity,    ForgetGravity);
	//fl_set_object_gravity(TreeForm, NorthWestGravity);

	fl_set_object_dblbuffer(TreeFormHorSlider,  TRUE);
	fl_set_object_dblbuffer(TreeFormVertSlider, TRUE);
	
	//tree_form_scale_cb();
    }

    focus_save();
    if(!TreeForm->visible){
	tree_search_clear_mode(),

	// freezing may interfere with centering ? (on alpha)
	//fl_freeze_form(TreeForm);
	fl_set_form_minsize(TreeForm, 40, 40);
	fl_set_form_maxsize(TreeForm, tool_max_ww, tool_max_wh);
	fl_set_initial_placement(TreeForm, tree_wx, tree_wy, tree_ww, tree_wh);
	fl_show_form(TreeForm, FL_PLACE_FREE, TRUE, "UCTree:");

	if (last_wintitle) {
	    fl_free(last_wintitle);
	    last_wintitle = NULL;
	}
	// icon stuff is 85x67 on SGI's.
	static Pixmap pixmap =0;
	if(pixmap == 0){
	    pixmap = XCreatePixmapFromBitmapData(
	        fl_display, TreeForm->window, tree_bits, tree_width,
	        tree_height,
	        WhitePixel(fl_display, fl_screen),
	        BlackPixel(fl_display, fl_screen),
	        1
	    );
	}
	set_window_icon(TreeForm->window, pixmap);
	XSetIconName(fl_display, TreeForm->window, "UCTree");
	tree_center();
	//fl_unfreeze_form(TreeForm);
    } else {
	XMapRaised(fl_display, TreeForm->window);
    }

    if(current_node){
	tool_focus_move(TreeForm, current_node->button);
	return;
    } else {
	tool_focus_move(TreeForm, TreeFormFileButton);
    }
}

tree_node_c::~tree_node_c(void){
    // don't delete subtree. That'd be a bit dangerous.
    // also, null all pointers for safety.
    if(this == source_node)  source_node = 0;
    if(this == target_node)  target_node = 0;
    if(this == current_node) current_node = 0;
    
    if(button){
	fl_hide_object(button);
	fl_delete_object(button);
	fl_free_object(button);
	button = NULL;
    }
    if(link_button){
	fl_hide_object(link_button);
	fl_delete_object(link_button);
	fl_free_object(link_button);
	link_button = NULL;
    }
    if(name){
	fl_free(name);
	name = NULL;
    } else {
        M_err("UCT tree_node_c destructor","destructor called twice (name==0)");
    }
    if(link_name){
	fl_free(link_name);
	link_name = NULL;
    }
    if(next_ID>0){
	next_ID--;
    } else {
	cerr << "hmm - destructor called with static next_ID == " << next_ID << endl;
    }
}

tree_node_c::tree_node_c(const char *name_iv){
    child0 = parent = next = prev = NULL;
    link_name = NULL;
    link_button = NULL;
    my_ID = next_ID;
    hidden_flag = no_perms_flag = symlink_flag = FALSE;
    next_ID++;
//    collapsed = FALSE;
    if(name_iv){
	name = strdup(name_iv);
    } else {
	cerr << "Warning: made a tree node class with a NULL name.\n";
	name = strdup("no name");
    }

    fl_addto_form(TreeForm);
    button = fl_add_button(FL_NORMAL_BUTTON,0,0,50,20,name);
    // frame box is marginally acceptable, all others but border_box are bad.
    // fl_set_object_boxtype(button,FL_FRAME_BOX);
    fl_set_object_boxtype(button,FL_BORDER_BOX);
    fl_set_object_lsize(button,tree_font_size);
    fl_set_object_lstyle(button,tree_font_style);
    fl_set_object_align(button,FL_ALIGN_LEFT | FL_ALIGN_INSIDE);
    fl_set_call_back(button,DirButtonCB,0);
    fl_set_object_resize(button, FL_RESIZE_NONE);
    fl_set_object_gravity(button, NorthWestGravity, ForgetGravity);

    if(tool_bw_flag){
	fl_set_object_color(button, FL_WHITE, FL_COL1);
	fl_set_object_lcol(button,FL_BLACK);
    } else {
	fl_set_object_color(button, tree_normal_bg_colour, FL_COL1);
	fl_set_object_lcol(button, tree_normal_fg_colour);
    }

    fl_end_form();

    button->argument = (long) this;
}

tree_node_c* tree_node_c::attempt_change_to_first(int level_count){
    if(level_count <= 0){
	if(parent){
	    return parent->child0;
	} else {
	    return this;
	}
    } else {
	if(child0){
	    tree_node_c *work_node = child0, *result = NULL;
	    // start from first.
	    while (work_node){
		result = work_node->attempt_change_to_first(level_count-1);
		if(result){
		    return result;
		} else {
		    work_node = work_node->next;
		}
	    }
	}
    }
    return NULL;
}

tree_node_c* tree_node_c::attempt_change_to_last(int level_count){
    if(level_count <= 0){
	tree_node_c *tmp_node = this;
	while(tmp_node->next){
	    tmp_node = tmp_node->next;
	}
	return tmp_node;
    } else {
	if(child0){
	    tree_node_c *work_node = child0, *result = NULL;
	    // go to last.
	    while(work_node->next){
		work_node = work_node->next;
	    }
	    // work up from last.
	    while (work_node){
		result = work_node->attempt_change_to_last(level_count-1);
		if(result){
		    return result;
		} else {
		    work_node = work_node->prev;
		}
	    }
	}
    }
    return NULL;
}

tree_node_c* tree_node_c::find_last(void){
    // find last in local subtree function.
    if(!child0) return this;
    tree_node_c *tmp_node = child0;
    while(tmp_node->next){
	tmp_node = tmp_node->next;
    }
    return tmp_node->find_last();
}

tree_node_c* tree_node_c::find_child(const char *child_name){
    if(!child_name){
	cerr << "Null name in find_child.\n";
	return NULL;
    }
    tree_node_c *tmp_node = child0, *result = NULL;
    while(tmp_node){
	result = tmp_node->find_path(child_name);
	if(result){
	    return result;
	} else {
	    tmp_node = tmp_node->next;
	}
    }
    return NULL;
}

tree_node_c* tree_node_c::find_node(tree_node_c *node){
    if(!node) return NULL;
    if(node == this) return node;
    if(child0){
	if(child0->find_node(node)) return node;
    }
    
    if(next){
	if(next->find_node(node)) return node;
    }
    return NULL;
}

tree_node_c* tree_node_c::find_object (FL_OBJECT *obj){
    tree_node_c *node;
    if(!obj) return NULL;
    if(obj == button) return this;
    if(child0){
	node = child0->find_object(obj);
	if(node) return node;
    }
    
    if(next){
	node = next->find_object(obj);
	if(node) return node;
    }
    return NULL;
}

tree_node_c* tree_node_c::find_path(const char *path){
    if(!path){
	cerr << "Null path in find_path.\n";
	return NULL;
    }
    char *name_p = name;             // working copy (!)
    
    if(path  [0] == '/') path++;     // skip slashes, if any.
    if(name_p[0] == '/') name_p++;
    
    // attempt direct match.
    int name_len = strlen(name_p);
    if(strncmp(path, name_p, name_len) == 0){
	path += name_len;
	if(path[0] == '\0'){
	    return this;
	} else {
	    tree_node_c *tmp_node = child0, *result = NULL;
	    while(tmp_node){
		result = tmp_node->find_path(path);
		if(result){
		    return result;
		} else {
		    tmp_node = tmp_node->next;
		}
	    }
	    return NULL;
	}
    } else {
	return NULL;
    }
}

void tree_node_c::delete_sub_trees_below(int cutoff_level){
    if(cutoff_level <= 0){
	while(child0){
	    delete_sub_tree(child0);
	}
    } else {
	cutoff_level--;
	tree_node_c *tmp_node  = child0;
	while(tmp_node){
	    tmp_node->delete_sub_trees_below(cutoff_level);
	    tmp_node = tmp_node->next;
	}
    }
}

tree_node_c *tree_node_c::find_y(tree_node_c *best_node, int exact_x, int target_y){
    tree_node_c *tmp_node = NULL;
    if(button->x > exact_x){
	// we're off, cut.
    } else if(button->x == exact_x){
	// check yourself
	if(!best_node){
	    best_node = this;
	} else {
	    if(fabs(button->y - target_y) < fabs(best_node->button->y - target_y)){
		best_node = this;
	    }
	}
	// check the rest on the same level. 
	tmp_node = next;
	while(tmp_node){
	    if(fabs(tmp_node->button->y - target_y) < fabs(best_node->button->y - target_y)){
		best_node = tmp_node;
	    }
	    tmp_node = tmp_node->next;
	}
    } else {
	// recurse.
	tmp_node = child0;
	while(tmp_node){
	    best_node = tmp_node->find_y(best_node, exact_x, target_y);
	    tmp_node = tmp_node->next;
	}	    
    }
    return best_node;
}

/* call like find deepest (0, maxlev), with max_lev (<=) = 0;
 */
tree_node_c* tree_node_c::find_deepest(int my_level, int &max_level){
    // find deepest node in local subtree function.
    if(my_level>max_level){
	max_level = my_level;
    }

    int old_max_level;

    if(!child0) return this;

    tree_node_c
	*tmp_node  = child0,
	*result    = NULL,    // subtree query
	*deep_node = child0   // safeguard
    ;
    while(tmp_node){
	old_max_level = max_level;
	result = tmp_node->find_deepest(my_level+1,max_level);
	if(max_level > old_max_level){
	    deep_node = result;
	    //if(dbg_flag) cerr << "deep node: " << deep_node->get_path() << endl;
	}
	tmp_node = tmp_node->next;
    }
    return deep_node;
}

tree_node_c* tree_node_c::find_pattern(const char *pattern, int &current_index, int pattern_index){
    if(wildcard_match(pattern, name)){
	if(current_index == pattern_index){
	    return this;
	} else {
	    current_index++;
	}
    }
    if(!child0){
	return NULL;
    }
    tree_node_c
	*tmp_node  = child0,
	*result    = NULL    // subtree query
    ;
    while(tmp_node){
	result = tmp_node->find_pattern(pattern, current_index, pattern_index);
	if(result){
	    return result;
	} else {
	    tmp_node = tmp_node->next;
	}
    }
    return NULL;
}

tree_node_c* tree_node_c::left_arrow(void){
    if(!parent) return NULL;
    
    if(parent->child0 != this || !tool_get_meta()){
	return parent;
    } else {
	tree_node_c *tmp_node = parent;
	while(tmp_node){
	    if(!tmp_node->parent){
		return tmp_node;
	    }
	    if(tmp_node->parent->child0 != tmp_node){
		return tmp_node;
	    } else {
		tmp_node = tmp_node->parent;
	    }
	}
	return NULL;
    }
}

tree_node_c* tree_node_c::right_arrow(void){
    if(child0){
	if(tool_get_meta()){
	    tree_node_c *tmp_node = child0;
	    while(tmp_node){
		if(tmp_node->child0){
		    tmp_node = tmp_node->child0;
		} else {
		    return tmp_node;
		}
	    }
	} else {
	    return child0;
	}
    }
    if(symlink_flag){
	if(!link_name){
	    read_link();
	}
	if(link_name){
	    char resolved_name_buf[MAXPATHLEN+1];
	    const char *path = get_path();
	    if(!path) return NULL;                          // safeguards.
	    if(strlen(path)<=0) return NULL;
	    if(real_path(path, resolved_name_buf)){
		return tree_construct_path(resolved_name_buf);
	    } else {
		return NULL;
	    }
	} else {
	    return NULL;
	}
    }
	    
    // step-through function.
    if(next){
	if(tool_get_meta()){
	    tree_node_c *tmp_node = next;
	    while(tmp_node){
		if(tmp_node->child0){
		    return tmp_node->child0;
		} else {
		    tmp_node = tmp_node->next;
		}
	    }
	    return next;   // if not found ... 
	} else {
	    return next;
	}
    }
    tree_node_c *tmp_node = parent;
    while(tmp_node){
	if(tmp_node->next){
	    return tmp_node->next;
	} else {
	    tmp_node = tmp_node->parent;
	}
    }
    return NULL;
}

tree_node_c* tree_node_c::down_arrow(void){
    tree_node_c *tmp_node, *result;
    if(tool_get_meta()){
	if(next){
	    tmp_node = next;
	    while(tmp_node->next){
		tmp_node = tmp_node->next;
	    }
	    return tmp_node;
	} else {
	    tmp_node = parent;	
	}
    }
    if(next) return next;
    tmp_node = parent;
    int level_count = 1;

    while(tmp_node){
	if(tmp_node->next){
	    result = tmp_node->next->attempt_change_to_first(level_count);
	    if(result){
		return result;
	    } else {
		tmp_node = tmp_node->next;
	    }
	} else {
	    tmp_node = tmp_node->parent;
	    level_count++;
	}
    }
    return child0;  // good default action.
}

tree_node_c* tree_node_c::up_arrow(void){
    tree_node_c *tmp_node, *result;
    if(tool_get_meta()){
	if(prev){
	    return parent->child0;
	} else {
	    tmp_node = parent;	
	}
    }
    if(prev) return prev;
    tmp_node = parent;	
    int level_count = 1;

    while(tmp_node){
	if(tmp_node->prev){
	    result = tmp_node->prev->attempt_change_to_last(level_count);
	    if(result){
		return result;
	    } else {
		tmp_node = tmp_node->prev;
	    }
	} else {
	    tmp_node = tmp_node->parent;
	    level_count++;
	}
    }
    return parent;  // good default action. 
}

void tree_node_c::_get_path(void){
    if(parent){
	parent->_get_path();
	int len = strlen(path_buf);
	if(len>0){
	    if(path_buf[len-1] != '/'){
		strcat(path_buf, "/");
	    }
	}
	/*
	if(symlink_flag && allow_link_flag){
	    if(link_name[0] != '/'){
		strcat(path_buf, link_name);
	    } else {
		strcpy(path_buf, link_name);
	    }
	} else {
	    strcat(path_buf, name);
	}
	*/
	strcat(path_buf, name);
	
    } else {
	/*
	if(symlink_flag && allow_link_flag){
	    strcpy(path_buf, link_name);
	} else {
	    strcpy(path_buf, name);
	}
	*/
	strcpy(path_buf, name);
    }
}

const char *tree_node_c::get_path(void){
    //_get_path(allow_link_flag);
    _get_path();
    return path_buf;
}

void tree_node_c::realign(void){
    button->w = button_w;
    button->h = button_h;
    if(!parent){
	// root
	button->x = left_margin + root_rel_x;
	button->y = top_margin + root_rel_y;
	button->w = fl_get_string_width(tree_font_style, tree_font_size, " / ", 3);
	fl_set_object_align(button, FL_ALIGN_CENTER);
	TreeSourceBox->visible = FALSE;
	TreeTargetBox->visible = FALSE;
    } else {
	if(!prev){
	    // go right
	    button->x = parent->button->x + parent->button->w + button_x_gap;
	    button->y = parent->button->y;
	} else {
	    // go down (coord: up :-)
	    button->x = prev->button->x;
	    tree_node_c *tmp_node = prev->find_last();
	    button->y = tmp_node->button->y + button->h + button_y_gap;
	}
    }
    if(button->x < left_margin ||
       button->x + button->w > TreeForm->w - right_margin ||
       button->y < top_margin ||
       button->y + button->h > TreeForm->h - bottom_margin
    ){
	//fl_hide_object(button);
	button->visible = FALSE;
    } else {
	if(!button->visible){
	    fl_show_object(button);
	}
    }
    if(this == source_node){
	int h = (int) floor(button->h/2), w = frame_margin;
	if(h < frame_margin) w = h;
	TreeSourceBox->x = button->x-w;
//	TreeSourceBox->y = button->y+h+1;
	TreeSourceBox->y = button->y;
	TreeSourceBox->w = w;
	TreeSourceBox->h = h;
	TreeSourceBox->visible = button->visible;
    }
    if(this == target_node){
	int h = (int) floor(button->h/2), w = frame_margin;
	if(h < frame_margin) w = h;
	TreeTargetBox->x = button->x-w;
//	TreeTargetBox->y = button->y;
	TreeTargetBox->y = button->y+h+1;
	TreeTargetBox->w = w;
	TreeTargetBox->h = h;
	TreeTargetBox->visible = button->visible;
    }
    if(symlink_flag && link_name){
	if(!link_button){
	    make_link_button();	    
	} else {
	    char buf[MAXPATHLEN+5];
	    sprintf(buf, "-> %s", link_name);
	    int w = fl_get_string_width(tree_font_style, tree_font_size, buf, strlen(buf))+5;
	    link_button->w = w;
	}	    
	link_button->x = button->x + button->w;
	link_button->y = button->y;
	link_button->h = button->h;
	if(!button->visible){
	    // left, bottom, top
	    link_button->visible = FALSE;
	} else {
	    // right ? (clipping would be nice).
	    if(link_button->x + link_button->w > TreeForm->w - right_margin){
		link_button->visible = FALSE;
	    } else {
		link_button->visible = TRUE;
	    }
	}
    }
    if(child0) child0->realign();  // go right first
    if(next)   next->realign();    // go down next
}

void tree_node_c::make_link_button(void){
    if(!TreeForm) return;
    if(link_button) return;

    fl_addto_form(TreeForm);
    char buf[MAXPATHLEN+5];
    sprintf(buf, "-> %s", link_name);
    int w = fl_get_string_width(tree_font_style, tree_font_size, buf, strlen(buf))+5;
    link_button = fl_add_button(FL_NORMAL_BUTTON,0, 0, w, 20 ,buf);

    //fl_set_object_align(link_button, FL_ALIGN_LEFT);
    fl_set_object_boxtype(link_button,FL_FLAT_BOX);
    fl_set_object_lsize  (link_button,tree_font_size);
    fl_set_object_lstyle (link_button,tree_font_style);
    fl_set_call_back     (link_button,LinkButtonCB,0);
    
    if(tool_bw_flag){
	fl_set_object_color(link_button, FL_WHITE, FL_COL1);
	fl_set_object_lcol (link_button,FL_BLACK);
    } else {
	fl_set_object_color(link_button, tree_link_bg_colour, FL_COL1);
	fl_set_object_lcol (link_button, tree_link_fg_colour);
    }
    fl_end_form();
    link_button->argument = (long) this;
}

static void tree_retrunc(void){
    char buf[NAME_MAX+1];
    for(int i=0; i<=tree_name_length; i++){  // one more.
	buf[i] = ' ';
    }
    buf[tree_name_length+1] = '\0';
    button_w = fl_get_string_width(tree_font_style, tree_font_size, buf, strlen(buf));
    if(tree_font_style == 8 || tree_font_style == 9){
	// variable
	button_w = (FL_Coord) (2.5 * (double) button_w);
    }
    /*
    cerr << "w: " << button_w << ", size: " << tree_font_size << ", style: ";
    cerr << tree_font_style << ", namelength: " << tree_name_length << endl;
    cerr << "buf:`" << buf << "'\n";
    */

    int ffasc, fdesc;
    tree_char_height = fl_get_char_height(tree_font_style, tree_font_size, &ffasc, &fdesc);
    switch(tree_packing_level){
    case 1:
	button_x_gap = (int) floor(button_w / 4);
	button_h = (int) (tree_char_height+fdesc+2); break;
    case 2:
	button_x_gap = (int) floor(button_w / 8);
	button_h = (int) (tree_char_height+fdesc); break;
    default:
	button_x_gap = (int) floor(button_w / 3);
	button_h = (int) floor(1.2 * (tree_char_height+fdesc));
    }
    button_h = bound_value(button_h, tree_char_height+2, tree_char_height+10);

    button_x_gap = bound_value(button_x_gap, 2*frame_margin+1, 20);

    switch(tree_packing_level){
    case 1: button_y_gap = (int) floor(button_h / 8); break;
    case 2: button_y_gap = 1; break;
    default:
	button_y_gap = (int) floor(button_h / 4);
    }
    button_y_gap = bound_value(button_y_gap, 1, 10);
    
    tree_reshape();
}

void tree_reshape(void){
    if(!root)return;
//    if(!TreeForm->visible) return;
    fl_freeze_form(TreeForm);
    root->retrunc();
    root->realign();
    tree_refit();
    root->realign();
    tree_adjust_sliders();
    fl_redraw_form(TreeForm);
    fl_unfreeze_form(TreeForm);
    tree_set_complete_wintitle();
}

int tree_node_c::read_link(void){
    char link_buf[MAXPATHLEN+1];
    char *path = (char *) get_path();
    if(!path){
        cerr << "readlink of null path\n";
	return FALSE;
    }
    int n = readlink(path, link_buf, sizeof(link_buf));
    if(n<0){
        status_msg(2, "UCTree: couldn't readlink %s: %s",  path, strerror(errno));
	return FALSE;
    } else {
	if(n == 0){
	    status_msg(2, "UCTree:cannot handle empty symbolic link names (%s)", path);
	    link_name = NULL;
	    return FALSE;
	} else {
	    link_buf[n] = '\0';
	}
	link_name = strdup(link_buf);
	return TRUE;
    }
}

void tree_node_c::recolour(void){
    if(this == current_node){
	fl_set_object_color(button, tree_current_bg_colour, FL_COL1);
	fl_set_object_lcol (button, tree_current_fg_colour);
    } else {
	fl_set_object_color(button, tree_normal_bg_colour, FL_COL1);
	fl_set_object_lcol (button, tree_normal_fg_colour);
    }
    if(link_button){
	fl_set_object_color(link_button, tree_link_bg_colour, FL_COL1);
	fl_set_object_lcol (link_button, tree_link_fg_colour);
    }
    if(child0) child0->recolour();  // go right first
    if(next)   next->recolour();    // go down next
}

void tree_node_c::reconnect(tree_node_c *new_parent){
    if(!new_parent || !root || this==root || this==new_parent){
	cerr << "UCT: reconnect failed.\n";
	return;
    }
    if(new_parent == parent) return;
    tree_node_c *cut_node = cut_yourself();
    if(!cut_node){
	cerr << "UCT: cut_yourself failed.\n";
	return;
    } else {
	new_parent->insert_child(cut_node);
	return;
    }
}

void tree_node_c::retrunc(int recursive_flag){
    static char buf[MAXPATHLEN+5];
    if(!parent && strlen(name) == 1){
	// root
	strcpy(buf, name);
    } else {
	strcpy(buf, name); // no fine adjustment yet.
	int pad = tree_char_height/2;  // 2 *h/4
	int continue_flag	= TRUE,
	    pruned_flag		= FALSE,
		olddiff			= 0,
	    width, len
	;
	while(continue_flag){
	    len   = strlen(buf);
	    width = fl_get_string_width(tree_font_style, tree_font_size, buf, len);
	    if(width < button_w - pad || len <= 0){
		continue_flag = FALSE;
	    } else {
		buf[len-1]  = '\0';
		pruned_flag = TRUE;
		olddiff     = width - (button_w - pad);
	    }
	}
	if(pruned_flag && len >= 0){
	    int idx=len;
	    if(len > 0){
		if(olddiff > width / (4*len)){  // adaptive positioning of '*'
		    idx = len-1;
		} 
	    }
	    buf[idx]   = '*';
	    buf[idx+1] = '\0';
	}
    }
    if(buf[0] == '@'){
	status_msg(1, "UCTree: cannot handle directory names (%s) starting with `@'", buf);
	buf[0] = '?';
    }
    fl_set_object_label(button, buf);
    fl_set_object_lstyle(button, tree_font_style);
    fl_set_object_lsize(button, tree_font_size);

    if(symlink_flag && link_name){
	if(!link_button){
	    make_link_button();
	} else {
	    sprintf(buf, "-> %s", link_name);
	    int w = fl_get_string_width(tree_font_style, tree_font_size, buf, strlen(buf))+5;
	    link_button->w = w;
	    fl_set_object_lsize(link_button,tree_font_size);
	    fl_set_object_lstyle(link_button,tree_font_style);
	}
    }
    if(recursive_flag){
	if(child0) child0->retrunc();  // go right first
	if(next)   next->retrunc();    // go down next
    }
}

static void tree_adjust_sliders(void){
    {
	// use diffs between first and last button
	int y0, y1, ydiff, y_total, y_smax;
	double tmp_size;
	y0 = root->button->y;
	y1 = root->find_last()->button->y;

	ydiff   = abs(y0-y1) + button_h + 1;   // total height of all buttons.
	y_total = abs(TreeForm->h - top_margin - bottom_margin);   // abs for safety.
//	y_smax  = ydiff - y_total;
//	if(y_smax > 0){
//	    fl_set_slider_bounds(TreeFormVertSlider, 0.0, y_smax);   
//	    fl_set_slider_value(TreeFormVertSlider, y_smax - root_rel_y);
//	    tmp_size = (double) y_total / (double) ydiff;
	y_smax  = y_total - ydiff;
	if(y_smax < 0){
	    fl_set_slider_bounds(TreeFormVertSlider, y_smax, 0.0);   
	    fl_set_slider_value(TreeFormVertSlider, y_smax - root_rel_y);
	    tmp_size = fabs((double) y_total / (double) ydiff);
	    if(tmp_size > 0.5){
		tmp_size = 0.5;
	    }
	} else {
	    tmp_size = 1.0;
	}
	
	correct_slider_size(TreeFormVertSlider, tmp_size);
	fl_set_slider_size(TreeFormVertSlider, tmp_size);
    }
	
    {
	// use diffs between first and DEEPEST button
	// int maxlev = 0;
	int x0,xdiff,x_total, x_smax;
	double tmp_size;
	x0 = root->button->x;
//	int x1 = root->find_deepest(0, maxlev)->button->x;
//	xdiff   = fabs(x0-x1) + button_w + 1.0;   // total width of all buttons.
	
	int t_xmax = 0;
	root->find_geometrical_xmax(t_xmax);
	xdiff = abs(x0-t_xmax) + 1;
	    
	x_total = abs(TreeForm->w - left_margin - right_margin);   // abs for safety.
	x_smax  = xdiff - x_total;
	if(x_smax > 0){
	    fl_set_slider_bounds(TreeFormHorSlider, 0.0, x_smax);
	    fl_set_slider_value(TreeFormHorSlider,  abs(root_rel_x));
	    tmp_size = (double) x_total/ (double) xdiff;
	    if(tmp_size > 0.5){
		tmp_size = 0.5;
	    }
	} else {
	    tmp_size = 1.0;
	}
	fl_set_slider_size(TreeFormHorSlider, tmp_size);	
    }
}

int conditional_tree_dump (void){
    if(tree_has_changed_flag){
	int save_flag = FALSE;
	switch(save_tree_on_exit_state){
	case always_e: save_flag = TRUE; break;
	case ask_e:
	    switch(ask_multi_question(
		FALSE, 0, "Save Directory Tree to Database?",
		"Go!",        "GgyY",     NoSymbol,
		"Cancel",     "CcnN\033", XK_Escape,
		"Don't quit, oops", "Dd"
	    )) {
	    case 2:  return FALSE;
	    case 1:  save_flag = FALSE; break;
	    case 0:  save_flag = TRUE;  break;
	    default: break;
	    }
	    break;
	case dont_e:   
	default:     save_flag = FALSE; break;
	}
	if(save_flag){
	    tree_dump();
	}
    }
    return TRUE; // bye-bye
}

static char last_dump_path[MAXPATHLEN+1];

int tree_dump(void){
    strcpy(last_dump_path, "");
    if(!root) return FALSE;

    char fname[MAXPATHLEN+1];
    if(!tool_ldefault_dir){
	error_msg("UC: no default dir ?.");
	return FALSE;
    } else {
	sprintf(fname, "%s/uctreedb", tool_ldefault_dir);
    }

    if(access(fname, F_OK) == 0){
        char buffer[2*MAXPATHLEN+64];
        sprintf(buffer, "%s %s %s.old", cp_path, fname, fname);
        safe_system(buffer);  // make a copy of tree database file.
    }

    ofstream ofile(fname);
    if(!ofile){
	timed_error_notifier("UC:cannot open tree output file:", fname, notifier_secs);
	return FALSE;
    }
    
    ofile << "# Unix Cockpit version " << version_major << ".";
    ofile << version_minor << " directory tree data base.\n";
    ofile << "# Better don't edit this file by hand :-) \n\n";
    if(!root->dump_tree(ofile)){
	error_msg("/->dump_tree failed.");
    }
    ofile << "\n# end of UC tree. \n";
    ofile.close();
    tree_has_changed_flag = FALSE;
    return TRUE;
}

static int db_version_check(ifstream &ifile){
    char line[1024];
    if (!ifile.getline(line, 1024)) return FALSE;
    int iVersionMajor;
    if (sscanf(line, "# Unix Cockpit version %d", &iVersionMajor) <= 0) return FALSE;
    if (dbg_flag) cerr << iVersionMajor << endl;
    int val = atoi(version_major);
    return val <= iVersionMajor;
}

int tree_restore(int do_error_msg_flag = TRUE, int merge_flag = FALSE){

    if(dbg_flag) cerr << " start tree restore.\n";
    status_msg(1, "restoring tree database");
    
    char fname[MAXPATHLEN+1];
    if(!tool_ldefault_dir){
	error_msg("UC: no default dir ?.");
	return FALSE;
    } else {
	sprintf(fname, "%s/uctreedb", tool_ldefault_dir);
    }

    ifstream ifile(fname);
    if(!ifile){
	if(tool_gdefault_dir){
	    sprintf(fname, "%s/uctreedb", tool_gdefault_dir);
	    ifile.open(fname);
	}
    }
	
    if(!ifile){
	if(do_error_msg_flag){
	    timed_notifier("UC:cannot open tree input file (rescanning):", fname, notifier_secs);
	}
	tree_has_changed_flag = TRUE;
	return FALSE;
    }
    int bOldMode = FALSE;
    if (!db_version_check(ifile)) bOldMode = TRUE;

    if(root){
	if(!merge_flag){
	    fl_freeze_form(TreeForm);
	    while(root->child0){
		root->delete_sub_tree(root->child0);
	    }
	    fl_unfreeze_form(TreeForm);
	    current_node = NULL;
	}
    } else {
	root = new tree_node_c("/");
	if(!root){
	    return FALSE;
	}
    }

    const int bufsize = 2*MAXPATHLEN;
    char bbuf[bufsize];
    int continue_flag = TRUE;
    fl_freeze_form(TreeForm);
    tree_node_c *tmp_node;
    while(ifile.getline(bbuf, bufsize) && continue_flag){
	if(bbuf[0] == '#' || bbuf[0] == '\0' || bbuf[0] == ' ' || bbuf[0] != '/'){
	    continue;
	    // it is a comment, or empty line, or starts with a blank
	    // (redundant, right now), or doesn't start with a slash. 
	}
	char *dir_tok       = strtok(bbuf, " ");
	char *link_tok      = strtok(NULL, " ");
	char *perms_tok     = strtok(NULL, " ");
	char *hidden_tok    = bOldMode ? (char *)NULL : strtok(NULL, " ");
	char *link_name_tok = strtok(NULL, " ");
	if(!dir_tok){
	    error_msg("tree_restore: unrecognized line in tree database.");
	    continue;
	} else {
	    tmp_node = tree_construct_path(dir_tok, FALSE);
	    tmp_node->hidden_flag   = FALSE;
	    tmp_node->no_perms_flag = FALSE;
	    tmp_node->symlink_flag  = FALSE;
	    if(!tmp_node){
		continue_flag = FALSE;
		timed_error_notifier("UC tree: cannot construct path:", dir_tok, notifier_secs);
	    } else {
		if(!link_tok){
		    if(dbg_flag) cerr << "no link flag found, assuming no link.\n";
		} else {
		    tmp_node->symlink_flag = (link_tok[0] == 'l');
		    if(!perms_tok){
			if(dbg_flag) cerr << "no perms flag found, assuming perms ok.\n";
		    } else {
			tmp_node->no_perms_flag = (perms_tok[0] == 'p');
		    }
		    if (!bOldMode) {
			if(!hidden_tok){
			    if(dbg_flag) cerr << "no hidden flag found, assuming visible.\n";
			} else {
			    tmp_node->hidden_flag = (hidden_tok[0] == 'h');
			}
		    }
		    if(tmp_node->symlink_flag){
			if(!link_name_tok){
			    if(dbg_flag) cerr << "link name not found, reading.\n";
			    tmp_node->read_link();
			} else {
			    tmp_node->link_name = strdup(link_name_tok);
			}
		    }
		}
	    }
	}
    }
    tree_reshape();
    tree_form_sync();
    fl_unfreeze_form(TreeForm);
    ifile.close();
    if(dbg_flag) cerr << " end tree restore.\n";
    status_msg(1, "database loaded.");
    return TRUE;
}

tree_node_c* tree_node_c::find_geometrical_xmax(int &xmax){
    if(this == root){
	// ground the search. 
	xmax = root->button->x;
    }
    if(!child0){
	int my_xmax;
	if(symlink_flag && link_button){
	    my_xmax = link_button->x + link_button->w;
	} else {
	    my_xmax = button->x + button->w;
	}
	if(my_xmax > xmax){
	    xmax = my_xmax;
	    return this;
	} else {
	    return NULL;
	}
    }   

    tree_node_c	*tmp_node = child0, *result = NULL, *tmp_node_2 = NULL;
    while(tmp_node){
	tmp_node_2 = tmp_node->find_geometrical_xmax(xmax);
	if(tmp_node_2){
	    result = tmp_node_2;
	}
	tmp_node = tmp_node->next;
    }
    return result;
}

static void tree_refit(void){
    // shift-fit the tree so that no excess bottom or right margins persist.
    if (!root) return;
    if(root_rel_y < 0){
	// problem: resizing of window. Bottom margin open. Adjust by diff between
	// last button and bottom.
//	int ydelta = root->find_last()->button->y - bottom_margin;
	int ydelta = TreeForm->h - bottom_margin - button_h - root->find_last()->button->y;
	if(ydelta > 0){
//	    root_rel_y -= ydelta;
	    root_rel_y += ydelta;
//	    if(root_rel_y < 0){
	    if(root_rel_y > 0){
		root_rel_y = 0;
	    }
	}
    }
    if(root_rel_x < 0){
	// problem: resizing of window. right margin open. Adjust by diff between
	// deepest button and right.
	// int maxlev=0;
//	int xdelta = TreeForm->w - root->find_deepest(0,maxlev)->button->x - button_w - right_margin;
	int xmax = root->button->x;
	root->find_geometrical_xmax(xmax);
	int xdelta = TreeForm->w - xmax - right_margin;		
	if(xdelta > 0.0){
	    root_rel_x += xdelta;
	    if(root_rel_x > 0){
		root_rel_x = 0;
	    }
	}
    }
}

static void draw_arrow(int x, int y, int x1, int y1, int col, int which){
    int arrow_size = bound_value((int) floor(button_h / 3.0), 1, button_x_gap/2-1);

    switch(which){
    case 2: // down
	fl_diagline(x, y1-1, arrow_size,    -arrow_size+2, col);
	fl_diagline(x, y1-1, -arrow_size+2, -arrow_size+2, col);
	break;
    case 8: // up
	fl_diagline(x, y, arrow_size,    arrow_size, col);
	fl_diagline(x, y, -arrow_size+2, arrow_size, col);
	break;
    case 4:
	fl_diagline(x, y, arrow_size, arrow_size,    col);
	fl_diagline(x, y, arrow_size, -arrow_size+2, col);
	break;
    case 6:
	fl_diagline(x1, y, -arrow_size+2, arrow_size,    col);
	fl_diagline(x1, y, -arrow_size+2, -arrow_size+2, col);
    default:
	break;
    }
}

void tree_node_c::clip_line(int x,int y,int w,int h,int col,int up_arrow, int down_arrow, int left_arrow, int right_arrow){
    int x1 = x+w, y1 = y+h;
    // don't draw anything if the line is completely invisible !
    // use _half_ the button margins to determine visibility for y. 
    if(x1 < left_margin || 
       x > TreeForm->w - right_margin || 
       y1 < top_margin + button_h/2.0 ||         
       y > TreeForm->h - bottom_margin - button_h/2.0
    ){
	return;
    }
    // now, clip. remember to adjust w with x and h with y (done via x,x1; y,y1)
    
    if(y < top_margin){
	y = top_margin;
	up_arrow = TRUE;
    }
    if(y1> TreeForm->h - bottom_margin){
	y1 = TreeForm->h - bottom_margin;
	down_arrow = TRUE;	
    }
    if(x < left_margin){
	x = left_margin;
	left_arrow = TRUE;
    }
    if(x1> TreeForm->w - right_margin){
	x1 = TreeForm->w - right_margin;
	right_arrow = TRUE;
    }

    if(up_arrow)    draw_arrow(x,y,x1,y1,col,8);
    if(down_arrow)  draw_arrow(x,y,x1,y1,col,2);
    if(left_arrow)  draw_arrow(x,y,x1,y1,col,4);
    if(right_arrow) draw_arrow(x,y,x1,y1,col,6);

    fl_diagline(x,y,x1-x,y1-y,col);
}

void tree_node_c::draw_lines(void){
    int x,y,w,h;
    int ua = FALSE,
	da = FALSE,
	la = FALSE,
	ra = FALSE;
    
    if(!parent){
	// root
	// cerr << "no parent. \n";
	if(dbg_flag)cerr << "root: draw_lines.\n";
    } else {
	y = button->y + (int) floor(button->h/2.0);
	if(!prev){
	    // connect to left (full step):
	    x = parent->button->x + parent->button->w;
	    w = button_x_gap;
	    h = 1;
	    if(!parent->button->visible)la=TRUE;
	    if(!button->visible)ra=TRUE;
	    clip_line(x,y,w,h,tree_lines_fg_colour,FALSE,FALSE,la,ra);
	} else {
	    // connect left to downline (halfstep).
	    x = parent->button->x + parent->button->w + button_x_gap/2;
	    w = button->x - x;
	    h = 1;
	    if(!button->visible)ra=TRUE;
	    clip_line(x,y,w,h,tree_lines_fg_colour,FALSE,FALSE,la,ra);
	    if(!next){
		// downline (last in subs, and more than 1 subs).
		x = parent->button->x + parent->button->w + button_x_gap/2;
//		h = parent->button->y + (int) floor(parent->button->h/2) - y;
		h = y - parent->button->y - (int) floor(parent->button->h/2);
		w = 1;
		if(!parent->button->visible){
//		    if(parent->button->y + parent->button->h > TreeForm->h - top_margin){
		    if(parent->button->y <= top_margin){
			ua=TRUE;
		    }
		}
		if(!button->visible){
//		    if(button->y <= bottom_margin){
		    if(button->y + button->h > TreeForm->h - bottom_margin){
			da=TRUE;
		    }
		}
//		clip_line(x,y,w,h,tree_lines_fg_colour,ua,da,FALSE,FALSE);
		clip_line(x,y-h,w,h,tree_lines_fg_colour,ua,da,FALSE,FALSE);
	    }
	}
    }
    if(symlink_flag && link_button){
	if(button->visible && !link_button->visible){
	    // draw it like ->
	    x = button->x+button->w + button_x_gap/4;
	    y = button->y + (int) floor(button->h/2);
	    w = button_x_gap/2;
	    h = 1;
	    clip_line(x,y,w,h,tree_lines_fg_colour,FALSE,FALSE,FALSE,ra);

	    x = button->x+button->w;
	    w = button_x_gap;
	    draw_arrow(x,y,x+w,y+h,tree_lines_fg_colour,6);
	}
    }
    if(child0) child0->draw_lines();  // go right first
    if(next)   next->draw_lines();    // go down next
}

void tree_node_c::set_current(int flag){
    if(!button){
	cerr << "No button in tree_node_c (flag = " << flag << ")." << endl;;
	return;
    }
    int new_col  = tool_bw_flag ? FL_WHITE : tree_normal_bg_colour,
	new_lcol = tool_bw_flag ? FL_BLACK : tree_normal_fg_colour
    ;
    if(flag){
	new_col  = tool_bw_flag ? FL_BLACK : tree_current_bg_colour;
	new_lcol = tool_bw_flag ? FL_WHITE : tree_current_fg_colour;
    }
    fl_set_object_color(button, new_col, FL_COL1);
    fl_set_object_lcol (button, new_lcol);
}

static void tree_set_wintitle_backend(const char *title){
    if(!title) return;
    if (last_wintitle && (strcmp(last_wintitle, title) == 0)) return;
    if (last_wintitle) {
	fl_free(last_wintitle);
    }
    last_wintitle = strdup(title);
    XStoreName(fl_display, TreeForm->window, title);
    toolbox_attach_title();
}

void tree_set_complete_wintitle(void){
    static int reset_cursor_flag = FALSE;
    if(TreeForm->visible && target_node && current_node){
        if(!cnode_is_stable()) {
	    tree_set_wintitle_backend("UCTree (no current node)");
	    return;
	}
	char
	    buffer			[2*MAXPATHLEN],
	    cwd_buf			[MAXPATHLEN],
	    target_cwd_buf	[MAXPATHLEN],
	    nl_buf			[64]
	;
	const char
	    *header_text
	;

	strcpy(cwd_buf, current_node->get_path());
	if (dbg_flag) cerr << cwd_buf << endl;
	header_text = tree_target_mode_flag ? "SELECT NEW TARGET" : "UCTree";
	strcpy(target_cwd_buf, target_node->get_path());

	if (!cwd_buf || !target_cwd_buf || !current_node->name) {
	    program_unstable("NULL from get_path!");
	    return;
	}

	if(shove_flag){
	    tree_set_wintitle_backend("UCTree: 2D-SpeedScroll");
	    return;
	}
	if (tree_scanning_flag) {
	    fl_set_global_cursor(XC_watch);
	    reset_cursor_flag = TRUE;
	    sprintf(buffer, "%s: Scanning %s ...", header_text, cwd_buf);
	    tree_set_wintitle_backend(buffer);
	    return;
	}

	if (xdnd_flag || (deact_node && dnd_flag)) {
	    // tracking.
	    if (xdnd_flag) {
		sprintf(buffer, "D&D %s from #%d",
		    tool_dnd_op_str,
		    source_browser->get_ID()
	        );
	    } else {
		sprintf(buffer, "UCTree: drag & drop %s directory `%s'",
	            dnd_copy_flag ? "_copy_" : "move",
		    current_node->name 
	        );
	    }
	    if(current_node == target_node || dnd_notarget_flag){
		strcat(buffer, " ...");
	    } else {
		strcat(buffer, " to `");
		strcat(buffer, target_cwd_buf);
		strcat(buffer, "'");
	    }
	    tree_set_wintitle_backend(buffer);
	    return;
	}
	sprintf(nl_buf, "(%d node", tree_node_c::next_ID);
	if(tree_node_c::next_ID>1){
	    strcat(nl_buf, "s");
	}
	int max_lev = 0;
	root->find_deepest(0,max_lev);
	int my_lev = 0;
	tree_node_c *tmp_node = current_node;
	while(tmp_node){
	    tmp_node = tmp_node->parent;
	    my_lev++;
	}
	char lev_buf[64];
        lev_buf[0] = '\0';
	sprintf(lev_buf, ", level %d of %d)", my_lev, max_lev+1);

	strcat(nl_buf, lev_buf);

	if (reset_cursor_flag) {
	    reset_cursor_flag = FALSE;
	    fl_reset_global_cursor();
	}
	sprintf(buffer, "%s%s%s %s", header_text, tool_host_string, cwd_buf, nl_buf);
	tree_set_wintitle_backend(buffer);
    }
}

static void tree_change_source_to_node(tree_node_c *node){
    if(!node) return;
    const char *cwd_buf = node->get_path();
    if(cwd_buf){
	int update_tree_flag = node->symlink_flag;
	if(tool_change_dir(cwd_buf, FALSE, TRUE, update_tree_flag, TRUE, FALSE)){
	    // no current, brwsers yes, tree no, home no.
	    tree_form_sync();
	    /*
	     * code fails if change_dir fails only partially. 
	    tree_set_source(node);
	    if(source_browser == target_browser){
		tree_set_target(node);
	    }
	    */
	}
    } else {
	cerr << "tree_change_source: no path?\n";
    }
}

static void tree_change_source_to_current(void){
    if(!cnode_is_stable()) return;
    tree_change_source_to_node(current_node);
}

static void tree_change_target_to_node(tree_node_c *node){
    if(!node) return;
    const char *cwd_buf = node->get_path();
    if(cwd_buf){
	int update_tree_flag = node->symlink_flag;
	if(tool_change_target_dir(cwd_buf, TRUE, update_tree_flag)){
	    // no current, brwsers yes, tree no, home no.
	    tree_set_target(node);
	    if(source_browser == target_browser){
		tree_set_source(node);
	    }
	}
    } else {
	cerr << "tree_change_target: no path?\n";
    }
}

static void tree_change_target_to_current(void){
    if(!cnode_is_stable()) return;
    tree_change_target_to_node(current_node);
}

static void
tree_set_current(tree_node_c *node, int always_center_flag, int tool_sync_flag){
     // no cnod check in this function, it can deal with a NULL cnode.
    if(!node){
	cerr << "Null new current node !\n";
	return;
    }
    if(!node->button){
	if(dbg_flag)cerr << "Null button in new current node !\n";
	return;
    }
    if(dbg_flag) cerr << "attempt to set current: " << node->get_path() << endl;
    if(node == current_node){
	if(dbg_flag) cerr << "current stays the same: " << node->get_path() << endl;
	if(!node->button->visible){
	    tree_center(FALSE);
	}
	return;
    } else {
	if(!node->restat(TRUE, FALSE)){
	    tree_remove_invalid_node(node);
	    return;
	}
	if(current_node){
	    current_node->set_current(FALSE);
    	if(dbg_flag) cerr << "    old current: " << current_node->get_path() << endl;
	}
	current_node = node;
	if(dbg_flag) cerr << "    new current: " << current_node->get_path() << endl;
	current_node->set_current(TRUE);
	if(tree_auto_sync_dir_flag && tool_sync_flag){
	    tree_change_source_to_current();
	}
	tree_set_complete_wintitle();
	tree_center(always_center_flag);
    }
}

tree_node_c* tree_node_c::cut_yourself(void){
    // take yourself out of the tree, leave the tree intact.
    if(this == root || !parent){
	cerr << "UCT: cut_yourself failed.\n";
	return NULL;
    }
    if(this == parent->child0){
	if(parent->child0->next){
	    parent->child0 = parent->child0->next;
	    parent->child0->prev = NULL;
	} else {
	    parent->child0 = NULL;
	}
    } else {
	if(!prev) return NULL;  // consistency errcheck.
	
	prev->next = next;
	if(next){
	    next->prev = prev;
	}
    }
    next = prev = parent = NULL;
    return this;
}

int tree_node_c::dump_tree(ofstream &ofile){
    if(!ofile){
	cerr << "tree_node_c::dump_tree: Cannot dump to invalid out file stream.\n";
	program_unstable("tree_node_c::dump_tree: Cannot dump to invalid out file stream.");
	return FALSE;
    }
    const char *path = get_path();
    //char pbuf[sizeof(path_buf)];
    //char* last_slash;

    if(path){
	ofile << path;
	/*
	int short_flag = FALSE;
	strcpy(pbuf, path);
	last_slash = strrchr(pbuf, '/');
	last_slash[0] = '\0';  
	if(this != root){
	    if(strcmp(pbuf, last_dump_path) == 0){
		if(last_slash[1] != '\0'){
		    cerr << "is short: " << last_slash+1 << endl;
		    ofile << last_slash+1;
		    short_flag = TRUE;
		}
	    }
	}
	
	if(!short_flag){
	    ofile << path;
	    strcpy(last_dump_path, pbuf);
	}
	*/
	if(symlink_flag){
	    ofile << " l";
	} else {
	    ofile << " n";
	}
	if(no_perms_flag){
	    ofile << " p";
	} else {
	    ofile << " n";
	}
	if(hidden_flag){
	    ofile << " h";
	} else {
	    ofile << " v";
	}
	if(symlink_flag){
	    if(!link_name){
		read_link();
	    }
	    if(link_name){
		ofile << " " << link_name;
	    }
	}
	ofile << endl;
	tree_node_c *tmp_node = child0;
	while(tmp_node){
	    if(!tmp_node->dump_tree(ofile)){
		return FALSE;
	    }
	    tmp_node = tmp_node->next;
	}
    }
    return TRUE;
}

static void LinkButtonCB(FL_OBJECT *, long val){
    if(val>0){
	tool_got_event_flag = TRUE;
	tree_node_c *tmp_node = (tree_node_c *) val;
	const char *path = tmp_node->get_path();
	if(dbg_flag) cerr << "LinkbuttonCB:->get_path: " << path << endl;
	char resolved_name_buf[MAXPATHLEN+1];
	char *r_path = real_path(path, resolved_name_buf);
	if(!r_path){
	    status_msg(0, "couldn't resolve: %s", resolved_name_buf);
	    return;
	} else {
	    if(dbg_flag) cerr << "r_path: " << r_path << endl;
	    if(dbg_flag) cerr << "resolved: " << resolved_name_buf << endl;
	    tmp_node = tree_construct_path(resolved_name_buf);
	    tree_set_current(tmp_node);
	}
    }
}

static void DirButtonCB(FL_OBJECT *button, long val){
    // slightly incomplete, but harmlessly so. (dbl click timing only)
    sginap(15);  // 10 was a little fast.
    if(!root || val <= 0){
	cerr << "UCT: no root or negative val.\n";
	return;
    }
    tree_node_c *node = root->find_node((tree_node_c *) val);
    if(!node){
	M_err("UCT/DirButtonCB","val %d maps to invalid node", val);
	node = (tree_node_c *) val;
	if(node){
	    M_err("attempt match", "button %d, name %s", (long) node->button, node->name);
	}
	return;
    }
    //M_err("DirButtonCB", "button %d, name %s", (long) node->button, node->name);
    tool_got_event_flag = TRUE;
    tree_default_actions();

    if(fl_keymask & Button3Mask){
	if(dbg_flag) cerr << "Button 3.\n";
	if(!current_node || button != current_node->button){
	    tree_set_current(node);
	}
	if(tool_get_ctrl()){
	    tree_collapse_subtree_all_cb(0);
	} else {
	    tree_collapse();
	}
    } else if(fl_keymask & Button2Mask){
	if(dbg_flag) cerr << "Button 2.\n";
	if(!current_node || button != current_node->button){
	    tree_set_current(node);
	}
	if(tool_get_ctrl()){
	    tree_expand_subtree_all_cb(0);
	} else {
	    tree_expand(TRUE);
	}
    } else if(fl_keymask & Button1Mask){ 
	if(dbg_flag)cerr << "Button 1.\n";
	if(!current_node || button == current_node->button){
	    if(tree_target_mode_flag){
		tree_set_target_cb(NULL, 0);		
	    } else {
		if(!tree_auto_sync_dir_flag){        // only if not already done before.
		    tree_change_source_to_current();
		}
	    }
	} else {
	    tree_set_current(node);
	}
    } else {
	// weird mask.
	// M_err("DirButtonCB", "No buttons in fl_keymask");
    }
}

void TreeFormHorLeftButtonCB(FL_OBJECT *, long){
    tree_default_actions();
    root_rel_x += button_w + button_x_gap;
    new_root_rel();
}

void TreeFormHorRightButtonCB(FL_OBJECT *, long){
    tree_default_actions();
    root_rel_x -= button_w + button_x_gap;
    new_root_rel();
}

void TreeFormVertDownButtonCB(FL_OBJECT *, long){
    tree_default_actions();
//    root_rel_y += button_h + button_y_gap;
    root_rel_y -= button_h + button_y_gap;
    new_root_rel();
}

void TreeFormVertUpButtonCB(FL_OBJECT *, long){
    tree_default_actions();
//    root_rel_y -= button_h + button_y_gap;
    root_rel_y += button_h + button_y_gap;
    new_root_rel();
}

void TreeFormHorSliderCB(FL_OBJECT *, long){
    tree_default_actions();
    int new_x =  (int) -fl_get_slider_value(TreeFormHorSlider);
    if(new_x == root_rel_x) return;
    root_rel_x = new_x;
    new_root_rel();
}

void TreeFormVertSliderCB(FL_OBJECT *, long){
    tree_default_actions();
    double min,max;
    fl_get_slider_bounds(TreeFormVertSlider, &min, &max);
    int new_y = (int) (min - fl_get_slider_value(TreeFormVertSlider));
    if(new_y == root_rel_y) return;
    root_rel_y = new_y;
    new_root_rel();
}

void new_root_rel(void){
    static int old_x = 0, old_y = 0, first = TRUE;
    // clip.
    if(root_rel_x > 0) root_rel_x = 0;
    if(root_rel_y > 0) root_rel_y = 0;

    // use diffs between first and last button
    int y0,y1,ydiff,y_total, y_smax;
    y0 = root->button->y;
    y1 = root->find_last()->button->y;
    ydiff   = abs(y0-y1) + button_h + 1;   // total height of all buttons.
    y_total = abs(TreeForm->h - top_margin - bottom_margin);   // abs for safety.
    y_smax  = y_total - ydiff;
    if(y_smax < 0){
	if(root_rel_y < y_smax){
	    root_rel_y = y_smax;
	}
    } else {
	root_rel_y = 0;
    }
    
    // use diffs between first and DEEPEST button
    // int maxlev = 0;
    int x0,xdiff,x_total, x_smax;
    x0 = root->button->x;
    
    int t_xmax = 0;
    root->find_geometrical_xmax(t_xmax);
    xdiff = abs(x0-t_xmax) + 1;

    x_total = abs(TreeForm->w - left_margin - right_margin);   // abs for safety.
    x_smax  = xdiff - x_total;
    if(x_smax > 0){
	if(-root_rel_x > x_smax){
	    root_rel_x = -x_smax;
	}
    } else {
	root_rel_x = 0;
    }

    if (old_x == root_rel_x && old_y == root_rel_y && !first) return;
    first = FALSE;
    old_x = root_rel_x;
    old_y = root_rel_y;
    
    fl_freeze_form(TreeForm);

    // sync sliders, if necessary:
    int tmp_int;
    double min, max;
    fl_get_slider_bounds(TreeFormVertSlider, &min, &max);
    tmp_int = (int) (min - fl_get_slider_value(TreeFormVertSlider));
    if(tmp_int != root_rel_y){
	fl_set_slider_value(TreeFormVertSlider, min - root_rel_y);
    }
    tmp_int = (int) -fl_get_slider_value(TreeFormHorSlider);
    if(tmp_int != root_rel_x){
	fl_set_slider_value(TreeFormHorSlider, -root_rel_x);
    }
    
    root->realign();
    fl_redraw_form(TreeForm);
    fl_unfreeze_form(TreeForm);
}

static int cnode_is_stable(void){
    if(!current_node){
	M_warn("cnode_is_stable", "No current node in tree! Resetting.");
	current_node = source_node ? source_node : (target_node ? target_node : root);
	return FALSE;
    } else if(!current_node->button){
	M_warn("cnode_is_stable", "No current node->button in tree! Resetting.");
	current_node = source_node ? source_node : (target_node ? target_node : root);
	return FALSE;
    } else {      
	return TRUE;
    }
}

static tree_node_c *expand_last_node = NULL;

static void tree_collapse(void){
    tree_has_changed_flag = TRUE;
    if(!cnode_is_stable()) return;
    if(!root) return;
    if(!TreeForm->visible) return;
    tree_default_actions();

    expand_last_node = NULL;  // side effect.
    int maxlevel = 0;
    tree_node_c *tmp_node = current_node->find_deepest(0, maxlevel);
    if(!tmp_node){
	cerr << "uc::tree_c::null node found in collapse.\n";
	return;
    }
    if(tmp_node == current_node){
	if(tmp_node == root){
	    timed_error_notifier("Cannot remove the root tree node.", notifier_secs);
	    return;
	} else {
	    fl_freeze_form(TreeForm);
	    if(tmp_node->next){
		tree_set_current(tmp_node->next);        // down ?
	    } else if(tmp_node->prev){
		tree_set_current(tmp_node->prev);        // up ?
	    } else {
		tree_set_current(current_node->parent);  // parent ! (always do update)
	    }
	    tmp_node->parent->delete_sub_tree(tmp_node);
	    tree_reshape();
	    fl_unfreeze_form(TreeForm);
	    return;
	}
    } else {
	// remove the whole last level (THUNDER!) :-)
	fl_freeze_form(TreeForm);
	current_node->delete_sub_trees_below(maxlevel-1);
	tree_reshape();
	fl_unfreeze_form(TreeForm);
	return;	
    }
}

void TreeCollapseCB	(FL_OBJECT *, long){
    if(tool_get_ctrl()){
	tree_collapse_subtree_all_cb(0);
    } else {
	tree_collapse();
    }
}

void TreeExpandCB(FL_OBJECT *, long){
    if(tool_get_ctrl()){
	tree_expand_subtree_all_cb(0);
    } else {
	tree_expand(TRUE);
    }
}
static void tree_expand(int allow_iteration_flag){
    static int level_count = 0;
    
    if(!cnode_is_stable()) return;
    if(!root) return;
    if(!TreeForm->visible) return;
    tree_default_actions();

    if(expand_last_node == current_node && allow_iteration_flag){
	level_count++;
    } else {
	level_count = 0;
    }
    expand_last_node = current_node;
    fl_freeze_form(TreeForm);
    if(tool_get_ctrl()){
	tree_restat(current_node, TRUE);
	tree_scan(current_node, SCAN_ALL_LEVELS, -1, TRUE);
    } else {
	// allows for iterative deepening.
	// only rescan at level_count level, all others just traverse
	if(level_count){
	    indicator_simple_bar("subdirectory expand");
	    indicator_check("0");
	}
	if(level_count == 0){
	    // only once. 
	    tree_restat(current_node, TRUE);
	}
	// tree_restat_below(current_node, level_count, TRUE);
	tree_scan_below(current_node, level_count, -1, TRUE);
	if(level_count){
	    indicator_end();
	}
    }
    
    tree_reshape();
    fl_unfreeze_form(TreeForm);    
}

void TreeCenterCB(FL_OBJECT *, long){
    if(tool_get_meta()){
	tree_set_target_cb(NULL, 0);
    } else {
	tree_center(TRUE);
    }
}

void tree_center(int always_center_flag){
    if(!cnode_is_stable()) return;
    if(!always_center_flag && current_node->button->visible){
	return;
    }
    if(current_node == root){
	root_rel_x = 0;
	root_rel_y = 0;
    } else {
//	root_rel_y = root->button->y - current_node->button->y - ((TreeForm->h - bottom_margin - top_margin)/2);
	root_rel_y = root->button->y - current_node->button->y + ((TreeForm->h - bottom_margin - top_margin)/2);
        root_rel_x = root->button->x - current_node->button->x - current_node->button->w + ((TreeForm->w - left_margin - right_margin)/2);
    }
    new_root_rel();
}

void tree_end(void){
    tree_default_actions();
    tree_set_current(root->find_last());
}

static void tree_home(void){
    tree_default_actions();
    tree_set_current(root);
}

void TreeHomeCB(FL_OBJECT *, long){
    if(tool_get_meta()){
		tree_cd_home_event_cb(NULL, 0);
    } else {
		tree_home();
    }
}

void tree_left_arrow(void){
    if(!cnode_is_stable()) return;
    tree_default_actions();

    tree_node_c *tmp_node = current_node->left_arrow();
    if(tmp_node){
	tree_set_current(tmp_node);
    } else {
	//cerr << "null in tree_left_arrow. No more parents.\n";
    }
}

void tree_right_arrow(void){
    if(!cnode_is_stable()) return;
    tree_default_actions();
    tree_node_c *tmp_node = current_node->right_arrow();
    if(tmp_node){
	tree_set_current(tmp_node);
    } else {
	//cerr << "null in tree_right_arrow. bottom in tree.\n";
    }
}

void tree_down_arrow(int override){
    if(!cnode_is_stable()) return;
    if(!override && !down_arrow_key_verify()) return;    
    
    tree_default_actions();
    tree_node_c *tmp_node = current_node->down_arrow();
    if(tmp_node){
	tree_set_current(tmp_node);
    } else {
	if(dbg_flag) cerr << "null in tree_down_arrow. bottom in lev. change right?.\n";
	//tree_right_arrow();
    }
}

void tree_up_arrow(int override){
    if(!cnode_is_stable()) return;
    if(!override && !up_arrow_key_verify()) return;    
    tree_default_actions();
    tree_node_c *tmp_node = current_node->up_arrow();
    if(tmp_node){
	tree_set_current(tmp_node);
    } else {
	if(dbg_flag) cerr << "null in tree_up_arrow. top in lev. change left ?.\n";
	//tree_left_arrow();
    }
}

void tree_page_down(void){
    if(!cnode_is_stable()) return;
    tree_default_actions();
    if(tool_get_meta()){
	tree_node_c *tmp_node = current_node->parent, *result;
	int level_count = 1;
	while(tmp_node->parent){
	    tmp_node = tmp_node->parent;
	    level_count++;
	}
	result = root->attempt_change_to_last(level_count);
	if(result == current_node){
	    // don't stay static. go right.
	    result = current_node->child0;
	}
	if(result){
	    tree_set_current(result);
	}	
	return;
    }
    int new_y = current_node->button->y + (TreeForm->h - bottom_margin - top_margin);

    tree_node_c *tmp_node = root->find_last();
    if(!tmp_node) return;
    if(tmp_node->button->y < new_y){
	new_y = tmp_node->button->y;
    }

    tmp_node = root->find_y(NULL, current_node->button->x, new_y);
    if(!tmp_node){
	return;
    }
    if(tmp_node == current_node){
	tmp_node = current_node->down_arrow();
    }
    if(tmp_node){
	tree_set_current(tmp_node);
    }
	
    //root_rel_y += TreeForm->h - bottom_margin - top_margin;
    //new_root_rel();
}

void tree_page_up(void){
    if(!cnode_is_stable()) return;
    if(current_node == root) return;
    tree_default_actions();
    if(tool_get_meta()){
	tree_node_c *tmp_node = current_node->parent, *result;
	int level_count = 1;
	while(tmp_node->parent){
	    tmp_node = tmp_node->parent;
	    level_count++;
	}
	result = root->attempt_change_to_first(level_count);
	if(result == current_node){
	    // don't stay static.
	    result = current_node->parent;
	}
	if(result){
	    tree_set_current(result);
	}	
	return;
    }
    int new_y = current_node->button->y - (TreeForm->h - bottom_margin - top_margin);

    if(root->button->y > new_y){
	new_y = root->button->y;
    }

    tree_node_c *tmp_node = root->find_y(NULL, current_node->button->x, new_y);

    if(tmp_node == current_node){
	tmp_node = current_node->up_arrow();
    }
    if(!tmp_node) return;
    tree_set_current(tmp_node);
}

static void tree_default_actions(void){
    tree_search_clear_mode();
}

void TreeEnterKeyCB(FL_OBJECT*, long){
    if(tree_target_mode_flag){
	tree_set_target_cb(NULL, 0);		
    } else {
	if(!tree_auto_sync_dir_flag){
	    tree_change_source_to_current();
	}
    }
}

static int tree_key_event(XKeyEvent *, KeySym keysym, char key){
    if(keysym == XK_KP_Enter || keysym == XK_Return){
	if((xshell_visible_flag && xshell_needs_enter_flag) || 
	    tool_get_ctrl()){
	    return FALSE;
	} else {
	    TreeEnterKeyCB(NULL, 0);
	    return TRUE;
	}
    }
    if(tool_get_ctrl()){
	int retval = TRUE;
	tool_disable_ctrl();
	switch(keysym){
	case XK_a: case XK_A: tree_home        (); break;
	case XK_b: case XK_B: tree_left_arrow  (); break;
	case XK_e: case XK_E: tree_end         (); break;
	case XK_f: case XK_F: tree_right_arrow (); break;
	case XK_l: case XK_L:
	    tree_default_actions   ();
	    tree_center            (TRUE);
	    break;
	case XK_n: case XK_N: tree_down_arrow  ();           break;
	case XK_p: case XK_P: tree_up_arrow    ();           break;
	case XK_r: case XK_R: tree_restat      (root);       break;
	case XK_t: case XK_T: tree_set_target_cb(NULL, 0);   break;
	case XK_v: case XK_V: tree_page_down   ();           break;
	default:
	    retval = FALSE;break;
	}
	tool_enable_ctrl();
	return retval;
    }
    if(tool_get_alt() || tool_get_esc()){
//	cerr << "alt detected, working on " << (int) key << endl;
	switch(key){
	case '-': TreeCollapseCB(NULL, 0);             return TRUE;
	case '+': TreeExpandCB  (NULL, 0);             return TRUE;
	case 'a':
	case 'A': TreeFormActionsButtonCB(NULL, TRUE); return TRUE;
	case 'd':
	case 'D': dir_menu_cb(); return TRUE; 
	case 'e':
	case 'E': TreeFormEditButtonCB(NULL, TRUE);    return TRUE;
	case 'f':
	case 'F': TreeFormFileButtonCB(NULL, TRUE);    return TRUE;
	case 'o':
	case 'O': TreeFormOptionsButtonCB(NULL, TRUE); return TRUE;
	case 'q':
	case 'Q': tree_close_form();                   return TRUE;
	case 'w':
	case 'W': windows_menu_cb();                   return TRUE; 
	default: break;
	}
	    
	switch(keysym){
	case XK_less:         tree_home      (); return TRUE;
	case XK_greater:      tree_end       (); return TRUE;
	case XK_v: case XK_V: tree_page_up   (); return TRUE;
	default:                                 return FALSE;
	}
    }
    return FALSE;
}

char tree_new_target_dir[MAXPATHLEN+1];

void tree_set_target_cb(FL_OBJECT *, long){
    if(!cnode_is_stable()) return;
    if(dbg_flag) cerr << "tree_set_target_cb !\n";
    if(tree_target_mode_flag){
	if(target_browser != source_browser){
	    tree_change_target_to_current();
	}
	const char *cwd_buf = current_node->get_path();
	if(cwd_buf){
	    strcpy(tree_new_target_dir, cwd_buf);
	    action_work_form_target_sync(tree_new_target_dir);
	} else {
	    cerr << "tree_set_target_cb: no path?\n";
	}
	tree_close_form();
	return;
    } else {
	if(target_browser != source_browser){
	    tree_change_target_to_current();
	} else {
	    if(!tree_auto_sync_dir_flag){
		tree_change_source_to_current();
	    }
	}
    }
}

static void tree_cd_home_event_cb(FL_OBJECT *, long){
    char *home_buf = getenv("HOME");
    if(home_buf){
	tree_node_c *tmp_node = tree_construct_path(home_buf);
	if(tmp_node){
	    tree_set_current(tmp_node, FALSE, FALSE);
	}
    }
    if(!tree_target_mode_flag){
	cd_home_event_cb(NULL, 0);
    }
}

static void tree_remove_invalid_node(tree_node_c *node){
    if(!node)return;
    if(dbg_flag)cerr << "removing invalid node.\n";
    status_msg(0, "tree: invalid node, removing (%s).", node->get_path());
    fl_freeze_form(TreeForm);
    if(node->parent){
	node->parent->delete_sub_tree(node);
    } else {
	tree_destroy_node(node);
    }
    tree_reshape();
    fl_unfreeze_form(TreeForm);    
}

static const char* tree_dnd_tracking(FL_Coord x, FL_Coord y){
    if(dbg_flag)cerr<< "tree_dnd_tracking\n";
    if(!root || tree_target_mode_flag) return NULL;
    if(!xdnd_flag && (!dnd_flag || !deact_node)) return NULL;
    
    FL_OBJECT *obj;
    if(deact_node){
	fl_activate_object(deact_node->button);
    }
    obj = fl_find_last(TreeForm, FL_FIND_MOUSE, x, y);
    if(deact_node){
	fl_deactivate_object(deact_node->button);
    }
    if(!obj){
	if(dbg_flag)cerr<<"no ob below mouse\n";
	return NULL;
    }
    tree_node_c *track_node = root->find_object(obj);
    if(!track_node){
	if(dbg_flag)cerr<<"no tracking node below mouse\n";
	dnd_notarget_flag = TRUE;
	tree_set_complete_wintitle();
	return NULL;
    } else {
	dnd_notarget_flag = FALSE;
    }	
    if(track_node == target_node){
	if(dbg_flag)cerr<<"track ob is target node.\n";
	// this is clever enough not to redraw the window title always.
	tree_set_complete_wintitle();
	return track_node->get_path();
    }
    if(!track_node->restat(TRUE, FALSE)){
	tree_remove_invalid_node(track_node);
	return NULL;
    }
    tree_set_target(track_node);
    tree_set_complete_wintitle();
    return track_node->get_path();
}

static int tree_dnd_backend(tree_node_c *from_node, tree_node_c *new_target_node){
    if(!new_target_node || !from_node){
	cerr << "UCT: not a target node.\n";
	return FALSE;
    } else if(from_node == root){
	// can't move root dir.
	status_msg(2, "can't move root directory.");
	return FALSE;
    } else if(new_target_node == from_node){
	// don't do a thing
	return FALSE;
    } else if(new_target_node == from_node->parent){
	// we're a child already
	status_msg(2, "directory is already direct child of parent.");
	return FALSE;
    } else {
	tree_node_c *tmp_node = new_target_node->parent;
	while(tmp_node){
	    if(tmp_node == from_node){
		// target is subdir of from.
		status_msg(2, "cannot move parent into child directory.");
		return FALSE;
	    }
	    tmp_node = tmp_node->parent;
	}
    }
    if(dnd_copy_flag){
	cerr << "shift = copy.\n";
    }
    if(dbg_flag)cerr << "dndmove " << from_node->name << " to " << new_target_node->name << endl;
    //use getpath :-)
    const char *from_ptr = from_node->get_path();
    if(!from_ptr){
	cerr << "UCT: get_path failed.\n";
	return FALSE;
    } 
    char from_buf[MAXPATHLEN];
    strcpy(from_buf, from_ptr);
    
    const char *to_ptr = new_target_node->get_path();
    if(!to_ptr){
	cerr << "UCT: get_path failed.\n";
	return FALSE;
    } 
    char to_buf[2*MAXPATHLEN];
    sprintf(to_buf, "%s/%s", to_ptr, from_node->name);
    if(dbg_flag){
	cerr << "dnd from: " << from_buf << endl;
	cerr << "dnd to:   " << to_buf << endl;
    }
    
    if(rename(from_buf, to_buf) == 0){
	// need to adjust tree structure, too !
	fl_freeze_form(TreeForm);
	from_node->reconnect(new_target_node);
	tree_reshape();

	// take care of possible inconsistencies due to the move. (all browsers and shell)
	if(from_node->find_node(source_node)){
	    tree_change_source_to_node(source_node);
	}
	if(source_browser != target_browser && dnd_old_target_node){
	    if(from_node->find_node(dnd_old_target_node)){
		tree_change_target_to_node(dnd_old_target_node);
	    }
	}
	browser_c *tmp_browser = source_browser;
	do {
	    if(tmp_browser != source_browser &&
	       tmp_browser != target_browser){
		const char *cwd_buf = tmp_browser->get_cwd();
		unsigned int len = strlen(from_buf);
		if(cwd_buf && len > 0){
		    if(strlen(cwd_buf) >= len &&
		       strncmp(cwd_buf,from_buf,len) == 0){
			char buf[2*MAXPATHLEN];
			sprintf(buf, "%s%s", to_buf, cwd_buf+len);
			//cerr << "would update tmp_browser from " << cwd_buf << " to " << buf << endl;
			tmp_browser->set_cwd_buf(buf);
			tmp_browser->set_complete_wintitle();
		    }
		}
	    }
	    tmp_browser = tmp_browser->next_browser;
	} while(tmp_browser != source_browser);
	fl_unfreeze_form(TreeForm);
	
	return TRUE;
    } else {
	if(errno == EXDEV){
	    timed_error_notifier("Sorry, drag & drop move failed.",
				 "Can't move between file systems (use copy instead)."
	    );
	} else {
	    timed_strerror_notifier("Sorry, drag & drop move failed.");
	}
	return FALSE;
    }
}

static int tree_dnd_process_xy(tree_node_c *from_node, FL_Coord x, FL_Coord y){
    // return if a change has happened, so that we can restore the original
    // target node if necessary.
    if(!from_node || !root){
	cerr << "UCT: no from-node or root.\n";
	return FALSE;
    }

    if(x < left_margin || x > TreeForm->w ||
       y > TreeForm->h - bottom_margin || y < 0){
	// quick death.
	return FALSE;
    }

    // is it on a button ? (obj under the mouse)
    FL_OBJECT *obj;
    obj = fl_find_last(TreeForm, FL_FIND_MOUSE, x, y);
    if(!obj || obj == TreeFormFreeObj) return FALSE;
    tree_node_c *new_target_node = root->find_object(obj);
    return tree_dnd_backend(from_node, new_target_node);
}

int tree_move_path2dir(const char *source_path, const char *target_dir, int reshape_flag){
    if(!source_path || !target_dir || !root) return FALSE;
    tree_node_c *from_node = root->find_path(source_path);
    if(!from_node) return FALSE;
    tree_node_c *new_target_node = tree_construct_path(target_dir);
    if(!new_target_node) return FALSE;
    from_node->reconnect(new_target_node);
    if(reshape_flag){
	tree_reshape();
    }
    return !reshape_flag;
}

void tree_rename_path(const char *source_path, const char *target_path){
    if(!source_path || !target_path || !root) return;
    if(dbg_flag)cerr << "tree rename path: " << source_path << ", " << target_path << endl;
    tree_node_c *from_node = root->find_path(source_path);
    if(!from_node) return;
    
    // don't want/can't deal with non-abs names here.
    const char *last_slash = strrchr(target_path, '/');
    if(!last_slash || *(last_slash+1) == '\0') return; 

    if(strncmp(source_path, target_path, last_slash-target_path) != 0) return;
	
    if(from_node->name){
	fl_free(from_node->name);
    }
    from_node->name = strdup(last_slash+1);
    from_node->retrunc(FALSE);
}

#include "bitmaps/folder.bm"
#ifdef SGI	    
#include "bitmaps/folder_mask.bm"
#endif	

static Cursor folder_cursor = 0;
void make_folder_cursor(void){
    if(!folder_cursor){
	Pixmap pixmap = XCreatePixmapFromBitmapData(
	    fl_display, TreeForm->window,
	    folder_bits, folder_width, folder_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
	Pixmap mask_pixmap =
#ifdef SGI	    
	XCreatePixmapFromBitmapData(
	    fl_display, TreeForm->window,
	    folder_mask_bits, folder_mask_width, folder_mask_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
#else
	None;
#endif	
	XColor fg_c, bg_c, junk;
	if(XLookupColor(fl_display, fl_map(fl_vmode), "red", &junk, &fg_c)){
	    if(XLookupColor(fl_display, fl_map(fl_vmode), "white", &junk, &bg_c)){
		folder_cursor = XCreatePixmapCursor(
		    fl_display, pixmap, mask_pixmap,
		    &fg_c, &bg_c, folder_width/2, folder_height/2
	        );
		return;
	    }
	}
	cerr << "folder cursor creation failed.\n";
	//fl_set_cursor_colors(folder_cursor, FL_RED, FL_WHITE);
	folder_cursor = XC_fleur;
    }
}

static int tree_handle_keyrelease(XKeyEvent * xkev) {
    if (!xkev) return FALSE;
    KeySym keysym;
    char keybuf[1];
    tool_got_event_flag = TRUE;
    // gobble all.
    XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);
    if (keysym == XK_Escape) {
	//cerr << "escape FALSE.\n";
	tool_escape_flag = FALSE;
    }
    return TRUE;
}

static int tree_handle_keypress(XKeyEvent * xkev) {
    if (!xkev) return FALSE;
    KeySym keysym;
    char keybuf[1];
    tool_got_event_flag = TRUE;
    if(dbg_flag) cerr << "tree: found an X key press event.\n";
    XLookupString(xkev, keybuf, sizeof(keybuf), &keysym, 0);

    if(is_end(keysym, xkev)){
	tree_end();
	return TRUE;
    } else if(is_down(keysym, xkev)){
	tree_down_arrow();
	return TRUE;
    } else if(is_page_down(keysym, xkev)){
	tree_page_down();
	return TRUE;
    } else if(is_left(keysym, xkev)){
	tree_left_arrow();
	return TRUE;
    } else if(is_begin(keysym, xkev)){
	TreeCenterCB(NULL, 0);
	return TRUE;
    } else if(is_right(keysym, xkev)){
	tree_right_arrow();
	return TRUE;
    } else if(is_home(keysym, xkev)){
	TreeHomeCB(NULL, 0);
	return TRUE;
    } else if(is_up(keysym, xkev)){
	tree_up_arrow();
	return TRUE;
    } else if(is_page_up(keysym, xkev)){
	tree_page_up();
	return TRUE;
    }

    switch(keysym){
    case XK_F1:
	if(dbg_flag) cerr << "tree got F1.\n";
	help_event_cb(tree_form_help_func);
	return TRUE;
    case XK_F2:
	if(tool_get_meta()){
	    OptionsButtonCB(NULL,0);
	    return TRUE;
	} else {
	    TreeFormOptionsButtonCB(NULL, TRUE); 
	    return TRUE;
	}
    case XK_F3:
	tree_mkdir();
	return TRUE;
    case XK_F4:
	tree_raise_form();
	return TRUE;
    case XK_F5:
	tree_view();
	return TRUE;
    case XK_F6:
	tree_edit();
	return TRUE;
    case XK_F7:
	tree_rename();
	return TRUE;
    case XK_F8:
	tree_move();
	return TRUE;
    case XK_F9:
	tree_duplicate();
	return TRUE;
    case XK_F10:
	tree_copy();
	return TRUE;
#ifdef DEC_ALPHA
    case XK_F13:
#endif
    case XK_F11:
	tree_delete();
	return TRUE;
    case SunXK_F37: if(!sun_keyboard_flag) return FALSE;
    case XK_F12:
	ExitButtonCB(NULL, 0);
	return TRUE;
    case XK_Escape:
	tool_escape_flag = TRUE;
	tree_search_clear_mode(TRUE);
	return TRUE;
    case XK_KP_Tab:
    case XK_Tab:
	if(dbg_flag)cerr << "tab\n"; 
	if(!tree_target_mode_flag){
	    if(tool_get_ctrl()){
		xkev->state = xkev->state & ~ControlMask;
		xshell_sendevent_callback(xkev, XK_Tab);
	    } else {
		TabKeyCB(NULL, 1);
	    }
	}
	return TRUE;
    case XK_KP_Add:   
	TreeExpandCB(NULL, 0);
	return TRUE;
    case XK_KP_Subtract:
	TreeCollapseCB(NULL, 0);
	return TRUE;
    case XK_s: case XK_S:
	// needs to be preemptive, before search key cb.
	if(xkev->state & ControlMask){
	    tree_search_event_cb(NULL, 0);
	    return TRUE;
	}
	break;
    }
    //printf("keysym: %x\n", (unsigned int) keysym);
    if(tree_search_key_cb(xkev, keysym, keybuf[0])) return TRUE;
    if(tree_key_event(xkev, keysym, keybuf[0])) return TRUE;
    if(direct_window_access_cb(xkev, keysym, TRUE)) return TRUE;
    return xshell_sendevent_callback(xkev, keysym);
}

int tree_handle_free(FL_OBJECT *obj, int event, FL_Coord x, FL_Coord y, int key, void *xev){
    static FL_Coord w=0, h=0, start_x=0, start_y=0;
    switch(event){
    case FL_DRAW:
	tool_got_event_flag = TRUE;
	if(dbg_flag)cerr << "tree FL_DRAW event.\n";
	if(TreeForm->w != w || TreeForm->h != h){
	    if(dbg_flag) cerr << "new tree size\n";
	    w = TreeForm->w;
	    h = TreeForm->h;
	    tree_form_scale_cb();
	}
	if(root){
	    root->draw_lines();
	    //cerr << "drawing done. Press return to continue.\n";
	    //char c;
	    //cin >> c;
	}
	return TRUE;
    case FL_PUSH:
	if(dbg_flag)cerr << "handle_free: push\n";
	start_x = x;  // seed values.
	start_y = y;
	if(shove_flag){
	    if(dbg_flag)cerr << "and hold: shove == true\n";
	    fl_set_cursor_color(XC_fleur, FL_RED, FL_WHITE);
	    fl_set_cursor(TreeForm->window, XC_fleur);  
	    tree_set_complete_wintitle();
	} else if(dnd_flag && deact_node){
	    if(dbg_flag)cerr << "and dnd.\n";
	    if(deact_node != current_node){
		// no sync;
		tree_set_current(deact_node, FALSE, FALSE);
		if(deact_node != current_node){
		    // important error checking.
		    deact_node = NULL;
		    dnd_flag = FALSE;
		    tree_set_complete_wintitle();
		    return FALSE;
		}
	    }
	    // dnd_copy_flag = shiftkey_down(fl_keymask);  // important at start.
	    dnd_old_target_node = target_node;
	    make_folder_cursor();
	    //fl_set_cursor(TreeForm->window, folder_cursor);
	    XDefineCursor(fl_display, TreeForm->window, folder_cursor);
	    tree_set_complete_wintitle();    // important to do it here once.
	}
	return TRUE;
    case FL_RELEASE:
	tool_got_event_flag = TRUE;
	if(dbg_flag)cerr << "handle_free: release\n";
	if(shove_flag){
	    fl_reset_cursor(TreeForm->window);
	    shove_flag = FALSE;
	    tree_set_complete_wintitle();
	} else if(dnd_flag){
	    fl_reset_cursor(TreeForm->window);
	    dnd_flag = FALSE;
	    if(deact_node){
		fl_activate_object(deact_node->button);
	    }
	    if(tree_dnd_process_xy(deact_node,x,y)){
		// need a cleverer solution for this.
		tree_set_target(dnd_old_target_node);
	    } else {
		// restore old
		tree_set_target(dnd_old_target_node);
		tree_set_complete_wintitle();
	    }
	    deact_node = NULL;
	}
	return TRUE;
    case FL_MOUSE:
	if(dbg_flag)cerr << "handle_free: mouse position at x " << x << ",y " << y << ",key " << key << endl;
	tool_got_event_flag = TRUE;
	if(dbg_flag)cerr << "shove,dnd: " << shove_flag << "," << dnd_flag << endl;
	if(shove_flag || dnd_flag){
	    FL_Coord dx, dy;
	    if(CheckPeekWindowEvent(fl_display, TreeForm->window, MotionNotify)){
		if(dbg_flag)cerr << "compressing.\n";
		return TRUE;
	    } else if(dnd_flag){
		// start at boundaries. position/speed based.
		// don't dnd-scroll if inside of TreeFormBox.
		dx = dy = 0;
		if(x < left_margin) dx = x-left_margin;
		if(x > TreeForm->w) dx = x-TreeForm->w;
		if(y < 0) dy = y;
		if(y > TreeForm->h - bottom_margin) dy = y - (TreeForm->h - bottom_margin);
	    } else {
		// get the mouse pos for shoving
		dx = x-start_x;
		dy = y-start_y;
	    }
	    if(dx || dy){
		root_rel_x -= dx;
		root_rel_y -= dy;
		new_root_rel();
	    } else {
		// we're not scrolling, attempt tracking.
		if(dnd_flag){
		    tree_dnd_tracking(x,y);
		}
	    }
	}
	return TRUE;
    default:
	break;
    }

    if (event != FL_KEYBOARD) {
	return FALSE;
    }
    
    if(!xev || !obj){
	if(dbg_flag) cerr  << "tree handle free called with Null xev or obj.\n";
	return FALSE;
    }
    if(dbg_flag)cerr << "tree handle free called.\n";
    
    switch(((XEvent *)xev)->type){
    case KeyPress:
	return tree_handle_keypress((XKeyEvent *) xev);
    case KeyRelease:
	return tree_handle_keyrelease((XKeyEvent *) xev);
    default:
	if(dbg_flag)cerr << "tree event (" << event << ") X(" << ((XEvent *)xev)->type << ") unprocessed.\n";
	break;
    }
    return FALSE;
}

void tree_form_help_func(void){
    if(bIsSearchMode && TreeForm->visible){
	load_help_browser(tree_search_help_text, "Directory Tree Search");
    } else {
	if(tree_target_mode_flag){
	    load_help_browser(tree_target_mode_help_text, "Tree Select Target Mode");
	} else {
	    load_help_browser(tree_help_text, "Directory Tree");
	}
    }
}

static void tree_form_scale_cb(void){
    if(dbg_flag)cerr << "tree_form_scale_cb.\n";

    TreeFormBox->w = TreeForm->w - box_unit;
    TreeFormBox->y = box_unit;
    TreeFormBox->h = TreeForm->h - box_unit - TreeFormBox->y;
    TreeFormBox->x = box_unit;
	
    TreeFormVertSlider->w = box_unit;
    TreeFormVertSlider->h = TreeForm->h - 4*box_unit;
    TreeFormVertSlider->y = 2*box_unit;
    TreeFormVertSlider->x = 0;

    TreeFormHorSlider->w = TreeForm->w - 3*box_unit;
    TreeFormHorSlider->h = box_unit;
    TreeFormHorSlider->y = TreeForm->h - box_unit;
    TreeFormHorSlider->x = 2 * box_unit;

    TreeFormHorLeftButton->w = box_unit;
    TreeFormHorLeftButton->h = box_unit;
    TreeFormHorLeftButton->x = box_unit;
    TreeFormHorLeftButton->y = TreeFormHorSlider->y;

    TreeFormHorRightButton->w = box_unit;
    TreeFormHorRightButton->h = box_unit;
    TreeFormHorRightButton->x = TreeFormHorSlider->x + TreeFormHorSlider->w;
    TreeFormHorRightButton->y = TreeFormHorSlider->y;

    TreeFormVertDownButton->w = box_unit;
    TreeFormVertDownButton->h = box_unit;
    TreeFormVertDownButton->x = 0;    
    TreeFormVertDownButton->y = TreeFormHorSlider->y - box_unit;
    
    TreeFormVertUpButton->w = box_unit;
    TreeFormVertUpButton->h = box_unit;
    TreeFormVertUpButton->x = 0;    
    TreeFormVertUpButton->y = box_unit;

    fl_freeze_form(TreeForm);
    if (root) root->realign();
    tree_refit();
    if (root) root->realign();
    tree_adjust_sliders();
    fl_unfreeze_form(TreeForm);
}

static char select_path_prefix[MAXPATHLEN+1];

typedef struct {
    unsigned char bit0:1;
    unsigned char bit1:1;
    unsigned char bit2:1;
    unsigned char bit3:1;
    unsigned char bit4:1;
    unsigned char bit5:1;
    unsigned char bit6:1;
    unsigned char bit7:1;
} bitfield;

#if defined(AIX)
// need C linkage for select function for IBM C.
extern "C" {
#endif

#if defined(HP) || defined(LINUX)
  // new Suse Linux 7.0 uses const, too
int scandir_select(const struct dirent *entry){
#elif defined(SGI) || defined(DEC_ALPHA) || defined(AIX) || defined(SOLARIS) || defined(UNIXWARE) || defined(SCO) || defined(SINIX)
int scandir_select(struct dirent *entry){
#else
int scandir_select(struct direct *entry){   // SUN
#endif
    if(!entry){
	return FALSE;
    }
    const char *name = entry->d_name;
    /* don't want dotdots in our list: */
    if(strcmp(name, ".") == 0){
	//cerr << "killing dot." << endl;
	return FALSE;
    }
    if(strcmp(name, "..") == 0){
	//cerr << "killing dotdot." << endl;
	return FALSE;
    }

    // cerr << "select on " << name << endl;
    /*
    char buffer[1024];
    strcpy(buffer, select_path_prefix);
    strcat(buffer, "/");
    strcat(buffer, name);
    */
    //cerr << "full name: " << name << endl;

    struct stat statbuf;
#if defined(SUNOS)
    if(lstat((char *)name, &statbuf) == 0){
#else
    if(lstat(name, &statbuf) == 0){
#endif
	if(S_ISDIR(statbuf.st_mode)){
	    // cerr << "is a directory ! " << endl;
#if defined(HP) || defined(LINUX)
	    ((struct dirent *)entry)->d_ino = FALSE;
#else
	    entry->d_ino = FALSE;
#endif
	    return TRUE;
	} else if(S_ISLNK(statbuf.st_mode)){
	    // follow the link ... 
	    if(stat(name, &statbuf) == 0){
		if(S_ISDIR(statbuf.st_mode)){
		    if(dbg_flag) cerr << "found a symlink: " << name << endl;
#if defined(HP) || defined(LINUX)
		    ((struct dirent *)entry)->d_ino = TRUE;
#else
		    entry->d_ino = TRUE;
#endif
		    return TRUE;
		} else {
		    return FALSE;
		}
	    } else {
	        status_msg(2, "UCTree: error in stat of file %s in dir %s: %s", name, select_path_prefix, strerror(errno));
		return FALSE;
	    }
	} else {
	    return FALSE;
	}
    } else {
        status_msg(2, "UCTree: error in lstat of file %s in dir %s: %s", name, select_path_prefix, strerror(errno));
	return FALSE;
    }
}

#if defined(AIX)
// need C linkage for select function for IBM C.
}
#endif

static void tree_destroy_node(tree_node_c *tmp_node){
    if(!tmp_node) return;
    if(tmp_node == root) return;
    while(tmp_node->child0){
	tmp_node->delete_sub_tree(tmp_node->child0);
    }
    delete tmp_node;
}
    

void tree_node_c::delete_sub_tree(tree_node_c *sub_root){
    // complete 
    if(!sub_root) return;

    // safeguard: search
    tree_node_c *tmp_node = child0;
    int found_flag = FALSE;
    while(tmp_node && !found_flag){
	if(tmp_node == sub_root){
	    found_flag = TRUE;
	} else {
	    tmp_node = tmp_node->next;
	    if(!tmp_node){
		program_unstable("tree_node_c: attempt to delete an invalid subtree.");
		return;
	    }
	}
    }

    // erase sub_root from child list. 
    if(sub_root == child0){
	if(child0->next){
	    child0 = child0->next;
	    child0->prev = NULL;
	} else {
	    // else one child only.
	    child0 = NULL;
	}
    } else {
	sub_root->prev->next = sub_root->next;
	if(sub_root->next){
	    sub_root->next->prev = sub_root->prev;
	}
    }

    // erase sub_roots subtrees.
    while(sub_root->child0){
	sub_root->delete_sub_tree(sub_root->child0);
    }

    // kill the bugger.
    delete sub_root;
}

tree_node_c *tree_node_c::insert_new(const char *new_name){
    if(!new_name) return NULL;
    tree_node_c *tmp_node = new tree_node_c(new_name);
    if(!tmp_node) return NULL;
    insert_child(tmp_node);
    return tmp_node;
}

tree_node_c *tree_node_c::insert_child(tree_node_c *new_child){
    if(!new_child){
	cerr << "UCT: insert_child failed.\n";
	return NULL;
    }
    new_child->parent = this;
    if(!child0){
	child0 = new_child;
	return new_child;
    } else {
	tree_node_c *work_node = child0, *old_work_node = NULL;
	while(work_node){
	    old_work_node = work_node;
	    if(strcmp(work_node->name, new_child->name) < 0){
		work_node = work_node->next;
	    } else {
		if(work_node == child0){
		    // before old child0
		    child0 = new_child;
		    child0->next = work_node;
		    work_node->prev = child0;
		    return new_child;
		} else {
		    // somewhere in the middle
		    work_node->prev->next = new_child;
		    new_child->prev  = work_node->prev;
		    work_node->prev = new_child;
		    new_child->next  = work_node;
		    return new_child;
		}
	    }
	}
	// last in line
	old_work_node->next = new_child;
	new_child->prev  = old_work_node;
	return new_child;
    }
}
	

static tree_node_c *tree_scan_find_child(const char *name, tree_node_c *old_child0){
    tree_node_c *tmp_node = old_child0;
    while(tmp_node){
	if(strcmp(name, tmp_node->name) == 0){
	    return tmp_node;
	} else {
	    tmp_node = tmp_node->next;
	}
    }
    return NULL;
}

static void tree_scan_delete_from_oldlist(tree_node_c *tmp_node, tree_node_c **old_child0_p){
    tree_node_c *old_child0 = *old_child0_p;
    // rip the child out of the old child list. If it is the first,
    // update the old_child0 pointer. This is NOT an optimization.
    // if the child gets reconnected while staying in the old list :-(
    if(tmp_node == old_child0){
	old_child0 = tmp_node->next;
	if(old_child0){
	    old_child0->prev = NULL;
	}
	*old_child0_p = old_child0;
    } else {
	// look back.
	if(tmp_node->prev){
	    tmp_node->prev->next = tmp_node->next;
	    tmp_node->prev = NULL;
	} else {
	    cerr << "rescan: strange: no prev in tmp_node != old_child0";
	    const char *path = tmp_node->get_path();
	    if(path){
		cerr << path << endl;
	    } else {
		cerr << "stranger: no path!\n";
	    }
	}
	// look forward.
	if(tmp_node->next){
	    tmp_node->next->prev = tmp_node->prev;
	    tmp_node->next = NULL;
	}
    }
}

int tree_delete_path(const char *path, int reshape_flag){
    if(!path || !root) return FALSE;
    tree_node_c *delete_node = root->find_path(path);
    if(!delete_node || delete_node == root) return FALSE;
    if(delete_node->cut_yourself()){
	tree_destroy_node(delete_node);
	if(reshape_flag){
	    tree_reshape();
	}
	return !reshape_flag;
    } else {
	return FALSE;
    }
}

tree_node_c *tree_construct_path(const char *path, int reshape_flag){
    if(!root) return NULL;
    // cut into pieces. Trace down until no more matches. Then, local
    // mini scan.
    if(!path){
	cerr << "tree_node_c:: cannot construct a Null path.\n";
	return NULL;
    }
    if(path[0] == '\0'){
	cerr << "tree_node_c:: cannot construct an empty path.\n";
	return NULL;
    }
    if(path[0] != '/'){
	if(dbg_flag) cerr << "tree_node_c:: construct: relative path.\n";
	// return NULL;
    }
    if(strcmp(path, root->name) == 0){
	return root;
    }
    /* buggy code from multi-component times.
    if(path[1] == '\0'){
	if(strcmp(path, root->name) == 0){
	    return root;
	} else {
	    M_err("tree_construct_path","couldn't match `%s' to root `%s'.\n",
		path, (root->name ? root->name : "NULL rootname")
	    );
	    return NULL;
	}
    }
    */
    tree_node_c *max_match_node = root, *tmp_node = NULL;
    /*
    // quick fix.
    tree_node_c *max_match_node = root->find_path(path), *tmp_node = NULL;
    if(max_match_node){
	return max_match_node;
    } else {
	max_match_node = root;
    }
    */
    
    // construct maximum match.
    char path_copy[MAXPATHLEN+1];
    strcpy(path_copy, strip_trailing_slashes(path+1));      // this would have failed for "" path!

    char *tok = strtok(path_copy, "/");
    
    while(tok){
	if(dbg_flag) cerr << "tree_construct_path: tok is: " << tok << endl;
	if(strcmp(tok, ".") == 0){
	    if(dbg_flag) cerr << "tree_construct_path: skipping `.'\n";
	    // skip.
	} else if(strcmp(tok, "..") == 0){
	    if(dbg_flag) cerr << "tree_construct_path: attempting `..'\n";
	    if(max_match_node != root){
		if(max_match_node->parent){
		    max_match_node = max_match_node->parent;
		} else {
		    cerr << "uctree: No parent in tree construct path!\n";
		    return NULL;
		}
	    }
	    // else stay root.
	} else {
	    // else it is a real path component.
	    tmp_node = max_match_node->find_child(tok);
	    if(!tmp_node){
		tmp_node = max_match_node->insert_new(tok);
		if(!tmp_node){
		    cerr << "tree_error_c::insert_new failed.\n";
		    return NULL;
		} else {
		    if(reshape_flag){
			tree_reshape();
		    }
		}
	    }
	    max_match_node = tmp_node;
	}
	tok = strtok(NULL, "/");
    }
    return max_match_node;
}

int tree_node_c::restat(int quiet_flag, int subtree_flag){
    const char *path = get_path();
    if(!path){
	cerr << "NULL path in restat.\n";
	return FALSE;
    }
    struct stat statbuf;
    if(stat(path, &statbuf) == 0){  
	if(S_ISDIR(statbuf.st_mode)){
	    if(subtree_flag){
		tree_node_c *tmp_node = child0, *last_good_node = NULL;
		while(tmp_node){
		    if(!tmp_node->restat(quiet_flag)){
			last_good_node = tmp_node->prev;
			tmp_node->parent->delete_sub_tree(tmp_node);
			// tricky restart of the restat operation:
			tmp_node = last_good_node;
			if(!tmp_node){
			    tmp_node = child0;
			}
		    } else {
			tmp_node = tmp_node->next;
		    }
		}
	    }
	} else {
	    if(!quiet_flag){
		timed_error_notifier("UC: restat: The following file is no directory:", path, notifier_secs);
	    };
	    status_msg(2, "Not a directory: %s", path);
	    tree_has_changed_flag = TRUE;
	    return FALSE;
	}
    } else {
	if(!quiet_flag){
	    timed_error_notifier("UC: restat: The following directory cannot be accessed:", path, notifier_secs);
	}
	status_msg(2, "Cannot stat directory: %s", path);
	
	tree_has_changed_flag = TRUE;
	return FALSE;
    }
    return TRUE;
}

static int tree_restat_below(tree_node_c *node, int recurse_level, int quiet_flag){
    if(!node){
	cerr << "NULL pointer in tree_restat_below!\n";
	return FALSE;
    }
    /*
    if(!scan_continue_flag){
	return TRUE;
    }
    */
    if(recurse_level <= 0){
	return tree_restat(node, quiet_flag);
    } else {
	recurse_level--;
    }
    tree_node_c *tmp_node = node->child0;
    while(tmp_node){  // && scan_continue_flag){
	tree_restat_below(tmp_node, recurse_level, quiet_flag);
	tmp_node = tmp_node->next;
    }
    return TRUE;
}

static int tree_restat(tree_node_c *node, int quiet_flag){
    if(!node) return FALSE;
    const char *cwd_buf = node->get_path();
    if(cwd_buf){
	char buffer[MAXPATHLEN+64];
	sprintf(buffer, "UCTree: Verifying %s (no interrupt) ...", cwd_buf);
	tree_set_wintitle_backend(buffer);
    }
    
    fl_freeze_form(TreeForm);
    int retval = node->restat(quiet_flag);
    if(!retval){
	// node itself is corrupt.
	status_msg(0, "Node %s is invalid.", node->get_path());
	status_msg(0, "removing subtree.");
	
	if(node == root){
	    timed_error_notifier("Tree data structure is corrupted.",
				 "Remove ~/uctreedb and restart.",
				 notifier_secs);
	    return FALSE;
	} else {
	    tree_node_c *parent = node->parent;
	    fl_freeze_form(TreeForm);
	    parent->delete_sub_tree(node);
	    tree_set_current(parent);
	    fl_unfreeze_form(TreeForm);
	}
    }
    
    tree_reshape();
    fl_unfreeze_form(TreeForm);
    tree_set_wintitle_backend("UCTree: Verification done.");
    return retval;
}

#if defined(SUN) 
typedef int (*intfunc_pt)();
#endif

static int _tree_scan(tree_node_c *parent, int max_levels, int max_nodes, int rescan_flag){
    // 2 simultaneous constraints.
    if(!scan_continue_flag){
	return TRUE;
    }
    if(max_levels < 0 && max_levels != SCAN_ALL_LEVELS){
	return TRUE;
    }
    if(!indicator_step()){
	scan_continue_flag = FALSE;
	if(dbg_flag) cerr << "interrupted." << endl;
	return TRUE;
    }
    if(!parent){
	cerr << "NULL pointer in tree_scan!\n";
	return FALSE;
    }
    if(parent->symlink_flag){
	if(dbg_flag) cerr << "not following a symlink (" << parent->name << ")\n";
	parent->read_link();
	return TRUE;
    }
    if(dbg_flag) cerr << "operating on " << parent->name << ", level " << max_levels << endl;
    if(max_levels != SCAN_ALL_LEVELS){
	max_levels--;
    }
    parent->no_perms_flag = FALSE;
    
    strcpy(select_path_prefix, parent->get_path());
    if(safe_chdir(select_path_prefix) != 0){
	status_msg(2, "UCTree: can't change directory to %s: %s", select_path_prefix, strerror(errno));
	parent->no_perms_flag = TRUE;
	return FALSE;
    }
#if defined(SUNOS)
    struct direct **namelist;	
    int n_entries = scandir(".", &namelist, (intfunc_pt) scandir_select, (intfunc_pt) alphasort);
#else
    struct dirent **namelist;
    int n_entries = scandir(".", &namelist, scandir_select, alphasort);
#endif
    if(n_entries == -1){
        status_msg(2, "UCTree: scandir error for %s: %s", select_path_prefix, strerror(errno));
	if(errno == EPERM){
	}
	parent->no_perms_flag = TRUE;
	return FALSE;
    }

    tree_node_c
	*tmp_node   = NULL,
	*work_node  = NULL,
	*old_child0 = parent->child0  // needs to be remembered.
    ;

    int i=0; 
    while(i<n_entries){
	// printf("#%d> %s\n", i,  namelist[i]->d_name);
	
	tmp_node = tree_scan_find_child(namelist[i]->d_name, old_child0);
	if(rescan_flag && tmp_node){
	    //cerr << "found child: " << tmp_node->name << endl;
	    tree_scan_delete_from_oldlist(tmp_node, &old_child0);
	    // now insert the free child
	    if(i==0){
		parent->child0 = tmp_node;
	    } else {
		work_node->next = tmp_node;
		tmp_node->prev = work_node;
	    }
	    // tmp_node->parent = parent;
	    work_node = tmp_node;
	} else {
	    tmp_node = new tree_node_c(namelist[i]->d_name);
	    if(i==0){
		parent->child0 = tmp_node;
	    } else {
		work_node->next = tmp_node;
		tmp_node->prev = work_node;
	    }
	    tmp_node->parent = parent;
	    work_node = tmp_node;
	}
	if(namelist[i]->d_ino > 0){
	    work_node->symlink_flag = TRUE;
	    work_node->read_link();
	    work_node->realign();
	}
	_tree_scan(work_node, max_levels, max_nodes, rescan_flag);
	if(i%3 == 0 && scan_continue_flag){
	    scan_continue_flag = indicator_ministep(FALSE);
	}
	i++;
    }
//    cerr << "done" << endl;
//    cerr << endl;
    if(rescan_flag){
	if(old_child0){
	    if(dbg_flag) cerr << "seems that old list members are still there. Removing ... \n";
	    work_node = old_child0;
	    while(work_node){
		tmp_node = work_node->next;
		tree_destroy_node(work_node);
		work_node = tmp_node;
	    }
	}
    }
    for(i=0; i<n_entries; i++){
	//fl_free(namelist[i]->d_name);  // coredumps !
#ifdef SUN	
	fl_free((char *) namelist[i]);
#else
	fl_free(namelist[i]);
#endif
    }
#ifdef SUN	
    fl_free((char *)namelist);
#else
    fl_free(namelist);
#endif
    char buffer[64];
    sprintf(buffer, "%d", tree_node_c::next_ID);
    if(scan_continue_flag){
	scan_continue_flag = indicator_check(buffer);
    }
    return TRUE;
}

static int tree_scan(tree_node_c *parent, int max_levels, int max_nodes, int rescan_flag){
    if(max_levels == SCAN_ALL_LEVELS){
	indicator_simple_bar("Full subdirectory scan");
	indicator_check("0");
    }    
    tree_scanning_flag = TRUE;
    tree_set_complete_wintitle();
    scan_continue_flag = TRUE;
    int retval = _tree_scan(parent, max_levels, max_nodes, rescan_flag);
    tree_scanning_flag = FALSE;
    if(max_levels == SCAN_ALL_LEVELS){
	indicator_end();
    }
    safe_chdir(source_browser->get_cwd());
    return retval;
}

/* This version supports iterative deepening. It assumes that the tree up
 * to recurse_level level is consistent (ie, has just been rescanned).
 * It then only does new scanning on the recurse_level.
 */
static int _tree_scan_below(tree_node_c *parent, int recurse_level, int max_nodes, int rescan_flag){
    tree_has_changed_flag = TRUE;
    if(!parent){
	cerr << "NULL pointer in tree_scan!\n";
	return FALSE;
    }
    if(!scan_continue_flag){
	return TRUE;
    }
    if(recurse_level <= 0){
	return tree_scan(parent, 0, max_nodes, rescan_flag);
    } else {
	recurse_level--;
    }
    tree_node_c *tmp_node = parent->child0;
    while(tmp_node && scan_continue_flag){
	_tree_scan_below(tmp_node, recurse_level, max_nodes, rescan_flag);
	tmp_node = tmp_node->next;
    }
    return TRUE;
}

static int tree_scan_below(tree_node_c *parent, int recurse_level, int max_nodes, int rescan_flag){
    tree_scanning_flag = TRUE;
    tree_set_complete_wintitle();
    scan_continue_flag = TRUE;
    int retval = _tree_scan_below(parent, recurse_level, max_nodes, rescan_flag);
    tree_scanning_flag = FALSE;
    return retval;
}


/*
    can't do this, it has to be done in the first instance the form is made. 
    fl_addto_form(TreeForm);
    extern int handle_lines_free(FL_OBJECT *obj, int event, FL_Coord mx, FL_Coord my, char key);
    fl_add_free(FL_NORMAL_FREE, 0, 0, TreeForm->w,TreeForm->h,"", handle_lines_free);
    fl_end_form();
 */

static void tree_set_source(tree_node_c *new_node){
    if(source_node != new_node){
	source_node = new_node;
	if(TreeSourceBox->visible){
	    // paint over old position :-)
	    fl_set_object_color(TreeSourceBox, tree_lines_bg_colour, FL_COL1);
	    fl_redraw_object(TreeSourceBox);
	}
	if(source_node){
	    TreeSourceBox->visible = source_node->button->visible;
	    int h = (int) floor(source_node->button->h/2.0), w = frame_margin;
	    if(h < frame_margin) w = h;
	    TreeSourceBox->x = source_node->button->x-w;
	    TreeSourceBox->y = source_node->button->y;
	    TreeSourceBox->w = w;
	    TreeSourceBox->h = h;
	    fl_set_object_color(TreeSourceBox, tree_source_colour, FL_COL1);
	    // hack:
	    fl_redraw_object(TreeSourceBox);
	}
    }
}

static void tree_set_target(tree_node_c *new_node){
    if(target_node != new_node){
	target_node = new_node;
	if(TreeTargetBox->visible){
	    // paint over old position :-)
	    fl_set_object_color(TreeTargetBox, tree_lines_bg_colour, FL_COL1);
	    fl_redraw_object(TreeTargetBox);
	}
	if(target_node){
	    TreeTargetBox->visible = target_node->button->visible;
	    int h = (int) floor(target_node->button->h/2.0), w = frame_margin;
	    if(h < frame_margin) w = h;
	    TreeTargetBox->x = target_node->button->x-w;
	    TreeTargetBox->y = target_node->button->y+h+1;
	    TreeTargetBox->w = w;
	    TreeTargetBox->h = h;
	    fl_set_object_color(TreeTargetBox, tree_target_colour, FL_COL1);
	    fl_redraw_object(TreeTargetBox);
	}
    }
}

void tree_form_sync(void){
    if(!root) return;
    if(!source_is_stable()) return;
//    if(!TreeForm->visible) return;
    tree_default_actions();
	
    // search.
    tree_node_c *tmp_node = root->find_path(source_browser->get_cwd());
    if(!tmp_node){
	//cerr << "not found: " << source_browser->get_cwd() << "constructing ...\n";
	tmp_node = tree_construct_path(source_browser->get_cwd());
	if(!tmp_node){
	    tmp_node = root;
	}
    }
    tree_set_source(tmp_node);
    if(!tree_target_mode_flag){
	tree_set_current(tmp_node, FALSE, FALSE);   // external sync, don't update again.
    }

    tree_sync_to_target_browser();
    tree_set_complete_wintitle();
}

int do_tree_new_target(void){
    callback_table.restrict_all();
    //tree_unrestrict_cbs();
    tree_new_target_done_flag = FALSE;
    fl_deactivate_all_forms();
    int deactivated_count = TreeForm->deactivated;
    // redundant, but logically better.
    fl_activate_form_always(TreeForm);  
    do_tree(TRUE);
    while(!tree_new_target_done_flag){
	tool_mung_callbacks();
    }
    while(deactivated_count){
	fl_deactivate_form(TreeForm);
	deactivated_count--;
    }
    fl_activate_all_forms();
    callback_table.unrestrict_all();
    return (strlen(tree_new_target_dir) > 0);
}

void do_tree(int tree_target_mode_flag_iv){
    if(!source_is_stable()){
	return;
    }
    if(tree_target_mode_flag_iv){
	tree_target_mode_flag = TRUE;
	strcpy(tree_new_target_dir, "");
    }
    tree_target_mode_was_visible_flag = TreeForm->visible;
    fl_activate_form_always(TreeForm);
    
    if(!root){
	if(!tree_restore(FALSE)){
	    // build from scratch
	    const char *cwd_buf = source_browser->get_cwd();
	    char *root_dir = "/";
	    int max_levels = 1, max_nodes = 50;
	    root = new tree_node_c(root_dir);
	    if(!root) return;
	    char buffer[MAXPATHLEN+64];
	    sprintf(buffer, "Scanning `%s', max levels %d, max nodes %d", cwd_buf, max_levels, max_nodes);
	    
	    tree_node_c *cwd_node = tree_construct_path(source_browser->get_cwd());
	    if(!cwd_node){
		cwd_node = root;
	    }
	    
	    indicator_simple_bar(buffer);
	    indicator_check("0");
	    fl_freeze_form(TreeForm);
	    tree_scan(cwd_node, max_levels, max_nodes, FALSE);
	    tree_scanning_flag = FALSE;	    
	    fl_unfreeze_form(TreeForm);
	    indicator_end();
	}
	tree_retrunc();
	tree_form_sync();
    }
    if(tree_target_mode_flag){
	tree_node_c *cwd_node = tree_construct_path(target_browser->get_cwd());
	if(cwd_node){
	    tree_set_current(cwd_node, FALSE, FALSE);
	}
    }
    tree_show_form();
	//toolbox_attach();
    tree_set_complete_wintitle();
}

/* desired functionality:
 * if text field is hidden, erase it and show the key as first key.
 * search for the entry.
 * if text field is visible, get the text, append this key and search.
 * if search successful, display new text. Else, show status msg 'not
 * found'.
 * we DO support search again, so we DO need to remember the
 * position of the current node. Make set_current_fname return the
 * success of setting the line.
 * potentially freeze the form if show object and set_current both
 * redraw.
 */

/* wrap is possible if we found at least one pattern of this kind,
 * but failed on the last search. Additionally, allow it only if we
 * did not try extending first. 
 */
static int
    wrap_possible_flag   = FALSE,
    pattern_index = 0
;

static char tree_last_search_buf[NAME_MAX+1] = "";

void tree_search_event_cb(FL_OBJECT *, long){
    if(!cnode_is_stable()) return;
    if(bIsSearchMode){
	if(strlen(TreeFormText->label) <= 0){
	    pattern_index = -1;
	    // research for previous search.
	    fl_set_object_label(TreeFormText, tree_last_search_buf);
	} else {
	    if(pattern_index > 0 && wrap_possible_flag){
		// found it at least once, so we can wrap.
		pattern_index = -1;
	    }
	}
	tree_search_again_event_cb(NULL, 0);
	return;
    }
    pattern_index = 0;
    search_last_node = current_node;
    tree_search_set_mode();
}

void tree_search_again_event_cb(FL_OBJECT *, long){
    if(dbg_flag)cerr << "again called.\n";
    char *label_buf = TreeFormText->label, keybuf[NAME_MAX+1];
    if(!label_buf) return;
    if(strlen(label_buf) == 0) return;

    strcpy(keybuf, label_buf);

    pattern_index++;
    int dummy = 0;
    tree_node_c *tmp_node = root->find_pattern(keybuf, dummy, pattern_index);
    if(tmp_node){
	tree_set_current(tmp_node, FALSE, FALSE);
	status_msg(1, "`%s' matches directory `%s'.", keybuf, tmp_node->get_path());
	fl_set_object_label(TreeFormText, keybuf);
	wrap_possible_flag = FALSE;
    } else {
	XBell(fl_display, 0);
	status_msg(1, "no more matches for `%s'.", keybuf);
	wrap_possible_flag = TRUE;
    }
}

static int tree_search_key_cb(XKeyEvent *, KeySym keysym, char key){
    int erase_flag = FALSE;
    if(!bIsSearchMode || !root){
	return FALSE;
    }
    if(IsModifierKey(keysym)){
	return FALSE;
    }
    //cerr << "tree search: key is: " << key << endl;
    if(keysym == XK_KP_Enter || keysym == XK_Return){
	tree_search_clear_mode();
	return TRUE;
    }
    
    if(keysym == XK_BackSpace || keysym == XK_Delete
#ifdef XK_KP_Delete
	|| keysym == XK_KP_Delete
#endif
    ){
	erase_flag = TRUE;
    } else {
	if(!isprint(key)){
	    // return true so that it isn't forwarded to the xshell.
	    //cerr << "key is not printable.\n";
	    return TRUE;
	}
    }   
    char *label_buf = TreeFormText->label, keybuf[NAME_MAX+1];
    if(!label_buf) return FALSE;
    int len = strlen(label_buf);

    if(len == 0){
	// first character.
	sprintf(keybuf, "%c*", key);
    } else {
	// sophisticated iterated search.
	if(erase_flag){
	    if(len <= 1){
		return TRUE;
	    }
	    label_buf[len-2] = '\0';
	} else {
	    // eliminate the trailing '*' in the label. 
	    // overwrite it with the new character, and append a '*'. 
	    label_buf[len-1] = key;   
	}
	sprintf(keybuf, "%s*", label_buf);
    }
    int dummy = 0;
    tree_node_c *tmp_node = root->find_pattern(keybuf, dummy, 0);
    if(tmp_node){
	tree_set_current(tmp_node, FALSE, FALSE);
	status_msg(1, "`%s' matches directory `%s'.", keybuf, tmp_node->get_path());
	fl_set_object_label(TreeFormText, keybuf);
	wrap_possible_flag = FALSE;
	strcpy(tree_last_search_buf, keybuf);
    } else {
	XBell(fl_display, 0);
	status_msg(1, "no match for `%s'.", keybuf);
	wrap_possible_flag = TRUE;
    }
    return TRUE;
}

static void tree_search_clear_mode(int bRestoreLastNode) {
    if (!bIsSearchMode) return;
    bIsSearchMode = FALSE;

    fl_freeze_form(TreeForm);
    fl_set_object_label(TreeFormText, "");

    fl_set_object_color(TreeFormText, FL_COL1, FL_BLACK);
    fl_set_object_lcol(TreeFormText, FL_BLACK);

    if(bRestoreLastNode && search_last_node){
	tree_set_current(search_last_node,FALSE, FALSE);
    }
    fl_unfreeze_form(TreeForm);
}

static void tree_search_set_mode(void) {
    if (bIsSearchMode) return;
    bIsSearchMode = TRUE;
    fl_set_object_color(TreeFormText, FL_BLACK, FL_COL1);
    fl_set_object_lcol(TreeFormText, FL_WHITE);
}

/**** EXTERNAL DND ****/
const char *tree_xdnd_track(FL_Coord mx, FL_Coord my){
    xdnd_flag = TRUE;
    //cerr << "mx, my" << mx << "," << my << endl;
    const char *tmp_buf = tree_dnd_tracking(mx, my);
    xdnd_flag = FALSE;
    return tmp_buf;
}

/**** TREE MENU SUPPORT ****/
static void fsize_cb(int size){
    if(size > 0){
	tree_font_size = size;
	tree_retrunc();
    }
}

static void fstyle_cb(int style){
    if(style > 0){
	tree_font_style = style;
	//cerr << "Can't change fontstyle yet.\n";
	tree_retrunc();
    }
}

static void nlen_cb(int len){
    tree_name_length = len;
    tree_retrunc();
}

static void tree_verify_root_cb(int){
    if(ask_question(FALSE, "Verification takes 20-30 secs/Gigabyte. Continue ?")){
	tree_restat(root, TRUE);
    }
}

static void tree_verify_subtree_cb(int){
    if(!cnode_is_stable()) return;
    if(ask_question(FALSE, "Verification takes 20-30 secs/Gigabyte. Continue ?")){
	tree_restat(current_node, TRUE);
    }
}

static void tree_expand_subtree_1_cb(int){
    tree_expand(TRUE);
}

static void tree_expand_subtree_all_cb(int){
    if(!cnode_is_stable()) return;
    if(ask_question(FALSE, "Rescanning all subdirectories takes 2-3 minutes/Gigabyte. Continue ?")){
	fl_freeze_form(TreeForm);
	tree_restat(current_node, TRUE);
	tree_scan(current_node, SCAN_ALL_LEVELS, -1, TRUE);
	tree_reshape();
	fl_unfreeze_form(TreeForm);
    }
}

static void tree_reload_db_cb(int){
    if(ask_question(FALSE, "Reload the tree database from disc ?")){
	if(!tree_restore()){
	    timed_error_notifier("Reloading failed.", notifier_secs);
	}
    }
}

static void tree_save_db_cb(int){
    status_msg(1, "Saving tree database ...");
    tree_dump();
    status_msg(1, "database saved.");
}

static void tree_collapse_subtree_1_cb(int){
    fl_freeze_form(TreeForm);
    tree_collapse();
    fl_unfreeze_form(TreeForm);
}

static void tree_collapse_subtree_all_cb(int){
    if(!cnode_is_stable()) return;
    if(current_node == root){
	if(!ask_question(TRUE, "This will collapse the whole tree. Continue ?")) return;
    }
    fl_freeze_form(TreeForm);
    if(!current_node->child0 && current_node != root){
	// remove even the current node if it hasn't children.
	tree_collapse();
    } else {
	// remove all children, but not the current node.
	while(current_node->child0){
	    tree_collapse();
	}
    }
    /*
    if(current_node != root){
	tree_collapse();
    }
    */
    fl_unfreeze_form(TreeForm);
}

static void do_help(int){
    tree_form_help_func();
}

static void do_ghelp(int){
    info_show_form();
}

static void do_search(int){
    tree_search_event_cb(NULL, 0); 
}

static void do_set_target(int){
    tree_set_target_cb(NULL, 0);
}

static void do_set_source(int){
    TreeEnterKeyCB(NULL, 0);
}

static void do_tree_colours(int){
    tree_colour_show_form();
}

static void tree_packed_cb(int val){
    if(val != tree_packing_level){
	tree_packing_level = val;
	status_msg(2, "packing switched to %d", val); 
	tree_retrunc();
    } else {
	status_msg(2, "tree packing: no change.");
    }
}

static void tree_hide_cb(int){
    tree_close_form();
}

void TreeFormFileButtonCB(FL_OBJECT *, long bFocus){
    static long
	db_menu   = 0
    ;
    long file_menu = 0;

    if (db_menu == 0) {
	db_menu = new_pup();
	if (!db_menu) return;
	add_to_pup(db_menu, "Database%t", 0);
	add_to_pup(db_menu, "Load%f",     (long) tree_reload_db_cb, "Ll");
	add_to_pup(db_menu, "Save%f",     (long) tree_save_db_cb,   "Ss");
    }

    file_menu = new_pup();
    if (!file_menu) return;
    add_to_pup(file_menu, "File (Alt-f, Esc-f)%t", 0);

    if (!tree_target_mode_flag) {
	add_to_pup(file_menu, "DataBase%m", db_menu, "Bb");
	long dir_menu = get_dir_menu();
	if(dir_menu>0){
	    add_to_pup(file_menu, "Directory (Alt-d, Esc-d)%m", dir_menu, "Dd");
	}
    }
    add_to_pup(file_menu, "set Source (Enter)%f",  (long) do_set_source, "Ss");
    add_to_pup(file_menu, "set Target (Ctrl-t)%f", (long) do_set_target, "Tt");
    add_to_pup(file_menu, "Quit (Alt-q, Esc-q)%f", (long) tree_hide_cb,  "Qq");

    if (bFocus) {
	tool_focus_move(TreeForm, TreeFormFileButton);
    }

    do_pup(file_menu);
    free_pup(file_menu);
}

void TreeFormEditButtonCB(FL_OBJECT *, long bFocus){
    static long 
	collapse_menu = 0, 
	edit_menu     = 0,
	expand_menu   = 0,
	verify_menu   = 0
    ;

    if (collapse_menu == 0) {
	collapse_menu = new_pup();
	if (!collapse_menu) return;
	add_to_pup(collapse_menu, "Collapse (Mouse-3)%t", 0);
	add_to_pup(collapse_menu, "1 Sublevel      (keypad -, Alt -, Esc -)%f",      (long) tree_collapse_subtree_1_cb,   "1");
	add_to_pup(collapse_menu, "All Sublevels (keypad Ctrl -)%f", (long) tree_collapse_subtree_all_cb, "Aa");
    }
    if (expand_menu == 0) {
	expand_menu = new_pup();
	if (!expand_menu) return;
	add_to_pup(expand_menu, "Expand (Mouse-2)%t", 0);
	add_to_pup(expand_menu, "1 Sublevel      (keypad +, Alt +, Esc +)%f",      (long) tree_expand_subtree_1_cb,   "1");
	add_to_pup(expand_menu, "All Sublevels (keypad Ctrl +)%f", (long) tree_expand_subtree_all_cb, "Aa");
    }

    if (verify_menu == 0) {
	verify_menu = new_pup();
	if (!verify_menu) return;
	add_to_pup(verify_menu, "Verify%t", 0);
	add_to_pup(verify_menu, "All Sublevels%f", (long) tree_verify_subtree_cb, "Aa");
	add_to_pup(verify_menu, "Everything%f",    (long) tree_verify_root_cb,    "Ee");
    }

    if (edit_menu == 0) {
	edit_menu = new_pup();
	if (!edit_menu) return;
	add_to_pup(edit_menu, "Edit (Alt-e, Esc-e)%t",   0);
	add_to_pup(edit_menu, "Collapse %m",             collapse_menu,  "Cc");
	add_to_pup(edit_menu, "Expand %m",               expand_menu,    "Ee");
	add_to_pup(edit_menu, "Verify %m",               verify_menu,    "Vv");
	add_to_pup(edit_menu, "Search (Ctrl-s)%f",       (long) do_search,     "Ss");
	add_to_pup(edit_menu, "Help (F1)%f",	         (long) do_help,       "Hh");
	add_to_pup(edit_menu, "Global Help (Meta-F1)%f", (long) do_ghelp,      "lL");
    }
    if (bFocus) {
	tool_focus_move(TreeForm, TreeFormEditButton);
    }
    do_pup(edit_menu);
}

static void do_global(int){
    OptionsButtonCB(NULL,0);   
}

void TreeFormOptionsButtonCB(FL_OBJECT *, long bFocus){
    static long 
	fsize_menu    = make_fsize_menu((long) fsize_cb),
	fstyle_menu   = make_fstyle_menu((long) fstyle_cb),
	nlen_menu     = 0, 
	packed_menu   = 0
    ; 
    long 
	options_menu  = 0,
	windows_menu  = get_windows_menu()
    ;
    if (!fsize_menu || !fstyle_menu | !windows_menu) {
	cerr << "Tree: Can't make menus.\n";
	return;
    }
    if (tree_target_mode_flag) {
	timed_notifier("No tree options are available in Select New Target mode.");
	return;
    }

    if (nlen_menu == 0) {
	nlen_menu = new_pup();
	if (!nlen_menu) return;
	add_to_pup(nlen_menu, "Length%t", 0);
	add_to_pup(nlen_menu, "4%x4",   (long) nlen_cb, "4");
	add_to_pup(nlen_menu, "5%x5",   (long) nlen_cb, "5");
	add_to_pup(nlen_menu, "6%x6",   (long) nlen_cb, "6");
	add_to_pup(nlen_menu, "7%x7",   (long) nlen_cb, "7");
	add_to_pup(nlen_menu, "8%x8",   (long) nlen_cb, "8");
	add_to_pup(nlen_menu, "10%x10", (long) nlen_cb, "1");
	add_to_pup(nlen_menu, "12%x12", (long) nlen_cb, "2");
	add_to_pup(nlen_menu, "15%x15", (long) nlen_cb, "");
	add_to_pup(nlen_menu, "20%x20", (long) nlen_cb, "0");
	add_to_pup(nlen_menu, "30%x30", (long) nlen_cb, "3");
    }
    
    if (packed_menu == 0) {
	packed_menu = new_pup();
	if (!packed_menu) return;
	add_to_pup(packed_menu, "Packing%t", 0);
	add_to_pup(packed_menu, "Narrow%x2", (long) tree_packed_cb, "Nn");
	add_to_pup(packed_menu, "Medium%x1", (long) tree_packed_cb, "Mm");
	add_to_pup(packed_menu, "Wide%x0",   (long) tree_packed_cb, "Ww");
    }
    
    options_menu = new_pup();
    if (!options_menu) return;
    add_to_pup(options_menu, "Options (Alt-o, Esc-o, F2)%t", 0);
    add_to_pup(options_menu, "Font Size %m",                 fsize_menu,    "zZ");
    add_to_pup(options_menu, "Font Style %m",                fstyle_menu,   "tT");
    add_to_pup(options_menu, "Name Length %m",               nlen_menu,     "Ll");
    add_to_pup(options_menu, "Packed Display %m",            packed_menu,   "Pp");
    add_to_pup(options_menu, "Windows (Alt-w, Esc-w%m",      windows_menu,  "Ww");
    add_to_pup(options_menu, "Colours%x0",                   (long) do_tree_colours, "Cc");
    add_to_pup(options_menu, "Global Options (Meta-F2) %f",  (long) do_global,  "Gg");

    if (bFocus) {
	tool_focus_move(TreeForm, TreeFormOptionsButton);
    }

    do_pup(options_menu);
    free_pup(options_menu);
}

static void functions_cb(int val){
    switch(val){
    case 1:  help_event_cb(tree_form_help_func); break;
    case 2:  TreeFormOptionsButtonCB(NULL, 0);   break;
    case 3:  tree_mkdir();          break;
    case 4:  tree_raise_form();     break;
    case 5:  tree_view();           break;
    case 6:  tree_edit();           break;
    case 7:  tree_rename();         break;
    case 8:  tree_move();           break;
    case 9:  tree_duplicate();      break;
    case 10: tree_copy();           break;
    case 11: tree_delete();         break;
    case 12: ExitButtonCB(NULL, 0); break;
    default: error_msg("functions_cb: Invalid value %d.", val);
    }
}

void TreeFormActionsButtonCB(FL_OBJECT *, long bFocus){
    static long	functions_menu = 0;
    if (tree_target_mode_flag) {
	timed_notifier("No tree actions are available in Select New Target mode.");
	return;
    }
    if (functions_menu == 0) {
	functions_menu = new_pup();
	if (functions_menu <= 0) return;
	add_to_pup(functions_menu, "Action Keys (Alt-a, Esc-a)%t", 0);
	add_to_pup(functions_menu, "F1  Help%x1",              (long) functions_cb, "Hh");
	add_to_pup(functions_menu, "F2  Options%x2",           (long) functions_cb, "Oo");
	add_to_pup(functions_menu, "F3  Make Dir%x3",          (long) functions_cb, "kK");
	add_to_pup(functions_menu, "F4  Raise Tree Window%x4", (long) functions_cb, "Ww");
	add_to_pup(functions_menu, "F5  Dir Info%x5",          (long) functions_cb, "Ii");
	add_to_pup(functions_menu, "F6  Source Browser%x6",    (long) functions_cb, "Ss");
	add_to_pup(functions_menu, "F7  Rename Dir %x7",       (long) functions_cb, "Rr");
	add_to_pup(functions_menu, "F8  Move Dir %x8",         (long) functions_cb, "Mm");
	add_to_pup(functions_menu, "F9  Duplicate Dir %x9",    (long) functions_cb, "pP");
	add_to_pup(functions_menu, "F10 Copy Dir %x10",       (long) functions_cb, "Cc");
	add_to_pup(functions_menu, "F11 Delete Dir %x11",     (long) functions_cb, "Dd");
	add_to_pup(functions_menu, "F12 Exit UC %x12",        (long) functions_cb, "xX");
    }
    if (bFocus) {
	tool_focus_move(TreeForm, TreeFormActionsButton);
    }

    do_pup(functions_menu);
}


/**** FUNCTION KEY OPERATIONS ****/

/* local to current node, not source dir. ignore ctrl for a while */

void tree_mkdir(void){
    if(!cnode_is_stable()) return;
	if(!TreeForm || !TreeForm->visible) return;
    const char *path = current_node->get_path();
    if(!path) return;
    char buf[2*MAXPATHLEN];
    sprintf(buf, "%s/", path);
    const char *result = OKCancel("Tree Make Directory", buf, tree_mkdir_help_text, "Tree Make Directory");
    if(!result || !strlen(result)) return;
    if(result[0] != '/'){
	// correct for relative pathnames.
	sprintf(buf, "%s/%s", path, result);
    } else {
	strcpy(buf, result);
    }
    if(mkdir(buf, mkdir_mode_mask) != 0){
	timed_strerror_notifier("Can't make directory:");
	return;
    }
    tree_construct_path(buf);

    path = current_node->get_path();  // no aliasing, please.
    if(!path) return;                 // supersafe code.
    // we miss a half-correct source fname adjust here. May not be necessary,
    // the browsers should sense it anyway .. but the adjusting of cfn. Hm.
    tool_update_matching_dirs(path);  
}

void tree_view(void){
    if(!cnode_is_stable()) return;
    dirinfo(current_node->get_path());
}

void tree_edit(void){
    if(!cnode_is_stable()) return;
    if(!source_browser) return;
    source_browser->freeze();
    tree_change_source_to_current();
    tool_set_source(source_browser, TRUE);
    source_browser->win_raise();
    source_browser->unfreeze();
}

static char tree_action_pathbuf[MAXPATHLEN];

static int tree_move_path(const char *source_path, const char *target_dir, const char *new_name);

static int tree_action_check(const char *action){
    if(!action || !cnode_is_stable() || !source_browser || !target_browser) return FALSE;
    if(current_node == root){
	char buf[MAXPATHLEN];
	sprintf(buf, "Cannot %s root directory.", action);
	timed_notifier(buf);
	return FALSE;
    }
    const char *path = current_node->get_path();
    if(!path || !get_basename(path)) return FALSE;

    strcpy(tree_action_pathbuf, path);
    return TRUE;
}

static void tree_renamemove_action(int move_flag){
    char target_buf[2*MAXPATHLEN]; //, pathbuf[MAXPATHLEN];

    const char *result;
    if(move_flag){
	sprintf(target_buf, "Tree move directory `%s' to:", tree_action_pathbuf);
	result = OKCancel(target_buf, tree_action_pathbuf, tree_move_help_text, "Tree Move Directory");
    } else {
	sprintf(target_buf, "Tree Rename Directory `%s' to:", get_basename(tree_action_pathbuf));
        result = OKCancel(target_buf, get_basename(tree_action_pathbuf), tree_rename_help_text, "Tree Rename Directory");
    }

    if(!result || !strlen(result)) return;
    if(result[0] == '/'){
	// abs pathname. Use it.
	strcpy(target_buf, result);
    } else {
	// make a valid absolute path.
	sprintf(target_buf, "%s/%s", get_dirname(tree_action_pathbuf), result);  
    }
    strcpy(target_buf, strip_trailing_slashes(target_buf));

    int new_name_flag = FALSE;
    struct stat statbuf;
    if(stat(target_buf, &statbuf) == 0){
	// is it a directory ?
	if(!S_ISDIR(statbuf.st_mode)){
	    // no, and don't remove target file. (cp would fail anyway)
	    timed_notifier((move_flag ? "Sorry, can't move: target file exists.":"Sorry, can't rename: target file exists."));
	    return;
	}
	// extra check:
	char buf[MAXPATHLEN];
	strcpy(buf, target_buf);
	sprintf(target_buf, "%s/%s", buf, get_basename(tree_action_pathbuf));
	if(stat(target_buf, &statbuf) == 0){
	    // it's a dir, but we exist nevertheless even then !
	    timed_notifier("Sorry, target file exists. ");
	    return;
	} else {
	    // check recursivity !
	    if(dir_is_child(target_buf, tree_action_pathbuf)){
		timed_notifier("Sorry, can't move/rename to self: target is child, must avoid infinite loop.");
		return ;
	    }
	}
    } else {
	new_name_flag = TRUE;
    }
    
    if(rename(tree_action_pathbuf, target_buf) != 0){
	timed_strerror_notifier("Tree Rename failed:");
	return;
    }

    tree_move_path(tree_action_pathbuf, get_dirname(target_buf), new_name_flag ? get_basename(target_buf) : (char *) NULL);
}

static int tree_move_path(const char *source_path, const char *target_dir, const char *new_name){
    // abs path's only, will be realpathed. If !new_name, no rename.
    // Assume new dir+name are already real. Only tree+browser consistency maintenance.
    if(!source_path || !target_dir || source_path[0] != '/' || target_dir[0] != '/'){
	M_err("tree_move_path", "invalid ptrs");
	return FALSE;
    }
    char source_rpath[MAXPATHLEN], target_rdir[MAXPATHLEN];
    if(!real_path(source_path, source_rpath) || !real_path(target_dir,target_rdir)){
	M_err("tree_move_path", "realpath failed.");
	return FALSE;
    }

    // local source node
    tree_node_c *lsource_node = root->find_path(source_rpath);
    if(!lsource_node) return FALSE;
    if(new_name){
	if(lsource_node->name) fl_free(lsource_node->name);
	lsource_node->name = strdup(new_name);
	lsource_node->retrunc(FALSE);
    }

    tree_node_c *ltarget_node = tree_construct_path(target_dir, FALSE);
    if(!ltarget_node) return FALSE;

    if(lsource_node->parent != ltarget_node){
	fl_freeze_form(TreeForm);
	lsource_node->reconnect(ltarget_node);
	tree_reshape();
	fl_unfreeze_form(TreeForm);
    }
   
    tree_set_complete_wintitle();

    // also force update of cwd buf in all browsers. Notice path component
    // complication. check wether browser cwd contains pathbuf. If yes,
    // assemble new dir and chdir ? Or rather, set the cwdbuf directly ?
    // the latter seems a bit better .. but also the shell would need a
    // cd command -> change all non-source browsers directly (is not necessarily
    // another way anyway), and change source browser via toolchdir ? Nope,
    // tool changedir will fail ! Attempt safe_chdir to the one first.
    
    browser_c *tmp_browser = source_browser;
    do {
	const char *cwd = tmp_browser->get_cwd();
	if(cwd && strlen(cwd) && strstr(cwd, source_rpath)){
	    //cerr << "cwd: " << cwd << ", path: " << source_rpath << endl;
	    char cwd_buf[MAXPATHLEN];
	    strcpy(cwd_buf, target_rdir);
	    strcat(cwd_buf, "/");
	    if(new_name){
		strcat(cwd_buf, new_name);
	    } else {
		strcat(cwd_buf, get_basename(source_rpath));
	    }
	    strcat(cwd_buf, cwd + strlen(source_rpath));
	    //cerr << cwd_buf << endl;
	    // avoid one browser update by not updating browser bufs,
	    // and by doing the chdir here.
	    // 
	    tmp_browser->set_cwd_buf(cwd_buf);
	    tmp_browser->set_complete_wintitle();
	    if(tmp_browser == source_browser){
		// do the chdir, and update shell.
		safe_chdir(cwd_buf);
		xshell_cd(cwd_buf);
		xshell_sync_title();
	    }
	}
	tmp_browser = tmp_browser->next_browser;
    } while(tmp_browser != source_browser);
    return TRUE;
}

void tree_rename(void){
    if(!tree_action_check("rename")) return;
    tree_renamemove_action(FALSE);
}

void tree_move(void){
    if(!tree_action_check("move")) return;
    tree_renamemove_action(TRUE);
}

static int tree_delete_action(void){
    char buf[MAXPATHLEN], buf2[MAXPATHLEN];
    struct stat statbuf;
    if(lstat(tree_action_pathbuf, &statbuf) != 0) return FALSE;

    // ask specially for non-empty or wp directories.
    int ask_flag = FALSE;
    if(!S_ISLNK(statbuf.st_mode)){
	sprintf(buf, "Directory `%s' is ", get_basename(tree_action_pathbuf));
	int n_files_in_dir = files_in_dir(tree_action_pathbuf);
	if(n_files_in_dir > 2){
	    sprintf(buf2, "a non-empty directory (%d files) ", n_files_in_dir);
	    strcat(buf, buf2);
	    ask_flag = TRUE;
	}
	if(!try_access(&statbuf, W_OK)){
	    if(ask_flag) strcat(buf, "and is ");
	    ask_flag = TRUE;
	    strcat(buf, "write protected");
	}
	if(ask_flag){
	    strcat(buf, ". Force tree delete ?");
	    if(!ask_question(TRUE, buf)) return FALSE;
	}
    }

    // reset current node to something meaningful
    tree_node_c *delete_node = current_node;
    if(current_node->prev){
	tree_set_current(current_node->prev);
    } else if(current_node->next){
	tree_set_current(current_node->next);
    } else {
	tree_set_current(current_node->parent);
    }
    if(current_node == delete_node) return FALSE;  // something went very wrong
    
    // salvage browser & shell dirs. Use tool change dir this time,
    // because we're attempting this before bad things happen.
    browser_c *tmp_browser = source_browser;
    do {
	const char *cwd = tmp_browser->get_cwd();
	if(cwd && strlen(cwd) && strstr(cwd, tree_action_pathbuf)){
	    if(tmp_browser == source_browser){
		tree_change_source_to_current();
	    } else if(tmp_browser == target_browser){
                // extra check s!=t not necessary due to aliasing.
		tree_change_target_to_current();
	    } else {
		const char *path = current_node->get_path();
		if(!path) path = "/";
		tmp_browser->set_cwd_buf(path);
		tmp_browser->set_complete_wintitle();
	    }
	}
	tmp_browser = tmp_browser->next_browser;
    } while(tmp_browser != source_browser);

    // cerr << "would now force-delete: " << tree_action_pathbuf << endl;
    int td_flag = FALSE;
    fl_freeze_form(TreeForm); // freeze over the restat.
    if(delete_dir(tree_action_pathbuf)){
	td_flag = TRUE;
    } else {
	// cerr << "delete dir failed on " << tree_action_pathbuf << endl;
	td_flag = !delete_node->restat(TRUE, TRUE);
    }
    if(td_flag){
	tree_remove_invalid_node(delete_node);
    }
    fl_unfreeze_form(TreeForm);
    return TRUE;
}

static int duplicopy_flag = FALSE;

static int tree_duplicopy_action(void){
    char target_buf[MAXPATHLEN];
    struct stat statbuf;
    if(stat(tree_action_pathbuf, &statbuf) != 0) return FALSE;
    const char *result;
    if(duplicopy_flag){
	sprintf(target_buf, "Tree Copying `%s' to:", tree_action_pathbuf);
	result = OKCancel(target_buf, tree_action_pathbuf, tree_copy_help_text, "Tree Copy");
    } else {
	sprintf(target_buf, "Tree Duplicating `%s' to:", get_basename(tree_action_pathbuf));
	result = OKCancel(target_buf, get_basename(tree_action_pathbuf), tree_duplicate_help_text, "Tree Duplicate");
    }
    if(!result || (strcmp(result, tree_action_pathbuf) == 0)) return FALSE;
    if(result[0] == '/'){
	// abs pathname. Use it.
	strcpy(target_buf, result);
    } else {
	// make a valid absolute path.
	sprintf(target_buf, "%s/%s", get_dirname(tree_action_pathbuf), result);  
    }
    
    if(stat(target_buf, &statbuf) == 0){
	// is it a directory ?
	if(!S_ISDIR(statbuf.st_mode)){
	    // no, and don't remove target file. (cp would fail anyway)
	    timed_notifier("Sorry, target file exists. ");
	    return FALSE;
	}
	char buf[MAXPATHLEN];
	strcpy(buf, target_buf);
	sprintf(target_buf, "%s/%s", buf, get_basename(tree_action_pathbuf));
	if(stat(target_buf, &statbuf) == 0){
	    // it's a dir, but we exist nevertheless even then !
	    timed_notifier("Sorry, target file exists. ");
	    return FALSE;
	} else {
	    // check recursivity !
	    if(dir_is_child(buf, tree_action_pathbuf)){
		timed_notifier("Sorry, can't copy to self: target is child, must avoid infinite loop.");
		return FALSE;
	    }
	}
    }   // else hope it's a good path !

    if(do_fork_duplicate(tree_action_pathbuf, target_buf)){
	tree_construct_path(target_buf);
	return TRUE;
    } else {
	// partial duplicopy ?
	if(stat(target_buf, &statbuf) == 0){
	    tree_construct_path(target_buf);
	}
	return FALSE;
    }
}

void tree_delete(void){
    if(!tree_action_check("delete")) return;
    char buf[MAXPATHLEN], buf2[MAXPATHLEN];
    sprintf(buf, "Tree Delete directory `%s' ?", tree_action_pathbuf);
    sprintf(buf2, "Tree Delete %s", get_basename(tree_action_pathbuf));
    fork_action(tree_delete_action, buf, NULL, buf2, TRUE);
}

void tree_copy(void){
    if(!tree_action_check("copy")) return;
    char buf[MAXPATHLEN], buf2[MAXPATHLEN];
    sprintf(buf, "Tree Copy directory `%s' ?", tree_action_pathbuf);
    sprintf(buf2, "Tree Copy %s", get_basename(tree_action_pathbuf));
    duplicopy_flag = TRUE;
    fork_action(tree_duplicopy_action, buf, NULL, buf2);
}

void tree_duplicate(void){
    if(!tree_action_check("duplicate")) return;
    char buf[MAXPATHLEN], buf2[MAXPATHLEN];
    sprintf(buf, "Tree Duplicate directory `%s' ?", tree_action_pathbuf);
    sprintf(buf2, "Tree Duplicate %s", get_basename(tree_action_pathbuf));
    duplicopy_flag = FALSE;
    fork_action(tree_duplicopy_action, buf, NULL, buf2);
}

