/*
    utility.C:
    Define some always used functions.
    written 1991 by Henrik A. Klagges
    see file copyright.txt about copying policy
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>  // for HP
#include "utility.h"

extern "C" {
    double drand48();
    long lrand48();
    void srand48(long seedval);
}

/*
 *	Array handling: copying.
 *	Null-Pointer dest's are newed.
 */

void copy_array(
    double *dest,
    const double *src,
    int size
){
    if(dest == NULL){
	error_exit("Can't copy_array (double) into null-pointer.\n");
    }
    for(int i=0; i<size; i++){
	dest[i] = src[i];
    }
}

void copy_array(
    float *dest,
    const float *src,
    int size
){
    if(dest == NULL){
	error_exit("Can't copy_array (float) into null-pointer.\n");
    }
    for(int i=0; i<size; i++){
	dest[i] = src[i];
    }
}
void copy_array(
    int *dest,
    const int *src,
    int size
){
    if(dest == NULL){
	error_exit("Can't copy_array (int) into null-pointer.\n");
    }
    for(int i=0; i<size; i++){
	dest[i] = src[i];
    }
}

/*
 *	Array handling: return a max or min.
 */

double max_of_array(
    const double *array,
    int size
){
    double tmp_double = 0.0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_double = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] > tmp_double){
		tmp_double = array[i];
	    }
	}
    }
    return tmp_double;
}

double min_of_array(
    const double *array,
    int size
){
    double tmp_double = 0.0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_double = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] < tmp_double){
		tmp_double = array[i];
	    }
	}
    }
    return tmp_double;
}

float max_of_array(
    const float *array,
    int size
){
    float tmp_float = 0.0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_float = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] > tmp_float){
		tmp_float = array[i];
	    }
	}
    }
    return tmp_float;
}

float min_of_array(
    const float *array,
    int size
){
    float tmp_float = 0.0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_float = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] < tmp_float){
		tmp_float = array[i];
	    }
	}
    }
    return tmp_float;
}

int max_of_array(
    const int *array,
    int size
){
    int tmp_int = 0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_int = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] > tmp_int){
		tmp_int = array[i];
	    }
	}
    }
    return tmp_int;
}

int min_of_array(
    const int *array,
    int size
){
    int tmp_int = 0;
    if(size <=0){
	error_exit("Cannot max of an %d array.\n", size);
    } else {
	tmp_int = array[0];
	for(int i=1; i<size; i++){
	    if(array[i] < tmp_int){
		tmp_int = array[i];
	    }
	}
    }
    return tmp_int;
}

/*	Array handling: setting to a value	*/

void set_array(double *array, int size, double value){
    for(int i=0; i<size; i++){
	array[i] = value;
    }
}

void set_array(float *array, int size, float value){
    for(int i=0; i<size; i++){
	array[i] = value;
    }
}

void set_array(int *array, int size, int value){
    for(int i=0; i<size; i++){
	array[i] = value;
    }
}

void set_array(unsigned short *array, int size, unsigned short value){
    for(int i=0; i<size; i++){
	array[i] = value;
    }
}

/*
 *	Array handling: standard stream output.
 */
 
void dump_array(
    const int *array,
    int size,
    int n_columns
){
    int i;
    for(i=0; i<size; i++){
	if(i%n_columns == 0){
	    cerr << "\t";
	}
	cerr << "(" << i << ") " << array[i] << "\t";
	if(i%n_columns == n_columns-1){
	    cerr << "\n";
	}
    }
    if(i%n_columns != n_columns-1){
	cerr << "\n";
    }	// if not already, push a trailing newline.
}

void dump_array(
    const double *array,
    int size,
    int n_columns
){
    int i;
    for(i=0; i<size; i++){
	if(i%n_columns == 0){
	    cerr << "\t";
	}
	cerr << "(" << i << ") " << array[i] << "\t";
	if(i%n_columns == n_columns-1){
	    cerr << "\n";
	}
    }
    if(i%n_columns != n_columns-1){
	cerr << "\n";
    }	// if not already, push a trailing newline.
}

void dump_array(
    const float *array,
    int size,
    int n_columns
){
    int i;
    for(i=0; i<size; i++){
	if(i%n_columns == 0){
	    cerr << "\t";
	}
	cerr << "(" << i << ") " << array[i] << "\t";
	if(i%n_columns == n_columns-1){
	    cerr << "\n";
	}
    }
    if(i%n_columns != n_columns-1){
	cerr << "\n";
    }	// if not already, push a trailing newline.
}

/* 
 * 	Standard stream input: 
 */
 
int ask_ok(const char* question){
//    static char c;
    cerr << question << " y/n: ";
/*
  cin >> c;
  switch(c){
  case 'y'    : return(1); 
  case 'Y'    : return(1); 
  case 'j'    : return(1); 
  case 'J'    : return(1); 
  case '\n'    : return(1); 
  }
  */
    return(0);
}

void ask_confirmation(const char* message){
    static char c;
    cerr << message;
    cin >> c;
}

void get_continue(void){
    char c;
    cin.get(c);
}

/*
 *	Standard stream output:
 */
 
int w_manip_length = 14;
/*
ostream &w_manip(ostream &o){
	o.width(w_manip_length);
	return(o);
}
*/

/*
 *	Error handling:
 *	NEVER rewrite this with //cerr ! I couldn't guarantee its initiali-
 *	zation if a call to this function is made during static initiali-
 *	zation.
 */
 
void error_exit(
    char *format_string, 
    ...
){
    va_list arg_pointer; 							
    va_start(arg_pointer , format_string); 			
    fprintf(stderr, "Fatal error: ");				
    vfprintf(stderr, format_string, arg_pointer);	
    va_end(arg_pointer);
    exit(-1);
}

void error_msg(
    char *format_string, 
    ...
){
    va_list arg_pointer; 							
    va_start(arg_pointer , format_string); 			
    fprintf(stderr, "Error: ");				
    vfprintf(stderr, format_string, arg_pointer);	
    fprintf(stderr, "\n");
    va_end(arg_pointer);
}

void not_implemented(
    char *format_string, 
    ...
){
    va_list arg_pointer; 							
    va_start(arg_pointer , format_string); 			
    fprintf(stderr, "Sorry, not implemented: ");				
    vfprintf(stderr, format_string, arg_pointer);	
    va_end(arg_pointer);
    exit(-1);
}
/*
 *	number manipulators:
 */
 
float bound_value(
    float variable, 
    float lower, 
    float upper
){
    if(variable > upper){
        return(upper);
    } else if(variable < lower){
        return(lower);
    } else {
        return(variable);
    }
}

int bound_value(
    int variable, 
    int lower, 
    int upper
){
    if(variable > upper){
        return(upper);
    } else if(variable < lower){
        return(lower);
    } else {
        return(variable);
    }
} 

int entier(float value){	
    static int tmp_int;
    tmp_int = (int) value;
    if( ((float) tmp_int) > value ){
	tmp_int--;
    }	// very very ugly
    return(tmp_int);
}	// end entier 

int odd(int value){			
    return(value % 2);
}	// end odd
 
/*
 *	Random number support (autoinitializing seeds !)
 */
static int 
    drand48_initialized = 0		// guarantee random initialization.
    ;
#define	DRAND48_INIT_MACRO 			\
if(drand48_initialized == 0){	\
				    srand48(getpid());			\
									    drand48_initialized = 1;	\
													    }

int random_bit(void){
    DRAND48_INIT_MACRO;
    return((int) (lrand48() % 2) );
}	// end random_bit

int random_int(int range){	
    DRAND48_INIT_MACRO;
    return((int) (lrand48() % range) );
}	// end random_int

float random_float(float range){
    DRAND48_INIT_MACRO;
    return(range * random_float_01());
}	// end random_float

float random_float_01(void){
    DRAND48_INIT_MACRO;
    return((float) drand48());
}	// end random_float_01

float random_symmetric_range(float range){    // return one out of [-range, range[
    DRAND48_INIT_MACRO;
    return(((drand48() *2.0)-1.0) * range);
}

double random_float(){
    DRAND48_INIT_MACRO;
    return(drand48());
}	// end random_float

long msecs_diff(struct timeval snd_time, struct timeval fst_time){
    return (
	(snd_time.tv_sec  - fst_time.tv_sec)  * 1000 +
	(snd_time.tv_usec - fst_time.tv_usec) / 1000
    );
}
