/*
    utility.h:
    Define some always used functions.
    written 1991 by Henrik A. Klagges
    see file copyright.txt about copying policy
    this would be nicer with templates (1993).
*/

#ifndef UTILITY_H
#define UTILITY_H

#include <math.h>       // drand48() & lrand48 functions 
#include <unistd.h>		// getpid to seed rand
#include <stdarg.h>		// stdarg AFTER stdio: vfprintf depends	
#include <sys/time.h>
#include <sys/types.h>	// getpid

/*
 *	Array handling: copying.
 *	Null-Pointer dest's raise an error_exit.
 */
void 
	copy_array(double *dest, const double	*src, int size),
	copy_array(float  *dest, const float	*src, int size),
	copy_array(int 	  *dest, const int 	*src, int size)
;

/*	Array handling: return a max or min. */
double
	max_of_array(const double *array, int size),
	min_of_array(const double *array, int size)
;
float
	max_of_array(const float *array, int size),
	min_of_array(const float *array, int size)
;
int
	max_of_array(const int *array, int size),
	min_of_array(const int *array, int size)
;

/*	Array handling: setting to a value	*/
void
	set_array(double 		 *array, int size, double 		  value),
	set_array(float  		 *array, int size, float  		  value),
	set_array(int    		 *array, int size, int    		  value),
	set_array(unsigned short *array, int size, unsigned short value)
;

/*	Array handling: standard stream output. */ 
void
	dump_array(const double	*array, int size, int n_columns),
	dump_array(const float 	*array, int size, int n_columns),
	dump_array(const int 	*array, int size, int n_columns)
;

/* 
 * 	Standard stream input: 
 */
 
int ask_ok				(const char* question);
void 	
	ask_confirmation	(const char* message),
	get_continue		(void)
;

/*	Standard stream output:	*/
 
extern int w_manip_length;		// a manipulator to be inserted into ostreams.
//ostream &w_manip(ostream &o);	// it sets the width to w_manip_length.

/*	Error handling:	*/
 
void 
    error_exit		(char *format_string, ...),
    error_msg		(char *format_string, ...),
    not_implemented	(char *format_string, ...)
;

/*	number manipulators:	*/

inline float min(float a, float b){
	if(a>=b){
		return(b);
	} else {
		return(a);
	}
}

inline float max(float a, float b){
	if(a>=b){
		return(a);
	} else {
		return(b);
	}
}

inline int min(int a, int b){
	if(a>=b){
		return(b);
	} else {
		return(a);
	}
}

inline int max(int a, int b){
	if(a>=b){return(a);} else {return(b);}
}

inline double signum(double a){
	if(a>0.0){return(1.0);} else if (a<0.0){return(-1.0);} else {return(0.0);}
}

inline float signum(float a){
	if(a>0.0){return(1.0);} else if (a<0.0){return(-1.0);} else {return(0.0);}
}

inline int signum(int a){
	if(a>0){return(1);} else if (a<0){return(-1);} else {return(0);}
}

int bound_value(int variable, int lower, int upper);
float bound_value(float variable, float lower, float upper);

int     
	entier      	(float value), 	// entier(3.5) == 3, entier(-3.5) == -4 
	odd         	(int value)   	// return true (1) if value is odd
;

/*
 *	Random number support (autoinitializing seeds !)
 */
 
int 
	random_bit  	(void),              	// returns 0 or 1
	random_int  	(int range)     	 	// random_int(5) returns 1 out of 0,1,2,3,4
;
float   
	random_float	(float range),   		// return one out of [0,range[
    random_float_01	(void),                 // return [0,1[
	random_symmetric_range(float range)		// return one out of [-range, range[
;

inline void toggle_bool(int &a_bool){
    a_bool = !a_bool;
}

long msecs_diff(struct timeval snd_time, struct timeval fst_time);

#endif	// UTILITY_H
