
#include "utility.h"
#include "wildcard.h"

/* Shell like pattern matching for meta characters *, ?, [class] */
/* Check string s against pattern p and return 1 on match 0 else */

int wildcard_match(const char *p, const char *s){
    if(!p || !s) return 0;

    register int lc, ok, ex;
    
    for( ; *p; s++, p++) {
	switch(*p) {
	case '\\':                /* Literal match next char */
	    ++p;
	    /*FALLTHRU*/
	default:                  /* Literal match char */
	    if(*s != *p)
		return(0);
	    continue;
	case '?':                 /* Match any char */
	    if(*s == '\0')
		return(0);
	    continue;
	case '*':                 /* Match any chars */
	    if(*++p == '\0')        /* Matches all */
		return(1);
	    for( ; *s; s++)
		if(wildcard_match(p, s))
		    return(1);
	    return(0);
	case '[':                 /* Class */
	    if ( (ex = (p[1] == '^' || p[1] == '!')) )
		++p;
	    for(lc = 0400, ok = 0; *++p && *p != ']'; lc = *p)
		if(*p == '-' ? *s <= *++p && *s >= lc : *s == *p)
		    ok = 1;
	    if(ok == ex)
		return(0);
	    continue;
	}
    }
    return(*s == '\0');
} 

#ifdef OLDCODE
int wildcard_match(const char *pattern, const char *string){
    if(!pattern){
	error_msg("wildcard match called with NULL pattern string.");
	return 0;
    }
    if(!string){
	error_msg("wildcard match called with NULL match string.");
	return 0;
    }
    while (1) {
	if(*pattern == '*'){
	    pattern++;
	    /* end of pattern string ? we match it all :-) */
	    if(*pattern == '\0'){
		return 1;
	    } else {
		/* next pattern char */
		/* what do we do ? If we can match the rest of string
		 * straight away, great (ret 1). 
		 * if we can't match it, we test 
		 * wether it is exhausted (ret 0)
		 * if not, we advance a letter (doing * after all) and retry
		 * to match.
		 */
		while(!wildcard_match(pattern, string)){
		if(*string == '\0'){
		return 0;
	    } else {
		string++;
	    }
	    }
		return 1;
	    }
	    } else if(*pattern == '\0'){
		/* end of pattern string ... end of source string ? */
		return (*string == '\0');
	    } else if(*pattern == '?'){
		/* match one more character */
		if (*string == '\0'){
		/* gasp, couldn't match */
		return 0;
	    } else {
		/* step one forward */
		pattern++;
		string++;
	    }
	} else if (*pattern == *string){
	    /* pointwise match */
	    pattern++;
	    string++;
	} else {
	    /* pointwise mismatch */
	    return 0;
	}
    }
}    
#endif

#ifdef  UNDER_DEVELOPMENT
//#include <iostream.h>

main(){
    char patbuf[256], stringbuf[256];
    while(1){
	cout << "give pattern: ";
	cin.getline(patbuf, 256);
	cout << "give string:  ";
	cin.getline(stringbuf, 256);
	if(wildcard_match(patbuf,stringbuf)){
	    cout << "Match !" << endl;
	} else {
	    cout << "Don't match." << endl;
	}
    }
    return 0;
}

#endif // UNDER_DEVELOPMENT

