/*
 * functionality for mini-window manager.
 */
#include <iostream.h>
#include <sys/param.h>

#include "browser.h"
#include "edit_forms_aux.h"
#include "error_notifier_forms_aux.h"
#include "flspecial.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "linklist.h"
#include "menu.h"
#include "options_forms.h"
#include "options_forms_aux.h"
#include "resource.h"
#include "stack.h"
#include "tree_forms.h"
#include "tree_forms_aux.h"
#include "bitmaps/wm.bm"
#include "wm_forms.h"
#include "wm_forms_aux.h"
#include "xshell.h"

int wm_visible_flag = TRUE,
    wm_wx = 10,
    wm_wy = 10,
    wm_ww = 185,
    wm_wh = 60
;

const char *wm_ID_string = "wm";

prm_entry_t wm_prm_table[] = {
    {"wx",           int_e,   (void *) &wm_wx}, 
    {"wy",           int_e,   (void *) &wm_wy}, 
    {"ww",           int_e,   (void *) &wm_ww}, 
    {"wh",           int_e,   (void *) &wm_wh}, 
    {"visible_flag", int_e,   (void *) &wm_visible_flag}, 
    {"",             int_e,   NULL}
};

static int is_hidden_flag = FALSE, is_real_hidden_flag = FALSE;

#ifdef NO_TEMPLATE
static stack_c wid_stack(75);
#else
static stack_c<Window> wid_stack(75);
#endif

static Window wid_stack_get(int index);

static long get_window_list_menu(void);
static void get_window_stack(int bWmOkFlag = FALSE); 

static void var_sync(void){
    wm_wx = WmForm->x;
    wm_wy = WmForm->y;
    wm_ww = WmForm->w;
    wm_wh = WmForm->h;
}

void wm_save_defaults(ofstream &ofile){
    if(WmForm->visible){
        var_sync();
    }
    resource_save_defaults(ofile, wm_prm_table, wm_ID_string);
}    

int wm_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    M_trace("wm_load_default", "start");
    return resource_load_default(
        wm_prm_table,  
        wm_ID_string,   
        obj_tok, name_tok, val_tok
    );
}

void wm_sync_defaults(void){
    // after loading. 
    if(wm_visible_flag){
	wm_show_form();
    } else {
	wm_close_form();
    }
}

void wm_close_form(void){
    static int first = TRUE;
    if(WmForm->visible){
	if(is_hidden_flag && is_real_hidden_flag){
	    timed_notifier("Cannot close window manager while rest of UC is hidden.");
	    return;
	}
	if(first){
	    first = FALSE;
	    if(!OptionsForm->visible && tool_looser_flag){
		timed_notifier("use Ctrl-F2 Options to get the UC mini window manager back.");
	    }
	}
	//focus_restore();
	var_sync();
	fl_hide_form(WmForm);
    }
    wm_visible_flag = FALSE;
    options_sync_defaults();
}

void WmListButtonCB(FL_OBJECT *, long){
    if (!is_hidden_flag) {
	windows_menu_cb();
    }
}

/*
void WmCloseReturnButtonCB(FL_OBJECT *, long){
    wm_close_form();
}
*/

static void wm_help_func(void){
    load_help_browser(wm_help_text, "UC Mini Window Manager");
}

void WmIconifyButtonCB(FL_OBJECT *, long){
    WmHideButtonCB(NULL, TRUE);
}

static void wm_hide_sync(void){
    fl_freeze_form(WmForm);
    if(is_hidden_flag){
	fl_set_object_label(WmHideButton, "Show");
	fl_set_object_shortcut(WmHideButton, "Ss", TRUE);

	// can't have two buttons with same shortcut. Emulate
	// it by inserting the magic underline character into the label.
	char buffer[64];
	sprintf(buffer, "S%chow", *fl_ul_magic_char);
	fl_set_object_label(WmIconifyButton, buffer);
	//fl_set_object_shortcut(WmIconifyButton, "Ss", TRUE);

	if(is_real_hidden_flag){
	    fl_deactivate_object(WmPushButton);
	    fl_deactivate_object(WmPopButton);
	    fl_deactivate_object(WmDirectionButton);
	    fl_deactivate_object(WmSlowButton);
	    fl_deactivate_object(WmMediumButton);
	    fl_deactivate_object(WmFastButton);
	    fl_deactivate_object(WmListButton);
	}
    } else {
	fl_set_object_label(WmHideButton, "Hide");
	fl_set_object_shortcut(WmHideButton, "Hh", TRUE);

	fl_set_object_label(WmIconifyButton, "Icon");
	fl_set_object_shortcut(WmIconifyButton, "Ii", TRUE);

	fl_activate_object(WmPushButton);
	fl_activate_object(WmPopButton);
	fl_activate_object(WmDirectionButton);
	fl_activate_object(WmSlowButton);
	fl_activate_object(WmMediumButton);
	fl_activate_object(WmFastButton);
	fl_activate_object(WmListButton);
    }
    fl_unfreeze_form(WmForm);
}

static void wm_busy(void){
    fl_set_cursor(WmForm->window, XC_watch);
}

static void wm_notbusy(void){
    fl_reset_cursor(WmForm->window);
}

void WmHideButtonCB(FL_OBJECT *, long iconify_flag){
    wm_busy();
    is_hidden_flag = !is_hidden_flag;
    if(is_hidden_flag){
	if(iconify_flag){
	    tool_iconify();
	} else {
	    tool_hide();
	}
    } else {
	tool_show();
    }
    wm_notbusy();
}
	
void WmPushButtonCB(FL_OBJECT *, long){
    wm_busy();
    tool_push();
    wm_notbusy();
}

void WmPopButtonCB(FL_OBJECT *, long){
    wm_busy();
    tool_pop();
    wm_notbusy();
}

static int direction = 8;  // like the symbols.

void WmDirectionButtonCB(FL_OBJECT *, long){
    int lflag = fl_keymask & Button2Mask;
    
    fl_freeze_form(WmForm);
    switch(direction){
    case 8: direction = lflag ? 4 : 6; break;
    case 6: direction = lflag ? 8 : 2; break;
    case 2: direction = lflag ? 6 : 4; break;
    case 4: 
    default:
	direction = lflag ? 2 : 8;
	break;
    }
    switch(direction){
    case 2:
	fl_set_object_label(WmDirectionButton, "@2");
	fl_set_object_label(WmSlowButton, "@2>");
	fl_set_object_label(WmMediumButton, "@2>>");
	fl_set_object_label(WmFastButton, "@2");
	break;
    case 4:
	fl_set_object_label(WmDirectionButton, "@4");
	fl_set_object_label(WmSlowButton, "@4>");
	fl_set_object_label(WmMediumButton, "@4>>");
	fl_set_object_label(WmFastButton, "@4");
	break;
    case 6:
	fl_set_object_label(WmDirectionButton, "@6");
	fl_set_object_label(WmSlowButton, "@6>");
	fl_set_object_label(WmMediumButton, "@6>>");
	fl_set_object_label(WmFastButton, "@6");
	break;
    case 8:
    default:
	fl_set_object_label(WmDirectionButton, "@8");
	fl_set_object_label(WmSlowButton, "@8>");
	fl_set_object_label(WmMediumButton, "@8>>");
	fl_set_object_label(WmFastButton, "@8");
	break;
    }
    fl_unfreeze_form(WmForm);
}

static void move_it(int step){
    switch(direction){
    case 8:
	tool_move_relative(0, -step);
	break;
    case 6:
	tool_move_relative(step, 0);
	break;
    case 2:
	tool_move_relative(0, step);
	break;
    case 4:
    default:
	tool_move_relative(-step, 0);
	break;
    }
}

void WmSlowButtonCB(FL_OBJECT *, long){
    move_it(fl_keymask & Button2Mask ? -1 : 1);
}

void WmMediumButtonCB(FL_OBJECT *, long){
    move_it(fl_keymask & Button2Mask ? -10 : 10);
}

void WmFastButtonCB(FL_OBJECT *, long){
    move_it(fl_keymask & Button2Mask ? -100 : 100);
}

static int wm_form_raw_cb(FL_FORM *form, void *xev){
    XEvent *xevent = (XEvent *)xev;
    if(form != WmForm){
	return FALSE;
    }
    /*
    if(adjust_return_button(WmForm, xevent)){
	return TRUE;
    }
    */
    return standard_form_event_cb(WmForm, xevent, wm_close_form, wm_help_func);
}

void wm_raise_form(void){
    if(WmForm->visible){
	XMapRaised(fl_display, WmForm->window);
	//cerr << "raising wm !\n";
    }
}

void wm_show_form(int move_focus_flag){
    /* set up the menu */
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_register_call_back(WmForm, FL_ALL_EVENT, wm_form_raw_cb);
	fl_set_object_shortcut(WmExitButton,    "Ee", TRUE);
	fl_set_object_shortcut(WmListButton,    "Ll", TRUE);
	fl_set_object_shortcut(WmHideButton,    "Hh", TRUE);
	fl_set_object_shortcut(WmIconifyButton, "Ii", TRUE);
	fl_set_object_shortcut(WmPushButton,    "Pp", TRUE);
	fl_set_object_shortcut(WmPopButton,     "oO", TRUE);
	//fl_set_object_shortcut(WmCloseReturnButton, "Cc^M", TRUE);
    }
    //set_to_ret_button(WmCloseReturnButton);

    if(move_focus_flag) focus_save();
    if(!WmForm->visible){
	fl_set_initial_placement(WmForm, wm_wx, wm_wy, wm_ww, wm_wh);
	fl_set_form_minsize(WmForm, 40, 20);
	char buffer[MAXPATHLEN+64];
	if(strlen(tool_host_string)>1){
	    sprintf(buffer, "WM%s", tool_host_string);
	} else {
	    strcpy(buffer, "UC Mini-WM");
	}
	fl_show_form(WmForm, FL_PLACE_FREE, TRUE, buffer);
    } else {
	XMapRaised(fl_display, WmForm->window);
    }
    if(move_focus_flag) tool_focus_move(WmForm, WmDirectionButton);

    static Pixmap pixmap =0;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, WmForm->window, wm_bits, wm_width,
	    wm_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    set_window_icon(WmForm->window, pixmap);
    
    wm_visible_flag = TRUE;
    options_sync_defaults();
}

/* window management cb's */
static void do_browser(int ID){
  // with browsers, it's ok to use the value directly. 
  if(ID > 0 && ID < browser_c::next_ID){
    sginap(10);
    tool_set_source(ID, TRUE);
    //source_browser->win_raise();
  } else {
    cerr << "Invalid browser ID (" << ID << ") in do_browser.\n";
  }
}

class index_c {
public:
  Window window;
  int nIndex;
};

static linklist_c winlist;

#ifdef OLDCODE
static void do_window(int nID){
    nID--;
    for (int i=0; i<winlist.getsize(); i++){
      index_c *pIndex = (index_c *) winlist[i];
      if (pIndex && nID == pIndex->nIndex) {
	FL_FORM *form = fl_win_to_form(pIndex->window);
	if(form) {
	  fl_raise_form(form);
	  tool_focus_move(form);
	  return;
	}
      }
    }
    cerr << "unknown ID (" << nID << ") in do_window.\n";
}
#endif

static void do_tree_cb(int){
  sginap(10);
  do_tree(); 
}

static Window xshell_window_id;

static void do_shell_cb(int){
  sginap(10);
  XMapRaised(fl_display, xshell_window_id);
  focus_move(
    xshell_wx + xshell_ww/2, 	
    xshell_wy + xshell_wh/2, 	 	
    xshell_window_id
  );
}

static long get_window_list_menu(void){
    static long window_list_menu = 0;
    if  (window_list_menu > 0){
      free_pup(window_list_menu);
    }
    window_list_menu = new_pup();
    if (!window_list_menu) return 0;

    char buffer[MAXPATHLEN + 64], digit[8];

    // empty list now.
    while(winlist.getsize()){
      index_c *pIndex = (index_c *) winlist.getval();
      if (pIndex) delete pIndex;
      winlist.remove();
    }

    add_to_pup(window_list_menu, "UC Window List%t",  0);

    // we start our list at 1.
    int nItems = 1;

    // browsers first.
    for (int i=1; i<browser_c::next_ID; i++, nItems++) {
	browser_c *browser = tool_find_browser_id(i);
	if (!browser) {
	    cerr << "Invalid browser ID (" << i << ") in get_window_list_menu.\n";
	    return 0;
	}
	if (nItems<10){
	    sprintf(digit, "%d", nItems);
	    sprintf(buffer, "%d:  Browser #%d: %s (Ctrl-%d)    %cf", 
		nItems, i, browser->get_cwd(), i, '%');
	    add_to_pup(window_list_menu, buffer, (long) do_browser, digit);
	} else {
	    sprintf(buffer, "     Browser #%d: %s (Ctrl-%d)    %cf", 
		i, browser->get_cwd(), i, '%');
	    add_to_pup(window_list_menu, buffer, (long) do_browser, "");
	}
    }

    // editors next
    int nAdded = AddEditorsToMenu(window_list_menu, browser_c::next_ID);
    if (nAdded < 0) {
      cerr << "AddEditors failed.\n";
    } else {
      nItems += nAdded;
    }

    // tree next
    if (TreeForm && TreeForm->visible) {
	add_to_pup(window_list_menu, "     Tree %f", (long) do_tree_cb, "Tt");
        nItems++;
    }

    // shell next
    xshell_window_id = xshell_get_win();
    if (xshell_window_id != None) {
	add_to_pup(window_list_menu, "     Shell %f", (long) do_shell_cb, "Ss");
	nItems++;
    }

    // the remaining windows stink (you can see RCS for code).
    return window_list_menu;
  }

static void do_hide(int){
    tool_hide();
    if(!WmForm->visible){
	wm_show_form();
    }
}

static void do_icon(int){
    tool_iconify();
}

static void do_move(int){
    wm_show_form(TRUE);
}

static void do_pop(int){
    tool_pop();
}

static void do_push(int){
    tool_push();
}

long get_windows_menu(void){
    static long windows_menu = 0;
    if (windows_menu > 0){
      free_pup(windows_menu);
    }
    windows_menu = new_pup();
    if(windows_menu > 0){
      add_to_pup(windows_menu, "Windows (Alt-w, Esc-w)%t",  0);

      long window_list_menu = get_window_list_menu();
      if (window_list_menu > 0) {
	add_to_pup(windows_menu, "Switch to %m", window_list_menu, "Ss");
      }
      add_to_pup(windows_menu, "Hide All %f",    (long) do_hide, "Hh");
      add_to_pup(windows_menu, "Iconify All %f", (long) do_icon, "Ii");
      add_to_pup(windows_menu, "Move All %f",    (long) do_move, "Mm");
      add_to_pup(windows_menu, "Push All %f",    (long) do_push, "Pp");
      add_to_pup(windows_menu, "Pop All %f",     (long) do_pop,  "oO");
    } else {
      M_err("get_windows_menu:","unable to make new pup");
    }
    return windows_menu;
}

void windows_menu_cb(void){
    long menu = get_windows_menu();
    if(menu){
	do_pup(menu);
    }
}

/* watch for vroots. stack independent of forms, just via ID's.
 * what about only 20 forms max in forms.c ?
 */

static int is_uc_wid(Window window){
    if(window == WmForm->window) return FALSE; // exception.
    FL_FORM *form = fl_win_to_form(window);
    if(form) return (form->visible);
    if(xshell_check_win(window)) return TRUE;
    return FALSE;
}

static void wid_stack_push(Window window){
#ifdef NO_TEMPLATE
    Window *new_w = new Window;
    *new_w = window;
    wid_stack.push(new_w);
#else
    wid_stack.push(window);
#endif
}

static Window wid_stack_get(int index){
#ifdef NO_TEMPLATE
    Window *w = (Window *) wid_stack[index];
    if(w){
	return *w;
    } else {
	return 0;
    }
#else
    return wid_stack[index];
#endif
}

static void query_tree(Window w, int level){
    level++;
    if(level > 4) return; // optimize a bit. 3 works, 4 is safe even with tvtwm.
    Window *children, dummy;
    unsigned int nchildren;
    if (!XQueryTree(fl_display, w, &dummy, &dummy, &children, &nchildren)){
	M_warn("UC","XQueryTree failed.");
	return;
    }
    for(unsigned int i=0; i<nchildren; i++){
	if(is_uc_wid(children[i])){
	    wid_stack_push(children[i]);
	} else {
	    query_tree(children[i], level);
	}
    }
    if(children){
	XFree((char *)children);
    }
}

static int (*tmp_handler)(Display *, XErrorEvent *);

static int local_handler(Display *, XErrorEvent *){
    M_warn("get_stacking_order","window ID's changed.");
    return 0;
}

static void get_stacking_order(void){
    // empty stack first.
    while(wid_stack.size() > 0){
#ifdef NO_TEMPLATE
	delete (Window *) wid_stack.pop();
#else
	wid_stack.pop();
#endif
    }
    //cerr << "start get_stacking_order.\n";
    int level = 0;
    query_tree(RootWindow(fl_display, fl_screen), level);
    //cerr << "end.\n";
}

static void get_window_stack(int bWmOkFlag){
    // just get id's. Don't care for stacking order. fast.
    while(wid_stack.size() > 0){
#ifdef NO_TEMPLATE
	delete (Window *) wid_stack.pop();
#else
	wid_stack.pop();
#endif
    }
#ifdef UNUSED_CODE
    FL_FORM **fws, **fw, *form;
    for (fw = forms, fws = fw + formnumb; fw < fws; fw++){
	form = *fw;
	if((form != WmForm || bWmOkFlag) && form->visible){
	    wid_stack_push(form->window);
	}
    }
#endif
    
    Window window = xshell_get_win();
    if(window != None){
	wid_stack_push(window);
    }
}

static void tool_hiconify(int hide_flag){
    is_hidden_flag = TRUE;
    is_real_hidden_flag = hide_flag;
    fl_freeze_all_forms();
    tmp_handler = XSetErrorHandler(local_handler);
    get_stacking_order();
    int sz = wid_stack.size();
    for(int i=0; i<sz; i++){
	if(hide_flag){
	    XWithdrawWindow(fl_display, wid_stack_get(i), fl_screen);
	} else {
	    XIconifyWindow(fl_display, wid_stack_get(i), fl_screen);
	}
    }
    XSetErrorHandler(tmp_handler);
    fl_unfreeze_all_forms();
    wm_hide_sync();
}

void tool_iconify(void){
    tool_hiconify(FALSE);
}

void tool_hide(void){
    tool_hiconify(TRUE);
}

void tool_show(void){
    is_hidden_flag = FALSE;
    is_real_hidden_flag = FALSE;
    fl_freeze_all_forms();
    tmp_handler = XSetErrorHandler(local_handler);
    // use old stacking order that was created upon iconification.
    int sz = wid_stack.size();
    for(int i=0; i<sz; i++){
	XMapRaised(fl_display, wid_stack_get(i));
    }
    XSetErrorHandler(tmp_handler);
    wm_hide_sync();
    wm_raise_form();
    fl_unfreeze_all_forms();
}

void tool_pop(void){
    fl_freeze_all_forms();
    tmp_handler = XSetErrorHandler(local_handler);
    get_stacking_order();
    int sz = wid_stack.size();
    for(int i=0; i<sz; i++){
	XRaiseWindow(fl_display, wid_stack_get(i));
    }
    XSetErrorHandler(tmp_handler);
    wm_raise_form();
    fl_unfreeze_all_forms();
}

void tool_push(void){
    fl_freeze_all_forms();
    tmp_handler = XSetErrorHandler(local_handler);
    get_stacking_order();
    while(wid_stack.size() > 0){
#ifdef NO_TEMPLATE	
	Window *tmp_w = (Window *) wid_stack.pop();
	if(tmp_w){
	    XLowerWindow(fl_display, *tmp_w);
	    delete tmp_w;
	}
#else
	XLowerWindow(fl_display, wid_stack.pop());
#endif
    }
    XSetErrorHandler(tmp_handler);
    fl_unfreeze_all_forms();
}

FL_Coord wm_forms_fudge_dx = 0, wm_forms_fudge_dy = 0;
/*
static Window pred_window = 0;
static time_t timeout_time = 0;

#ifdef SGI_IRIX4
int configure_predicate(Display *, XEvent *xev, char *){
#else
Bool configure_predicate(Display *, XEvent *xev, XPointer){
#endif
    time_t the_time = time(NULL);
    if(the_time-timeout_time >= 2){
	cerr << "wait configure: timeout (2secs).\n";
	return TRUE;
    }
    if(xev->xany.window != pred_window) return FALSE;
    // StructureNotifyMask;
    return (xev->type == ConfigureNotify);
    //may,maybe add || xev->type == ReparentNotify.
}

void wait_configure(Window window){
    if(!window) return;
    timeout_time = time(NULL);
    XEvent xev;
    pred_window = window;
    XPeekIfEvent(fl_display, &xev, configure_predicate, NULL);
}
*/

// one click == 10 msecs.
void wait_configure(Window window, int clicks){
    if(!window) return;
    int click = 0;
    do {
	// cerr << "click: " << click << endl;
	if(CheckPeekWindowEvent(fl_display, window, ConfigureNotify)) return;
	sginap(1);
	click++;
    } while(click < clicks);
    M_err("wait_configure","timeout");
}

void wm_compute_fudges(FL_FORM *form){
    if(!form){
	M_err("wm_compute_fudges:", "NULL form");
	return;
    } else if(!form->visible){
	M_err("wm_compute_fudges:", "invisible form");
	return;
    }
    if(fl_wmstuff.rep_method == FL_WM_SHIFT){
	// make a good first guess for forms.
	wm_forms_fudge_dx = -fl_wmstuff.rpx;    
	wm_forms_fudge_dy = -fl_wmstuff.rpy;
    }
    FL_Coord wx,wy,ww,wh,old_wx,old_wy,raw_dx=1,raw_dy=1;

    // get & do
    fl_get_win_geometry(form->window, &old_wx, &old_wy, &ww, &wh);
    M_warn("old geometry: ", "old_wx %d, old_wy %d", old_wx, old_wy);
    fl_set_form_position_direct(form, old_wx+raw_dx+wm_forms_fudge_dx, old_wy+raw_dy+wm_forms_fudge_dy);

    // wait for done, longer than normal, this is important.
    wait_configure(form->window, 500);

    // now check again.
    fl_get_win_geometry(form->window, &wx, &wy, &ww, &wh);
    M_warn("new geometry: ", "wx %d, wy %d", wx, wy);

    // difference
    FL_Coord fdx, fdy;
    fdx = old_wx + raw_dx - wx;
    fdy = old_wy + raw_dy - wy;
    M_warn("forms auto-fudge: "," dx %d, dy %d, fdx %d, fdy %d",wm_forms_fudge_dx,wm_forms_fudge_dy,fdx,fdy);

    wm_forms_fudge_dx += fdx;
    wm_forms_fudge_dy += fdy;
    M_warn("new f auto-fudge: "," dx %d, dy %d",wm_forms_fudge_dx,wm_forms_fudge_dy);
    // no further correction necessary, form will be hidden.
}

void tool_move_relative(int raw_dx, int raw_dy){
    // assume fudges are stable.
    fl_freeze_all_forms();
    tmp_handler = XSetErrorHandler(local_handler);
    xshell_move_relative(raw_dx, raw_dy);  // raw dx, dy please.

    get_window_stack();
    int sz = wid_stack.size();
    for(int i=0; i<sz; i++){
	FL_FORM *form = fl_win_to_form(wid_stack_get(i));
	if(form){
	    // works also for non-visible forms, it seems.
	    FL_Coord ww,wh,old_wx, old_wy;
	    // extra safety margin: get the geometry explicitly here. 
	    fl_get_win_geometry(form->window, &old_wx, &old_wy, &ww, &wh);

	    M_warn("old geometry: ", "wx %d, wy %d", old_wx, old_wy);
	    M_warn("form geometry: ", "f->wx %d, f->wy %d, rdx %d, rdy %d, fdx %d, fdy %d",
		  form->x, form->y, raw_dx, raw_dy, wm_forms_fudge_dx, wm_forms_fudge_dy);

	    fl_set_form_position_direct(form, old_wx+raw_dx+wm_forms_fudge_dx, old_wy+raw_dy+wm_forms_fudge_dy);
	    // extra safety: wait for configure only for visible windows
	    if(form->visible){
		wait_configure(form->window);
	    }
	    // doesn't work (even if no freezing, and only visible forms) ?
	}
    }
    XSync(fl_display, False);    // new geometries now, please. (done via wait ?)
    XSetErrorHandler(tmp_handler);
    fl_unfreeze_all_forms();
}

