/*
 * xshell.C: 
 * define xshell interface.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>

#include <iostream.h>
#include <math.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

#include "browser.h"
#include "cb_fkey.h"
#include "client_msg.h"
#include "dir_menu.h"
#include "error_notifier_forms_aux.h"
#include "file_utility.h"
#include "fkeybar_forms_aux.h"
#include "flspecial.h"
#include "forms.h"
#include "forms_aux.h"
#include "global.h"
#include "gl_utility.h"
#include "help_forms_aux.h"
#include "help_text.h"
#include "license_forms_aux.h"
#include "menu.h"
#include "options_forms_aux.h"
#include "panel.h"
#include "pathnames.h"
#include "resource.h"
#include "status_forms_aux.h"
#include "toolbox_forms_aux.h"
#include "tree_forms_aux.h"
#include "utility.h"
#include "wm_forms.h"
#include "wm_forms_aux.h"
#include "bitmaps/xshell.bm"
#include "xshell.h"
#include "xshell_forms.h"

#ifdef AIX
#   include <sys/ioctl.h>
#   include <sys/types.h>
#endif

#ifdef LINUX_GLIBC2
#       include <errno.h>
#endif

static void tool_sync_geometry(void);
static void xshell_compute_wxy(void);

#ifndef SGI
#define USE_XTERM
#endif

//#define USE_XTERM
// SGI uses xwsh, not winterm. period

/* Global variables used below */

static Window xshell_win_id =0;

int
    xshell_update_count = 50,
    xshell_font_h = 15,
    xshell_font_w = 9,
    xshell_wh     = 90,
    xshell_ww     = 400
//    xshell_wbw    = 0,    // win border width
//    xshell_wth    = 24    // win top height. hack.
;    

int xshell_x_fudge  = 0,
    xshell_y_fudge  = 0,
    xshell_geo_cols = 80,
    xshell_geo_rows = 8,
    // values are recomputed before use.
    // these are fallback 
    // wy uses an inverted geometry because I cannot estimate
    // the shells window height in advance, it is too font-dependent.
    xshell_wx = 0, 	
    xshell_wy = 0 	
;

FL_Coord 
    // offset for the whole window array,
    // measured from the bottom left corner.
    // most values are computed dynamically.
    nXOffset	= 0,		
    nYOffset	= 0,
    nWinBorderW	= 0,
    nWinTitleH	= 0
;

static char xshell_options_string[MAXPATHLEN];

prm_entry_t xshell_prm_table[] = {
    {"columns",        int_e,  (void *) &xshell_geo_cols}, 
    {"rows",           int_e,  (void *) &xshell_geo_rows}, 
    {"win_x",          int_e,  (void *) &xshell_wx}, 
    {"win_y ",         int_e,  (void *) &xshell_wy}, 
    {"fudge_x ",       int_e,  (void *) &xshell_x_fudge}, 
    {"fudge_y ",       int_e,  (void *) &xshell_y_fudge}, 
    {"options_string", char_p_e, (void *) &xshell_options_string},
    {"update_count",   int_e,  (void *) &xshell_update_count},
    {"visible_flag",   int_e,  (void *) &xshell_visible_flag}, 
    {"",               int_e,  NULL}
};

const char
#ifdef USE_XTERM
    *xshell_class_name   = "XTerm",
//    *xshell_class_name   = "HPterm",
#else
    *xshell_class_name   = "XWsh",
#endif
    *xshell_ID_string    = "xshell",
    *xshell_win_title    = "UCShell"
    ;

static char xshell_instance_name[MAXHOSTNAMELEN+10]; // host + PID.

char *xshell_path = NULL;
    
int xshell_visible_flag = TRUE;

static long
    global_send_event_sent_count = 0,
    global_send_event_received_count = 0
    ;

static int delayed_search_xshell(void);
static int search_xshell(void);
static void var_sync(void);
static int xshell_check_fifo(void);
static void xshell_close_fifo(void);
static void xshell_init_fifo(void);
static void xshell_options_defaults(void);

static int dbg_flag = tool_debug_flag;
//static int dbg_flag = TRUE;

#include <X11/keysym.h>

/* two-layered system:
 * a) attempt StringToKeysym - works with alphanums.
 * b) attempt one-2-one translation.
 * on failure, return 0;
 * new edition:
 * realize that there is no one-to-one back translation function that makes a valid x key
 * event out of a character. I can deduce the keysym with my table, but not the possible
 * state modifications (shift). So, the new approach is to determine this via trial and
 * error. I should put that information (the shiftflag) into an array. However, the lookup
 * function is local and therefore fast.
 * If I communicate normal events from the browser or the tree to the shell, I can 
 * avoid going through this bottleneck - which means it doesn't have to be fast.
 * It is only used for the occasional selection forwarding. 
 * German keyboards use XK_Mode_Switch for Alt-Graph, which switches the character set.
 * That messes up the whole affair, because the state gets an offset from 0x2000. 
 */

class CKeyInfo {
public:
    char key;
    int  bValid;
    unsigned int nState;
    KeyCode keycode;
    CKeyInfo(void);
};

CKeyInfo::CKeyInfo(void){
    bValid  = FALSE;
    key     = '\0';
    keycode = 0;
    nState  = 0;
}

static CKeyInfo aCKeyInfo[256];

static int TableLookup(char key, unsigned int &nState, KeyCode &keycode) {
    int nIndex = (int) key;
    if (nIndex <0 || nIndex > 255) {
	cerr << "UCShell: Invalid Index (" << nIndex << ", " << key << ") for TableLookup.\n";
	return FALSE;
    }
    if (!aCKeyInfo[nIndex].bValid) {
	return FALSE;
    }
    nState  = aCKeyInfo[nIndex].nState;
    keycode = aCKeyInfo[nIndex].keycode;
    return TRUE;
}

static void TableInsert(char key, unsigned int nState, KeyCode keycode){
    int nIndex = (int) key;
    if (nIndex <0 || nIndex > 255) {
	cerr << "Invalid Index (" << nIndex << ") for TableInsert.\n";
	return;
    }
    aCKeyInfo[nIndex].bValid  = TRUE;
    aCKeyInfo[nIndex].nState  = nState;
    aCKeyInfo[nIndex].keycode = keycode;
}

static int IsSameKey(char c0, char c1){
    if (c0 == c1) return TRUE;
    if (c0 == 13 && c1 == 10) return TRUE;
    if (c0 == 10 && c1 == 13) return TRUE;
    return FALSE;
}

static KeyCode GL2KeyCode(char key, unsigned int &state){
    /*
      static int shift_flags [256], first = TRUE;
      if(first){
      first = FALSE;
      for(int i=0; i<sizeof(shift_flags), i++){
      shift_flags[i] = -1;  // means not initialized. 0 means false, 1 means true.
      }
      }
      */
    KeyCode kc;
    if (TableLookup(key, state, kc)) {
	return kc;
    }

    char buf[2];
    buf[0] = key; buf[1] = '\0';
    if (dbg_flag) cerr << "\nAttempting translation of '" << (int) key << "', '" << key << "'\n";

    KeySym ks;
    ks = XStringToKeysym(buf);
    if(ks == NoSymbol){
	if(dbg_flag) cerr << "string 2 keysym failed. lowlevel now.\n";
	switch(key){
	case 10:  ks = XK_Return; break;
	case 13:  ks = XK_Return; break;
	case ' ': ks = XK_space               ; break;
	case '!': ks = XK_exclam              ; break;
	case '@': ks = XK_at                  ; break;
	case '#': ks = XK_numbersign          ; break;
	case '$': ks = XK_dollar              ; break;
	case '%': ks = XK_percent             ; break;
	case '&': ks = XK_ampersand           ; break;
//	case '': ks = XK_apostrophe           ; break;
	case '\'':ks = XK_quoteright          ; break;
	case '(': ks = XK_parenleft           ; break;
	case ')': ks = XK_parenright          ; break;
	case '*': ks = XK_asterisk            ; break; 
	case '+': ks = XK_plus                ; break;
	case ',': ks = XK_comma               ; break;
	case '-': ks = XK_minus               ; break;
	case '.': ks = XK_period              ; break;
	case '/': ks = XK_slash               ; break;
	case ':': ks = XK_colon               ; break;
	case ';': ks = XK_semicolon           ; break;
	case '<': ks = XK_less                ; break;
	case '=': ks = XK_equal               ; break;
	case '>': ks = XK_greater             ; break;
	case '?': ks = XK_question            ; break;
	case '[': ks = XK_bracketleft         ; break;
	case '\\':ks = XK_backslash           ; break;
	case ']': ks = XK_bracketright        ; break;
	case '^': ks = XK_asciicircum         ; break;
	case '_': ks = XK_underscore          ; break;
//	case '': ks = XK_grave                ; break;
	case '`': ks = XK_quoteleft           ; break;
	case '{': ks = XK_braceleft           ; break;
	case '|': ks = XK_bar                 ; break;
	case '}': ks = XK_braceright          ; break;
	case '~': ks = XK_asciitilde          ; break;
// also XK_dead_tilde on Linux.
#ifndef XK_dead_tilde 
#define XK_dead_tilde 0xFE53 
#endif

#ifdef XK_quotedbl
	case '"': ks = XK_quotedbl            ; break;
#endif
#ifdef XK_section
	case '': ks = XK_section             ; break;
#endif
#ifdef XK_ssharp
	case '': ks = XK_ssharp              ; break;
#endif
/*
#ifdef XK_
	case '': ks = XK_            ; break;
#endif
*/
	default:
	    if(dbg_flag) cerr << "no symbol for buf " << buf << endl;
	    return 0; 
	}
    } 
    kc = XKeysymToKeycode(fl_display, ks);
    if (kc == 0) {
	// special cases try alternatives:
	switch(key){
	case '~': 
	    ks = XK_dead_tilde; 
	    kc = XKeysymToKeycode(fl_display, ks);
	    break;
	default:
	    break;
	}
	if (kc == 0) {
	    cerr << "Can't generate keycode for keysym '" << ks << "'\n";
	    return 0;
	}
    }
    
    XKeyEvent xkev;
    memset(&xkev, 0, sizeof(XKeyEvent));   // just in case.
    xkev.type        = KeyPress;  // always
    xkev.display     = fl_display;
    xkev.window      = xshell_win_id;
    xkev.root        = RootWindow(fl_display, fl_screen);
    xkev.state       = 0;
    xkev.keycode     = kc;     // computed
    xkev.time        = CurrentTime;
    xkev.same_screen = True;
    xkev.subwindow   = None;
    xkev.x = xkev.y = xkev.x_root = xkev.y_root = 0;

    KeySym keysym;
    char keybuf[2];
    int continue_flag, nStateOffset = 0;

    for (int i=0; i<2; i++) {
	switch(i) {
	case 0:	nStateOffset = 0x0000; break;
	case 1:	nStateOffset = 0x2000; break;
	}
	// reset state between loops.
	xkev.state    = 0;
	continue_flag = TRUE;
	while(continue_flag){
	    // conditionally switch the character set:
	    xkev.state |= nStateOffset;

	    // first attempt, with shift off.
	    state = xkev.state;
	    keybuf[1] = '\0';  // saveguard.

	    XLookupString(&xkev, keybuf, sizeof(keybuf), &keysym, 0);
	    if(IsSameKey(keybuf[0], key)){
		if (dbg_flag) {
		    cerr << "GL2K key match (no shift) for '" << keybuf[0] << "'\n";
		}
		if(keysym == ks){
		    if (dbg_flag) cerr << "GL2K keysym match! (no shift)\n";
		    TableInsert(key, xkev.state, kc);
		    return kc;
		} 
		if (dbg_flag) cerr << "GL2K keysym MISmatch (no shift)\n";
	    }
    
	    // next attempt, with shift on.
	    xkev.state |= ShiftMask;
	    state = xkev.state;
	    XLookupString(&xkev, keybuf, sizeof(keybuf), &keysym, 0);
	    if(IsSameKey(keybuf[0], key)){
		M_info2("GL2K", "key match (shift)(%c).", keybuf[0]);
		if(keysym == ks){
		    M_info2("GL2K", "keysym match! (shift)");
		    TableInsert(key, xkev.state, kc);
		    return kc;
		} 
		M_info2("GL2K", "keysym MISmatch (shift)");
	    } else {
		if (dbg_flag) {
		    cerr << "Lookup failed (keybuf: '" << (int) keybuf[0] 
			 << "', key: '" << (int) key << "', state: "
			 << xkev.state << ")\n";
		}
	    }
	    // neither normal nor normal/shift. attempt Mod1Mask, then Mod2Mask etc
	    if(xkev.state & Mod1Mask){
		xkev.state = Mod2Mask;
		M_info2("GL2Key", "Mod2");
	    } else if(xkev.state & Mod2Mask){
		xkev.state = Mod3Mask;
		M_info2("GL2Key", "Mod3");
	    } else if(xkev.state & Mod3Mask){
		xkev.state = Mod4Mask;
		M_info2("GL2Key", "Mod4");
	    } else if(xkev.state & Mod4Mask){
		xkev.state = Mod5Mask;
		M_info2("GL2Key", "Mod5");
	    } else if(xkev.state & Mod5Mask){
		// end of transfer, duude !
		continue_flag = FALSE;
	    } else {
		xkev.state = Mod1Mask;
		M_info2("GL2Key", "Mod1");
	    }
	}
    }
    // XK_Return's always end up here. It can come in as code 10, it get's output
    // as code 13. No problem.
    cerr << "possibly illegal x key code generated, double mismatch.\n";
    // hopeful, but wrong:
    state = ShiftMask;	
    return kc;
}

// pump an ascii string to the xshell. No controls allowed.
// check for window on first character transmitted. 
int xshell_string (const char *string, int add_newline_flag){
    if(!xshell_visible_flag) return FALSE;
    if(!string) return FALSE;
    int len = strlen(string);
    for(int i=0; i<len; i++){
	if(!xshell_key(string[i], FALSE, FALSE, (i == 0))) return FALSE;
    }
    int retval = TRUE;
    if(add_newline_flag){
	retval =  xshell_key('\n', FALSE, FALSE, TRUE);
	xshell_needs_enter_flag = FALSE;
    }
    XFlush(fl_display);
    return retval;
}

/* explicit validity hack */ 

static int xwa_flag = TRUE, quiet_flag = FALSE;
static int xwa_handler(Display *, XErrorEvent *xev){
    M_warn("shell lost.","xwa_handler called.");
	if (!quiet_flag) {
    	xshell_check_errorevent(xev);     // important. 
	}
    xwa_flag = FALSE;
    return 0;
}

static int (*tmp_handler)(Display *, XErrorEvent *xev);

static int xshell_win_id_valid(XWindowAttributes *win_attr_in = NULL, int bQuiet = FALSE){
	quiet_flag = bQuiet;
    if(!fl_display || !xshell_win_id) return FALSE;
    // cerr << "xshell_win_id_valid\n";
    XWindowAttributes win_attr, *win_attr_p;
    win_attr_p = win_attr_in ? win_attr_in : &win_attr;           // allow us to get the winattr.
    xwa_flag = TRUE;  // reset for error handler.
    tmp_handler = XSetErrorHandler(xwa_handler);
    XGetWindowAttributes(fl_display, xshell_win_id, win_attr_p);  // check call.
    XSetErrorHandler(tmp_handler);
    return xwa_flag;
}

// pump a (control) character to the cockpit shell.
// deduce wether the shift mask needs to be set.

int xshell_key (char key, int ctrl_flag, int sync_flag, int check_flag){
    tool_got_event_flag = TRUE;
    if(!xshell_visible_flag) return FALSE;

    if(key == '\n'){
	xshell_needs_enter_flag = FALSE;
    }
    XKeyEvent xkev;
    memset(&xkev, 0, sizeof(XKeyEvent));   // just in case.

    unsigned int mods   = 0;

    KeyCode kc = GL2KeyCode(key, mods);
    M_info2("xshell_key", "pumping: %c, mods s(%d), ctrl(%d)", key, mods, ctrl_flag);

    if(kc == 0){
	if(dbg_flag) cerr << "No keycode for key ";
	if(dbg_flag) fprintf(stderr, "%c", key);
        if(dbg_flag) cerr << endl;
	return FALSE;
    } else {
	if(dbg_flag){
	    fprintf(stderr, "keycode found for key '%c', code: %d \n", key, (int) kc);
	}
    }

    // no type specific mods or details for KeyPress.
    if(ctrl_flag)  mods |= ControlMask;

    xkev.type        = KeyPress;  // always
    xkev.display     = fl_display;
    xkev.window      = xshell_win_id;
    xkev.root        = RootWindow(fl_display, fl_screen);
    xkev.keycode     = kc;     // computed
    xkev.state       = mods;     // computed
    xkev.time        = CurrentTime;
    xkev.same_screen = True;
    xkev.subwindow   = None;
    xkev.x = xkev.y = xkev.x_root = xkev.y_root = 0;

    if(check_flag){
	if (!xshell_win_id_valid()){
	    return FALSE;
	}
    }
    if (XSendEvent(fl_display,xshell_win_id,True,KeyPressMask,(XEvent *) &xkev) == 0){
	cerr << "XSendEvent failed.\n";
	return FALSE;
    } else {
	global_send_event_sent_count++;
    }
    if(dbg_flag) cerr << "event sent.\n";
    if(sync_flag){
	XFlush(fl_display);
    }
    return TRUE;
}

/** ERROR HANDLING **/

static char xshell_error_buf[MAXPATHLEN+1] = "";

int xshell_check_errorevent(XErrorEvent *err){
    if(!err) return FALSE;
    if(err->resourceid == xshell_win_id && xshell_visible_flag && xshell_win_id > 0){
	XGetErrorText(fl_display, err->error_code, xshell_error_buf, sizeof(xshell_error_buf));
//	if(err->error_code == BadWindow){
	xshell_win_id = 0;
	// was a real error
	fl_qenter(XSHELL_ERROR_HANDLER_CMSG, 1);
	xshell_visible_flag = FALSE;
//	}
	return TRUE;
    } else {
	return FALSE;
    }
}

void xshell_error_handler_cb(FL_OBJECT *, long){
    cerr << "UC:cockpit shell lost.\n";
    timed_error_notifier(
	"X protocol error: Did you exit the cockpit shell ?",
	xshell_error_buf, notifier_secs
    );
    if(ask_question(FALSE, "Attempt to restart the cockpit shell ?")){
	xshell_visible_flag = TRUE;
	xshell_show();
    }
    options_sync_defaults();
}

/**** SNATCHING XSHELL EVENTS ****/

static int n_xshell_children = 0;
static const int n_xshell_children_max = 20;

static Window xshell_children[n_xshell_children_max];

void xshell_select_inputs(void){
    n_xshell_children = 0;
    XSelectInput(fl_display, xshell_win_id, KeyPressMask | StructureNotifyMask);

    Window *children, dummy;
    unsigned int i, nchildren;
  
    if(dbg_flag) cerr << "parent win: " << xshell_win_id << endl;
    
    if (!XQueryTree(fl_display, xshell_win_id, &dummy, &dummy, &children, &nchildren)){
	cerr << "couldn't query tree on shell window children.\n";
	return ;
    }
    
    for(i=0; i<nchildren; i++){
	if(dbg_flag) cerr << "selecting on 1st order children: " << children[i] << endl;
	xshell_children[n_xshell_children] = children[i];
	XSelectInput(fl_display, children[i], KeyPressMask);
	n_xshell_children++;
	if(n_xshell_children >= n_xshell_children_max){
	    cerr << "UCShell: too many children.\n";
	    return;
	}
    }
    if(children){
	XFree((char *)children);
    }
    for(i=0; i<nchildren; i++){
	unsigned int nchildren2;
	Window *children2;
	if(dbg_flag) cerr << "subparent: " << xshell_children[i] << endl;
	if (!XQueryTree(fl_display, xshell_children[i], &dummy, &dummy, &children2, &nchildren2)){
	    cerr << "couldn't query tree on shell window children.\n";
	    return ;
	}
	for(unsigned int j=0; j<nchildren2; j++){
	    if(dbg_flag) cerr << "selecting on 2nd order children: " << children2[j] << endl;
	    xshell_children[n_xshell_children] = children2[j];
	    XSelectInput(fl_display, children2[j], KeyPressMask);
	    n_xshell_children++;
	    if(n_xshell_children >= n_xshell_children_max){
		cerr << "UCShell: too many children.\n";
		return;
	    }
	}
	if(children2){
	    XFree((char *)children2);
	}
    }
    if(dbg_flag)cerr << "total children: " << n_xshell_children << endl;
}

/**** STARTING ****/

static void make_new_name(void){
    strcpy(xshell_instance_name, "UCShell");
    char hostbuf[MAXHOSTNAMELEN];
    if(gethostname(hostbuf, MAXHOSTNAMELEN) < 0) return;
    pid_t pid = getpid();
    if(pid < 0) return;
    sprintf(xshell_instance_name, "%s_%d", hostbuf, pid);
    if(dbg_flag) cerr << xshell_instance_name << endl;
}

static int start_xshell(void){
    if(!xshell_path){
#ifdef USE_XTERM
	xshell_path = strdup(xterm_path);
#else
	xshell_path = strdup(xwsh_path);   // not ! winterm.
#endif
    }
    make_new_name();

    status_msg(1, "Starting cockpit shell window `%s'.", xshell_path);
    // can use system, because I need to XQueryTree for the name anyway.
    char buf[1024], geobuf[1024];
    static int show_first = TRUE;
    if(show_first && tool_looser_flag){
	xshell_options_defaults();
	show_first = FALSE;
	xshell_compute_wxy();
	sprintf(geobuf, "%dx%d+%d-%d", xshell_geo_cols, xshell_geo_rows, xshell_wx, xshell_wy);
    } else {
	sprintf(geobuf, "%dx%d+%d+%d", xshell_geo_cols, xshell_geo_rows, 
	    xshell_wx + xshell_x_fudge, 
	    xshell_wy + xshell_y_fudge
        );
    }
    /* -sl: scroll save lines. -sb: turn scrollbar on. 
     * -bw: border width (portable ?)
     * -wf: wait for mapping before starting shell subprocess. Try to avoid 
     * race condition.
     */
#ifdef USE_XTERM
    sprintf(buf, "%s -sl 640 -name %s -title %s -wf -geometry %s -sb %s ",
        xshell_path, xshell_instance_name, xshell_win_title, geobuf, xshell_options_string
    );
#else
    // -nomenu ? not ! fonts work.
    sprintf(buf, "%s -sl 640 -autofork -name %s -title %s -icontitle UCShell -geometry %s %s",
        xshell_path, xshell_instance_name, xshell_win_title, geobuf, xshell_options_string
    );
#endif
#ifndef USE_XTERM
    sprintf(geobuf, " -xrm %s*autoFork:on ", xshell_instance_name);
    strcat(buf, geobuf);
#endif
    sprintf(geobuf, " -xrm %s*allowSendEvents:on ", xshell_instance_name);
    strcat(buf, geobuf);

/*
 * didn't work anywhere yet (taken from Linux 2.0)
 sprintf(geobuf, " -xrm %s*allowSends:on ", xshell_instance_name);
 strcat(buf, geobuf);
 sprintf(geobuf, " -xrm %s*SmeBSB.allowSends:on ", xshell_instance_name);
 strcat(buf, geobuf);
 */
    // this added to Linux 2.0 made it work:
    sprintf(geobuf, " -xrm *allowSendEvents:on ");
    strcat(buf, geobuf);

    sprintf(geobuf, " -xrm %s*keyMapFileAfter:on ", xshell_instance_name);
    strcat(buf, geobuf);
    sprintf(geobuf, " -xrm %s*kpamKeyMap: ", xshell_instance_name);   // app mode disabled
    strcat(buf, geobuf);
    sprintf(geobuf, " -xrm %s*kpnmKeyMap: ", xshell_instance_name);   // normal mode disabled
    strcat(buf, geobuf);
    sprintf(geobuf, " -xrm %s*ckmdKeyMap: ", xshell_instance_name);   // cursor keys disabled disabled
    strcat(buf, geobuf);
    sprintf(geobuf, " -xrm %s*ckmeKeyMap: ", xshell_instance_name);   // cursor keys  enabled disabled
    strcat(buf, geobuf);
    // sprintf(geobuf, " -xrm %s* ", xshell_instance_name);
    // strcat(buf, geobuf);
    
    char *keymap_fname = strdup(tmpnam(NULL));
    if(!keymap_fname){
	timed_strerror_notifier("Couldn't get a temporary file name.", notifier_secs);
	return FALSE;
    }
    ofstream keymap_file(keymap_fname);
    if(!keymap_file){
	timed_strerror_notifier("Couldn't open temporary file.", notifier_secs);
	return FALSE;
    }
    for(int i=1; i<=12; i++){
	keymap_file << "F" << i << "(any): send(\"\");\n";
    }
    /* interferes with tcsh cmd repetition.
       keymap_file << "Up(any): func(\"up-line\");\n";
       keymap_file << "Down(any): func(\"down-line\");\n";
       keymap_file << "KP_Up(any): func(\"up-line\");\n";
       keymap_file << "KP_Down(any): func(\"down-line\");\n";
       */
    keymap_file << "KP_Home(any): func(\"home\");\n";
    keymap_file << "KP_End(any): func(\"end\");\n";
    keymap_file << "KP_Prior(any): func(\"up-page\");\n";
    keymap_file << "KP_Next(any): func(\"down-page\");\n";
    keymap_file.close();
#ifdef USE_XTERM
    strcat(buf, " &");
#else
    strcat(buf, " -keymap ");
    strcat(buf, keymap_fname);
    strcat(buf, " ");
#endif
    
    safe_system(buf);
    unlink(keymap_fname);
    fl_free(keymap_fname);
    return TRUE;
    // start only
}

int xshell_show(void){
    if(!fl_display) return FALSE;

    // always start a new shell.
    if(!start_xshell()) return FALSE;
    if(!delayed_search_xshell()){
	char buffer[1024];
	sprintf(buffer, "Timeout: Unable to execute/locate cockpit shell `%s'", xshell_path);
	timed_error_notifier(buffer, "Check path names and option strings. Window layout will be wrong.", notifier_secs);
	xshell_visible_flag = FALSE;
	options_sync_defaults();
	return FALSE;
    }

    // always reinit a newly found/shown shell.
    static Pixmap pixmap =0;
    if(pixmap == 0){
	pixmap = XCreatePixmapFromBitmapData(
	    fl_display, xshell_win_id, xshell_bits, xshell_width,
	    xshell_height,
	    WhitePixel(fl_display, fl_screen),
	    BlackPixel(fl_display, fl_screen),
	    1
	);
    }
    set_window_icon(xshell_win_id, pixmap);
    fl_set_winproperty(xshell_win_id, FL_COMMAND_PROP);
	
    xshell_sync_title();
    xshell_select_inputs();
    xshell_init_fifo();
    status_msg(1, "shell started.");
    
    var_sync();
    tool_sync_geometry();
    /* doesn't work:
       XReparentWindow(fl_display, xshell_win_id, source_browser->form->window, 0, 0);
       */
    return TRUE;
}

/*
 * winterms/xshells don\'t seem to eat this.
 */
/*
static void xshell_wm_delete(void){
    Atom atom_type, atom_data;
    XClientMessageEvent xcm;
    //cerr << "is delete win.\n";
    atom_type = XInternAtom(fl_display, "WM_PROTOCOLS", 0);
    atom_data = XInternAtom(fl_display, "WM_DELETE_WINDOW", 0);
    xcm.type = ClientMessage;
    xcm.display = fl_display;
    xcm.window = xshell_win_id;
    xcm.message_type = atom_type;
    xcm.format = 32;
    xcm.data.l[0] = atom_data;
    // 0xfff possibly wrong ! Rather, use a Mask. difficult for cm\'s
    if (XSendEvent(fl_display,xshell_win_id,True,0xfff,(XEvent *) &xcm) == 0){
	cerr << "XSendEvent failed.\n";
    } else {
	cerr << "XSendEvent succeeded.\n";
    }
}
*/

int xshell_check_win(Window window){
    return (window == xshell_win_id && xshell_visible_flag && xshell_win_id_valid());
}

Window xshell_get_win(void){
    if(xshell_visible_flag && xshell_win_id_valid()){
	return xshell_win_id;
    } else {
	return None;
    }
}

void xshell_cd(const char *dir){
    if(!dir || !dir[0]) return;
    if(xshell_needs_enter_flag){
	xshell_send_interrupt();
    }
    xshell_string("cd ");
    if(strpbrk(dir, " \t\n")){
	// quote the damn thing.
	xshell_key('\'', FALSE, FALSE, FALSE);
	xshell_string(dir, FALSE);
	xshell_key('\'', FALSE, FALSE, FALSE);
    } else {
	xshell_string(dir, TRUE);
    }
}

void xshell_send_interrupt(void){
    if(!xshell_win_id_valid()) return;
    struct termios tdes; 
    if(tcgetattr(0, &tdes) < 0){
	xshell_key('c', TRUE, FALSE, TRUE);     
    } else {
	char key = tdes.c_cc[VINTR];
	char kkey= key;
	if(iscntrl(key)){
	    kkey += 'a'-1;
	}
	xshell_key(kkey, iscntrl(key), FALSE, TRUE);
	M_info2("xshell_close","corrected kkey is: %c", kkey);
    }
}

void xshell_close(void){
    // always close if win id is valid. 
    if(!xshell_win_id_valid()) return;
    
    var_sync();
    
	// try to be nice first (DestroyWindow leaves a hanging
	// xterm process on HP 9.0.
      // why not ? send stdin-intr char.
      xshell_send_interrupt();
      // add extra newline, for those really fucked-up situations.
      xshell_key('\n', FALSE, FALSE, TRUE);
    
      xshell_string("exit", TRUE);
      XFlush(fl_display);
      XSync(fl_display, False);

	// allow the exit to happen before Destroy.
	int is_valid_flag = TRUE, count = 0;
	while (is_valid_flag && count < 10) {
		fl_msleep(50);
		is_valid_flag = xshell_win_id_valid(NULL, TRUE);
		count++;
	}

	if (is_valid_flag) {	
    	XDestroyWindow(fl_display, xshell_win_id);
    	XFlush(fl_display);
    	XSync(fl_display, False);
	}

    xshell_close_fifo();
    //xshell_wm_delete();
    xshell_visible_flag = FALSE;
    xshell_win_id = 0;
    options_sync_defaults();
}

static void xshell_compute_wxy(void) {
    nXOffset	= 0;
    nYOffset	= 40;
    nWinBorderW	= fl_wmstuff.rpx + fl_wmstuff.bw;
    nWinTitleH	= fl_wmstuff.rpy + fl_wmstuff.bw;
    xshell_wx	= 0;
    xshell_wy	= fkeybar_wh + nWinTitleH + nWinBorderW + nYOffset;
    if(fl_wmstuff.rep_method == FL_WM_SHIFT){
    } else {
	xshell_wy += nWinBorderW;
    }
}

static void tool_sync_geometry_split(void) {
    // nWinBorderW = fl_wmstuff.rpx;  good for 4Dwm, mwm. not twm.
    toolbox_split_flag = TRUE;

    fkeybar_wx	= nXOffset + nWinBorderW;
    fkeybar_wy	= fl_scrh - (nYOffset + fkeybar_wh + nWinBorderW);
    fkeybar_ww	= xshell_ww;

    toolbox_wx	= fkeybar_wx + fkeybar_ww + 2 * nWinBorderW;
    toolbox_wy	= fkeybar_wy - (xshell_wh + nWinBorderW + nWinTitleH);
    toolbox_wh	= fkeybar_wh + xshell_wh + nWinBorderW + nWinTitleH;

    wm_wx	= toolbox_wx + toolbox_ww + 2 * nWinBorderW;
    wm_wy	= fkeybar_wy + fkeybar_wh - wm_wh;

    status_wx	= wm_wx;
    status_wy	= toolbox_wy;
    status_ww	= wm_ww;
    status_wh	= toolbox_wh - (wm_wh + nWinBorderW + nWinTitleH);

    FL_Coord tmp_h;
    tmp_h = 20*xshell_font_h;
    bound_value(tmp_h, 200, 2*(toolbox_wh + nWinTitleH));
    FL_Coord tmp_max_w = xshell_ww + toolbox_ww + wm_ww + (2*nWinBorderW);

    // fonts can be off, so I need the actual width of a reasonable
    // 45 + fname or so characters for the browsers. Divide by ten because 
    // string is that wide. 
    // 25 pixels for the slider and extras.
    FL_Coord tmp_browser_w = (FL_Coord) (25.0 + (4.5 + 2.0)*fl_get_string_width(
	fl_get_panel_fontstyle(source_browser->panel),
	fl_get_panel_fontsize (source_browser->panel),
	"          ",
	10
    ));
    //cerr << "Width: " << tmp_browser_w << endl;

    source_browser->win_x = fkeybar_wx;

    // no wider than the whole thing:
    if (tmp_browser_w > tmp_max_w) {
	tmp_browser_w = tmp_max_w;
	source_browser->win_w = tmp_browser_w;
	target_browser->win_w = tmp_browser_w;
    } else if (tmp_browser_w < tmp_max_w / 2) {
	// no narrower than half of the whole thing, either:
	tmp_browser_w = tmp_max_w / 2;
	source_browser->win_w = tmp_browser_w;
	target_browser->win_w = tmp_max_w - tmp_browser_w;   // avoid one-pixel-off errors
    } else {
	// normal case, both the same width and overlapping.
	source_browser->win_w = tmp_browser_w;
	target_browser->win_w = tmp_browser_w;
    }
 
    source_browser->win_h = target_browser->win_h = tmp_h;
    source_browser->win_y = target_browser->win_y = 
	toolbox_wy - (nWinTitleH + nWinBorderW + source_browser->win_h)
    ;
    target_browser->win_x = source_browser->win_x + tmp_max_w + 2*nWinBorderW
	- target_browser->win_w
    ;

    tree_ww = (int) (1.5 * source_browser->win_w);
    tree_wh = source_browser->win_h + toolbox_wh;
    tree_wx = (int) (0.25 * source_browser->win_w) + nWinBorderW;
    tree_wy = source_browser->win_y + nWinTitleH;
}

static void tool_sync_geometry(void){
    if(!xshell_visible_flag)return;
    if(!tool_looser_flag) return;
    M_info2("tool_sync_geometry", "active");
    tool_sync_geometry_split();
    return;

    FL_Coord wbw, wth;
    wth = fl_wmstuff.rpy + fl_wmstuff.bw;
    wbw = fl_wmstuff.rpx + fl_wmstuff.bw;
    // wbw = fl_wmstuff.rpx;  good for 4Dwm, mwm. not twm.
    // must be here, override implicit rescaling later.
    // Mini-WM
    //wm_wy     = fl_scrh - (wm_wh + 2*wbw + wth);
    wm_wy     = fl_scrh - (wm_wh + wbw);

    // Status
    status_wx = xshell_wx + xshell_ww + 3*wbw;
    if(fl_wmstuff.rep_method == FL_WM_NORMAL){
	// SGI 4Dwm hack. test.
	status_wx -= wbw;  
    }
    wm_wx     = status_wx;
    
    status_ww = wm_ww;
    status_wh = xshell_wh - (wm_wh + wbw + wth);
    status_wy = fl_scrh - (xshell_wh + wbw);

    // fonts can be off, so I need the actual width of a reasonable
    // 70 or so characters for the browsers. 10 for the slider and extras.
    int tmp_w = (int) (8.0*fl_get_string_width(
	fl_get_panel_fontstyle(source_browser->panel),
	fl_get_panel_fontsize (source_browser->panel),
	"          ",
	10
    ));

    // sanity check:
    toolbox_ww = 124;
    if(tmp_w < xshell_ww) tmp_w = xshell_ww;

    source_browser->win_x = target_browser->win_x = wbw;
    source_browser->win_w = target_browser->win_w = tmp_w;
    source_browser->win_h = target_browser->win_h = 12*xshell_font_h;
    source_browser->win_y = status_wy - (source_browser->win_h + wbw + wth);
    target_browser->win_y = source_browser->win_y - (target_browser->win_h + wbw + wth);

    //wm_wx = source_browser->win_x;
    //wm_wy = target_browser->win_y - (WmForm->h + wbw + wth);
	
    status_ww = source_browser->win_w + toolbox_ww + 2*wbw - status_wx + wbw;
    wm_ww     = status_ww;
    
    toolbox_wh = 2*source_browser->win_h + wbw + wth;
    toolbox_wx = source_browser->win_x + source_browser->win_w + 2*wbw;
    toolbox_wy = status_wy - (toolbox_wh + wbw + wth);
    
    tree_ww = status_wx + status_ww - wbw;
    tree_wh = status_wh + 2 * source_browser->win_h;
    tree_wx = wbw;
    tree_wy = wbw;
}

static void var_sync(void){
    // cerr << "xshell_var_sync\n";

    // good to start with this here, to avoid errors.
    XWindowAttributes win_attr;
    if(!xshell_win_id_valid(&win_attr)) return;

    fl_get_win_geometry(xshell_win_id, &xshell_wx, &xshell_wy, &xshell_ww, &xshell_wh);

    if(fl_wmstuff.rep_method == FL_WM_SHIFT){
	// cerr << "WM SHIFT\n";
	// works well for geometry memorizing (mwm, 4Dwm, twm, tvtwm)
	xshell_wx -= fl_wmstuff.rpx + fl_wmstuff.bw;
	xshell_wy -= fl_wmstuff.rpy + fl_wmstuff.bw;
    } else {
	// nothing. ?
    }
    // cerr << "classic wx,wy="<< xshell_wx << "," << xshell_wy << endl;
    // new_xy();

    // if I can get the hints, good. otherwise, problems. If I have the incs,
    // good again, I don't need the rest. Otherwise, uhoh. I would need to guess
    // at the fontsize. I also need to guess (set!) the slider width.
    
    // needed to get size in font increments.
    XSizeHints *hints = XAllocSizeHints();
    long supplied;
    if(!XGetWMNormalHints(fl_display, xshell_win_id, hints, &supplied)){
	cerr << "xshell: couldn't get window size hints.\n";
    } else {
	hints->flags &= supplied;
	long flags = hints->flags;
	if((flags & PBaseSize) && (flags & PResizeInc)){
	    M_debug("xshell", "sufficient information for geometry calculation.");
            // here's the solution: use the actual width, not the original one !
	    xshell_geo_cols  = (xshell_ww - hints->base_width ) / hints->width_inc;
	    xshell_geo_rows  = (xshell_wh - hints->base_height) / hints->height_inc;
	    xshell_font_h    = hints->height_inc;
	    xshell_font_w    = hints->width_inc;
	    M_debug("xshell", "geometry: %dx%d", xshell_geo_cols, xshell_geo_rows);
	    M_debug("xshell", "font %dx%d", xshell_font_w, xshell_font_h);
	} else {
	    M_debug("xshell","insufficient information for geometry calculation.");
	    M_debug("flags:","PSize %d, PBaseSize %d, PResizeInc %d", 
		flags&PSize, flags&PBaseSize, flags&PResizeInc);
	}
	XFree((char *)hints);
    }
}

void xshell_save_defaults(ofstream &ofile){
    if(xshell_visible_flag){
	var_sync();
    }
    resource_save_defaults(ofile, xshell_prm_table, xshell_ID_string);
}

void xshell_sync_defaults(void){
    if(xshell_visible_flag){
	xshell_show();
    } else {
	xshell_close();
    }
}

/*
screen     shell   browser        status msg  help  menus  view toolbox
1280x1024  9x15    FL_NORMAL (12) FL_SMALL    brows brwser brws statusmsg
1024x768   8x13    FL_SMALL  (10) FL_SMALL    
 800x600   7x13    FL_SMALL  (10) FL_TINY     
 640x480   6x12    FL_TINY   (8)  FL_TINY     
 */

static void xshell_only_options_defaults(void){
#ifdef USE_XTERM
    strcpy(xshell_options_string, " -fg white -bg black -cr green -ms red ");
#else
    strcpy(xshell_options_string, " -fg white -bg black -selfg yellow -selbg grey -cursorbg green ");
#endif
    if(fl_scrw >= 1279){
	strcat(xshell_options_string, " -fn 9x15 ");
    } else if(fl_scrw >= 1023){
	strcat(xshell_options_string, " -fn 8x13 ");
    } else if(fl_scrw >= 799){
	strcat(xshell_options_string, " -fn 7x13 ");
    } else {
	strcat(xshell_options_string, " -fn 6x12 ");
    }
}

static void xshell_options_defaults(void){
    xshell_only_options_defaults();
    if(fl_scrw >= 1279){
#if defined(HP) || defined(SUNOS)
	source_browser->set_fontsize(FL_MEDIUM_SIZE);
	tool_fontsize = tool_menu_fontsize = help_fontsize   = FL_MEDIUM_SIZE;
	status_fontsize = FL_NORMAL_SIZE;
#else
	source_browser->set_fontsize(FL_NORMAL_SIZE);
	tool_fontsize = tool_menu_fontsize = help_fontsize   = FL_NORMAL_SIZE;
	status_fontsize = FL_SMALL_SIZE;
#endif
    } else if(fl_scrw >= 1023){
#if defined(HP) || defined(SUNOS)
	source_browser->set_fontsize(FL_NORMAL_SIZE);
	tool_fontsize= tool_menu_fontsize = help_fontsize   = FL_NORMAL_SIZE;
	status_fontsize = FL_NORMAL_SIZE;
#else
	source_browser->set_fontsize(FL_SMALL_SIZE);
	tool_fontsize= tool_menu_fontsize = help_fontsize   = FL_SMALL_SIZE;
	status_fontsize = FL_SMALL_SIZE;
#endif
    } else if(fl_scrw >= 799){
	source_browser->set_fontsize(FL_SMALL_SIZE);
	tool_fontsize = tool_menu_fontsize = help_fontsize   = FL_SMALL_SIZE;
	status_fontsize = FL_TINY_SIZE;
    } else {
	source_browser->set_fontsize(FL_TINY_SIZE);
	tool_fontsize = tool_menu_fontsize = help_fontsize   = FL_TINY_SIZE;
	status_fontsize = FL_TINY_SIZE;
    }
    target_browser->sync_display_opts(source_browser);
}

void xshell_init(void){
    // call externally, before showing or loading defaults.
    strcpy(xshell_options_string,"");
}

int xshell_load_default(
    const char *obj_tok, 
    const char *name_tok, 
    const char *val_tok
){
    return resource_load_default(
        xshell_prm_table, 
        xshell_ID_string, 
        obj_tok, name_tok, val_tok
    );
}
    
/* Window_With_Instance_Name:
 * routine to locate a window with a given instance and class name on a display.
 * If no window with the given name is found, 0 is returned.
 * If more than one window has the given name, the first
 * one found will be returned.  Only top and its subwindows
 * are looked at.  Normally, top should be the RootWindow.
 * level limits the search depth.
 */
Window Window_With_Instance_Name(
    Display *dpy,
    Window top,
    const char *class_name,
    const char *instance_name,
    int level				 
){
    Window *children, dummy;
    unsigned int i, nchildren;
    Window w=0;

    if(level<0){
	return 0;
    } else {
	level--;
    }

    XTextProperty text_prop;
    if(!XGetTextProperty(dpy, top, &text_prop, XA_WM_CLASS)){
	// failure. Some win's don't have class :-) !
	// if(dbg_flag) cerr << "UC: X: couldn't get text property.\n";
    } else {
	int count;
	char **list_return;
	if(XTextPropertyToStringList(&text_prop, &list_return, &count)){
	    int found_flag = FALSE;
	    if(dbg_flag){
		for(i=0; i<(unsigned int) count; i++){
		    cerr << "item no: " << i << ": " << list_return[i] << endl;
		}
	    }
	    if(count > 2){
		if(strcmp(list_return[0], instance_name) == 0 && 
		    strcmp(list_return[1], class_name) == 0){
		    found_flag = TRUE;
		    if(dbg_flag) cerr << "found ucshell !\n";
		}
	    }
	    XFreeStringList(list_return);
	    if(found_flag) return top;
	}
    }
    
    if (!XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren)){
	return 0;
    }
    for(i=0; i<nchildren; i++){
	w = Window_With_Instance_Name(dpy, children[i], class_name, instance_name, level);
	if(w)
	    break;
    }
    if(children){
	XFree((char *)children);
    }
    return w;
}

static int delayed_search_xshell(void){
    int found_flag = FALSE;
    int i=0;
    sginap(100);  // a preroll :-)
    do {
	if(dbg_flag) cerr << "searching for shell ...\n";
	sginap(50);
	found_flag = search_xshell();
	i++;
    } while (i<10 && !found_flag);
    return found_flag;
}

static int search_xshell(void){
    xshell_win_id = Window_With_Instance_Name(fl_display, RootWindow(fl_display, fl_screen), xshell_class_name, xshell_instance_name, 3);
    if (!xshell_win_id){
	if(dbg_flag) cerr << "No window with name " << xshell_win_title << " exists !\n";
	return FALSE;
    } else {
	if(dbg_flag) cerr << "xshell: Window (" << xshell_win_title << ") id: " << xshell_win_id << "\n";
	return TRUE;
    }
    // possibly return valid () ? Naw, overkill.
}

void xshell_sync_title(void){
    if(!source_is_stable()) return;
    if(!xshell_visible_flag) return;
    if(!xshell_win_id_valid()) return;
    
    const char *cwd_buf = source_browser->get_cwd();
    if(!cwd_buf) return;

    char buffer[MAXPATHLEN+64];
    sprintf(buffer, "%s%s%s", xshell_win_title, tool_host_string, cwd_buf);
    XStoreName(fl_display, xshell_win_id, buffer);
    XFlush(fl_display);
}

int xshell_needs_enter_flag = FALSE;

int xshell_sendevent_callback(XKeyEvent *xev, KeySym keysym){
    // always sync and check. 
    // cerr << "xshell sendevent cb.\n";
    M_info("xshell_send", "start");	   
    tool_got_event_flag = TRUE;
    if(!xshell_visible_flag) return FALSE;
    if(!xshell_win_id_valid()) return FALSE;

    // filter the modifier keys.
    if(IsModifierKey(keysym)){
	return FALSE;
    }
    
    if(keysym == XK_KP_Enter ||
	keysym == XK_Return ||
	((keysym == XK_c || keysym == XK_C) && tool_get_ctrl())
    ){
	source_browser->time_stamp();
	if(dbg_flag) cerr << "xshell setting needs_enter false.\n";
	xshell_needs_enter_flag = FALSE;
    } else {
	if(dbg_flag) cerr << "xshell setting needs_enter TRUE.\n";
	xshell_needs_enter_flag = TRUE;
    }
    
    // do it on foot, so that I have only one way of dealing with
    // key send events. 
 
    XKeyEvent xkev;
    memset(&xkev, 0, sizeof(XKeyEvent));   // just in case.

    xkev.type    = KeyPress;  // always
    xkev.display = fl_display;
    xkev.keycode = xev->keycode;
    xkev.state   = xev->state;
    xkev.x = xkev.y = xkev.x_root = xkev.y_root = 0;
    xkev.time = CurrentTime;
    xkev.same_screen = True;
    xkev.subwindow = None;
    xkev.root = RootWindow(fl_display, fl_screen);
    xkev.window = xshell_win_id;

    if (XSendEvent(fl_display,xshell_win_id,True,KeyPressMask,(XEvent *) &xkev) == 0){
	cerr << "XSendEvent failed.\n";
	return FALSE;
    } else {
	global_send_event_sent_count++;
    }
    if(!xshell_needs_enter_flag){
	// it just happened.
	source_browser->time_stamp_check();
    }
    if(dbg_flag) cerr << "event sent.\n";
    XFlush(fl_display);
    return TRUE;
}

#ifdef UNDER_DEVELOPMENT

main(){
    search_xshell();
    //XSynchronize(fl_display,True);
    while(1){
	char buf[16];
	cin >> buf;
	xshell_key(buf[0]);
    }
}

#endif

void xshell_help_func(void){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	HelpButtonCB(NULL, 0);
    } else {
	load_help_browser(shell_help_text, "Shell");
    }
}

static void do_close(int){
    xshell_close();    
}

static void do_colours_fonts(int){
    load_help_browser(shell_colours_fonts_help_text, "Shell Colours & Fonts");

}

static void xshell_menu_cb(void){
    static long windows_menu = get_windows_menu(), options_menu = 0;
    long
	dir_menu = get_dir_menu(),
	xshell_menu = new_pup()
	;
    if(xshell_menu <= 0){
	return;
    }
    if(options_menu <= 0){
	options_menu = new_pup();
	add_to_pup(options_menu, "Options%t", 0);
	add_to_pup(options_menu, "Colours & Fonts%f", (long) do_colours_fonts, "Cc");
    }
	
    
    add_to_pup(xshell_menu, "UCShell%t", 0);
    add_to_pup(xshell_menu, "Close %f", (long) do_close,    "Cc");
    if(dir_menu>0){
	add_to_pup(xshell_menu, "Directory %m", dir_menu, "Dd");
    }
    add_to_pup(xshell_menu, "Options %m",   options_menu,  "Oo");
    if(windows_menu>0){
	add_to_pup(xshell_menu, "Windows %m",   windows_menu,  "Ww");
    }
    do_pup(xshell_menu);
    free_pup(xshell_menu);
}

static void xshell_options_func(void){
    int tmp_flag = tool_ctrl_on_flag;
    if(tool_get_meta()){
	tool_ctrl_on_flag = tmp_flag;     // restore
	OptionsButtonCB(NULL, 0);
    } else {
	xshell_menu_cb();
    }
}

int xshell_check_event(XEvent *xev){
    XEvent event;
    if(!xshell_visible_flag) return FALSE;
    if(xshell_check_fifo()) return TRUE;

    if(dbg_flag) cerr << "xshell: looking for X event.\n";
    int found_flag = FALSE;
    int i=0;

    if(xev){
	if(xev->type == KeyPress){
	    if(xev->xkey.window == xshell_win_id){
		found_flag = TRUE;
	    } else {
		while(i<n_xshell_children && !found_flag){
		    if(xev->xkey.window == xshell_children[i]){
			found_flag = TRUE;
		    }
		    i++;
		}
	    }
	    if(found_flag){
		event = *xev;
	    }
	}
    }
    if(!found_flag){
	// check directly.
	found_flag = XCheckWindowEvent(fl_display,xshell_win_id,KeyPressMask,&event);
	i=0;
	while(i<n_xshell_children && !found_flag){
	    found_flag = XCheckWindowEvent(fl_display,xshell_children[i],KeyPressMask,&event);
	    i++;
	}
    }
    if(found_flag){
	KeySym keysym;
	char keybuf[2];
	if(dbg_flag) cerr << "found a desired KeyPressMask X event.\n";
	switch(event.type){
	case KeyPress:
	    if(dbg_flag) cerr << "found an X key press event.\n";
	    if(event.xkey.send_event){
		if(dbg_flag) cerr << "event is synthetic ...\n";
		global_send_event_received_count++;
		if(global_send_event_received_count > global_send_event_sent_count){
		    xshell_close();
		    load_help_browser(xshell_conflict_help_text, "UCShell event conflict");
		    timed_error_notifier("Sorry: X event conflict in UCShell.", (double) 0.0);
		    return FALSE;
		}
	    }
	
	    XLookupString((XKeyEvent *)&event, keybuf, sizeof(keybuf), &keysym, 0);
	    fl_keymask = event.xkey.state; // for alt/ctrl hacking.
	    switch(keysym){

	    case XK_w:
	    case XK_W:
		if (tool_get_alt() || fl_keysym_pressed(XK_Escape)) {
		    windows_menu_cb(); 
		}
		break;
	    case XK_F1:       xshell_help_func();    break;
	    case XK_F2:       xshell_options_func(); break;
	    case XK_F3:       MkDirButtonCB    (NULL, 0); break;
	    case XK_F4:       TreeButtonCB     (NULL, 0); break;
	    case XK_F5:       QuickEditButtonCB(NULL, 0); break;
	    case XK_F6:       EditButtonCB     (NULL, 0); break;
	    case XK_F7:       RenameButtonCB   (NULL, 0); break;
	    case XK_F8:       MoveButtonCB     (NULL, 0); break;
	    case XK_F9:       DuplicateButtonCB(NULL, 0); break;
	    case XK_F10:      CopyButtonCB     (NULL, 0); break;
	    case SunXK_F36:   if(!sun_keyboard_flag) return FALSE;
#ifdef DEC_ALPHA
	    case XK_F13:
#endif
	    case XK_F11:
		DeleteButtonCB   (NULL, 0); break;
	    case SunXK_F37: if(!sun_keyboard_flag) return FALSE;
	    case XK_F12:      ExitButtonCB     (NULL, 0); break;
		/*
		  case XK_Linefeed: 
		  case XK_Return:
		  case XK_KP_Enter: xshell_check_fifo();        break;
		  */
	    default:
		if(!event.xkey.send_event){
		    xshell_needs_enter_flag = TRUE;
		}
		break;
	    }
	    return TRUE;
	    /* 
	       case ButtonRelease:
	       if(dbg_flag) cerr << "found an X BUTTON release event.\n";
	       if(event.xbutton.button == Button2){
	       if(dbg_flag) cerr << "possible selection transfer.\n";
	       xshell_check_fifo(); 
	       }
	       break;
	       */
	default:
	    if(dbg_flag) cerr << "not a key press.\n";
	    return TRUE;
	}
    } else {
	// if(dbg_flag) cerr << "found NO X key press event.\n";
    }
    return FALSE;
}

/* format: shell name, command header, command trailer */


static char fifo_name[MAXPATHLEN+1];
static int fifo_fd = -1, persistent_fd = -1;
static int fifo_made_flag = FALSE;

static int xshell_make_fifo(void){
    char *tmp_nam = tempnam(NULL, "ucp");
    if(!tmp_nam){
	cerr << "couldn't generate temporary name.\n";
	return 0;
    }
    strcpy(fifo_name, tmp_nam);
    if(mkfifo(fifo_name, S_IRUSR | S_IWUSR) < 0){
	cerr << "mkfifo failed for " << fifo_name << ".\n";
	return 0;
    } else {
	if(dbg_flag) cerr << "mkfifo succeeded for " << fifo_name << ".\n";
	fifo_made_flag = TRUE;
    }

    // open for reading and writing -> persistent.
    // or use O_RDWR.
    if((fifo_fd = open(fifo_name, O_NDELAY | O_RDONLY | O_CREAT)) == -1){
	cerr << "couldn't open input fifo: " << fifo_name << endl;
	return 0;
    }
    if((persistent_fd = open(fifo_name, O_WRONLY)) == -1){
	cerr << "couldn't open output fifo: " << fifo_name << endl;
	return 0;
    }

    if(dbg_flag) cerr << "opened input/output fifo: " << fifo_name << endl;
    return TRUE;
}

static void xshell_close_fifo(void){
    if(fifo_fd != -1){
	close(fifo_fd);
	fifo_fd = -1;
    }
    if(persistent_fd != -1){
	close(persistent_fd);
	persistent_fd = -1;
    } 
    if(fifo_made_flag){
	if(unlink(fifo_name)<0){
	    if(errno != ENOENT){
		cerr << "couldn't remove fifo: " << fifo_name << ", " << strerror(errno) << endl;
	    }
	    return;
	} else {
	    if(dbg_flag) cerr << "removed fifo: " << fifo_name << endl;
	}
    }
}

static void bash_init(void){
    if(!xshell_make_fifo()) return;
    /* key function.
     * cd() { builtin cd "$@" ; pwd >> fifo_name ; }
     */
    xshell_string("cd() { builtin cd ");
    xshell_key('"', FALSE, FALSE, TRUE);
    xshell_key('$', FALSE, FALSE, TRUE);
    xshell_key('@', FALSE, FALSE, TRUE);
    xshell_key('"', FALSE, FALSE, TRUE);
    xshell_string(" ; pwd >> ");
    xshell_string(fifo_name);
    xshell_string("; }", TRUE);
    xshell_string("clear", TRUE);
}

/* don't use echo -n, the newlines are great separators if multiple writes
 * have been made to the fifo.
 */

static void ctcsh_init(void){
    if(!xshell_make_fifo()) return;
    // key alias (tcsh:cwdcmd does not support output redirection)
//    xshell_string("alias cd ' cd \\!*; echo $cwd >>");  // note the \\ !
    xshell_string("alias cd ' cd \\!*; echo $cwd >>");  // note the \\ !
    xshell_string(fifo_name);
    xshell_string("'", TRUE);
    // auxiliary aliases.
    xshell_string("alias popd ' popd \\!*; echo $cwd >>");  
    xshell_string(fifo_name);
    xshell_string("'", TRUE);
    xshell_string("alias pushd ' pushd \\!*; echo $cwd >>");  
    xshell_string(fifo_name);
    xshell_string("'", TRUE);
    xshell_string("clear", TRUE);
}

static void ksh_init(void){
    if(!xshell_make_fifo()) return;
    /* key function. (we check if the no. of args is > 0)
     * alias cd=_cd
     * function _cd
     * {
     * \cd $1 >/dev/null; echo $PWD >> name
     * }
     * use echo, not print. ksh on BSD doesn't have print!
     */
    xshell_string("alias cd=_cd", TRUE);
    xshell_string("function _cd", TRUE);
    xshell_string("{", TRUE);
    xshell_string("\\cd $1 > /dev/null; echo $PWD >> ");
    xshell_string(fifo_name, TRUE);
    xshell_string("}", TRUE);
    xshell_string("clear", TRUE);
}

static void sh_init(void){
    if(!xshell_make_fifo()) return;
    /* key function.
     * ucd() { cd "$@" ; pwd >> fifo_name ; }
     */
    xshell_string("ucd() { cd ");
    xshell_key('"', FALSE, FALSE, TRUE);  // did not manage the escaping correctly.
    xshell_key('$', FALSE, FALSE, TRUE);
    xshell_key('@', FALSE, FALSE, TRUE);
    xshell_key('"', FALSE, FALSE, TRUE);
    xshell_string(" ; pwd >> ");

    xshell_string(fifo_name);
    xshell_string("; }", TRUE);
    xshell_string("clear", TRUE);
    xshell_string("echo Note: Use \\'ucd\\' instead of \\'cd\\' if you want UC to chdir with sh.", TRUE);
    XStoreName(fl_display, xshell_win_id, "Note: Please use 'ucd' instead of 'cd' if you want UC to chdir with sh.");
    XFlush(fl_display);
}

static void rc_init(void){
    if(!xshell_make_fifo()) return;
    /* key function.
     * chpwd() { pwd >> fifo_name ; }
     */
    xshell_string("fn cd { builtin cd $* && pwd >> ");
    xshell_string(fifo_name);
    xshell_string(" }", TRUE);
    xshell_string("clear", TRUE);
}

static void zsh_init(void){
    if(!xshell_make_fifo()) return;
    /* key function.
     * chpwd() { pwd >> fifo_name ; }
     */
    xshell_string("chpwd() { pwd >> ");
    xshell_string(fifo_name);
    xshell_string("; }", TRUE);
    xshell_string("clear", TRUE);
}

static void xshell_init_fifo(void){
    if(fifo_made_flag){
	xshell_close_fifo();
    }
    const char *shell_iv = getenv("SHELL");
    char shell[MAXPATHLEN+1];
    
    if(shell_iv){
	strcpy(shell, get_basename(shell_iv));
    } else {
	uid_t uid = getuid();
	struct passwd *pw_ent = getpwuid(uid);
	if(!pw_ent){
	    cerr << "UC:Couldn't get SHELL or login shell.\n";
	    return;
	}
	shell_iv = get_basename(pw_ent->pw_shell);
	if(shell_iv){
	    strcpy(shell, shell_iv);
	} else {
	    cerr << "UC:Couldn't get SHELL or login shell.\n";
	    return;
	} 
    }
    if(strcmp(shell, "bash") == 0){	
	status_msg(2, "initialising bash.");
	bash_init();
    } else if(strcmp(shell, "csh") == 0){
	status_msg(2, "initialising C-shell.");
	ctcsh_init();
    } else if(strcmp(shell, "ksh") == 0){
	status_msg(2, "initialising Korn shell.");
	ksh_init();
    } else if(strcmp(shell, "pdksh") == 0){
	status_msg(2, "initialising PD Korn shell.");
	ksh_init();
    } else if(strcmp(shell, "rc") == 0){	
	status_msg(2, "initialising rc shell.");
	rc_init();
    } else if(strcmp(shell, "sh") == 0){	
	status_msg(2, "initialising bourne shell.");
	sh_init();
    } else if(strcmp(shell, "tcsh") == 0){
	status_msg(2, "initialising TC shell.");
	ctcsh_init();
    } else if(strcmp(shell, "zsh") == 0){
	status_msg(2, "initialising Z shell.");
	zsh_init();
    } else {
	char buffer[MAXPATHLEN*2];
	sprintf(buffer, "UC: Unknown or unsupported shell: `%s'", shell);
	status_msg(2, "unknown shell (%s). ", shell);
	timed_notifier(buffer, "Supported are: bash, csh, ksh, pdksh, rc, sh, tcsh & zsh.");
    }
}

static int xshell_check_fifo(void){
    if(!fifo_made_flag) return FALSE;
    //return FALSE;
    const int bufsize = 10 * MAXPATHLEN+1;
    char buffer[bufsize+1];
    int n_read = read(fifo_fd, buffer, bufsize);
    if(n_read > 0){
	if(dbg_flag) cerr << "read nbytes: " << n_read << endl;
	buffer[bufsize] = '\0';
	if(n_read < bufsize){
	    buffer[n_read] = '\0';
	}
	if(n_read >= 1){
	    // strip trailing newline.
	    if(buffer[n_read-1] == '\n'){
		buffer[n_read-1] = '\0';
	    }
	}
	if(n_read > 0){
	    if(dbg_flag) cerr << "buffer is: " << buffer << endl;
	}
	char *last_newline = strrchr(buffer, '\n');
	if(last_newline){
	    if(dbg_flag) cerr << "buffer contains newlines.\n";
	    last_newline++;
	} else {
	    last_newline = buffer;
	}

	char path_buf[MAXPATHLEN+1];
	char *dir = real_path(last_newline, path_buf);
	if(dir){
	    if(access(dir, X_OK) == 0){
		struct stat statbuf;
		if(stat(dir, &statbuf) < 0){
		    cerr << "shell error: cannot stat file `" << dir << "' !\n";
		    return FALSE;
		}
		if(S_ISDIR(statbuf.st_mode) || S_ISLNK(statbuf.st_mode)){
		    tool_change_dir(dir, FALSE, TRUE, TRUE, FALSE, FALSE);
		} else {
		    cerr << "UC: shell directory path is not a directory: " << dir << endl;
		}
	    } else {
		cerr << "UC: shell directory path (" << dir << ") inaccessible (" 
		     << strerror(errno) << endl;
	    }
	} else {
	    cerr << "UC: invalid directory path from shell (NULL) !\n";
	    cerr << "offending path component: " << path_buf << endl;
	}
	return TRUE;
    } else {
	if(dbg_flag) cerr << "xshell didn't change dir.\n";
	return FALSE;
    }
}

/* window moving
 * need this 2nd method of x,y for fvwm (incompatible to classic var_sync)
 * cannot move wm frame directly.
 */
static FL_Coord fudge_dx = 0, fudge_dy = 0;

static void xshell_compute_fudges(FL_Coord target_dx, FL_Coord target_dy){
    // do your own var_sync. 
    var_sync();
    if(fl_wmstuff.rep_method == FL_WM_SHIFT){
	// make a good first guess.
	fudge_dx = fl_wmstuff.rpx + fl_wmstuff.bw;
	fudge_dy = fl_wmstuff.rpy + fl_wmstuff.bw;   // mwm good. twm doesn't like rpy.
    }
    FL_Coord old_wx = xshell_wx, old_wy = xshell_wy, raw_dx = 5, raw_dy = 5, fdx, fdy;
    XMoveWindow(fl_display, xshell_win_id, old_wx+raw_dx+fudge_dx, old_wy+raw_dy+fudge_dy);
    // XSync(fl_display, False);

    // wait for done and resync. Important to wait for a long time here!
    wait_configure(xshell_win_id, 250); 
    var_sync();

    // calc difference
    fdx = old_wx + raw_dx - xshell_wx;
    fdy = old_wy + raw_dy - xshell_wy;
    // correct factors.
    fudge_dx += fdx;
    fudge_dy += fdy;

    M_warn("xshell auto-fudge: "," fdx %d, fdy %d", fudge_dx, fudge_dy);

    // translate anew with target correction on top.
    XMoveWindow(fl_display, xshell_win_id, old_wx+target_dx+fudge_dx, old_wy+target_dy+fudge_dy);
}

void xshell_move_relative(int dx, int dy){
    if (!xshell_win_id_valid()) return;
    static int first = TRUE;
    if(first){
	first = FALSE;
	xshell_compute_fudges(dx, dy);
	return;
    }
    var_sync();
    M_warn("XShell move relative", "wx=%d, wy=%d, dx=%d, dy=%d, rpx=%d, rpy=%d, bw=%d, fdx=%d, fdy=%d", 
	xshell_wx, xshell_wy, dx, dy, fl_wmstuff.rpx, fl_wmstuff.rpy, fl_wmstuff.bw, fudge_dx, fudge_dy);
    
    dx += fudge_dx;
    dy += fudge_dy;
    XMoveWindow(fl_display, xshell_win_id, xshell_wx + dx, xshell_wy + dy);
    //XSync(fl_display, False);
    wait_configure(xshell_win_id);
}

void xshell_close_form(void){
    if(XShellForm->visible){
	focus_restore_start();
	fl_hide_form(XShellForm);
	focus_restore_finish();
    }
}

void xshell_options_context_func(void){
    load_help_browser(xshell_options_help_text, "XShell Options");
}

static int xshell_options_raw_cb(FL_FORM *form, void *xev){
    return standard_form_event_cb(form, (XEvent *) xev, xshell_close_form, xshell_options_context_func);
}

void xshell_show_form(void){
    static int first = TRUE;
    if(first){
	first = FALSE;
	fl_set_input_color(XShellOptionsInput, FL_WHITE, FL_RED);
	fl_set_input_return(XShellOptionsInput, TRUE);
	fl_set_counter_bounds   (XShellXFudgeCounter, -100.0, 100.0);
	fl_set_counter_bounds   (XShellYFudgeCounter, -100.0, 100.0);
	fl_set_counter_step     (XShellXFudgeCounter, 1.0, 1.0);
	fl_set_counter_step     (XShellYFudgeCounter, 1.0, 1.0);
	fl_set_counter_precision(XShellXFudgeCounter, 0);
	fl_set_counter_precision(XShellYFudgeCounter, 0);
	fl_register_call_back(XShellForm, FL_ALL_EVENT, xshell_options_raw_cb);
    }
    fl_set_input(XShellOptionsInput, xshell_options_string);
    fl_set_counter_value(XShellXFudgeCounter, (double) xshell_x_fudge);
    fl_set_counter_value(XShellYFudgeCounter, (double) xshell_y_fudge);

    focus_save();
    if(!XShellForm->visible){
	fl_show_form(XShellForm, FL_PLACE_CENTER, TRUE, "UCShell Options:");
    } else {
	XMapRaised(fl_display, XShellForm->window);
    }
    tool_focus_move(XShellForm, XShellCloseReturnButton);
}

void XShellCloseReturnButtonCB(FL_OBJECT *, long){
    xshell_close_form();
}

void XShellDefaultsButtonCB(FL_OBJECT *, long){
    xshell_only_options_defaults();
    fl_set_input(XShellOptionsInput, xshell_options_string);
}

void XShellHelpButtonCB(FL_OBJECT *, long){
    xshell_options_context_func();
}

void XShellApplyButtonCB(FL_OBJECT *, long){
    if(xshell_visible_flag){
	//if(!ask_question(TRUE, "This will close and reopen your shell window. Continue ?"))return;
	xshell_close();
	// sginap(10);
	//tool_mung_events();
    }
    // XFlush(fl_display);
    fl_set_global_cursor(XC_watch);
    xshell_visible_flag = TRUE;
    xshell_show();
    options_sync_defaults();
    fl_reset_global_cursor();
}

void XShellOptionsInputCB(FL_OBJECT *, long){
    const char *result = fl_get_input(XShellOptionsInput);
    if(result){
	strcpy(xshell_options_string, result);
    }
}

void XShellXFudgeCounterCB(FL_OBJECT *, long){
    xshell_x_fudge = (int) fl_get_counter_value(XShellXFudgeCounter);
}

void XShellYFudgeCounterCB(FL_OBJECT *, long){
    xshell_y_fudge = (int) fl_get_counter_value(XShellYFudgeCounter);
}
